#!/usr/bin/perl 

use strict;
use warnings;

use DBI;
use Getopt::Long;

use F5::DbUtils;
use F5::ImportExportPolicy;
use F5::ASMReady;


# get command line and validate args
my ($class, $policy, $help);
GetOptions (    "class|c=s" => \$class,
                "policy|p=s" => \$policy,
                "help|h" => \$help
    ) or die( "Cannot parse command-line options" );
if ( $help ) {
    print "Usage: $0 --class=httpclass_name --policy=/path/to/policy.tgz";
    exit 1;
}
die "--class and --policy arguments are required" if (not defined $class) or (not defined $policy);
die "policy ($policy) not found" if not -f $policy;

# wait for ASM to be ready
my $timeout = 300;
my $iteration_sleep_time = 5;
my $waited = 0;
my $asm_ready;

while ((! $asm_ready) && ($waited < $timeout)) {

        $asm_ready = F5::ASMReady::is_asm_ready();
        $waited+= sleep($iteration_sleep_time);
}
if (not $asm_ready) {
    die "gave up waiting for ASM to be ready after $timeout seconds...\n";
}

# get the account for the class name given
my $dbh = F5::DbUtils::get_dbh();

# wait a bit (300 secs) for the class to be created
my $account_id;
$waited = 0;
while ((! $account_id) && ($waited < $timeout)) {
	($account_id) = $dbh->selectrow_array(
                'SELECT account_id FROM DCC.ACCOUNTS WHERE policy_name = ?',
                undef,
                "/Common/$class" );
        $waited+= sleep($iteration_sleep_time);
}

if (not $account_id) {
    die "No account found for class ($class) after $timeout seconds...\n";
}

# import the policy
my $import_export = F5::ImportExportPolicy->new(dbh => $dbh);
my $result = $import_export->import_policy (
    filename => $policy,
);

die "Failed to import the policy: $result->{error_msg}" if !$result->{completed_successfully};

my $policy_name = $result->{policy_name};
my $policy_id   = $result->{policy_id};


# set the policy active
$result = F5::SetActive::set_active(dbh => $dbh, 
                                    expanded_return_value => 1,
                                    do_update_core => 1, 
                                    policy_id => $policy_id,
                                    account_id => $account_id,);

if(!$result->{completed_successfully}) {
    die "set active failed: $result->{error_message}";
}
print "Successfully activated policy '$policy_name' (id=$policy_id) for httpclass '$class' (id=$account_id)";
$dbh->disconnect;
exit 0;
