#!/usr/bin/perl
use strict;
use warnings;

BEGIN
{
    if(defined $ENV{F5_INSTALL_MODE} && $ENV{F5_INSTALL_MODE} eq 'install') {
        exit 0;
    }
}

use Getopt::Long;
use F5::Logger qw(:LEGACY_ERROR_CONSTANTS);
use F5::Provisioning;
use F5::ASMConfig::EasyClient;

use constant DEFAULT_DEBUG_LEVEL  => 1;
use constant DEFAULT_LOG_FILE     => '/ts/log/bigdb_var.log';
use constant DEFAULT_PRINT_TARGET => 1;


my $CMDLINE = join ' ', ($0,@ARGV);

# init default logger before anything else
my $DEFAULT_LOGGER = F5::Logger->new(DEFAULT_LOG_FILE, DEFAULT_PRINT_TARGET);

my %args = handle_commandline_arguments();
# after processing args we can init the custom logger
my $LOGGER = F5::Logger->new($args{log_file}, $args{print_target});

log_print($DBG, undef, "Executed: $CMDLINE");

if (!F5::Provisioning::is_asm_on()) {
    log_print($INF, undef, "ASM is not provisioned. Exiting");
    exit 0;
}

eval {
    my $asm_config_client = F5::ASMConfig::EasyClient->new(username => 'bigdb_var');
    if ($asm_config_client) {
        $asm_config_client->update_mcp_db($asm_config_client->{session},
                                          F5::ASMConfig::CallFlags->new,
                                          F5::ASMConfig::McpDb->new({mcp_db_name => $args{var_name}, mcp_db_value => $args{var_value}}),
                                          F5::ASMConfig::SetActiveFlags->new({do_update_core => $args{flg_update_core}}));
    }
};
if ($@) {
    log_print($INF, undef, "Couldn't connect to ASM Config ($@)");
}

exit 0;

sub log_print {
    my ($level, $action, $msg) = @_;
    return if($level eq $DBG && (not $args{debug}));
    $LOGGER->log_print($level, $action, $msg);
}

sub handle_commandline_arguments {
	my %arguments;
	my $rv = GetOptions (
		"debug=i"       => \$arguments{debug},
		"log_file=s"    => \$arguments{log_file},
		"print_target"  => \$arguments{print_target},
		"var_name=s" => \$arguments{var_name},
		"var_value=s" => \$arguments{var_value},
		"flg_update_core=i" => \$arguments{flg_update_core},
		"help"   => \$arguments{help}
	);
	unless ($rv) {
		$DEFAULT_LOGGER->log_print($ERR, undef, 
			usage() . "Problems processing commandline arguments\n",
			);
		exit 1;
	}

	if ($arguments{help}) {
		print usage();
		exit 0;
	}

	unless (defined $arguments{var_name}) {
		$DEFAULT_LOGGER->log_print($ERR, undef, 
			usage() . "Missing variable name\n",
			);
		exit 1;
	}
	$arguments{var_value} = '' unless defined $arguments{var_value};
	$arguments{flg_update_core} = 1 unless defined $arguments{flg_update_core};

	$arguments{debug} = DEFAULT_DEBUG_LEVEL unless defined $arguments{debug};
	$arguments{log_file} = DEFAULT_LOG_FILE unless $arguments{log_file};
	$arguments{print_target} = DEFAULT_PRINT_TARGET unless defined $arguments{print_target};

	return %arguments;
}

sub usage {
	return qq{
USAGE:
	$0 <mandatory arguments>

Mandatory arguments:

	--var_name='bigdb variable name'
		bigdb variable name

	--var_value='value of the bigdb variable'
		value of the bigdb variable

Optional arguments:

	--debug='debug level'
		0 for no debug information logging
		1 for logging all information
	--log_file='log filename'
		fullpath/filename of log file to use for logging
	--print_target='print target'
		print target used for logging

Optionally, using --help will issue this help message.
	\n};
}
