#!/usr/bin/perl

use strict;
use Storable qw(fd_retrieve store_fd);
use Getopt::Long;
use XML::Twig;

#$Header$
my $input_filename;
my $output_filename;

my $option_objects_lower_case = 0;
my $option_objects_to_http = 0;
my $option_objects_to_https = 0;
my $option_modify_objects = 0;
my $option_filetypes_lower_case = 0;
my $search_for = '';
my $replace_with = '';
my $verbose = 0;
my $template_system_id = 0;
my $help = 0;
my $policy_description = '';

GetOptions("objects_lower_case"   => \$option_objects_lower_case,
           "objects_to_http"       => \$option_objects_to_http,
           "objects_to_https"       => \$option_objects_to_https,
           "modify_objects"       => \$option_modify_objects,
           "filetypes_lower_case" => \$option_filetypes_lower_case,
           "v|verbose"            => \$verbose,
           "i|input_file=s"       => \$input_filename,
           "o|output_file=s"      => \$output_filename,
           "search_for=s"         => \$search_for,
           "replace_with=s"       => \$replace_with,
           "inject_template_system_id=i" => \$template_system_id,
           "h|help"               => \$help,
	   "policy_description=s" => \$policy_description,
	   );

if($help)
{
   print usage();
   exit 0;
}

if((! defined $input_filename) && (! defined $output_filename))
{
   print usage();
   exit 1;
}

if($input_filename eq $output_filename)
{
   print usage();
   exit 1;
}

if($option_objects_lower_case + $option_objects_to_http + $option_objects_to_https + $option_modify_objects > 1)
{
   print usage();
   exit 1;
}


open(my $filetype_check, '-|', '/usr/bin/file', $input_filename) or die "cannot determine policy file type";
my $filetype_check_output = <$filetype_check>;
close $filetype_check;

if ($filetype_check_output =~ /gzip/) {
    gzip_format();
} 
elsif ($filetype_check_output =~ /XML/) {
    xml_format();
}
else {
    print "the input file doesn't seem to be a policy file.";
}

#--------------------------------------------------------------------------------------

sub xml_format {
    if ($option_objects_lower_case || $option_modify_objects || $option_filetypes_lower_case || $verbose || $search_for || $replace_with  ) {
        print usage();
        exit 1;
    }

    my $twig = XML::Twig->new(pretty_print => 'indented');
    $twig->parsefile($input_filename);
    my $root= $twig->root;

    my $description_node = $root->first_child('description');
    $description_node->set_text($policy_description);

    my $urls = $root->first_child('urls');
    my @url_list = $urls->children('url');
    my $protocol = $option_objects_to_https ? 'HTTPS' : 'HTTP';

    my %url_map;
    $url_map{HTTP} = [];
    $url_map{HTTPS} = [];

    foreach my $url (@url_list) {
        push (@{$url_map{$url->{'att'}->{'protocol'}}}, $url->{'att'}->{'name'});
    }

    foreach my $url (@url_list) {
        if (! scalar grep {$_ eq $url->{'att'}->{'name'} } @{$url_map{$protocol}}  ) {
            $url->set_att( protocol => $protocol);
        }
    }

    open (my $out, ">", $output_filename) or die "cannot open $output_filename";
    $twig->print(\*$out);
    close $out;
}

#--------------------------------------------------------------------------------------

sub gzip_format {
  open(my $fh, "/bin/gzip -q -d -c '$input_filename' |");
  open(my $outf, "| /usr/bin/gzip > '$output_filename'");
  
  my $import_file_version;
  my $objects_cache = {};
  my $filetypes_cache = {};
  my @data_set = ();
  
  # preprocessing phase
  while(!eof($fh)) {
    my $import_data = fd_retrieve($fh);
    if(! $import_file_version)
    {
       $import_file_version = defined($import_data->{_export_file_version}) ?
                              $import_data->{_export_file_version} :
                              1;
    }
  
    if (exists $import_data->{PL_OBJECTS}) {
  
      # build the cache of objects
      for my $val (values %{$import_data->{PL_OBJECTS}})
      {
         $objects_cache->{$val->{app_id}}->{$val->{uri}} = 1;
      }
    }
    if (exists $import_data->{PL_FILETYPES}) {
  
      # build the cache of objects
      for my $val (values %{$import_data->{PL_FILETYPES}})
      {
         $filetypes_cache->{$val->{name}} = 1;
      }
    }
    push @data_set, $import_data;
  }
  
  close($fh);
  
  my $policy_id;
  my $is_flg_policy_templates_settings_exists = 0;
  
  # edit phase
  for my $import_data (@data_set)
  {
    # objects
    if (exists $import_data->{PL_OBJECTS}) {
  
      if($option_objects_lower_case)
      {
         for my $val (values %{$import_data->{PL_OBJECTS}}) {
           # change to lower case if the lower case does not exist in the policy
           my $new_uri = lc($val->{uri});
           if (! defined $objects_cache->{$val->{app_id}}->{$new_uri})
           {
              print "object $val->{uri} changed to $new_uri (".($val->{app_id} ? "https" : "http").")\n" if $verbose;
              $objects_cache->{$val->{app_id}}->{$new_uri} = 1;
              $val->{uri} = $new_uri;
           }
         }
      }
  
      if($option_objects_to_http || $option_objects_to_https)
      {
         my $target_app_id = ($option_objects_to_http) ? 0 : 1;
         my $target_app_name = ($option_objects_to_http) ? "http" : "https";
         for my $val (values %{$import_data->{PL_OBJECTS}}) {
           # change to https if the https object does not exist in the policy
           if (! defined $objects_cache->{$target_app_id}->{$val->{uri}})
           {
              print "object $val->{uri} $target_app_name\n" if $verbose;
              $val->{app_id} = $target_app_id;
           }
    
         }
      }
  
      if($option_modify_objects)
      {
         for my $val (values %{$import_data->{PL_OBJECTS}}) {
           my $new_uri = $val->{uri};
           $new_uri =~ s/$search_for/$replace_with/g;
  
           # change to https if the https object does not exist in the policy
           if (! defined $objects_cache->{$val->{app_id}}->{$new_uri})
           {
              print "object $val->{uri} changed to $new_uri (".($val->{app_id} ? "https" : "http").")\n" if $verbose;
              $objects_cache->{$val->{app_id}}->{$new_uri} = 1;
              $val->{uri} = $new_uri;
           }
         }
      }
    }
  
    # object types
    if (exists $import_data->{PL_FILETYPES}) {
  
      if($option_filetypes_lower_case)
      {
         for my $val (values %{$import_data->{PL_FILETYPES}}) {
           # change to lower case if the lower case does not exist in the policy
           my $new_extension = lc($val->{name});
           if (! defined $filetypes_cache->{$new_extension})
           {
              print "object type $val->{name} changed to $new_extension\n" if $verbose;
              $val->{name} = $new_extension;
              $filetypes_cache->{$new_extension} = 1;
           }
         }
      }
    }
  
    if (exists $import_data->{PL_POLICIES}) {
      $policy_id = $import_data->{PL_POLICIES}->{id};
    }
  
    if (exists $import_data->{PL_POLICY_TEMPLATES_SETTINGS}) {
      for my $key (keys %{$import_data->{PL_POLICY_TEMPLATES_SETTINGS}}) {
          $import_data->{PL_POLICY_TEMPLATES_SETTINGS}->{$key}->{policy_template_system_id} = $template_system_id;
      }
      $is_flg_policy_templates_settings_exists = 1;
    }
  }
  
  if (! $is_flg_policy_templates_settings_exists) {
    push (@data_set, {PL_POLICY_TEMPLATES_SETTINGS => { 
      1 => {
          policy_template_system_id => $template_system_id,
          policy_id => $policy_id,
      }
    }});
  }
  
  # save phase
  for my $import_data (@data_set)
  {
    store_fd($import_data, $outf);
  }
  
  close($outf);
}

#--------------------------------------------------------------------------------------

sub usage
{
   return "usage: $0 -i|--input_file <input_file> -o|--output_file <output_file> <edit_options>
edit_options are: --objects_to_http, --objects_to_https, --modify_objects, --filetypes_lower_case, --objects_lower_case, --inject_template_system_id
when using modify_objects, the following should be specified:
--search_for
--replace_with 
NOTES: 
   only use one of the object editing options at a time
   using the same input and output files is not supported

   when the input_file is an XML policy, you can specify the following options only:
      input_file,
      output_file,
      objects_to_http
      objects_to_https
"; 
}

