#!/usr/bin/perl

use strict;
use warnings;

use Text::CSV_XS;
use IO::File;
use Getopt::Long;
use YAML::XS;
use F5::ASMConfig::Utils;
use Scalar::Util qw(looks_like_number);

my ($base_dir);
my @databases;
GetOptions('base_dir|d=s' => \$base_dir,
           "db=s" => \@databases,);


my $schema = F5::ASMConfig::Utils::get_entity_schema(base_dir => $base_dir,
                                                     extra_paths => ["factory_data/*schema.yaml",
                                                                     "$ENV{DEVFS}/var/ts/install/db/factory_data/*schema.yaml"],
             );


for my $db (@databases) {
  # Determine load order of tables in new format
  my @new_tables = map { /$db\.(.*)\.sql/; $1 } glob("table_schema/$db.*.sql");
  my %new_tables = map { ($new_tables[$_], $_) } 0..$#new_tables;

  my $current_table_index = 0;
  while ($current_table_index <= $#new_tables) {
    my $table = $new_tables[$current_table_index];
    my $table_schema = getfile("table_schema/$db.$table.sql");

    my $max_parent_index = 0;
    while($table_schema =~ /foreign key\s*.*?\s+references\s+\`?(.*?)\`?\s*\(/ig) {
      my $foreign_table = $1;
      $foreign_table =~ s/^.*\.//;
      if ($new_tables{$foreign_table} && 
	  $new_tables{$foreign_table} > $current_table_index && 
	  $new_tables{$foreign_table} > $max_parent_index) {
	$max_parent_index = $new_tables{$foreign_table};
      }
    }

    # Move this table after max_parent position
    if ($max_parent_index) {
      splice(@new_tables, $current_table_index, 1);
      splice(@new_tables, $max_parent_index, 0, $table);
      %new_tables = map { ($new_tables[$_], $_) } 0..$#new_tables;
    }
    else {
      $current_table_index++;
    }
  }

  my @source_files = ();
  push @source_files, (lc($db) . '_old_templ.sql') if(-f lc($db) . '_old_templ.sql');
  for my $new_table (@new_tables) {
    push(@source_files, grep { -e $_ } ("table_schema/$db.$new_table.sql", 
					"factory_data/$db.$new_table.sql",
                                        "$ENV{DEVFS}/var/ts/install/db/factory_data/$db.$new_table.csv",
                                        "$ENV{DEVFS}/var/ts/install/db/factory_data/$db.$new_table.yaml",
                                        "$ENV{DEVFS}/var/ts/install/db/factory_data/$db.$new_table.schema.yaml",
                                        "factory_data/$db.$new_table.csv",
                                        "factory_data/$db.$new_table.yaml",
                                        "factory_data/$db.$new_table.schema.yaml",
                                       ));
  }
  
  my %table_settings;
  open(my $outf, '>', lc($db) . '.sql');
  for my $source_file (@source_files) {
    my $source_fh = IO::File->new($source_file);

    if ($source_file =~ /^.+?\.([^\.]*)(\.schema)?\.(csv|yaml)$/) {
        my $table_name = $1;
        my $schema_file = $2;
        my $source_type = $3;
        local $" = ',';
        my $csv;
        my $table_def;
        my $field_names;

        my @field_value_insert_list;
        if ($source_type eq 'csv') {
            $csv = Text::CSV_XS->new({ binary => 1, eol => $/ ,  blank_is_undef => 1  });

            $field_names = $csv->getline($source_fh);
        }
        elsif ($source_type eq 'yaml') {
            $table_def = $schema_file ? $schema->{$table_name}
                                      : YAML::XS::LoadFile($source_file);

            my %field_names;
            for my $row (@{$table_def->{table_data_rows}}) {
                for my $k (keys %$row) {
                    $field_names{$k} = 1;
                }
            }
            $field_names = [ keys %field_names ];

            if ($table_def->{generate_rest_uuid}) {
                require F5::Utils::Rest;
                push(@$field_names, 'rest_uuid');
            }
        }

        while (1) {
            my $field_values;
            if ($source_type eq 'csv') {
                $field_values = $csv->getline($source_fh);
                last unless $field_values;
            }
            elsif ($source_type eq 'yaml') {
                last unless @{$table_def->{table_data_rows}};
                my $row = shift(@{$table_def->{table_data_rows}});

                if ($table_def->{generate_rest_uuid}) {
                    $row->{rest_uuid} = F5::Utils::Rest::generate_uuid($row->{$table_def->{generate_rest_uuid}->{field_name}});
                }

                $field_values = [ @{$row}{@$field_names} ];
            }

            next if (defined $field_values->[0] && $field_values->[0] =~ /^##/);
            # The first line must always be the header. Any other line can include remark with a ## at the beginning of the line
            for my $field_index (0..$#$field_values) {
                my $field_name = $field_names->[$field_index];
                if (! defined($field_values->[$field_index])) {
                    $field_values->[$field_index] = defined($table_settings{$table_name}{$field_name}{default}) ? $table_settings{$table_name}{$field_name}{default} 
                                                                                      : 'NULL';
                }
                elsif ($field_values->[$field_index] eq '\N') {
                    $field_values->[$field_index] = 'NULL';
                }
                else {
                    $field_values->[$field_index] =~ s/\\/\\\\/g;
                    $field_values->[$field_index] =~ s/\'/\'\'/g;
                    $field_values->[$field_index] =~ s/\n/\\n/g;
                    $field_values->[$field_index] =~ s/\r/\\r/g;
                    if (!looks_like_number($field_values->[$field_index]) || ($table_settings{$table_name}{$field_name}{type} && $table_settings{$table_name}{$field_name}{type} =~ /text/)) {
                        $field_values->[$field_index] = "'$field_values->[$field_index]'";
                    }
                }
            }

            push (@field_value_insert_list, "(@$field_values)");

            if (scalar @field_value_insert_list == 100) {
                print $outf "INSERT INTO $table_name (@$field_names) VALUES " .
                  join (',', @field_value_insert_list) .
                    ";\n";
                @field_value_insert_list = ();
            }
        }

        if (scalar @field_value_insert_list > 0) {
            print $outf "INSERT INTO $table_name (@$field_names) VALUES " .
              join (',', @field_value_insert_list) .
                ";\n";
        }
    }
    else {
        my $table_name;
        if ($source_file =~ /^.*table_schema.+?\.([^\.]*)\.(sql)$/) {
            $table_name = $1;
        }
        while (my $line = <$source_fh>) {
            if ($table_name && $line =~ /^[ ]*([^ ]+)[ ]*([^ ]+)/) {
                my $field = $1;
                my $type = $2;
                $table_settings{$table_name}{$field}{type} = $type;
            }
            if ($line =~ /^[ ]*([^ ]+).*default[ ]*([^, ]*)/i) {
                my $field = $1;
                my $value = $2;
                $value =~ s/'//g;
                if (!looks_like_number($value) || $table_settings{$table_name}{$field}{type} =~ 'text') {
                    $value = "'$value'";
                }
                $table_settings{$table_name}{$field}{default} = $value;
            }
            print $outf $line;
        }
    }
    close $source_fh;
    print $outf "\n\n";
  }

  close $outf;
}




sub getfile {
  my($filename) = @_;
  open(my $fh, $filename);
  my $contents = join('', <$fh>);
  close $fh;
  return $contents;
}
