#!/usr/bin/perl
use strict;
use warnings;

use F5::BigipUtils;
use F5::Deploy;

use constant DEFAULT_XPATH_FILENAME => 'query.xp';
use constant DEFAULT_NAMESPACE_FILENAME => 'namespace_mapping.xp';
use constant DEFAULT_SCHEMA_FILENAME => 'schema.xsd';
use constant DEFAULT_SCHEMA_FILENAME_WSDL => 'base.xsd';
use constant DEFAULT_CONFIG_FILENAME => 'config.xml.cfg';
use constant DEFAULT_DEBUG_FILENAME => 'debug.bitmask';
use constant START_RED => "\033[31;1m";
use constant START_GREEN => "\033[32;1m";
use constant END_COLOR => "\033[0m";
use constant OK => START_GREEN.'OK    '.END_COLOR;
use constant NOK => START_RED.'NOT OK'.END_COLOR;
use constant SUCCESSFUL_DIR => START_GREEN.'DIR PASS'.END_COLOR;
use constant FAILED_DIR => START_RED.'DIR FAIL'.END_COLOR;
use constant FATAL_DIR => START_RED.'DIR FATAL'.END_COLOR;
use constant XML_PARSER_EOF_DELIM => 'file_end_delimiter';
use constant VERSION_FILE_ON_BIGIP => '/VERSION';
use constant PRODUCT_PROCESSOR => 'processor'; # xml processor
use constant PRODUCT_XPATH => 'xpath';
use constant PRODUCT_JSON => 'json';
use constant PRODUCT_XDEFENSE => 'xdefense'; # xml processor in ASM lite
use constant PRODUCT_WSS => 'wss'; # encryption and decryption engine
use constant XPATH_COMPILER_EXEC => 'TrafficShield/xml/processor/compiler/xpathcompiler.pl';
use constant SUFFIX_EXPECTED_OUTPUT_FILENAME => '.exp';
use constant WSS_EXEC => 'TrafficShield/xml/processor/xml_wss_test/Debug/xml_wss_test';

my %dynamic_debug_hash_table = ("ESSENTIAL","0x1",
	"SAX_EVENTS","0x2",
	"XMLPARSER","0x4",
	"WSS_PARSER","0x8",
	"XPATHPARSER","0x10",
	"FETCHPARSER","0x20",
	"CANONICALIZER","0x40",
	"TOKENIZER","0x80",
	"CONFIG","0x100",
	"CALLBACK","0x200",
	"KEYSTORE","0x400",
	"SCHEMA_LOAD","0x800",
	"JSON_PARSER","0x1000",
	"SCHEMA_VALIDATOR","0x2000",
	"SCHEMA_EVENTS","0x8000",
	"XMLPARSER_TRIE_DUMP","0x10000",
	"TRIE","0x20000",
	"HASH","0x40000",
	"MEMORY_FOOTPRINT","0x80000",
	"DYNAMIC_MEMORY","0x100000",
	"MEMORY_UTIL","0x200000",
	"CACHING","0x400000",
	"XPATH_COMPILE","0x900000",
	"XPATH_ALGORITHM","0x1000000",
	"WSS_PARSER_ESSENTIAL","0x2000000",
	"BUFFER_MANAGER","0x4000000",
	);




use Getopt::Std;
use File::Basename;
use File::Path;
use Cwd;
use Data::Dumper;
use Digest::MD5 qw(md5_hex);
use XML::Twig;

my ($client_root, $client_root_without_home, $client_arch, @ORIG_ARGV, $amir_tr);

sub usage {
  my($mess) = @_;

  my $amir_help = `$amir_tr --help 2>&1`;

  die "${\($mess ? qq/Error: $mess\n/ : '')}
lib/generic Test Runner Help (libxml2, reactivity or all processors)
--------------------------------------------------------------------

$amir_help

XML Group Test Runner Help (F5 processor)
-----------------------------------------

Usage: ${\(basename $0)} [flags] [test|testrootdir]...
  -1                          run xml_parser_test for each xml instance (default: multi mode)
  -a                          read from and write to failed_tests.txt, skip from skip_tests.txt
  -b <executable>             choose the driver to run your tests (default: xml/processor/xml_parser_test/Debug/xml_parser_test)
  -c                          print statistics for each shell command run
  -d <debug-bitmask>          when using DYNAMIC_DEBUG, turn on requested MODs (from debug.h) with a bitmask (example: ENCODING|HASH == 0x8001)
  -e                          run xpath instead of xml processor
  -f <p|n|w>                  'p' - run only positive (_w or _v) tests, 'f' - run only negative (_x, _s or _d) tests, 'w' - run only wellformedness/defense tests, excluding _s tests.
  -g                          run with valgrind for memory debugging: creates valgrind_errors.txt, valgrind_tests.txt
  -h                          print help
  -i <filename>               run tests listed in filename
  -j                          run json_tests.
  -k                          run wss_xml_test (Encryption/Decryption) instead of xml processor
  -l                          if -g, then use --leak-check=full
  -m                          run xdefense instead of xml processor
  -n                          pretend mode
  -o <filename>               list failed tests in filename
  -p <f5|libxml2|reactivity|all>  xml processor to run (default: f5)
                                     if this argument contains anything but F5, then the current
                                     command is replaced with Amir's test_runner.pl, which is then run
                                     with the original given arguments, help for which preceeds this section.
  -r                          force schema compilation (default: s.dat is cached unless schema or schema compiler changed)
  -s <filename>               save run output to filename
  -t <iterations>             test iterations (\\d*[KMB]?) optional number followed by optional K=1000, M=1,000,000, B=1,000,000,000
  -v                          verbose mode
  -w                          first convert wsdl files to xsd
  -x <filename>               skip over tests listed in filename (listed tests can also be directory nodes or patterns)
  -z <num_iter>               randomize test's sequence - argument should be a number specifying the number of iterations(tests).
  -W                          only convert wsdl files (and don't run xml processor)
  -R                          remove generated files (like the valuable compiled schema) after run to save space

  This program runs some xml_processor tests.

";
}

if (on_bigip()) {
}else{
	($client_root) = `p4 info` =~ m|Client root:\s*(.*)$|m;
	($client_root_without_home = $client_root) =~ s|/.*?/.*?/|/|;

        ($client_arch) = `uname -p`;
        if($client_arch =~ /x86_64/) {
           $ENV{LD_LIBRARY_PATH}  = "/usr/lib64;".$ENV{HOME}."$client_root_without_home/devfs_x86_64/usr/lib64/;".$ENV{HOME}."$client_root_without_home/devfs_x86_64/lib64/";
        }
        else {
           $ENV{LD_LIBRARY_PATH}  = "/usr/lib;".$ENV{HOME}."$client_root_without_home/devfs/usr/lib/;".$ENV{HOME}."$client_root_without_home/devfs/lib/";
        }
	$amir_tr = "$client_root/TrafficShield/xml/processor/tests/lib/generic/test_runner.pl";

}

# if -p reactivity|libxml2, use Amir's test_runner
for (my $i=0; $i<@ARGV; $i++) {
    if ($ARGV[$i] =~ /^-p/) {
	last if $ARGV[$i+1] =~ /^F5$/i;
	exec "$amir_tr @ARGV";
    }
}

use vars qw($opt_1 $opt_a $opt_b $opt_c $opt_d $opt_e $opt_f $opt_g $opt_h $opt_i $opt_j $opt_k $opt_l $opt_m $opt_n $opt_o $opt_p $opt_r $opt_s $opt_t $opt_v $opt_w $opt_x $opt_z $opt_W $opt_R );
getopts("1ab:cd:ef:ghi:jklmno:p:rs:t:vwx:z:WR") or usage;
usage if $opt_h;
our @expected_result_errors;

$SIG{INT} = sub { print "${\(basename $0)} caught INT; exiting...\n"; exit };

sub my_system {
    my ($cmd) = @_;

    if ($opt_n) {
	print "$cmd\n";
    } else {
	system $cmd;
    }
}

my $multi_run = not $opt_1;
my $auto_verbose;
my $regression_dir = Cwd::getcwd;
my %verrs;
my @valstrings = ('Invalid read', 'Invalid write', 'Mismatched free', 'definitely lost');

my(@list, @skip, $arg);

$opt_i = 'failed_tests.txt' if $opt_a;
$opt_o = 'failed_tests.txt' if $opt_a;
$opt_x = 'skip_tests.txt' if $opt_a;
$opt_w = 1 if $opt_W;

my $test_filter = '';
if ($opt_f) {
    if ($opt_f eq 'p') {
	$test_filter = '_[wv]\.xml$';
    } elsif ($opt_f eq 'n') {
	$test_filter = '_[xsd]\.xml$';
    } elsif ($opt_f eq 'w') {
	$test_filter = '_[xwvd]\.xml$';
    }
} 

my %mult = ('k' => 1000, 'm' => 1000000, 'b' => 1000000000);

if ($opt_t) {
    usage "-t $opt_t is illegal" unless $opt_t =~ /^(\d*)([KMB]?)$/i;
    $opt_t = ($1 || 1) * ($mult{$2} || 1);
}

my $v_string = "";
if ($opt_g) {
    $v_string = "valgrind ";
    if ($opt_l) {
	$v_string .= "--leak-check=full ";
    }
}

my $c_string = $opt_c ? '/usr/bin/time -v ' : '';

chomp(@list = `cat $opt_i`) if $opt_i;
chomp(@skip = `cat $opt_x`) if $opt_x;

my $product = PRODUCT_PROCESSOR;
$product = PRODUCT_XPATH    if $opt_e;
$product = PRODUCT_XDEFENSE if $opt_m;
$product = PRODUCT_WSS      if $opt_k;
$product = PRODUCT_JSON     if $opt_j;

my $running_folder = "Debug";
if (defined $opt_d)
{
  $running_folder="Dynamic_Debug";
  if(exists $dynamic_debug_hash_table{$opt_d})
  {
      $opt_d = $dynamic_debug_hash_table{$opt_d}; 
  }
 
  
  
}


my $exec_cmd;
if ($product eq PRODUCT_PROCESSOR) {
    $exec_cmd = $opt_b ||= on_bigip() ? "../xml_parser_test" : "$client_root/TrafficShield/xml/processor/xml_parser_test/$running_folder/xml_parser_test";
    
}
elsif ($product eq PRODUCT_XPATH) {
    
    $exec_cmd = $opt_b ||= on_bigip() ? "../xml_xpath_test" : "$client_root/TrafficShield/xml/processor/xml_xpath_test/$running_folder/xml_xpath_test";
     
}
elsif ($product eq PRODUCT_JSON) {
    
    $exec_cmd = $opt_b ||= on_bigip() ? "../json_test" : "$client_root/TrafficShield/xml/processor/json_test/$running_folder/json_test";
     
}
elsif ($product eq PRODUCT_XDEFENSE) {
    $exec_cmd = $opt_b ||= on_bigip() ? "../xml_defense_test" : "$client_root/TrafficShield/xml/processor/xml_defense_test/$running_folder/xml_defense_test";
    
    
    
} elsif ($product eq PRODUCT_WSS) {
	$exec_cmd = $opt_b = $client_root."/".WSS_EXEC;
	$multi_run = 0;# this part should be deprecated and removed once CR 121114 is done
	$test_filter = qr/(?:req|res)_[^\/\.\s]+_[vs]\.xml$/; # CR 121272
} else { 
  usage("Unsupported product |$product|\n");
}

$opt_b = './ximline.pl' if ($opt_b and ($opt_b eq 'new'));

push @ARGV, @list;
push @ARGV, '.' unless @ARGV;
my @tests=();
foreach $arg (@ARGV) {
    $arg =~ s/^\s+//;
    $arg =~ s/\s+$//;
    $arg = "$regression_dir/$arg" unless $arg =~ m|^/|;
    if ($arg =~ /\.xml$/ or ($opt_j and $arg =~ /\.js$/)) {
	next unless $arg =~ /$test_filter/;
	my $d = dirname($arg);
	push @tests,{
	    dir=>$d,
	    config=>$d.'/'.DEFAULT_CONFIG_FILENAME,
	    schema=>$d.'/'.DEFAULT_SCHEMA_FILENAME,
	    xpath=>$d.'/'.DEFAULT_XPATH_FILENAME,
	    namespace=>$d.'/'.DEFAULT_NAMESPACE_FILENAME,
	    debug=>$d.'/'.DEFAULT_DEBUG_FILENAME,
	    xml=>$arg,
	};
	$auto_verbose++;
    } else {
	push @tests, get_dir_tests($arg, 0);
    }
}
#die Data::Dumper::Dumper(\@tests);

my $do_schema_validation=0;
my $have_compiled_xpath;
my @executed_tests;
my @successful_tests;
my @failed_tests;
my @fatal_tests;

if ($opt_s) {
    open(RUNOUT, ">$opt_s") || die;
} 

if ($opt_g) {
    open(VALERRORS, ">valgrind_errors.txt") || die;
    open(VALTESTS, ">valgrind_tests.txt") || die;
} 

my ($sdat, $sdf, $cdf, $nsd, $ncd, $ret, $cache_dir, $cmd, $xdat);
my $schema_code_path = on_bigip()? F5::Deploy::map_path('/usr/lib/perl5/site_perl/F5/XML/Schema/') : "$client_root/TrafficShield/xml/processor/xsd2das/";
my $schema_compiler_source_ls = `ls -ltR $schema_code_path`;
my $xsd2das_location = on_bigip()? F5::Deploy::map_path('/usr/share/ts/bin/xml-processor/') : "$client_root/TrafficShield/xml/processor/xsd2das/";
foreach my $test (@tests) {
	print "\n","="x5,"[",$test->{dir},"]","="x5,"\n";
	if ($opt_w) {
	    $cmd = "$amir_tr -wsdl -norun_tests $test->{dir}";
	    my_system $cmd;
	    next if $opt_W;
	    if ($? >> 8) {
		fatal_dir($test, "wsdl conversion", $cmd);
	    }
	}
	$do_schema_validation = 0;
	if ($product eq PRODUCT_PROCESSOR) {
	    $test->{schema}=~s,/+,/,g if $test->{schema};
	    $test->{schema} = resolve_schema_filename($test->{schema});
	    if ($test->{schema} && -f $test->{schema}) {
		$cache_dir = "/tmp$test->{dir}";
		mkpath $cache_dir;
		$sdat = "$cache_dir/s.dat";
		$sdf = "$cache_dir/schema_sources.md5";
		$cdf = "$cache_dir/schema_compiled.md5";
		($ret, $nsd) = schema_needs_compile($test, $sdat, $sdf, $cdf);
		if ($ret) { # needs compile ? 
		    my $no_permission = give_me_writability_or_give_me_death($sdat, $sdf, $cdf);
		    my $cmd = "$c_string $xsd2das_location/schema2das.pl $test->{schema} > $sdat";
		    my $ret = my_system $cmd;
		    die "caught CTRL-C, exiting" if ($ret & 127) == 2;
		    if ($no_permission or $ret or (not (-f $sdat and -s $sdat))) {
			fatal_dir($test, "schema compilation", $cmd);
			next;
		    }
		    write_digests($sdf, $cdf, $nsd, $sdat);
		}
		$do_schema_validation=1;
	    }
	}

	$have_compiled_xpath = 0;
	if ($opt_e && $test->{xpath} && -f $test->{xpath}) {
	    $cache_dir = "/tmp$test->{dir}";
	    mkpath $cache_dir;
	    $xdat = "$cache_dir/xpath.dat";
	    my $cmd = $client_root.'/'.XPATH_COMPILER_EXEC." -x $test->{xpath} -n $test->{namespace} > $xdat";
	    my $ret = my_system $cmd;
	    die "caught CTRL-C, exiting" if ($ret & 127) == 2;
	    $have_compiled_xpath = 1;
	}



	my @xml_filenames;
	if ($test->{xml}) {
		push @xml_filenames,$test->{xml};
	}else{
		opendir(my $dir_fh,$test->{dir}) or die "Couldn't open dir ".$test->{dir}.":$!\n";
		foreach my $f (sort readdir($dir_fh)) {
			my $filename = $test->{dir};
			$filename.='/' unless $filename=~m,/$,;
			$filename.=$f;
			next unless -f $filename and ($filename=~/\.xml$/ or ($opt_j and $filename=~/\.js$/));
			next unless $filename =~ /$test_filter/;
			push @xml_filenames,$filename;
		}
		closedir $dir_fh;
	}
	my @failed_tests_in_dir;
	my @successful_tests_in_dir;
	my @fatal_tests_in_dir;
	my $output_accumulation='';
	my $executed_tests_so_far=@executed_tests;
	
	chomp($opt_d = `cat $test->{debug}`) if -f $test->{debug};
	my $d_string = $opt_d ? "--debug " . eval $opt_d : "";

	my @parser_feedback;
	if ($multi_run) {
		if ($opt_z) {
			die "option j is |$opt_z| is not numeric. Aborting!\n" if $opt_z!~/^\d+$/;
			my @xml_randomized_filenames;
			for(my $i=0; $i<$opt_z; ++$i) {
				push @xml_randomized_filenames,$xml_filenames[int(rand(scalar(@xml_filenames)))];
			}
			@xml_filenames = @xml_randomized_filenames; # evil!!!
		}
		push @executed_tests,@xml_filenames;
		my $parse_line = "$c_string $v_string $opt_b $d_string ";
		
		if ($product eq PRODUCT_PROCESSOR) {
			$parse_line.= qq{ --namespace };
		}
		$parse_line.= qq{ --strict --schema \@} . $sdat if $do_schema_validation;
		$parse_line.= qq{ \@} . $xdat if $have_compiled_xpath;
		$parse_line.= qq{ --config } . $test->{config} if -f $test->{config};
        
		$parse_line.= qq{ --iter $opt_t } if $opt_t;
		$parse_line.= " ".join " ", @xml_filenames;
		$parse_line.= qq{ 2>&1};
                my $cmdline = on_bigip() ? '' : qq{export LD_LIBRARY_PATH="}.$ENV{LD_LIBRARY_PATH}.qq{"; };
                $cmdline .= $parse_line; 
		print("$cmdline\n"), next if $opt_n;
		my $parser_out = qx{$cmdline};
		my $return_code = $?;
		print RUNOUT $parser_out if $opt_s;
		$parser_out=' ' unless defined $parser_out && length($parser_out);
		my @results = split XML_PARSER_EOF_DELIM, $parser_out;
		push @results, ("Test group crashed") x (@xml_filenames - @results);

		for (my $i=0; $i<@xml_filenames; ++$i) {
			my $result_analysis = 
				analyze_result($product,$xml_filenames[$i],$return_code,$results[$i],undef);
			push @parser_feedback, {
				%$result_analysis,
				output => $results[$i],
				xml_filename => $xml_filenames[$i],
				test => $executed_tests_so_far + $i,
			};
		}
	} else { ## not a multi run
		foreach my $xml_filename (sort @xml_filenames) {
			push @executed_tests,$xml_filename;
			my $parse_line = "$c_string $v_string $opt_b $d_string ";
                        $parse_line =~ s/^\s+//; # trim leading whitespaces
			if ($product eq PRODUCT_PROCESSOR) {
				$parse_line.= qq{ --namespace };
			}

			my $expected_output_filename = undef;
			my $result_buffer_filename = undef;
			my $wss_files=undef;
			if ($product eq PRODUCT_WSS) {
				die "\n\n===> Could not find file:\n$xml_filename\n\n"
					unless -f $xml_filename;
				$wss_files = get_wss_files_for_test($client_root,$test->{dir});
				my $wss_mode = $wss_files->{get_SOAP_mode}->($xml_filename);
				$wss_files->{mode} = $wss_mode;
				$expected_output_filename = $wss_files->{get_expected_results_filename}->($xml_filename);
				$parse_line.= qq{ --config $wss_files->{config}};
				$parse_line.= qq{ --dump $wss_files->{result_buf}};
				$parse_line.= qq{ --$wss_mode };
			} else { #  not wss
				$parse_line.= qq{ --strict --schema \@} . $sdat if $do_schema_validation;
				$parse_line.= qq{ \@} . $xdat if $have_compiled_xpath;
				$parse_line.= qq{ --config } . $test->{config} if -f $test->{config};
			}

			$parse_line.= qq{ --iter $opt_t } if $opt_t;
			$parse_line.= qq{ $xml_filename 2>&1 };
			my $cmdline = on_bigip() ? '' : qq{export LD_LIBRARY_PATH=}.$ENV{LD_LIBRARY_PATH}.qq{; };
                        $cmdline .= $parse_line; 
			print("$cmdline\n"), next if $opt_n;
                        my $parser_out = qx{$cmdline};
			my $return_code = $?;
			#$parse_line.= qq{ >  };
			#$parse_line.= qq{ out.foo };
                        #my $return_code = system($parse_line);
                        #my $FHOUT;
                        #my $parser_out = "";
                        #open $FHOUT, "out.foo";
                        #while(<$FHOUT>) {
                        #   $parser_out .= $_;
                        #}
                        #close $FHOUT;
			print RUNOUT $parser_out if $opt_s;
			my $result_analysis = analyze_result($product,$xml_filename,$return_code,$parser_out,$wss_files);
			$parser_out='' unless defined $parser_out;
			push @parser_feedback,{ 
				output => $parser_out,
				%$result_analysis,
				xml_filename => $xml_filename,
				test => scalar(@executed_tests),
			};
		}
	}
	foreach my $r (@parser_feedback) {
		
		$output_accumulation.= "\ttest #$r->{test}: ";
		if ($r->{ok}) {
			push @successful_tests_in_dir,$r->{xml_filename};
			$output_accumulation.= OK.": ".$r->{xml_filename}."\n";
		}else{
			if ($r->{crash}) {
				push @fatal_tests_in_dir,$r->{xml_filename};
			}else{
				push @failed_tests_in_dir,$r->{xml_filename};
			}
			$output_accumulation.=NOK.": ".$r->{xml_filename}."\n";
		}
		print $r->{output} if $opt_v || ($auto_verbose && @tests == 1);;
		print @expected_result_errors if ($auto_verbose && @tests == 1);
	}
	push @successful_tests, @successful_tests_in_dir;
	push @failed_tests, @failed_tests_in_dir;
	push @fatal_tests, @fatal_tests_in_dir;
	my $total_bad_string = (@fatal_tests or @failed_tests) ? "          TOTAL(" . START_RED . "FATAL " . END_COLOR . scalar(@fatal_tests) . "/ " . START_RED . "FAILED " . END_COLOR . scalar(@failed_tests) . "/ " . START_GREEN . "PASSED " . END_COLOR . scalar(@successful_tests) . ")" . END_COLOR : "";
	my $all_tests_in_dir = scalar(@successful_tests_in_dir) + scalar(@failed_tests_in_dir);
	if (@successful_tests_in_dir==@xml_filenames and not exists $test->{xml}) {
		print "\ttests #".$executed_tests_so_far.'-#'.scalar(@executed_tests).": ";
		print SUCCESSFUL_DIR," (".$all_tests_in_dir." tests) $total_bad_string\n";
	} elsif (@failed_tests_in_dir==@xml_filenames and not exists $test->{xml}) {
		print "\ttests #".$executed_tests_so_far.'-#'.scalar(@executed_tests).": ";
		print FAILED_DIR," (".$all_tests_in_dir." tests) $total_bad_string\n";
	} elsif (@fatal_tests_in_dir==@xml_filenames and not exists $test->{xml}) {
		print "\ttests #".$executed_tests_so_far.'-#'.scalar(@executed_tests).": ";
		print FATAL_DIR," (".$all_tests_in_dir." tests) $total_bad_string\n";
	}else {
		print $output_accumulation;
		print "Directory summary:  ".scalar(@failed_tests_in_dir)."/".$all_tests_in_dir." failed.  ".
			scalar(@successful_tests_in_dir)."/".$all_tests_in_dir." successful.\n" if @tests > 1;
	}
	rmtree $cache_dir if $opt_R;	
}
if (@fatal_tests) {
	print "Fatal test";
	if (@fatal_tests==1) {
		print " is";
	} else {
		print "s are";
	}
	print ":\n";
	map { print "\t$_\n" } @fatal_tests;
}
if (@failed_tests) {
	print "Failed test";
	if (@failed_tests==1) {
		print " is";
	} else {
		print "s are";
	}
	print ":\n";
	map { print "\t$_\n" } @failed_tests;
}
print "Summary:  ".scalar(@fatal_tests)."/".scalar(@executed_tests)." fatal.\t".scalar(@failed_tests)."/".scalar(@executed_tests)." failed.\t".scalar(@successful_tests)."/".scalar(@executed_tests)." successful.\n" 
	if @executed_tests;

if ($opt_g) {
    print "Valgrind Summary:\n";
    foreach my $v (@valstrings) {
	printf("%20s: %d\n", $v, $verrs{$v}||0);
    }
}

my $t;
if ($opt_o) {
    open(OUTFILE, ">$opt_o") || die "Can't open $opt_o for write";
    foreach $t (@fatal_tests, @failed_tests) {
	print OUTFILE "$t\n";
    }
}

sub fatal_dir {
    my ($test, $operation, $command) = @_;

    print FATAL_DIR;
    # list fatals
    my $fata_test_count=0;
    my @bad_schema_tests;
    warn "problems in $operation while executing\n\t$command\n";
    if ($test->{xml}) {
	push @bad_schema_tests, "$test->{dir}/$test->{xml}";
    } else {
	opendir(my $dir_fh,$test->{dir}) or die "Couldn't open dir ".$test->{dir}.":$!\n";
	foreach my $f (readdir($dir_fh)) {
	    my $filename = $test->{dir};
	    $filename.='/' unless $filename=~m,/$,;
	    $filename.=$f;
	    next unless -f $filename and $filename=~/_[wxvsd]\.xml$/;
	    next unless $filename =~ /$test_filter/;
	    push @bad_schema_tests, $filename;
	}
	closedir $dir_fh;
    }
    push @fatal_tests, @bad_schema_tests;
    push @executed_tests, @bad_schema_tests;
    $fata_test_count += @bad_schema_tests;
    print "($fata_test_count tests)";
    print ": ignoring tests due to $operation problems: ".$test->{dir}."\n";
}

sub get_dir_tests {
	my ($regression_dir, $level) = @_;
	my $found_test;
	my @tests;
	my $have_files = 0;
	
	$regression_dir =~ s/\/$//;
	
	opendir(my $dir_fh,$regression_dir) or die "Couldn't open dir $regression_dir: $!\n";
	foreach my $f (sort readdir($dir_fh)) {
		next if $f=~/^\.\.?$/;
		next if $f=~/xmlsec-crypto-config/; # see: CR 121272
		my $d = $regression_dir.'/'.$f; 
		next if scalar(grep($d =~ /$_/, @skip));
		#next unless -d $d;
		if(-d $d) {
			foreach my $test (get_dir_tests($d, $level+1)) {
				push @tests, $test; 
			}
		}
		else {
			$have_files++ if -f $d and ($d =~ /\.xml$/ or ($opt_j and $d =~ /\.js$/));
		}
	}
	if($have_files > 0) {
		push @tests, {
				dir=>$regression_dir,	
				config=>$regression_dir.'/'.DEFAULT_CONFIG_FILENAME,
				schema=>$regression_dir.'/'.DEFAULT_SCHEMA_FILENAME,
				debug=>$regression_dir.'/'.DEFAULT_DEBUG_FILENAME,
				xpath=>$regression_dir.'/'.DEFAULT_XPATH_FILENAME,
				namespace=>$regression_dir.'/'.DEFAULT_NAMESPACE_FILENAME,
		};
	}

	if ($level == 1 and scalar @tests < 1) {
		print "\n","="x5,"[",$regression_dir,"]","="x5,"\n";
		print "No tests found in $regression_dir \n\n";
	}
	
	closedir $dir_fh;
	return @tests if scalar @tests >= 1;
	return;
}

sub is_parser_error {
    my ($parser_out) = @_;
    if ($opt_k) {
      return 1 if $parser_out && $parser_out=~/Internal error|Document malformed|Certificate expired|Certificate invalid|Decryption failure|Encryption failure|Signing failed|Signature verification failed|Timestamp missing|Timestamp invalid|Timestamp expired|Timestamp exiped to far in the future|Timestamp unsigned/;
      #return 1 if $parser_out && $parser_out=~/Invalid or unsupported type of security token|Invalid or unsupported type of algorithm|Invalid or unauthenticated or unauthenticatable security token|Invalid signature|Decryption failure|Encryption failure|Referenced security token is unavailable|Unsupported namespace|A Security header with mustUnderstand=\"1\" was not understood|Message expired|The security token could not be authenticated or authorized|Document is not wellformed|Schema of incoming document is invalid|Signature generation failed|Signature verification failed/;
    }
    return 1 if $parser_out && $parser_out=~/(?:Schema validation|Namespace validation|Syntax error|Syntax warning|Document malformed|Parser attack|Defence alert|Internal error):\s|Cannot continue|Test group crashed/;
    return 0;
    
}

sub extract_query_value_and_count {
    my($text) = @_;
    my $res;
    my ($pat, $tag);
    my $index = 0;

    while($text =~ /(<xp(query|value)>.*?<\/xp\2>)/g) {
	($pat, $tag) = ($1, $2);
	$res .= "$index $pat\n";
	$index++ if $tag eq 'value';
    }
    return $res || '';
}

sub diffcmd {
    my ($seq1, $seq2) = @_;
    my $res;

    my $file1 = "/tmp/diff-$$-1";
    my $file2 = "/tmp/diff-$$-2";

    open(FILE, ">$file1") || die;
    print FILE $seq1;
    
    open(FILE, ">$file2") || die;
    print FILE $seq2;

    close FILE;
    
    $res = `diff $file1 $file2`;
    unlink $file1, $file2;

    return $res;
}

sub is_expected_result_error {
    my ($xml_filename, $parser_out) = @_;
    my $expected_filename;
    my $exp_text;
    my ($exp, $act);
    my $error_text;
    my $pattern;
    
    @expected_result_errors = ();
    
    # don't enforce expected result in case of iterative execution.
    if ($opt_j || $opt_t || $opt_z) {
      return 0;
    }

    # expected xpath file
    ($expected_filename = $xml_filename) =~ s/\.xml$/.exp/;
    if (-f $expected_filename) {
	$exp_text = `cat $expected_filename`;
	$exp = extract_query_value_and_count($exp_text);
	$act = extract_query_value_and_count($parser_out);
	if ($error_text = diffcmd($exp, $act)) {
	    push @expected_result_errors, "ERROR: expected(<) and actual(>) outputs differ:\n$error_text";
	    #print $error_text;
	    return 1;
 	}
   }

    # plain expected file
    ($expected_filename = $xml_filename) =~ s/\.xml$/.e/;
    if (-f $expected_filename) {
	chomp($pattern = (`cat $expected_filename`)[0]);
	if ($parser_out !~ m|$pattern|s) {
	    push @expected_result_errors, $error_text = "ERROR: Expected results not found\n";
	    print $error_text;
	    return 1;
	}
    }
    
    return 0;
}

sub analyze_result {
	my ($product,$xml_filename,$return_code,$parser_out,$product_files) = @_;
	my $ok=0;
	my $crash=0;
	my $expected_diff_ok=1;

	if ($product eq PRODUCT_WSS) {

		if ($return_code == -1) {
			warn "failed to execute product $product\n";
		} elsif ($return_code & 127) {
			warn sprintf("Crash with return code $return_code\n");
			return { ok=>0, crash=>1,};
		}


		if ($return_code!=0){
			warn "CRASH: product $product crashed with return code: $return_code\n";
			return { crash=>1, ok=>0 };
		}
		my $expected_output_filename = $product_files->{get_expected_results_filename}->($xml_filename);
		my $result_buffer_filename = $product_files->{result_buf};

	        if ('response' eq $product_files->{mode}){
                        ## disable xmlsec , it does not support some of encryption types
			# try with xmlsec -- to see if it works or the response is invalid or somehow bad
			#my $xmlsec_return_code = $product_files->{xmlsec_decrypt}->(
			#	$product_files->{xmlsec_config},
			#	$product_files->{xmlsec_keys},
			#	$product_files->{result_buf},
			#	$product_files->{xmlsec_output}
			#);
			#if ($xmlsec_return_code == -1) {
			#	warn "failed to execute xmlsec\n";
			#} elsif ($xmlsec_return_code & 127) {
			#	warn sprintf("Crash with return code $xmlsec_return_code\n");
			#	return { ok=>0, crash=>1,};
			#} 
			#if ($xmlsec_return_code !=0) {
			#	return { ok => 0, crash => 0, why => $xmlsec_return_code};
			#}
			# now decrypt using wss in order to get output that can be diffed with original
			my $wss_return_code = $product_files->{wss_decrypt}->(
				$product_files->{config},
				$product_files->{result_buf},
				$product_files->{wss_output}
			);
			if ($wss_return_code == -1) {
				warn "failed to execute wss to self decrypt the message encrypted with wss\n";
			} elsif ($wss_return_code & 127) {
				warn sprintf("Crash with return code $wss_return_code\n");
				return { ok=>0, crash=>1,};
			} 
			if ($wss_return_code !=0) {
				return { ok => 0, crash => 0, why => $wss_return_code};
			}
			$result_buffer_filename = $product_files->{wss_output}; # compare expected with xmlsec decrypted
		}

		if (not -f $expected_output_filename) {
                     if ($product_files->{get_test_tag}->($xml_filename) eq 'v') {
                             warn "Missing file expected in test: |".get_filename($expected_output_filename)."|\n";
                             $expected_diff_ok=0; # fail tests with no result buffer
                     }
		}else{
                   if ($product_files->{get_test_tag}->($xml_filename) eq 's') {
                      warn "Unnessesary expected results file in test: |".get_filename($expected_output_filename)."|\n";
                      $expected_diff_ok=1; # we don't force expected files existance for 's' tests
                   }
                   else {
			my @xml_pped=();
			if (not -f $result_buffer_filename) {
				warn "Missing result buffer file (result of wss operation): |".get_filename($result_buffer_filename)."|\n";
				$expected_diff_ok=0; # fail tests with no result buffer
			} else {
				my $xmlpp = $product_files->{xml_pp_do}->($result_buffer_filename);
				my $pped_res = $xmlpp->{pped_filename};
				if ($xmlpp->{xmlpp_rc}!=0) {
                                    my $errmesg="\n\nERROR: failed whitespace elimination:\n".$xmlpp->{xmlpp_out}."\n";
                                    warn $errmesg;
                                    push @expected_result_errors,$errmesg;
				}else{
                                    push @xml_pped,$pped_res;
                                       my $xmlpp =  $product_files->{xml_pp_do}->($expected_output_filename);
                                       my $pped_exp = $xmlpp->{pped_filename};
                                       if ($xmlpp->{xmlpp_rc}!=0) {
                                               my $errmesg="\n\nERROR: failed whitespace elimination:\n".$xmlpp->{xmlpp_out}."\n";
                                               warn $errmesg;
                                               push @expected_result_errors,$errmesg;
                                       }else{
                                               push @xml_pped,$pped_exp;
                                               my $cmdline = "diff ".join(' ',@xml_pped);
                                               my $diff_result = qx{$cmdline};
                                               if ($diff_result!~/^\s*$/) {
                                                       my $errmesg = "\n\nERROR: wss expected result buffer and actual result buffer are not identical:\n$diff_result\n";
                                                       warn $errmesg;
                                                       push @expected_result_errors, $errmesg;
                                                       
                                                       $expected_diff_ok = 0;
                                               } 
                                       }
				}
			}
			unlink $_ for @xml_pped;
                   }
                }

	}
	
	my $result_switches;
	if ($product eq PRODUCT_JSON) {
		($result_switches)=$xml_filename=~/_([wxvsd])\.js$/;
	} else {
		($result_switches)=$xml_filename=~/_([wxvsd])\.xml$/;
	}

	if (!$result_switches || $return_code) {
		$ok = 0; # crash!!
		$crash=1; 
	}else{
	    my @switches=split //,$result_switches;
		$crash=0;
		if (is_parser_error($parser_out) || is_expected_result_error($xml_filename, $parser_out)) {
			foreach my $s (@switches) {
				if ($s=~/^[dsx]$/) {
					$ok=1;
					last;
				} elsif ($s!~/^[wv]$/) {
					warn "Unsupported test tagging -- expected result unknown\n";
				}
			}
		} else {
			foreach my $s (@switches) {
				if ($s=~/^[wv]$/) {
					$ok=1;
					last;
				} elsif ($s!~/^[sxd]$/) {
					warn "Unsupported test tagging -- expected result unknown\n";
				}
			}
		}
	}

	if ($opt_g) {	    
	    my $val_flag = 0;
	    my @valput = $parser_out =~ m/(==\d+==.*\n)/gm;
	    my $valput = "@valput";
	    foreach my $v (@valstrings) {
		if ($valput =~ /$v/) {
		    $verrs{$v}++;
		    $val_flag = 1;
		}
	    }
	    if ($val_flag) {
		print VALERRORS "\n$xml_filename\n", @valput;
		print VALTESTS  "$xml_filename\n";
	    }
	}

	if ($product eq PRODUCT_WSS) {
		$ok = $ok && $expected_diff_ok;
	}

	return { ok => $ok, crash => $crash, };
}

sub resolve_schema_filename {
	my ($schema_filename) = @_;
	my $schema_default_name = DEFAULT_SCHEMA_FILENAME;
	my $wsdl_schema_default_name = DEFAULT_SCHEMA_FILENAME_WSDL;
	return $schema_filename if $schema_filename=~/$schema_filename$/ && -f $schema_filename;
	$schema_filename=~s/$schema_default_name$/$wsdl_schema_default_name/;
	return $schema_filename if -f $schema_filename;
	warn "===> Couldn't find neither a $schema_default_name nor a $wsdl_schema_default_name\n";
	return 0;
}

sub give_me_writability_or_give_me_death {
    my @files = @_;
    my $errors;

    for (@files) {
	(warn "No permission to write $_", $errors++) unless !-f $_ || -w _;
    }
    # non-zero return value denotes error
    return $errors;
}

sub schema_needs_compile {

    # The Legend of Many Similar Short Variable Names
    #
    # sd = source digest
    # cd = compiled digest
    # f  = file
    # o  = old
    # n  = new

    my ($test, $sdat, $sdf, $cdf) = @_;
    my ($osd, $ocd, $ncd);
    my $schema_source = `ls -lt $test->{dir}/*.xsd`;
    my $nsd = md5_hex($schema_compiler_source_ls . $schema_source);

    if (!$opt_r && -r $sdf && -r $cdf && -s $sdat) {
	chomp($osd = `cat $sdf`);
	chomp($ocd = `cat $cdf`);
	$ncd = md5_hex(bin_read("cat $sdat"));
	if ($nsd eq $osd && $ncd eq $ocd) {
	    return(0);
	}
    }
    return(1, $nsd);
}

sub write_digests {
    my ($sdf, $cdf, $nsd, $sdat) = @_;

    open(DIGEST, ">$sdf") || die "$sdf :$!\n";
    print DIGEST "$nsd\n";
    close DIGEST;

    my $ncd = md5_hex(bin_read("cat $sdat"));
    open(DIGEST, ">$cdf") || die "$cdf :$!\n";
    print DIGEST "$ncd\n";
    close DIGEST;
}

sub bin_read {
	my ($cmd) = @_;
	open my $sfh, "$cmd |" or die "Cannot perform $cmd: $!\n";
	binmode $sfh;
	my $data = join '',<$sfh>;
	close $sfh;
	return $data;
}

sub on_bigip { F5::BigipUtils::is_docker() || -f VERSION_FILE_ON_BIGIP}

sub get_wss_files_for_test {
	my ($client_root,$test_dir) = @_;
	my $factory_dir = "$client_root/TrafficShield/mng/db/factory_data/";
	my $required_files_for_a_wss_test = {
		factory_ns_map => "$factory_dir/PLC.PL_XML_SOAP_PREDEFINED_PREFIX_NAMESPACE_MAPPING.csv",
		factory_req_queries => "$factory_dir/PLC.PL_XML_SOAP_PREDEFINED_SPECIFIC_ELEMENTS.csv",
		factory_res_queries => "$factory_dir/PLC.PL_XML_SOAP_PREDEFINED_SPECIFIC_ELEMENTS.csv",
		user_defined_ns_map => $test_dir.'/'."user_defined_ns_mapping.xp",
		user_defined_res_queries => $test_dir.'/'."user_defined_encryption_queries.xp",
		user_defined_req_equries => $test_dir.'/'."user_defined_verification_queries.xp",
		ns_map => $test_dir.'/'."glued_ns_mapping.xp",
		req_queries => $test_dir.'/'."glued_request_queries.xp",
		res_queries => $test_dir.'/'."glued_response_queries.xp",
		req_queries_dat => $test_dir.'/'."request_queries.dat",
		res_queries_dat => $test_dir.'/'."response_queries.dat",
		config => $test_dir.'/'.DEFAULT_CONFIG_FILENAME,
		result_buf => $test_dir.'/'."result_buf",
		wss_output => $test_dir.'/'."wss_output.xml",
		xmlsec_config => $test_dir.'/'."xmlsec-crypto-config",
		xmlsec_keys => $test_dir.'/'."mykeys.xml",
		xmlsec_output => $test_dir.'/'."xmlsec_output.xml",
		get_SOAP_mode => sub {
			my ($filename) = @_;
			my $errstr= "Unexpected filename while trying to determine whether test input file is a request test or a response test\n";
			die $errstr unless $filename;
			if ($filename=~m,[^/]*req_[^/]+\.xml$,) {
				return 'request';
			} elsif ($filename=~m,[^/]*res_[^/]+\.xml$,) {
				return 'response';
			} 
			die $errstr;
		},
		xml_pp_do => sub {
			my($f)=@_;
			my $xml_pped_filename = $f.".no_ws";
			my $twig=XML::Twig->new( 
				pretty_print => 'none',
				comments => 'drop',
				pi => 'drop',
			);
			eval{$twig->xparse($f);};
			my $error_message = $@;
			my $rc=$error_message?1:0;
			my $outfile = $xml_pped_filename;
			open my $outfh,">:utf8",$outfile or die "cannot open $outfile for writing: $!\n";
			$twig->print(\*$outfh);
			close($outfh);
			my $xmlpp_out = '';
			$xmlpp_out = "xml_pp_do: $f: $error_message\n"
				if $rc;
			return { pped_filename => $xml_pped_filename, xmlpp_out => $xmlpp_out, xmlpp_rc => $rc};
		},
		get_test_tag => sub {
			my ($filename)=@_;
			my ($tag) = ($filename=~/_([vs])\.xml$/);
			warn "Cannot find test flag in |$filename|\n"
				unless $tag;
			return $tag;
		},
		get_expected_results_filename => sub { 
			my ($filename) = @_;
			my $errstr = "Unexpected filename while trying to construct expected results filename\n";
			die $errstr unless $filename;
			die $errstr if $filename!~/\.xml$/;
			return $filename.SUFFIX_EXPECTED_OUTPUT_FILENAME;
		},
		xmlsec_decrypt => sub {
			my ($xmlsec_config_filename,$xmlsec_keys_filename,$encrypted,$output_filename) = @_;
			my $encrypted_data_occurrences=0;
			my $twig=XML::Twig->new( 
				map_xmlns => {
					"http://www.w3.org/2001/04/xmlenc#" => 'xmlenc',
				},
				twig_handlers => {
					'xmlenc:EncryptedData' => sub { ++$encrypted_data_occurrences; },
				}
			);
			eval {$twig->xparse($encrypted);};
			die "Failed to parse $encrypted when trying to count occurrences of EncryptedData elements: $@\n"
				if ($@);
			my $number_of_xmlsec_executions = $encrypted_data_occurrences;
			my $in = $encrypted;
			my $out = $output_filename;
			system("touch $output_filename");
			my $new_in = $in."xmlsec_altered";
			for (my $i=0;$i<$number_of_xmlsec_executions;++$i) {
				my $cmdline = "xmlsec1 decrypt --crypto-config $xmlsec_config_filename --keys-file $xmlsec_keys_filename --output $out $in";
				my $output_xmlsec = qx/$cmdline/;
				return $? if $?;
				system("cp $out $new_in");
				$in = $new_in;
			}
		},
		wss_decrypt => sub {
			my ($config,$encrypted,$output) = @_;
                        # create new config
                        my $new_config = $config;
                        $new_config =~ s/cfg/cfs/;
                        my $client_cert = undef;
			my $client_twig = undef;
                        my $server_cert = undef;
			my $server_twig = undef;
			my $twig=XML::Twig->new( 
				pretty_print => 'none',
				comments => 'drop',
				pi => 'drop',
     				#map_xmlns => {
				#	"http://www.f5.com/ASM/xmlparser.conf" => '',
				#},
                                keep_original_prefix => 1,
                                twig_handlers => {
					'/policy/profile/wss/client_credentials/client_data/key_file' => sub {
                                          $client_cert = $_->first_child->text unless($client_cert);
					  $client_twig = $_ unless($client_twig);
                                          },
					'/policy/profile/wss/server_credentials/server_data/key_file' => sub {
					  $server_cert = $_->first_child->text unless($server_cert);
					  $server_twig = $_ unless($server_twig);
                                          #$_->set_text( $client_cert);
                                          },
				},
			);
			#eval{$twig->xparse($config);};
                        $twig->xparse($config);
			my $error_message = $@;
			my $rc=$error_message?1:0;
			$server_twig->set_text( $client_cert) if($server_twig and $client_cert);
			$client_twig->set_text( $server_cert) if($client_twig and $server_cert);
			open my $outfh,">:utf8",$new_config or die "cannot open $new_config for writing: $!\n";
			$twig->print(\*$outfh);
			close($outfh);

			my $cmdline = on_bigip() ? '' : "export LD_LIBRARY_PATH=\"".$ENV{LD_LIBRARY_PATH}."\"; ";
			$cmdline .= $exec_cmd ." --request --config $new_config --dump $output $encrypted 2>&1";
			system("touch $output");
			my $parser_out = qx{$cmdline};
			return $? if $?;
		},
	};
	my $csv2hash = sub {
		my ($csv_filename) = @_;
		open my $fact,"<",$csv_filename or die "Cannot open $csv_filename for reading: $!\n";
		my $h={};
		my $header_line = <$fact>;
		while(my $l=<$fact>){
			next if $l=~/^\s*$/;
			chomp $l;
			my($k,$v)=split ',',$l;
			push @{$h->{$k}},$v;
		}
		return $h;
	};

	my $glue = sub {
		my ($glued_filename,$factory_lines,@files_to_glue) = @_;
		open my $out,">",$glued_filename or die "Cannot open $glued_filename for writing: $!\n";
		print $out $factory_lines;
		foreach my $f (@files_to_glue) {
			if (not -f $f) {
#				warn "\n\t Ignoring missing file |$f| while attempting to glue it into |$glued_filename|\n\n";
				next;
			}
			open my $in,"<",$f or die "Cannot open $f for reading: $!\n\n";
			print $out join('',<$in>);
			close($in);
		}
		close $out  or die "Cannot close $glued_filename after writing: $!\n\n";
	};

	for my $f (
		$required_files_for_a_wss_test->{factory_ns_map},
		$required_files_for_a_wss_test->{factory_req_queries},
		$required_files_for_a_wss_test->{factory_res_queries}
	) {
		 die "Missing required factory file: $f\n" unless -f $f;
	}

	my $csv_hash = $csv2hash->($required_files_for_a_wss_test->{factory_ns_map});
	my $lines='';
	while(my($k,$v)=each %$csv_hash){
		foreach my $i (@$v) {
			$lines.=$k.' '.$i."\n";
		}
	}
	$glue->($required_files_for_a_wss_test->{ns_map},
			$lines,
			$required_files_for_a_wss_test->{user_defined_ns_map});
	$lines='';
	$csv_hash = $csv2hash->($required_files_for_a_wss_test->{factory_req_queries});
	while(my($k,$v)=each %$csv_hash){
		next unless $k eq 'decrypt';
		foreach my $i (@$v) {
			$lines.=$i."\n";
		}
	}
	$glue->($required_files_for_a_wss_test->{req_queries},
			$lines,
			$required_files_for_a_wss_test->{user_defined_req_equries});
	
	
	
	
	
	
	$lines='';
	$csv_hash = $csv2hash->($required_files_for_a_wss_test->{factory_res_queries});
	while(my($k,$v)=each %$csv_hash){
		next unless $k eq 'encrypt';
		foreach my $i (@$v) {
			$lines.=$i."\n";
		}
	}
	$glue->( $required_files_for_a_wss_test->{res_queries},
			$lines,
			$required_files_for_a_wss_test->{user_defined_res_queries});

	my $xpath_compile = sub {
		my ($ns_map,$queries,$dat_file) = @_;
		my $cmdline = $client_root.'/'.XPATH_COMPILER_EXEC." -x $queries -n $ns_map > $dat_file";
		qx/$cmdline/;
	};

	$xpath_compile->(
			$required_files_for_a_wss_test->{ns_map},
			$required_files_for_a_wss_test->{req_queries},
			$required_files_for_a_wss_test->{req_queries_dat}
	);
	$xpath_compile->(
			$required_files_for_a_wss_test->{ns_map},
			$required_files_for_a_wss_test->{res_queries},
			$required_files_for_a_wss_test->{res_queries_dat}
	);

	return $required_files_for_a_wss_test;
}

sub get_filename {
	my ($filename)=@_;
	my ($name,$path,$suffix) = File::Basename::fileparse($filename);
	return $name;
}

