#!/usr/bin/perl

use strict;
use warnings;
no warnings 'uninitialized';

use Getopt::Long;
use HTTP::Daemon;
use HTTP::Status;
use HTTP::Headers;
use JSON::PP;
use POSIX qw(setsid WNOHANG);

use F5::DbUtils;
use F5::Violation;
use F5::SystemUtils;
use F5::License;

use constant PORT => 8090;


$|++;


main(@ARGV);

#----------------------------------------------------------------------------------------

sub main {
    my @args = @_;

    my ($daemon, $port, $support_id, $last);
    GetOptions ("daemon" => \$daemon, "port=i" => \$port, "support_id=s" => \$support_id, "last" => \$last);
    if (! ($daemon || $support_id || $last)) {
        show_usage();
        exit;
    }

    if ($daemon && ! F5::License->new->is_production()) {
       start_daemon($port);
    }
    elsif($support_id) {
        print_request_by_support_id($support_id);
    }
    elsif($last) {
        print_last_request();
    }

}

#----------------------------------------------------------------------------------------

sub show_usage {
    my $port = PORT;

    print <<'MSGEND';
Usage:
print request message by support ID: $0 -s <support ID>
print last request message: $0 -l
MSGEND

    if (! F5::License->new->is_production()) {
        print <<'MSGEND';
activate as daemon (default port $port): $0 -d [-p <port num>]
MSGEND
    }

}

#----------------------------------------------------------------------------------------

sub start_daemon {
    my ($port) = @_;

    if (fork) {
        exit;
    }
    else {
        $SIG{CHLD} = sub { while( waitpid(-1, WNOHANG)>0 ) {} };
        $SIG{HUP} = sub {};

        setsid();
        open STDIN, '/dev/null';
        open STDOUT, '>/dev/null';

        run_daemon($port);
      }
}

#----------------------------------------------------------------------------------------

sub run_daemon {
    my ($port) = @_;

    $port ||= PORT;

    my $d = HTTP::Daemon->new(
        LocalPort => $port,
        Listen => 20,
        Proto  => 'tcp',
        Reuse  => 1,
        Timeout => 60*60 ) || die "Cannot start HTTP Daemon\n";

    warn "Start request viewer daemon\n";

    open STDERR, '>&STDOUT';

    while (1) {
        next unless my $conn = $d->accept();
        unless (fork) {
            procReq($conn);
            exit 0;
        }
    }

}

#----------------------------------------------------------------------------------------

sub procReq {
    my ($conn) = @_;
    local $SIG{PIPE} = 'IGNORE';

    $conn->autoflush(1);

    getParams($conn);

    $conn->close;
}

#----------------------------------------------------------------------------------------

sub getParams {
    my ($conn) = @_;

    my $r = $conn->get_request();
    return unless $r;

    my $quer = $r->uri;

    my $dbh = load_env();
    if ($quer =~ m|^/supportid\?id=(\d*)|) {
        my $support_id = $1;
        my $message_cont = request_by_support_id($dbh, $support_id);
        send_msg_response($conn, $message_cont);
    }
    elsif ($quer =~ m|^/lastprotobuf|) {
        my $message_cont = last_request($dbh);
        send_msg_response($conn, $message_cont);
    }
    elsif ($quer =~ m|^/backup_supportid\?id=(\d*)|) {
        my $support_id = $1;
        backup_supportid($dbh, $support_id, $conn);
    }
    elsif ($quer =~ m|^/show_backup|) {
        my $message_cont = proc_backup_proto($conn, $r);
        send_msg_response($conn, $message_cont);
    }
    elsif ($quer =~ m|^/insert_backup|) {
        my $message_cont = insert_backup_proto($dbh, $conn, $r);
        send_msg_response($conn, $message_cont);
    }
    else {
        showInp($conn);
    }

}

#----------------------------------------------------------------------------------------

sub send_msg_response {
    my ($conn, $message_cont) = @_;

    my $response = new HTTP::Response(200);
    $response->header('Content-Type' => 'application/json; charset=utf-8');
    $response->content($message_cont);
    $conn->send_response($response);
}

#----------------------------------------------------------------------------------------


sub showInp {
    my ($conn) = @_;

    my $content =
'<html><body>

<form action="/supportid">
  request support id: <input type="text" name="id"><br>
  <input type="submit" value="Get Protobuf Support ID">
</form>
<form action="/lastprotobuf">
  <input type="submit" value="Get Last Protobuf">
</form>
<form action="/backup_supportid">
  backup support id: <input type="text" name="id"><br>
  <input type="submit" value="Backup Protobuf Support ID">
</form>
<form enctype="multipart/form-data" action="/show_backup" method="POST">
  <input type="file" name="file"><br>
  <input type="submit" value="Show Backup Protobuf">
</form>
<form enctype="multipart/form-data" action="/insert_backup" method="POST">
  <input type="file" name="file"><br>
  <input type="submit" value="Insert Backup Protobuf">
</form>

</body></html>
';

    my $response = new HTTP::Response(200);
    $response->content($content);
    $conn->send_response($response);
}

#----------------------------------------------------------------------------------------

sub insert_backup_proto {
    my ($dbh, $conn, $r) = @_;
    my $protobuf_bin = get_backup_proto($conn, $r);

    my $request_log = F5::RequestLog->new(dbh => $dbh);
    $request_log->init_run_insert;
    $request_log->load_policy_mapping;
    $request_log->protobuf_message_received(raw_protobuf => $protobuf_bin);
    $request_log->{batch_insert}->flush;

    sleep 5;

    F5::SystemUtils::run_system_tool(command_line_params => ['pkill', 'asmlogd']);

    return parse_proto($protobuf_bin);
}

#----------------------------------------------------------------------------------------

sub backup_supportid {
    my ($dbh, $support_id, $conn) = @_;

    my $protobuf_bin = protobin_by_support_id($dbh, $support_id);
    my $filename = $support_id. ".pbf";

    my $h = HTTP::Headers->new;
    $h->header('Content-Type' => "application/octet-stream; file=$filename");
    $h->header('Content-Disposition' => "attachment; filename=$filename");
    $h->header('Content-Title' => "$filename");

    my $response = new HTTP::Response(200, undef, $h);
    $response->content($protobuf_bin);
    $conn->send_response($response);
}

#----------------------------------------------------------------------------------------

sub proc_backup_proto {
    my ($conn, $r) = @_;
    my $protobuf_bin = get_backup_proto($conn, $r);
    return parse_proto($protobuf_bin);
}

#----------------------------------------------------------------------------------------

sub get_backup_proto {
    my ($conn, $r) = @_;
    my $buff = $r->content;

    $buff =~ /^(.+?)\n/m;
    my $sep = $1;
    $sep =~ s/[\n\r]//g;
    $buff =~ s/$sep[\r\n]+//m;
    $buff =~ s/$sep.*//gm;
    my @message_lines = split(/\n/m, $buff);

    my $protobuf;
    my $cnf = 1;
    for my $line (@message_lines) {
        if ($cnf) {
            unless ($line =~ /^\w/) {
                $cnf = 0;
            }
            next;
        }
        $protobuf .= $line. "\n" ;
    }

    return $protobuf;
}

#----------------------------------------------------------------------------------------

sub print_last_request {
    my $dbh = load_env();
    my $message_cont = last_request($dbh);
    print "$message_cont\n";
}

#----------------------------------------------------------------------------------------

sub print_request_by_support_id {
    my ($support_id) = @_;

    my $dbh = load_env();
    my $message_cont = request_by_support_id($dbh, $support_id);
    print "$message_cont\n";
}

#----------------------------------------------------------------------------------------

sub last_request {
    my ($dbh) = @_;

    my $sth = $dbh->prepare(<<EndQuery);
SELECT protobuf_file_id, protobuf_file_seek_pos, protobuf_record_size, id
  FROM PRX.REQUEST_LOG where id in
  (SELECT max(id) from PRX.REQUEST_LOG)
EndQuery

    $sth->execute();
    my ($protobuf_file_id, $protobuf_file_seek_pos, $protobuf_record_size) = $sth->fetchrow_array;

    my $protobuf_bin = protobuf_from_file($protobuf_file_id, $protobuf_file_seek_pos, $protobuf_record_size);
    return parse_proto($protobuf_bin);
}

#----------------------------------------------------------------------------------------

sub request_by_support_id {
    my ($dbh, $support_id) = @_;

    my $protobuf_bin = protobin_by_support_id($dbh, $support_id);
    return parse_proto($protobuf_bin);
}

#----------------------------------------------------------------------------------------

sub protobin_by_support_id {
    my ($dbh, $support_id) = @_;

    my $sth = $dbh->prepare(<<EndQuery);
SELECT protobuf_file_id, protobuf_file_seek_pos, protobuf_record_size
  FROM PRX.REQUEST_LOG
  WHERE support_id=?
EndQuery

    $sth->execute($support_id);
    my ($protobuf_file_id, $protobuf_file_seek_pos, $protobuf_record_size) = $sth->fetchrow_array;

    return protobuf_from_file($protobuf_file_id, $protobuf_file_seek_pos, $protobuf_record_size);
}

#----------------------------------------------------------------------------------------

sub protobuf_from_file {
    my ($protobuf_file_id, $protobuf_file_seek_pos, $protobuf_record_size) = @_;

    my $rec = {
        protobuf_file_id => $protobuf_file_id,
        protobuf_file_seek_pos => $protobuf_file_seek_pos,
        protobuf_record_size => $protobuf_record_size
    };

    my $protobuf_bin = F5::RequestLog->get_protobuf(rec => $rec);
    return $protobuf_bin;
}

#----------------------------------------------------------------------------------------

sub parse_proto {
    my ($protobuf_bin) = @_;

    my $bd = bd::Main->new();
    $bd->unpack($protobuf_bin);
    my $bd_data = $bd->to_hashref();

    add_viol_ids( $bd_data );

    my $message_cont = JSON::PP->new->utf8->pretty->indent_length(5)->encode($bd_data);
    return $message_cont;
}

#----------------------------------------------------------------------------------------

sub load_env {
    require bd::Main;
    require F5::RequestLog;

    my $dbh = F5::DbUtils::get_dbh() || die "Cannot connet to DB: " . $DBI::errstr . "\n";
    return $dbh;
}

#----------------------------------------------------------------------------------------

sub add_viol_ids {
    my ($ent) = @_;

    my $viol_names = get_viol_names();

    my $refType = ref $ent;
    if ( $refType =~ /ARRAY/ ) {
        for my $ael (@$ent) {
            add_viol_ids($ael);
        }
    }
    elsif ( $refType =~ /HASH/ ) {
        for my $hel ( keys %$ent ) {
            if ($hel eq 'viol_set') {
                my @viol_idxs = F5::Violation::Utils::get_indexes_from_viol_set($ent->{$hel});
                my @names = map { $viol_names->{ $_ }.", id:$_" } @viol_idxs;
                $ent->{__viol_set_ids__} = \@names;
                delete $ent->{'viol_set'};
            }
            else {
                add_viol_ids($ent->{$hel});
            }
        }
    }
}

#----------------------------------------------------------------------------------------

sub get_viol_names {
    my $viol_names = {};
    for my $viol_name (keys %F5::Violation::) {
        my $viol_const = F5::Violation->can($viol_name);
        if ( $viol_name =~ /^VIOL_/ ) {
            my $viol_index = $viol_const->();
            $viol_names->{$viol_index} = $viol_name;
        }
    }

    return $viol_names;
}

#----------------------------------------------------------------------------------------


