#ifndef _XMLPARSER_CALLBACKS_H_
#define _XMLPARSER_CALLBACKS_H_

#include "encoding_types.h"
#include "logger_consts.h"
#include "xmlerrors_desc.h"

#ifdef __cplusplus
extern "C" {
#endif

/*
    callback structures
*/

typedef int (*pfn_invalidDocumentReportCallback) (void *, //parser handle
                          xml_parser_error_info error,
                          xml_error_location location,
                          xml_parser_error_info_ex error_ex,
                          void * userptr);
typedef int (*pfn_internalErrorReportCallback) (void *, //parser handle
                        xml_parser_error_info error,
                        void * userptr);

typedef int (*pfn_alertReportCallback) (void *, //parser handle
                    xml_parser_error_info error,
                    xml_error_conf_property property,
                    void * userptr);

typedef int (*pfn_checkNegsigCallback) (void *, //parser handle
                    char * data,
                    int data_len,
                    char * raw_data,
                    int raw_data_len,
                    unsigned int hash_value,
                    int location,
                    void * userptr,
                    unsigned int is_sensitive,
                    unsigned int depth);

typedef int (*pfn_usernameTokenCallback) (void *, //parser handle
                    char * user_name,
                    int user_name_len,
                    char * user_password,
                    int user_password_len,                          
                    void * userptr
                    );

typedef int (*pfn_soapMethodCallback) (void *,  //parser handle
                       char * namespace_name,
                       int namespace_len,
                       char * data,
                       int data_len,
                       void * userptr,
                       encoding_type_t enc_type);
typedef int (*pfn_sensitiveDataCallback) (void *, //parser handle
                    unsigned int start_data_index,
                    unsigned int data_length,
                    void * userptr);

typedef int (*pfn_wssErrorCallback) (void *, 
                                     xml_parser_error_info, 
                                     xml_error_location, 
                                     void *userptr);

typedef int (*pfn_debugCallback) (MODULE_NAME module_id, 
                                  LOG_LEVEL log_level,
                                  const char * data);


typedef struct xml_parser_callbacks_t {
    pfn_invalidDocumentReportCallback       invalidDocumentReportCallback;
    pfn_internalErrorReportCallback         internalErrorReportCallback;
    pfn_alertReportCallback                 alertReportCallback;
    pfn_checkNegsigCallback                 checkNegsigCallback;
    pfn_soapMethodCallback                  soapMethodCallback;
    pfn_sensitiveDataCallback               sensitiveDataCallback;
    void * user_ptr;
    void * thread_data;
} xml_parser_callbacks_t;

typedef struct _wss_parser_callbacks_t {
    pfn_wssErrorCallback       wssErrorCallback;
    pfn_alertReportCallback    alertReportCallback;
    pfn_usernameTokenCallback  usernameTokenCallback;
    void * user_ptr;
} wss_parser_callbacks_t;

typedef struct _json_parser_callbacks_t {
    pfn_invalidDocumentReportCallback       malformedDocument;
    pfn_alertReportCallback                 defenseViolation;
    pfn_checkNegsigCallback                 negativeSignatures;
    pfn_sensitiveDataCallback               sensitiveDataCallback;
    void * user_ptr;
} json_parser_callbacks_t;

xml_error_location set_error_location(unsigned char* buf, int index, int len);

#define THROW_MALFORMED_DOC_CALLBACK(_pd, _code, _level, _error_location)\
{\
    xml_parser_error_info parser_error = {_code, _level, XMLPARSER_ERROR_TYPE_DOCUMENT_MALFORMED, PARSER_ERROR_MALFORMED_DOC};\
    xml_parser_error_info_ex error_ex = {xml_parser_error_desc[_code]};\
    if (_pd->callbacks->invalidDocumentReportCallback) {\
        switch(_pd->callbacks->invalidDocumentReportCallback (_pd, (parser_error), (_error_location), (error_ex), _pd->callbacks->user_ptr)) {\
        case XML_PARSE_ABORT:                       \
            return XML_PARSER_STATUS_CALLBACK_ABORT;\
        default:                                    \
            break;                                  \
      }                                             \
    }                                               \
}

#define THROW_INVALID_DOC_CALLBACK(_callbacks, _pd, _code, _level, _type, _error, _error_location, _error_ex)\
{\
    xml_parser_error_info parser_error = {_code, _level, _type, _error};\
    xml_parser_error_info_ex error_ex = {_error_ex};\
    if (_callbacks->invalidDocumentReportCallback) {\
        switch(_callbacks->invalidDocumentReportCallback (_pd, (parser_error), (_error_location), (error_ex), _callbacks->user_ptr)) {\
        case XML_PARSE_ABORT:                       \
            return XML_PARSER_STATUS_CALLBACK_ABORT;\
        default:                                    \
            break;                                  \
      }                                             \
    }                                               \
}

#define THROW_ALERT_REPORT_CALLBACK(_callbacks, _pd, _error, _property)\
{\
    if (_callbacks->alertReportCallback) {\
        _callbacks->alertReportCallback(_pd, (_error), (_property), _callbacks->user_ptr);                                  \
    }\
}

#define THROW_DEFENSE_ALERT_REPORT_CALLBACK(_callbacks, _pd, _error, _property)\
{\
    if (_callbacks->alertReportCallback) {\
        switch(_callbacks->alertReportCallback(_pd, (_error), (_property), _callbacks->user_ptr)) {\
        case XML_PARSE_ABORT:                       \
            return XML_DEFENSE_PARSER_ERROR_DEFENCE_EXIT;\
        default:                                    \
            break;                                  \
        }\
    }\
}

  
#define THROW_INTERNAL_ERROR_CALLBACK(_callbacks, _pd, _error)\
{\
    xml_parser_error_info parser_error = {XML_PARSER_ERROR_INTERNAL,\
                                          XMLPARSER_ERROR_LEVEL_FATAL,\
                                          XMLPARSER_ERROR_TYPE_INTERNAL,\
                                          _error};\
    if (_callbacks->internalErrorReportCallback) {\
      switch(_callbacks->internalErrorReportCallback (_pd, (parser_error), _callbacks->user_ptr)) { \
      case XML_PARSE_ABORT:                       \
          return XML_PARSER_STATUS_CALLBACK_ABORT;\
      default:                                    \
          break;                                  \
      }                                           \
    }\
  }

#define THROW_CHECK_NEGSIG_PI_CALLBACK(_pd) \
    { \
        if (_pd->callbacks->checkNegsigCallback && \
            _pd->thread_data->pi_value_buf_len) {\
            switch(_pd->callbacks->checkNegsigCallback (_pd, \
                                                        (char *)(_pd->thread_data->pi_value_buf), \
                                                        (_pd->thread_data->pi_value_buf_len), \
                                                        NULL, \
                                                        0, \
                                                        _pd->thread_data->pi_hash_value, \
                                                        XMLPARSER_LOCATION_PI_VALUE, \
                                                        _pd->callbacks->user_ptr, \
                                                        0, \
                                                        0)) {\
                case XML_PARSE_ABORT:\
                    return XML_PARSER_STATUS_CALLBACK_ABORT;\
                default:\
                    break;\
            }\
            _pd->thread_data->pi_value_buf_len = 0; \
        } \
    }

#define THROW_CHECK_NEGSIG_COMMENT_CALLBACK(_pd) \
    { \
        if (_pd->callbacks->checkNegsigCallback && \
            _pd->thread_data->comment_value_buf_len) {\
            switch(_pd->callbacks->checkNegsigCallback (_pd, \
                                                        (char *)(_pd->thread_data->comment_value_buf), \
                                                        (_pd->thread_data->comment_value_buf_len), \
                                                        NULL,\
                                                        0,\
                                                        _pd->thread_data->comment_hash_value, \
                                                        XMLPARSER_LOCATION_COMMENT_VALUE, \
                                                        _pd->callbacks->user_ptr, \
                                                        0, \
                                                        0)) {\
            case XML_PARSE_ABORT:\
                return XML_PARSER_STATUS_CALLBACK_ABORT;\
            default:\
                break;\
            }\
            _pd->thread_data->comment_value_buf_len = 0; \
        }\
    }

#define THROW_CHECK_NEGSIG_ELEMENT_VALUE_CALLBACK(_pd, _sens) \
    {\
        if (_pd->callbacks->checkNegsigCallback && \
            _pd->thread_data->value_buf_len ) {\
            switch(_pd->callbacks->checkNegsigCallback (_pd, \
                                                        (char *)(_pd->thread_data->value_buf), \
                                                        (_pd->thread_data->value_buf_len), \
                                                        NULL,\
                                                        0,\
                                                        _pd->thread_data->value_hash_value, \
                                                        XMLPARSER_LOCATION_ELEMENT_VALUE, \
                                                        _pd->callbacks->user_ptr, \
                                                        _sens , \
                                                        0 )) {\
                case XML_PARSE_ABORT:\
                    return XML_PARSER_STATUS_CALLBACK_ABORT;\
                default:\
                    break;\
            }\
            _pd->thread_data->value_buf_len = 0; \
        }\
    }
    
#define THROW_CHECK_NEGSIG_ATTRIBUTE_VALUE_CALLBACK(_pd, _idx, _sens) \
    {\
        attribute_t* att = ATT_BY_INDEX(_idx);\
        if (_pd->callbacks->checkNegsigCallback && \
            att->value_buf_end) { \
            switch(_pd->callbacks->checkNegsigCallback (_pd, \
                                                        (char *)(&_pd->thread_data->value_buf[att->value_buf_start]), \
                                                        (unsigned int)(att->value_buf_end - att->value_buf_start), \
                                                        NULL,\
                                                        0,\
                                                        att->value_hash,\
                                                        XMLPARSER_LOCATION_ATTRIBUTE_VALUE, \
                                                        _pd->callbacks->user_ptr,\
                                                        _sens,\
                                                        0)) { \
                case XML_PARSE_ABORT: \
                    return XML_PARSER_STATUS_CALLBACK_ABORT; \
                default: \
                    break; \
            } \
        }\
    }
   

#define THROW_SOAP_METHOD_CALLBACK(_pd, _namespace_start, _namespace_len, _val_start, _len) \
    if (_pd->callbacks->soapMethodCallback) {\
        switch(_pd->callbacks->soapMethodCallback (_pd, _namespace_start, _namespace_len, (char *)_val_start, _len, _pd->callbacks->user_ptr, _pd->encoding_type)) {\
            case XML_PARSE_ABORT:\
                return XML_PARSER_STATUS_CALLBACK_ABORT;\
            default:\
                break;\
        }\
    }
#define THROW_SENSITIVE_DATA_CALLBACK(_pd, _start, _len) \
    if (_pd->callbacks->sensitiveDataCallback) {\
        _pd->has_icap_data = 1;/* set icap check, this is a workaround for bug 559541*/\
        switch(_pd->callbacks->sensitiveDataCallback (_pd, _start, _len, _pd->callbacks->user_ptr)) {\
            case XML_PARSE_ABORT:\
                return XML_PARSER_STATUS_CALLBACK_ABORT;\
            default:\
                break;\
        }\
    }

#define THROW_WSS_ERROR_CALLBACK(_callbacks, _pd, _code, _level, _error, _location)\
{\
    xml_parser_error_info parser_error = {_code, _level, 0, _error};\
    if (_callbacks->wssErrorCallback) {\
        switch(_callbacks->wssErrorCallback(_pd, parser_error, (_location), _callbacks->user_ptr)) {\
        case XML_PARSE_ABORT:                       \
            return WSS_FAILURE;                     \
        default:                                    \
            break;                                  \
        }\
    }\
}

///// JSON CALLBACKS MACROS
#define THROW_JSON_CHECK_NEGSIG_VALUE_CALLBACK(_pd, _data, _len, _raw_data, _raw_len, _hash, _vt, _s, _depth) \
    {\
        if ((pd->max_value_length < (_raw_len)) && (_vt != JSON_LOCATION_NAME))\
        {\
               _pd->max_value_length = (_raw_len);\
        }\
        if (_pd->callbacks->negativeSignatures && \
            _raw_len && _raw_data != NULL) {\
            switch(_pd->callbacks->negativeSignatures ( _pd, \
                                                       (_data), \
                                                       (_len), \
                                                       (_raw_data), \
                                                       (_raw_len), \
                                                       (_hash), \
                                                        _vt, \
                                                        _pd->callbacks->user_ptr, \
                                                        _s, \
                                                        _depth)) { \
                case JSON_PARSE_ABORT:\
                    return JSON_PARSER_SUCCESS;\
                default:\
                    break;\
            }\
        }\
        pd->value_buf_len = 0; \
    }

#define THROW_JSON_DEFENSE_REPORT_CALLBACK(_pd, _error, _property)\
{\
    if (_pd->callbacks->defenseViolation) {\
        switch(_pd->callbacks->defenseViolation(_pd,    \
                                                (_error),   \
                                                (_property),    \
                                                _pd->callbacks->user_ptr)){ \
            case JSON_PARSE_ABORT:\
                if ((_pd)->max_buf > (_pd)->buf) {\
                    THROW_JSON_CHECK_NEGSIG_VALUE_CALLBACK(_pd, \
                                                           (char *)_pd->buf, \
                                                           (unsigned int)(_pd->max_buf - _pd->buf), \
                                                           (char *)_pd->buf, \
                                                           (unsigned int)(_pd->max_buf - _pd->buf), \
                                                           _pd->current_string_hash, \
                                                           JSON_LOCATION_VALUE, \
                                                           CUR_JSON_ITEM(_pd)->is_sensitive, \
                                                           GET_JSON_STACK_DEPTH(_pd)); \
                }\
                return JSON_PARSER_SUCCESS;\
            default:\
                break;\
        }\
    }\
}


#define THROW_JSON_MALFORMED_CALLBACK(_pd, _code, _level, _error, _error_location)\
{\
    if (_pd->callbacks->malformedDocument) { \
        _pd->block_parsing = 1;\
        xml_parser_error_info parser_error = {_code, _level, JSONPARSER_ERROR_TYPE_DOCUMENT_MALFORMED, _error};\
        xml_parser_error_info_ex error_ex = {(char *)json_parser_error_desc[_code]};\
        switch(_pd->callbacks->malformedDocument (_pd, (parser_error), (_error_location), (error_ex), _pd->callbacks->user_ptr)) {\
        case XML_PARSE_ABORT:          \
            return JSON_PARSER_SUCCESS;\
        default:                       \
            break;                     \
        }                              \
    }                                  \
}


#define THROW_JSON_SENSITIVE_DATA_CALLBACK(_pd, _start_doc, _start_p, _end_p) \
    if (_pd->callbacks->sensitiveDataCallback) {\
        _pd->callbacks->sensitiveDataCallback (_pd, (unsigned int)((unsigned long)_start_p - (unsigned long)_start_doc), (unsigned int)((unsigned long)_end_p - (unsigned long)_start_p), _pd->callbacks->user_ptr);\
        CUR_JSON_ITEM(_pd)->is_sensitive = (PREV_JSON_ITEM(pd) != NULL) ? PREV_JSON_ITEM(pd)->is_sensitive : 0;\
    }

///////////////////////////////////////////////////////////
// location: current parser position
///////////////////////////////////////////////////////////
#define XML_LOCATION_CUR_PARSER_POS(_start_doc, _prev_char) \
        (unsigned char *)_prev_char, (unsigned int)(_prev_char - (unsigned long)_start_doc), 1
		
#define XML_LOCATION_CUR_TAG() \
        (unsigned char *)pd->tag_start, (unsigned int)(pd->tag_start - (unsigned long)chunk_buf), (unsigned int)(pd->last_buf - pd->tag_start)


///////////////////////////////////////////////////////////
// location: current start element name
///////////////////////////////////////////////////////////
#define XML_LOCATION_CUR_START_ELT_NAME(_pd, _start_doc) \
        (unsigned char *)(XML_PARSER_ITEM(_pd)->element_prefix_start), \
        (unsigned int)(XML_PARSER_ITEM(_pd)->element_prefix_start - (unsigned long)_start_doc), \
        (unsigned int)(XML_PARSER_ITEM(_pd)->element_name_end - XML_PARSER_ITEM(_pd)->element_prefix_start)

///////////////////////////////////////////////////////////
// location: current start attribute name
///////////////////////////////////////////////////////////
#define XML_LOCATION_CUR_START_ATT_NAME(_pd, _start_doc) \
        (unsigned char *)_pd->attribute_prefix_start, \
        (unsigned int)(_pd->attribute_prefix_start - (unsigned long)_start_doc), \
        (unsigned int)(_pd->attribute_name_end - _pd->attribute_prefix_start)


  
#ifdef __cplusplus
}
#endif

#endif //_XMLPARSER_CALLBACKS_H_

