package F5::ASM::COAPI;
use strict;
use warnings;

use F5::BigipUtils;
use F5::DbUtils;
use F5::ASM::Tmsh;
use YAML::XS;

eval {
    if (! F5::BigipUtils::is_install_mode()) {
        require F5::COAPI;
    }
};

my $use_file_store = 0;

our @ISA;
if (eval {require F5::COAPI}) {
    @ISA  = qw(F5::COAPI);
}
else {
    require F5::ASM::COAPI::File;
    @ISA = qw(F5::ASM::COAPI::File);
    $use_file_store = 1;
}

my $login_name = 'root';

sub set_login_name {
    my($class, $new_login_name) = @_;
    $login_name = $new_login_name;
}

sub Login {
    my($new_login_name) = @_;

    return if F5::BigipUtils::is_install_mode();

    $login_name = $new_login_name;
    my $rv;
    eval {
        $rv = $use_file_store ? F5::ASM::COAPI::File::Login($new_login_name)
                              : F5::COAPI::Login($new_login_name);
    };
    if ($@) {
        return;
    }
    return $rv;
}

sub bypass_validation {
    if (!$login_name || substr($login_name,0,1) ne '%') {
        Login('%asmconfig'); #Special name that will still trigger CMI sync events (tm_daemon/mcp/framework/request_group.cpp)
    }
}


sub TransactionalSave {
    my($transaction) = @_;

    return if F5::BigipUtils::is_install_mode();

    if (F5::Cluster::Utils::is_clustering_enabled() && ! F5::Cluster::Utils::is_master_blade()) {
        # Silently ignore saves to mcp on secondary blade - otherwise, mcp issues an error
        return;
    }

    my $rv = $use_file_store ? F5::ASM::COAPI::File::TransactionalSave($transaction)
                             : F5::COAPI::TransactionalSave($transaction);
    if (!$rv) {
        die $F5::COAPI::error . "\n";
    }
}

sub is_fflag_enabled {
    my ($flag, %params) = @_;

    return if F5::BigipUtils::is_install_mode();

    my $rv = 0;
    if (my ($result) = @{F5::ASM::COAPI::Query->new('fflag_get_request', { name => $flag }, do_cache => 1)}) {
        $rv = $result->{value} ? 1 : 0;  # Normalize integral fflags to a boolean result
    }
    return $rv;
}

#----------------------------------------------------------------------

sub SaveAll {
    my(%params) = @_;
    my $logger = $params{logger};

    if ((F5::Cluster::Utils::is_clustering_enabled() && ! F5::Cluster::Utils::is_master_blade()) || F5::BigipUtils::is_docker()) {
        # Silently ignore saves to mcp on secondary blade - otherwise, mcp issues an error
        return;
    }

    # Save sys config only if mcp was loaded successfully
    my $mcp_state = F5::ASM::Tmsh::Query->new(cmd => [ qw(tmsh -a show sys mcp-state field-fmt) ],
                                              parse_response => 1);

    my $mcp_state_last_load = 'unknown';
    if ($mcp_state && $mcp_state->{completed_successfully} && 
        $mcp_state->{response}->{'mcpd-state'} &&
        $mcp_state->{response}->{'mcpd-state'}->{'last-load'}) {

        $mcp_state_last_load = $mcp_state->{response}->{'mcpd-state'}->{'last-load'};
    }

    if ($mcp_state_last_load =~ /\w+-config-load-succeed/) { # at upgrade 'full-config-load-succeed'
        F5::ASM::Tmsh::Query->new(cmd => [ qw(tmsh save sys config partitions all) ],
                                              logger => $logger,
                                              parse_response => 1);
        $logger->info('MCP state was persisted') if $logger;
    }
    else {
        $logger->info("Skipped persisting MCP state (MCP was not ready, mcp state = $mcp_state_last_load)") if $logger;
    }
}

#======================================================================

package F5::ASM::COAPI::Query;

our $APPLY_FILTER = 1;

our @ISA;
if (eval {require F5::COAPI}) {
    @ISA  = qw(F5::COAPI::Query);
}
else {
    @ISA = qw(F5::ASM::COAPI::File::Query);
    $APPLY_FILTER = 0; #Trust F5::ASM::COAPI::File
}

my $RETRIES = 2;
my $UNLIKELY_DELIMITER = '###UNLIKELY_DELIMITER###';

my $table_cache = {};

sub new {
    my ($self, $table, $filter, %params) = @_;

    return if F5::BigipUtils::is_install_mode();

    my $retrieve_as_hash = $params{retrieve_as_hash};
    my $hash_key_field_name;
    if (ref($retrieve_as_hash) eq 'HASH' && $retrieve_as_hash->{key_field_name}) {
        $hash_key_field_name = $retrieve_as_hash->{key_field_name};
    }

    my $counter = 0;
    my $do_cache = $params{do_cache};
    my $direct_coapi_filter = $filter;
    if ($params{direct_coapi_filter_key}) {
        # Nasty hack: In certain cases, MCP query performs very poorly with multiple search criteria.
        # Thus, we will need to pass through only the specified key field, and filter for all others afterwards.
        # (However, it's not clear that this will filter correctly in all cases, and thus this is not the default
        # behavior.)
        $direct_coapi_filter = { $params{direct_coapi_filter_key} => $filter->{$params{direct_coapi_filter_key}} };
    }

    while ($counter++ <= $RETRIES) {
        my $query;
        eval {
            my $key = 'data';
            if ($do_cache) {
                my $cache_filter = undef;
                if ($table eq 'fflag_get_request' && $filter && ref($filter)) {
                   # Special table for feature flags, only returns when flag is specified
                   $key = $filter->{name};
                   $cache_filter = $direct_coapi_filter;
                   $filter = undef;
                }

                if (!$table_cache->{$table}->{$key}) {
                    $table_cache->{$table}->{$key} = $query = $self->SUPER::new($table, $cache_filter);
                }
                $query = $table_cache->{$table}->{$key};
            }
            else {
                $query = $self->SUPER::new($table, $direct_coapi_filter);
            }
        };
        if ( $@ && ($F5::COAPI::error =~ /The requested .* \(.*\) was not found\./) ){
            return [];
        }

        if ($query) {
            if ($APPLY_FILTER && $filter && ref($filter)) {
                if ($do_cache) {
                    my @filter_fields = sort keys %$filter;
                    my $filter_key = join ("___", @filter_fields);
                    if (!$table_cache->{$table}->{filters}->{$filter_key}) {
                        for my $rec (@$query) {
                            my $rec_key = join ($UNLIKELY_DELIMITER, map {$rec->{$_}} @filter_fields);
                            push (@{$table_cache->{$table}->{filters}->{$filter_key}->{$rec_key}}, $rec);
                        }
                    }
                    my $filter_value =  join ($UNLIKELY_DELIMITER, map {$filter->{$_}} @filter_fields);
                    $query = $table_cache->{$table}->{filters}->{$filter_key}->{$filter_value};
                }
                else {
                    $query = [ grep { filter_match(filter => $filter, rec => $_) } @$query ];
                }
            }
            $query = [ map { $_ = bless $_, 'F5::ASM::COAPI::Object' } @$query ];

            if ($hash_key_field_name) {
                my %hashed_query = map { ($_->{$hash_key_field_name} => $_) } @$query;
                return %hashed_query;
            }

            return $query;
        }

        sleep 1;
    }
    return [];
}

sub clear_cache {
    my ($self, %params) = @_;
    if ($params{table}) {
        delete $table_cache->{$params{table}};
    }
    else {
        $table_cache = {};
    }
}

sub filter_match {
    my(%params) = @_;
    my $filter = $params{filter};
    my $rec = $params{rec};

    for my $k (keys %$filter) {
        if ($rec->{$k} ne $filter->{$k}) {
            return;
        }
    }
    return 1;
}

#======================================================================

package F5::ASM::COAPI::Object;

our @ISA;
if (eval {require F5::COAPI}) {
    @ISA  = qw(F5::COAPI::Object);
}
else {
    @ISA = qw(F5::ASM::COAPI::File::Object);
}

use F5::Cluster::Utils;

sub new {
    my($class, @params) = @_;

    return if F5::BigipUtils::is_install_mode();

    if ((F5::Cluster::Utils::is_clustering_enabled() && ! F5::Cluster::Utils::is_master_blade()) ) {
        # Silently return a dummy hash on secondary blade - otherwise, mcp issues an error
        return bless({}, $class);
    }

    $class->SUPER::new(@params);
}

sub save {
    my($self, @params) = @_;

    return if F5::BigipUtils::is_install_mode();

    if ((F5::Cluster::Utils::is_clustering_enabled() && ! F5::Cluster::Utils::is_master_blade())) {
        # Silently ignore saves to mcp on secondary blade - otherwise, mcp issues an error
        return;
    }

    $self->SUPER::save(@params) || die $F5::COAPI::error . "\n";    
}

sub delete {
    my($self, @params) = @_;

    return if F5::BigipUtils::is_install_mode();

    if ((F5::Cluster::Utils::is_clustering_enabled() && ! F5::Cluster::Utils::is_master_blade())) {
        # Silently ignore saves to mcp on secondary blade - otherwise, mcp issues an error
        return;
    }

    $self->SUPER::delete(@params) || die $F5::COAPI::error . "\n";
}

sub delete_on_save {
    my($self, @params) = @_;

    return if F5::BigipUtils::is_install_mode();

    if ((F5::Cluster::Utils::is_clustering_enabled() && ! F5::Cluster::Utils::is_master_blade())) {
        # Silently ignore saves to mcp on secondary blade - otherwise, mcp issues an error
        return;
    }

    # The COAPI delete_on_save() function does not return any value
    $self->SUPER::delete_on_save(@params);
}

# this is a setter that will set only defined values
# setting undefined values into a coapi object throws an error
sub set_value {
    my ($self, $field, $value) = @_;

    $self->{$field} = $value if defined $value;
}

our $DBUTILS_CALLBACK_NAME = 'F5::ASM::COAPI::transactional_action';

sub transactional_action {
    my ($self, %params) = @_;
    my $action = $params{action};
    my $dbh = $params{dbh};
    my $success_action = $params{success_action};
    my $ignore_error = $params{ignore_error};
    my $logger = $params{logger};

    if ($action eq 'delete') {
        $self->delete_on_save();
    }

    my $callback_rec;
    my $mcp_objects_to_commit = [];
    my $success_actions = [];

    if ($dbh && ! $dbh->{AutoCommit}) {
        $callback_rec = F5::DbUtils::get_callback_by_name(dbh => $dbh,
                                                          name => $DBUTILS_CALLBACK_NAME,
                                                         );
        # On all transactional actions after the first, a callback will have already been created.
        # If so, callback params will use references to the params already associated with that
        # callback via closure scope.
        if ($callback_rec) {
            $mcp_objects_to_commit = $callback_rec->{callback_params}->{mcp_objects_to_commit};
            $success_actions = $callback_rec->{callback_params}->{success_actions};
        }
    }

    push(@$mcp_objects_to_commit, $self);
    if ($success_action) {
        push(@$success_actions, $success_action);
    }

    my $callback = sub {
                         eval {
                             F5::ASM::COAPI::TransactionalSave($mcp_objects_to_commit);
                         };
                         if (my $err = $@) {
                             my $err_msg = "MCP Validation error - $err";
                             if ( $ignore_error ) {
                                 $logger->debug("Ignoring error: $err_msg") if $logger;
                             }
                             else {
                                 if ($dbh && !$dbh->{AutoCommit}) {
                                     $dbh->rollback;
                                 }
                                 die "$err_msg\n";
                             }
                         }
                         else {
                             for $success_action (@$success_actions) {
                                 if ($success_action->{func}) {
                                     $success_action->{func}->();
                                 }
                                 elsif ( $success_action->{action} && $success_action->{obj}) {
                                     my $obj_action = $success_action->{action};
                                     $success_action->{obj}->$obj_action;
                                 }
                             }
                         }
    };

    if (!$dbh || $dbh->{AutoCommit}) {
        $callback->();
    }
    else {
        if (! $callback_rec) {
            # Add a DbUtils callback only on the first time
            F5::DbUtils::add_callback(dbh => $dbh,
                                      callback => $callback,
                                      name => $DBUTILS_CALLBACK_NAME,
                                      callback_params => {
                                                          mcp_objects_to_commit => $mcp_objects_to_commit,
                                                          success_actions => $success_actions
                                                         },
                                     );
        }
    }
}


#======================================================================

1;
