package F5::ASMConfig::Cache;

use strict;
use warnings;

use F5::ObjUtils;
use F5::ModuleLoader;
use F5::Db::Single;
use F5::ASM::COAPI;
use F5::BigipUtils;

my $policy_cache;
my $pb_config_cache;
my $rest_external_struct_cache;
my %viol_index_2_violation_cache;
my %encoding_name_2_type_cache;
my $enable_enforcement_by_rating;
my $sig_tag_cache;
#----------------------------------------------------------------------

sub get_cached_violations {
    my(%params) = @_;
    my $dbh = $params{dbh};

    # fill { viol_index => violation } cache once
    if (!%viol_index_2_violation_cache && $dbh) {
        my @violations = F5::ASMConfig::Entity::Violation->retrieve_all(dbh => $dbh);
        %viol_index_2_violation_cache = map {($_->{viol_index} => $_)} @violations;
    }

    return \%viol_index_2_violation_cache;
}

#----------------------------------------------------------------------

sub get_cached_encoding_type {
    my(%params) = @_;

    my $dbh = $params{dbh};
    my $encoding_name = $params{encoding_name};
    my $logger = $params{logger};

    if (!$encoding_name) {
        my $caller = (caller(1))[3];
        my $err = "The caller:[$caller] did not pass in a value for 'encoding_name' to retrieve the 'encoding_type' for -- aborting.";
        $logger->error($err) if $logger;
        die $err;
    }

    # fill encoding { name => type } cache once
    if (!%encoding_name_2_type_cache && $dbh) {
        my @encodings = F5::ASMConfig::Entity::Encoding->retrieve_all(dbh => $dbh,);
        %encoding_name_2_type_cache = map { ( $_->{encoding_name} => $_->{encoding_type}) } @encodings;
    }

    # we have no idea what is the encoding_type for this encoding_name
    #     default encoding type - multibyte
    if( !$encoding_name_2_type_cache{$encoding_name} ) {
        $encoding_name_2_type_cache{$encoding_name} = F5::ASMConfig::EncodingType::ENCODING_TYPE_MULTI_BYTE();
    }
    
    return $encoding_name_2_type_cache{$encoding_name};
}

#----------------------------------------------------------------------

sub has_policy_id {
    my $policy = shift;

    return ($policy && defined $policy->{policy_id} && $policy->{policy_id} > 0) ? 1 : 0 ;
}

#----------------------------------------------------------------------

sub get_cached_policy {
    my(%params) = @_;

    my $dbh = $params{dbh};
    my $policy = $params{policy};
    my $context = $params{context} || 'default';

    return if (!$policy);

    if (! has_policy_id($policy)) {
        my $found_policy;
        if (!$policy->{policy_name} && $policy->{leaf_name}) {
            F5::ASMConfig::Entity::Policy->policy_name_fixup(%params, object => $policy);
        }

        if ($policy->{policy_name}) {
            ($found_policy) = grep { $_->{policy_name} eq $policy->{policy_name} } values %{$policy_cache->{$context}};
        }
        elsif ($policy->{account_id}) {
            ($found_policy) = grep { $_->{account_id} eq $policy->{account_id} } values %{$policy_cache->{$context}};
        }
        elsif ($policy->{policy_name_crc}) {
            ($found_policy) = grep { $_->{policy_name_crc} eq $policy->{policy_name_crc} } values %{$policy_cache->{$context}};
        }
        elsif ($policy->{rest_uuid}) {
            ($found_policy) = grep { $_->{rest_uuid} eq $policy->{rest_uuid} } values %{$policy_cache->{$context}};
        }
        if ($found_policy) {
            $policy = $found_policy;
        }
    }

    if (has_policy_id($policy) && $policy_cache->{$context}->{$policy->{policy_id}} && ! $params{force_reload}) {
        if ($dbh) {
            $policy_cache->{$context}->{$policy->{policy_id}}->{dbh} = $dbh;
        }
        return $policy_cache->{$context}->{$policy->{policy_id}};
    }

    if (($dbh || $params{use_singleton_dbh}) && 
        ! $params{bypass_load} && 
        (has_policy_id($policy) || $policy->{policy_name} || $policy->{account_id} || $policy->{policy_name_crc} || $policy->{rest_uuid})) {

        if (! grep { defined $_ } values %$policy) {
            # Empty policy passed; return undef, which will match everything
            return;
        }

        # Not found in cache; load a new one
        my $original_policy = { %$policy };
        if (! $dbh) {
            $dbh = F5::Db::Single->instance;
        }
        $policy = get_policy(%params, dbh => $dbh);
        if (! $policy) {
            $policy = F5::ASMConfig::Entity::Policy->new(object => $original_policy, dbh => $dbh, call_flags => $params{call_flags});
            $policy->{_failed_to_load} = 1;
            if (! has_policy_id($policy)) {
                $policy->{policy_id} = -1;
            }
        }
        else {
            $policy->{_failed_to_load} = 0;
        }

        if (! $policy->{_failed_to_load} && has_policy_id($policy)) {
            $policy_cache->{$context}->{$policy->{policy_id}} = $policy;
        }

        return $policy;
    }
    return;
}

#----------------------------------------------------------------------

sub get_policy {
    my(%params) = @_;
    my $full_policy = F5::ASMConfig::Entity::Policy->new(dbh => $params{dbh},
                                                         minimal_load => 1,
                                                         get_criteria => $params{policy},
                                                         call_flags => $params{call_flags},
                                                        );
    return $full_policy;
}

#----------------------------------------------------------------------

sub clear_cached_policy {
    my(%params) = @_;
    my $context = $params{context} || 'default';
    my $policy = get_cached_policy(%params, bypass_load => 1);
    if (has_policy_id($policy)) {
        delete $policy_cache->{$context}->{$policy->{policy_id}};
    }
}

#----------------------------------------------------------------------

sub get_cached_pb_config {
    my (%params) = @_;
    my $policy = $params{policy};
    if (! $policy->{policy_id}) {
        $policy = get_cached_policy(%params);
    }
    my $policy_id = $policy->{policy_id} || "default";
    if (!$pb_config_cache->{$policy_id}) {
        my %get_params = (dbh => $params{dbh});
        if ($policy) {
            $get_params{policy} = $policy;
            $get_params{get_criteria} = {};
        }
        else {
            $get_params{preload_default} = 1;
        }

        my $pb_config = F5::ASMConfig::Entity::PBConfig->new(%get_params);
        $pb_config_cache->{$policy_id} = $pb_config;
    }

    return $pb_config_cache->{$policy_id};
}

#----------------------------------------------------------------------

sub get_cached_rest_external_structs {
    my (%params) = @_;
    my $entity_type = $params{entity_type};
    my $field_name = $params{field_name};
    my $field_value_list = $params{field_value_list};
    my $dbh = $params{dbh};
    my $call_flags = $params{call_flags};
    my $policy;
    my $is_expanded = $params{is_expanded} || 0;

    if ($entity_type ne 'F5::ASMConfig::Entity::Policy') {
        #Do not consider the passed in policy if we are retreiving an actual list of policies
        $policy = $params{policy};
    }
 
    my $key = $entity_type . $field_name . "_${is_expanded}_";
    if ($policy && $entity_type->policy_id_field()) {
        $key .= $policy->{policy_id};
    }

    my $rest_reference_fields = $entity_type->rest_reference_fields();
    my $verbose_references = !defined($call_flags->{verbose_references}) || $call_flags->{verbose_references};
    my $selected_fields = $is_expanded ? $params{rest_selected_fields}
                                       : $verbose_references ? $rest_reference_fields
                                       : undef;

    my %retrieve_opts = (dbh => $dbh,
                         policy => $policy,
                         call_flags => F5::ASMConfig::CallFlags->new({rest_call => 1}),
                         populate_diff_structs => $is_expanded,
                         rest_selected_fields => $selected_fields,
                         minimal_load => $is_expanded ? 0 : 1,
                         selected_fields => $is_expanded ? undef : [],);

    my $use_simple_constructor = !$is_expanded;
    if ($use_simple_constructor && $verbose_references) {
        my $combined_field_def = {};
        for my $reference_field_def (values %$rest_reference_fields) {
           $combined_field_def = {%$reference_field_def, %$combined_field_def};
        }
        $use_simple_constructor = !$entity_type->requires_mapping(direction => 'to_entity', field_def => $combined_field_def);
    }

    if ($use_simple_constructor) {
        #Use a simplified constructor when the object is not expanded. We only need the rest_uuid of the object and the foreign id.
        $retrieve_opts{object_constructor} = sub {
                                           my(%params) = @_;
                                           my $rec = $params{rec};

                                           my $obj = $entity_type->new(dbh => $dbh,
                                                                       policy => $policy,
                                                                       object => { rest_uuid => $rec->{rest_uuid},
                                                                                   $field_name => $rec->{$field_name},},
                                                                       );
                                           if ($verbose_references) {
                                               for my $reference_field (keys %$rest_reference_fields) {
                                                   $obj->{$reference_field} = $rec->{$reference_field};
                                               }
                                           }
                                           return $obj;
                                         };
    }

    my @objects;
    my @missing_values;
    my $full_retrieve_needed = (!$field_value_list || !@$field_value_list);

    if ($rest_external_struct_cache->{$key}) {
        if ($full_retrieve_needed && $rest_external_struct_cache->{'FULL_CACHE'}->{$key}) {
            @objects = values %{$rest_external_struct_cache->{$key}};
            return \@objects;
        }
        elsif ($field_value_list && @$field_value_list) {  
            for my $value (@$field_value_list) {
                if ($rest_external_struct_cache->{$key}->{$value}) {
                    push (@objects, $rest_external_struct_cache->{$key}->{$value});
                }
                else {
                    push (@missing_values, $value);
                }
            }
        }
    }
    else {
        @missing_values = @$field_value_list;
    }

    if ($full_retrieve_needed) {
        @objects = $entity_type->retrieve_all(%retrieve_opts);
        my %cache = map { $_->{$field_name} => $_ } @objects;
        $rest_external_struct_cache->{$key} = \%cache;
        $rest_external_struct_cache->{'FULL_CACHE'}->{$key} = 1;
        return \@objects;
    }

    if ( @missing_values && !$rest_external_struct_cache->{'FULL_CACHE'}->{$key}) {
        #Don't bother searching for missing values if we already got the full cache
        my $search_filters = $entity_type->rest_search_filters();
        push (@$search_filters, F5::ASMConfig::SearchFilter->new(field_name => $field_name,
                                                                 field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                                 field_value_list => \@missing_values));
        my @new_objects = $entity_type->retrieve_all(%retrieve_opts,
                                                     search_filters => $search_filters);
        push (@objects, @new_objects);
        for my $object (@new_objects) {
            $rest_external_struct_cache->{$key}->{$object->{$field_name}} = $object; 
        }
    }

    return \@objects;
}

#----------------------------------------------------------------------

sub get_cached_enable_enforcement_by_rating {
    my(%params) = @_;
    my $dbh = $params{dbh};

    if( !defined $enable_enforcement_by_rating ){
     ($enable_enforcement_by_rating) = $dbh->selectrow_array(<<EndQuery, undef, "enable_enforcement_by_rating");
SELECT current_value
    FROM DCC.INTERNAL_CONFIG_ITEMS
WHERE field_name = ?
EndQuery

        if ( not defined $enable_enforcement_by_rating) {
            if (F5::BigipUtils::is_feature_enabled('standalone') && F5::BigipUtils::is_feature_enabled('enable_enforcement_by_rating')) {
                $enable_enforcement_by_rating = 1;
            }
            else {
                $enable_enforcement_by_rating = 0;
            }
        }
    }

    return $enable_enforcement_by_rating;
}

#----------------------------------------------------------------------

sub soft_reset {
    $policy_cache = {};
    $pb_config_cache = {};
    $rest_external_struct_cache = {};
    $sig_tag_cache = {};
    F5::ASM::COAPI::Query->clear_cache();
    $enable_enforcement_by_rating = undef;
}

#----------------------------------------------------------------------

sub get_cached_sig_tags {
    my (%params) = @_;
    my $dbh = $params{dbh};

    if (! %{$sig_tag_cache}) {
        $sig_tag_cache = $dbh->selectall_hashref(<<EndQuery, 'sig_tag');
SELECT DISTINCT(sig_tag)
  FROM PLC.NEGSIG_SIGNATURES
EndQuery
    }
    return $sig_tag_cache;
}

#----------------------------------------------------------------------

1;
