package F5::ASMConfig::Entity::Incident;

use strict;
use warnings;

use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::Incident/;

#----------------------------------------------------------------------

# Convert list of AttackType references to 64-bit integer value for storage in database
sub fill_attack_types {
    my($self, %params) = @_;

    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $call_flags = $params{call_flags};

    my $attackTypes = $self->get_new_value( store_obj => $store_obj,
                              existing_obj => $existing_obj,
                              call_flags => $call_flags,
                              field_name => "attack_types", );

    # Prepare bits vector filled with zero bits
    my $vec = Bit::Vector->new( 64 );

    # Enable bits in vec that correspond to the passed attack types entities
    for my $attackType (@$attackTypes) {
		$vec->Bit_On( $attackType->{attack_type_id} );
    }

    $store_obj->{attack_types_mask} = $vec->to_Dec();
    delete $store_obj->{attack_types};
}

#----------------------------------------------------------------------

# Convert list of Violation references to 128-bit binary blob for storage in database
sub fill_violations {
    my($self, %params) = @_;

    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $call_flags = $params{call_flags};

    my $violations = $self->get_new_value( store_obj => $store_obj,
                              existing_obj => $existing_obj,
                              call_flags => $call_flags,
                              field_name => "violations", );

    # Prepare bits vector filled with zero bits
    my $vec = Bit::Vector->new( 128 );

    # Enable bits in vec that correspond to the passed violation entities
    for my $violation (@$violations) {
        $vec->Bit_On( $violation->{viol_index} );
    }

    # Get hex string value from $vec and set it to be stored in database
    $store_obj->{violations_mask} = $vec->to_Hex();
    delete $store_obj->{violations};
}

#----------------------------------------------------------------------

sub instantiate_nested_struct_entities {
    my($self, %params) = @_;

    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $call_flags = $params{call_flags};

    fill_attack_types($self, %params);
    fill_violations($self, %params);

    return $self->SUPER::instantiate_nested_struct_entities(%params);
}

#----------------------------------------------------------------------

my $attack_types_cache = [];

sub get_attack_types {
    my ( $self, %params ) = @_;

    my $dbh = $params{dbh};
    my $attack_types_mask = $params{field_value};

    my @attackTypes = ();
    my $vec = Bit::Vector->new_Dec( 64, $attack_types_mask );

    for my $pos ( $vec->Index_List_Read ) {
        my $attack_type_id = $pos + 1;
        my $attackType;

        if ( $attack_types_cache->[$attack_type_id] ) {
            $attackType = $attack_types_cache->[$attack_type_id];
        }
        else {
            $attackType = F5::ASMConfig::Entity::AttackType->new( dbh => $dbh, get_criteria => { attack_type_id => $attack_type_id } );
            $attack_types_cache->[$attack_type_id] = $attackType;
        }

        push( @attackTypes, $attackType );
    }

    return @attackTypes;
}

#----------------------------------------------------------------------

sub map_attack_types {
    my ( $self, %params ) = @_;

    my $attack_types_mask = $params{field_value};
    my $direction         = $params{direction};
    my $dbh               = $params{dbh};

    return unless defined($attack_types_mask);

	my @attackTypesList = ();

    if ( $direction eq 'to_entity' ) {
       @attackTypesList = get_attack_types($self, %params);
    }

    return \@attackTypesList;
}

#----------------------------------------------------------------------

my $violations_cache = [];

sub get_violations {
    my ( $self, %params ) = @_;
    my $dbh = $params{dbh};
    my $violations = $params{field_value};

    my @violation_list = ();
    my $vec = Bit::Vector->new_Hex( 128, $violations );

    for my $pos ( $vec->Index_List_Read ) {
        my $viol_index = $pos + 1;
        my $violation;

        if ( $violations_cache->[$viol_index] ) {
            $violation = $violations_cache->[$viol_index];
        }
        else {
            $violation = F5::ASMConfig::Entity::Violation->new( dbh => $dbh, get_criteria => { viol_index => $viol_index } );
            $violations_cache->[$viol_index] = $violation;
        }

        push( @violation_list, $violation );
    }

    return @violation_list;
}

#----------------------------------------------------------------------

sub map_violations {
    my ( $self, %params ) = @_;

    my $violations = $params{field_value};
    my $direction  = $params{direction};
    my $dbh        = $params{dbh};

    return unless defined($violations);

    my @violation_list = ();

    if ( $direction eq 'to_entity' ) {
       @violation_list = get_violations($self, %params);
    }

    return \@violation_list;
}

#----------------------------------------------------------------------

sub get_rest_uuid {
    my ($self, %params) = @_;
    my $object = $params{object} || $self;
    my $rest_uuid = $object->{rest_uuid} || $object->{incident_id} || $self->{rest_uuid} || $self->{incident_id};
    return $rest_uuid;
}

#----------------------------------------------------------------------

sub self_link {
    my ($self, %params) = @_;
    my $link = $params{base_link} || '';
    $link .= '/' . $self->element_rest_path_name() . "/" . $self->get_rest_uuid;
    $link = $self->add_ver_to_self_link(%params,
                                        'link' => $link);
    return $link;
}

#--------------------------------------------------------------------

sub suppress_events {
    my ($self, %params) = @_;

    if (($params{event_type} && ($params{event_type} == F5::ASMConfig::EventType::ADD)) ||
        ($params{event_type} && ($params{event_type} == F5::ASMConfig::EventType::UPDATE) && $params{store_obj} && $params{store_obj}->{incident_status} == F5::ASMConfig::IncidentStatus::INCIDENT_CLOSED())) {
       return 0;
    }

    return 1;
}

#--------------------------------------------------------------------
1;
