package F5::ASMConfig::Entity::PBConfig;

use strict;
use warnings;

use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::Entity::PolicySignatureSet;
use F5::ASMConfig::Entity::Blocking;
use F5::ASMConfig::ASMConfig;
use F5::BigipVersionUtils;
use F5::Violation;
use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::PBConfig/;

#----------------------------------------------------------------------

our %viol_to_field_mapping = ( 
                              F5::Violation::VIOL_XML_SOAP_ATTACHMENT() => 'flg_content_profiles', 
                              F5::Violation::VIOL_XML_WEB_SERVICES_SECURITY() => 'flg_content_profiles',
                              F5::Violation::VIOL_XML_FORMAT() => 'flg_content_profiles',
                              F5::Violation::VIOL_XML_SCHEMA() => 'flg_content_profiles', 
                              F5::Violation::VIOL_XML_MALFORMED() => 'flg_content_profiles',
                              F5::Violation::VIOL_XML_SOAP_METHOD() => 'flg_content_profiles',
                              F5::Violation::VIOL_JSON_FORMAT() => 'flg_content_profiles',
                              F5::Violation::VIOL_JSON_MALFORMED() => 'flg_content_profiles',
                              F5::Violation::VIOL_GWT_MALFORMED() => 'flg_content_profiles',
                              F5::Violation::VIOL_GWT_FORMAT() => 'flg_content_profiles',
                              F5::Violation::VIOL_REQUEST_MAX_LENGTH() => 'flg_request_length',
                              F5::Violation::VIOL_EVASION() => 'flg_evasion',
                              F5::Violation::VIOL_CSRF() => 'flg_csrf_urls',
                              F5::Violation::VIOL_ENCODING() => 'flg_encoding_failed_to_convert',
                              F5::Violation::VIOL_HTTP_PROTOCOL() => 'flg_http_checks',
                              F5::Violation::VIOL_QUERY_STRING_LENGTH() => 'flg_filetypes_lengths',
                              F5::Violation::VIOL_REQUEST_LENGTH() => 'flg_filetypes_lengths',
                              F5::Violation::VIOL_POST_DATA_LENGTH() => 'flg_filetypes_lengths',
                              F5::Violation::VIOL_URL_LENGTH() => 'flg_filetypes_lengths',
                              F5::Violation::VIOL_COOKIE_LENGTH() => 'flg_check_max_cookie_header_length',
                              F5::Violation::VIOL_PARAMETER_VALUE_METACHAR() => 'flg_parameters_value_meta',
                              F5::Violation::VIOL_PARAMETER_NAME_METACHAR() => 'flg_parameters_name_meta',
                              F5::Violation::VIOL_URL_METACHAR() => 'flg_urls_metachars',
                              F5::Violation::VIOL_HEADER_LENGTH() => 'flg_check_max_http_header_length',
                              F5::Violation::VIOL_METHOD() => 'flg_methods',
                              F5::Violation::VIOL_PARAMETER_VALUE_LENGTH() => 'flg_parameters_lengths',
                              F5::Violation::VIOL_CROSS_ORIGIN_REQUEST() => 'flg_urls_cross_origin_request',
                            );

our %field_to_viol_mapping;
for my $viol (keys %viol_to_field_mapping) {
    push (@{$field_to_viol_mapping{$viol_to_field_mapping{$viol}}}, $viol);
}

my $UNIFIED_LEARNING_VERSION = '12.0.0';

sub retrieve_nested_structures {
    my ($self, %params) = @_;
    $self->SUPER::retrieve_nested_structures(%params);

    my $dbh = $self->{dbh};
    my $policy = $self->{policy};

    #if rest_version < 12.0.0
    my $ver = $params{ver} || $params{call_flags}->{rest_version};
    if ( $ver && F5::BigipVersionUtils::compare_versions(version1 => $UNIFIED_LEARNING_VERSION,
                                                         version2 => $ver) > 0) {

        $self->populate_blocking_settings();
    
    }
}

#----------------------------------------------------------------------

sub populate_for_event {
    my ($self, %params) = @_;
    $self->populate_diff_structs(%params);
}

#----------------------------------------------------------------------

#Cheap population for backwards compatibility.
sub populate_blocking_settings {
    my ($self, %params) = @_;
    return if ($self->{blocking_settings} && @{$self->{blocking_settings}});

    my @blocking_settings = @{$self->{dbh}->selectall_arrayref(<<EndQuery, { Slice => {} }, $self->{policy}->{policy_id})};
SELECT viol_index, flg_learn
  FROM PLC.PL_BLOCKING
 WHERE policy_id = ?
EndQuery

    @blocking_settings = map {F5::ASMConfig::Blocking->new(flg_learn => $_->{flg_learn}, 
                                                           violation => F5::ASMConfig::Violation->new(viol_index => $_->{viol_index}))} @blocking_settings;
    $self->{blocking_settings} = \@blocking_settings;

    my ($are_signatures_in_learning) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $self->{policy}->{policy_id});
SELECT EXISTS(
  SELECT flg_learn
    FROM PLC.PL_POLICY_NEGSIG_SETS
   WHERE policy_id = ?
     AND flg_learn = 1)
EndQuery

    $self->{flg_signatures} = $are_signatures_in_learning;

    #Needed for backwards compatibility and pb config cascading changes.
    for my $blocking (@{$self->{blocking_settings}}) {
        if ($blocking->{violation} && $viol_to_field_mapping{$blocking->{violation}->{viol_index}}) {
            $self->{$viol_to_field_mapping{$blocking->{violation}->{viol_index}}} ||= $blocking->{flg_learn};
        }
    }
}

#----------------------------------------------------------------------
sub before_store {
    my($self, %params) = @_;

    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};

    my @fields = qw(
                     flg_enabled
                     level_global_threshold
                     flg_track_site_changes_untrusted
                     flg_track_site_changes_trusted
                     flg_urls
                     flg_websocket_urls
                   );

   my %field_values;
   my %field_display_name;

   for my $field (@fields) {
      $field_values{$field} = $self->get_new_value( store_obj => $store_obj,
                                                    existing_obj => $existing_obj,
                                                    field_name => $field, );

      $field_display_name{$field} = $self->get_field_display_name_string(field_name => $field);
   }

   if ( $self->has_value_changed(%params, check_fields => 'auto_apply_frequency') && $store_obj->{auto_apply_frequency} != F5::ASMConfig::AutoApplyFrequency::AUTO_APPLY_SCHEDULED() ) {
       $store_obj->{flg_auto_apply_on_all_days} = 1;
       $store_obj->{flg_auto_apply_at_all_times} = 1;
   }

   if ($self->has_value_changed(%params, check_fields => 'flg_content_profile_classification')) {
       $store_obj->{flg_classify_parameters} = $store_obj->{'flg_content_profile_classification'};
       $store_obj->{flg_classify_urls} = $store_obj->{'flg_content_profile_classification'};
   }
   
   if ( $self->has_value_changed(%params, check_fields => ['auto_apply_start_time', 'auto_apply_end_time']) ) {
       my $start_time = $self->get_new_value( store_obj => $store_obj,
                                              existing_obj => $existing_obj,
                                              field_name => 'auto_apply_start_time', );
       my $end_time = $self->get_new_value( store_obj => $store_obj,
                                            existing_obj => $existing_obj,
                                            field_name => 'auto_apply_end_time', );
       my $start_dt = Date::Parse::str2time($start_time);
       my $end_dt = Date::Parse::str2time($end_time);
       my $min_diff = ($end_dt - $start_dt)/60;
       $min_diff = $min_diff < 0 ? $min_diff + 24 * 60 : $min_diff;
       if ( $min_diff < 5 ) {
           die "Period of less than 5 minutes is not allowed: $start_time - $end_time\n";
       }
   }

    if ( $self->has_value_changed(%params, check_fields => 'level_global_threshold') ) {
        if (! $store_obj->{pbconfig_parameter}) {
            $store_obj->{pbconfig_parameter} = F5::ASMConfig::Entity::PBConfigParameter->new( dbh => $self->{dbh},
                                                                                              policy => $self->{policy});
        }

        if (! $store_obj->{pbconfig_cookie}) {
            $store_obj->{pbconfig_cookie} = F5::ASMConfig::Entity::PBConfigCookie->new( dbh => $self->{dbh},
                                                                                        policy => $self->{policy});
        }


        if ($field_values{level_global_threshold}) {
            $store_obj->{pbconfig_parameter}->{collapse_parameters_occurrences} = $field_values{level_global_threshold};
            $store_obj->{pbconfig_parameter}->{collapse_parameters} = 1;
            $store_obj->{pbconfig_cookie}->{collapse_cookies_occurrences} = $field_values{level_global_threshold};
            $store_obj->{pbconfig_cookie}->{collapse_cookies} = 1;
        }
        else {
            $store_obj->{pbconfig_parameter}->{collapse_parameters} = 0;
            $store_obj->{pbconfig_cookie}->{collapse_cookies} = 0;
        }
    }

   if($params{validate_value} && $field_values{flg_enabled}) {
       if ( $field_values{level_global_threshold} && $field_values{level_global_threshold} == 1 ) {
             die "Illegal value for $field_display_name{level_global_threshold}. Legal values are 0 ('no-collapse') or positive integer number larger than 1.\n";
       }
    
       if ($field_values{flg_track_site_changes_untrusted} && !$field_values{flg_track_site_changes_trusted}) {
           die "Illegal value for $field_display_name{flg_track_site_changes_untrusted}. Cannot track untrusted changes without tracking trusted changes.\n";
       }
       
       #additional validation to insure logical configuration
       my @extra_validations = (
           {
            validated_mode => "le",
            field1 => "rule_loosen_time_trusted",
            field2 => "rule_tighten_time",
            message => 'Trusted Time Period Minimum in "Accept as Legitimate" rule must be smaller or equal to Time Period Minimum in "Stabilize" rule.',
           },
           {
            validated_mode => "le",
            field1 => "rule_loosen_time_untrusted",
            field2 => "rule_tighten_time",
            message => 'Untrusted Time Period Minimum in "Accept as Legitimate" rule must be smaller or equal to Time Period Minimum in "Stabilize" rule.',
           },
           {
            validated_mode => "le",
            field1 => "rule_sitechanges_time_trusted",
            field2 => "rule_loosen_time_trusted",
            prerequisite_field => "flg_track_site_changes_trusted",
            message => 'Trusted Time Period Minimum in "Accept as Legitimate" rule must be smaller or equal to Trusted Time Period Minimum in "Stabilize" rule.',
           },
           {
            validated_mode => "le",
            field1 => "rule_sitechanges_time_trusted_max",
            field2 => "rule_loosen_time_trusted_max",
            prerequisite_field => "flg_track_site_changes_trusted",
            message => 'Trusted Time Period Maximum in "Track Site Changes" rule must be smaller or equal to Trusted Time Period Maximum in "Accept as Legitimate" rule.',
           },
           {
            validated_mode => "le",
            field1 => "rule_sitechanges_sources_trusted",
            field2 => "rule_loosen_sources_trusted",
            prerequisite_field => "flg_track_site_changes_trusted",
            message => 'Trusted Sources in "Track Site Changes" rule must be smaller or equal to Trusted Sources in "Accept as Legitimate" rule.',
           },
           {
            validated_mode => "le",
            field1 => "rule_sitechanges_time_untrusted",
            field2 => "rule_loosen_time_untrusted",
            prerequisite_field => "flg_track_site_changes_untrusted",
            message => 'Untrusted Time Period Minimum in "Track Site Changes" rule must be smaller or equal to Untrusted Time Period Minimum in "Accept as Legitimate" rule.',
           },
           {
            validated_mode => "le",
            field1 => "rule_sitechanges_time_untrusted_max",
            field2 => "rule_loosen_time_untrusted_max",
            prerequisite_field => "flg_track_site_changes_untrusted",
            message => 'Untrusted Time Period Maximum in "Track Site Changes" rule must be smaller or equal to Untrusted Time Period Maximum in "Accept as Legitimate" rule.',
           },
           {
            validated_mode => "le",
            field1 => "rule_sitechanges_sources_untrusted",
            field2 => "rule_loosen_sources_untrusted",
            prerequisite_field => "flg_track_site_changes_untrusted",
            message => 'Untrusted Sources in "Track Site Changes" rule must be smaller or equal to Untrusted Sources in "Accept as Legitimate" rule.',
           },
           {
            validated_mode => "special",
            field1 => "rule_loosen_time_untrusted_max",
            field2 => "rule_loosen_time_untrusted",
            field3 => "rule_loosen_sources_untrusted",
            message => 'Untrusted Period Maximum should be larger than the Untrusted Period Minimum in the "Accept as Legitimate" rule.',
           },
           {
            validated_mode => "special",
            field1 => "rule_loosen_time_untrusted_max",
            field2 => "rule_loosen_time_untrusted",
            field3 => "rule_loosen_sources_untrusted",
            message => 'Untrusted Period Maximum should be larger than the Untrusted Period Minimum in the "Accept as Legitimate" rule.',
           },
           {
            validated_mode => "special",
            field1 => "rule_loosen_time_trusted_max",
            field2 => "rule_loosen_time_trusted",
            field3 => "rule_loosen_sources_trusted",
            message => 'Trusted Period Maximum should be larger than the Trusted Period Minimum in the "Accept as Legitimate" rule.',
           },
           {
            validated_mode => "special",
            field1 => "rule_loosen_time_trusted_max",
            field2 => "rule_loosen_time_trusted",
            field3 => "rule_loosen_sources_trusted",
            message => 'Trusted Period Maximum should be larger than the Trusted Period Minimum in the "Accept as Legitimate" rule.',
           },
           {
            validated_mode => "special",
            field1 => "rule_sitechanges_time_untrusted_max",
            field2 => "rule_sitechanges_time_untrusted",
            field3 => "rule_sitechanges_sources_untrusted",
            prerequisite_field => "flg_track_site_changes_untrusted",
            message => 'Untrusted Period Maximum should be larger than the Untrusted Period Minimum in the "Track Site Changes" rule.',
           },
           {
            validated_mode => "special",
            field1 => "rule_sitechanges_time_untrusted_max",
            field2 => "rule_sitechanges_time_untrusted",
            field3 => "rule_sitechanges_sources_untrusted",
            prerequisite_field => "flg_track_site_changes_untrusted",
            message => 'Untrusted Period Maximum should be larger than the Untrusted Period Minimum in the "Track Site Changes" rule.',
           },
           {
            validated_mode => "special",
            field1 => "rule_sitechanges_time_trusted_max",
            field2 => "rule_sitechanges_time_trusted",
            field3 => "rule_sitechanges_sources_trusted",
            prerequisite_field => "flg_track_site_changes_trusted",
            message => 'Trusted Period Maximum should be larger than the Trusted Period Minimum in the "Track Site Changes" rule.',
           },
           {
            validated_mode => "special",
            field1 => "rule_sitechanges_time_trusted_max",
            field2 => "rule_sitechanges_time_trusted",
            field3 => "rule_sitechanges_sources_trusted",
            prerequisite_field => "flg_track_site_changes_trusted",
            message => 'Trusted Period Maximum should be larger than the Trusted Period Minimum in the "Track Site Changes" rule.',
           },
        );
       for my $validation (@extra_validations) {
           my $prerequisite_field = 1;
           if($validation->{prerequisite_field}) {
               $prerequisite_field = $self->get_new_value( store_obj => $store_obj,
                                                           existing_obj => $existing_obj,
                                                           field_name => $validation->{prerequisite_field}, );
           }
           if($prerequisite_field) {
               my $field1 = $self->get_new_value( store_obj => $store_obj,
                                               existing_obj => $existing_obj,
                                               field_name => $validation->{field1}, );
               my $field2 = $self->get_new_value( store_obj => $store_obj,
                                               existing_obj => $existing_obj,
                                               field_name => $validation->{field2}, );
               if($validation->{validated_mode} eq "le") {
                   if($field1 > $field2) {
                      die $validation->{message}."\n"; 
                   }
               }
               elsif($validation->{validated_mode} eq "special") {
                   my $field3 = $self->get_new_value( store_obj => $store_obj,
                                                   existing_obj => $existing_obj,
                                                   field_name => $validation->{field2}, );
                   if($field3 > 1 && $field1 < ($field2 / ($field3 -1)) ) {
                       die $validation->{message}."\n";
                   }
               }
           }
       }
   }

###Only if ver < 12.0.0
   my $ver = $params{ver} || $params{call_flags}->{rest_version};
   if ( $ver && F5::BigipVersionUtils::compare_versions(version1 => $UNIFIED_LEARNING_VERSION,
                                                        version2 => $ver) > 0) {

       for my $field (keys %field_to_viol_mapping) {
           if ($self->has_value_changed(%params, check_fields => $field)) {
               for my $viol_index (@{$field_to_viol_mapping{$field}}) {
                   my $blocking = F5::ASMConfig::Entity::Blocking->new(dbh => $self->{dbh},
                                                                       policy => $self->{policy},
                                                                       get_criteria => {},
                                                                       search_filters => [F5::ASMConfig::SearchFilter->new(field_name => 'violation__viol_index',
                                                                                                                           field_value => $viol_index,
                                                                                                                           field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ())]);
                   $blocking->store_update(call_flags => $params{call_flags},
                                            store_obj => { flg_learn => $self->get_new_value(%params, field_name => $field) });
               }
           }
       }
       if ($self->has_value_changed(%params, check_fields => 'flg_signatures')) {
            my $flg_signatures = $self->get_new_value(%params, field_name => 'flg_signatures');
            my $learning_sets_iter = F5::ASMConfig::Entity::PolicySignatureSet->retrieve_all(dbh => $self->{dbh},
                                                                                  policy => $self->{policy},
                                                                                  get_criteria => {},
                                                                                  search_filters => [F5::ASMConfig::SearchFilter->new(field_name => 'flg_learn',
                                                                                                                                   field_value => $flg_signatures,
                                                                                                                                   field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE())
                                                                                                     ],);
           while (my $set = $learning_sets_iter->get_next()) {
               $set->store_update(call_flags => $params{call_flags},
                                  store_obj => { flg_learn => $flg_signatures });
           }
       }
    }
}

#----------------------------------------------------------------------

sub from_rest {
    my ($self, %params) = @_;
    my $rv = $self->SUPER::from_rest(%params);
    if (defined $rv->{learning_mode}) {
        $rv->{flg_learning} = ($rv->{learning_mode} != F5::ASMConfig::LearningMode::LEARNING_MODE_NONE()) ? 1 : 0;
        $rv->{flg_enabled}  = ($rv->{learning_mode} == F5::ASMConfig::LearningMode::LEARNING_MODE_AUTOMATIC()) ? 1 : 0;
    }
    if ($rv->{flg_enabled}) {
        #For Backwards compatibility
        $rv->{flg_learning} = 1;
    }
    return $rv;
}

#----------------------------------------------------------------------

sub map_fields_to_entity {
    my($class, %params) = @_;

    $class->SUPER::map_fields_to_entity(%params);
    my $rec = $params{rec};

    if (! defined $rec->{learning_mode}) {
        if ($rec->{flg_enabled}) {
            $rec->{learning_mode} = F5::ASMConfig::LearningMode::LEARNING_MODE_AUTOMATIC();
        }
        elsif ($rec->{flg_learning}) {
            $rec->{learning_mode} = F5::ASMConfig::LearningMode::LEARNING_MODE_MANUAL();
        }
        else {
            $rec->{learning_mode} = F5::ASMConfig::LearningMode::LEARNING_MODE_NONE();
        }
    }
}

#----------------------------------------------------------------------

sub map_rest_level_global_threshold {
    my($self, %params) = @_;
    my $policy_id = $params{policy}->{id};
    my $direction = $params{direction};
    my $dbh = $self->{dbh};

    my $level_global_threshold = 0;

    if ($direction eq "to_rest") {
        my ($collapse_parameters_occurrences) = $dbh->selectrow_array("SELECT IF (collapse_parameters, collapse_parameters_occurrences, 0) FROM PLC.PL_POLICY_BUILDER_CONFIG_PARAMETER WHERE policy_id = ?",
                                                                        undef, $policy_id);
        my ($collapse_cookies_occurrences) = $dbh->selectrow_array("SELECT IF (collapse_cookies, collapse_cookies_occurrences, 0) FROM PLC.PL_POLICY_BUILDER_CONFIG_COOKIE WHERE policy_id = ?",
                                                                        undef, $policy_id);

        $level_global_threshold = ($collapse_parameters_occurrences || $collapse_cookies_occurrences);
    }
    else {
        $level_global_threshold = $params{field_value};
    }

    return $level_global_threshold;
}

#----------------------------------------------------------------------

sub affects_policy {
    my ($self, %params) = @_;
    return $self->{flg_enabled} != $params{policy}->{is_pb_enabled};
}

#----------------------------------------------------------------------
1;
