package F5::ASMConfig::Entity::ResponsePage;

use strict;
use warnings;

use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::ASMConfig;
use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::ResponsePage/;

my @CONFIGURABLE_RESPONSE_ACTION_TYPES = (
                                          F5::ASMConfig::ResponseActionType::RESPONSE_ACTION_TYPE_CUSTOM(),
                                          F5::ASMConfig::ResponseActionType::RESPONSE_ACTION_TYPE_REDIRECT(),
                                         );
my %CONFIGURABLE_RESPONSE_ACTION_TYPES = map { ($_ => 1) } @CONFIGURABLE_RESPONSE_ACTION_TYPES;
my @INVALID_MACROS_IN_REDIRECT_URL = ('<%TS.script.erase_cookies%>');

#----------------------------------------------------------------------

sub before_store {
    my($self, %params) = @_;

    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $dbh = $params{dbh} || $self->{dbh};

    my @fields = qw(
                    response_type
                    response_action_type
                    response_header
                    response_content
                    redirect_url
                    settings
                    ajax_custom_content
                    ajax_action_type
                    ajax_redirect_url
                    ajax_popup_message
                   );

   my %field_values;
   my %field_display_name;

   if (defined $existing_obj->{response_type} ){
       $store_obj->{response_type} = $existing_obj->{response_type};
   }

   for my $field (@fields) {
      $field_values{$field} = $self->get_new_value( store_obj => $store_obj,
                                                    existing_obj => $existing_obj,
                                                    field_name => $field, );

      $field_display_name{$field} = $self->get_field_display_name_string(%params, field_name => $field);
   }

    if ($self->has_value_changed(%params, check_fields => 'response_action_type') &&
        ! $CONFIGURABLE_RESPONSE_ACTION_TYPES{$field_values{response_action_type}}) {

        my($default_page) = __PACKAGE__->retrieve_all(dbh => $dbh,
                                                      logger => $params{logger},
                                                      default_table => 'PLC.PL_ALTERNATE_RESPONSE_DEFAULTS',
                                                      search_filters => [
                                                                         F5::ASMConfig::SearchFilter->new(field_name => q{response_type},
                                                                                                          field_value => $field_values{response_type},
                                                                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ()),
                                                                         F5::ASMConfig::SearchFilter->new(field_name => q{response_action_type},
                                                                                                          field_value => $field_values{response_action_type},
                                                                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ())
                                                                        ]
                                                     );
        if (! $default_page) {
            my $error_message = "Invalid $field_display_name{response_action_type} for this $field_display_name{response_type}";
            die { error_message => $error_message,
                    display_message => $error_message,
                    error_level     => 'info',
                    error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                };
        }

        $store_obj->{response_header} = $field_values{response_header} = $default_page->{response_header};
        $store_obj->{response_content} = $field_values{response_content} = $default_page->{response_content};
    }

   if ( $params{validate_value} && $field_values{response_type} != F5::ASMConfig::ResponseType::RESPONSE_TYPE_AJAX &&
         defined  $store_obj->{settings} ){
        my $error_message = "Invalid field $field_display_name{settings} for this $field_display_name{response_type}";
        die { error_message => $error_message,
              display_message => $error_message,
              error_level     => 'info',
              error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
            };
   }

   if ( $params{validate_value} && $field_values{response_type} == F5::ASMConfig::ResponseType::RESPONSE_TYPE_XML() &&
          ( $field_values{response_action_type} != F5::ASMConfig::ResponseActionType::RESPONSE_ACTION_TYPE_SOAP_FAULT() && 
            $field_values{response_action_type} != F5::ASMConfig::ResponseActionType::RESPONSE_ACTION_TYPE_CUSTOM() ) ){
         my $error_message = "Invalid $field_display_name{response_action_type}. (valid values are 'soap-fault' or 'custom')";
         die { error_message => $error_message,
               display_message => $error_message,
               error_level     => 'info',
               error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
             };
   }

   if ( $params{validate_value} && $field_values{response_action_type} == F5::ASMConfig::ResponseActionType::RESPONSE_ACTION_TYPE_CUSTOM &&
        !$field_values{response_header} ){
         my $error_message = "$field_display_name{response_header} is required";
         die { error_message => $error_message,
               display_message => $error_message,
               error_level     => 'info',
               error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
             };
   }

   if ( $params{validate_value} && $field_values{response_action_type} == F5::ASMConfig::ResponseActionType::RESPONSE_ACTION_TYPE_REDIRECT) {
       if (!$field_values{redirect_url} ) {
           my $error_message = "$field_display_name{redirect_url} is required";
           die { error_message => $error_message,
                   display_message => $error_message,
                   error_level     => 'info',
                   error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
               };
       }
       for my $invalid_macro_in_redirect_url (@INVALID_MACROS_IN_REDIRECT_URL) {
           if ($field_values{redirect_url} =~ /\Q$invalid_macro_in_redirect_url\E/) {
               my $error_message = "$field_display_name{redirect_url} may not contain \"$invalid_macro_in_redirect_url\"";
               die { error_message => $error_message,
                       display_message => $error_message,
                       error_level     => 'info',
                       error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                   };
           }
       }
   }

  if ( $params{validate_value} && $field_values{response_action_type} == F5::ASMConfig::ResponseActionType::RESPONSE_ACTION_TYPE_REDIRECT &&
        $field_values{response_type} == F5::ASMConfig::ResponseType::RESPONSE_TYPE_HIJACK ){
         my $error_message = "$field_display_name{redirect_url} may not be set for this $field_display_name{response_type}.";
         die { error_message => $error_message,
               display_message => $error_message,
               error_level     => 'info',
               error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
             };
   }

    my $empty_field;

    if ( $field_values{response_type} == F5::ASMConfig::ResponseType::RESPONSE_TYPE_AJAX() ||
         $field_values{response_type} == F5::ASMConfig::ResponseType::RESPONSE_TYPE_AJAX_LOGIN()){
          if ($field_values{ajax_action_type} == F5::ASMConfig::AjaxActionType::AJAX_ACTION_TYPE_REDIRECT()){
             $empty_field = $field_display_name{ajax_redirect_url} if ( !$field_values{ajax_redirect_url} || $field_values{ajax_redirect_url} eq '');
          }
          elsif ( $field_values{ajax_action_type} == F5::ASMConfig::AjaxActionType::AJAX_ACTION_TYPE_ALERT_POPUP() ) {
             $empty_field = $field_display_name{ajax_popup_message} if ( !$field_values{ajax_popup_message} || $field_values{ajax_popup_message} eq '');
          }
          elsif ( $field_values{ajax_action_type} == F5::ASMConfig::AjaxActionType::AJAX_ACTION_TYPE_CUSTOM() ) {
             $empty_field = $field_display_name{ajax_custom_content} if ( !$field_values{ajax_custom_content} || $field_values{ajax_custom_content} eq '');
          }
    }

    if ( $params{validate_value} && $empty_field ) {
        my $error_message = "$empty_field is required";
        die { error_message => $error_message,
                display_message => $error_message,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
            };
    }

    if ($field_values{response_type} == F5::ASMConfig::ResponseType::RESPONSE_TYPE_CAPTCHA() ||
        $field_values{response_type} == F5::ASMConfig::ResponseType::RESPONSE_TYPE_CAPTCHA_FAIL()) {

        if ($self->has_value_changed(%params, check_fields => 'response_header')) {
            my $error_message = "CAPTCHA $field_display_name{response_header} may not be modified";
            die { error_message => $error_message,
                    display_message => $error_message,
                    error_level     => 'info',
                    error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                };
        }

        my $error_message;
        my %optional_tokens = ('audio' => 1);
        for my $captcha_token (qw(change solution submit image support_id audio)) {
            my $pattern = '%ASM.captcha.' . $captcha_token . '%';
            my $regex = qr/$pattern/;
            my $occurrences_count = () = ($field_values{response_content} =~ /$regex/g);

            if ((! exists $optional_tokens{$captcha_token}) && ($occurrences_count == 0)) {
                $error_message = "Required CAPTCHA token in $field_display_name{response_content} is missing (\"$pattern\")";
            }
            if ($occurrences_count > 1) {
                $error_message = "Required CAPTCHA token in $field_display_name{response_content} may only occur once (\"$pattern\")";
            }

            if ($error_message) {
                die { error_message => $error_message,
                        display_message => $error_message,
                        error_level     => 'info',
                        error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                    };
            }
        }
    }
}

#----------------------------------------------------------------------

sub rest_search_filters {
    my ($class, %params) = @_;
    if ($class->is_feature_disabled(feature_name => 'grpc_enforcement')) {
        return [ F5::ASMConfig::SearchFilter->new(field_name => 'response_type',
                                                  field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE(),
                                                  field_value => F5::ASMConfig::ResponseType::RESPONSE_TYPE_GRPC()),
               ];
    }
    return [];
}

#----------------------------------------------------------------------

1;
