package F5::ASMConfig::Entity::Violation;

use strict;
use warnings;

use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::ASMConfig;
use F5::CRC;
use F5::Utils::Rest;
use F5::Cluster::Utils;
use F5::ASMConfig::Cache;

use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::Violation/;

our $MIN_USER_VIOL_INDEX = 100;
our $MAX_USER_VIOL_INDEX = 120;

#----------------------------------------------------------------------

sub before_store {
    my($self, %params) = @_;

    my $store_mode = $params{store_mode};
    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};

    delete $store_obj->{flg_is_user_defined}; # user may not change flg_is_user_defined
    my %fields = %{$self->fields};
    my @factory_fields = grep {!( defined $fields{$_}->{extern_db_table} && $fields{$_}->{extern_db_table} eq 'PLC.VIOLATIONS_PROPERTIES')} keys %fields;
    if ( $existing_obj && %$existing_obj && (!$existing_obj->{flg_is_user_defined}) ){
        for my $factory_field ( @factory_fields ) {
            delete $store_obj->{$factory_field};
        }
    }

    if ($store_mode eq 'add' || $self->has_value_changed(%params, check_fields => [ qw(viol_name) ])) {

        my $violation_name = $self->get_new_value( store_obj => $store_obj,
                                                   existing_obj => $existing_obj,
                                                   field_name => 'viol_name', );

        $store_obj->{viol_name_crc} = F5::CRC::get_crc32($violation_name);
        $store_obj->{icontrol_name} = $violation_name;
        $store_obj->{bd_icontrol_name} = $violation_name;
    }

    if ($self->has_value_changed(%params, check_fields => [ 'flg_is_deleted' ])) {
        if ($self->get_new_value(%params, field_name => 'flg_is_deleted')) {
            $store_obj->{delete_time} = time;
        }
        else {
            $store_obj->{delete_time} = 0;
        }
    }

    if ( $store_mode eq 'add' ){
        $store_obj->{flg_is_user_defined} = 1;
        $store_obj->{section_type} = F5::ASMConfig::SectionType::SECTION_TYPE_USER_VIOLATIONS();
        $store_obj->{created_time} = time();
        my $available_index = $self->get_viol_index(dbh => $self->{dbh});
        $store_obj->{viol_index} = $available_index ? $available_index : die "No available index found/n";
    }

    if ($self->has_value_changed(%params, check_fields => [ 'attack_type_id' ])) {
        if ( $store_obj->{attack_type_id} ) {
            $store_obj->{attack_type_code} = Bit::Vector->new_Enum(64, $store_obj->{attack_type_id} - 1)->to_Dec;
        }
    }

    if ( !defined $existing_obj->{severity} && !defined $store_obj->{severity}) {
        $store_obj->{severity} =F5::ASMConfig::ViolationSeverity::VIOLATION_SEVERITY_EMERG();        
    }
}

#----------------------------------------------------------------------

sub after_store {
    my($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $viol_index = $self->get_new_value(%params, field_name => 'viol_index');
    my %rec = (viol_index => $viol_index,
               flg_alarm => 0,
               flg_reject => 0,
               flg_learn => 0,
               flg_session_awareness => 0,
               security_level_id => 1, # ASM full
              );

    F5::DbUtils::insert_row(dbh => $dbh,
                            table_name => 'PLC.PL_SECURITY_LEVEL_DEFAULTS',
                            ignore => 1,
                            rec => \%rec);

    my @policy_ids = @{$dbh->selectcol_arrayref("SELECT id FROM PLC.PL_POLICIES")};
    for my $policy_id (@policy_ids) {
        F5::DbUtils::insert_row(dbh => $dbh,
                                table_name => 'PLC.PL_BLOCKING',
                                ignore => 1,
                                rec => { %rec,
                                         policy_id => $policy_id },
                               );
    }

    F5::Utils::Rest::populate_uuids(dbh => $dbh,
                                    rest_entities => ['F5::ASMConfig::Entity::Blocking'],
                                   );
}

#----------------------------------------------------------------------

sub map_fields_to_entity {
    my ($class_or_self, %params) = @_;

    $class_or_self->SUPER::map_fields_to_entity(%params);

    my $obj = $params{rec} || $class_or_self;

    my $is_ignore_bewaf_enabled = $class_or_self->is_ignore_bewaf_enabled();

    $obj->{flg_accessible_violation} = 1;
    if ($obj->{feature_flag}) {
        $obj->{flg_accessible_violation} = $class_or_self->has_feature_flag(feature_flag => $obj->{feature_flag});

        if ($obj->{viol_index} && $obj->{viol_index} == F5::Violation::VIOL_LEAKED_CREDENTIALS()) {
            # Load dynamically to prevent F5::COAPI and SOAP::Lite usage at build time
            require F5::License;
            if (F5::License::is_credential_stuffing_licensed()) {
                $obj->{flg_accessible_violation} = 1;
            }
            else {
                $obj->{flg_accessible_violation} = 0;
            }
        }
    }

    if ($obj->{viol_index} && $obj->{viol_index} == F5::Violation::VIOL_THREAT_ANALYSIS() && !$is_ignore_bewaf_enabled) {
        # hide 'Mitigation action determined by Threat Analysis Platform' viol from GUI
        $obj->{flg_accessible_violation} = 0;
    }
    
   if ($obj->{viol_index} && $obj->{viol_index} == F5::Violation::VIOL_THREAT_CAMPAIGN() && F5::License::is_bewaf_licensed()) {
       # show 'Threat Campaign detected' viol in GUI
       $obj->{flg_accessible_violation} = 1;
    }

    # show/hide violations managed by bewaf
    my %bewaf_violations = (
       F5::Violation::VIOL_BLOCKING_CONDITION() => 1,
       F5::Violation::VIOL_CONVICTION() => 1,
       F5::Violation::VIOL_THREAT_ANALYSIS() => 1
    );

    if ($is_ignore_bewaf_enabled) {
        # show these violations always when asm.ignore_bewaf is true
        $bewaf_violations{F5::Violation::VIOL_MALICIOUS_DEVICE()} = 1;
    }

    if ($obj->{viol_index} && $bewaf_violations{$obj->{viol_index}}) {
        $obj->{flg_accessible_violation} = $is_ignore_bewaf_enabled ? 1 : 0;
    }

    my %violation_rating_viols = (
       F5::Violation::VIOL_RATING_NEED_EXAMINATION() => 1,
       F5::Violation::VIOL_RATING_THREAT() => 1,
    );

    if ($obj->{viol_index} && $violation_rating_viols{$obj->{viol_index}} ) {
       if ( F5::ASMConfig::Cache::get_cached_enable_enforcement_by_rating(%params) ) {
           $obj->{flg_accessible_violation} = 1;
       }
       else {
           $obj->{flg_accessible_violation} = 0;
       }
    }
}

#----------------------------------------------------------------------

sub retrieve_all {
    my($class, %params) = @_;

    if (! $params{search_filters}) {
        return $class->SUPER::retrieve_all(%params);
    }

    my @select_filters = grep {$_->{field_name} eq "flg_accessible_violation"} @{$params{search_filters}};
    if (! scalar @select_filters) {
        return $class->SUPER::retrieve_all(%params);
    }

    my @objs = $class->SUPER::retrieve_all(%params, page_size => 0);

    $class->apply_search_filters(result_set => \@objs,
                                 %params,
                                 search_filters => \@select_filters,
                                );

    $class->apply_paging(result_set => \@objs,
                             %params);

    my $total_record_count = scalar @objs;
    if (wantarray) {
        return @objs;
    }
    else {
        my $iterator = F5::ASMConfig::Entity::Iterator->new(
                                                            get_next_function => sub {return shift(@objs);},
                                                            get_all_function => sub {return @objs;},
                                                            total_record_count => $total_record_count,
                                                           );
        return $iterator;
    }
}

#----------------------------------------------------------------------

sub additional_field_mapping {
    my($class, %params) = @_;

    my $dbh = $params{dbh};
    my %additional_field_mapping = (
                                    "delete_time" => { full_db_field_name => F5::DbUtils::is_sqlite($dbh) ?
                                                                            "datetime(delete_time, 'unixepoch')"
                                                                          : 'FROM_UNIXTIME(delete_time, \'%Y-%m-%d %H:%i:%s\')' },
                                   );

    return \%additional_field_mapping;
}

#----------------------------------------------------------------------

sub get_viol_index{
  my ($class,%params) = @_;

  my $dbh = $params{dbh};

  my $ids = $dbh->selectall_hashref(<<EndQuery,'viol_index');
SELECT viol_index,created_time,flg_is_deleted,delete_time
  FROM PLC.VIOLATIONS
 WHERE flg_is_user_defined = 1
EndQuery

  my $optional_replace_id = 0;
  foreach my $index ($MIN_USER_VIOL_INDEX..$MAX_USER_VIOL_INDEX) {
    if (!$ids->{$index}) {
        return $index;
    }
    else {
        if ($ids->{$index}->{flg_is_deleted}){
            if ($optional_replace_id) {
                if ($ids->{$optional_replace_id}->{delete_time} > $ids->{$index}->{delete_time}){
                    $optional_replace_id = $index;
                }
            }else{
                $optional_replace_id = $index;
            }
        }
    }
  }

  if ($optional_replace_id){
       $dbh->do(<<EndQuery, undef, $optional_replace_id);
DELETE FROM PLC.VIOLATIONS WHERE viol_index = ?
EndQuery

       return $optional_replace_id;
  }

  return 0;
}

sub is_ignore_bewaf_enabled {
  my $ignore_bewaf = F5::Cluster::Utils::get_mcp_val("asm.ignore_bewaf");
  return 0 if (! defined $ignore_bewaf);
  return ($ignore_bewaf eq 'true') ? 1 : 0;
}

#----------------------------------------------------------------------

1;
