package F5::ASMConfig::Handler;

use strict;
use warnings;

use YAML::XS;
use File::Basename;
use File::Path;
use File::Copy qw(move);
use File::Find;
use File::stat;
use Storable qw(thaw);
use Socket;
use Symbol;
use IO::Socket::INET6;
use Time::HiRes;
use List::Util;
use Fcntl;
use POSIX qw(:errno_h :sys_wait_h);
use Scalar::Util qw(reftype);

use F5::Deploy;
use F5::Logger;
use F5::DbUtils;
use F5::ASMConfig::ASMConfig qw(preload);
use F5::ASMConfig::Handler::Base;
use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::Utils;
use F5::ASMConfig::Cache;
use F5::IO::Selector;
use F5::IO::Storable;
use F5::ASM::COAPI;
use F5::ActionItemHandler;
use F5::ObjUtils;
use F5::ASMConfig::Entity::DeviceGroupMember;
use F5::BatchInsert;
use F5::BigIpLog;
use F5::Cfg;
use F5::ModuleLoader;
use F5::DeviceGroup;
use F5::Cgc::Channel::DeviceGroup;
use F5::LockFactory;
use F5::BigipUtils;
use F5::DateTimeUtils;
use F5::ASMConfig::Client;
use F5::Mcp::UserRole;
use F5::Utils::SharedMemory; 
use File::Basename;
use F5::ObjUtils qw(freeze dclone store_fd);
use F5::NwdUtils;
use F5::Cluster::Utils;
use F5::BigipVersionUtils;

$SIG{PIPE} = 'IGNORE';

my $DEFAULT_CONFIG_FILE = '/etc/ts/tools/asm_config_server.cfg';
our $DEFAULT_EVENT_QUEUE_PORT = 9783;
our $DEFAULT_RPC_TO_EVENT_PORT = 9784;

our $DUPLICATE_FOUND_DBI_ERROR = '(DBD::mysql::db do failed: Duplicate entry)|(DBD::SQLite::db do failed: columns .* are not unique)|DBD::SQLite::db do failed: UNIQUE constraint failed:';
our $NONEXISTENT_RECORD_ERROR = 'nonexistent record error';
our $MISSING_FOREIGN_KEY_ERROR = 'Cannot add or update a child row: a foreign key constraint fails';
our $DEADLOCK_FOUND_ERROR = 'Deadlock found when trying to get lock; try restarting transaction';
our $LOCK_TIMEOUT_ERROR = 'Lock wait timeout exceeded; try restarting transaction';

our $DEFAULT_MAX_QUEUED_METHOD_CALL_TOTAL_SIZE = 2_000_000; # 2 MB (approx) 
our $DEFAULT_MAX_QUEUED_METHOD_CALL_TOTAL_SIZE_BEFORE_ACTIVE = 500_000; # 500 kb (approx)

our $SYNC_DIR = F5::Deploy::map_path("/ts/var/sync");

my $ASM_UCS_INSTALL_PROC_FILENAME = F5::Deploy::map_path('/var/ts/var/install/ucs_install.pid');
our $ASM_CONFIG_PROCESS_PRIORITY_FILENAME = F5::Deploy::map_path('/usr/share/ts/config/asm_config_process_priority.yaml');
my $rpc_process_priority_was_set = 0;

my $handler_modules_loaded = 0;
our $TSCONFD_USER_NAME = "tsconfd";
my $POLICY_BUILDER_USER_NAME = "Policy Builder";
my $BOTD_USER_NAME = "botd";
my $POLICY_INDEX = 2;
my $CALL_FLAG_INDEX = 1;
our $SYNC_REREQUEST_TIME = 5 * 60; # 5 Minutes
our $DEFAULT_CHECK_MEMORY_SIZE_INTERVAL = 10;
our $STALLED_ASYNC_CALLS_INTERVAL = 60;

our $verbose_logger = F5::Logger->get_logger(logger_name => 'asmconfigverbose');
our $ASM_DATA_PARTITION = '/var/asmdata1';
our $TEMP_FILE_DIR = "$ASM_DATA_PARTITION/tmp";

my $MAX_DEADLOCK_RETRY_COUNT = 3;

#----------------------------------------------------------------------

sub new {
  my($class, %params) = @_;
  my $self = {};
  bless $self, $class;
  $self->init(%params);

  return $self;
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;

  $self->{element_types} = {};
  $self->load_handler_modules;
  $self->{logger} = $params{logger} || F5::Logger->get_logger(logger_name => 'asm_config_server.pl');
  $self->{zero_padded_major_minor_version} = F5::BigipVersionUtils::zero_padded_major_minor(F5::BigipVersionUtils::bigip_version());
  $self->{config_file} = $params{config_file} || $DEFAULT_CONFIG_FILE;
;
  $self->{rpc_to_event_port} = $params{rpc_to_event_port} || F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'RpcToEventPort') || $DEFAULT_RPC_TO_EVENT_PORT;
  $self->{persistence_frequency_count} =  F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'PersistenceFrequencyCount') || 5;
  $self->{persistence_frequency_units} =  F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'PersistenceFrequencyUnits') || "secs";
  $self->{max_rpc_calls_per_async_handler} =  F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'MaxRpcCallsPerAsyncHandler') || 10000;
  $self->{max_memory_size} =  F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'MaxMemorySize') || (450 * 1024 * 1024); #450MB
  $self->{async_max_memory_size} =  F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'AsyncMaxMemorySize') || 200_000_000;
  $self->{async_max_idle_time} =  F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'AsyncMaxIdleTime') || (60 * 5); #5 Minutes
  $self->{max_memory_size_check_time} = time();
  $self->{memory_limit_exceeded} = 0;
  $self->{check_memory_size_interval} = F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'CheckMemorySizeInterval') || $DEFAULT_CHECK_MEMORY_SIZE_INTERVAL;
  $self->{max_queued_method_call_total_size} = F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'MaxQueuedMethodCallTotalSize') || $DEFAULT_MAX_QUEUED_METHOD_CALL_TOTAL_SIZE;
  $self->{max_queued_method_call_total_size_before_active} = F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'MaxQueuedMethodCallTotalSizeBeforeActive') || $DEFAULT_MAX_QUEUED_METHOD_CALL_TOTAL_SIZE_BEFORE_ACTIVE;
  $self->reinit(initial_creation => 1, dbh => $params{dbh});

  if (F5::BigipUtils::is_docker()) {
      $self->{local_device_name} = F5::BigipUtils::get_docker_container_id();
  }
  else {
      $self->{local_device_name} = F5::DeviceGroup::get_local_device_name();
  }

  if (F5::BigipUtils::is_feature_enabled('cmi_shared_memory')) {
      #Shared variable to reflect if we are currently a member of the asm device group (controls if we will forward messages or not)
      ($self->{shared_state_ipc}, $self->{shared_state}) = F5::Utils::SharedMemory::get_shared_memory(key => $F5::Utils::SharedMemory::SHARED_STATE_KEY);
      ($self->{known_times_ipc}, $self->{known_times}) = F5::Utils::SharedMemory::get_shared_memory(key => $F5::Utils::SharedMemory::KNOWN_TIME_KEY);
  }
  else {
      $self->{shared_state} = {};
      $self->{known_times} = {};
  }

  $self->{shared_state}{local_device_name} = $self->{local_device_name};
  my $device = F5::DeviceGroup::get_device($self->{local_device_name});
  $self->{shared_state}{configsync_ip} = $device->{configsync_ip};
  my $mcp_device_group = F5::DeviceGroup::get_device_group($self->{local_device_name});
  $self->{device_group} = $mcp_device_group->{name};
  if ($self->{shared_state}{device_group} &&
      $self->{shared_state}{device_group} ne $self->{device_group}
     ) {
      #device group changed while ASM was down!
      $self->{shared_state}{sync_init} = time();
      $self->{shared_state}{last_sync_request_time} = 0;
      $self->clean_device_group_state(device_group => $self->{shared_state}{device_group}); 
  }

  $self->{shared_state}{device_group} = $self->{device_group};  
  if ($self->{device_group}) {
      $self->{shared_state}{is_auto_sync} = $mcp_device_group->{autosync_enabled};
      my @devices_in_group = F5::DeviceGroup::get_devices_in_device_group($self->{device_group});
      $self->{shared_state}{devices_in_group} = \@devices_in_group;

      if (my @device_members = F5::ASMConfig::Entity::DeviceGroupMember->retrieve_all(dbh => $self->{dbh})) {
          $self->{logger}->debug("Loading known_times from database") if $self->{logger};
          for my $device_member (@device_members) {
              $self->{known_times}{$device_member->{host}} ||= $device_member->{known_time};
          }
      }
      my %in_group = map {( $_ => 1)} @devices_in_group;
      for my $known_host (keys %{$self->{known_times}}) {
          if (!$in_group{$known_host}) {
              delete $self->{known_times}->{$known_host};
              F5::ASMConfig::Entity::DeviceGroupMember->delete_many(dbh => $self->{dbh}, 
                                                                    search_criteria => F5::ASMConfig::DeviceGroupMember->new({host => $known_host}));
          }
          elsif (defined $self->{known_times}{$known_host}) {
               $self->{shared_state}{sync_init} = 0;
          }
      }
  }

  $self->{shared_state}{local_hostname} = F5::Cluster::Utils::get_mcp_val('hostname');
  $self->{alarm_on_destroy} = defined $params{alarm_on_destroy} ? $params{alarm_on_destroy} : 1;
  $self->{guarantee_event_sending} = $params{guarantee_event_sending} || 0;
  # Global setting for send_events_to_dispatcher will be propagated to call_flags
  $self->{send_events_to_dispatcher} = defined $params{send_events_to_dispatcher} ? $params{send_events_to_dispatcher} :
    F5::BigipUtils::is_feature_enabled('send_events_to_dispatcher');
}

#----------------------------------------------------------------------
sub clean_device_group_state {
    my ($self, %params) = @_;
    my $device_group = $params{device_group};

    # Only clean up sync objects where there is no longer a device_group on the folder
    my @devices_in_group = F5::DeviceGroup::get_devices_in_device_group($device_group);
   
    my $folder_name =  F5::DeviceGroup::get_asm_sync_folder_name($device_group);
    my $folders = F5::ASM::COAPI::Query->new("folder");
    if ($folders && @$folders) {
        if (my ($sync_folder) = grep { $_->{name} eq $folder_name} @$folders) {
            if ( !$sync_folder->{devicegroup}) {
                F5::DeviceGroup::clean_asm_device_sync(device_group => $device_group,
                                                       logger => $self->{logger},);
            }
        }
    }

    F5::ASMConfig::Entity::DeviceGroupMember->delete_many(dbh => $self->{dbh});
    for my $known_host (keys %{$self->{known_times}}) {
        delete $self->{known_times}->{$known_host};
    }

    $self->{shared_state}{devices_in_group} = [];
    return;
}

#----------------------------------------------------------------------

sub reinit {
  my($self, %params) = @_;
  unless ($params{initial_creation}) {
    # Rename process to something else (to not be counted by nwd)
    if ($params{is_async_process}) {
      $0 = 'asm_config_server_rpc_handler_async.pl';
    }
    elsif ($params{is_relay_handler_process}) {
      $0 = 'asm_config_server.pl';  ## We want this to be counted by nwd
    }
    else {
      $0 = 'asm_config_server_rpc_handler.pl';
    }
  }
  F5::ASM::COAPI::Login("root");

  $self->{is_relay_handler_process} = $params{is_relay_handler_process};
  $self->{dbh} = $params{dbh} || F5::DbUtils::get_dbh(db_opts => {
                                                                  RaiseError => 1,
                                                                  PrintError => 0,
                                                                 });

  $self->{known_function_calls} = {};
  $self->{events} = [];
  $self->{send_pbconfig_event} = {};
  $self->{post_commit_hooks} = [];
  $self->{section_policy_cache} = {};
  $self->{sessions} = [ 'dummy' ]; # pre-seeded with dummy to use session IDs from 1 upwards
  $self->{session_last_event_subscription_id} = {};
  $self->{is_async_process} = $params{is_async_process};
  if ($self->{is_async_process} && $self->{session} && ($self->{session}->{user_name} eq $POLICY_BUILDER_USER_NAME ||
                                                        $self->{session}->{user_name} eq $BOTD_USER_NAME) ) {
      $self->{batched_insertor} =  F5::BatchInsert->new(dbh => $self->{dbh});
      $self->{is_long_lived_async_handler_process} = 1;
  }
  $self->{in_transaction} = 0;
  $self->{async_call_id} = 1;
  $self->{async_children} = {};
  $self->{rpc_entry_depth} = 0;
  $self->{transaction_relay_queue} = [];
  $self->{relay_transaction_rpc_depth_stack} = [];
  $self->{update_mcp_timestamp} = 0;
  $self->{max_memory_size_check_time} = time();
  delete $self->{transaction_locker};
  delete $self->{transactional_saves};
  $self->{rpc_queue} = [];
  $self->{action_item_handler} = F5::ActionItemHandler->get_instance(dbh => $self->{dbh},
                                                                     logger => $self->{logger},
                                                                    );

  $self->{dcc_dbname} = 'DCC';
  $self->{plc_dbname} = 'PLC';

  my $cfg_file = $self->{config_file} || '/etc/ts/tools/asm_config_server.cfg';
  $self->{default_page_size} = F5::Cfg::cfg_get_config_item($cfg_file, 'General', 'AsmConfigDefaultPageSize') || 10000;
  $self->{max_child_policies} = F5::Cfg::cfg_get_config_item($cfg_file, 'General', 'AsmConfigMaxChildPolicies') || 255;

  if (! defined $self->{asmconfig_check_authz}) {
      my $config_check_authz = F5::Cfg::cfg_get_config_item($cfg_file, 'General', 'asmconfig_check_authz');
      $self->{asmconfig_check_authz} = defined $config_check_authz ? $config_check_authz : 1;
  }
}


#----------------------------------------------------------------------
sub spawn_relay_handler {
  my($self) = @_;

  if (my $pid = fork) {
    # Parent
    return $pid;
  }
  else {
    # Child
    $self->{dbh}->{InactiveDestroy} = 1;
    my $batched_insertor = F5::BatchInsert->new(dbh => $self->{dbh});
    $self->reinit(is_relay_handler_process => 1,
                  batched_insertor => $batched_insertor,
                 );

    $self->{logger}->info("Relay handler starting") if $self->{logger};
    if ((!$self->{shared_state}{device_group}) && (F5::BigipUtils::is_device_bigip())) {
        F5::ASMConfig::Entity::DeviceGroupMember->delete_many(dbh => $self->{dbh});
        for my $known_host (keys %{$self->{known_times}}) {
            delete $self->{known_times}->{$known_host};
        }
    }

    my $MAX_CHANNEL_ATTEMPTS = 5;
    my $channel_attempts = 0;
    $self->{shared_state}{last_sync_request_time} = 0;
    $self->{shared_state}{last_sync_request_host} = 0;
    while (1) {
        #Wait until we are a relay member
        while (!$self->{shared_state}{device_group} || (F5::Cluster::Utils::is_clustering_enabled() && ! F5::Cluster::Utils::is_master_blade())) { 
            Time::HiRes::usleep(500_000);
        }
        $self->{device_group} = $self->{shared_state}{device_group};
        ## Open reading connection to CGC here
        my %channels;
        $self->subscribe_to_cgc_channels(\%channels);

        if (!$channels{$F5::Cgc::Channel::DeviceGroup::ASM_CGC_LOW_PRI_TO_ALL} ||
            !$channels{$F5::Cgc::Channel::DeviceGroup::ASM_CGC_HIGH_PRI_TO_SINGLE}) {
            if ($self->{shared_state}{device_group}) {
                $channel_attempts++;
                if ($channel_attempts > $MAX_CHANNEL_ATTEMPTS) {
                     $self->{logger}->warn("CGC Failure.  We are a member of device group " . $self->{shared_state}{device_group} . ", but a channel could not be established.  Shutting down relay_listener") if $self->{logger};
                     exit;
                }
                $self->{logger}->info("CGC Failure.  We are a member of device group " . $self->{shared_state}{device_group} . ", but a channel could not be established") if $self->{logger};
                Time::HiRes::usleep(500_000);
            }
            next;
        }
        else {
            $channel_attempts = 0;
        }

        my $startup_sync_device_group_member;

        if (my @devices_in_group = F5::DeviceGroup::get_devices_in_device_group($self->{shared_state}{device_group}) && (F5::BigipUtils::is_device_bigip())) {
            my $best_peer = $self->find_best_peer(devices => \@devices_in_group,
                                                  include_self => 1);
            if ($best_peer->{device}) {
                $self->{shared_state}{sync_init} = 0;
            }


            if ($best_peer->{device} && $best_peer->{device} eq $self->{local_device_name}) {
                if ($self->{known_times}{$self->{local_device_name}}
                   && $self->{shared_state}{is_auto_sync}) {
                    my $asm_config_settings = F5::ASMConfig::Entity::AsmConfigSettings->new( dbh => $self->{dbh})->retrieve(get_criteria => {});
                    if ($asm_config_settings->{flg_recommend_sync_when_policy_is_not_applied}) {
                        $self->update_mcp_timestamp();
                    }
                }
            } 
            elsif ($best_peer->{device} && $best_peer->{timestamp} gt ($self->{known_times}->{$best_peer->{device}} ||= 0) ) {
                $self->{logger}->info("We are a member of a device group, but we are missing information about it.") if $self->{logger}; 
                $startup_sync_device_group_member = F5::ASMConfig::DeviceGroupMember->new({host => $best_peer->{device},
                                                                                           known_time => $best_peer->{timestamp},});
            }
        }
    
        my $DEFAULT_POLL_TIME = 180;
        my $EMERGENCY_MAX_WAIT_TIME = 60;
        my $MAX_UNRESOLVED_EMERGENCY_WAIT_TIME = 60 * $EMERGENCY_MAX_WAIT_TIME;

        my $persistency_count = $self->{persistence_frequency_units} eq "secs" ? time() : 0;
        my $resubscribe = 0;
        $self->{logger}->debug("Relay Listener is a member of DG '" . $self->{shared_state}{device_group} . "' and will listen to events") if $self->{logger}; 
        while ($self->{shared_state}{device_group} && !$resubscribe && !$self->{shared_state}{resubscribe}
               && (! F5::Cluster::Utils::is_clustering_enabled() || F5::Cluster::Utils::is_master_blade())) {
           
            $self->check_memory_size();
            if (($self->{persistence_frequency_units} eq "calls" && $persistency_count >= $self->{persistence_frequency_count}) ||
                ($self->{persistence_frequency_units} eq "secs" && (time() - $persistency_count >= $self->{persistence_frequency_count}))) {
                $self->{logger}->debug("Persisting known times") if $self->{logger};
                $persistency_count = $self->{persistence_frequency_units} eq "secs" ? time() : 0;
                for my $host (keys %{$self->{known_times}}) {
                    my $device_group_member = F5::ASMConfig::DeviceGroupMember->new({host => $host,
                                                                                     known_time => $self->{known_times}{$host}});
                    eval {
                        $self->update_device_group_member(undef, F5::ASMConfig::CallFlags->new({add_on_updating_nonexistent_record => 1}), undef, $device_group_member);
                    };
                    $self->{logger}->info("Failed to update device group member: $@->{error_message}") if $@ && $self->{logger};
                }
            }
            my @channels_to_poll = ($channels{$F5::Cgc::Channel::DeviceGroup::ASM_CGC_HIGH_PRI_TO_SINGLE});
            my @poll_times;
            if ($self->{shared_state}{emergency} || $self->{shared_state}{sync_init}) {
                push(@poll_times, $EMERGENCY_MAX_WAIT_TIME);
            }
            else {
                push @channels_to_poll, $channels{$F5::Cgc::Channel::DeviceGroup::ASM_CGC_LOW_PRI_TO_ALL},
                                        $channels{$F5::Cgc::Channel::DeviceGroup::ASM_CGC_MCP_EVENTS};
                push (@poll_times, (($startup_sync_device_group_member || $self->{shared_state}{data_protection_peer}) ? 1 : $DEFAULT_POLL_TIME));
            }
            if ($self->{shared_state}{last_sync_request_time}) {
                my $poll_time = F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $SYNC_REREQUEST_TIME,
                                                                                   minimum_interval => 1,
                                                                                   start_time => $self->{shared_state}{last_sync_request_time},
                                                                                  );
                push (@poll_times , $poll_time);
            }

            my $timeout = List::Util::min(@poll_times);

            # poll returns a list of waiting chans, but we only handle them one at a time
            if (my ($waiting_channel) = F5::Cgc::Channel::DeviceGroup->poll(channels => \@channels_to_poll,
                                                                            timeout => $timeout,
                                                                            resubscribe_ref => \$resubscribe,
                                                                            logger => $self->{logger},
                                                                           )) {
                if ($resubscribe || $self->{shared_state}{resubscribe}) {
                    $self->{logger}->debug("Going to resubscribe CGC channel") if $self->{logger};
                    last;
                }

                my $relayed_call;
                eval {
                    $relayed_call =  $waiting_channel->receive_deserialized();
                };
                if (my $err = $@) {
                    $self->{logger}->error("Error on receive a relayed call: $err") if $self->{logger};
                    die $err;
                }
                if (!$relayed_call) {
                    $self->{logger}->debug("Did not receive a relayed call") if $self->{logger};
                    next;
                }

                if ($relayed_call->{function_name} ne "sync_receive_file_part" ) {
                    $self->{logger}->trace($relayed_call) if $self->{logger};
                }
                else {
                    $self->{logger}->trace("Relayed call function name is sync_receive_file_part - skip logging the call trace") if $self->{logger};
                }

                if ($relayed_call->{device_group_member}) {
                    #MCP update call             
                    $self->{logger}->debug("Received MCP update call") if $self->{logger};
                    if (!$self->{known_times}->{$relayed_call->{device_group_member}->{host}}
                         || $relayed_call->{device_group_member}->{known_time} gt $self->{known_times}->{$relayed_call->{device_group_member}->{host}}
                       ) {
                        my $reason = $self->{shared_state}{is_auto_sync} ? "MCP indicates missed Event $relayed_call->{device_group_member}->{host}:$relayed_call->{device_group_member}->{known_time}"
                                                                         : "Manual Synchronization request detected";
                        $self->{logger}->info($reason) if $self->{logger};
                        $self->sync_request_ucs(@{$relayed_call->{rpc_params}});
                    }
                    next;
                }
                if ($self->{persistence_frequency_units} eq "calls") {
                    $persistency_count++;
                }

                # Skip if set_active came from self or lock issue on local known times
                my $known_time = $self->{known_times}{$relayed_call->{originator}};
                if ($relayed_call->{originator} eq $self->{local_device_name} 
                    && (!$relayed_call->{mark_time} || $relayed_call->{time_change_detected} || !$known_time)) {
                    $self->{logger}->debug("Ignoring call from self") if $self->{logger};
                    next;
                }

                # Skip if we're already past this event from this sender
                if ($known_time && $known_time ge $relayed_call->{timestamp} && 
                    !$relayed_call->{time_change_detected}) {
                    $self->{logger}->debug("Already seen event from this sender (originator = $relayed_call->{originator}, call timestamp = $relayed_call->{timestamp}, seen timestamp for originator = $known_time)") if $self->{logger};
                    next;
                }

                my $function_name = $relayed_call->{function_name};

                my $ONE_DAY_AGO = (Time::HiRes::time - (60 * 60 * 24));
                if ($relayed_call->{timestamp} && $relayed_call->{timestamp} lt $ONE_DAY_AGO) {
                    my $timestamp = F5::DateTimeUtils::convert_epoch_to_datetime($relayed_call->{timestamp});
                    $self->{logger}->info("Received old relayed call to '$function_name' from '$relayed_call->{originator}' sent at: $timestamp. Ignoring.") if $self->{logger};
                    next;
                }

                if ($relayed_call->{previous_timestamp} && !$relayed_call->{in_emergency_state}
                    && ( !$known_time || $known_time lt $relayed_call->{previous_timestamp} )
                    && ( !exists($self->{sync_info}->{timestamp}) || $self->{sync_info}->{timestamp} lt $relayed_call->{timestamp}) 
                   ) {
                        $self->{logger}->info("Received out of order call from $relayed_call->{originator}.  Requesting full sync.") if $self->{logger};
                        $self->sync_request_ucs(@{$relayed_call->{rpc_params}});
                        next;
                }                

                #Skip processing if the message would be seen as part of an autosync failover group, but update the known_time
                if ($relayed_call->{failover_device_group} && $relayed_call->{user_name} &&

                    $relayed_call->{user_name} eq $TSCONFD_USER_NAME) {
                    my $fo_dg = F5::DeviceGroup::get_failover_device_group();
                    if ($relayed_call->{failover_device_group} eq $fo_dg->{name}) {
                        $self->{logger}->debug("Received an MCP triggered call from a peer in autosync failover device group '$relayed_call->{failover_device_group}'. Ignoring") if $self->{logger};
                        if ($relayed_call->{mark_time} && 
                            ($relayed_call->{originator} ne $self->{local_device_name})) {

                            $self->{known_times}{$relayed_call->{originator}} = $relayed_call->{timestamp};
                        }
                        next;
                    }
                }
                
                my $orig_rpc_params = dclone($relayed_call->{rpc_params});  # Keep this clean version in case we need to ask for a sync, or relay it back to sender

                $self->{logger}->debug("Begin relayed call to $function_name (originator = $relayed_call->{originator}, call timestamp = $relayed_call->{timestamp})") if $self->{logger};
                eval {
                    $self->$function_name(@{$relayed_call->{rpc_params}});
                };
                if (my $ex = $@) {
                    if (ref($ex) && $ex->{error_code} == F5::ASMConfig::ErrorCode::MCP_VALIDATION_ERROR()) {
                         $self->{logger}->info("Relayed call to $function_name failed MCP validation.  Ignoring on relayed call") if $self->{logger};
                         if ($relayed_call->{mark_time}) {
                             $self->{known_times}{$relayed_call->{originator}} = $relayed_call->{timestamp};
                         }
                    }
                    else {
                        $self->{logger}->debug("Relayed call to $function_name ended in error (originator = $relayed_call->{originator}, call timestamp = $relayed_call->{timestamp})") if $self->{logger};
    
                        # Check is failure is 'expected' due to event from before ucs start end times or if sent from "insane" host
                        if ($relayed_call->{in_emergency_state}) {
                            $self->{logger}->debug("Call received from device in inconsistent state -- ignoring failed call") if $self->{logger};
                            next;
                        }
                        if ($self->{sync_info}->{timestamp}
                                && $self->{sync_info}->{timestamp} ge $relayed_call->{timestamp}) {
                            $self->{logger}->debug("Received call timestamp ($relayed_call->{timestamp}) is earlier than loaded Sync timestamp ($self->{sync_info}->{timestamp}) -- ignoring failed call") if $self->{logger};
                            next;
                        }
                        if ($relayed_call->{originator} eq $self->{local_device_name}) {
                            $self->{logger}->debug("Ignoring failed call that originated on this device");
                            next;
                        }
                        if ($self->{shared_state}{emergency}) {
                            $self->{logger}->warn("Error during '$function_name' while in sync recovery state.  Giving up.  State may be inconsistent with other peers.") if $self->{logger};
                            $self->{shared_state}{emergency} = 0;
                            next;
                        }
                        $self->{logger}->info("Relayed call to $function_name ended in error: Requesting a full sync.") if $self->{logger};
                        $self->sync_request_ucs(@{$orig_rpc_params});
                    }
                }
                elsif ($relayed_call->{mark_time}) {
                    if ($relayed_call->{originator} ne $self->{local_device_name}) {
                        $self->{known_times}{$relayed_call->{originator}} = $relayed_call->{timestamp};
                    }
                    if (exists($self->{sync_info}{$relayed_call->{originator}}) && $self->{sync_info}{$relayed_call->{originator}} gt $relayed_call->{timestamp}) {
                       #forward this back to the originator as they might have missed it
                       $self->{logger}->debug("Sending '$function_name' ($relayed_call->{originator}:$relayed_call->{timestamp}) back to originator as they sent it while receiving a sync") if $self->{logger};
                       $self->send_to_relay(group_id      => $relayed_call->{originator},
                                            priority      => "low",
                                            function_name => $function_name,
                                            rpc_params    => $orig_rpc_params,
                                            );
                    }
                }
                $self->{logger}->debug("End relayed call to $function_name (call $relayed_call->{originator}:$relayed_call->{timestamp}))") if $self->{logger};
            }
            elsif ($self->{shared_state}{emergency} && (time() - $self->{shared_state}{emergency} >= $EMERGENCY_MAX_WAIT_TIME) ) {
                #Defensive coding around erroneous immediate returns from channel poll
                $self->{logger}->warn("Sync recovery state timed out. State may be inconsistent with other peers") if $self->{logger};
                $self->{shared_state}{emergency} = 0;
                if ($self->{sync_info}{out_of_sync_start} && (time() - $self->{sync_info}{out_of_sync_start} >= $MAX_UNRESOLVED_EMERGENCY_WAIT_TIME) ) {
                    $self->{logger}->error("Maximum total time to resolve recovery state exceeded -- restarting sync handler");
                    exit();
                }
            }
            elsif ($startup_sync_device_group_member) {
                if (!$self->{known_times}->{$startup_sync_device_group_member->{host}} ||
                     $startup_sync_device_group_member->{known_time} gt $self->{known_times}->{$startup_sync_device_group_member->{host}}) {
                    $self->{logger}->info("MCP indicates missed Event $startup_sync_device_group_member->{host}:$startup_sync_device_group_member->{known_time}") if $self->{logger};
                    my $call_flags = F5::ASMConfig::CallFlags->new({originator  => $startup_sync_device_group_member->{host},
                                                                    originator_time => $startup_sync_device_group_member->{known_time}});
                    my @rpc_params = (undef, $call_flags);
                    $self->sync_request_ucs(@rpc_params);
                }
                else {
                    $self->{logger}->info("Device " . $self->{local_device_name} . " is in sync with the device group (" . $self->{device_group}  . ")") if $self->{logger};
                }
                undef $startup_sync_device_group_member;
            }
            elsif (my $peer = $self->{shared_state}{data_protection_peer}) {
                $self->{logger}->info("Device " . $self->{local_device_name} . " is requesting Data Protection Configuration from $peer") if $self->{logger};
                my $call_flags = F5::ASMConfig::CallFlags->new({originator  => $peer,
                                                                originator_time => time()});
                my @rpc_params = (undef, $call_flags, undef, );
                $self->sync_request_data_protection(@rpc_params);
                $self->{shared_state}{data_protection_peer} = '';
            }
            elsif ($self->{shared_state}{last_sync_request_time} &&
                   (F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $SYNC_REREQUEST_TIME,
                                                                       start_time => $self->{shared_state}{last_sync_request_time}) == 0)) {
                $self->{logger}->info("Attempting to resolve inconsistent state.") if $self->{logger};
                my @rpc_params = (undef,
                                  F5::ASMConfig::CallFlags->new({originator  => $self->{shared_state}{last_sync_request_host},
                                                                 originator_time => time(),})
                                 );
                $self->sync_request_ucs(@rpc_params);
            }

            $batched_insertor->flush_old_records(interval => 3);
        }
        $self->close_cgc_channels();
    } 
    $batched_insertor->flush;
    # Should not reach here
    die "Relay handler shutdown";
  }
}

#----------------------------------------------------------------------
sub close_cgc_channels {
    my ($self) = @_;

    my $MIN_CHANNEL_TIME = 2;
    my $now = Time::HiRes::time();
    if ($self->{cgc_channel_last_access_time} && (($now - $self->{cgc_channel_last_access_time}) < $MIN_CHANNEL_TIME)) {
        sleep $MIN_CHANNEL_TIME;
    }

    if (exists $self->{cgc_channels}) {
        $self->{logger}->info("Closing CGC listening Channels") if $self->{logger};
        for my $channel_key (keys %{$self->{cgc_channels}}) {
            if ($self->{cgc_channels}->{$channel_key}) {
                eval { $self->{cgc_channels}->{$channel_key}->close(); };
                undef $self->{cgc_channels}->{$channel_key};
            }
        }
        delete $self->{cgc_channels};
        F5::Cgc::Channel::DeviceGroup::reset();
        $self->{cgc_channel_last_access_time} = $now;
    }
    return;
}
#----------------------------------------------------------------------
sub subscribe_to_cgc_channels {
    my ($self, $chan_ref) = @_;
   
    $self->close_cgc_channels();
    $self->{shared_state}{resubscribe} = 0;
    $self->{local_device_name} = $self->{shared_state}->{local_device_name};
    $self->{device_group} = $self->{shared_state}->{device_group};
    $self->{logger}->info("Opening CGC listening Channels for $self->{device_group} on $self->{local_device_name}") if $self->{logger};
    my %services = ($F5::Cgc::Channel::DeviceGroup::ASM_CGC_LOW_PRI_TO_ALL     => $self->{device_group},
                    $F5::Cgc::Channel::DeviceGroup::ASM_CGC_HIGH_PRI_TO_SINGLE => $self->{local_device_name},
                    $F5::Cgc::Channel::DeviceGroup::ASM_CGC_MCP_EVENTS         => $F5::Cgc::Channel::DeviceGroup::ASM_CGC_LOCAL_GROUP,);

    for my $service (keys %services) {
        $chan_ref->{$service} = $self->get_cgc_channel(group_id => $services{$service},
                                                       service  => $service,
                                                       channel_type => "subscriber",);
    }
    $self->{logger}->trace($chan_ref) if $self->{logger};
    return;
}

#----------------------------------------------------------------------

sub clientBegin {
    my($self, $iprot, $oprot) = @_;
    $self->reinit;
}

#----------------------------------------------------------------------

sub load_handler_modules {
  my($self) = @_;

  if ($handler_modules_loaded) {
      return;
  }
  no strict 'refs';
  my @classes;
  my %loaded_methods;

  my @entity_classes = F5::ModuleLoader::load_from_base_class('F5::ASMConfig::Entity::Base');
  my @handler_classes = F5::ModuleLoader::load_from_base_class('F5::ASMConfig::Handler::Base');
  for my $class (@handler_classes) {
    $self->{logger}->debug("Loading RPC functions from $class") if $self->{logger};

    my @handled_functions = eval('@' . $class . '::handled_functions');
    for my $function (@handled_functions) {
        # Accommodate old (pre-hash) handled_functions defintions
        if (! ref($function)) {
            $function = { function_name => $function };
        }

        $loaded_methods{$function->{function_name}} = 1;
        my $local_function_name = 'F5::ASMConfig::Handler::' . $function->{function_name};
        my $remote_function_name = $class . '::' . $function->{function_name};
        my $entity_class = ${$class . "::entity_class"};
        no warnings 'redefine';
        *{$local_function_name} = $self->get_install_function(remote_function_name => $remote_function_name,
                                                              entity_class => $entity_class,
                                                              %$function,
                                                             );
    }
    push(@classes, $class);
  }

  for my $class (@entity_classes) {
      my $entity_function_name = $class;
      $entity_function_name =~ s/.*:://;
      $entity_function_name =~ s/([a-z0-9])([A-Z])/${1}_${2}/g;
      $entity_function_name = lc $entity_function_name;

      if ($class->can('element_type') && $class->element_type) {
          $self->{element_types}->{$class->element_type} =  { element_type_name => $class->element_type_name,
                                                              class => $class,
                                                              entity_function_name => $entity_function_name,
                                                            };
      }

      my $relayable = ($class->can('relayable')) ? $class->relayable()
                                                 : 1;
      my %standard_criteria_and_store_locations = (criteria_index => 3,  
                                                   store_index    => 4);

      my %base_functions = ( add               => {relayable => $relayable,
                                                   store_index => 3,
                                                   write_access => 1,},
                             get               => {} ,
                             get_default       => {} ,
                             delete            => {relayable => $relayable,
                                                   write_access => 1,},
                             update            => {relayable => $relayable,
                                                   write_access => 1,
                                                   %standard_criteria_and_store_locations},
                             update_dont_check => {plural_function_name => 1,
                                                   write_access => 1,
                                                   relayable => $relayable,},
                             search            => {},
                             reset_defaults    => {plural_function_name => 1,
                                                   write_access => 1,
                                                   relayable => $relayable,},
                             update_many       => {plural_function_name => 1,
                                                   relayable => $relayable,
                                                   write_access => 1,
                                                   %standard_criteria_and_store_locations,},
                             delete_many       => {plural_function_name => 1,
                                                   write_access => 1,
                                                   relayable => $relayable,},
                           );

      for my $action (keys %base_functions) {
          my $function_name = $action . '_' . $entity_function_name;
          if ($base_functions{$action}->{plural_function_name}) {
              if ($class->can('element_type_name_plural') && $class->element_type_name_plural ){
                  $function_name = $action . '_' . $class->element_type_name_plural;
              }
          }
          next if ($loaded_methods{$function_name});

          $loaded_methods{$function_name} = 1;
          my $local_function_name = 'F5::ASMConfig::Handler::' . $function_name;
          no warnings 'redefine';
          *{$local_function_name} = $self->get_install_function(method => "base_$action",
                                                                entity_class   => $class,
                                                                function_name  => $function_name,
                                                                %{$base_functions{$action}},
                                                               );
      }
  }
  F5::ASMConfig::Entity::Base->set_element_type_cache($self->{element_types});
  $handler_modules_loaded = 1;
}

#----------------------------------------------------------------------

sub get_install_function {
    my($self, %params) = @_;
    my $remote_function_name = $params{remote_function_name};
    my $method = $params{method};
    my $entity_class = $params{entity_class};
    my @policy_param_indexes;
    if (!$params{authz_params}) {
        if ($entity_class && $entity_class->can("get_policies")) {
            $params{authz_params} = [($params{criteria_index} || $params{store_index} || $POLICY_INDEX + 1 )];
        }
        else {
            $params{authz_params} = [$POLICY_INDEX];
        }
    }

    my $install_function = sub {
        my($self, @params) = @_;

        $self->{logger}->debug("Begin $params{function_name}") if $self->{logger};
        if ($params{function_name} ne "sync_receive_file_part" && $params{function_name} ne 'add_request_log_protobuf') {
            $self->{logger}->trace(\@params) if $self->{logger};
        }
        if (!$params[$CALL_FLAG_INDEX]) {
            $params[$CALL_FLAG_INDEX] = F5::ASMConfig::CallFlags->new;
        }

        $self->{session} = $params[0];
        my $call_flags = $params[$CALL_FLAG_INDEX];

        my $singleton_process_lock;
        if ($call_flags->{singleton_process_lock_name}) {
            $singleton_process_lock = F5::LockFactory::get_locker(lock_filename => $call_flags->{singleton_process_lock_name},
                                                                  logger => $self->{logger},);
            my $has_singleton_process_lock = $singleton_process_lock->lock(dont_block => 1);
            if (! $has_singleton_process_lock) {
                $self->{logger}->info("Another process with lock $call_flags->{singleton_process_lock_name} is still running -- skipping") if $self->{logger};
                return;
            }
        }

        if (! $self->{is_async_process} && (! $call_flags || ! $call_flags->{from_rpc_queue})) {
            $self->check_memory_size();
        }

        if ( $call_flags && $call_flags->{act_as_gui_user_name} ) {
            $self->{session}->{gui_user_name} = $call_flags->{act_as_gui_user_name};
        }

        if(not defined $self->{session}->{partition}) {
            $self->{session}->{partition} = F5::ASMConfig::Constants::PARTITION_COMMON();
            $call_flags->{partition_was_defined} = 0;
        } elsif (not defined $call_flags->{partition_was_defined}) {
            $call_flags->{partition_was_defined} = 1;
        }

        if ($call_flags->{internal_flags} && @{$call_flags->{internal_flags}}) {
            for my $internal_flag (@{$call_flags->{internal_flags}}) {
                $call_flags->{internal_flags_hash}->{$internal_flag} = 1;
            }
        }

        my $remote_device_ip;
        if ($call_flags->{remote_device}) {
            $remote_device_ip = $call_flags->{remote_device}->{mng_ip_address};
            undef($call_flags->{remote_device});
        }
         

        #For EM Backwards compatiblity
        if ($call_flags->{extra_fields} && grep { $_ eq "extra_virtual_servers" } @{$call_flags->{extra_fields}}) {
            $self->{em_client} = 1;
        }

        if (! $self->{send_events_to_dispatcher}) {
            $call_flags->{suppress_events} = 1;
            $call_flags->{send_events_to_dispatcher} = 0;
        }

        if (!$call_flags->{force_populate_rest_uuid} && !F5::BigipUtils::is_feature_enabled('populate_rest_uuids')) {
            $call_flags->{populate_rest_uuid} = 0;
        }

        if (defined $params[$POLICY_INDEX] && ref($params[$POLICY_INDEX]) && $params[$POLICY_INDEX]->{policy_name}) {
            $self->from_em_slash_replace(obj => $params[$POLICY_INDEX],
                                         field_name => 'policy_name');
        }

        my $shifted_index = 0;
        if (($entity_class && $entity_class eq "F5::ASMConfig::Entity::Policy") ||
            (defined $params[$POLICY_INDEX] && ref($params[$POLICY_INDEX]) !~ /^F5::ASMConfig::(?:Entity::)?Policy$/)) {
            $shifted_index = 1;
        }

        if (! $self->{is_relay_handler_process} && ! $rpc_process_priority_was_set) {
            # Only load once per child RPC process

            my $asm_config_process_priority_cfg = YAML::XS::LoadFile($ASM_CONFIG_PROCESS_PRIORITY_FILENAME);
            if ($asm_config_process_priority_cfg->{user_name} &&
                $self->{session} &&
                $self->{session}->{user_name} && 
                $asm_config_process_priority_cfg->{user_name}->{$self->{session}->{user_name}} &&
                $asm_config_process_priority_cfg->{user_name}->{$self->{session}->{user_name}}->{nice}) {

                my $nice = $asm_config_process_priority_cfg->{user_name}->{$self->{session}->{user_name}}->{nice};
                $self->{logger}->debug("setting nice:[$nice] for pid:[$$]") if $self->{logger};
                setpriority(0, $$, $nice);
            }
            $rpc_process_priority_was_set = 1;
        }

        if (!$self->{is_relay_handler_process} && !$self->{is_async_process} && $call_flags->{is_asynchronous}) {
            $self->pass_call_to_asynchronous_process(function_name => $params{function_name},
                                                     rpc_params => \@params);
            $self->{logger}->debug("End $params{function_name} (after async handoff)") if $self->{logger};
            my $return_index = -1;
            
            if (defined $params{store_index}) {
                $return_index = $params{store_index};
                
                # NOTE: before splice
                if ($shifted_index) {
                    $return_index -= 1;
                }
                
                # Cope with wrong store_index in deprecated methods (e.g. update_parameter_add_update_negsigs)
                $return_index = -1 unless $params[$return_index];
            }

            return $params[$return_index];
        }

        if ($remote_device_ip) {
            my $function_name = $params{function_name};
            my $client = F5::ASMConfig::Client->new(device => F5::ASMConfig::Device->new({mng_ip_address => $remote_device_ip}), is_ssl => 1);
            my $rv;

            eval {
                $rv = $client->$function_name(@params);
            };
            if ($@) {
                if (ref($@) && (ref($@) eq 'F5::ASMConfig::ASMConfigException')) {
                    $self->{logger}->debug("Failed to call \"$params{function_name}\" on \"$remote_device_ip\"") if $self->{logger};
                    $self->{logger}->debug($@) if $self->{logger};
                    die $@;
                }
                else {
                    return $self->handle_error(action => $function_name,
                                               error_thrown => $@,
                                               error_code => F5::ASMConfig::ErrorCode::UNREACHABLE_HOST,
                                               call_flags => $call_flags,
                                               entity_class => $params{entity_class},
                                              );

                    $self->log_error_and_rollback(exception => $@);
                }
            }
            return $rv;
        }

        $self->{rpc_entry_depth}++;

        my $err;
        my $retval;
        my $locker;
        for my $attempt (1..$MAX_DEADLOCK_RETRY_COUNT) {
            my $attempt_count_str = "";
            if ($attempt > 1) {
                $attempt_count_str = ", attempt $attempt";
            }

            $self->{dbh} = F5::DbUtils::refresh_dbh(dbh => $self->{dbh},
                                                    db_opts => {
                                                                RaiseError => 1,
                                                                PrintError => 0,
                                                               }
                                                   );

            if ($self->{batched_insertor}) {
                $self->{batched_insertor}->set_dbh( $self->{dbh} );
            }
            if ($self->{action_item_handler}) {
                $self->{action_item_handler}->set_dbh(dbh => $self->{dbh} );
            }

            $self->{logger}->trace("Check transaction at call start (rpc_entry_depth = $self->{rpc_entry_depth}, in_transaction = $self->{in_transaction})") if $self->{logger};
            if (!$self->{in_transaction} && $self->{rpc_entry_depth} == 1) {
                F5::ASMConfig::Cache::soft_reset();
                delete $self->{rpc_calls_to_relay};
                if ($params{exclusive_lock_file}) {
                    $locker = F5::LockFactory::get_locker(lock_filename => $params{exclusive_lock_file},
                                                          logger => $self->{logger},);
                    $locker->lock();
                }
                $self->{logger}->trace("Beginning transaction") if $self->{logger};
                $self->{dbh}->begin_work;
                # Refresh COAPI handle to MCP
                F5::ASM::COAPI::Login("root");
                $self->{events} = [];
                $self->{send_pbconfig_event} = {};
                $self->{post_commit_hooks} = [];
                $self->{section_policy_cache} = {};
            }
            elsif ($self->{in_transaction} && $params{exclusive_lock_file}) {
                $self->{transaction_locker}->{$params{exclusive_lock_file}} ||= F5::LockFactory::get_locker(lock_filename => $params{exclusive_lock_file},
                                                                                                            logger => $self->{logger},);
                $self->{transaction_locker}->{$params{exclusive_lock_file}}->lock(single_lock => 1);
            }

            if ($self->{rpc_entry_depth} == 1) {
                F5::ASM::COAPI::Login("root");
                $self->{authz_checked} = 0;
            }

            #Ensure integrity of RPC arguments
            # todo: move to handlers/entities?
            if ( !($self->{is_relay_handler_process} && $self->{rpc_entry_depth} == 1) && $shifted_index) {
                splice(@params, $POLICY_INDEX, 0, undef);
            
            }

            if (defined $params[$POLICY_INDEX] && ref($params[$POLICY_INDEX]) =~ /^F5::ASMConfig::(?:Entity::)?Policy$/ ) {
                $params[$POLICY_INDEX] = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, policy => $params[$POLICY_INDEX]);
            }

            #Ensure criteria is populated if this is an old style call
            if (defined($params{criteria_index})) {
                if (ref($params[$params{criteria_index}]) && !ref($params[$params{store_index}])) {
                    $params[$params{store_index}] = dclone($params[$params{criteria_index}]);
                }
            }

            $self->{device_group} = $self->{shared_state}{device_group};
            $self->{local_device_name} = $self->{shared_state}{local_device_name};

            if ($params{filename_arguments} && $self->{is_relay_handler_process}) {
                $self->extract_from_params(rpc_params   => \@params, 
                                           arguments    => $params{filename_arguments},
                                           strip_paths  => 1,
                                           prepend_path => $SYNC_DIR);
            }

            local $SIG{ALRM} = sub { die "Timeout triggered\n" };
            eval {
                if ($self->{rpc_entry_depth} == 1 && $call_flags->{call_timeout}) {
                    alarm $call_flags->{call_timeout};
                }

                if ($self->{asmconfig_check_authz} && !$self->{authz_checked} && !$params{skip_authz}) {
                    $self->check_authz(rpc_params => \@params,
                                       %params,
                                      );
                }

                if ($self->{session}->{user_name} && 
                    $self->{session}->{user_name} eq $TSCONFD_USER_NAME && 
                    $call_flags->{is_asynchronous}) {

                    my $last_config_load_time = $self->{shared_state}{last_config_load_time};
                    if ($call_flags->{call_time} && $last_config_load_time &&
                        $call_flags->{call_time} < $last_config_load_time) {

                        $self->{logger}->debug("Ignoring obsolete tsconfd call to $remote_function_name") if $self->{logger};
                        $self->{logger}->trace("Last load config time was " . $last_config_load_time) if $self->{logger};
                        $self->{logger}->trace("Current tsconfd call time is " . $call_flags->{call_time}) if $self->{logger};

                        # similar to how we die, to ignore, in check_ucs_load
                        die { error_level => "info",
                              error_message => "Received updates from tsconfd that preceed the last config load time.  Ignoring.",
                              error_code =>  F5::ASMConfig::ErrorCode::ILLEGAL_OPERATION(),
                            };
                    }

                    #Reject asynchronous calls during UCS load.  (updates for CMI informations (device group data) is synchronous)
                    $self->check_ucs_load();
                }

                if ($remote_function_name) {
                    $self->{logger}->debug("Calling $remote_function_name" . $attempt_count_str) if $self->{logger};
                    { 
                        no strict 'refs';
                        $retval = &{$remote_function_name}($self, @params);
                    }
                    $self->{logger}->debug("Back from $remote_function_name") if $self->{logger};
                }
                else {
                    $self->{logger}->debug("Calling $method" . $attempt_count_str) if $self->{logger};
                    $retval = $self->$method(rpc_params => \@params,
                                             entity_class => $entity_class);
                    $self->{logger}->debug("Back from $method") if $self->{logger};
                }

                if ($retval){
                    if ($retval->{events}) {
                        push @{$self->{events}}, @{$retval->{events}};
                    }
                }

                if ($call_flags->{send_result_as_event_type}) {
                    my %event_opts = (
                                      session => $self->{session},
                                      policy => $params[$POLICY_INDEX],
                                      call_flags => $call_flags,
                                      suppress_save_to_log => 1,
                                      do_policy_version_update => 0,
                                      skip_inheritance => 1,
                                     );

                    if ($retval && $retval->{return_value} && ! $retval->{ignored_error}) {
                        push(@{$self->{events}}, $self->create_event(%event_opts,
                                                                     element => $retval->{return_value},
                                                                     event_type => $call_flags->{send_result_as_event_type},
                                                                    ));
                    }
                    else {
                        my $element;
                        if ($retval && $retval->{return_value} && $entity_class) {
                            $element = $entity_class->new(object => $retval->{return_value});
                        }

                        push(@{$self->{events}}, $self->create_event(%event_opts,
                                                                     element => $element,
                                                                     event_type => $call_flags->{send_no_result_as_event_type} || F5::ASMConfig::EventType::ASYNC_NO_RESULT,
                                                                    ));
                    }
                }
                if ($retval && $retval->{events} && @{$retval->{events}} && $params{relayable} && $entity_class && $entity_class->add_pbconfig_event()) {
                    $self->add_pbconfig_event($self->{session}, $call_flags, $params[$POLICY_INDEX]);
                }
            };
            $err = $@;
            if ( $err && ref($err) && reftype($err) eq 'HASH' && ($err->{error_code} == F5::ASMConfig::ErrorCode::DEADLOCK_FOUND()
                || $err->{error_code} == F5::ASMConfig::ErrorCode::LOCK_TIMEOUT()) && $self->{rpc_entry_depth} == 1 ) {

                if (! $self->{dbh}->{AutoCommit}) {
                    $self->{dbh}->rollback();
                }
                next;
            }
            last;
        }

        if ($self->{rpc_entry_depth}) {
            $self->{rpc_entry_depth}--;
        }
        #Word of warning: Do not add any calls out to handlers on the $self after this block as it will clear out everything from the call (rpc_entry_depth can be 0)

        if ($self->{rpc_entry_depth} == 0 && $call_flags->{call_timeout}) {
            alarm 0;
        }       

        if ($err) {
            $self->log_error_and_rollback(exception => $err);
        }

        my $relayable = $call_flags->{relayable} || $params{relayable} || 0;
        if ($relayable == F5::ASMConfig::RelayabilityType::RELAYABILITY_TYPE_CHECK_RELAYABLE()) {
            $relayable = $self->check_relayable(entity_class => $entity_class);
        }

        $self->{logger}->trace("Check transaction at call end (rpc_entry_depth = $self->{rpc_entry_depth}, in_transaction = $self->{in_transaction})") if $self->{logger};
        if (! $self->{in_transaction} && $self->{rpc_entry_depth} == 0) {
            if (!$self->{batched_insertor}) {
                $self->save_events_to_config_log(events => $self->{events});
            }
            if ( ! $self->{dbh}->{AutoCommit} ) {
                eval {
                    $self->{logger}->trace("Committing transaction") if $self->{logger};
                    $self->{dbh}->commit; 
                };
                if ($@) {
                    $self->log_error_and_rollback(exception => F5::ASMConfig::ASMConfigException->new(error_message => $@,
                                                                                                      error_code => F5::ASMConfig::ErrorCode::MCP_VALIDATION_ERROR()),
                                                  );
                }
            }
        }

        my $writes_to_mcp = defined $call_flags->{writes_to_mcp} ? $call_flags->{writes_to_mcp}
                                                                 : $params{writes_to_mcp};
        if ($self->{session}->{gui_user_name} && !$call_flags->{skip_mcp} &&
            ($writes_to_mcp || ( !defined($writes_to_mcp) &&
               ( ($params{write_access} || ( ! defined $params{write_access} && $relayable))
                 && $retval->{return_value} && ref($retval->{return_value}) =~ /F5::ASMConfig::Entity::/ && $retval->{return_value}->is_mcp_object )
               )
            )) {
            $self->{persist_to_mcp} = 1;
            delete $call_flags->{writes_to_mcp};
        }

        if ($retval && ref($retval->{return_value}) =~ /F5::ASMConfig::Entity/) {
            $retval->{return_value}->add_post_commit_hook(handler => $self, call_flags => $call_flags, writes_to_mcp => $writes_to_mcp);
        }

        eval {
            if (! $self->{in_transaction} && $self->{rpc_entry_depth} == 0) {
                $self->post_commit_actions(events => $self->{events}, call_flags => $call_flags);
                $self->{send_pbconfig_event} = {};
                $self->{post_commit_hooks} = [];

                $locker->unlock() if $locker;
                $singleton_process_lock->unlock() if $singleton_process_lock;
                if ($self->{transaction_locker}) {
                    for my $locker (values %{$self->{transaction_locker}}) {
                        $locker->unlock_all();
                    }
                    delete $self->{transaction_locker};
                }
                delete $self->{transactional_saves};

                if ($self->{batched_insertor}) {
                    $self->save_events_to_config_log(events => $self->{events});
                }
                $self->write_event_messages_to_syslog(events => $self->{events});
                if (! defined $call_flags->{send_events_to_dispatcher} || $call_flags->{send_events_to_dispatcher}) {
                    $self->send_events_to_dispatcher(events => $self->{events});
                }
                $self->{events} = [];
            }
        };
        if ($@) {
            $self->log_error_and_rollback(exception => $@);
        }

        #Send to relay if this is not the relay handler, and function relayable flag on
        if (!$call_flags->{do_not_relay} &&
            $self->{device_group} &&
            !$self->{is_relay_handler_process} &&
            ($relayable || $call_flags->{relayable}) &&            
            ($self->{rpc_entry_depth} == 0 || ($self->{relay_transaction_rpc_depth} && ($self->{rpc_entry_depth} == $self->{relay_transaction_rpc_depth}))) &&
            $self->{device_group} eq $self->{shared_state}{device_group} &&
            !($retval && $retval->{ignored_error}) &&    # Never relay calls that failed with ignored errors
            (! F5::Cluster::Utils::is_clustering_enabled() || F5::Cluster::Utils::is_master_blade()) # Only forward from a primary device 
           ) {
            my $can_update_mcp;

            if (F5::BigipUtils::is_device_bigip()) {
                my $asm_config_settings = F5::ASMConfig::Entity::AsmConfigSettings->new( dbh => $self->{dbh})->retrieve(get_criteria => {});
                $can_update_mcp = $asm_config_settings->{flg_recommend_sync_when_policy_is_not_applied} || $params{recommend_sync} || $self->{shared_state}{sync_init};
            }
            if ($self->{shared_state}{is_auto_sync} && !$self->{shared_state}{sync_init}) {

                my @rpc_calls_to_relay;
                if ($self->{rpc_calls_to_relay} && $self->{rpc_calls_to_relay}->{$params{function_name}}) {
                    @rpc_calls_to_relay = @{delete $self->{rpc_calls_to_relay}->{$params{function_name}}};
                }
                else {
                    @rpc_calls_to_relay = ({function_name => $params{function_name},
                                            rpc_params    => \@params,});
                }

                # skip mcp in relay for synced folders
                my $relayed_skip_mcp = 0; # means do not change in relayed call
                if ($retval && $retval->{return_value} &&
                    ref($retval->{return_value}) =~ /F5::ASMConfig::Entity::/ &&
                    $retval->{return_value}->mcp_folder() &&
                    F5::DeviceGroup::is_folder_synchronized(device_group => $self->{device_group},
                                                            folder => $retval->{return_value}->mcp_folder(),
                                                            logger => $self->{logger},
                                                           )
                   ) {
                    $relayed_skip_mcp = 1; # means set the flag in relayed call
                }

                # consider import policy result retval
                if (!$relayed_skip_mcp && $retval &&
                    $retval->{return_value} && ref($retval->{return_value}) =~ /F5::ASMConfig::ImportPolicyRes/ &&
                    $retval->{return_value}->{policy} && ref($retval->{return_vailue}->{policy}) =~ /F5::ASMConfig::Entity::Policy/ &&
                    $retval->{return_value}->{policy}->mcp_folder() &&
                    F5::DeviceGroup::is_folder_synchronized(device_group => $self->{device_group},
                                                            folder => $retval->{return_value}->{policy}->mcp_folder(),
                                                            logger => $self->{logger},
                                                           )
                   ) {
                    $relayed_skip_mcp = 1; # means set the flag in relayed call
                }

                for my $rpc_call_to_relay (@rpc_calls_to_relay) {
                    my $function_name = $rpc_call_to_relay->{function_name};
                    my @rpc_params = @{$rpc_call_to_relay->{rpc_params}};
                    my $relayed_call_flags = $rpc_params[$CALL_FLAG_INDEX];

                    $relayed_call_flags->{skip_mcp} = 1 if $relayed_skip_mcp;

                    if (defined $rpc_params[$POLICY_INDEX]) {
                        $relayed_call_flags->{originating_policy_version_time} = $rpc_params[$POLICY_INDEX]->{policy_version_time};

                        if (ref($rpc_params[$POLICY_INDEX]) eq "F5::ASMConfig::Entity::Policy") {
                            $rpc_params[$POLICY_INDEX] = dclone($rpc_params[$POLICY_INDEX]);
                            $rpc_params[$POLICY_INDEX]->remove_non_unique_record_parts(dbh => $self->{dbh});
                        }
                    }
                    $relayed_call_flags->{originator} = $self->{local_device_name};
                    $relayed_call_flags->{originator_zero_padded_major_minor_version} = $self->{zero_padded_major_minor_version};
                    $relayed_call_flags->{originating_policy_version_device_name} = $self->{shared_state}->{local_hostname};
                    $relayed_call_flags->{originator_is_failover_standby} = F5::Cluster::Utils::is_failover_standby();

                    if ($params{filename_arguments}) {
                        for my $filename ($self->extract_from_params(rpc_params => \@rpc_params, arguments => $params{filename_arguments})) {
                            $self->sync_send_file(@rpc_params[0,1,2], $filename);
                        }

                        if ( $call_flags->{sync_files} ) {
                             for my $filename ( @{$call_flags->{sync_files}} ){
                                $self->sync_send_file(@rpc_params[0,1,2], $filename);
                             }
                        }
                    }

                    if ($params{criteria_index} && $rpc_params[$params{criteria_index}] && $entity_class) {
                        $rpc_params[$params{criteria_index}] = dclone($rpc_params[$params{criteria_index}]);
                        $entity_class->purify_store_object(object => $rpc_params[$params{criteria_index}], dbh => $self->{dbh});
                    }
                    if ($params{store_index} && $entity_class) {
                        $rpc_params[$params{store_index}] = dclone($rpc_params[$params{store_index}]);
                        $entity_class->purify_store_object(object => $rpc_params[$params{store_index}], dbh => $self->{dbh});
                    }

                    my $mark_time = defined($params{update_mcp_timestamp}) ? $params{update_mcp_timestamp} : 1;  #don't mark time for set_active, send_cluster_sync calls

                    my %relay_params = (function_name => $function_name,
                                        rpc_params    => \@rpc_params,
                                        group_id      => $self->{device_group},
                                        priority      => "low",
                                        mark_time     => $mark_time,
                                        can_update_mcp => $can_update_mcp);
                    if ($self->{in_transaction}) {
                         push @{$self->{transaction_relay_queue}}, \%relay_params;
                    }
                    else {
                        $self->send_to_relay(%relay_params);
                    }
                }
            }
            else { 
                #Update the MCP timestamp for every would-be-relayed transaction if we are in a manual_sync device_group
                if (!$self->{shared_state}{is_auto_sync}) {
                    $self->{known_times}{$self->{local_device_name}} = Time::HiRes::time;
                }
                if ($can_update_mcp) {
                    $self->{update_mcp_timestamp} = 1;
                }
            }
        }

        $self->{logger}->debug("End $params{function_name}") if $self->{logger};

        #linked RPC commands to run after current call has finished and committed. (Async for tasks)
        if ($self->{rpc_entry_depth} == 0 && (!$self->{in_transaction} || $call_flags->{bulk_call})) {
             while (my $rpc_call = shift (@{$self->{rpc_queue}})) {
               my $queued_function = $rpc_call->{function_name};
               my $queued_rpc_params = $rpc_call->{rpc_params};
               #Set Async flag
               $queued_rpc_params->[$CALL_FLAG_INDEX]->{is_asynchronous} = $rpc_call->{is_asynchronous};
               $queued_rpc_params->[$CALL_FLAG_INDEX]->{from_rpc_queue} = 1;

               eval {
                   $self->$queued_function(@$queued_rpc_params);
               };
               if (my $err = $@) {
                   my $err_msg = $err;
                   my $die_after_log = 0;
                   my %log_options;
                   if (ref($err) && (ref($err) eq 'F5::ASMConfig::ASMConfigException')) {
                       $err_msg = $err->{display_message} || $err->{error_message};

                       if ($err->{error_code} == F5::ASMConfig::ErrorCode::ASYNC_SOCKET_FULL) {
                           $die_after_log = 1;
                           %log_options = (throttled_log_level => 'debug',
                                           throttling_interval => 10);
                       }
                   }

                   $self->{logger}->log(msg => "Could not complete call to $queued_function.  Error: $err_msg",
                                        log_level => 'info',
                                        %log_options) if $self->{logger};
                   if ($die_after_log) {
                       $self->log_error_and_rollback(do_log => 0,
                                                     exception => $err);
                   }
               }
            }
        }
        my $rv =  $retval ?  $retval->{return_value} : undef;

        $verbose_logger->trace("About to return value") if $verbose_logger;
        $verbose_logger->trace($retval) if $verbose_logger;

        return $rv;
    };

    return $install_function;
}

#----------------------------------------------------------------------

my %TYPE_CALL_MAPPING = ( F5::ASMConfig::EventType::ADD() => 'add',
                          F5::ASMConfig::EventType::UPDATE() => 'update',
                          F5::ASMConfig::EventType::DELETE() => 'delete',);

sub get_function_name {
    my ($self, %params) = @_;
    my $element_type = $params{element_type};
    my $event_type = $params{event_type};
    my $is_many = $params{is_many};
    my $class = $self->{element_types}->{$element_type}->{class};

    my $function_name;

    if ( $is_many ){
        $function_name = $TYPE_CALL_MAPPING{$event_type} . '_many_'. $class->element_type_name_plural;
    }
    else {
        $function_name = $TYPE_CALL_MAPPING{$event_type} . '_' . $class->element_type_name;
    }

    return $function_name;
}

#----------------------------------------------------------------------

sub get_cached_inheriting_policy_ids {
    my ($self, %params) = @_;
    my $policy = $params{policy};
    my $section_type = $params{section_type};

    if (!exists $self->{section_policy_cache}->{$policy->{policy_id}}->{$section_type}) {
        $self->{section_policy_cache}->{$policy->{policy_id}}->{$section_type} = {};
        if (!$policy->{child_policy_ids}) {
            $policy->{child_policy_ids} = $policy->child_policy_ids(dbh => $self->{dbh}, policy_id => $policy->{policy_id});
        }
    
        my $section_search_filters = [ F5::ASMConfig::SearchFilter->new(field_name => "policy_id",
                                                                        field_value_list => $policy->{child_policy_ids},
                                                                        field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN()),
                                       F5::ASMConfig::SearchFilter->new(field_name => "child_inheritance_status",
                                                                        field_value_list => [F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_MANDATORY(),
                                                                                             F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_ACCEPTED()],
                                                                        field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN()),
                                     ];
    
        my $section_iter = F5::ASMConfig::Entity::Section->retrieve_all(dbh => $self->{dbh},
                                                                        search_filters => $section_search_filters,
                                                                        get_criteria => { section_type => $section_type });
    
        while (my $section = $section_iter->get_next()) {
            $self->{section_policy_cache}->{$policy->{policy_id}}->{$section_type}->{$section->{policy}->{policy_id}} = $section->{policy};
        }
    }

    return $self->{section_policy_cache}->{$policy->{policy_id}}->{$section_type};
}

#----------------------------------------------------------------------

sub propagate_inheritance_rpc {
    my($self, %params) = @_;
    my $session       = $self->{session};
    my $policy        = $params{policy};
    my $function_name = $params{function_name},
    my $call_flags    = F5::ASMConfig::CallFlags->new(%{ $params{call_flags} || {} });
    my $criteria_obj  = $params{criteria_obj};
    my $store_obj     = $params{store_obj};
    my $section_type  = $params{section_type};
    my $use_policy_filter  = $params{use_policy_filter};

    return unless ($policy && $policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT());
    
    my @inheriting_policy_ids = keys %{$self->get_cached_inheriting_policy_ids(%params)};
    return unless @inheriting_policy_ids;

    $call_flags->{inherited_call} = 1;
    
    my $store_entity = dclone($store_obj);
    $store_entity->remove_non_inheritance_fields();
    $store_entity->set_inherited();

    if ($use_policy_filter) {
        $call_flags->{search_filters} ||= [];
        push(@{$call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new(
                                                                           field_name => "policy_id",
                                                                           field_value_list => \@inheriting_policy_ids,
                                                                           field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN()));
    }
    for my $policy_id (@inheriting_policy_ids) {
        my @rpc_params = ($session, $call_flags, $use_policy_filter ? undef : F5::ASMConfig::Policy->new(policy_id => $policy_id));
        if ($criteria_obj) {
            push (@rpc_params, $criteria_obj);
        }

        push (@rpc_params, $store_entity);
        $self->$function_name(@rpc_params);
        last if $use_policy_filter;
    }
}

#----------------------------------------------------------------------

sub propagate_inheritance_events {
    my($self, %params) = @_;
    my $session = $self->{session};
    my $policy = $params{policy};
    my $events = $params{events};
    my $call_flags = $params{call_flags} || F5::ASMConfig::CallFlags->new();

    my $local_events = [];   
    return $local_events unless ($policy && $policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT());

    my $saved_events = delete($self->{events});
    my %section_policy_cache;

    $call_flags->{inherited_call} = 1;
    $call_flags->{add_on_updating_nonexistent_record} = 1;
    $call_flags->{log_note} = "Inherited from $policy->{policy_name}";

    for my $event (@$events) {
        my $element_type = $event->{element_type};
        my $element_type_name = $self->{element_types}->{$element_type} ? $self->{element_types}->{$element_type}->{element_type_name} : undef;
        my $entity = $element_type_name ? $event->{element}->{$element_type_name} : undef;
        next if (!$entity || !$entity->can('skip_inheritance') || $entity->skip_inheritance(event => $event));

        if ($entity && (my $section_type = $entity->get_section_type())) {
            if ($self->get_cached_inheriting_policy_ids(policy => $policy, section_type => $section_type)) {
                if ($event->{event_type} == F5::ASMConfig::EventType::ADD()) {
                    my $function_name = $self->get_function_name(element_type => $element_type,
                                                                 event_type => F5::ASMConfig::EventType::UPDATE(),);
                        #TODO-Layering: Ensure required entities are stored first (ie. URL for Login Pages)
                        $self->propagate_inheritance_rpc(function_name => $function_name,
                                                         call_flags => $call_flags, 
                                                         policy => $policy,
                                                         store_obj => $entity,
                                                         section_type => $section_type,);
                }
                else {
                    my $use_cross_policy_filter = $entity->can_update_many_cross_policy(event => $event);
                    my $criteria_entity;
                    if (@{$entity->unique_record_parts()}) {
                        $criteria_entity = dclone($entity);
                        $criteria_entity->remove_non_unique_record_parts();
                    }
                    else {
                        $criteria_entity = $self->{element_types}->{$element_type}->{class}->new();
                    }

                    my $function_name = $self->get_function_name(element_type => $element_type,
                                                                 event_type => $event->{event_type},
                                                                 is_many => $use_cross_policy_filter);

                    $self->propagate_inheritance_rpc(function_name => $function_name,
                                                     call_flags => $call_flags,
                                                     policy => $policy,
                                                     criteria_obj => $criteria_entity,
                                                     store_obj => $entity,
                                                     section_type => $section_type,
                                                     use_policy_filter => $use_cross_policy_filter,);
                } 
            }
        }
    }
    if ($self->{events}) {
        $local_events = delete $self->{events};
    }
    $self->{events} = $saved_events;
    return $local_events;
}

#----------------------------------------------------------------------

sub check_relayable {
    my ($self, %params) = @_;
    my $relayable = F5::ASMConfig::RelayabilityType::RELAYABILITY_TYPE_IS_NOT_RELAYABLE;
    my $entity_class = $params{entity_class};
    if ($self->{events} && $entity_class) {
        my $element_type_name = $entity_class->element_type_name;
        for my $event (@{$self->{events}}) {
            if ($event->{element}->{$element_type_name} && $event->{element}->{$element_type_name}->relayable() == F5::ASMConfig::RelayabilityType::RELAYABILITY_TYPE_IS_RELAYABLE()) {
                $relayable = F5::ASMConfig::RelayabilityType::RELAYABILITY_TYPE_IS_RELAYABLE();
                last;
            }
        }
    }
    return $relayable;
}

#----------------------------------------------------------------------

sub check_authz {
    my($self, %params) = @_;
    my $rpc_params = $params{rpc_params};
    my $session = $self->{session};
    $self->{authz_checked} = 1;

    if ($self->{is_relay_handler_process} || ! $session->{gui_user_name}) {
        # If no gui_user_name is specified or this is a relayed command, allow all access
        return 1;
    }

    my $access_type = 'read';
    if (($params{relayable} && ! defined $params{write_access}) || $params{write_access}) {
        $access_type = 'write';
    }

    # Determine user role
    my $user;
    if ($session->{gui_user_name} eq 'root') {
        # iControl accessed from localhost is reported as root and should be granted admin access
        $user = [{ role => F5::Mcp::UserRole::ROLE_ADMINISTRATOR }];
    }
    else {
        $user = F5::ASM::COAPI::Query->new('user_role_partition', { user => $session->{gui_user_name} });
        if (! $user || ! @$user) {
            return $self->handle_error(error_thrown => "User $session->{gui_user_name} does not exist",
                                       error_code => F5::ASMConfig::ErrorCode::NONEXISTENT_USER,
                                      );
        }
    }

    my %allowed_partitions;
    for my $user_role (@$user) {

        # TODO take the default allowed min roles for read/write

        if ($user_role->{role} == F5::Mcp::UserRole::ROLE_ADMINISTRATOR ||
            $user_role->{role} == F5::Mcp::UserRole::ROLE_APPLICATION_SECURITY_ADMINISTRATOR || 
            $user_role->{role} == F5::Mcp::UserRole::ROLE_RESOURCE_ADMIN) {

            # Can access all partitions
            $session->{has_admin_access} = 1;
            return 1;
        }
        elsif (($user_role->{role} == F5::Mcp::UserRole::ROLE_APPLICATION_SECURITY_EDITOR ||
		$user_role->{role} == F5::Mcp::UserRole::ROLE_APPLICATION_SECURITY_OPERATIONS_ADMINISTRATOR)
		&& $access_type eq 'write') {
            $allowed_partitions{$user_role->{partition}} = 1;
            my $function_name = $params{function_name} || '';
            if (defined $params{entity_class}) {
                my $entity_config = $params{entity_class}->entity_config;
                if ( $entity_config->{authz_permissions} && $entity_config->{authz_permissions}->{minimum_write_level}
                     && F5::ASMConfig::AuthzRole::AUTHZ_ROLE_EDITOR() lt $entity_config->{authz_permissions}->{minimum_write_level} ) {
                    delete $allowed_partitions{$user_role->{partition}};
                    next;
                }
            }

            $session->{has_admin_access} = 0;
            for my $request_policy ($self->get_request_policies(%params)) {
                if($request_policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()) {
                    return $self->handle_error(error_thrown => "User $session->{gui_user_name} does not have permission to call $function_name on a Parent Policy",
                                               error_code => F5::ASMConfig::ErrorCode::PERMISSION_DENIED
                                              );
                }
            }

            if ($user_role->{partition} eq F5::ASMConfig::Constants::PARTITION_ALL || $params{allow_action_on_all_partitions}) {
                # Can access all partitions
                return 1;
            }
            else {
                $allowed_partitions{$user_role->{partition}} = 1;
            }
        }
        elsif ($access_type eq 'read') {
            # Check read restrictions
            my $function_name = $params{function_name} || '';
            my $has_limited_read_access = 0;
            if (defined $params{entity_class}) {
                my $entity_config = $params{entity_class}->entity_config;
                # Check read level
                if ( $entity_config->{authz_permissions} && $entity_config->{authz_permissions}->{minimum_read_level} ) {
                    $allowed_partitions{$user_role->{partition}} = 1;
                    my $minimum_read_level = $entity_config->{authz_permissions}->{minimum_read_level};
                    my $current_user_role = ($user_role->{role} == F5::Mcp::UserRole::ROLE_APPLICATION_SECURITY_EDITOR
                        || $user_role->{role} == F5::Mcp::UserRole::ROLE_APPLICATION_SECURITY_OPERATIONS_ADMINISTRATOR)
                        ? F5::ASMConfig::AuthzRole::AUTHZ_ROLE_EDITOR() : F5::ASMConfig::AuthzRole::AUTHZ_ROLE_GUEST();

                    if ( $entity_config->{authz_permissions}->{minimum_read_level_condition} && $entity_config->{authz_permissions}->{minimum_read_level_condition}->{db_variable} ) {
                        my $minimum_read_level_condition = F5::Cluster::Utils::get_mcp_val($entity_config->{authz_permissions}->{minimum_read_level_condition}->{db_variable});

                        if ($current_user_role lt $minimum_read_level
                            && $minimum_read_level_condition eq $entity_config->{authz_permissions}->{minimum_read_level_condition}->{value}) {
                            $has_limited_read_access = 1;
                        }
                    }
                    elsif ( $current_user_role lt $minimum_read_level ) {
                        $has_limited_read_access = 1;
                    }

                    if ( $has_limited_read_access ) {
                        delete $allowed_partitions{$user_role->{partition}};
                    }
                    elsif ($user_role->{partition} eq F5::ASMConfig::Constants::PARTITION_ALL || $params{allow_action_on_all_partitions}) {
                        # Can read all partitions
                        return 1;
                    }
                }
                else {
                    # Other users have read access only
                    if ($user_role->{partition} eq F5::ASMConfig::Constants::PARTITION_ALL || $params{allow_action_on_all_partitions}) {
                        # Can read all partitions
                        return 1;
                    }
                    $allowed_partitions{$user_role->{partition}} = 1;
                    $allowed_partitions{F5::ASMConfig::Constants::PARTITION_COMMON} = 1;
                }
            }
            else {
                # Other users have read access only
                if ($user_role->{partition} eq F5::ASMConfig::Constants::PARTITION_ALL || $params{allow_action_on_all_partitions}) {
                    # Can read all partitions
                    return 1;
                }
                $allowed_partitions{$user_role->{partition}} = 1;
                $allowed_partitions{F5::ASMConfig::Constants::PARTITION_COMMON} = 1;
            }
        }
    }

    if ($access_type eq 'read' && $self->limit_search_partitions_by_authz(%params,
                                                                          allowed_partitions => \%allowed_partitions)) {
        return 1;
    }

    # Access can only be granted if the policy is in the correct partition
    for my $request_partition ($self->get_request_partitions(%params)) {
        if (!$allowed_partitions{$request_partition}) {
            my $function_name = $params{function_name} || '';
            if ($request_partition) {
                $function_name .= " on partition $request_partition";
            }
            return $self->handle_error(error_thrown => "User $session->{gui_user_name} does not have permission to call $function_name",
                                       error_code => F5::ASMConfig::ErrorCode::PERMISSION_DENIED,
                                      );
        }
    }
    return 1;
}

#----------------------------------------------------------------------

sub to_em_slash_replace {
    my ($self, %params) = @_;
    $params{obj}->{$params{field_name}} =~ s|/|⁄|g;
}

#----------------------------------------------------------------------

sub from_em_slash_replace {
    my ($self, %params) = @_;
    if ($params{obj}->{$params{field_name}} =~ s|⁄|/|g) {
        $self->{em_client} = 1;
    }
}

#----------------------------------------------------------------------

sub get_request_partitions {
    my ($self, %params) = @_;
    my @authz_objects;
    if ($params{authz_objects}) {
        @authz_objects = @{$params{authz_objects}};
    }
    elsif ($params{authz_params} && $params{rpc_params}) {
        my @arguments = $self->extract_from_params(rpc_params => $params{rpc_params}, arguments => $params{authz_params});
        if ($params{entity_class} && $params{entity_class}->can('get_policies')) {
            for my $arg (@arguments) {
                for my $policy ($params{entity_class}->new(dbh => $self->{dbh}, object => $arg)->get_policies()) {
                    push (@authz_objects, F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, policy => $policy, session => $params{session}));
                }
            }
        }
        else {
            @authz_objects = @arguments;
        }
    }
    
    if (@authz_objects) {
        my %partitions;
        for my $obj (@authz_objects) {
            my $partition = F5::ASMConfig::Constants::PARTITION_COMMON;
            if ($obj->{policy_name} && ($obj->{policy_name} =~ m|^/([^/]*)|)) {
                $partition = $1; #Strip non partition path
            }
            elsif ($obj->{folder_name} && ($obj->{folder_name} =~ m|^/([^/]*)|)) {
                $partition = $1; #Strip non partition path
            }
            elsif ($obj->{partition_id} || $obj->{partition}) {
                $partition = $obj->{partition_id} || $obj->{partition};
            }
            elsif ($self->{session}->{partition}) {
                $partition = $self->{session}->{partition};
            }            
            $partitions{$partition} = 1;
        }
        return (keys %partitions);
    }

    return (F5::ASMConfig::Constants::PARTITION_COMMON);
}

#----------------------------------------------------------------------

sub limit_search_partitions_by_authz {
    my ($self, %params) = @_;

    my $search_limit_handled = 0;
    my $rpc_params = $params{rpc_params};
    my $function_name = $params{function_name} || '';
    my $allowed_partitions = $params{allowed_partitions} || [];
    my $authz_limit_field = $params{authz_limit_field};
    my $authz_limit_comparison = $params{authz_limit_comparison} || F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN();

    if ($authz_limit_field) {
        # These are the only calls that need to provide filtered results based on accessible partitions.

        if (! $rpc_params->[$CALL_FLAG_INDEX]) {
            $rpc_params->[$CALL_FLAG_INDEX] = F5::ASMConfig::CallFlags->new;
        }
        my $search_filter = F5::ASMConfig::SearchFilter->new(field_name => "partition_id",
                                                             field_value_list => [keys %$allowed_partitions],
                                                             field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN());

        if ((!$params{entity_class} || $params{entity_class} ne 'F5::ASMConfig::Entity::Policy') && $authz_limit_field ne 'partition_id') {
            my @allowed_policies;
            # Determine all policies within the allowed partitions
            my $policy_iter = F5::ASMConfig::Entity::Policy->retrieve_all(dbh => $self->{dbh},
                                                                          search_filters => [$search_filter]);
            while (my $policy = $policy_iter->get_next()) {
               push(@allowed_policies, $policy->{policy_id});
            }

            $search_filter = F5::ASMConfig::SearchFilter->new(
                                                  field_name => $authz_limit_field,
                                                  field_comparison => $authz_limit_comparison,
                                                  field_value_list => \@allowed_policies,
                                                  );
        }

        push(@{$rpc_params->[$CALL_FLAG_INDEX]->{search_filters}}, $search_filter);

        $search_limit_handled = 1;
    }

    return $search_limit_handled;
}

#----------------------------------------------------------------------

sub get_internal_config_items {
    my ($self, %params) = @_;

    my $config_item_name = $params{config_item_name};

    my $search_filter;
    if ($config_item_name) {
        $search_filter = F5::ASMConfig::SearchFilter->new({ field_name => 'config_item_name',
                                                            field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                                            field_value => $config_item_name });
    }
    else {
        $search_filter = F5::ASMConfig::SearchFilter->new();
    }

    my @internal_config_items = F5::ASMConfig::Entity::InternalConfigItem->retrieve_all(dbh => $self->{dbh},
                                                                                        search_filters => [ $search_filter ]);

    return wantarray ? @internal_config_items :
                       @internal_config_items ? $internal_config_items[0] : {} ;
}

#----------------------------------------------------------------------

sub get_request_policies {
    my ($self, %params) = @_;
    my @policies;
    my @arguments = $self->extract_from_params(rpc_params => $params{rpc_params}, arguments => $params{authz_params});
    if ($params{entity_class} && $params{entity_class}->can('get_policies')) {
        for my $arg (@arguments) {
            for my $policy ($params{entity_class}->new(dbh => $self->{dbh}, object => $arg)->get_policies()) {
                push (@policies, $policy);
            }
        }
    }
    else {
        for my $arg (@arguments) {
            if (ref($arg) eq 'F5::ASMConfig::Entity::Policy') {
                push (@policies, $arg);
            }
        }
    }
   return @policies;
}

#----------------------------------------------------------------------

sub extract_from_params {
    my ($self, %params) = @_;
    my $rpc_params = $params{rpc_params};
    my $arguments = $params{arguments};
    my $strip_paths = $params{strip_paths};
    my $prepend_path = $params{prepend_path};
    my @args;
    for my $argument (@$arguments) {
        my $arg_ref;
        my ($array_index, @hash_keys) =  split (/\./,  $argument);
        if (@hash_keys) {
            my $last_key = pop @hash_keys;
            my $arg_finder = $rpc_params->[$array_index];
            for my $dereffer (@hash_keys) {
                $arg_finder = $arg_finder->{$dereffer};
            }
            $arg_ref = \$arg_finder->{$last_key};
        }
        else {
            $arg_ref = \$rpc_params->[$array_index];
        }
        if ($$arg_ref) {
            if ($strip_paths) {                
                $$arg_ref = ($prepend_path ? "$prepend_path/" : "") . basename($$arg_ref);
            }
            push @args, $$arg_ref;
        }
    }
    return @args;
}

#----------------------------------------------------------------------
sub send_to_relay {
    my ($self, %params) = @_;

    my $service = $params{service};
    if (!$service) { 
        $service = $params{group_id} eq $self->{device_group} ? $F5::Cgc::Channel::DeviceGroup::ASM_CGC_LOW_PRI_TO_ALL
                                                              : $F5::Cgc::Channel::DeviceGroup::ASM_CGC_HIGH_PRI_TO_SINGLE;
    }

    $self->{logger}->debug("Opening CGC publisher Channels for service '$service'") if $self->{logger};
    my $channel = $self->get_cgc_channel(group_id     => $params{group_id},
                                         service      => $service,
                                         channel_type => "publisher",);
    $self->{logger}->trace($channel) if $self->{logger};
    $self->{logger}->debug("Sending $params{function_name} to $params{group_id}.") if $self->{logger};
    my $timestamp;
    my $previous_timestamp;
    my $time_change_detected;
    my $known_times_lock = F5::LockFactory::get_locker(type => "IPC",
                                                       ipc  => $self->{known_times_ipc});

    if ($params{mark_time}) {
        $known_times_lock->lock();
        $timestamp = Time::HiRes::time;
        if ($self->{shared_state}{recovery_time}) {
            #Do not send a previous timestamp if we just recovered from a ucs load
            $self->{shared_state}{recovery_time} = 0;
        }
        else {
            $previous_timestamp = $self->{known_times}{$self->{local_device_name}};
        }
        if ($self->{known_times}{$self->{local_device_name}} &&
            $self->{known_times}{$self->{local_device_name}} gt $timestamp) {
            $time_change_detected = 1;
        }
        $self->{known_times}{$self->{local_device_name}} = $timestamp;            
    }
    else {
        $timestamp = Time::HiRes::time;
    }

    #Set an originator_time in the call_flags so that if this call triggers a UCS sync in the group, the config can be correlated and only written once
    $params{rpc_params}->[$CALL_FLAG_INDEX]->{originator_time} = int($timestamp);
    $params{rpc_params}->[$CALL_FLAG_INDEX]->{out_of_sync} = $self->{shared_state}{out_of_sync};
    eval {
 
        my $msg = { function_name      => $params{function_name},
                    rpc_params         => $params{rpc_params},
                    originator         => $self->{local_device_name},
                    timestamp          => $timestamp,
                    mark_time          => $params{mark_time},
                    in_emergency_state => $self->{shared_state}{emergency},
                    previous_timestamp => $previous_timestamp,
                    time_change_detected => $time_change_detected,
                    device_group_member => $params{device_group_member},
                    user_name           => $self->{session}->{user_name},
                  };
        # If this is 
        # 1) A config call (mark_time is set)
        # 2) That was triggered by MCP (username is tsconfd)
        # 3) And we are a memeber of an autosync failover group
        # Then populate the failover dg in the message so other members of that dg will ignore this triggered event.
        if ($params{mark_time} && $self->{session}->{user_name} eq $TSCONFD_USER_NAME) {
            my $fo_dg = F5::DeviceGroup::get_failover_device_group();
            if ($fo_dg->{name} && $fo_dg->{autosync_enabled}) {
                $msg->{failover_device_group} = $fo_dg->{name};
            }
        }
        F5::ObjUtils::remove_unfreezable_references($msg);
        $channel->send($msg);
    };
    if ($@) {
        $self->{logger}->info("Failed on sending $params{function_name} to $params{group_id}: $@") if $self->{logger};
        $self->{logger}->trace(\%params) if $self->{logger};
    }
    if ($params{mark_time}) {        
        $known_times_lock->unlock();
        if ($params{can_update_mcp}) {
            $self->{update_mcp_timestamp} = 1;
        }
    }
    return;
}

#----------------------------------------------------------------------
sub update_mcp_timestamp {
    my ($self) = @_;
    return if !$self->{device_group} || !$self->{logger};
    $self->{logger}->debug("Updating asm device sync timestamp") if $self->{logger};
    my $timestamp = int($self->{known_times}{$self->{local_device_name}} || 0);
    eval {
        my $asm_sync_folder = F5::DeviceGroup::get_or_create_asm_sync_folder($self->{device_group});
        my $object = F5::ASM::COAPI::Object->new("asm_device_sync");
        $object->{device_name} =  $asm_sync_folder . "/" . basename($self->{local_device_name});
        $object->{timestamp} = $timestamp;
        $object->save();
        $self->{update_mcp_timestamp} = 0;
    };
    $self->{logger}->info("Unable to update MCP state: $@") if $@ && $self->{logger};
    return;
}

#----------------------------------------------------------------------

sub get_cgc_channel {
    my ($self, %params) = @_;
    my $group_id     = $params{group_id};
    my $service      = $params{service};
    my $channel_type = $params{channel_type};
    my $key          = "$group_id-$service-$channel_type";

    $self->{cgc_channel_last_access_time} = Time::HiRes::time();
    if (!exists ($self->{cgc_channels}{$key})) {
        my %additional_params;
        if ($channel_type eq 'subscriber') {
            $additional_params{is_listen} = 1;
        }
        my $channel = F5::Cgc::Channel::DeviceGroup->new(group_id     => $group_id,
                                                         logger       => $self->{logger},
                                                         service      => $service,
                                                         channel_type => $channel_type,
                                                         %additional_params
                                                        );
        if ($channel) {
            $self->{cgc_channels}{$key} = $channel;
        }
        return $channel;
    }
    return $self->{cgc_channels}{$key};
}

#----------------------------------------------------------------------
sub find_best_peer {
    my ($self, %params) = @_;
    my @devices = @{$params{devices}};
    my $best_peer = {timestamp => 0};

    my $asm_sync_folder = F5::DeviceGroup::get_asm_sync_folder($self->{device_group});
    return if !$asm_sync_folder;

    for my $device (@devices) {
        next if ($device eq $self->{local_device_name} && !$params{include_self});
        eval {
            my $device_mcp_timestamp =  F5::ASM::COAPI::Query->new("asm_device_sync", {device_name =>  "$asm_sync_folder/" . basename($device)});
            if ($device_mcp_timestamp && @$device_mcp_timestamp) {
                if ($device_mcp_timestamp->[0]->{timestamp} > $best_peer->{timestamp}) {
                    $best_peer->{timestamp} = $device_mcp_timestamp->[0]->{timestamp};
                    $best_peer->{device} = $device;
               }
            }
        };
    }
    return $best_peer;
}
#----------------------------------------------------------------------

sub log_error_and_rollback {
    my($self, %params) = @_;
    my $exception = $params{exception};
    my $do_log = defined $params{do_log} ? $params{do_log} : 1;

    if (! $self->{in_transaction} && ! $self->{dbh}->{AutoCommit}) {
        $self->{dbh}->rollback;
        if ($self->{transaction_locker}) {
            for my $locker (values %{$self->{transaction_locker}}) {
                $locker->unlock_all();
            }
            delete $self->{transaction_locker};
        }
        delete $self->{transactional_saves};
    }
    elsif ($self->{in_transaction}) {
        $self->{transaction_error} = 1; #Only finish the rollback when we are at the base rpc_entry_depth (let the error bubble up first)
        $self->rollback($self->{session});
    }

    my $error_msg;
    my $internal_err;
    my $log_level = 'error';
    my $error_code = $F5::ASMConfig::ErrorCode::UNKNOWN;
    my $additional_error_messages;
    if (ref($exception)) {
       $error_msg = $exception->{error_message};
       $internal_err = $exception->{internal_error} || $error_msg;
       $log_level = $exception->{error_level} || 'error';
       $error_code = $exception->{error_code};
       $additional_error_messages = $exception->{additional_error_messages};
       if ($exception->{events}) {
           $self->send_events_to_dispatcher(%$exception);
       }
    }
    else {
        $internal_err = $error_msg = $exception;
    }
    chomp($error_msg);
    chomp($internal_err);
    if ($do_log) {
        $self->{logger}->$log_level($error_msg) if $self->{logger};
    }
    die F5::ASMConfig::ASMConfigException->new({error_message  => $error_msg,
                                                internal_error => $internal_err,
                                                error_code     => $error_code,
                                                additional_error_messages => $additional_error_messages,
                                                rest_code      => $self->map_code_to_rest(error_code => $error_code)});
}

#----------------------------------------------------------------------

my %code_mapping = ( F5::ASMConfig::ErrorCode::SUCCESS()          => F5::ASMConfig::RestCode::REST_OK(),
                     F5::ASMConfig::ErrorCode::KNOWN_DB_ERRORS()  => F5::ASMConfig::RestCode::REST_NOT_FOUND(),
                     F5::ASMConfig::ErrorCode::DUPLICATE_RECORD() => F5::ASMConfig::RestCode::REST_BAD_REQUEST(),
                     F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS() => F5::ASMConfig::RestCode::REST_BAD_REQUEST(),
                     F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED()      => F5::ASMConfig::RestCode::REST_UNAUTHORIZED(),
                     F5::ASMConfig::ErrorCode::KNOWN_OPERATION_ERRORS()  => F5::ASMConfig::RestCode::REST_BAD_REQUEST(),
                     F5::ASMConfig::ErrorCode::NOT_IMPLEMENTED()         => F5::ASMConfig::RestCode::REST_NOT_IMPLEMENTED(),
                     F5::ASMConfig::ErrorCode::KNOWN_NETWORK_ERRORS()    => F5::ASMConfig::RestCode::REST_INTERNAL_SERVER_ERROR(),
                     F5::ASMConfig::ErrorCode::KNOWN_AUTHENTICATION_ERRORS() => F5::ASMConfig::RestCode::REST_UNAUTHORIZED(),
                    );
                     
sub map_code_to_rest {
    my ($self, %params) = @_;
    my $error_code = $params{error_code};
    my $rest_code = F5::ASMConfig::RestCode::REST_INTERNAL_SERVER_ERROR(); #Default
    if ($error_code) {
        my $error_class = int($error_code /100) * 100;
        if ($code_mapping{$error_code} || $code_mapping{$error_class}) {
            $rest_code = $code_mapping{$error_code} || $code_mapping{$error_class};
        }
    }
    return $rest_code;
}

#----------------------------------------------------------------------

sub login {
  my($self, $user_name) = @_;

  $user_name ||= "SYSTEM";
  push(@{$self->{sessions}}, $user_name);
  my $session_id = ($$ * 100) + $#{$self->{sessions}};
  $self->{local_device_name} = $self->{shared_state}->{local_device_name};
  $self->{logger}->debug("Login creating new session (local_device_name = $self->{local_device_name}, session_id = $session_id, user_name = $user_name)") if $self->{logger};

  return F5::ASMConfig::Session->new({user_name => $user_name,
                                      session_id => $session_id});
}

#----------------------------------------------------------------------

sub ping {
  my($self, $session) = @_;
  my $session_id = $session->{session_id} || '';
  my $user_name = $session->{user_name} || '';
  $self->check_memory_size();
  return 1;
}

#----------------------------------------------------------------------

sub log_info {
  my($self, $session, $info_message) = @_;
  $self->{logger}->info($info_message) if $self->{logger};
  $self->check_memory_size();
  return 1;
}

#----------------------------------------------------------------------

sub handle_stalled_async_calls {
    my($self, $session) = @_;

    if (! $self->{is_async_process}) {
        # Since this handler function is defined directly here, it doesn't go through
        # get_install_function handling, so instead this will hand off to async directly
        $self->pass_call_to_asynchronous_process(function_name => 'handle_stalled_async_calls',
                                                 rpc_params => [ $session ]);
    }
    else {
        my %found_stalled_files;
        find({ wanted => sub {
                   if ($File::Find::name =~ /saved_async_calls/) {
                       my $stalled_file_mtime = stat($File::Find::name)->mtime;
                       if (F5::DateTimeUtils::get_time_until_maximum_interval(start_time => $stalled_file_mtime,
                                                                              maximum_interval => $STALLED_ASYNC_CALLS_INTERVAL) == 0) {
                           $found_stalled_files{$File::Find::name} = $stalled_file_mtime;
                           $self->{logger}->debug("Adding saved async call file $File::Find::name as candidate for trigger") if $self->{logger};
                       }
                       else {
                           $self->{logger}->debug("Async call file $File::Find::name is too recent -- skipping") if $self->{logger};
                       }
                   }
               }
             },
             $TEMP_FILE_DIR);

        if (%found_stalled_files) {
            # Take the oldest file that seems to be abandoned
            my @found_stalled_files_by_mtime = sort { $found_stalled_files{$a} <=> $found_stalled_files{$b} } keys %found_stalled_files;
            $self->{logger}->debug("Loading async call file $found_stalled_files_by_mtime[0]") if $self->{logger};
            $self->load_async_calls_from_file(filename => $found_stalled_files_by_mtime[0]);
        }
    }
    return 1;
}

#----------------------------------------------------------------------

sub post_commit_actions {
    my ($self, %params) = @_;
    my $call_flags = $params{call_flags};
    my @queue;
    my %policies_to_apply;
    my $full_sync = 0;
    my $skip_cluster_sync = $call_flags->{skip_cluster_sync} || 0;
    my $do_table_checksums;

    for my $post_commit_hook (@{$self->{post_commit_hooks}}) {
        $self->{logger}->debug("Executing post commit hook $post_commit_hook->{name}") if $self->{logger};
        eval {
            $post_commit_hook->{func}->();
        };
        if (my $err_msg = $@) {
            my $level = "error";
            if (ref $err_msg) {
                $level = $err_msg->{error_level} || "error";
                $err_msg = $err_msg->{error_message};
            }
            $self->{logger}->$level("Could not execute post commit hook $post_commit_hook->{name}: $err_msg") if $self->{logger};
        }
    }

    if ($params{events}) {

        for my $event (@{$params{events}}) {
            if($event->{element_type} && $event->{element_type} == F5::ASMConfig::ElementType::SET_ACTIVE()
               && $event->{element}->{set_active}->{delayed_update_core}) {
    
                my $set_active_settings = $policies_to_apply{$event->{policy}->{policy_id}};
                if (!$set_active_settings) {
                    $set_active_settings = {policy => $event->{policy}};
                    $policies_to_apply{$event->{policy}->{policy_id}} = $set_active_settings;
                    push (@queue, $set_active_settings);
                }
                for my $field (qw(active_policy_was_changed original_account_id do_table_checksums)) {
                    $set_active_settings->{$field} ||= $event->{element}->{set_active}->{$field};
                }
                $full_sync ||= $set_active_settings->{active_policy_was_changed};
                $do_table_checksums ||= $set_active_settings->{do_table_checksums};
            }
        }
    
        if (@queue) {
            my $username = $self->{session}->{gui_user_name} || $self->{session}->{user_name};
            require F5::SetActive::Impl;
            F5::SetActive::Impl::delayed_update_core(dbh => $self->{dbh},
                                                     logger => $self->{logger},
                                                     username => $username,
                                                     full_sync => $full_sync,
                                                     do_table_checksums => $do_table_checksums,
                                                     skip_cluster_sync => $skip_cluster_sync,
                                                     update_core_queue => \@queue,);
        }
    
        #Add one pbconfig event for any policy that needs one.
        for my $policy_id (keys %{$self->{send_pbconfig_event}}) {
            my $pbconfig = F5::ASMConfig::Entity::PBConfig->new(dbh => $self->{dbh},
                                                                get_criteria => {policy_id => $policy_id},
                                                                policy => $self->{send_pbconfig_event}->{$policy_id});
            if ($pbconfig) {
                $pbconfig->populate_diff_structs();
                push (@{$self->{events}}, $self->create_event(event_type => F5::ASMConfig::EventType::UPDATE,
                                                              policy => $pbconfig->{policy},
                                                              session => $self->{session},
                                                              element => $pbconfig,
                                                              call_flags => $call_flags,
                                                              user_name => $self->{session}->{user_name},
                                                              skip_inheritance => 1,
                                                              suppress_save_to_log => 1,
                                                             ));
            }
        }

        #Add a persist_to_mcp event to trigger the cron task
        if ($self->{persist_to_mcp}) {
            push (@{$self->{events}}, $self->create_event(event_type => F5::ASMConfig::EventType::UPDATE,
                                                          session => $self->{session},
                                                          element => F5::ASMConfig::Element->new({ element_type => F5::ASMConfig::ElementType::PERSIST_TO_MCP } ),
                                                          call_flags => $call_flags,
                                                          user_name => $self->{session}->{user_name},
                                                          suppress_save_to_log => 1,
                                                          ));
            delete $self->{persist_to_mcp};
        }
    }

}
#----------------------------------------------------------------------

sub add_post_commit_hook {
    my ($self, %params) = @_;
    push (@{$self->{post_commit_hooks}}, \%params);
    $self->{logger}->debug("Added post commit hook $params{name}") if $self->{logger};
}

#----------------------------------------------------------------------

sub send_events_to_dispatcher {
    my($self, %params) = @_;

    return unless $params{events};
    if (! $self->{event_dispatcher}) {
        my %event_dispatcher_opts = (port => $self->{rpc_to_event_port},
                                     logger => $self->{logger},
                                    );

        if ($self->{guarantee_event_sending}) {
            $event_dispatcher_opts{reconnect_on_failure} = 1;
        }
        else {
            $event_dispatcher_opts{flush_write_attempt_limit} = 2;
        }

        $self->{event_dispatcher} = F5::IO::Storable->new(%event_dispatcher_opts);
    }

    for my $event (@{$params{events}}) {
        next unless $event;
        next if $event->{suppress_send_to_dispatcher};
        delete $event->{call_flags};
        F5::ObjUtils::remove_unfreezable_references($event);
        my $event_type = $event->{event_type} || 'undef';
        $self->{logger}->debug("Forwarding event to dispatcher (event type = $event_type)") if $self->{logger};

        $self->{event_dispatcher}->write_message($event);
    }

    if (! $self->{event_dispatcher}->flush()) {
        $self->{logger}->error("Events could not be sent to dispatcher") if $self->{logger};
    }
}

#----------------------------------------------------------------------

my %multi_policy_events = ( 
                          );

sub save_events_to_config_log {
  my($self, %params) = @_;

  return unless $params{events};
  my @events;
  for my $event (@{$params{events}}) {
    next if $event->{suppress_save_to_log};

    my @event_policies;
    if ($event->{element} && $event->{element_type} && $multi_policy_events{$event->{element_type}}) {
        if ($multi_policy_events{$event->{element_type}} eq "all") {
            @event_policies = F5::ASMConfig::Entity::Policy->retrieve_all(dbh => $self->{dbh}, logger => $self->{logger},);
        }
    }
    elsif ($event->{element_type} && $event->{element_type} == F5::ASMConfig::ElementType::POLICY()) {
        @event_policies = ($event->{element}->{policy});
    }
    else {
        @event_policies = ($event->{policy});
    }

    for my $event_policy (@event_policies) {
        next unless defined $event_policy;
        my $event_type = $event->{event_type} || 'undef';
        $self->{logger}->debug("Saving event to config log (event type = $event_type)") if $self->{logger};

        my $config_log = F5::ASMConfig::Entity::ConfigLog->new(dbh => $self->{dbh},
                                                               logger => $self->{logger},
                                                               object => $event,
                                                               policy => $event_policy);

        eval {
            $config_log->store_add(batched_insertor => $self->{batched_insertor},
                                   session => $self->{session},
                                  );
        };
        if (my $err_msg = $@) {
            my $level = "error";
            if (ref $err_msg) {
                $level = $err_msg->{error_level} || "error";
                $err_msg = $err_msg->{error_message};           
            }
            $self->{logger}->$level("Couldn't save to config log ($err_msg)") if $self->{logger};
        }
    }
  }
}

#----------------------------------------------------------------------

sub write_event_messages_to_syslog {
    my($self, %params) = @_;

    return unless $params{events};
    my @events;
    for my $event (@{$params{events}}) {

        if ($event->{element} && $event->{element}->{element_type} &&
            $event->{element}->{element_type} == F5::ASMConfig::ElementType::PBCONFIG &&
            $event->{update_thrift_obj} &&
            defined $event->{update_thrift_obj}->{flg_enabled}) {

            if ($event->{update_thrift_obj}->{flg_enabled}) {
                $self->write_to_syslog(msg => "Policy Builder was enabled for security policy \"$event->{policy}->{policy_name}\".");
            }
            else {
                $self->write_to_syslog(msg => "Policy Builder was disabled for security policy \"$event->{policy}->{policy_name}\".");
            }
        }

        if ($event->{rule_number} && $event->{rule_number} == 1 && $event->{element}) {
            my $element_type_name = $self->{element_types}->{$event->{element}->{element_type}}
                                    ? $self->{element_types}->{$event->{element}->{element_type}}->{element_type_name}
                                    : undef;
            my $entity = $event->{element}->{$element_type_name};
            my $entity_display_name = $entity ? $entity->get_display_name : '';
            my $element_type_display_name = $entity ? $entity->get_element_type_display_name: '';

            $self->write_to_syslog(msg => "Policy Builder detected site change and loosened security policy \"$event->{policy}->{policy_name}\" for element \"$element_type_display_name ($entity_display_name)\".");
        }
    }
}

#----------------------------------------------------------------------

sub create_event {
  my($self, %params) = @_;

  my $suppress_save_to_log = $params{suppress_save_to_log} || $params{call_flags}->{suppress_logging};
  my $do_policy_version_update = defined $params{do_policy_version_update} ? $params{do_policy_version_update} : 1;

  my $element = undef;
  my $search_result = undef;
  if ($params{element}) {
    if (ref($params{element}) eq 'F5::ASMConfig::Element') {
      $element = $params{element};
    }
    elsif (ref($params{element}) eq 'F5::ASMConfig::SearchResult') {
      $search_result = $params{element};
    }
    else {
      $params{element}->populate_for_event(%params);
      $element = F5::ASMConfig::Element->new({
                                              element_type => $params{element}->element_type,
                                              $params{element}->element_type_name => $params{element},
                                             });
      $suppress_save_to_log ||= $params{element}->suppress_modified_flag_update();
    }

    if ($element && $element->element_type == F5::ASMConfig::ElementType::CONFIG_LOG) {
        # no events on config log operations
        return;
    }
  }
  my $originating_host = $params{call_flags}->{originator} || $self->{local_device_name};
  if ($element && $element->{element_type} &&
      $self->{element_types}->{$element->{element_type}} &&
      $self->{element_types}->{$element->{element_type}}->{element_type_name} &&
      $element->{$self->{element_types}->{$element->{element_type}}->{element_type_name}}) {

      my $entity = $element->{$self->{element_types}->{$element->{element_type}}->{element_type_name}};
      if ($params{policy}) {
        if ($entity->affects_policy(policy => $params{policy},
                                  entity => $entity,)) {
            $params{policy} = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, policy => $params{policy}, force_reload => 1);
        }
      }
      elsif ($entity->policy_id_field()) {
          $params{policy} = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh},
                                                                    policy => $entity->{policy});
      }
  }

  my $event_time = time;

  my $event = F5::ASMConfig::Event->new({
                                         %params,
                                         element => $element,
                                         event_time => $event_time,
                                         user_name => $params{session}->{user_name},
                                         gui_user_name => $params{session}->{gui_user_name},
                                         device_name => $originating_host,
                                         search_result => $search_result,
                                        });

  if ($params{call_flags}) {
      for my $call_flag_key (qw(log_note rule_number rule_is_trusted suggestion_rest_uuid event_tags send_result_as_event_type)) {
          $event->{$call_flag_key} = $params{call_flags}->{$call_flag_key};
      }
  }

  if ($element) {
    $event->{element_type} = $element->{element_type};
    if (ref ($params{element}) ne 'F5::ASMConfig::Element') {
      $event->{element_name} = $params{element}->get_full_display_name(empty_string_value => "");
      $event->{element_id} = $params{element}->get_unique_id;
    }
  }
  $event->{description} = $params{description} || '';
  if ($params{update_thrift_obj}) {
      $event->{update_thrift_obj} = $params{update_thrift_obj};
      $event->{description} = ref($params{element})->get_entity_changes_string(entity => $params{update_thrift_obj},
                                                                               dbh => $self->{dbh},
                                                                               full_thrift_obj => $params{element},
                                                                              );
  }

  my $policy_for_version;
  if ($do_policy_version_update) {
      my $policy_for_version = $params{policy};
      if (! $policy_for_version && $element && $element->{element_type} eq F5::ASMConfig::ElementType::POLICY) {
          $policy_for_version = $element->{policy};
      }
  }

  my $full_description;
  $event->{suppress_save_to_log} = $suppress_save_to_log;
  if (( ! $event->{suppress_save_to_log}) || 
      ($do_policy_version_update && $policy_for_version) &&
      $event->{element} && $event->{element}->{element_type}) {

      my $element_type_name = $self->{element_types}->{$event->{element}->{element_type}}
                              ? $self->{element_types}->{$event->{element}->{element_type}}->{element_type_name}
                              : undef;
      if (defined $element_type_name) {
          my $entity = $event->{element}->{$element_type_name};
          if (defined $entity) {
               my $entity_display_name = $entity->get_display_name();
               my $element_type_display_name = $entity->get_element_type_display_name();
               my %event_types = F5::ASMConfig::Utils::get_event_type_strings();
               if ($event->{element_type} == F5::ASMConfig::ElementType::SET_ACTIVE) {
                   $entity_display_name = $event->{policy}->{policy_name};
               }

               my @audit_parts;
               if ($event->{policy} && $event->{policy}->{policy_name}) {
                   push(@audit_parts, "policy = $event->{policy}->{policy_name}");
               }

               if ($params{session}->{gui_user_name}) {
                   push(@audit_parts, "username = $params{session}->{gui_user_name}");
               }
               elsif ($params{session}->{user_name}) {
                   push(@audit_parts, "component = $params{session}->{user_name}");
               }
               if ($params{session}->{client_ip_address}) {
                   push(@audit_parts, "client IP = $params{session}->{client_ip_address}");
               }
               my $audit_parts = '{ audit: ' . join(', ', @audit_parts) . ' }';


               $full_description = "$element_type_display_name $entity_display_name \[$event_types{$params{event_type}}\]" .
                                   ($event->{description} eq '' ? '' : ": $event->{description}") . " $audit_parts";
          }
      }
  }

  if ( ! $event->{suppress_save_to_log} && $full_description) {
      $event->{syslog_message} = "ASMConfig change: $full_description";
  }


  if ($do_policy_version_update) {
      my $policy_for_version = $params{policy};
      if (! $policy_for_version && $element && $element->{element_type} eq F5::ASMConfig::ElementType::POLICY) {
          $policy_for_version = $element->{policy};
      }

      if ($policy_for_version) {
          $event->{is_policy_version_event} = 1;
          $event->{policy_version_time} = $params{policy_version_time} || $params{call_flags}->{originating_policy_version_time} || $event_time;
          $event->{policy_version_policy_name} = $params{policy_version_policy_name} || $policy_for_version->{policy_name};
          $event->{policy_version_last_change} = $params{policy_version_last_change} || $full_description;
          $event->{policy_version_device_name} = $params{policy_version_device_name} ||
                                                 $params{call_flags}->{originating_policy_version_device_name} ||
                                                 $self->{shared_state}{local_hostname};

          # Update policy's version time for event reference
          $policy_for_version->{policy_version_time} = $event->{policy_version_time};
          $policy_for_version->{policy_version_policy_name} = $event->{policy_version_policy_name};
          $policy_for_version->{policy_version_last_change} = $event->{policy_version_last_change};
          $policy_for_version->{policy_version_device_name} = $event->{policy_version_device_name};
      }
  }

  my @events = ($event); 
  if (!$params{skip_inheritance} && !$params{call_flags}->{skip_inheritance}) {
      push (@events, @{$self->propagate_inheritance_events(%params,
                                                           events => [$event],)});
  }
  

  return @events;
}

#----------------------------------------------------------------------

sub base_add {
  my($self, %params) = @_;
  my($session, $call_flags, $policy, $thrift_obj) = @{$params{rpc_params}};

  my $dbh = $self->{dbh};

  my $entity = $params{entity_class}->new(dbh => $dbh,
                                          logger => $self->{logger},
                                          preload_default => 1,
                                          rest_call => $call_flags->{rest_call},
                                          object => $thrift_obj,
                                          policy => $policy);

  $self->{logger}->debug("Adding " . $entity->get_debug_string) if $self->{logger};
  my $rv;
  eval {
    $rv = $entity->store_add(%params,
                             thrift_obj => $thrift_obj,
                             call_flags => $call_flags,
                             session => $session,
                            );

    $self->update_tightened_entity(thrift_obj => $thrift_obj,
                                   entity_class => $params{entity_class},
                                   policy => $policy,
                                  );
  };
  if (my $err = $@) {
    my $duplicate_record_error = (
            ( ref($err) && $err->{error_code} == F5::ASMConfig::ErrorCode::DUPLICATE_RECORD() ) ||
            (!ref($err) && $self->classify_error(error_thrown => $err) == F5::ASMConfig::ErrorCode::DUPLICATE_RECORD() )
          ) ? 1 : 0 ;

    if ($call_flags->{update_on_adding_duplicate_record} && $duplicate_record_error) {
      return $self->base_update(%params,
                                rpc_params => [$session, $call_flags, $policy, $entity, $thrift_obj],);
    }
    else {
      if ($call_flags->{return_existing_on_duplicate_add} && $duplicate_record_error) {
          $entity = $params{entity_class}->new(dbh => $dbh,
                                               logger => $self->{logger},
                                               policy => $policy,
                                               get_criteria => $thrift_obj);
          return { return_value => $entity };
      }
      
      return $self->handle_error(action => 'add',
                                 error_thrown => $err,
                                 call_flags => $call_flags,
                                 entity_class => $params{entity_class},
                                 entity => $entity,
                                 thrift_obj => $thrift_obj,
                                );
    }
  }

  my $events = [];
  my $suppress_events_res = $entity->suppress_events(session => $session,
                                                     call_flags => $call_flags,
                                                     store_obj => $thrift_obj,
                                                     event_type => F5::ASMConfig::EventType::ADD);
  if (!$params{suppress_events} && !$suppress_events_res  && !$call_flags->{suppress_events}) {
    $events = [ $self->create_event(event_type => F5::ASMConfig::EventType::ADD,
                                                 session => $session,
                                                 policy => $policy,
                                                 element => $entity,
                                                 call_flags => $call_flags,
                                                 session => $session,
                                                 update_thrift_obj => $thrift_obj,
                                                )
                           ];
    for my $created_entity (@{$entity->get_linked_entities(automatically_created => 1)}) {
        push(@$events, $self->create_event(event_type => F5::ASMConfig::EventType::ADD,
                                           session => $session,
                                           policy => $policy,
                                           element => $created_entity,
                                           call_flags => $call_flags,
                                           session => $session,
                                          ));
    }
    if(exists $rv->{modified_entities}) {
        push(@$events, $self->create_events_for_sub_entities(modified_entities => $rv->{modified_entities},
                                                            session => $session,
                                                            policy => $policy,
                                                            call_flags => $call_flags,)
        );
    }
  }

  my $retval = { return_value => $entity,
                 events => $events,
               };

  return $retval;
}

#----------------------------------------------------------------------

sub update_tightened_entity {
    my($self, %params) = @_;
    my $thrift_obj = $params{thrift_obj};
    my $entity_class = $params{entity_class};
    my $policy = $params{policy};

    my $dbh = $self->{dbh};
    my @tightens_fields = grep { /^tightens_/ && defined $thrift_obj->{$_} } keys %$thrift_obj;
    return unless @tightens_fields;

    my %tightened_entity_get_criteria;
    $tightened_entity_get_criteria{is_wildcard} = 1;

    for my $tightens_field (@tightens_fields) {
        my $get_criteria_field = $tightens_field;
        $get_criteria_field =~ s/^tightens_//;

        $tightened_entity_get_criteria{$get_criteria_field} = $thrift_obj->{$tightens_field};
    }

    my $tightened_entity = $entity_class->new(dbh => $dbh,
                                              logger => $self->{logger},
                                              minimal_load => 1,
                                              get_criteria => \%tightened_entity_get_criteria,
                                              policy => $policy);

    if ($tightened_entity) {
        my $updated_field = ($tightened_entity->{has_tightening} == F5::ASMConfig::TighteningMode::TIGHTENING_MODE_ALWAYS) ? "last_tightening_time"
                                                                                                                           : "last_updated";
        $tightened_entity->store_update(store_obj => { $updated_field => time });
        $self->{logger}->debug("Updated $updated_field for " . $tightened_entity->get_debug_string) if $self->{logger};
    }
    else {
        $self->{logger}->info("Couldn't find entity for updating tightening time") if $self->{logger};
    }
}

#----------------------------------------------------------------------

sub create_events_for_sub_entities {
    my($self, %params) = @_;
    my $modified_entities = $params{modified_entities};

    return unless $modified_entities;

    my %event_opts = (
                        session => $params{session},
                        policy => $params{policy},
                        call_flags => $params{call_flags},
                     );

    my @events;
    for my $modified_entity (@{$modified_entities}) {
        push(@events, $self->create_event(event_type => $modified_entity->{event_type},
                                          element => $modified_entity->{existing_entity},
                                          update_thrift_obj => $modified_entity->{entity},
                                          %event_opts)
        );
    }
    return @events;
}

#----------------------------------------------------------------------

sub create_events_for_changed_entities {
    my($self, %params) = @_;

    my $session = $params{session};
    my $policy = $params{policy};
    my $entities_changed = $params{entities_changed};
    my $call_flags = $params{call_flags};
    my $event_type = $params{event_type} || F5::ASMConfig::EventType::UPDATE;
    my $suppress_save_to_log = defined  $params{suppress_save_to_log} ? $params{suppress_save_to_log} : 0;

    my @events = ();
    for my $entity (@$entities_changed) {
        if (! $entity->suppress_events(%params) && !$call_flags->{suppress_events} && !$entity->{__suppress_events}) {
            my $entity_policy = $entity->{policy} || $policy;
            $entity_policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, policy => $entity_policy);
            push(@events, $self->create_event(event_type => $event_type,
                                              session => $session,
                                              policy => $entity_policy,
                                              element => $entity,
                                              call_flags => $call_flags,
                                              session => $session,
                                              update_thrift_obj => $entity->{__store_obj},
                                              suppress_save_to_log => $suppress_save_to_log,
                                              skip_inheritance => 1));
        }
    }

    if (!$params{skip_inheritance} && !$call_flags->{skip_inheritance}) {
        push (@events, @{$self->propagate_inheritance_events(%params,
                                                             events => \@events,)});
    }

    return @events;
}

#----------------------------------------------------------------------

sub base_get {
  my($self, %params) = @_;
  my($session, $call_flags, $policy, $thrift_obj) = @{$params{rpc_params}};

  my $dbh = $self->{dbh};
  my $default_rd =  defined $session->{default_rd} ? $session->{default_rd} : 0;
  $thrift_obj ||= {};
  my $entity = $params{entity_class}->new(session => $session,
                                          search_filters => $call_flags->{search_filters},
                                          call_flags => $call_flags, 
                                          dbh => $dbh,
                                          logger => $self->{logger},
                                          get_criteria => $thrift_obj,
                                          policy => $policy,
                                          default_rd => $default_rd,
                                          extra_fields => $call_flags->{extra_fields});
  unless ($entity) {
    return $self->handle_error(action => 'get',
                               error_thrown => $NONEXISTENT_RECORD_ERROR,
                               call_flags => $call_flags,
                               entity => $entity,
                               entity_class => $params{entity_class},
                               thrift_obj => $thrift_obj,
                              );
  }
  my $retval = { return_value => $entity,
               };
  return $retval;
}

#----------------------------------------------------------------------

sub base_get_default {
  my($self, %params) = @_;
  my($session, $call_flags, $policy) = @{$params{rpc_params}};

  my $dbh = $self->{dbh};
  my $entity = $params{entity_class}->new(dbh => $dbh,
                                          logger => $self->{logger},
                                          policy => $policy,
                                          preload_default => 1,
                                         );

  if ( exists $entity->{in_staging} ){
    $entity->{in_staging} = 1;
  }

  my $retval = { return_value => $entity,
               };
  return $retval;
}

#----------------------------------------------------------------------

sub base_update {
  my($self, %params) = @_;
  my($session, $call_flags, $policy, $search_criteria, $store_obj) = @{$params{rpc_params}};

  my $dbh = $self->{dbh};
  my $default_rd =  defined $session->{default_rd} ? $session->{default_rd} : 0;
  my $do_update = 0;
  my @events;

  $search_criteria ||= {};
  #Do not modify the original objects so that it can be sent over CMI as it came in
  if (!$store_obj) {
      $store_obj = $search_criteria;
  }
  $search_criteria = dclone($search_criteria);
  $store_obj = dclone($store_obj);

  my $entity = $params{entity_class}->new(dbh => $dbh,
                                          logger => $self->{logger},
                                          get_criteria => $search_criteria,
                                          search_filters => $call_flags->{search_filters},
                                          policy => $policy,
                                          default_rd => $default_rd,
                                          extra_fields => $call_flags->{extra_fields},);

  unless ($entity) {
    if ($call_flags->{add_on_updating_nonexistent_record}) {
      my $add_object = {%$search_criteria, %$store_obj};
      return $self->base_add(%params,
                             rpc_params => [$session, $call_flags, $policy, $add_object],);
    }
    else {
      return $self->handle_error(action => 'update',
                                 error_thrown => $NONEXISTENT_RECORD_ERROR,
                                 call_flags => $call_flags,
                                 entity => $entity,
                                 entity_class => $params{entity_class},
                                 thrift_obj => $search_criteria,
                                );
    }
  }

  # Convert search_criteria to unique record parts
  $entity->convert_search_criteria_to_unique_record_parts(search_criteria => $search_criteria);

  if ($params{update_func}) {
    $params{update_func}->(entity => $entity,
                           store_obj => $store_obj);
  }

  my $rv;
  my $entity_before_update = $params{entity_class}->new(object => $entity, dbh => $self->{dbh});
  eval {
    $self->{logger}->debug("Updating " . $entity->get_debug_string) if $self->{logger};
    $rv = $entity->store_update(store_obj => $store_obj,
                                call_flags => $call_flags,
                                session => $session,
                                extra_fields => $call_flags->{extra_fields},
                                %params,
                               );
  };

  if ($@) {
      my $error_thrown = $@;
      my $unique_record_parts = $params{entity_class}->unique_record_parts(get_criteria => $search_criteria);
      my $thrift_obj = $search_criteria;
      if ($params{entity_class}->has_value_changed(existing_obj => $entity_before_update,
                                                   store_obj => $store_obj,
                                                   check_fields => $unique_record_parts)) {
          for my $unique_field (@$unique_record_parts) {
               if ($store_obj->{$unique_field}) {
                $thrift_obj->{$unique_field} = $store_obj->{$unique_field};
               }
          }
      }
    return $self->handle_error(action => 'update',
                               error_thrown => $error_thrown,
                               call_flags => $call_flags,
                               entity => $entity,
                               entity_class => $params{entity_class},
                               thrift_obj => $thrift_obj,
                              );
  }

  my $suppress_events_res = $entity->suppress_events(session => $session,
                                                     call_flags => $call_flags,
                                                     store_obj => $store_obj,
                                                     event_type => F5::ASMConfig::EventType::UPDATE);
  if (!$params{suppress_events} && !$suppress_events_res && !$call_flags->{suppress_events}) {
      if ( $rv->{db_change_made} ) {
          my %event_opts = (
                            session => $session,
                            policy => $policy,
                            call_flags => $call_flags,
                            session => $session,
                           );

          my $pb_unique_record_parts = $params{entity_class}->pb_unique_record_parts(get_criteria => $search_criteria);
          if ($params{entity_class}->has_value_changed(existing_obj => $entity_before_update,
                                                       store_obj => $store_obj,
                                                       check_fields => $pb_unique_record_parts)) {
              # Send delete and add events

              push(@events,
                   $self->create_event(event_type => F5::ASMConfig::EventType::DELETE,
                                       element => $entity_before_update,
                                       skip_inheritance => 1,
                                       %event_opts)
                  );

              push(@events,
                   $self->create_event(event_type => F5::ASMConfig::EventType::ADD,
                                       element => $entity,
                                       update_thrift_obj => $store_obj,
                                       %event_opts)
                  );

          }
          else {
              # Send update event
              push(@events,
                   $self->create_event(event_type => F5::ASMConfig::EventType::UPDATE,
                                       element => $entity,
                                       update_thrift_obj => $store_obj,
                                       %event_opts)
                  );
          }
      }
      if(exists $rv->{modified_entities}) {
          push(@events, $self->create_events_for_sub_entities(modified_entities => $rv->{modified_entities},
                                                              session => $session,
                                                              policy => $policy,
                                                              call_flags => $call_flags,)
          );
      }
  }

  if (my $func = $params{entity_class}->after_update_func()) {
      no strict 'refs';
      my $rv = $func->($self, policy => $policy,
                               call_flags => $call_flags,
                               update_thrift_obj => $store_obj,
                               element => $entity,
                               session => $session,
                               entity_class => $params{entity_class});

      if ($rv->{events}) {
          push (@events, @{$rv->{events}});
      }
  }


  my $retval = { return_value => $entity,
                 events => \@events };

  return $retval;
}

#----------------------------------------------------------------------

sub base_delete {
  my($self, %params) = @_;
  my($session, $call_flags, $policy, $thrift_obj) = @{$params{rpc_params}};

  my $dbh = $self->{dbh};
  my $default_rd =  defined $session->{default_rd} ? $session->{default_rd} : 0;

  my $entity = $params{entity_class}->new(dbh => $dbh,
                                          logger => $self->{logger},
                                          get_criteria => $thrift_obj,
                                          search_filters => $call_flags->{search_filters},
                                          policy => $policy,
                                          default_rd => $default_rd);
  unless ($entity) {
    return $self->handle_error(action => 'delete',
                               error_thrown => $NONEXISTENT_RECORD_ERROR,
                               call_flags => $call_flags,
                               entity_class => $params{entity_class},
                               thrift_obj => $thrift_obj,
                              );
  }

  # Convert search_criteria to unique record parts
  $entity->convert_search_criteria_to_unique_record_parts(search_criteria => $thrift_obj);

  my $rv;
  eval {
    $rv  = $entity->delete(%params,
                           call_flags => $call_flags,
                           session => $session,);
  };
  if ($@) {
    return $self->handle_error(action => 'delete',
                               error_thrown => $@,
                               call_flags => $call_flags,
                               entity => $entity,
                               thrift_obj => $thrift_obj,
                              );
  }

  my @events;
  if ($rv && !$params{suppress_events} && ! $entity->suppress_events(session => $session, call_flags => $call_flags) && !$call_flags->{suppress_events}) {
    push(@events,
         $self->create_event(event_type => F5::ASMConfig::EventType::DELETE,
                             session => $session,
                             policy => $policy,
                             element => $entity,
                             call_flags => $call_flags,
                             session => $session,
                            ));

      if (ref($rv) && $rv->{deleted_entities}) {
          push(@events,
               $self->create_events_for_changed_entities(event_type => F5::ASMConfig::EventType::DELETE,
                                                         session => $session,
                                                         policy => $policy,
                                                         entities_changed => $rv->{deleted_entities},
                                                         call_flags => $call_flags,
                                                         session => $session,
                                                        )
              );

      }
          
  }

  my $retval = { return_value => $entity,
                 events => \@events,
               };
  return $retval;
}

#----------------------------------------------------------------------

sub base_search {
  my($self, %params) = @_;
  my($session, $call_flags, $policy, $thrift_obj) = @{$params{rpc_params}};

  my $dbh = $self->{dbh};
  $thrift_obj ||= {};
  my $default_rd =  defined $session->{default_rd} ? $session->{default_rd} : 0;

  my $result_set = $params{entity_class}->retrieve_all(call_flags => $call_flags,
                                                       session => $session,
                                                       dbh => $dbh,
                                                       logger => $self->{logger},
                                                       get_criteria => $thrift_obj,
                                                       find_single_record => 0,
                                                       search_filters => $call_flags->{search_filters},
                                                       sort_orders => $call_flags->{sort_orders},
                                                       page_size => $call_flags->{page_size},
                                                       start_page => $call_flags->{start_page},
                                                       grouping_fields => $call_flags->{grouping_fields},
                                                       grouping_aggregation_fields => $call_flags->{grouping_aggregation_fields},
                                                       extra_fields => $call_flags->{extra_fields},
                                                       policy => $policy, 
                                                       default_rd => $default_rd);

  $result_set->{total_record_count} = 0 unless defined $result_set->{total_record_count};

  my @entities = $result_set->get_all;
  my $search_result = $self->entities_to_search_result(entities => \@entities,
                                                       call_flags => $call_flags,
                                                       total_record_count => $result_set->{total_record_count},
                                                       entity_class => $params{entity_class},
                                                       );

  my $retval = { return_value => $search_result,
               };
  return $retval;
}

#----------------------------------------------------------------------

sub base_update_dont_check {
  my($self, %params) = @_;
  my($session, $call_flags, $policy, $thrift_obj) = @{$params{rpc_params}};

  my $dbh = $self->{dbh};
  my %updated_entities;

  my %fields = %{$params{entity_class}->fields};
  my @dont_check_fields = grep { $fields{$_}->{set_via_dont_check} } keys %fields;


  my @events;

  for my $dont_check_field (@dont_check_fields) {
    if (defined $thrift_obj->{$dont_check_field}) {

      my $criteria_obj;
      $criteria_obj->{$dont_check_field} = zero_not($thrift_obj->{$dont_check_field});
      if ($fields{$dont_check_field}->{only_applies_to_wildcard}) {
        $criteria_obj->{is_wildcard} = 1;
      }

      my $update_obj;
      $update_obj->{in_staging} = 1;
      $update_obj->{$dont_check_field} = $thrift_obj->{$dont_check_field};
    
      if (my $length_field = $fields{$dont_check_field}->{length_field}) {
        $update_obj->{$length_field} = $thrift_obj->{$length_field};
      }
    
      my $field_entities_changed = $params{entity_class}->update_many( session         => $session, 
                                                                       call_flags      => $call_flags, 
                                                                       policy          => $policy, 
                                                                       search_criteria => $criteria_obj, 
                                                                       store_obj       => $update_obj,
                                                                       dbh             => $dbh
                                                                      );
      push(@events, $self->create_events_for_sub_entities(modified_entities => $field_entities_changed->{modified_entities},
                                                          session           => $session,
                                                          call_flags        => $call_flags,
                                                          policy            => $policy,
                                                        )
      );

      for my $entity (@{$field_entities_changed->{updated_entities}}) {
          my $entity_id = $entity->get_id;
          if ($updated_entities{$entity_id}) {
              # Combine store_obj for user log
              $entity->{__store_obj} = { %{$updated_entities{$entity_id}->{__store_obj}},
                                         %{$entity->{__store_obj}} };
          }
        $updated_entities{$entity_id} = $entity;
      }
    }
  }

  push(@events, $self->create_events_for_changed_entities(
                                                         session          => $session,
                                                         call_flags       => $call_flags,
                                                         policy           => $policy,
                                                         entities_changed => [ values %updated_entities ])
  );
      
  my $retval = { return_value => scalar @events,
                 events => \@events,
               };

  return $retval;
}

#----------------------------------------------------------------------

sub zero_not {
  my($val) = @_;
  return ($val ? 0 : 1);
}

#----------------------------------------------------------------------

our %log_level_mapping = ( F5::ASMConfig::LogLevel::LOG_LEVEL_TRACE => 'trace',
                           F5::ASMConfig::LogLevel::LOG_LEVEL_DEBUG => 'debug',
                           F5::ASMConfig::LogLevel::LOG_LEVEL_INFO => 'info',
                           F5::ASMConfig::LogLevel::LOG_LEVEL_WARN => 'warn',
                           F5::ASMConfig::LogLevel::LOG_LEVEL_ERROR => 'error',
                           F5::ASMConfig::LogLevel::LOG_LEVEL_FATAL => 'fatal',
                          );

sub handle_error {
  my($self, %params) = @_;
  my $entity = $params{entity};
  my $entity_class = $params{entity_class};
  my $error_thrown = $params{error_thrown};
  my $error_code = $params{error_code};
  my $display_message = $params{display_message};
  my $call_flags = $params{call_flags};
  my $action = $params{action} || '';
  my $thrift_obj = $params{thrift_obj};
  my $error_level = $params{error_level};
  my $events = $params{events};
  my $additional_error_messages = $params{additional_error_messages};

 ## error_thrown = internal_error
 ## display_message = error_message
  if (ref($error_thrown)) {
       $error_level     = $error_thrown->{error_level}     if defined($error_thrown->{error_level});
       $display_message = $error_thrown->{display_message} if defined($error_thrown->{display_message});
       $error_code      = $error_thrown->{error_code}      if defined($error_thrown->{error_code});
       $additional_error_messages = $error_thrown->{additional_error_messages}      if defined($error_thrown->{additional_error_messages});
       $error_thrown    = $error_thrown->{error_message};
  }

  my $entity_display_string = '';
  my $print_entity_class = '';  
  my $entity_as_string  = '';
  if ($entity_class) {
      $print_entity_class = ($entity ? ref($entity) : $entity_class);
      $print_entity_class =~ s/^.*:://;

      my $entity_display_name = '';
      my $entity_id = 0;
      my $entity_type_display_name = '';
      my $name_field = $entity_class->name_field();

      if ($thrift_obj) {
          my $display_entity = $entity_class->new(dbh => $self->{dbh},
                                              logger => $self->{logger},
                                              object => $thrift_obj,
                                              instantiate_entities => 1,
                                              full_retrieve_on_instantiate => 1,);
          if (!$name_field || defined($display_entity->{$name_field})) {
              $entity_display_name = $display_entity->get_full_display_name(%params);
          }
          $entity_id = $display_entity->get_id;
      }

      if ($entity) {
          $entity_as_string = $entity->get_debug_string;
          $entity_type_display_name = $entity->get_element_type_display_name;
          if (!$entity_display_name || $entity_display_name eq 'UNNAMED') {
              $entity_display_name = $entity->get_full_display_name(%params);
          }
          $entity_id ||= $entity->get_id;
      }
      else {
          $entity_as_string = $print_entity_class;
          $entity_type_display_name = $entity_class->get_element_type_display_name;
      }

      $entity_type_display_name ||= $print_entity_class;
      if ($entity_display_name) {
          $entity_display_name = " '$entity_display_name'";
      }
      elsif (!$entity_display_name && $entity_id) {
          $entity_display_name = " (id: $entity_id)";
      }
      $entity_display_string = $entity_type_display_name . $entity_display_name;
  }

  if (!defined($error_code)) {
      $error_code = $self->classify_error(error_thrown => $error_thrown,);
  }

  if (my $ignored_error = $self->is_error_ignored(call_flags => $call_flags, action => $action, error_code => $error_code)) {
      $error_level = ($ignored_error->{log_level} && exists $log_level_mapping{$ignored_error->{log_level}}) ? $log_level_mapping{$ignored_error->{log_level}} : 'debug';
      $self->{logger}->$error_level("Ignoring failed $action on error $error_code") if $self->{logger};
      if ( ! $ignored_error->{throw_exception_on_error} ) {
          my $return_value = defined $params{ignored_rv} ? $params{ignored_rv} : $thrift_obj;
          return { return_value => $return_value, ignored_error => 1 };
      }
  }

  my $internal_error;
  my $error_message;
  if ($error_code == F5::ASMConfig::ErrorCode::SUCCESS()) {
      return { return_value => $thrift_obj };
  }
  elsif ($error_code == F5::ASMConfig::ErrorCode::DUPLICATE_RECORD() ) {
      $error_level ||= 'info';
      $self->{logger}->$error_level("Failed $action on duplicate record for $entity_as_string") if $self->{logger};
      $internal_error = "Failed $action on duplicate record for $print_entity_class found -- $action aborted";
      if ($display_message) {
          $error_message  = (($entity_display_string && $action) ? "Could not $action the $entity_display_string.  $display_message"
                                                                 : $display_message);
      }
      elsif ($entity_class && $entity_class->policy_id_field()) {
          $error_message = "$entity_display_string already exists in this policy.";
      }
      else {
          $error_message = "$entity_display_string already exists.";
      }
  }
  elsif ($error_code == F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD() ) {
      $error_level ||= 'info';
      $self->{logger}->$error_level("Failed $action on nonexistent record for $entity_as_string") if $self->{logger};
      $internal_error = "Failed $action on nonexistent record for $print_entity_class found -- $action aborted";
      $error_message = "Could not $action the $entity_display_string, No matching record was found.";
  }
  elsif ($error_code == F5::ASMConfig::ErrorCode::MISSING_FOREIGN_KEY()) {
      $error_level ||= 'info';
      $self->{logger}->$error_level("Failed $action on missing foreign key for $entity_as_string") if $self->{logger};
      $internal_error = "Failed $action on missing foreign key for $print_entity_class found -- $action aborted";
      $error_message = "Could not $action the $entity_display_string, internal data inconsistency was detected.";
  }
  elsif ($error_code == F5::ASMConfig::ErrorCode::DEADLOCK_FOUND()) {
      $error_level ||= 'info';
      $self->{logger}->$error_level("Failed $action: $DEADLOCK_FOUND_ERROR") if $self->{logger};
      $internal_error = "Failed $action: DEADLOCK_FOUND_ERROR";
      $error_message = "Could not $action the $entity_display_string, $DEADLOCK_FOUND_ERROR";
  }
  elsif ($error_code == F5::ASMConfig::ErrorCode::LOCK_TIMEOUT()) {
      $error_level ||= 'info';
      $self->{logger}->$error_level("Failed $action: $LOCK_TIMEOUT_ERROR") if $self->{logger};
      $internal_error = "Failed $action: LOCK_TIMEOUT_ERROR";
      $error_message = "Could not $action the $entity_display_string, $LOCK_TIMEOUT_ERROR";
  }
  else {
      $display_message ||= $error_thrown;
      #if it's a known error then default to info, otherwise default to error
      $error_level ||= ($error_code ==  F5::ASMConfig::ErrorCode::UNKNOWN()) ? 'error' : 'info';
      $self->{logger}->$error_level("Failed $action for $entity_as_string") if ($self->{logger} && $entity_as_string);
      $self->{logger}->$error_level("Code: $error_code, Error message = " . $error_thrown) if $self->{logger};
      $internal_error = ($print_entity_class ? "Failed $action for $print_entity_class found -- $action aborted : Error message = $error_thrown"
                                             : "Failed $action : Error message = $error_thrown");
      $error_message  = (($entity_display_string && $action) ? "Could not $action the $entity_display_string.  $display_message"
                                                             : $display_message);
  }
  $error_message =~ s/ at \/usr.*//;
  die {
       error_level => $error_level,
       internal_error => $internal_error,
       error_message => $error_message,
       error_code    => $error_code,
       events        => $events,
       additional_error_messages => $additional_error_messages,
      };

}

#----------------------------------------------------------------------
sub classify_error {
    my ($self, %params) = @_;
    my $error_thrown = $params{error_thrown};
    my $error_code = F5::ASMConfig::ErrorCode::UNKNOWN();

    if ($error_thrown =~  /$DUPLICATE_FOUND_DBI_ERROR/) {
        $error_code = F5::ASMConfig::ErrorCode::DUPLICATE_RECORD();
    }
    elsif ($error_thrown =~ /$NONEXISTENT_RECORD_ERROR/) {
        $error_code = F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD();
    }
    elsif ($error_thrown =~ /$MISSING_FOREIGN_KEY_ERROR/) {
        $error_code = F5::ASMConfig::ErrorCode::MISSING_FOREIGN_KEY();
    }
    elsif ($error_thrown =~ /Signature.'.*'.is invalid:/) {
        $error_code = F5::ASMConfig::ErrorCode::INVALID_SIGNATURE();
    }
    elsif ($error_thrown =~ /^MCP Validation error - /) {
        $error_code = F5::ASMConfig::ErrorCode::MCP_VALIDATION_ERROR();
    }
   elsif ($error_thrown =~ /$DEADLOCK_FOUND_ERROR/) {
        $error_code = F5::ASMConfig::ErrorCode::DEADLOCK_FOUND();
    }
    elsif ($error_thrown =~ /$LOCK_TIMEOUT_ERROR/) {
        $error_code = F5::ASMConfig::ErrorCode::LOCK_TIMEOUT();
    }

    return $error_code;
}

#----------------------------------------------------------------------
sub is_error_ignored {
    my ($self, %params) = @_;
    my $call_flags = $params{call_flags};
    my $action     = $params{action};
    my $error_code = $params{error_code};

    if ( $call_flags->{rest_ignored_errors} ) {
        my $rest_error_code = $self->map_code_to_rest(error_code => $error_code);
        for my $rest_ignored_error (@{$call_flags->{rest_ignored_errors}}) {
        my $error_code = $rest_ignored_error->{error_code};
            if ( $error_code eq "*" || $error_code == $rest_error_code ) {
                return $rest_ignored_error;
            }
            elsif ( $error_code =~ m/\*/) {
               $error_code =~ s/\*/.*/g;
               if ( $rest_error_code =~ $error_code ) {
                   return $rest_ignored_error;
               }
            }
        }
    }

    if ($self->{is_relay_handler_process}) {
        push (@{$call_flags->{ignored_errors}}, F5::ASMConfig::IgnoredError->new({error_code => F5::ASMConfig::ErrorCode::KNOWN_MCP_ERRORS()}));
    }
    $self->convert_ignored_errors(call_flags => $call_flags);
    my $error_class = int($error_code /100) * 100;
    if (my ($ignored_error) = grep { (!$_->{action} || $_->{action} eq $action)
              && ($error_code == $_->{error_code} || $error_class == $_->{error_code})} @{$call_flags->{ignored_errors}}) {
        return $ignored_error;
    }
    return; 
}

#----------------------------------------------------------------------
my %error_mapping = ("ignore_all_known_database_conditions" => F5::ASMConfig::IgnoredError->new({error_code => F5::ASMConfig::ErrorCode::KNOWN_DB_ERRORS()}),
                   "ignore_duplicate_record_on_add" =>  F5::ASMConfig::IgnoredError->new({error_code => F5::ASMConfig::ErrorCode::DUPLICATE_RECORD(),
                                                                                          action     => "add"}),
                   "ignore_duplicate_record_on_update" =>  F5::ASMConfig::IgnoredError->new({error_code => F5::ASMConfig::ErrorCode::DUPLICATE_RECORD(),
                                                                                             action     => "update"}),
                   "ignore_nonexistent_record_on_update" => F5::ASMConfig::IgnoredError->new({error_code => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD(),
                                                                                              action     => "update"}),
                   "ignore_nonexistent_record_on_delete" => F5::ASMConfig::IgnoredError->new({error_code => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD(),
                                                                                              action     => "delete"}),
                   "ignore_missing_foreign_key_on_add" => F5::ASMConfig::IgnoredError->new({error_code => F5::ASMConfig::ErrorCode::MISSING_FOREIGN_KEY(),
                                                                                            action     => "add"}),
                   "ignore_missing_foreign_key_on_update" => F5::ASMConfig::IgnoredError->new({error_code => F5::ASMConfig::ErrorCode::MISSING_FOREIGN_KEY(),
                                                                                               action     => "update"}),
                    );



sub convert_ignored_errors {
    my ($self, %params) = @_;
    my $call_flags = $params{call_flags};

    for my $key (keys %error_mapping) {
        if ($call_flags->{$key}) {
            push (@{$call_flags->{ignored_errors}}, $error_mapping{$key});
        }
    }
    return;
}
#----------------------------------------------------------------------

sub pass_call_to_asynchronous_process {
    my($self, %params) = @_;
    my $function_name = $params{function_name};
    my $rpc_params = $params{rpc_params};
    $params{async_call_id} = $self->{async_call_id}++;
    my $attempt_count_limit = 3;

    for my $attempt_count (1 .. $attempt_count_limit) {
        eval {
            $self->{logger}->debug("Will attempt to pass call to async process (attempt $attempt_count)") if $self->{logger};

            #Reap any old async handlers
            my $child_process = waitpid(-1, WNOHANG);
            if ($child_process > 0 && $self->{async_children}->{$child_process}) {
                $self->{logger}->info("Async handler PID $child_process has finished") if $self->{logger};
                if (! $self->{async_children}->{$child_process}->{was_flush_completed}) {
                    $self->{async_children}->{$child_process}->{socket}->close();
                }
                delete $self->{async_children}->{$child_process};
            }

            # Attempt to flush inactive async children
            for my $inactive_async_child (grep { ! $_->{is_active} && ! $_->{was_flush_completed} } values %{$self->{async_children}}) {
                $self->{logger}->trace("Flushing async handler PID $inactive_async_child->{pid}") if $self->{logger};
                $inactive_async_child->{was_flush_completed} = $inactive_async_child->{socket}->flush(write_attempt_limit => 1);
                if ($inactive_async_child->{was_flush_completed}) {
                    $self->{logger}->debug("Flush complete for async handler PID $inactive_async_child->{pid}") if $self->{logger};
                    $inactive_async_child->{socket}->close();
                }
            }

            my $active_async_child;
            for my $child (values %{$self->{async_children}}) {
                if (-f $self->get_async_shutdown_filename(parent_pid => $$, child_pid => $child->{pid})) {
                    $self->{logger}->debug("Async child $child->{pid} is shutting down -- marking as inactive") if $self->{logger};
                    $child->{is_active} = 0;
                }
                if ($child->{is_active}) {
                    $active_async_child = $child;
                }
            }

            if (! $active_async_child && scalar values %{$self->{async_children}} < 2) {
                # We can only ever have 2 async children at a time. One that is currently finishing off a queue of requests, and the new one that is waiting to start.
                # If the new one has received max_rpc_calls_per_async_handler calls and the previous handler still has not yet finished, we treat this as a full socket and do NOT spawn a new handler.
                $active_async_child = $self->spawn_async_handler();
            }

            if (! $active_async_child) {
                my $error_msg = "Async processes are full - ignoring";
                $self->{logger}->log(msg => $error_msg,
                                     log_level => 'info',
                                     throttled_log_level => 'debug',
                                     throttling_interval => 10) if $self->{logger};
                die F5::ASMConfig::ASMConfigException->new({error_message => $error_msg});
            }

            $self->{logger}->debug("Dispatching call (call id $params{async_call_id}, $active_async_child->{rpc_call_count} calls handled) to pid $active_async_child->{pid}") if $self->{logger};
            my $socket = $active_async_child->{socket};
            F5::ObjUtils::remove_unfreezable_references(\%params);

            if (! $socket->write_message(\%params)) {
                if ($socket->is_socket_full()) {
                    my $error_msg = "Async process socket is full - ignoring";
                    $self->{logger}->log(msg => $error_msg,
                                         log_level => 'info',
                                         throttled_log_level => 'debug',
                                         throttling_interval => 10) if $self->{logger};
                    die F5::ASMConfig::ASMConfigException->new({error_code => F5::ASMConfig::ErrorCode::ASYNC_SOCKET_FULL,
                                                                error_message => $error_msg});
                }
                else {
                    my $error_msg = "Couldn't pass call to async process - ignoring";
                    $self->{logger}->error($error_msg) if $self->{logger};
                    die F5::ASMConfig::ASMConfigException->new({error_message => $error_msg});
                }
            }
            else {
                $active_async_child->{rpc_call_count}++;
                if ($active_async_child->{rpc_call_count} >= $self->{max_rpc_calls_per_async_handler}) {
                    $self->{logger}->info("Dispatched $active_async_child->{rpc_call_count} calls to pid $active_async_child->{pid} -- call limit reached for this process") if $self->{logger};
                    $active_async_child->{is_active} = 0;
                }
            }
        };
        if ($@) {
            my $err = $@;
            if ($attempt_count == $attempt_count_limit) {
                die $@;
            }
            $self->{logger}->trace("Sleep before next attempt to pass call to async process") if $self->{logger};
            sleep 1;
        }
        else {
            $self->{logger}->debug("Successfully passed call to async process") if $self->{logger};
            last;
        }
    }
}

#----------------------------------------------------------------------

sub spawn_async_handler {
    my($self, %params) = @_;

    # ZMQ handle cannot be used in both parent process and async child. Therefore,
    # delete the current handle (if one exists), and new ones will be created when
    # events need to be sent from each process.
    delete $self->{event_dispatcher};
    $self->close_cgc_channels();

    my $parent_pid = $$;
    my $child_socket = gensym;
    my $parent_socket = gensym;
    socketpair($child_socket, $parent_socket, AF_UNIX, SOCK_STREAM, PF_UNSPEC);

    $child_socket->autoflush(1);
    $parent_socket->autoflush(1);

    if (my $pid = fork) {
        # Parent
        # Increase socket buffer size to allow for more queued async calls
        my $previous_socket_sndbuf = unpack("I", getsockopt($child_socket, SOL_SOCKET, SO_SNDBUF));
        setsockopt($child_socket, SOL_SOCKET, SO_SNDBUF, pack("I", $previous_socket_sndbuf * 3));
        # Set socketpair socket to non-blocking (regular $socket->blocking(0) call doesn't work on these)
        my $socket_flags = fcntl($child_socket, F_GETFL, 0);
        if ($child_socket && $socket_flags) {
            fcntl($child_socket, F_SETFL, $socket_flags | O_NONBLOCK);
        }

        my $async_child = {
                           pid => $pid,
                           socket => $child_socket,
                           is_active => 1,
                           rpc_call_count => 0,
                          };
        $self->{logger}->info("Spawned new async pid $pid") if $self->{logger};
        $async_child->{socket} = F5::IO::Storable->new(socket => $child_socket,
                                                       logger => $self->{logger},
                                                       flush_write_attempt_limit => 1,
                                                       write_buffer_size_limit => 500_000,
                                                       before_close_handler => sub {
                                                           $self->{logger}->log("Async handler $pid has gone away",
                                                                                log_level => 'info',
                                                                                throttled_log_level => 'debug',
                                                                                throttling_interval => 10) if $self->{logger};
                                                           if ($self->{async_children}->{$pid}) {
                                                               $self->{async_children}->{$pid}->{is_active} = 0;
                                                               $self->{async_children}->{$pid}->{was_flush_completed} = 0;
                                                           }
                                                       },
                                             );

        close $parent_socket;
        $self->{async_children}->{$pid} = $async_child;
        return $async_child;
    }
    else {
        # Child
        close $child_socket;
        # Set socketpair socket to non-blocking (regular $socket->blocking(0) call doesn't work on these)
        my $socket_flags = fcntl($parent_socket, F_GETFL, 0);
        if ($parent_socket && $socket_flags) {
            fcntl($parent_socket, F_SETFL, $socket_flags | O_NONBLOCK);
        }

        $self->{dbh}->{InactiveDestroy} = 1;
        $self->reinit(is_async_process => 1,);
        if (!$self->{is_long_lived_async_handler_process}) {
            $self->{max_memory_size} = $self->{async_max_memory_size};
        }

        my $batched_insertor = $self->{batched_insertor};
        $self->{logger}->info("Async handler starting") if $self->{logger};
        my $pulse_msg_interval = 30;

        $self->{queued_method_calls} = [];
        $self->{queued_method_sizes} = [];
        my $parent_closed_connection = 0;
        my $new_method_call_was_read = 0;
        my $total_calls_received = 0;
        my $ignored_call_count = 0;
        my $close_with_parent = $self->{is_long_lived_async_handler_process};
        my $lock_log_attempt_level = 'debug';
        my $lock_dont_block = 1;
        my $has_parent_lock = 0;
        my $queue_limit_exceeded = 0;
        my $last_execute_time = time();
        my $max_queue_size = $self->{max_queued_method_call_total_size_before_active};
        my $async_max_idle_time = $self->{async_max_idle_time};

        my $selector = F5::IO::Selector->new;
        my $read_from_parent = F5::IO::Storable->new(socket => $parent_socket,
                                                     logger => $self->{logger},
                                                     read_message_handler => sub {
                                                         my($parent_connection, $method_call, $method_call_size) = @_;
                                                         my $queued_method_size_total = List::Util::sum(@{$self->{queued_method_sizes}}) || 0;

                                                         $self->{logger}->trace("Async method call added (method count =  " . (1 + scalar @{$self->{queued_method_calls}}) . ", usage = " . ($queued_method_size_total + $method_call_size) . ")") if $self->{logger};
                                                         push(@{$self->{queued_method_calls}}, $method_call);
                                                         push(@{$self->{queued_method_sizes}}, $method_call_size);
                                                         $new_method_call_was_read = 1;
                                                         $total_calls_received++;

                                                         if ($queued_method_size_total >= $max_queue_size) {
                                                             $queue_limit_exceeded = 1;
                                                             $self->{logger}->log(msg => "Async incoming queue has reached maximum call size (usage = $queued_method_size_total)",
                                                                                  log_level => 'info',
                                                                                  throttled_log_level => 'debug',
                                                                                  throttling_interval => 10) if $self->{logger};
                                                         }

                                                         if ($total_calls_received >= $self->{max_rpc_calls_per_async_handler}) {
                                                             $self->{logger}->info("Async handler has received $total_calls_received calls -- call limit reached for this process.") if $self->{logger};
                                                             $close_with_parent = 0;
                                                             $lock_log_attempt_level = 'info';
                                                             $lock_dont_block = 0;
                                                             $parent_connection->close();
                                                         }
                                                     },
                                                     before_close_handler => sub {
                                                         $parent_closed_connection = 1;
                                                     },
                                                     selector => $selector,
                                                    );
        my $last_pulse_time = time;
        my $calls_handled_count = 0;
        my $last_call_name = '';
        my $last_lock_attempt_time = 0;
        my $memory_limit_exceeded_interval = 5;
        my $memory_limit_exceeded_poll_time = 0;
        my $lock_attempt_interval = 2;
        my $lock_file = F5::Deploy::map_path("/ts/lock/asm_config_server_async_for_$parent_pid.lock");
        my $locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                                 logger => $self->{logger},);

        while (1) {
            if ($parent_closed_connection && ($close_with_parent || !@{$self->{queued_method_calls}})) {
                last;
            }

            my $poll_time = undef;
            $new_method_call_was_read = 0;

            my @potential_poll_times;

            if ($batched_insertor && $batched_insertor->has_waiting_records) {
                # Check for records to flush every second
                push(@potential_poll_times, 1);
            }

            if ($has_parent_lock && @{$self->{queued_method_calls}}) {
                if ($self->{memory_limit_exceeded}) {
                    push(@potential_poll_times, $memory_limit_exceeded_poll_time);
                }
                else {
                    push(@potential_poll_times, 0);
                }
            }

            if (!$has_parent_lock) {
                push(@potential_poll_times,
                     F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $lock_attempt_interval,
                                                                         start_time => $last_lock_attempt_time));
            }

            if ($pulse_msg_interval) {
                push(@potential_poll_times,
                     F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $pulse_msg_interval,
                                                                        start_time => $last_pulse_time));
            }

            if (@potential_poll_times) {
                $poll_time = List::Util::min(@potential_poll_times);
            }

            $self->{logger}->trace("About to poll for $poll_time seconds") if $self->{logger};

            if (! $queue_limit_exceeded && $selector->has_handles()) {
                my @readers = $selector->can_read($poll_time);

                for my $reader (@readers) {
                    $reader->handle_read;
                }
            }
            else {
                sleep $poll_time;
            }

            if ($pulse_msg_interval && 
                F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $pulse_msg_interval,
                                                                   start_time => $last_pulse_time) == 0) {
                my @additional_info;
                if ($last_call_name) {
                    push(@additional_info, "last call handled = $last_call_name");
                }
                if (!$has_parent_lock) {
                    push(@additional_info, "Waiting for previous handler to finish");
                    if (scalar @{$self->{queued_method_calls}}) {
                        push(@additional_info, "Async incoming queue has " . scalar @{$self->{queued_method_calls}} . " calls");
                    }
                }

                my $additional_info = '';
                if (@additional_info) {
                    $additional_info = '(' . join(', ', @additional_info) . ')';
                }
                $self->{logger}->info("$calls_handled_count calls handled since async startup $additional_info") if $self->{logger};
                $last_pulse_time = time;
            }

            if (!$has_parent_lock && 
                F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $lock_attempt_interval,
                                                                   start_time => $last_lock_attempt_time) == 0) {
                $has_parent_lock = $locker->lock(dont_block => $lock_dont_block,
                                                 log_attempt_level => $lock_log_attempt_level);
                $last_lock_attempt_time = time();
                if ($has_parent_lock) {
                    $max_queue_size = $self->{max_queued_method_call_total_size};
                    $self->load_async_calls_from_file(parent_pid => $parent_pid,
                                                     );
                }
            }

            if ($has_parent_lock && scalar @{$self->{queued_method_calls}} > 2) {
                $self->{logger}->debug("Async incoming queue has " . scalar @{$self->{queued_method_calls}} . " calls") if $self->{logger};
            }

            if ($has_parent_lock && !@{$self->{queued_method_calls}} &&
                F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $async_max_idle_time,
                                                                   start_time => $last_execute_time) == 0) {
                 $self->{logger}->info("No new calls for $async_max_idle_time for async handler -- exiting") if $self->{logger};
                 last;
            }

            if ($self->{memory_limit_exceeded}) {
                $self->write_async_shutdown_file(parent_pid => $parent_pid, child_pid => $$);
                $memory_limit_exceeded_poll_time = F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $memory_limit_exceeded_interval,
                                                                                                      start_time => $self->{memory_limit_exceeded_shutdown_start_time});
                if ($new_method_call_was_read) {
                    $self->{logger}->trace("New call was read for memory exceeded async process") if $self->{logger};
                    # Reset shutdown start time to ensure that all queued RPC calls are retrieved
                    $self->{memory_limit_exceeded_shutdown_start_time} = time;
                    next;
                }
                elsif ($memory_limit_exceeded_poll_time) {
                    $self->{logger}->trace("No new call was read, but will still poll for another $memory_limit_exceeded_poll_time seconds") if $self->{logger};
                    next;
                }
                else {
                    $self->{logger}->trace("No new calls were read for memory exceeded async process -- exiting") if $self->{logger};
                    last;
                }
            }

            # Skip actual method call handling if new method call was added, to read as many as possible out of the socket buffers
            if (! $new_method_call_was_read && $has_parent_lock && @{$self->{queued_method_calls}}) {
                $last_execute_time = time();
                my $params = shift(@{$self->{queued_method_calls}});
                shift(@{$self->{queued_method_sizes}});
                my $queued_method_size_total = List::Util::sum(@{$self->{queued_method_sizes}}) || 0;
                if ($queued_method_size_total < $max_queue_size) {
                    $queue_limit_exceeded = 0;
                }

                my $call_flags = $params->{rpc_params}->[$CALL_FLAG_INDEX];
                my $max_retry_attempt_upon_async_failure_count = $call_flags->{max_retry_attempt_upon_async_failure_count} || 1;

                for my $attempt_count (1..$max_retry_attempt_upon_async_failure_count) {
                    my $attempt_count_str = '';
                    if ($max_retry_attempt_upon_async_failure_count > 1) {
                        $attempt_count_str = ", attempt $attempt_count";
                    }

                    my $function_name = $params->{function_name};
                    $self->{logger}->debug("Begin async call to $function_name (call id $params->{async_call_id}$attempt_count_str)") if $self->{logger};
                    eval {
                        $self->$function_name(@{$params->{rpc_params}});
                        $last_call_name = $function_name;
                        $calls_handled_count++;
                    };
                    if ($@) {
                        my $err = $@;
                        $self->{logger}->debug("Async call to $function_name ended in error (call id $params->{async_call_id}$attempt_count_str)") if $self->{logger};
                        if (ref($@)) {
                            if (ref($@) eq 'F5::ASMConfig::ASMConfigException' && $attempt_count == 1) {
                                push @{$call_flags->{ignored_errors}}, F5::ASMConfig::IgnoredError->new({ error_code => $err->{error_code},
                                                                                                          throw_exception_on_error => 1, });
                            }
                        }
                        else {
                            # This exception was thrown outside of the eval block in get_install_function, and
                            # has never been reported.
                            $self->{logger}->error("Async call to $function_name ended in error ($@)") if $self->{logger};
                        }
                    }
                    else {
                        $self->{logger}->debug("End async call to $function_name (call id $params->{async_call_id}$attempt_count_str)") if $self->{logger};
                        last;
                    }
                }
            }
            eval {
                $self->check_memory_size();
            };

            if ((($batched_insertor && $batched_insertor->flush_old_records(interval => 3)) || !@{$self->{queued_method_calls}}) &&
                $self->{update_mcp_timestamp}) {

                $self->update_mcp_timestamp();
            }
        }

        if ($batched_insertor) {
            $batched_insertor->flush;
        }
        if ($self->{update_mcp_timestamp}) {
            $self->update_mcp_timestamp();
        }
        if ($self->{memory_limit_exceeded}) {
            $read_from_parent->close;
            unlink($self->get_async_shutdown_filename(parent_pid => $parent_pid, child_pid => $$));
            if (@{$self->{queued_method_calls}}) {
                $self->write_async_calls_to_file(parent_pid => $parent_pid,
                                                 queued_method_calls => $self->{queued_method_calls},
                                                 queued_method_sizes => $self->{queued_method_sizes},
                                                );
            }
        }

        $self->{logger}->debug("Async handler shutdown") if $self->{logger};
        if ($has_parent_lock) {
            $locker->unlock();
        }
        exit;                   # shutdown cleanly
    }
}

#----------------------------------------------------------------------

sub get_async_call_filename {
    my($self, %params) = @_;
    my $parent_pid = $params{parent_pid};
    mkpath($TEMP_FILE_DIR);
    return "$TEMP_FILE_DIR/saved_async_calls_for_pid_$parent_pid";
}

#----------------------------------------------------------------------

sub load_async_calls_from_file {
    my($self, %params) = @_;
    my $parent_pid = $params{parent_pid};

    $self->{logger}->debug("Checking for preserved async calls") if $self->{logger};
    my $async_call_filename = $params{filename};
    if (! $async_call_filename) {
        $async_call_filename = $self->get_async_call_filename(%params);
    }
    if (-f $async_call_filename) {
        my $async_call_file_size = -s $async_call_filename;
        if ($async_call_file_size > $self->{max_queued_method_call_total_size} * 1.5) {
            $self->{logger}->info("Preserved async call file $async_call_filename is larger than queue limit -- discarding contents  (size = $async_call_file_size)") if $self->{logger};
            unlink $async_call_filename;
            return;
        }

        $self->{logger}->info("Loading preserved async calls from file $async_call_filename (size = $async_call_file_size)") if $self->{logger};
        my $new_async_call_filename = $async_call_filename . "__handled_by_$$";
        move($async_call_filename, $new_async_call_filename);
        if (! -f $new_async_call_filename) {
            $self->{logger}->info("Could not move async call file to $new_async_call_filename -- skipping") if $self->{logger};
            return;
        }

        open(my $fh, '<', $new_async_call_filename);
        my $saved_queue = Storable::fd_retrieve($fh);
        close $fh;
        unlink $new_async_call_filename;
        if ($saved_queue) {
            if ($saved_queue->{queued_method_calls}) {
                unshift(@{$self->{queued_method_calls}}, @{$saved_queue->{queued_method_calls}});
                my $queued_method_call_count = scalar @{$saved_queue->{queued_method_calls}};
                $self->{logger}->info("Prepending $queued_method_call_count async calls") if $self->{logger};
            }
            if ($saved_queue->{queued_method_sizes}) {
                unshift(@{$self->{queued_method_sizes}}, @{$saved_queue->{queued_method_sizes}});
            }
        }
    }
}

#----------------------------------------------------------------------

sub write_async_calls_to_file {
    my($self, %params) = @_;
    my $parent_pid = $params{parent_pid};
    my $async_call_filename = $self->get_async_call_filename(%params);
    $self->{logger}->info("Saving remaining async calls for parent pid $parent_pid to file $async_call_filename") if $self->{logger};
    open(my $fh, '>', $async_call_filename);
    store_fd(\%params, $fh);
    close $fh;
}

#----------------------------------------------------------------------

sub get_async_shutdown_filename {
    my($self, %params) = @_;
    my $parent_pid = $params{parent_pid};
    my $child_pid = $params{child_pid};
    return "$TEMP_FILE_DIR/async_shutdown__parent_${parent_pid}__child_${child_pid}";
}

#----------------------------------------------------------------------

sub write_async_shutdown_file {
    my($self, %params) = @_;
    my $filename = $self->get_async_shutdown_filename(%params);
    if (! -f $filename) {
        $self->{memory_limit_exceeded_shutdown_start_time} = time;
        $self->{logger}->info("Writing async shutdown file $filename") if $self->{logger};
        open (my $outf, ">", $filename);
        print $outf $params{child_pid};
        close $outf;
    }
}

#----------------------------------------------------------------------

sub read_method_call {
  my($self, %params) = @_;
  my $socket = $params{socket};

  my $buf;
  recv($socket, $buf, 4, 0);
  unless ($buf) {
      return;
  }
  my $frozen_params_size = unpack('N', $buf);
  recv($socket, $buf, $frozen_params_size, 0);

  my $params = thaw($buf);
  return $params;
}

#----------------------------------------------------------------------

sub have_fields_to_update{
  my ($hash_a,$hash_b) = @_;

  return 0 if !defined $hash_a;  # we only want to compare defined values in the thrift_obj hash

  if (!ref $hash_a and !ref $hash_b){ # neither is a reference
   if ($hash_a ne $hash_b){
     return 1;
   }
  }
  elsif (ref $hash_a xor ref $hash_b){       # one's a reference, one isn't
     return 1;
  }
  else {
    if (ref($hash_a) eq 'SCALAR'){
       if ($hash_a ne $hash_b){
          return 1;
       }
    }
    elsif (ref($hash_a) eq 'ARRAY'){
       if (scalar @$hash_a != scalar @$hash_b){
         return 1;
       }

       for my $i(scalar @$hash_a){
         if (have_fields_to_update($hash_a->[$i],$hash_b->[$i]) ){
           return 1;
         }
       }
    } 
   else {
       for my $key (keys %$hash_a){
         if ( have_fields_to_update($hash_a->{$key},$hash_b->{$key}) ){
            return 1;
         }
       }
    }

 }
 return 0;
}

#----------------------------------------------------------------------

sub base_reset_defaults {
    my($self, %params) = @_;
    my($session, $call_flags, $policy) = @{$params{rpc_params}};
    my $entity_class = $params{entity_class};

    my @entities_changed = $entity_class->reset_defaults(policy => $policy, dbh => $self->{dbh}, session => $session, call_flags => $call_flags);

    my @events = $self->create_events_for_changed_entities(
        session => $session,
        call_flags => $call_flags,
        policy => $policy,
        entities_changed => \@entities_changed);
    my $retval = { return_value => 1, events => \@events};
    return $retval;
}

#----------------------------------------------------------------------

sub base_update_many {
    my($self, %params) = @_;

    my($session, $call_flags, $policy, $search_criteria, $store_obj) = @{$params{rpc_params}};
    my $entity_class = $params{entity_class};

    my $entities_changed;
    eval {
        $entities_changed = $entity_class->update_many(
                                                       session => $session,
                                                       call_flags => $call_flags,
                                                       policy => $policy,
                                                       search_criteria => $search_criteria,
                                                       store_obj => $store_obj,
                                                       page_size => $call_flags->{page_size} || (!$call_flags->{inherited_call} && $self->{default_page_size}),
                                                       start_page => (defined $call_flags->{start_page}) && ($call_flags->{start_page} > 0)
                                                                     ? $call_flags->{start_page}
                                                                     : 1,
                                                       dbh => $self->{dbh},
                                                      );
    };
    if ($@) {
        return $self->handle_error(action => 'update',
                                   error_thrown => $@,
                                   call_flags => $call_flags,
                                  );
    }

    if (! $entities_changed) {
        $entities_changed = {};
    }
    $entities_changed->{updated_entities} ||= [];
    $entities_changed->{modified_entities} ||= [];

    my @events = $self->create_events_for_changed_entities(
                                                           session => $session,
                                                           call_flags => $call_flags,
                                                           policy => $policy,
                                                           entities_changed => $entities_changed->{updated_entities});

    push(@events, $self->create_events_for_sub_entities(modified_entities => $entities_changed->{modified_entities},
                                                        session => $session,
                                                        call_flags      => $call_flags,
                                                        policy          => $policy,
                                                        )
    );

    my $rv;
    if ($call_flags->{rest_call}) {  #Return a search result if this is a rest call
        my @return_entities = grep { $_->element_type == $entity_class->element_type } (@{$entities_changed->{updated_entities}});
        $rv = $self->entities_to_search_result(entities => \@return_entities,
                                               call_flags => $call_flags,
                                               entity_class => $entity_class,);
    }
    else {
        $rv = scalar @events;
    }

    my $retval = { return_value => $rv, events => \@events, entities_changed => $entities_changed->{updated_entities} };
    return $retval;
}

#----------------------------------------------------------------------

sub base_delete_many {
    my($self, %params) = @_;
    my($session, $call_flags, $policy, $search_criteria) = @{$params{rpc_params}};
    my $entity_class = $params{entity_class};
    my $suppress_save_to_log = defined $params{suppress_save_to_log} ? $params{suppress_save_to_log} :  0;

    my @entities_changed;
    eval {
        push(@entities_changed, $entity_class->delete_many(
            session => $session,
            call_flags => $call_flags,
            policy => $policy,
            search_criteria => $search_criteria,
            find_single_record => 0,
            search_filters => $call_flags->{search_filters},
            page_size => $call_flags->{page_size} || (!$call_flags->{inherited_call} && $self->{default_page_size}),
            start_page => (defined $call_flags->{start_page}) && ($call_flags->{start_page} > 0) ? $call_flags->{start_page} : 1,
            dbh => $self->{dbh},));
    };
    if ($@) {
        return $self->handle_error(action => 'delete',
                                   error_thrown => $@,
                                   call_flags => $call_flags,
                                  );
    }

    my @events = $self->create_events_for_changed_entities(
                                                           session => $session,
                                                           call_flags => $call_flags,
                                                           policy => $policy,
                                                           entities_changed => \@entities_changed,
                                                           event_type => F5::ASMConfig::EventType::DELETE,
                                                           suppress_save_to_log => $suppress_save_to_log,
                                                          );


    my $rv;
    if ($call_flags->{rest_call}) {  #Return a search result if this is a rest call
        my @return_entities = grep { $_->element_type == $entity_class->element_type } @entities_changed;
        $rv = $self->entities_to_search_result(entities => \@return_entities,
                                               call_flags => $call_flags,
                                               entity_class => $entity_class,);
    } 
    else {
        $rv = scalar @events;
    }

    my $retval = { return_value => $rv, events => \@events, entities_changed => \@entities_changed};
    return $retval;
}

#----------------------------------------------------------------------

sub entities_to_search_result {
    my ($self, %params) = @_;
    my @entities = @{$params{entities}};
    my $call_flags = $params{call_flags};
    my $entity_class = $params{entity_class};

    my $result_record_count = scalar @entities;
    my $total_record_count = $params{total_record_count} || $result_record_count;
    my $max_page = 1;
    if ($call_flags->{page_size}) {
        $max_page = $total_record_count / $call_flags->{page_size};
        $max_page = int($max_page) + (int($max_page) == $max_page ? 0 : 1);
    }

    my ($element_type, $element_type_name);
    if ($entity_class) {
        $element_type = $entity_class->element_type; 
        $element_type_name = $entity_class->element_type_name; 
    }

    my $element_list;
    if ($call_flags->{rest_call}) {
        $element_list = [ 
                          map { ({
                                  element_type => ($element_type || $_->element_type),
                                  ($element_type_name || $_->element_type_name) => $_,
                                  grouping_count => $_->{grouping_count},
                                })
                              } @entities
                        ];
    }
    else {
        $element_list = [
                          map {
                            F5::ASMConfig::Element->new({
                                                         element_type => ($element_type || $_->element_type),
                                                         ($element_type_name || $_->element_type_name) => $_,
                                                         grouping_count => $_->{grouping_count},
                                                        })
                            } @entities
                        ];
    }
    my $search_result = F5::ASMConfig::SearchResult->new({
                                                        total_record_count => $total_record_count,
                                                        max_page => $max_page,
                                                        result_record_count => $result_record_count,
                                                        start_page => ($call_flags->{start_page} || 1),
                                                        element_list => $element_list,
                                                       });

    return $search_result;
}

#----------------------------------------------------------------------

sub write_to_syslog {
    my($self, %params) = @_;

    my $msg = $params{msg};
    F5::BigIpLog::bigip_ts_ts_info_msg_log($msg);
}

#----------------------------------------------------------------------
sub base_reset_default_rows{
  my($self, %params) = @_;
  my($session, $call_flags, $policy) = @{$params{rpc_params}};
  my $entity_class = $params{entity_class};

  my @events;
  my $retval = $self->base_delete_many(entity_class => $entity_class,
                                       rpc_params => [ $session, $call_flags, $policy ]);

  push ( @events, @{$retval->{events}} ) if defined $retval->{events};

  my $rv = $self->base_get_default_rows( entity_class => $entity_class, 
                                         default_table => $params{default_table},
                                         search_filters => $params{search_filters},
                                       );

  my $default_elements = $rv->{element_list};

  for my $default_element (@$default_elements){
     my $thrift_obj = $entity_class->new(object => $default_element->{$entity_class->element_type_name}, dbh => $self->{dbh} );
     $thrift_obj->{policy}->{policy_id} = $policy->{policy_id} if defined $policy;
     $retval = $self->base_add(entity_class => $entity_class,
                               rpc_params => [ $session, $call_flags, $policy,$thrift_obj ]);
     push ( @events, @{$retval->{events}} ) if defined $retval->{events};
  }

  return  { return_value => 1,
            events => \@events};
}
#----------------------------------------------------------------------
sub base_get_default_rows{
 my($self, %params) = @_;
 my $dbh =  $self->{dbh};

 my $result_set = $params{entity_class}->retrieve_all( dbh => $dbh,
                                                       logger => $self->{logger},
                                                       default_table => $params{default_table}, 
                                                       find_single_record => 0,
                                                       search_filters => $params{search_filters},
                                                      );
 my @entities = $result_set->get_all;

 my $result_record_count = scalar @entities;
 my $max_page = 1;

 my $search_result = $self->entities_to_search_result(entities => \@entities,
                                                      call_flags => $params{call_flags},
                                                      entity_class => $params{entity_class},
                                                      );
 return $search_result;
}

#----------------------------------------------------------------------

sub add_partition_search_filter {
    my ($self, %params) = @_;
    my $session    = $params{session};
    my $call_flags = $params{call_flags};
    my $search_field = $params{search_field} || 'partition_id';

    if ((!$call_flags->{rest_call} || $call_flags->{partition_was_defined}) && $session->{partition} ne F5::ASMConfig::Constants::PARTITION_ALL && !$self->{em_client}) {
        my %partitions = (F5::ASMConfig::Constants::PARTITION_COMMON => 1,
                          $session->{partition} => 1,
                         );
        my $search_filter = F5::ASMConfig::SearchFilter->new(field_name => $search_field,
                                                             field_value_list => [keys %partitions],
                                                             field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN());
        push (@{$call_flags->{search_filters}}, $search_filter);
    }
}

#----------------------------------------------------------------------

sub check_ucs_load {
   my ($self) = @_;

   if (-e $ASM_UCS_INSTALL_PROC_FILENAME) {
       if (open (my $fh, "<", $ASM_UCS_INSTALL_PROC_FILENAME)) {
           my $ucs_install_pid = <$fh>;
           chomp($ucs_install_pid);
           close($fh);
           if ( !$ucs_install_pid || !F5::NwdUtils::get_process_stat($ucs_install_pid)->{pid}) {
                unlink $ASM_UCS_INSTALL_PROC_FILENAME;
           }
           else {
               die { error_level => "info",
                     error_message => "Received updates from tsconfd while a UCS upgrade is running.  Ignoring.",
                     error_code =>  F5::ASMConfig::ErrorCode::ILLEGAL_OPERATION(),
               };
           }
       }
   }
}

#----------------------------------------------------------------------

sub check_memory_size {
    my ($self, %params) = @_;

    if (F5::BigipUtils::is_feature_enabled('asmconfig_check_memory_size') &&
        F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{check_memory_size_interval},
                                                           start_time => $self->{max_memory_size_check_time}) == 0) {
        $self->{last_memory_check_rss} = F5::NwdUtils::get_process_stat()->{rss};
        if ($self->{last_memory_check_rss} > $self->{max_memory_size}) {
            $self->{memory_limit_exceeded} = 1;
            if (! $self->{memory_limit_exceeded_start_time}) {
                $self->{memory_limit_exceeded_start_time} = time;
            }
        }
        $self->{max_memory_size_check_time} = time();
        $self->{logger}->debug("Process size: $self->{last_memory_check_rss}") if $self->{logger};
        if (! $self->{is_relay_handler_process} && exists $self->{cgc_channels} && F5::Cgc::Channel::DeviceGroup::has_config_changed()) {
            $self->{logger}->info("CGC topology has changed, closing publisher channels");
            $self->close_cgc_channels();
        }
    }
    if ($self->{memory_limit_exceeded}) {
        return $self->handle_error(error_thrown => "Process size ($self->{last_memory_check_rss}) has exceeded max size ($self->{max_memory_size})",
                                   error_code => F5::ASMConfig::ErrorCode::MEMORY_LIMIT_EXCEEDED,
                                  );
    }

    return;
}

#----------------------------------------------------------------------
sub DESTROY {
    my ($self) = @_;
    if (exists $self->{cgc_channels}) {
        $self->{logger}->debug("CGC channels exist.  Waiting to clear") if $self->{logger};
        sleep ($self->{sleep_on_destroy} || 3);  #Allow any pending CGC messages to be delivered before closing.
    }
    if ($self->{update_mcp_timestamp}) {
        $self->update_mcp_timestamp();
    }
    if ($self->{alarm_on_destroy}) {
        alarm 10;
    }
    return;
}

1;
