package F5::ASMConfig::Handler::PBConfig;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::Deploy;
use F5::ASMConfig::ASMConfig;
use F5::DeviceGroup;
use F5::Cluster::Utils;

use XML::Twig;
use File::Basename;
use YAML::XS;
use Memoize;
memoize('get_pb_affects_wildcard_config');


our @handled_functions = (
                           {
                             function_name => 'add_pbconfig',
                             relayable     => 1,
                             store_index   => 3,
                           },

                           {
                             function_name => 'update_pbconfig',
                             relayable => 1,
                             criteria_index => 3,
                             store_index => 4,
                           },
                           {
                             function_name => 'send_pbconfig_event',
                           },
                           {
                             function_name => 'add_pbconfig_event',
                           },
                         );

our $entity_class = 'F5::ASMConfig::Entity::PBConfig';
my $PURE_WILDCARD_PBCONFIG_FILENAME = F5::Deploy::map_path('/usr/share/ts/config/pb_affects_pure_wildcard_settings.yaml');

my %map_const_vals = ( F5::ASMConfig::TighteningMode::TIGHTENING_MODE_NEVER()     => 'F5::ASMConfig::TighteningMode::TIGHTENING_MODE_NEVER',
                       F5::ASMConfig::TighteningMode::TIGHTENING_MODE_EXCEPTION() => 'F5::ASMConfig::TighteningMode::TIGHTENING_MODE_EXCEPTION',
                       F5::ASMConfig::TighteningMode::TIGHTENING_MODE_ALWAYS()    => 'F5::ASMConfig::TighteningMode::TIGHTENING_MODE_ALWAYS',
                       F5::ASMConfig::TighteningMode::TIGHTENING_MODE_COMPACT()   => 'F5::ASMConfig::TighteningMode::TIGHTENING_MODE_COMPACT',
                      );

#----------------------------------------------------------------------

sub send_pbconfig_event {
  my($self, $session, $call_flags, $policy) = @_;

  my $dbh = $self->{dbh};

  my $pbconfig = F5::ASMConfig::Entity::PBConfig->new(dbh => $dbh,
                                                      get_criteria => {policy_id => $policy->{policy_id}},
                                                      policy => $policy);

  my $retval = { return_value => undef };

  if ($pbconfig) {
      $retval->{events} = [ $self->create_event(event_type => F5::ASMConfig::EventType::UPDATE,
                                                policy => $policy,
                                                session => $session,
                                                element => $pbconfig,
                                                call_flags => $call_flags,
                                                user_name => $session->{user_name},
                                                skip_inheritance => 1,
                                                suppress_save_to_log => 1,
                                               )
                          ];
  }

  return $retval;
}

#----------------------------------------------------------------------
sub add_pbconfig {
   my($self, $session, $call_flags, $policy, $pbconfig) = @_;
   if ($self->{is_relay_handler_process} && $pbconfig->{flg_enabled}
       && !F5::Cluster::Utils::is_failover_standby() && F5::DeviceGroup::is_device_active($call_flags->{originator})) {
       $pbconfig->{flg_enabled} = 0;
   }
   return $self->base_add(rpc_params => [$session, $call_flags, $policy, $pbconfig],
                          entity_class => $entity_class,
                         );
}
#----------------------------------------------------------------------

sub add_pbconfig_event {
    my($self, $session, $call_flags, $policy) = @_;
    if (!$call_flags->{suppress_events}) {
        my @policies;
        if ($policy) {
            @policies = ($policy);
        }
        elsif ($call_flags->{search_filters} && (my ($policy_filter) = grep { $_->{field_name} eq 'policy_id' } @{$call_flags->{search_filters}})) {
            @policies = map { F5::ASMConfig::Cache::get_cached_policy( policy => { policy_id => $_ }) } @{$policy_filter->{field_value_list}};
        }

        for my $policy (@policies) {
            $self->{send_pbconfig_event}->{$policy->{policy_id}} = $policy;
        }
    }
}

#----------------------------------------------------------------------

sub get_pb_affects_wildcard_config {
    my $config_filename = $ENV{PURE_WILDCARD_PBCONFIG_FILENAME} || $PURE_WILDCARD_PBCONFIG_FILENAME;
    return YAML::XS::LoadFile($config_filename);
    
}

#----------------------------------------------------------------------

sub update_pbconfig {
    my($self, $session, $call_flags, $policy, $pbconfig_criteria, $store_pbconfig) = @_;

    if ($self->{is_relay_handler_process} && $store_pbconfig && $store_pbconfig->{flg_enabled}
        && !F5::Cluster::Utils::is_failover_standby() && F5::DeviceGroup::is_device_active($call_flags->{originator})) {
        $store_pbconfig->{flg_enabled} = 0;
    }

    my $pb_affects_wildcard_config = get_pb_affects_wildcard_config();

    my $orig_pbconfig = F5::ASMConfig::Entity::PBConfig->new(dbh => $self->{dbh},
                                                             policy => $policy,
                                                             get_criteria => $pbconfig_criteria );
    my $was_pb_affected_field_changed = 0;

    for my $field (keys %$pb_affects_wildcard_config) {
        if (defined $store_pbconfig->{$field}) {
            $was_pb_affected_field_changed = 1;
        }
        elsif (my $composite_field = $entity_class->fields->{$field}->{rest_composite_field}) {
            if ($store_pbconfig->{$composite_field} && defined $store_pbconfig->{$composite_field}->{$field}) {
                $store_pbconfig->{$field} = $store_pbconfig->{$composite_field}->{$field};
                $was_pb_affected_field_changed = 1;
            }
        }
    }

    if ($orig_pbconfig && $orig_pbconfig->has_value_changed(store_obj => $store_pbconfig, check_fields => 'flg_learning')
        || $was_pb_affected_field_changed ) {

        $orig_pbconfig->populate_diff_structs();
    }

    my $retval =  $self->base_update(rpc_params => [$session, $call_flags, $policy, $pbconfig_criteria, $store_pbconfig],
                                     entity_class => $entity_class,
                                    );

    my $learning_changed = 0;

    if ( defined $store_pbconfig->{flg_learning} && 
         ($store_pbconfig->{flg_learning} ne $orig_pbconfig->{flg_learning}) ) {

        $learning_changed = 1;
        $policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, policy => $policy, force_reload => 1);
    }

    my $cascading_call_flags = F5::ASMConfig::CallFlags->new(%$call_flags);
    my @changed_fields;
    for my $field (keys %$pb_affects_wildcard_config) {
        if (my $composite_field = $entity_class->fields->{$field}->{rest_composite_field}) {
            $orig_pbconfig->{$field} = $orig_pbconfig->{$composite_field}->{$field};
            $retval->{return_value}->{$field} = $retval->{return_value}->{$composite_field}->{$field};
        }
        if ( $learning_changed || (defined $orig_pbconfig->{$field} && defined $store_pbconfig->{$field} && ( $store_pbconfig->{$field} ne $orig_pbconfig->{$field}))) {
            push (@changed_fields, $field);
        }
    }
    if (@changed_fields) {
        my $rv = apply_pb_changes_to_wildcard_config($self,(changed_fields => \@changed_fields,
                                                            policy => $policy,
                                                            pb_affects_wildcard_config => $pb_affects_wildcard_config,
                                                            update_thrift_obj => $store_pbconfig,
                                                            pbconfig => $retval->{return_value},
                                                            session => $session,
                                                            call_flags => $cascading_call_flags));
        if ($rv->{events}) {
            push(@{$retval->{events}}, @{$rv->{events}});
        }
    }
    return $retval;
  
}
#----------------------------------------------------------------------

sub apply_pb_changes_to_wildcard_config {
    my ($self, %params) = @_;
    my $pb_affects_wildcard_config = $params{pb_affects_wildcard_config} || get_pb_affects_wildcard_config();
    my @changed_fields = $params{changed_fields} ? @{$params{changed_fields}} : ();
    my $policy = $params{policy};
    my $pbconfig = $params{pbconfig};
    my $store_pbconfig = $params{update_thrift_obj};
    my $call_flags = $params{call_flags};
    my $session = $params{session};
    my $update_wildcard_entities = !defined($call_flags->{update_wildcard_entities}) || $call_flags->{update_wildcard_entities};

    #Do not apply changes when coming from diff
    return {} if $call_flags->{internal_flags_hash}->{skip_populate_defaults_on_add};

    my @events;

    if (!@changed_fields) {
        @changed_fields = grep { defined($store_pbconfig->{$_}) } (keys %$pb_affects_wildcard_config);
    }

    for my $field (@changed_fields) {
        my $class = $pb_affects_wildcard_config->{$field}->{class};
        if (!$pbconfig) {
            $pbconfig = F5::ASMConfig::Cache::get_cached_pb_config(dbh => $self->{dbh}, policy => $policy);
        }

        if ($policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()) {
            my $entity_config = $class->entity_config();
            next if $entity_config->{no_inheritance} || $entity_config->{no_inheritance_conditions};
        }

        my $store_object_vals = {};
        if ( (my $field_def = $pb_affects_wildcard_config->{$field}->{store_selected_values}) && $update_wildcard_entities ) {

            $pbconfig->populate_blocking_settings();

            if (!defined($pbconfig->{$field})) {
                $pbconfig->{$field} = $store_pbconfig->{$field};
            }

            set_store_obj_vals(self => $self,
                               pbconfig => $pbconfig,
                               base_field => $field,
                               field_def => $field_def,
                               store_obj_vals => $store_object_vals);
        }

        my $search_criteria;
        if ($class eq 'F5::ASMConfig::Entity::Parameter') {
            $search_criteria = { param_level => F5::ASMConfig::ParamLevel::PARAM_LEVEL_GLOBAL()};
        }

        if ($class->name_field()) {
            $search_criteria->{$class->name_field()} = "*";
        }

        $search_criteria->{is_wildcard} = 1;
        $call_flags->{ignore_nonexistent_record_on_update} = 1;
        $call_flags->{add_on_updating_nonexistent_record} = 0;
        $call_flags->{skip_inheritance} = 1;

        my $add_on_updating_nonexistent_record;
        if ( $update_wildcard_entities && $pbconfig->{$field} && $pbconfig->{flg_learning} &&
             ($pbconfig->{$field} == F5::ASMConfig::TighteningMode::TIGHTENING_MODE_EXCEPTION()
              || $pbconfig->{$field} == F5::ASMConfig::TighteningMode::TIGHTENING_MODE_COMPACT()) ) {

            $add_on_updating_nonexistent_record = 1;
        }

        my @search_criteria_list;
        if ($field eq 'flg_urls') {
            @search_criteria_list = ({%$search_criteria, app_id => 0});
            if (!$policy->{flg_is_protocol_free}) {
                push (@search_criteria_list, {%$search_criteria, app_id => 1});
            }
        }
        elsif ($field eq 'flg_websocket_urls') {
            @search_criteria_list = ({%$search_criteria, protocol => 0});
            if (!$policy->{flg_is_protocol_free}) {
                push(@search_criteria_list, {%$search_criteria, protocol => 1});
            }
        }
        else {
            @search_criteria_list = ($search_criteria);
        }

        for my $get_criteria (@search_criteria_list) {
            my $store_obj = $class->new(policy => $policy,
                                        dbh => $self->{dbh},
                                        object => { %$store_object_vals, has_tightening => $pbconfig->{$field} });

            my $exists = $class->new(dbh => $self->{dbh}, policy => $policy, get_criteria => $get_criteria);
            if ($exists) {
                delete $store_obj->{in_staging};
            }

            my $missing_entities_handler = $pb_affects_wildcard_config->{$field}->{missing_entities_handler};
            if ($update_wildcard_entities && $missing_entities_handler && ! $exists) {
                my $missing_entity = $class->$missing_entities_handler(dbh => $self->{dbh}, policy => $policy, store_obj => $store_obj);
                if ($missing_entity && $missing_entity->{__new_entity}) {
                    my $missing_rv = $self->base_add( entity_class => ref($missing_entity),
                                                      rpc_params => [ $session, $call_flags, $policy, $missing_entity ],);
                    if ($missing_rv && $missing_rv->{events}) {
                        push (@events, @{$missing_rv->{events}});
                    }
                }
            }

            my $rv;
            if ($exists) {
                $rv = $self->base_update( entity_class => $class,
                                          rpc_params => [ $session, $call_flags,$policy, $get_criteria, $store_obj ],);
            }
            elsif ($add_on_updating_nonexistent_record) {
                $store_obj->update_default_object_by_pb_settings(default_obj => $store_obj,
                                                                 policy => $policy);
                my $add_object = {%$get_criteria, %$store_obj};
                $rv = $self->base_add( entity_class => $class,
                                       rpc_params => [ $session, $call_flags,$policy, $add_object ],);
            }

            if ($rv && $rv->{events}) {
                 push (@events, @{$rv->{events}});
            }
        }
    }
    return { return_value => $store_pbconfig,
             events => \@events};
}

#----------------------------------------------------------------------
sub set_store_obj_vals{
  my(%params) = @_;
  my $self = $params{self};
  my $field_def = $params{field_def};
  my $store_obj_vals = $params{store_obj_vals};

  return unless $field_def;

  if ( $field_def->{apply_values}){
      for my $val (keys %{$field_def->{apply_values}}){
          if($field_def->{apply_values}->{$val} =~ /^F5::.*$/) {
              $field_def->{apply_values}->{$val} = eval($field_def->{apply_values}->{$val});
          }
          $store_obj_vals->{$val} = $field_def->{apply_values}->{$val};
      }
  }

  if ( $field_def->{check_field} ){
      for my $k ( keys %{$field_def->{check_field}}){
           my $field_name = $k;
           my $field_value;
           if ( $field_name eq 'tightening_mode' ){
               my $actual_field_name = $params{base_field};
               my $actual_field_value = $params{pbconfig}->{$actual_field_name};
               $field_value = $map_const_vals{$actual_field_value};
            }
            else{
                $field_value = $params{pbconfig}->{$field_name};
            }
           set_store_obj_vals(%params,
                           field_def => $field_def->{check_field}->{$k}->{$field_value});
       }
  }
 
}

#----------------------------------------------------------------------
1;
