package F5::ASMConfig::Handler::UpdateSignaturesTask;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::TaskBase/;
use F5::ASMConfig::Entity::UpdateSignaturesTask;
use F5::Utils::Rest;
use File::Copy;
use MIME::Base64;

our @handled_functions = (
                            {
                             function_name => 'add_update_signatures_task',
                             criteria_index => 3,
                             store_index    => 4,
                            },
                            {
                              function_name => 'update_update_signatures_task',
                              criteria_index => 3,
                              store_index => 4
                            },
                            { 
                             function_name => 'execute_update_signatures_task',
                            },
                            {
                             function_name => 'delete_update_signatures_task',
                            },
                            {
                             function_name => 'delete_many_update_signatures_tasks',
                            },
                         );

our $entity_class = 'F5::ASMConfig::Entity::UpdateSignaturesTask';

#----------------------------------------------------------------------

sub add_update_signatures_task {
  my($self, $session, $call_flags, $place_holder, $update_signatures_task) = @_;
  if ( !$update_signatures_task->{is_user_defined} ) {
      #Backwards compatibility mode for official update file
      $call_flags->{rest_response_parameters}->{resource_deprecated} = 1;
  }
  return F5::ASMConfig::Handler::TaskBase::add_task(self => $self,
                                                    rpc_params => [ $session, $call_flags, $place_holder, $update_signatures_task ],
                                                    entity_class => $entity_class,
                                                    task_type => 'update_signatures',);
}

#----------------------------------------------------------------------

sub update_update_signatures_task {
    my ($self, @rpc_params) = @_;
    return F5::ASMConfig::Handler::TaskBase::update_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                         task_type => 'update_signatures',);
}

#----------------------------------------------------------------------

my $LIVE_UPDATE_URL = "/mgmt/tm/live-update";
my $LIVE_UPDATE_DIR = '/var/lib/hsqldb/live-update/update-files';
my $TIMEOUT = 30*60; #30 Minutes
my $CHECK_INTERVAL = 3;

sub update_signatures {
    my (%params) = @_;
    my $self       = $params{self};
    my $session    = $params{session};
    my $call_flags = $params{call_flags};
    my $task       = $params{task};
    my $logger     = $self->{logger};

    my $filename = $task->{filename};
    my $remove_file = 0;
    eval {
        if ($filename) {
            $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{filename} );
        }
        if ($task->{file_contents}) {
            if (!$filename) {
                $task->{filename} = $task->{rest_uuid};
                $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{filename} );
                $remove_file = 1;
            }
            open (my $FH, '>', $filename) || $self->handle_error(action => 'execute',
                                                                 entity_class => $entity_class,
                                                                 thrift_obj  => $task,
                                                                 error_code  => F5::ASMConfig::ErrorCode::PERMISSION_DENIED(),
                                                                 msg         => "Could not open $filename for writing.",
                                                                 error_level  => "info", );
           my $contents = $task->{file_contents};
           if ($task->{is_base64}) {
               $contents = MIME::Base64::decode_base64($contents);
           }
           print $FH $contents;
           close ($FH);
        }

        if ($task->{is_user_defined}) {
            #Only add to transaction locker if we stay internal to ASMConfig
            if ($self->{in_transaction}) {
                $self->{transaction_locker}->{$F5::DbUtils::LOCK_FILE} ||= F5::LockFactory::get_locker(lock_filename => $F5::DbUtils::LOCK_FILE,
                                                                                                       logger => $logger,);
                $self->{transaction_locker}->{$F5::DbUtils::LOCK_FILE}->lock(single_lock => 1);
            }
            my $sigfile = F5::ASMConfig::Sigfile->new(filename => $filename,
                                                      flg_is_user_defined => $task->{is_user_defined});
            my $settings = F5::ASMConfig::Entity::SigfileUpdateSettings->new(dbh => $self->{dbh}, get_criteria => {}, logger => $logger);
            $settings->{allow_older_timestamp} = $task->{allow_older_timestamp};
            delete $settings->{check_interval}; #Indicates this is a manual call

            $self->begin_work($self->{session});
            my $rv = $self->update_sigfile($session, $call_flags, undef, $sigfile, $settings);
            if ($settings->{dry_run}) {
                $self->rollback($self->{session});
            }
            else {
                $self->commit($self->{session});
            }

            if ($rv->{completed_successfully} && $rv->{sigfile_status_id}) {
                $task->{sigfile_update_result} = $rv;
                $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_COMPLETED();
            }
            else {
                $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_FAILURE();
                $task->{message} = $rv->{error_msg};
            }
        }
        else {
            #Backwards compatibility mode for official update file
            $call_flags->{rest_response_parameters}->{resource_deprecated} = 1;
            my $installations = {};
            if ($filename) {
                my $update_file;
                if ($task->{filename}) {
                    #Check for existance
                    open (my $fh, '<', $filename) || die "File '$task->{filename}' not found\n";
                    my $md5 = uc(Digest::MD5->new->addfile($fh)->hexdigest());
                    close ($fh);
                    #currently no way to overwrite an existing file with the name name, so reuse any file with the same md5
                    my $existing_file_rv = F5::Utils::Rest::local_request(uri => "$LIVE_UPDATE_URL/update-files",
                                                                          parameters => { '$filter' => "md5 eq '$md5' or filename eq '$task->{filename}'", },
                                                                          logger => $logger,
                                                                         );
                    if ($existing_file_rv->{totalItems}) {
                        if (($update_file) = grep { $_->{md5} eq $md5 && $_->{filename} eq $task->{filename} } @{$existing_file_rv->{items}}) {
                            $logger->debug("Found existing file with md5 $md5");
                        }
                        else {
                            die "File '$task->{filename}' already exists with a different MD5\n";
                        }
                    }
                }

                if (!$update_file) {
                    #copy filename to remove the username from the file
                    $logger->debug("Adding $task->{filename} to Live Update");
                    copy($filename, "$LIVE_UPDATE_DIR/$task->{filename}");
                    $filename = "$LIVE_UPDATE_DIR/$task->{filename}";
                    if (my ($login,$pass,$tomcat_uid,$tomcat_gid) = getpwnam("tomcat")) {
                        chown $tomcat_uid, $tomcat_gid, $filename;
                    }

                    $update_file = F5::Utils::Rest::local_request(uri => "$LIVE_UPDATE_URL/update-files",
                                                                  method => "POST",
                                                                  body => { fileLocationReference => { link => $filename } },
                                                                  logger => $logger,
                                                                 );
                }
                my $update_type;
                if ($update_file->{selfLink} =~ /live-update\/([^\/]*)\/update-files/ ) {
                    $update_type = $1;
                }
                else {
                    die "Error in file upload\n";
                }

                my $rv = F5::Utils::Rest::local_request(uri => "$LIVE_UPDATE_URL/$update_type/installations",
                                                        parameters => { '$filter' => "updateFile.isFileAvailable eq true and updateFile.id eq '$update_file->{id}'",
                                                                        '$top'   => 1,
                                                                        '$orderby' => 'lastUpdateMicros desc',
                                                                        '$select'  => 'status,id',
                                                                       },
                                                        logger => $logger);
                my $create_file = 1;
                if ($rv->{totalItems}) {
                       my ($item) = @{$rv->{items}};
                       if ($item->{status} eq 'install-complete' || $item->{status} eq 'install-in-progress') {
                           $task->{message} = "The provided file is in state: $item->{status}";
                           $create_file = 0;
                       }
                       elsif ($item->{status} ne 'previously-installed' && $item->{status} ne 'error' ) {
                           $create_file = 0;
                           $installations->{$update_type} = $item;
                       }
                }
                if ($create_file) {
                    $installations->{$update_type} = F5::Utils::Rest::local_request(uri => "$LIVE_UPDATE_URL/$update_type/installations",
                                                                                    method => "POST",
                                                                                    body => { updateFileReference => { link => $update_file->{selfLink} },
                                                                                              status => 'install', },
                                                                                    logger => $logger,
                                                                                   );
                }
            }
            else {
                #No provided file, ensure the latest update file is installed.
                my @types = @{F5::ASMConfig::Entity::SigfileUpdateSettings->live_update_write_types()};
                for my $update_type (@types) {
                    $logger->debug("Checking for update for $update_type");
                    my $rv = F5::Utils::Rest::local_request(uri => "$LIVE_UPDATE_URL/$update_type/installations",
                                                     parameters => { '$filter' => 'updateFile.isFileAvailable eq true and updateFile.isFileManuallyUploaded eq false',
                                                                     '$top'   => 1,
                                                                     '$orderby' => 'updateFile.createDateTime desc, lastUpdateMicros desc',
                                                                     '$select'  => 'status,id',
                                                                   },
                                                     logger => $logger);
                   if ($rv->{totalItems}) {
                       my ($item) = @{$rv->{items}};
                       if ($item->{status} eq 'previously-installed' || $item->{status} eq 'error') {
                           #Create a new one
                           $installations->{$update_type} = F5::Utils::Rest::local_request(uri => "$LIVE_UPDATE_URL/$update_type/installations",
                                                                                           method => "POST",
                                                                                           body => { updateFileReference => $item->{updateFileReference},
                                                                                                     status => 'install', } ,
                                                                                           logger => $logger,
                                                                                          );
                       }
                       elsif ($item->{status} !~ /install/) {
                           #Statuses other than of install, install-in-progress, and install-complete
                           $installations->{$update_type} = $item;
                       }
                       #Otherwise it is already installed
                    }
                }
            }

            my %FINAL_STATUS = ('error' => 1,
                                'install-complete' => 1,
                               );

            for my $update_type (keys %$installations) {
                $logger->debug("Installing update for $update_type");
                my $installation = $installations->{$update_type};
                if ($installation->{status} ne 'install') {
                    $installation = F5::Utils::Rest::local_request(uri => "$LIVE_UPDATE_URL/$update_type/installations/$installation->{id}",
                                                                   method => "PATCH",
                                                                   body => { status => 'install' },
                                                                   logger => $logger,
                                                                  );
                }
                my $expire_time = time() + $TIMEOUT; #Give up after 30 minutes
                while (!$FINAL_STATUS{$installation->{status}}) {
                    if (time() > $expire_time) {
                        die "Maximum time exceeded\n";
                    }
                    sleep $CHECK_INTERVAL;
                    $installation = F5::Utils::Rest::local_request(uri => "$LIVE_UPDATE_URL/$update_type/installations/$installation->{id}",
                                                                   parameters => { '$select'  => 'status,id,errorMessage', },
                                                                   logger => $logger,
                                                                  );
                }
                if ($installation->{status} eq 'install-complete') {
                    $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_COMPLETED();
                    if ($update_type eq 'asm-attack-signatures') {
                        my $result = F5::ASMConfig::Entity::SigfileUpdateRes->new(dbh => $self->{dbh},
                                                                           get_criteria => {},
                                                                           selected_fields => ['sigfile_status_id'],
                                                                           sort_orders => [ { field_name => 'sigfile_status_id',
                                                                                              sort_direction => F5::ASMConfig::SortDirection::SORT_DIRECTION_DESC()}],
                                                                           );
                        $task->{sigfile_update_result} = $result;
                    }
                }
                else {
                    $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_FAILURE();
                    $task->{message} = $installation->{errorMessage};
                }
            }

            if (!%$installations) {
                $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_FAILURE();
                $task->{message} ||= "The most recent updates are already installed.";
            }
        }
    };
    my $err = $@;

    if ($remove_file) {
        unlink($filename);
    }

    die $err if $err;
}

#----------------------------------------------------------------------

sub execute_update_signatures_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::execute_task(self => $self,
                                                          rpc_params => \@rpc_params,
                                                          entity_class => $entity_class,
                                                          task_type => 'update_signatures',
                                                          execute_func => sub { update_signatures(@_); },
                                                          );
}

#----------------------------------------------------------------------

sub delete_update_signatures_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                        );
}

#----------------------------------------------------------------------

sub delete_many_update_signatures_tasks {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_many_tasks(self => $self,
                                                               rpc_params => \@rpc_params,
                                                               entity_class => $entity_class,
                                                              );
}

#----------------------------------------------------------------------

1;
