package F5::ASMConfig::Handler::XmlProfile;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::ASMConfig;
use F5::DbUtils;
use F5::XMLProfile;

our @handled_functions = (
                           {
                             function_name => 'add_xml_profile',
                             relayable => 1,
                             store_index => 3,
                           },
                           {
                             function_name => 'update_xml_profile',
                             relayable => 1,
                             criteria_index => 3,
                             store_index    => 4,
                           },        
                           {
                             function_name => 'delete_xml_profile',
                             relayable => 1,
                           },
                           {
                             function_name => 'handle_validation_files',
                           },
                           {
                             function_name => 'update_many_xml_profiles',
                             relayable => 1,
                             criteria_index => 3,
                             store_index    => 4,
                           },
                         );

our $entity_class = 'F5::ASMConfig::Entity::XmlProfile';
#---------------------------------------------------------

sub add_xml_profile{
  my($self, $session, $call_flags, $policy,$xml_profile) = @_;

  my $run_validation = defined($xml_profile->{validation_files}) && @{$xml_profile->{validation_files}};

  my $rv = $self->base_add(entity_class => $entity_class,
                           rpc_params => [ $session, $call_flags,$policy,$xml_profile ]);
  my $profile = $rv->{return_value};
  if ($run_validation) {
       _handle_validation_files( self => $self,
                                 profile => $profile,
                                 call_flags => $call_flags,
                                 policy => $policy,
                                 action => 'add_xml_profile',
                                );
  }
  return $rv;
}

#---------------------------------------------------------

sub update_xml_profile{
  my($self, $session, $call_flags, $policy,$xml_profile_criteria, $store_xml_profile) = @_;

  
  my $run_validation;
  if (defined($store_xml_profile->{validation_files})) {
      my $profile =  $entity_class->new(dbh => $self->{dbh},
                                        get_criteria => $xml_profile_criteria,
                                        call_flags => $call_flags);
      if ($profile) {
          $run_validation = $profile->has_value_changed(existing_obj => $profile,
                                                        store_obj => $store_xml_profile,
                                                        check_fields => ['validation_files', 'flg_follow_schema_links']);
      }
      else {
          $run_validation = 1;
      }
  }
  my $rv = $self->base_update(entity_class => $entity_class,
                              rpc_params => [ $session, $call_flags,$policy,$xml_profile_criteria, $store_xml_profile ]);

  my $profile = $rv->{return_value};
  if ($run_validation) {
      _handle_validation_files( self => $self,
                                profile => $profile,
                                call_flags => $call_flags,
                                policy => $policy,
                                action => 'update_xml_profile',
                               );
  }
  return $rv;
}


#---------------------------------------------------------

sub delete_xml_profile{
 my($self, $session, $call_flags, $policy,$xml_profile) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};

  my $rv = $self->base_delete(entity_class => $entity_class,
                              rpc_params => [ $session, $call_flags,$policy,$xml_profile ],
                            );

  F5::XMLProfile::clean_unreferenced_profile_files(dbh => $dbh,
                                                   logger => $logger);
  return $rv;
}
#---------------------------------------------------------

sub handle_validation_files {
    my($self, $session, $call_flags, $policy,$xml_profile, $validation_files_flags) = @_;
    my $dbh = $self->{dbh};
    my $logger = $self->{logger};
    my $rv =  _handle_validation_files( self => $self,
                                        profile => $xml_profile,
                                        policy => $policy,
                                        action => '',
                                        validation_files_flags => $validation_files_flags );
  
    my $res = F5::ASMConfig::ValidationFilesRes->new($rv);  
    my $retval = { return_value => $res };
    return $retval;  
}

#---------------------------------------------------------

sub _handle_validation_files{
  my (%params) = @_;

  my $self = $params{self};
  my $profile = $params{profile};
  my $action = $params{action};
  my $call_flags = $params{call_flags};
  my $policy = $params{policy};

  if ($call_flags->{rest_call} && ! $call_flags->{async_rest_call}) {
      return $self->handle_error(action => $action,
                                 entity_class => $entity_class,
                                 error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                 thrift_obj   => $profile,
                                 error_thrown => 'Can not set validationFiles directly. Please use the associate-xml-validation_files-to-xml-profile Task.',
                                 error_level  => 'info',
                                );
  }


  my $validation_files = $params{validation_files_flags} || F5::ASMConfig::ValidationFilesFlags->new();

  my %validation_files_flags;
  $validation_files_flags{profile_id} = $profile->{profile_id};
  $validation_files_flags{force_regenerate} = defined $validation_files->{force_regenerate} ?
                                                       $validation_files->{force_regenerate} : 1;
  $validation_files_flags{flg_follow_schema_links} = defined  $validation_files->{flg_follow_schema_links} ?
                                                       $validation_files->{flg_follow_schema_links} : $profile->{flg_follow_schema_links};
  $validation_files_flags{save_update} = defined  $validation_files->{save_update} ?
                                                    $validation_files->{save_update} : 1;

  my $xml_profile = F5::ASMConfig::Entity::XmlProfile->new(dbh => $self->{dbh}, object => $profile, policy => $policy);
  my $res = F5::XMLProfile::handle_validation_files(dbh => $self->{dbh},
                                                    logger => $self->{logger},
                                                    asm_config_handler => $self,
                                                    xml_profile => $xml_profile,
                                                    policy => $policy,
                                                    %validation_files_flags,
                                                    );
  if (!$res->{completed_successfully}) {
      my $error_thrown = "Validation failed.";
      my $error_code = F5::ASMConfig::ErrorCode::INVALID_XML_PROFILE();
      if ( defined $res->{missing_files} ) {
          if (!ref($res->{missing_files})) {
              $res->{missing_files} = [$res->{missing_files}];
          }
          $error_code = F5::ASMConfig::ErrorCode::XML_PROFILE_MISSING_FILES();
      }

      if ( defined $res->{messages} ) {
          $error_thrown .= "  " . (ref($res->{messages}) ? join (",", @{$res->{messages}}) : $res->{messages});
      }

      return $self->handle_error(action => $action,
                                 entity_class => $entity_class,
                                 error_code   => $error_code,
                                 thrift_obj   => $profile,
                                 error_thrown => $error_thrown,
                                 error_level  => 'info',
                                 additional_error_messages => $res->{missing_files},
                                );
  }
  return $res;
}
#---------------------------------------------------------

sub update_many_xml_profiles {
  my($self, $session, $call_flags, $policy, $xml_profile_criteria, $store_xml_profile) = @_;

  # cannot update validation files in many xml profiles - ever.
  # instead, clients need to update_xml_profile.
  if (defined($store_xml_profile->{validation_files})) {
      return $self->handle_error(action => 'update_many_xml_profiles',
                                 entity_class => $entity_class,
                                 error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                 error_thrown => 'Can not set validationFiles directly. Please use the associate-xml-validation_files-to-xml-profile Task.',
                                 error_level  => 'info',
                                );
  }

  return $self->base_update_many(entity_class => $entity_class,
                                 rpc_params => [ $session, $call_flags, $policy, $xml_profile_criteria, $store_xml_profile ],);
}
#---------------------------------------------------------

1;
