package F5::ASMReady;

use strict;
use warnings;

use F5::Deploy;
use F5::BigIpLog;
use F5::DateTimeUtils;
use F5::bss qw(READY_SEMA bss_release bss_is_released_no_block);
use File::Slurp;
use F5::BigipUtils;

use Fcntl qw/:flock/;           # import LOCK_* constants
use YAML::XS;

use constant {
    BSS_NAME => "asm",
};

# file indicating that ASM initialization complete
my $ASM_READY_FILE = F5::Deploy::map_path('/ts/recovery/asm_ready');

# same as in F5::ASMReady, F5::ProcessHandler, F5::BdAgent
my $BD_AGENT_CONFIG_END_LOCK_FILE = F5::Deploy::map_path('/ts/recovery/bd_agent_config_end.lock');

my $ASM_LOAD_COMPLETE_END_LOCK_FILE = F5::Deploy::map_path('/ts/recovery/asm_load_complete_end.lock');

my $ASM_UPGRADE_LOCK_FILE = F5::Deploy::map_path('/ts/recovery/asm_upgrade.lock');

our $BD_AGENT_RUN_DATA_FILENAME = F5::Deploy::map_path('/ts/recovery/bd_agent_run_data.yaml');
use constant BD_AGENT_GRACE => 1800;
use constant BD_AGENT_LAST_WRITE_GRACE => 60;
use constant BD_GRACE => 60;

our $lock_fh;

# --------------------------------------------
sub set_asm_ready {
    our $bss_release_once;
    if ( ! $bss_release_once ) {
        bss_release(BSS_NAME, READY_SEMA, 0) && die "error: bss_release(): $!";
        $bss_release_once++;
    }
    # create the file indicating that ASM started successfully
    open($lock_fh, ">", $ASM_READY_FILE );
    print $lock_fh $$;
    chmod(0644, $ASM_READY_FILE);
    flock($lock_fh, LOCK_EX);
}

# --------------------------------------------
sub is_asm_upgrade_lock {
    return YAML::XS::LoadFile($ASM_UPGRADE_LOCK_FILE) if (-f $ASM_UPGRADE_LOCK_FILE);
    return 0;
}

# --------------------------------------------
sub is_asm_ready {
    return 1 if F5::BigipUtils::is_docker();
    return bss_is_released_no_block(BSS_NAME, READY_SEMA);
}

# --------------------------------------------
sub wait_for_asm_to_be_ready{
   my %params = @_;

   my $waited = 0;
   my $iteration_sleep_time = $params{iterations} || 5;
   my $timeout = $params{timeout} || 180;

   while($waited < $timeout) {      
      if( is_asm_ready() ) {
        return 1;
      }
      $waited+= sleep($iteration_sleep_time);  
   }

   return 0;
}

# --------------------------------------------
sub remove_asm_ready_file {
    if ($lock_fh) {
        close($lock_fh);
    }
    unlink($ASM_READY_FILE) if -f $ASM_READY_FILE;
}

# --------------------------------------------
sub remove_asm_upgrade_lock {
    unlink $ASM_UPGRADE_LOCK_FILE if -f $ASM_UPGRADE_LOCK_FILE;
}

# --------------------------------------------
sub write_asm_upgrade_lock {
    File::Slurp::write_file( $ASM_UPGRADE_LOCK_FILE, YAML::XS::Dump { date => F5::DateTimeUtils::convert_epoch_to_datetime(time()), pid => $$ } );
}

# --------------------------------------------
sub remove_bd_agent_config_end_lock_file {
    unlink($BD_AGENT_CONFIG_END_LOCK_FILE) if -f $BD_AGENT_CONFIG_END_LOCK_FILE;
    unlink($BD_AGENT_RUN_DATA_FILENAME) if -f $BD_AGENT_RUN_DATA_FILENAME;
}

# --------------------------------------------
sub place_bd_agent_config_end_lock_file {
    open (my $outf, '>', $BD_AGENT_CONFIG_END_LOCK_FILE) or die "$!\n";
    print $outf "$$";
    close ($outf);
}

# --------------------------------------------
sub remove_asm_load_complete_end_lock_file {
    unlink($ASM_LOAD_COMPLETE_END_LOCK_FILE) if -f $ASM_LOAD_COMPLETE_END_LOCK_FILE;
}

# --------------------------------------------
sub write_asm_load_complete_end_lock_file {
    open (my $outf, '>', $ASM_LOAD_COMPLETE_END_LOCK_FILE) or die "$!\n";
    print $outf "$$";
    close ($outf);
}

# --------------------------------------------
sub exists_asm_load_complete_end_lock_file {
    return (-f $ASM_LOAD_COMPLETE_END_LOCK_FILE) ? 1 : 0 ;
}

# --------------------------------------------
# wait for bd_agent to finish pushing all config to bd before checking in on bd threads
# bd_agent will touch the file $BD_AGENT_CONFIG_END_LOCK_FILE when it's done
sub wait_for_bd_agent_config_end {
    my %params = @_;

   #TODO: Check if something is needed here for docker environment 
   return if F5::BigipUtils::is_docker();

    my $logger = $params{logger};
    my $timeout = BD_AGENT_GRACE;

    my $msg = 'Going to wait for bd_agent to finish pushing configuration to bd';
    $logger->info($msg) if $logger;
    F5::BigIpLog::bigip_ts_ts_info_msg_log($msg); # /var/log/asm

    while  ($timeout) {
        $timeout--;
        last if -f $BD_AGENT_CONFIG_END_LOCK_FILE;
        my $bd_agent_run_data =  get_bd_agent_run_data();
        if( $bd_agent_run_data && $bd_agent_run_data->{timestamp} ){
            if( F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => BD_AGENT_LAST_WRITE_GRACE,
                                                               start_time => $bd_agent_run_data->{timestamp}) == 0 ){
                $timeout = 0;
                my $msg = "Maximum interval since last configuration file sent exceeded (start time = $bd_agent_run_data->{timestamp})";
                $logger->info($msg) if $logger;
                F5::BigIpLog::bigip_ts_ts_info_msg_log($msg);
                last;
            }
        }
        sleep 1;
    }

    if ( !$timeout ) {
        $msg = 'Gave up waiting for bd_agent to finish pushing configuration to bd. Trying a normal asm start-up sequence anyway';
    }
    else {
        $msg = 'bd_agent has finished pushing configuration to bd';
    }
    $logger->info($msg) if $logger;
    F5::BigIpLog::bigip_ts_ts_info_msg_log($msg); # /var/log/asm
}

# --------------------------------------------
# wait for bd to start all threads
sub wait_for_bd_threads {
    my %params = @_;

    my $logger = $params{logger};
    my $nwd = $params{nwd};
    my $timeout = BD_GRACE;

    my $msg = 'Going to wait for bd to start all threads';
    $logger->info($msg) if $logger;
    F5::BigIpLog::bigip_ts_ts_info_msg_log($msg); # /var/log/asm

    while ($timeout) {
        $timeout--;
        last if $nwd->verify(proc => 'bd');
        sleep 1;
    }

    if ( !$timeout ) {
        $msg = 'Gave up waiting for bd to start all threads. Trying a normal asm start-up sequence anyway';
    }
    else {
        $msg = 'bd has started all threads';
    }
    $logger->info($msg) if $logger;
    F5::BigIpLog::bigip_ts_ts_info_msg_log($msg); # /var/log/asm
}

# --------------------------------------------

sub get_bd_agent_run_data {
    my $bd_agent_run_data = {};
    if (-f $BD_AGENT_RUN_DATA_FILENAME){
        $bd_agent_run_data = YAML::XS::LoadFile($BD_AGENT_RUN_DATA_FILENAME);
    }
    return $bd_agent_run_data;
}

#--------------------------------------------
sub update_bd_agent_run_data{
    my %params = @_;

    return YAML::XS::DumpFile($BD_AGENT_RUN_DATA_FILENAME, { timestamp => time() } );

}

#--------------------------------------------
1; 
