package F5::ASMStart;

use strict;
use warnings;

use F5::ProcessHandler;
use F5::PrepareFiles;
use F5::DbUtils;
use F5::RamCache;
use F5::ASMReady;
use F5::NwdUtils::Nwd;
use F5::BigIpLog;
use F5::BigipUtils;
use F5::Cluster::Utils;
use F5::Cfg;
use F5::Provisioning;
use F5::ASM::COAPI;

our $DEFAULT_ASMDATA_LARGE_PARTITION_SUPPORT = 'false';
our $QPROP_ASMDATA_LARGE_PARTITION_VARIABLE = 'tmos.asmdata_large_partition_supported';

our $ASMLOGD_CFG_FILE = '/etc/ts/tools/asmlogd.cfg';
our $PROTOBUF_CFG_SECTION = 'asmlogd';
our $PROTOBUF_LARGE_CFG_FLD = 'protobuf_file_total_max_size_large';
our $PROTOBUF_SMALL_CFG_FLD = 'protobuf_file_total_max_size_small';
our $PROTOBUF_ACTUAL_CFG_FLD = 'protobuf_file_total_max_size';

our $RESERVE_FOR_UPGRADE_LARGE_CFG_FLD = 'reserve_for_upgrade_large';
our $RESERVE_FOR_UPGRADE_SMALL_CFG_FLD = 'reserve_for_upgrade_small';

my $ASM_HIDDEN_FOLDER =  '/Common/asm-hidden/';

my @MODULES = qw(asm dos);

sub asm_start {
    my(%params) = @_;

    my $logger = $params{logger};

    F5::ASMReady::write_asm_load_complete_end_lock_file();
    F5::ASMReady::remove_asm_ready_file();

    my %modules_provisioned = ();
    for my $module (@MODULES) {
        # hold only the provisioned modules
        if (F5::Provisioning::is_module_enabled($module)) {
            $modules_provisioned{$module} = 1;
        }
    }

    die "None of " . join(', ',@MODULES) . "modules are provisioned"
        unless keys %modules_provisioned;

    if ($modules_provisioned{'dos'}) {
        my $obj = F5::ASM::COAPI::Object->new('folder');
        $obj->{name} = $ASM_HIDDEN_FOLDER;
        $obj->save();
    }

    # set signal handling as early as possible
    my $process_handler = F5::ProcessHandler->new(logger => $logger);
    set_signals( process_handler => $process_handler, logger => $logger);

    # set the request log protobuf files max size to 5 GB on a non-virtual platform
    # -------------------------------------------------------------
    my $is_asmdata_large_partition_supported = F5::Cluster::Utils::get_qprop_value(variable_name => $QPROP_ASMDATA_LARGE_PARTITION_VARIABLE);
    if (! $is_asmdata_large_partition_supported) {
        $logger->info("Could not read the value of [$QPROP_ASMDATA_LARGE_PARTITION_VARIABLE]");
        $is_asmdata_large_partition_supported = $DEFAULT_ASMDATA_LARGE_PARTITION_SUPPORT;
    }

    my $protobuf_file_size_cfg_field;
    my $reserve_for_upgrade_size_cfg_field;
    if ($is_asmdata_large_partition_supported eq 'true') {
        $protobuf_file_size_cfg_field = $PROTOBUF_LARGE_CFG_FLD;
        $reserve_for_upgrade_size_cfg_field = $RESERVE_FOR_UPGRADE_LARGE_CFG_FLD;
    }
    else {
        $protobuf_file_size_cfg_field = $PROTOBUF_SMALL_CFG_FLD;
        $reserve_for_upgrade_size_cfg_field = $RESERVE_FOR_UPGRADE_SMALL_CFG_FLD;
    }

    my $protobuf_file_size_cfg_value = F5::Cfg::cfg_get_config_item($ASMLOGD_CFG_FILE, $PROTOBUF_CFG_SECTION, $protobuf_file_size_cfg_field);
    my $reserve_for_upgrade_size_cfg_value = F5::Cfg::cfg_get_config_item($ASMLOGD_CFG_FILE, $PROTOBUF_CFG_SECTION, $reserve_for_upgrade_size_cfg_field);

    $protobuf_file_size_cfg_value = $protobuf_file_size_cfg_value - $reserve_for_upgrade_size_cfg_value;

    $logger->info("Setting $PROTOBUF_ACTUAL_CFG_FLD to: [$protobuf_file_size_cfg_value] KB");
    $logger->info("Reserved [$reserve_for_upgrade_size_cfg_value] KB for upgrade MySQL dumps");
    F5::Cfg::cfg_set_config_item($ASMLOGD_CFG_FILE, $PROTOBUF_CFG_SECTION, $PROTOBUF_ACTUAL_CFG_FLD, $protobuf_file_size_cfg_value);
    # -------------------------------------------------------------

    if (my $lock_info = F5::ASMReady::is_asm_upgrade_lock()) {
        my $err_msg = "An upgrade process, executed by PID '$lock_info->{pid}', was interrupted on '$lock_info->{date}'. It is very likely that ASM will start with a severe inconsistent internal state and critical errors";
        F5::BigIpLog::bigip_ts_ts_error_msg_log($err_msg); # /var/log/asm
    }

    my $msg = 'ASM starting';
    $logger->info($msg);
    F5::BigIpLog::bigip_ts_ts_info_msg_log($msg); # /var/log/asm

    my $dbh = F5::DbUtils::get_dbh();
    F5::RamCache::prepare_conf(dbh => $dbh);
    F5::PrepareFiles::init(dbh => $dbh);
    F5::PrepareFiles::PrepareSiteTree(site_id => 1, dbh => $dbh);
    F5::PrepareFiles::PrepareInternalConfigFiles(site_id => 1, dbh => $dbh);
    F5::PrepareFiles::PrepareCoreFiles(site_id => 1, dbh => $dbh);
    $dbh->disconnect; # to prevent handling within fork

    $process_handler->kill_process_group('all');
    F5::ASMReady::remove_bd_agent_config_end_lock_file();

    $process_handler->start_process_group_by_module(%modules_provisioned); # all required but mcp_bridge (tsconfd)

    my $nwd = F5::NwdUtils::Nwd->new(process_handler => $process_handler,
                                     logger => $logger);

    if ($modules_provisioned{'asm'} && !F5::BigipUtils::is_docker()) {
        # wait for bd_agent to finish pushing all config to bd before checking in on bd threads
        F5::ASMReady::wait_for_bd_agent_config_end(logger => $logger,
                                                  );

        # wait for bd to start all threads before set_asm_ready
        F5::ASMReady::wait_for_bd_threads(logger => $logger,
                                          nwd => $nwd,
                                         );
    }

    F5::ASMReady::set_asm_ready();

    # start mcp_bridge (tsconfd) after marking asm_ready
    #    so that initial load_complete would not get killed due to ASM being not ready
    $process_handler->start_process_group('mcp_bridge');

    $msg = 'ASM started successfully';
    $logger->info($msg);
    F5::BigIpLog::bigip_ts_ts_info_msg_log($msg); # /var/log/asm

    while (1) {
        sleep 1;
        $nwd->verify;
    }
}

#----------------------------------------------------------------------

sub set_signals {
    my(%params) = @_;

    my $process_handler = $params{process_handler};
    my $logger = $params{logger};

    my $sig_handler = sub {
        my ($signal) = @_;

        my $msg = "Caught signal SIG${signal}, Stopping ASM";
        $logger->info($msg);
        F5::BigIpLog::bigip_ts_ts_info_msg_log($msg); # /var/log/asm

        $process_handler->kill_process_group('all');

        $msg = 'ASM stopped';
        $logger->info($msg);
        F5::BigIpLog::bigip_ts_ts_info_msg_log($msg); # /var/log/asm

        exit(0);
    };

    $SIG{__WARN__} = sub { $logger->error(shift) };

    $SIG{PIPE} = 'IGNORE';

    # signals to catch
    $SIG{INT}  = $sig_handler;
    $SIG{HUP}  = $sig_handler;
    $SIG{TERM} = $sig_handler;
    $SIG{ABRT} = $sig_handler;
}

#----------------------------------------------------------------------

1;
