package F5::BdAgent;

use strict;
use warnings;

use Math::BigInt;
use File::Basename;
use Tie::RefHash;
use POSIX qw(mkfifo);
use Bit::Vector;
use Time::HiRes qw( ualarm gettimeofday tv_interval );
use List::Util;
use File::Slurp;

use F5::Deploy;
use F5::ASM::COAPI;
use F5::IO::Mtcl;
use F5::IO::Storable;
use F5::IO::Selector;
use F5::CfgConvert;
use F5::BdCfgTables;
use F5::BigIpLog;
use F5::DbUtils;
use F5::IpAddr;
use F5::Db::Single;
use F5::ASMReady;
use F5::DateTimeUtils;
use F5::Logger;
use F5::BigipUtils;
use F5::Utils::File;
use F5::Log::Job;

use constant {
    AGENT_MSG_CONFIG_END => 11,
    EV_TYPE_CONTENT => 1,

    MSG_QUEUE_MAX_SIZE => 500,
    MSG_QUEUE_HANDLE_PERIOD => 1.0, # floating seconds
    MSG_QUEUE_HANDLE_TIMEOUT => 500000, # microseconds

    DEBUG => 0,
    KEEP_BINARY_FILES => 0,
};

my $WAITING_FOR_ACK_TIME_LIMIT = 600; # 10 minutes
my $PULSE_MSG_INTERVAL = 30;
my $UPDATE_REQUEST_MAX_POLL_INTERVAL = 2;
my $UPDATE_REQUEST_CLEANING_AGE = 60*60; # 1 hour

my $TO_BD_PIPE_NAME = F5::Deploy::map_path('/var/ts/bd/ts_core_pipe');
my $FROM_BD_PIPE_NAME = F5::Deploy::map_path('/var/ts/bd/ts_agent_pipe');

# same as in F5::ASMReady, F5::ProcessHandler, F5::BdAgent
my $BD_AGENT_CONFIG_END_LOCK_FILE = F5::Deploy::map_path('/ts/recovery/bd_agent_config_end.lock');

my @core_msgs = qw(
                      CORE_MSG_ACK
                      CORE_MSG_ACK_MORE
                      CORE_MSG_ERROR
                      CORE_MSG_EVENT
                      CORE_MSG_LOG
                      CORE_MSG_INIT_READY
                 );
my %core_msgs = map { ($core_msgs[$_] => $_) } 0..$#core_msgs;

# event to handler method name (see 'BD event handler methods section' below)
my %EVENT_TYPE_TO_HANDLER = (
    'EV_TYPE_CORE_SYSTEM' => 'core_system_event_handler',
    'EV_TYPE_CONTENT' => 'core_content_event_handler',

    'EV_TYPE_CMD' => 'mng_cmd_event_handler',
    'EV_TYPE_INFO' => 'mng_info_event_handler',
    'EV_TYPE_SYSLOG' => 'mng_syslog_event_handler',
);

my %SYSLOG_SEVERITY = (
    0 => 'EMERG',
    1 => 'ALERT',
    2 => 'CRIT',
    3 => 'ERR',
    4 => 'WARNING',
    5 => 'NOTICE',
    6 => 'INFO',
);

# like in 'typedef enum _CORE_EVENT_CODE' from TrafficShield/common/include/core_events.h
my %CORE_EVENT_CODE_TO_MACRO = ( # event names in comment for convenience
    28 => 'TS_ERROR_MSG', # EV_CORE_ERROR
    29 => 'TS_ERROR_OUT_OF_MEMEORY', # EV_CORE_ERROR_OUT_OF_MEMEORY
    30 => 'TS_ERROR_DISK_FAILURE', # EV_CORE_ERROR_DISK_FAILURE
    31 => 'TS_ERROR_CARD_FAILURE', # EV_CORE_ERROR_CARD_FAILURE
    32 => 'TS_ERROR_SYS_FAILURE', # EV_CORE_ERROR_SYS_FAILURE
    33 => 'TS_ERROR_NETWORK_FAILURE', # EV_CORE_ERROR_NETWORK_FAILURE
    34 => 'TS_ERROR_SSL', # EV_CORE_ERROR_SSL
    35 => 'TS_ERROR_FULL_QUEUE', # EV_CORE_ERROR_FULL_QUEUE
    36 => 'TS_MANAGEMENT_COMMUNICATION_ERROR', # EV_CORE_MANAGEMENT_COMMUNICATION_ERROR
    38 => 'TS_CONFIGURATION_ERROR', # EV_CORE_CONFIGURATION_ERROR
    41 => 'TS_BAD_REQUEST', # EV_CORE_BAD_REQUEST
    42 => 'TS_BAD_RESPONSE', # EV_CORE_BAD_RESPONSE
    51 => 'TS_BD_FILTERED_HTML_EXCEEDS_MAX_LENGTH', # EV_CORE_BD_FILTERED_HTML_EXCEEDS_MAX_LENGTH
    52 => 'TS_BD_APPLICATION_ERROR_FILTERING_APPLIED', # EV_CORE_BD_APPLICATION_ERROR_FILTERING_APPLIED
    53 => 'TS_CONFIGURATION_ERROR', # EV_CORE_ILLEGAL_REG_EXP
    54 => 'TS_ERROR_HIGH_CPU', # EV_CORE_HIGH_CPU_UTILIZATION
    55 => 'TS_ERROR_SESSION_DB_OUT_OF_BOUNDS', # EV_CORE_SESSION_DB_OUT_OF_BOUNDS
    56 => 'TS_INFO_MSG', # EV_CORE_MISSING_ACCOUNT_CONFIG,
);

# like in 'typedef enum _CORE_EVENT_CODE' from TrafficShield/common/include/core_events.h
my %CORE_EVENT_CODE_TO_METHOD = ( # event names in comment for convenience
    56 => 'missing_account_config', # EV_CORE_MISSING_ACCOUNT_CONFIG
);

# like in 'typedef enum _EVENT_CODE' from TrafficShield/common/include/mng_events.h
my %MNG_EVENT_CODE_TO_MACRO = ( # event names in comment for convenience
    1001  => 'TS_ERROR_UNIT_STATE_CHANGED', # EV_UNIT_STATE_CHANGED,
    1003  => 'TS_INFO_MSG', # EV_UNIT_STARTED,
    1004  => 'TS_ERROR_UNIT_SHUTDOWN', # EV_UNIT_SHUTDOWN,
    1006  => 'TS_ERROR_WEBSERVER_FAILURE', # EV_WEBSERVER_FAILURE,
    1008  => 'TS_ERROR_BD_FAILURE', # EV_BD_FAILURE,
    1014  => 'TS_INFO_MSG', # EV_PROCESS_FAILURE,
    1030  => 'TS_ERROR_LEARNING_FAILED', # EV_LEARNING_REQUEST_FAILED,
);

# like in 'typedef enum _DCC_MSG' from TrafficShield/common/include/dcc_msg.h
my %DCC_EVENT_CODE_TO_MACRO = ( # event names in comment for convenience
    2 => 'TS_ERROR_UPDATE_CONFIG', # DCC_MSG_UPDATE_CONFIG
);

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = bless {}, $class;
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    $self->{single_run} = $params{single_run};
    $self->init_db(%params);

    $self->{f5_big_ip_log} = F5::BigIpLog->new;
    $self->{logger} = $params{logger};
    $self->{cfg_logger} = F5::Logger->get_logger(logger_name => 'bd_agent_cfg_startup');
    $self->{selector} = F5::IO::Selector->new;
    $self->{update_request_poll_dir} = $params{update_request_poll_dir};
    $self->{initial_full_dir} = $params{initial_full_dir} || $self->find_latest_full_dir(base_dir => F5::Deploy::map_path('/ts/files/site_1/config'));
    $self->{logger}->info("Initial load directory = $self->{initial_full_dir}") if $self->{logger};

    if ($self->{update_request_poll_dir}) {
        # Ignore all updates up to (and including) the current initial load directory
        my @updates = sort glob($self->{update_request_poll_dir} . "/*");
        $self->{last_handled_update_request} = $updates[-1];
        $self->{logger}->info("Ignoring all updates until $self->{last_handled_update_request}") if $self->{logger};
    }

    if (!$self->{single_run}) {
        $self->init_core_channel(%params);
        $self->init_config_channel(%params);
    }

    $self->{update_core_count} = 0;

    $self->init_updates(%params);
}

#----------------------------------------------------------------------

sub find_latest_full_dir {
    my($self, %params) = @_;
    my $base_dir = $params{base_dir};

    opendir(my $dirh, $base_dir);
    my @previous_dirs = sort { $a <=> $b } grep { /^\d+$/} readdir($dirh);
    closedir($dirh);

    if (@previous_dirs) {
        return "$base_dir/$previous_dirs[-1]/full";
    }
    return;
}

#----------------------------------------------------------------------

sub init_db {
    my($self, %params) = @_;

    $self->{dbh} = $params{dbh};
    my $dbh = $self->{dbh} || F5::Db::Single->instance;

    my $ary_ref = $dbh->selectcol_arrayref("SELECT viol_index, description FROM PLC.VIOLATIONS", { Columns=>[1, 2] });
    $self->{sub_viol_descriptions} = {@{$ary_ref}};

    $ary_ref = $dbh->selectcol_arrayref("SELECT http_check_code, description FROM PLC.HTTP_CHECK_CODES", { Columns=>[1, 2] });
    $self->{sub_viol_http_descriptions} = {@{$ary_ref}};

    $ary_ref = $dbh->selectcol_arrayref("SELECT field_name, current_value FROM DCC.INTERNAL_CONFIG_ITEMS", { Columns=>[1, 2] });
    $self->{internal_config_items} = {@{$ary_ref}};

    $ary_ref = $dbh->selectcol_arrayref("SELECT subviolation_code, subviolation_description FROM PLC.XML_WS_SUBVIOLATIONS_CODES", { Columns=>[1, 2] });
    $self->{sub_viol_wss_descriptions} = {@{$ary_ref}};

    $ary_ref = $dbh->selectcol_arrayref("SELECT normalization_code, description FROM PLC.NORMALIZATION_CODES", { Columns=>[1, 2] });
    $self->{sub_viol_norm_descriptions} = {@{$ary_ref}};

    # When ASM starts (bd_agent) then we set the end_time to the current time
    $self->update_row(table_name => 'DCC.NEW_EVENT_SYSTEM',
                      rec => { },
                      set_date_fields_to_now_unix => [ 'l_time' ],
                      where_fields => { l_time => 0 }
                     );
}

#----------------------------------------------------------------------

sub init_core_channel {
    my($self, %params) = @_;

    $self->{msg_queue} = [];

    mkfifo($FROM_BD_PIPE_NAME, 0777);
    mkfifo($TO_BD_PIPE_NAME, 0777);

    open(my $from_pipe_fh, '<', $FROM_BD_PIPE_NAME);
    $self->{from_pipe} = F5::IO::Mtcl->new(socket => $from_pipe_fh,
                                           logger => $self->{logger},
                                           read_message_handler => sub {
                                               $self->handle_bd_pipe_message(message => $_[1]);
                                           },
                                           selector => $self->{selector},
                                          );

    open(my $to_pipe_fh, '>', $TO_BD_PIPE_NAME);
    $self->{to_pipe} = F5::IO::Mtcl->new(socket => $to_pipe_fh,
                                         logger => $self->{logger});


    $self->{is_core_ready} = 0;
    $self->{startup_complete} = 0;
    $self->{next_ack_id} = 1;
    $self->{waiting_for_ack} = '';
    $self->{waiting_for_ack_time} = 0;

    $self->set_records_count();

    # these counters get reset only on init
    $self->{bd_table_ack_count} = 0;
    $self->{bd_other_message_count} = 0;
    $self->{added_records_sent_to_bd_overall} = 0;
    $self->{updated_records_sent_to_bd_overall} = 0;
    $self->{deleted_records_sent_to_bd_overall} = 0;
}

#----------------------------------------------------------------------

sub init_config_channel {
    my($self, %params) = @_;

    $self->{config_listener} = F5::IO::Storable->new(server => 1,
                                                     local_addr => '127.0.0.1',
                                                     port => 9786,
                                                     read_message_handler => sub {
                                                         $self->handle_config_message(connection => $_[0], message => $_[1]);
                                                     },
                                                     selector => $self->{selector},
                                                    );
}

#----------------------------------------------------------------------

sub init_updates {
    my($self, %params) = @_;

    $self->{next_updates} = [
                             {
                              files => [ $self->get_files(base_dir => $self->{initial_full_dir}),
                                       ],
                              after_complete => sub {
                                  $self->send_to_core(msg_code => AGENT_MSG_CONFIG_END);
                                  $self->{startup_complete} = 1;
                                  $self->{cfg_logger} = F5::Logger->get_logger(logger_name => 'bd_agent_cfg_ongoing');
                                  F5::ASMReady::place_bd_agent_config_end_lock_file();
                              },
                             }
                            ];
}

#----------------------------------------------------------------------

sub handle_config_message {
    my($self, %params) = @_;

    my $message = $params{message};
    my $connection = $params{connection};

    $self->{update_core_count}++;

    if ($message->{update_dir}) {
        my @files = sort glob("$message->{update_dir}/*");
        my $dcc_table_config = F5::BdCfgTables::get_dcc_table_config();

        my @update_files;
        for my $file (@files) {
            my $table_output_file = basename($file);
            $table_output_file =~ s/\.[^.]+$//;
            $table_output_file =~ s/__account_(\d+)$//;
            $table_output_file =~ s/^.*__//;
            my($table_def) = grep { $table_output_file eq $_->{output_file} } @$dcc_table_config;

            if (! $self->corresponding_protobuf_file_exists($file)) {
                push @update_files, $file;
            }
        }
        if (@update_files) {
            my %job_attributes;
            if ($message->{job}) {
                %job_attributes = %{$message->{job}};
            }
            $job_attributes{logger} = $self->{logger};
            $job_attributes{component} = 'bd_agent';
            my $job = F5::Log::Job->new(%job_attributes);
            $job->set_pending();

            push(@{$self->{next_updates}},
                 {connection => $connection,
                  files => \@update_files,
                  job => $job,
                 });
        }
        else {
            $connection->write_message({completed_successfully => 1,
                                        complete_datetime => time,
                                       });
        }

        $self->{logger}->debug("Received update command") if $self->{logger};
    }
}

#----------------------------------------------------------------------

sub start {
    my($self, %params) = @_;

    $self->{logger}->info('Starting bd_agent') if $self->{logger};

    my $last_pulse_time = time;

    while (1) {
        if ($self->{single_run} || ($self->{is_core_ready} && ! $self->{waiting_for_ack})) {
            if (@{$self->{next_updates}}) {
                $self->send_next_file;
            }
        }

        last if $self->{single_run} && !@{$self->{next_updates}};

        if ($self->{update_request_poll_dir}) {
            $self->check_for_update_requests();
        }

        if (!$self->{single_run}) {
            $self->handle_messages(last_pulse_time => $last_pulse_time);
        }

        if ($self->{waiting_for_ack_time} &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $WAITING_FOR_ACK_TIME_LIMIT,
                                                               start_time => $self->{waiting_for_ack_time}) == 0) {

            my $err_msg = "No ack received from BD after " . (time - $self->{waiting_for_ack_time}) . " seconds -- aborting";
            $self->{logger}->error($err_msg) if $self->{logger};
            die $err_msg;
        }

        if ($PULSE_MSG_INTERVAL &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $PULSE_MSG_INTERVAL,
                                                               start_time => $last_pulse_time) == 0) {

            $self->log_pulse_message;
            $last_pulse_time = time;
        }
    }
}

#----------------------------------------------------------------------

sub handle_messages {
    my($self, %params) = @_;
    my $poll_time = $self->get_poll_time(last_pulse_time => $params{last_pulse_time});
    $self->{logger}->trace("About to poll " . (defined $poll_time ? "for $poll_time seconds" : "until message received")) if $self->{logger};
    my @readers = $self->{selector}->can_read($poll_time);
    for my $reader (@readers) {
        $reader->handle_read;
    }
}

#----------------------------------------------------------------------

sub check_for_update_requests {
    my($self, %params) = @_;

    $self->{logger}->trace("Checking for update requests") if $self->{logger};
    my @updates = sort glob($self->{update_request_poll_dir} . "/*");
    my $update_request_filename = List::Util::first { $_ gt $self->{last_handled_update_request}} @updates;
    if ($update_request_filename) {
        $self->{logger}->info("Found new update (file = $update_request_filename)") if $self->{logger};
        my $update_request = YAML::XS::LoadFile($update_request_filename);
        $self->handle_config_message(message => $update_request);
        $self->{last_handled_update_request} = $update_request_filename;

        F5::Utils::File::cleanup_files(logger => $self->{logger},
                                       dir => $self->{update_request_poll_dir},
                                       age => $UPDATE_REQUEST_CLEANING_AGE,
                                      );
    }
}

#----------------------------------------------------------------------

sub get_poll_time {
    my($self, %params) = @_;
    my @potential_poll_times;
    my $last_pulse_time = $params{last_pulse_time};

    if ($self->{waiting_for_ack_time}) {
        push(@potential_poll_times,
             F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $WAITING_FOR_ACK_TIME_LIMIT,
                                                                start_time => $self->{waiting_for_ack_time}));
    }

    if ($PULSE_MSG_INTERVAL) {
        push(@potential_poll_times,
             F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $PULSE_MSG_INTERVAL,
                                                                start_time => $last_pulse_time));
    }

    if ($self->{update_request_poll_dir}) {
        push(@potential_poll_times, $UPDATE_REQUEST_MAX_POLL_INTERVAL);
    }

    return (@potential_poll_times) ? List::Util::min(@potential_poll_times) : undef;
}

#----------------------------------------------------------------------

sub log_pulse_message {
    my($self) = @_;
    my @additional_info;
    if ($self->{bd_table_ack_count}) {
        push(@additional_info, "$self->{bd_table_ack_count} BD acks");
    }

    if ($self->{bd_other_message_count}) {
        push(@additional_info, "$self->{bd_other_message_count} BD other messages");
    }

    for my $type (qw(added updated deleted)) {
        my $count = $self->{$type . '_records_sent_to_bd_overall'};
        push(@additional_info, "$count records $type") if defined $count;
    }

    my $additional_info = '';
    if (@additional_info) {
        $additional_info = '(' . join(', ', @additional_info) . ')';
    }
    $self->{logger}->info("$self->{update_core_count} update cores handled since startup $additional_info") if $self->{logger};
}

#----------------------------------------------------------------------

sub handle_bd_pipe_message {
    my($self, %params) = @_;

    my $message = $params{message};

    if ( DEBUG && $self->{logger} && $self->{logger}->is_trace()) {
        $self->{logger}->trace($message);
    }

    if (! defined $message->{msg_code}) {
        return;
    }

    if ($message->{msg_code} == $core_msgs{CORE_MSG_INIT_READY}) {
        $self->{is_core_ready} = 1;
        $self->{logger}->info("bd is ready") if $self->{logger};
    }
    elsif ($message->{msg_code} == $core_msgs{CORE_MSG_ACK}) {
        $self->{bd_table_ack_count}++;

        if ($message->{ack_id} eq $self->{waiting_for_ack}) {
            $self->{logger}->debug("Received ack $message->{ack_id} successfully") if $self->{logger};

            my $config_error;
            # overall lines sent to bd vs overall lines received in bd
            if (! defined($message->{records_received_by_bd})) {
                $config_error = "Could not determine the overall number of records received by enforcer";
            }
            elsif (! defined($self->{records_sent_to_bd_in_last_update})) {
                $config_error = "Could not determine the overall number of records sent to enforcer";
            }
            elsif ($self->{records_sent_to_bd_in_last_update} != $message->{records_received_by_bd}) {
                $config_error = "Some records sent to enforcer were not received (sent = $self->{records_sent_to_bd_in_last_update}, received = $message->{records_received_by_bd})";
            }

            if (!$config_error) {
                # by record type: sent to bd vs handled by bd
                for my $operation (qw(added updated deleted)) {
                    my $sent_field = $operation . '_records_sent_to_bd_in_last_update';
                    my $handled_field = 'records_' . $operation . '_by_bd';

                    if (!defined($message->{$handled_field})) {
                        $config_error = "Could not determine the number of records handled by enforcer (operation = $operation)";
                    }
                    elsif (!defined($self->{$sent_field})) {
                        $config_error = "Could not determine the number of records sent to enforcer (operation = $operation)";
                    }
                    elsif ($self->{$sent_field} != $message->{$handled_field}) {
                        $config_error = "Some records sent to enforcer were not handled (operation = $operation, sent = $self->{$sent_field}, handled = $message->{$handled_field})";
                    }
                }
            }

            if ($config_error) {
                my $log_error_message = "$config_error, ack ID = $message->{ack_id}";
                $self->{logger}->error($log_error_message) if $self->{logger};
                F5::BigIpLog::bigip_ts_ts_error_msg_log($log_error_message);
                $self->log_config_info(message => $message);
                if ($self->{internal_config_items}->{restart_upon_config_update_error}) {
                    my $err_msg = "Exiting due to a config failure in ack ID = $message->{ack_id}";
                    $self->{logger}->error($err_msg) if $self->{logger};
                    die $err_msg;
                }
            }

            my $this_update = $self->{next_updates}->[0];
            if (! @{$this_update->{files}}) {
                $self->{logger}->info("Update completed successfully") if $self->{logger};
                if ($this_update->{connection}) {
                    $this_update->{connection}->write_message({completed_successfully => 1,
                                                               complete_datetime => time});

                }

                if ($this_update->{after_complete}) {
                    $this_update->{after_complete}->();
                }

                $this_update->{job}->set_completed() if $this_update->{job};
                shift(@{$self->{next_updates}});
            }
        }
        else {
            my $err_msg = "Received some other ack! ($message->{ack_id}) -- aborting";
            $self->{logger}->error($err_msg) if $self->{logger};
            die $err_msg;
        }
        $self->{waiting_for_ack} = '';
        $self->{waiting_for_ack_time} = 0;
        $self->set_records_count();
    }
    else {
        # this happens first time in only
        $self->{start_time} = [gettimeofday()] unless $self->{start_time};

        $self->{bd_other_message_count}++;

        my $event_code = $message->get_event_code();
        $self->{logger}->debug("Got event code:[$event_code]") if $self->{logger};

        if ( $event_code == EV_TYPE_CONTENT ) {
            # oldest first
            push(@{$self->{msg_queue}},$message);

            # if queue reached MSG_QUEUE_MAX_SIZE -
            #    throw away 1 oldest messages of the head of the queue
            if ( scalar @{$self->{msg_queue}} > MSG_QUEUE_MAX_SIZE ) {
                 shift(@{$self->{msg_queue}});
            }

            if ( tv_interval($self->{start_time}) >= MSG_QUEUE_HANDLE_PERIOD ) {
                eval {
                    local $SIG{ALRM} = sub { die; };
                    ualarm(MSG_QUEUE_HANDLE_TIMEOUT);

                    # oldest handled first
                    while (my $queued_message = shift(@{$self->{msg_queue}}) ) {
                        $self->handle_message(event_code => $event_code,
                                              message => $queued_message,
                                             );
                    }

                    ualarm(0);
                };
                # reset last time stamp once - immediately after the eval above
                $self->{start_time} = [gettimeofday()];
            }
        }
        else {
            $self->handle_message(event_code => $event_code,
                                  message => $message,
                                 );
        }
    }
}

#----------------------------------------------------------------------

sub log_config_info {
    my($self, %params) = @_;
    my $message = $params{message};

    for my $operation (qw(added updated deleted)) {
        my $num = $self->{$operation . '_records_sent_to_bd_in_last_update'};
        $self->{logger}->info("Number of records sent to be $operation by agent = $num") if $self->{logger};
    }

    for my $operation (qw(added updated deleted received)) {
        my $num = $message->{'records_' . $operation . '_by_bd'};
        $self->{logger}->info("Number of records $operation by bd = $num") if $self->{logger};
    }

    $self->{logger}->trace($message) if $self->{logger};
}

#----------------------------------------------------------------------

sub handle_message {
    my($self, %params) = @_;
    my $event_code = $params{event_code} ;
    my $message = $params{message} ;

    my $event;
    eval {
        $event = $message->get_event();
    };
    if ($@) {
        # every time the ualarm above fires, for EV_TYPE_CONTENT messages, we'll get here
        #     thus, we lower log level to debug for EV_TYPE_CONTENT messages
        my $method = ( $event_code == EV_TYPE_CONTENT ) ? 'debug' : 'error' ;
        $self->{logger}->$method("Event parsing failed for event code:[$event_code]") if $self->{logger};
        if ( DEBUG ) {
            $self->{logger}->trace($message) if $self->{logger};
        }
    }
    else {
        my $handler_method = $EVENT_TYPE_TO_HANDLER{$event->{bd_event}};
        $self->{logger}->debug("Event parsing succeeded for event code:[$event_code]. Chosen method is:[$handler_method].") if $self->{logger};

        if ( DEBUG ) {
            $self->dump_event( handler => $handler_method, event => $event);
        }

        # this will fire one of the BD event handler methods below
        $self->$handler_method($event->{$event->{bd_event}}, $event->{COMMON_EVENT_TYPE});
    }
}

#----------------------------------------------------------------------

sub send_next_file {
    my($self) = @_;

    my $msg_data;
    my $msg_code;
    my $this_update = $self->{next_updates}->[0];
    my $next_file = shift(@{$this_update->{files}});
    return unless $next_file;

    F5::ASMReady::update_bd_agent_run_data();
    $this_update->{job}->set_started() if $this_update->{job};

    my %metadata;

    if (ref($next_file)) {
        #this case for iterator stored in queue that should return data
        my $config_data;
        my $last_iteration;
        ($config_data, $last_iteration, %metadata) = @{$next_file->()};
        $msg_data = $config_data;
        $msg_code = 1;
        if (! $last_iteration) {
            unshift @{$this_update->{files}}, $next_file;
        }
    }
    elsif (basename($next_file) eq 'bd.cfg') {
        open(my $inf, '<', $next_file);
        my $contents;
        sysread($inf, $contents, -s $next_file);
        close($inf);
        $msg_data = $contents;
        $msg_code = 0;
    }
    else {
        my %cfg_params = (from_file => $next_file,
                          return_as_string => 1);

        $self->{last_bd_filename} = $next_file;
        $self->{last_bd_file_count} = 1;

        $msg_code = 1;
        $self->{logger}->debug("About to convert $next_file") if $self->{logger};
        my $data;
        ($data, %metadata) = F5::CfgConvert::convert_table(%cfg_params,
                                                           logger => $self->{logger},
                                                           cfg_logger => $self->{cfg_logger});
        if (ref($data)) {
            my $config_data;
            my $last_iteration;
            ($config_data, $last_iteration, %metadata) = @{$data->()};
            $msg_data = $config_data;
            if (! $last_iteration) {
                unshift @{$this_update->{files}}, $data;
            }
        }
        else {
            $msg_data = $data;
        }
    }

    my %core_update_params = (
                              data => $msg_data || '',
                              msg_code => $msg_code,
                              added_records => $metadata{added_records} || 0,
                              updated_records => $metadata{updated_records} || 0,
                              deleted_records => $metadata{deleted_records} || 0,
                             );

    if (KEEP_BINARY_FILES || $self->{single_run}) {
        $self->save_update_file(%core_update_params);
    }

    #send_to_core should be called even if sending data is empty
    #otherwise blocking process and BD get SIGTERM after timeout
    $self->send_to_core(%core_update_params,
                        msg_type => $self->{startup_complete},
                       );

    $self->{logger}->info("Sent file $next_file") if $self->{logger};

    if ($self->{single_run} && !@{$this_update->{files}}) {
        $this_update->{job}->set_completed() if $this_update->{job};
        shift(@{$self->{next_updates}});
    }

    if ($self->{logger} && $self->{logger}->is_trace()) {
        if (! ref($next_file) ) {
            my $file_text = File::Slurp::read_file($next_file);
            $self->{logger}->trace("Sent file text was '$file_text'");
        }
        else {
            $self->{logger}->trace("Sent file was a ref") if $self->{logger};
        }
    }
}

#----------------------------------------------------------------------

sub save_update_file {
    my($self, %params) = @_;

    if (! ($self->{last_bd_filename} && $self->{last_bd_file_count})) {
        return;
    }

    my $filename = sprintf('%s__%.4d', $self->{last_bd_filename}, $self->{last_bd_file_count});
    File::Slurp::write_file("$filename.bin", $params{data});
    $self->{last_bd_file_count}++;
}

#----------------------------------------------------------------------

sub send_to_core {
    my($self, %params) = @_;
    return if $self->{single_run};

    %params = (msg_type => 0,
               data => '',
               msg_code => 0,
               %params);

    my $ack_id = $self->{next_ack_id};
    $self->{next_ack_id}++;

    $self->set_records_count(%params);

    my $message = F5::Mtcl->new(%params,
                                ack_id => $ack_id);
    if ( DEBUG && $self->{logger} && $self->{logger}->is_trace()) {
        $self->{logger}->trace(get_debug_message($message));
    }

    $self->{to_pipe}->write_message($message);
    $self->{logger}->debug("Sent message with ack_id $ack_id") if $self->{logger};
    $self->{waiting_for_ack} = $ack_id;
    $self->{waiting_for_ack_time} = time;
}

#----------------------------------------------------------------------

sub set_records_count {
    my($self, %params) = @_;

    $self->{records_sent_to_bd_in_last_update} = 0;
    for my $type (qw(added updated deleted)) {
        # these counters get reset on each update
        $self->{$type . '_records_sent_to_bd_in_last_update'} = $params{$type . '_records'} || 0;
        $self->{records_sent_to_bd_in_last_update} += $self->{$type . '_records_sent_to_bd_in_last_update'};

        # these are overall counters that are only reset on init
        $self->{$type . '_records_sent_to_bd_overall'} += $self->{$type . '_records_sent_to_bd_in_last_update'};
    }
}

#----------------------------------------------------------------------

sub get_files {
    my($self, %params) = @_;

    my $dcc_table_config = F5::BdCfgTables::get_dcc_table_config();
    my $base_dir = $params{base_dir};

    my @loadable_tables;
    for my $dcc_table (@$dcc_table_config) {
        my @current_table_files;
        push(@current_table_files, sort glob("$base_dir/$dcc_table->{output_file}.*"));
        push(@current_table_files, sort glob("$base_dir/$dcc_table->{output_file}__account*"));
        for my $current_table_file (@current_table_files) {
            if (! $self->corresponding_protobuf_file_exists($current_table_file)) {
                push(@loadable_tables, $current_table_file);
            }
        }
    }

    return @loadable_tables;
}

#----------------------------------------------------------------------

sub corresponding_protobuf_file_exists {
    my($self, $filename) = @_;

    if ($filename =~ /\.msgp$/) {
        my $corresponding_protobuf_file = $filename;
        $corresponding_protobuf_file =~ s/\.msgp$/.dat/;
        if (-e $corresponding_protobuf_file) {
            return 1;
        }
    }
    return;
}

#----------------------------------------------------------------------

sub get_debug_message {
    my($m) = @_;
    # unhandled messages - events go here
    my $message_copy = { %$m };
    if (length($message_copy->{data}) > 500) {
        $message_copy->{data_suffix} = substr($message_copy->{data}, -200);
        $message_copy->{data_prefix} = substr($message_copy->{data}, 0, 200);
        delete $message_copy->{data};
    }
    return $message_copy;
}

#----------------------------------------------------------------------

###########################################################
#    Start BD event handler methods section
###########################################################

sub missing_account_config {
    my($self, $event) = @_;

    my $account_name = $event->{desc};
    return unless $account_name;

    my $mcp_asm_policies = F5::ASM::COAPI::Query->new('asm_policy',{name => $account_name});
    return unless $mcp_asm_policies && scalar @$mcp_asm_policies;

    my $mcp_asm_policy = $mcp_asm_policies->[0];
    return unless exists $mcp_asm_policy->{is_active_policy};
    return if $mcp_asm_policy->{is_active_policy};

    $mcp_asm_policy->{is_active_policy} = 1;
    $mcp_asm_policy->save();

    return "bd_agent is asynchronously activating ASM security policy named '$account_name', per request from bd";
}

#----------------------------------------------------------------------

sub core_system_event_handler {
    my($self, $event, $common) = @_;

    my $desc = $event->{desc};

    my $method = $CORE_EVENT_CODE_TO_METHOD{$event->{code}} || 0;
    if ($method) {
        $desc = $self->$method($event);
    }

    if ($desc) {
        my $macro = $CORE_EVENT_CODE_TO_MACRO{$event->{code}} || 0;
        $self->big_ip_log($macro, $event->{code}, $desc);
    }
}

#----------------------------------------------------------------------

sub core_content_event_handler {
    my($self, $event, $common) = @_;

    my $severity = $SYSLOG_SEVERITY{$event->{severity_id}};
    my $status = ($event->{is_blocked}) ? 'blocked' : 'violation' ;
    my $asm_prefix = ($severity eq 'INFO') ? 'asm_' : '' ;
    my $macro = 'ts_' . $asm_prefix . 'request_' . $status . '_' . $severity;

    my $rejected_desc = $self->get_viol_desc($event->{viol_set_enforce}->as_bin());

    my $http_desc = $self->get_sub_viol_desc(sub_viol_value => $event->{non_rfc_sub_violations},
                                      db_map => $self->{sub_viol_http_descriptions},
                                     );

    my $wss_desc = $self->get_sub_viol_desc(sub_viol_value => $event->{wss_sub_violations},
                                     db_map => $self->{sub_viol_wss_descriptions},
                                    );

    my $et_desc = $self->get_sub_viol_desc(sub_viol_value => $event->{normalization_sub_violations},
                                    db_map => $self->{sub_viol_norm_descriptions},
                                   );

    my $str_app_id = ($event->{app_id}) ? 'HTTPS' : 'HTTP' ;


    my $src_ip = $self->ip_ntop($event->{source_ip_raw});
    my $dest_ip = $self->ip_ntop($event->{destination_ip_raw});
    my $xff_ip = $self->ip_ntop($event->{xff_ip_raw});

    # session_id shold be a lower-case hex string w/o the leading '0x'
    my $session_id = $event->{session_id}->as_hex(); # a signed hexadecimal string with prefixed 0x
    $session_id = lc(substr($session_id,2)); # lower-cased and '0x' stripped

    $self->big_ip_log($macro, undef,
        $rejected_desc,
        $http_desc,
        $wss_desc,
        $et_desc,
        $event->{virus_name} || "N/A",
        $event->{support_id}->bstr(),
        $src_ip,
        $xff_ip,
        $event->{client_remote_port},
        $dest_ip,
        $event->{client_local_port},
        $event->{rtdom_id},
        $event->{http_selector},
        $str_app_id,
        $event->{geo_location},
        $event->{first_byte},
        $event->{username} || "N/A",
        $session_id,
        $event->{violation_rate}
    );

    if ( DEBUG ) {
        $macro = 'ts_info_msg';
        $self->big_ip_log($macro, undef,
           'counter:['. $common->{counter} .'], '.
           'data_size:['.  $common->{data_size} .'], '.
           'event_id:['.  $common->{event_id}->bstr() .'], '.
           'event_type:['.  $common->{event_type} .'], '.
           'first_time:['.  $common->{first_time} .'], '.
           'last_time:['.  $common->{last_time} .'], '.
           'machine_id:['.  sprintf('%02X:%02X:%02X:%02X:%02X:%02X', @{$common->{machine_id}}) .'], '.
           'site_id:['.  $common->{site_id} .']'
        );
    }

}

#----------------------------------------------------------------------

sub mng_cmd_event_handler {
    my($self, $event, $common) = @_;

    my $rec = {
        f_time => $common->{first_time},
        l_time => $common->{last_time},
        counter => $common->{counter},
        mac_ID => sprintf('%02X:%02X:%02X:%02X:%02X:%02X', @{$common->{machine_id}}),
        event_crc => $common->{event_id}->bstr(),
        type => $common->{event_type},
        code => $event->{cmd_code},
        result_description => $event->{result_description},
    };

    $self->insert_row(table_name => 'DCC.NEW_EVENT_SYSTEM',
                      rec => $rec,
                     );

    my $macro = $DCC_EVENT_CODE_TO_MACRO{$event->{cmd_code}} || 0;
    $self->big_ip_log($macro, $event->{cmd_code}, $event->{result_description});
}

#----------------------------------------------------------------------

sub mng_info_event_handler {
    my($self, $event, $common) = @_;

    my $rec = {
        f_time => $common->{first_time},
        l_time => $common->{last_time},
        counter => $common->{counter},
        mac_ID => sprintf('%02X:%02X:%02X:%02X:%02X:%02X', @{$common->{machine_id}}),
        event_crc => $common->{event_id}->bstr(),
        type => $common->{event_type},
        code => $event->{code},
        state => $event->{state},
        result_description => $event->{desc},
    };

    $self->insert_row(table_name => 'DCC.NEW_EVENT_SYSTEM',
                      rec => $rec,
                     );

    my $macro = $MNG_EVENT_CODE_TO_MACRO{$event->{code}} || 0;
    $self->big_ip_log($macro, $event->{code}, $event->{desc});
}

#----------------------------------------------------------------------

sub mng_syslog_event_handler {
    my($self, $event, $common) = @_;

    my $rec = {
        f_time => $common->{first_time},
        l_time => $common->{last_time},
        counter => $common->{counter},
        mac_ID => sprintf('%02X:%02X:%02X:%02X:%02X:%02X', @{$common->{machine_id}}),
        event_crc => $common->{event_id}->bstr(),
        type => $common->{event_type},
        result_description => $event->{result_description},
    };

    $self->insert_row(table_name => 'DCC.NEW_EVENT_SYSTEM',
                      rec => $rec,
                     );
}


###########################################################
#    End BD event handler methods section
###########################################################

#----------------------------------------------------------------------

# MACRO is from TrafficShield/common/ts_bigip_log/bigip_ts_error_defs.in
# that MACRO, in lower case, is also an OO method in F5::BigIpLog
# so you can F5::BigIpLog->$method(ARGS) when $method = lc(MACRO);
# ARGS is an input per each '%' from the chosen MACRO - which is a C style printf modifier
# you can also view the prototype of $method here:
#     TrafficShield/mng/scripts/packages/xs_packages/BigIpLog/BigIpLog.xs
# there look for "bigip_ts_${method}_log".
# Note that BigIpLog.xs is auto-generated (you need to build tscommon and mng-scripts to get it)

sub big_ip_log {
    my ($self, $macro, $evt_code, @msg) = @_;
    $evt_code ||= 'none';

    if ( ! $macro ) {
        my $whowasi = (caller(1))[3];
        $self->{logger}->debug("Received unknown event number:[$evt_code] in method:[$whowasi].") if $self->{logger};
        return;
    }

    $macro = lc($macro);

    if ($self->{logger} && DEBUG ) {
        my $whowasi = (caller(1))[3];
        $self->{logger}->debug("Call details: macro:[$macro] caller:[$whowasi] event number:[$evt_code].");
        $self->{logger}->trace("ARGS to macro:[$macro] were:");
        $self->{logger}->trace(\@msg);
    }

    eval {
        $self->{f5_big_ip_log}->$macro(@msg);
    };
    if ($@) {
        $self->{logger}->debug('Error executing F5::BigIpLog->' . $macro . "(...) : [$@]") if $self->{logger};
    }
}

#----------------------------------------------------------------------
# called only when DEBUG is set to '1'

sub dump_event {
    my($self, %params) = @_;

    my $event = $params{event};
    my $handler = $params{handler};

    $self->{logger}->trace("The target handler is [$handler]. This is the event that I received:\n") if $self->{logger};
    $self->{logger}->trace($event) if $self->{logger};
}

#----------------------------------------------------------------------

sub get_sub_viol_desc {
    my($self, %params) = @_;

    my $sub_viol_value = $params{sub_viol_value};
    return "N/A" unless $sub_viol_value;

    my $db_map = $params{db_map};
    my $viol_vect = Bit::Vector->new_Dec(32, $sub_viol_value);

    my $desc = '';
    for my $bit ( $viol_vect->Index_List_Read() ) {
        $desc .= $db_map->{2 ** $bit} . ',';
    }
    chop($desc);

    return $desc || "N/A";
}

#----------------------------------------------------------------------

sub ip_ntop {
    my ($self, $ref) = @_;

    my $res = F5::IpAddr::ip_ntop(pack("L4", $ref->[0], $ref->[1], $ref->[2], $ref->[3]));

    return ( $res eq '::' ) ? "N/A" : $res ;
}

#----------------------------------------------------------------------

sub get_ip_binary_data {
    my ($self, $ref) = @_;

    return pack("L4", $ref->[0], $ref->[1], $ref->[2], $ref->[3]);
}

#----------------------------------------------------------------------

# tbd:sig
#     change bd_agent.h so that C::Include would unpack into a representation (char?)
#     that can be later used in F5::Violation::Utils::get_indexes_from_viol_set

sub get_viol_desc {
    my ($self, $bits) = @_;

    return "N/A" unless $bits; # will return on a bitstring that is all 0(s)

    my $rejected_desc;

    # $bits has a leading left '0b' notation, we remove it
    #    and left-pad it with '0's to have an even number of octets
    substr($bits, 0, 2, 0 x (8 - ((length($bits) - 2) % 8)));
    my $bits_len = length($bits);

    # traverse octets
    # each octet by itself is reversed - we reverse them back
    # the input is a C::Include->unpack product, not sure why the octets are reversed
    for (my $i=0 ; $i<$bits_len ; $i+=8) {
        my $octet = substr($bits, $i, 8);
        next if $octet eq '00000000'; # skip octet if no bits ON
        substr($bits, $i, 8, scalar reverse $octet);
    }

    # now, an index of each bit that it turned ON is a viol_index
    # Index_List_Read returns a list of indexes of bits that are turned ON in this octet (0-7)
    my $vec = Bit::Vector->new_Bin($bits_len, $bits);
    for my $bit ( $vec->Index_List_Read() ) {
        $rejected_desc .= $self->{sub_viol_descriptions}->{$bit} . ',';
    }

    chop($rejected_desc);
    return $rejected_desc || "N/A";
}

#----------------------------------------------------------------------

sub insert_row {
    my ($self, %params) = @_;
    my $dbh = $self->{dbh} || F5::Db::Single->instance;

    F5::DbUtils::insert_row(dbh => $dbh, %params);
}

#----------------------------------------------------------------------

sub update_row {
    my ($self, %params) = @_;
    my $dbh = $self->{dbh} || F5::Db::Single->instance;

    F5::DbUtils::update_row(dbh => $dbh, %params);
}

#----------------------------------------------------------------------

1;
