package F5::Cfg;

use strict;
use warnings;

use constant TRUE => 1;
use constant FALSE => 0;

# mysql password holder
use constant MYSQL_PASSWORD_FILE => '/var/db/mysqlpw';

use F5::Deploy;

#--------------------------------------------------------------------------------
#gets 
#   cfg_file (path/file) - file to search
#   section (may be undef) - section under search, if got undef searchs the hole file!
#   item - item to search
#returns
#   value - the item value (if not found - undef)  
sub cfg_get_config_item
{
    my ($cfg_file, $section, $item) = @_;
    my $lines = get_cfg_lines($cfg_file);
    if (not defined $lines)
    {
	print "F5::Cfg::cfg_get_config_item - Can't get lines from $cfg_file\n";
	return;
    }

    for (my $i = 0; $i < @$lines; $i++)
    {
	$_ = $lines->[$i];
	# no section case
	if(not defined $section){
            if(/^\s*$item\s*=\s*(.+?)\s*$/i)
            {
               return $1;
            }
	}       
	# with section
	else
	{    
            if(/^\s*\[\s*$section\s*\]/i)
	    {
		# found the section, now find the item, starting from the next line
		for (my $j = $i + 1; $j < @$lines; $j++)
		{
		    $_ = $lines->[$j];
		    return if (/\[.+\]/);
                    if(/^\s*$item\s*=\s*(.+?)\s*$/i)
                    {
                       return $1;
                    }
		}
	    }
	}
    }

    # item not found
    return;
}

#--------------------------------------------------------------------------------
#gets 
#   cfg_file (path/file) - file to modify
#   section (may be undef) - section containing the item, if undef, sets the first matching item
#   item - item to set
#   value - new value
#returns
#   TRUE (1)  - successfully set the item (even if it didn't change value)
#   FALSE (0) - didn't set the item
sub cfg_set_config_item 
{
    my ($cfg_file, $section, $item, $value) = @_;

    my $lines = get_cfg_lines($cfg_file);

    if (not defined $lines)
    {
	print "F5::Cfg::cfg_set_config_item - Can't get lines from $cfg_file\n";
	return;
    }
  
    # flag to indicate the item was changed
    my $changed = FALSE;
    # flag to indicate the item was found (not necessarily changed)
    my $found = FALSE;

    OUTER_LOOP:
    for (my $i = 0; $i < @$lines; $i++)
    {
	# no section case
	if (not defined $section)
	{
	    if ($lines->[$i] =~ s/^\s*($item)\s*=\s*(\S*)\s*$/$1=$value/i)
	    {
		$changed = TRUE if ($value ne $2);
		$found = TRUE;
		last OUTER_LOOP;
	    }
	}
	# with section
	else
	{    
	    if($lines->[$i] =~ /^\s*\[\s*$section\s*\]/i)
	    {
		# found the section, now find the item, starting from the next line
		for (my $j = $i + 1; $j < @$lines; $j++)
		{
		    last OUTER_LOOP if ($lines->[$j] =~ /\[.+\]/); # end of section

		    if ($lines->[$j] =~ s/^\s*($item)\s*=\s*(\S*)\s*$/$1=$value/i)
		    {
			$changed = TRUE if ($value ne $2);
			$found = TRUE;
			last OUTER_LOOP;
		    } # item found

		} # inner loop

	    } # matched section

	} # with section

    } # outer loop

    if ($changed)
    {
	# need to update the file
	my $tmp_file = $cfg_file . '.tmp';
	my $buf = join("\n", @$lines) . "\n";

	# cfg file is plain text

	if (not open(FILE, '>' . $tmp_file))
	  {
	    print "F5::Cfg::cfg_set_config_item: can't open file $tmp_file: $!";
	    return FALSE;
	  }

	print FILE $buf;
	close FILE;

	# rename the tmp file to the real file name

	if (rename $tmp_file, $cfg_file)
	{
	    # successfully updated the file
	    return TRUE;
	}
	else
	{
	    print "F5::Cfg::cfg_set_config_item: can't rename file $tmp_file to $cfg_file: $!";
	    unlink $tmp_file;
	    return FALSE;
	}
    }
    else
    {
	# item not changed, return $found
	return $found;
    }
}

#--------------------------------------------------------------------------------
#gets 
#   cfg_file (path/file) - file to search
#returns
#   secs_list - ref to list of sections 
sub cfg_get_config_sections
{
    my $cfg_file = shift;
    my @secs_list;

    my $lines = get_cfg_lines($cfg_file);

    if (not defined $lines)
    {
	print "F5::Cfg::cfg_get_config_sections - Can't get lines from $cfg_file\n";
	return;
    }
    
    foreach (@$lines)
    {
	push(@secs_list, $1) if(/^\s*\[\s*(\S+)\s*\]/i);
    }
	
    return \@secs_list;
}

#--------------------------------------------------------------------------------
#routine: get_cfg_lines
#arguments:
#   $cfg_filename - cfg file to open - plain text or encrypted
#return value:
#   $lines - refernce to array of lines in the cfg_file
sub get_cfg_lines
{
    my $cfg_filename = F5::Deploy::map_path(shift);
    my @lines;

    if (not -f $cfg_filename)
    {
	print "F5::Cfg::get_cfg_lines - File not found: $cfg_filename\n";
	return;
    }

    # cfg file is plain text

    if (not open(CFG, $cfg_filename))
      {
	print "F5::Cfg::get_cfg_lines - Can't open file $cfg_filename: $!\n";
	return;
      }

    local $/ = "\n";  # just in case $/ has been modified elsewhere
    @lines = <CFG>;
    chomp @lines;

    close CFG;
    return \@lines;
}
#--------------------------------------------------------------------------------
#gets
#   NONE
#
#returns
#   mysql password from /var/db/mysqlpw[.<user>] (if not found - undef)

sub get_mysql_password
{
   my (%params) = @_;

   my $user = $params{user} || cfg_get_config_item('/etc/ts/common/users.cfg', 'DATABASE', 'User');

   # open mysql-password holder
   return undef if (! open(FILE, ($user eq 'root') ?
                                  MYSQL_PASSWORD_FILE : MYSQL_PASSWORD_FILE . '.' . $user) );

   # read file content
   my $value = <FILE>;
   close FILE;

   return undef if (! defined $value);

   # remove new line 
   chomp $value;

   # decrypt file content if needed
   if (defined $params{encrypted}) {
      return $value;
   } else {
      require PassCrypt;
      return PassCrypt::decrypt_password($value);

   }
}

#--------------------------------------------------------------------------------

sub add_entire_line {
    my (%params) = @_;

    my $filename = F5::Deploy::map_path($params{filename});
    my $line_to_add = $params{line_to_add} . "\n";

    open (my $fh, "<", $filename) or die "can't open file $filename";
    my @lines = <$fh>;
    close $fh;

    for my $line (@lines) {
        return if $line eq $line_to_add;
    }
    
    push @lines, $line_to_add;
    open (my $out, ">", $filename) or die "can't open file $filename";
    for my $line (@lines) {
        print $out $line;
    }
    close $out;
}

#--------------------------------------------------------------------------------

sub remove_entire_line {
    my (%params) = @_;

    my $filename = F5::Deploy::map_path($params{filename});
    my $re_line_to_remove = $params{re_line_to_remove};

    open (my $fh, "<", $filename) or die "can't open file $filename";
    my @lines = <$fh>;
    close $fh;

    my @new_lines = grep {$_ !~ $re_line_to_remove} @lines;

    open (my $out, ">", $filename) or die "can't open file $filename";
    for my $line (@new_lines) {
        print $out $line;
    }
    close $out;
}

1;
