package F5::CommonUpgrade::LiveUpdate;

use base 'F5::CommonUpgrade::Base';

use F5::DbUtils;
use F5::DateTimeUtils;

use YAML::XS;
use File::Copy;
use File::Slurp qw(read_file write_file);

our $max_version = '14.1.0';
our $tables_handled = [ 
                     qw(
                        PLC.SIGFILE_FLAGS
                        PLC.THREAT_CAMPAIGN_FILE_FLAGS
                        PLC.SIGFILE_STATUS
                        PLC.THREAT_CAMPAIGN_FILE_STATUS 
                       ), 
                      ];

# Browser challenges will not be included in upgraded ASM data
my %UPDATE_TYPES = (
    ASM_ATTACK_SIGNATURES => 'sigs',
    BOT_SIGNATURES => 'botsigs',
    SERVER_TECHNOLOGIES => 'server_technologies',
    THREAT_CAMPAIGNS => 'threat_campaigns',
);

my $LIVE_UPDATE_IMPORT_FILE = '/var/lib/hsqldb/live-update/live-update-import.yaml';
my $TEMP_LIVE_UPDATE_IMPORT_FILE = $LIVE_UPDATE_IMPORT_FILE . '.tmp';

my $installation_collection = 'com.f5.liveupdate.entity.InstallationCollection';
my $install_schedule_collection = 'com.f5.liveupdate.entity.InstallScheduleCollection';
my $availability_Collection = 'com.f5.liveupdate.entity.AvailabilityCollection';

sub liveupdate_installation {
  my(%params) = @_;
  my $class = 'com.f5.liveupdate.entity.Installation';

  my $blessed_hash = {%params};
  bless $blessed_hash, $class;
  return $blessed_hash;
}

sub liveupdate_install_schedule {
  my(%params) = @_;
  my $class = 'com.f5.liveupdate.entity.InstallSchedule';

  my %params_with_defaults = (
         applyOnAllDays => 1,
         applyOnSaturdays => 0,
         applyOnSundays => 0,
         applyOnFridays => 0,
         applyOnWednesdays => 0,
         applyOnThursdays => 0,
         applyOnMondays => 0,
         applyOnTuesdays => 0,
         applyAtAllTimes => 1,
         frequency => 'NEVER',
         endTime => '23:59',
         startTime => '00:00',
         %params
        );

  my $blessed_hash = {%params_with_defaults};
  bless $blessed_hash, $class;
  return $blessed_hash;
}

sub liveupdate_link {
  my(%params) = @_;
  my $class = 'com.f5.liveupdate.entity.Link';

  my $blessed_hash = {%params};
  bless $blessed_hash, $class;
  return $blessed_hash;
}

sub datanucleus_list {
  my $class = 'org.datanucleus.store.types.wrappers.backed.List';
  my $blessed_list = [];
  bless $blessed_list, $class;
  return $blessed_list;
}

sub after_all_tables_complete {
    my($self, %params) = @_;

    my $common_upgrade = $params{common_upgrade} || {};
    my $logger = $params{logger} || $common_upgrade->{logger};
    my $dbh    = $params{dbh};
    my $policy_id = $params{policy_id};
    my $dbname = $params{dbname} || 'PLC';

    # skip in policy import
    if ($policy_id && $policy_id ne '*') {
        return;
    }

    $logger->info("Begin Live Update Common Upgrade") if $logger;

    my $status;

    my $import_data = {version => $max_version,
                       $installation_collection => [],
                       $install_schedule_collection => [],
                       # this list is not being populated, but we need an empty one
                       # so that LU does not get NullPointerException
                       $availability_Collection => [],
                      };

    my $ary_ref = $dbh->selectcol_arrayref("SELECT sig_id, rest_uuid FROM $dbname.NEGSIG_SIGNATURES", { Columns=>[1, 2] });
    my $uuid_by_sig_id = {@{$ary_ref}};

    my ($update_mode) = $dbh->selectrow_array("SELECT update_mode FROM $dbname.SIGFILE_FLAGS");
    my $frequency = ($update_mode eq 'auto-interval') ? 'REAL_TIME' : 'NEVER' ;
    $logger->debug("Sigfile flags update_mode is $update_mode, frequency is set to $frequency") if $logger;

    # install schedule is always set - regardless of actual installations
    # including BROWSER_CHALLENGES but not THREAT_CAMPAIGNS, which is set later
    for my $type (keys %UPDATE_TYPES , 'BROWSER_CHALLENGES') {
        next if $type eq 'THREAT_CAMPAIGNS';

        my $install_schedule = liveupdate_install_schedule(type => $type,
                                                           frequency => $frequency,
                                                          );
        push @{$import_data->{$install_schedule_collection}} , $install_schedule;
    }

    # ordered so that latest would be the first
    my $sigfile_status_rows = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} });
SELECT *
  FROM $dbname.SIGFILE_STATUS
 WHERE flg_is_user_defined = 0
 ORDER BY sigfile_status_id DESC
EndQuery

    my $initial_base_sigfile_status_id = 0;
    if ($sigfile_status_rows) {
        $initial_base_sigfile_status_id = $sigfile_status_rows->[-1]->{sigfile_status_id};
    }

    my $number_of_status_rows = scalar @$sigfile_status_rows;
    $logger->debug("There are $number_of_status_rows sigfile status rows") if $logger;

    # first row is the latest one - it's what is currently installed
    $status = 'INSTALL_COMPLETE';
    for my $row (@$sigfile_status_rows) {
        my $load_datetime = F5::DateTimeUtils::convert_datetime_to_epoch_in_ms($row->{load_datetime});
        my $file_timestamp = F5::DateTimeUtils::convert_datetime_to_file_timestamp($row->{sigfile_timestamp});
        my $create_datetime = F5::DateTimeUtils::convert_datetime_to_epoch_in_ms($row->{sigfile_timestamp});

        my $file_name = 'ASM-AttackSignatures_' . $file_timestamp . '.im';

        $logger->debug("Handling sigfile $file_name, with $row->{load_datetime} load datetime and $row->{sigfile_timestamp} create datetime") if $logger;
        $logger->trace("Load datetime $row->{load_datetime} converted to $load_datetime, create datetime $row->{sigfile_timestamp} converted to $create_datetime") if $logger;

        for my $type (keys %UPDATE_TYPES) {
            next if $type eq 'THREAT_CAMPAIGNS';

            if ($type eq 'ASM_ATTACK_SIGNATURES') {
                $row->{sigs_updated_count} = $row->{sigs_updated_minor_count} + $row->{sigs_updated_major_count};
            }

            # can't use the same $update_file in all types, for one row,
            # it'll create links in yaml, which we do not want
            # so, we create a new $update_file, every time
            my $update_file = {
                               isFileAvailable => 0,
                               isFileManuallyUploaded => 0,
                               fileLocation => '',
                               md5 => '',
                               type => $type,
                               createDateTime => $create_datetime,
                               fileName => $file_name,
                              };

            my $deleted_count = $row->{$UPDATE_TYPES{$type} . '_deleted_count'} || 0;
            my $added_count = $row->{$UPDATE_TYPES{$type} . '_added_count'} || 0;
            my $modified_count = $row->{$UPDATE_TYPES{$type} . '_updated_count'} || 0;

            $logger->debug("For update type of $type, there were $deleted_count deleted entities, $added_count added entities and $modified_count modified entities") if $logger;

            my $is_initial_base_sigfile = ($row->{sigfile_status_id} == $initial_base_sigfile_status_id);
            # Previous installations should only be created if at least one entity of this type exists, or
            # if this is the initial base sigfile
            if (! $deleted_count && ! $added_count && ! $modified_count && ! $is_initial_base_sigfile) {
                next;
            }
            my $readme = $row->{readme};
            $readme =~ s/\s*$//s;

            my $installation = liveupdate_installation(
                deletedEntitiesCount => $deleted_count,
                addedEntitiesCount => $added_count,
                modifiedEntitiesCount => $modified_count,
                type => $type,
                status => $status,
                loadDateTime => $load_datetime,
                lastUpdateMicros => $load_datetime . '000',
                updateFile => $update_file,
                modifiedEntityList => datanucleus_list(),
                addedEntityList => datanucleus_list(),
                deletedEntityList => datanucleus_list(),
                readme => $readme,
            );

            # Only true of ASM Signatures -- Server Technologies and Bot Signatures are counts only
            if ($type eq 'ASM_ATTACK_SIGNATURES') {
                my $modified_signatures = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} });
SELECT *
  FROM $dbname.SIGFILE_MODIFIED_SIGNATURES
 WHERE sigfile_status_id = $row->{sigfile_status_id}
EndQuery

                for my $modified_sig (@$modified_signatures) {
                    my $link = liveupdate_link(name => $modified_sig->{sig_name},
                                               link => 'https://localhost/mgmt/tm/asm/signatures/' . $uuid_by_sig_id->{$modified_sig->{sig_id}},
                                              );

                    my $queue = ($modified_sig->{change_level} eq 'added') ? 'addedEntityList' : 'modifiedEntityList' ;
                    push (@{$installation->{$queue}}, $link);
                }
            }

            push @{$import_data->{$installation_collection}} , $installation;
        }

        # all next rows are what was previously installed
        $status = 'PREVIOUSLY_INSTALLED';
    }

    # no TC prior to v14
    if (F5::DbUtils::check_table_exists(dbh => $dbh,  dbname => $dbname,  table_name => 'THREAT_CAMPAIGNS')) {
        $ary_ref = $dbh->selectcol_arrayref("SELECT threat_campaign_id, rest_uuid FROM $dbname.THREAT_CAMPAIGNS", { Columns=>[1, 2] });
        my $uuid_by_threat_campaign_id = {@{$ary_ref}};

        # ordered so that latest would be the first
        my $threat_campaign_status_rows = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} });
SELECT *
  FROM $dbname.THREAT_CAMPAIGN_FILE_STATUS
 ORDER BY threat_campaign_file_status_id DESC
EndQuery

        $number_of_status_rows = scalar @$threat_campaign_status_rows;
        $logger->debug("There are $number_of_status_rows threat campaign file status rows") if $logger;

        my $type = 'THREAT_CAMPAIGNS';
        $status = 'INSTALL_COMPLETE';
        for my $row (@$threat_campaign_status_rows) {
            my $load_datetime = F5::DateTimeUtils::convert_datetime_to_epoch_in_ms($row->{load_datetime});
            my $file_timestamp = F5::DateTimeUtils::convert_datetime_to_file_timestamp($row->{threat_campaign_file_timestamp});
            my $create_datetime = F5::DateTimeUtils::convert_datetime_to_epoch_in_ms($row->{threat_campaign_file_timestamp});

            my $file_name = 'ThreatCampaigns_' . $file_timestamp . '.im';

            $logger->debug("Handling sigfile $file_name, with $row->{load_datetime} load datetime and $row->{threat_campaign_file_timestamp} create datetime") if $logger;
            $logger->trace("Load datetime $row->{load_datetime} converted to $load_datetime, create datetime $row->{threat_campaign_file_timestamp} converted to $create_datetime") if $logger;

            my $update_file = {
                               isFileAvailable => 0,
                               isFileManuallyUploaded => 0,
                               fileLocation => '',
                               md5 => '',
                               type => $type,
                               createDateTime => $create_datetime,
                               fileName => $file_name,
                              };

            my $deleted_count = $row->{$UPDATE_TYPES{$type} . '_deleted_count'} || 0;
            my $added_count = $row->{$UPDATE_TYPES{$type} . '_added_count'} || 0;
            my $modified_count = $row->{$UPDATE_TYPES{$type} . '_updated_count'} || 0;

            $logger->debug("For update type of $type, there were $deleted_count deleted entities, $added_count added entities and $modified_count modified entities") if $logger;

            # Previous installations should only be created if at least one entity of this type exists
            next unless $deleted_count || $added_count || $modified_count;

            my $installation = liveupdate_installation(
                deletedEntitiesCount => $deleted_count,
                addedEntitiesCount => $added_count,
                modifiedEntitiesCount => $modified_count,
                type => $type,
                status => $status,
                loadDateTime => $load_datetime,
                lastUpdateMicros => $load_datetime . '000',
                updateFile => $update_file,
                modifiedEntityList => datanucleus_list(),
                addedEntityList => datanucleus_list(),
                deletedEntityList => datanucleus_list(),
            );

            my $modified_threat_campaigns = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} });
SELECT *
  FROM $dbname.THREAT_CAMPAIGN_FILE_MODIFIED_THREAT_CAMPAIGNS
 WHERE threat_campaign_file_status_id = $row->{threat_campaign_file_status_id}
EndQuery

            for my $modified_threat_campaign (@$modified_threat_campaigns) {
                my $link = liveupdate_link(name => $modified_threat_campaign->{threat_campaign_display_name},
                                           link => 'https://localhost/mgmt/tm/asm/threat-campaigns/' . $uuid_by_threat_campaign_id->{$modified_threat_campaign->{threat_campaign_id}},
                                          );

                my $queue = ($modified_threat_campaign->{change_level} eq 'added') ? 'addedEntityList' : 'modifiedEntityList' ;
                push (@{$installation->{$queue}}, $link);
            }

            push @{$import_data->{$installation_collection}} , $installation;

            $status = 'PREVIOUSLY_INSTALLED';
        }

        my ($tc_update_mode) = $dbh->selectrow_array("SELECT update_mode FROM $dbname.THREAT_CAMPAIGN_FILE_FLAGS");
        my $frequency = ($tc_update_mode eq 'auto-interval') ? 'REAL_TIME' : 'NEVER' ;
        $logger->debug("Threat campaign file flags update_mode is $tc_update_mode, frequency is set to $frequency") if $logger;

        my $install_schedule = liveupdate_install_schedule(type => $type,
                                                           frequency => $frequency,
                                                          );
        push @{$import_data->{$install_schedule_collection}} , $install_schedule;
    }

    # all done, dump $import_data to a temporary yaml file
    YAML::XS::DumpFile($TEMP_LIVE_UPDATE_IMPORT_FILE, $import_data);

    #------------------------------------------------
    # for dome reason, 'YAML::XS::DumpFile' dumps
    #    '!!perl/hash:com.f5.liveupdate.entity.Link' instead of '!com.f5.liveupdate.entity.Link'
    # and
    #    '!!perl/array:com.f5.liveupdate.entity.Link' instead of '!com.f5.liveupdate.entity.Link'
    # though 'YAML::XS::LoadFile' reads a yaml w/o the '!perl/array:' or '!perl/hash:' notations
    # perfectly well, into Perl blessed hash/array.
    # the problem is that LU blows up on reading the yaml with these Perl notations being added
    #
    # remove the '!perl/array:' or '!perl/hash:' notations from yaml file - 
    #------------------------------------------------
    my $yaml_contents = File::Slurp::read_file($TEMP_LIVE_UPDATE_IMPORT_FILE, {binmode => ':utf8'});
    $yaml_contents =~ s|!perl/hash:||g;
    $yaml_contents =~ s|!perl/array:||g;
    File::Slurp::write_file($TEMP_LIVE_UPDATE_IMPORT_FILE, {binmode => ':utf8'}, $yaml_contents);
    #------------------------------------------------

    # move the temporary yaml file, to a file that will be automatically imported (and deleted) by LU
    File::Copy::move($TEMP_LIVE_UPDATE_IMPORT_FILE,$LIVE_UPDATE_IMPORT_FILE);

    $logger->info("End Live Update Common Upgrade") if $logger;
}

1;
