package F5::ConfigSet;

use strict;
use warnings;

use File::Spec;
use File::Slurp;
use File::Temp;
use File::Path qw(make_path rmtree);
use File::Basename;
use File::Find;
use File::Copy;
use MIME::Base64;
use JSON;
use YAML::XS;
use Digest::MD5;
use Scalar::Util qw(reftype);

use F5::Deploy;
use F5::DeviceConfig;
use F5::Utils::Account;
#TODO: We can't use a mix of EasyClient and local dbh when using SQLite. This will use Handler directly until everything can be done via client.
#use F5::ASMConfig::GuaranteedClient;
use F5::ASMConfig::Handler;
use F5::SystemUtils;
use F5::SetActive;
use F5::PrepareConf::Base;
use F5::PrepareConf::LoggingProfile;
use F5::Mcp::FwLoggerType;
use F5::Mcp::FwLogRemoteProtocol;
use F5::Mcp::FwLogFilterKey;
use F5::Mcp::SyslogFacility;
use F5::DataProtection;
use F5::DbUtils;
use F5::ObjUtils qw(dclone);
use F5::Logger;
use F5::Db::Single;
use F5::BdAgent;
use F5::BatchInsert;
use F5::Utils::Policy;
use F5::Utils::File;

my $ACCOUNT_DATA_DIR = $F5::PrepareConf::Base::ACCOUNT_DATA_DIR;
my $CONFIG_DIR = F5::Deploy::map_path('/var/ts/files/site_1/config');
my $FULL_CONFIG_DIR = "$CONFIG_DIR/full";
my $TMP_DIR = '/tmp';
my $K_BYTES = 1024;
my $DEFAULT_POLICY_MAP_FILENAME = "$CONFIG_DIR/policy_path.map";
my $UPDATES_DIR = F5::Deploy::map_path('/ts/var/update_files');

my %updates = ( signatures => {
                                update_file => "$UPDATES_DIR/signatures/signature_update.yaml",
                                version_file => "$UPDATES_DIR/signatures/version",
                                package      => "F5::NegativeSignatures",
                                update_rpc   => "update_sigfile",
                                object_class => "F5::ASMConfig::Sigfile",
                                settings_class => "F5::ASMConfig::SigfileUpdateSettings",
                              },
                threat_campaigns => {
                                update_file => "$UPDATES_DIR/threat_campaigns/threat_campaign_update.yaml",
                                version_file => "$UPDATES_DIR/threat_campaigns/version",
                                package      => "F5::ThreatCampaignFile",
                                update_rpc   => "update_threat_campaign_file",
                                object_class => "F5::ASMConfig::ThreatCampaignFile",
                                settings_class => "F5::ASMConfig::ThreatCampaignFileUpdateSettings",
                              },
               );


#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = {};
    bless $self, $class;
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;
    my $input_file = $params{input_file};
    my $config_set_file = $params{config_set_file};
    my $previous_file = $params{previous_file};

    $self->{save_for_reuse} = $params{save_for_reuse};

    $self->{logger} = $params{logger} || F5::Logger->new('/ts/log/config_set.log');
    $self->{require_cookie_seed} = defined($params{require_cookie_seed}) ? $params{require_cookie_seed} : 1;
    F5::Db::Single->init;

    if ($previous_file || $self->{save_for_reuse}) {
        $self->{calculate_checksums} = 1;
    }
    if ($previous_file) {
        $self->load(config_set_file => $previous_file);
    }

    if ($config_set_file) {
        $self->load(config_set_file => $config_set_file);
    }

    $self->{dbh} = F5::Db::Single->instance();
    $self->{client} = $params{client} || F5::ASMConfig::Handler->new(username => 'config_set_compiler', alarm_on_destroy => 0, dbh => $self->{dbh});

    my %job_attributes = (logger => $self->{logger},
                          component => 'config_set',
                          operationId => 'init',
                         );
    if ($input_file) {
        $job_attributes{configSetSize} = -s $input_file;
        $job_attributes{configSetChecksum} = Digest::MD5->new->add($input_file)->hexdigest;
    }

    $self->{job} = F5::Log::Job->new(%job_attributes);
    $self->{job}->set_started();

    if ($input_file) {
        if ($params{dry_run}) {
            $self->{logger}->info('Running in "Dry Run" mode.') if $self->{logger};
        }
        eval {
            $self->create(%params);
        };
        my $err = $@;
        die $err if $err;
    }
    elsif ($params{update_type}) {
        $self->handle_update(filename => $params{update_file},
                             update_type => $params{update_type},);
    }

    $self->{job}->set_completed();
}

#----------------------------------------------------------------------

sub create {
    my($self, %params) = @_;

    my $apply_security_updates        = defined $params{apply_security_updates} ? $params{apply_security_updates} : 1;
    my $delete_unused_policies        = defined $params{delete_unused_policies} ? $params{delete_unused_policies} : 1;
    my $apply_user_defined_signatures = defined $params{apply_user_defined_signatures} ? $params{apply_user_defined_signatures} : 1;
    my $verify_policies               = defined $params{verify_policies} ? $params{verify_policies} : 1;

    if ($params{input_file}) {
        my $input_file = $params{input_file};

        $self->{config} = $self->load_file(filename => $input_file);
    }
    if ($params{config}) {
        $self->{config} = $params{config};
    }

    # Avoid standard set active during config set building
    my $call_flags = F5::ASMConfig::CallFlags->new(skip_set_active => 1, suppress_logging => 1);

    if (!$self->{previous_config}) {
        my $previous_config = F5::DbUtils::select_singleton(dbh => $self->{dbh}, table_name => 'PLC.CONFIG_SET');
        if ($previous_config && $previous_config->{config_set}) {
            $self->{previous_config} = YAML::XS::Load($previous_config->{config_set});
            $self->{calculate_checksums} = 1;
        }
    }

    $self->{client}->begin_work(); ## Prevent commits until we are certain that all the policies and updates are valid

    if ($self->{previous_config} && defined($self->{previous_config}->{cookie_seed}) &&
         (!defined($self->{config}->{cookie_seed}) || $self->{previous_config}->{cookie_seed} eq $self->{config}->{cookie_seed}) ) {
        $self->{logger}->info("Specified cookie seed matches from previous run -- skipping") if $self->{logger};
        $self->{config}->{cookie_seed} = $self->{previous_config}->{cookie_seed};
    }
    else {
        if (! $self->{config}->{cookie_seed} && $self->{require_cookie_seed}) {
            #Set as a random string if not supplied
            $self->{logger}->info("Creating random cookie seed") if $self->{logger};
            my @chars = ('a'..'z', 'A'..'Z', '0'..'9');
            $self->{config}->{cookie_seed} = join '', map { @chars[rand @chars] } 1 .. 16;
        }

        if ( $self->{config}->{cookie_seed}) {
            $self->{logger}->info("Generating new DataProtection") if $self->{logger};
            my $data_protection_rv = $self->{client}->update_data_protection($self->{client}->{session},
                                                                             $call_flags,
                                                                             undef,
                                                                             F5::ASMConfig::DataProtection->new(secure_key => F5::DataProtection::generate_key($self->{config}->{cookie_seed})));
        }
    }

    my $prepare_sigs = $self->{previous_config} ? $self->{previous_config}->{prepare_sigs} : 0;
    my $prepare_policies = {};
    if ($apply_user_defined_signatures) {
        my $uds_rv = $self->handle_user_defined_signatures(%params, call_flags => $call_flags, validate_requirements => 0);
        $prepare_sigs ||= $uds_rv->{prepare_sigs};
        if ($uds_rv->{prepare_policies}) {
            $prepare_policies = {%$prepare_policies, %{$uds_rv->{prepare_policies}}};
        }
    }

    if ($apply_security_updates) {
        for my $update_type (keys %updates) {
            my $filename = "${update_type}_filename";
            if ($self->{config}->{$filename}) {
                if ($self->{calculate_checksums}) {
                    $self->{config}->{"${filename}_md5"} = $self->get_file_md5(filename => $self->{config}->{$filename});
                }
                if ($self->{previous_config} &&
                    ($self->{previous_config}->{"${filename}_md5"} || '') eq $self->{config}->{"${filename}_md5"}) {
                     $self->{logger}->info("Specified $update_type file matches signatures from previous run -- skipping") if $self->{logger};
                }
                else {
                    my $rv = $self->handle_update(dry_run => $params{dry_run},
                                                  call_flags => $call_flags,
                                                  filename => $self->{config}->{$filename},
                                                  im_file => 1,
                                                  skip_set_sigs_active => 1,
                                                  validate_requirements => 0,);
                    $prepare_sigs ||= $rv->{prepare_sigs};
                    if ($rv->{prepare_policies}) {
                        $prepare_policies = {%$prepare_policies, %{$rv->{prepare_policies}}};
                    }
                }
            }
            else {
                my $rv = $self->handle_update(%params,
                                              call_flags => $call_flags,
                                              update_type => $update_type,
                                              skip_set_sigs_active => 1,
                                              validate_requirements => 0,);
                $prepare_sigs ||= $rv->{prepare_sigs};
                if ($rv->{prepare_policies}) {
                    $prepare_policies = {%$prepare_policies, %{$rv->{prepare_policies}}};
                }
            }
        }
    }

    my %existing_policies;
    my %policies_to_import;
    my $policy_path_map;
    my @vs_config;

    if (!$verify_policies) { 
        $self->{logger}->info("Specified policies will not be verified and must match pre-built configuration");
    }

    for my $policy_key (keys %{$self->{config}->{policies}}) {
        my $policy = $self->{config}->{policies}->{$policy_key};
        my $previous_policy;
        if ($self->{previous_config} && $self->{previous_config}->{policies} && $self->{previous_config}->{policies}->{$policy_key}) {
            $previous_policy = $self->{previous_config}->{policies}->{$policy_key};
        }
        my $filename;
        my $delete_temp_file = 0;
        if ($policy->{import_contents}) {
            my $fh;
            ($fh, $filename) = File::Temp::tempfile();
            print $fh MIME::Base64::decode_base64($policy->{import_contents});
            close $fh;
            $delete_temp_file = 1;
        }
        else {
            $filename = File::Spec->rel2abs($policy->{import_filename});
        }
        my $src_config_line = delete $policy->{src_config_line};
        my $import_filename_md5 = $self->get_file_md5(filename => $filename,
                                                      src_config_line => $src_config_line);
        $policy->{import_filename_md5} = $import_filename_md5;

        my $import_policy_name = $policy->{name};
        if ($verify_policies) {
            if (!$import_policy_name) {
                if ($policy_path_map->{$policy->{import_filename_md5}}) {
                    $import_policy_name = $policy_path_map->{$import_filename_md5}->{policy_name};
                }
                else {
                    my $import_rv;
                    eval {
                        $import_rv = $self->{client}->import_policy($self->{client}->{session},
                                                                    $call_flags,
                                                                    F5::ASMConfig::ImportExportProperties->new(filename => $filename,
                                                                                                               force_policy_name_flg => 1,
                                                                                                               allow_encoding_change => 1,
                                                                                                               get_policy_attributes_only => 1,
                                                                                                               resolve_external_references => 1,));
                    };
                    if (!$import_rv || !$import_rv->{completed_successfully}) {
                         my $err = $@;
                         if ($import_rv && $import_rv->{msgs}) {
                            $err = join ("\n", map { $_->{msg} } @{$import_rv->{msgs}});
                         }
                         elsif (ref $err) {
                             $err = $err->{display_message} || $err->{error_message};
                         }
                         my $msg = "Failed to import Policy '$policy_key' from '$filename': $err";
                         $self->{logger}->error($msg)  if $self->{logger};
                         die {"error_message" => $msg,
                              "error_line_number" => $src_config_line};
                    }
                
                    $import_policy_name = $import_rv->{policy_name};
                    if ($import_rv->{external_reference_count}) {
                        $policy->{content} = $import_rv->{content};
                    }
                }
            }
            $existing_policies{$import_policy_name}->{src_config_line} = $src_config_line;
            $policy_path_map->{$policy->{import_filename_md5}}->{policy_name} = $import_policy_name;
            $policy_path_map->{$policy->{import_filename_md5}}->{description} = $filename;
            if ($policies_to_import{$import_policy_name}) {
                 my $msg = "Duplicate policy names found '$import_policy_name'";
                 $self->{logger}->error($msg)  if $self->{logger};
                 die {"error_message" => $msg,
                      "error_line_number" => $src_config_line};
            }
        }

        if (!$policy->{vs_info} && $policy->{logging}) {
            my $vs_name = $policy->{vs_name} || "vs1";
            $policy->{vs_info} = { $vs_name => {
                                                 logging => $policy->{logging}
                                               }
                                  };
        }

        my $logging_config = $self->build_logging_config(policy => $policy);
        push (@vs_config,  @$logging_config);
        delete $policy->{vs_info};

        if (!$verify_policies) {
             #Skip policy imports
            $self->{logger}->debug("Policies verification is disabled. Skipping policy imports.")  if $self->{logger};
        }
        elsif (!$prepare_policies->{$import_policy_name} && $previous_policy && are_hashes_equivalent($previous_policy, $policy)) {
            $self->{logger}->info("Policy $policy_key matches signature from previous run -- skipping")  if $self->{logger};
        }
        else {
            $self->{logger}->info("Preparing importing for policy $policy_key")  if $self->{logger};
            my $import_export_properties = F5::ASMConfig::ImportExportProperties->new(filename => $filename,
                                                                                      src_config_line => $src_config_line,
                                                                                      policy_name => $import_policy_name,
                                                                                      force_policy_name_flg => 1,
                                                                                      overwrite => 1,
                                                                                      allow_encoding_change => 1,
                                                                                      policy_description => $filename, #store for policy_path_map_generation
                                                                                      content => $policy->{content},
                                                                                      do_validation => 1,
                                                                                     );
            $policies_to_import{$import_policy_name} = {import_export_properties => $import_export_properties,
                                                        delete_temp_file => $delete_temp_file,};
        }
    }

    my @policy_names = keys %policies_to_import;
    for my $policy_name (@policy_names) {
        my $import_policy = $policies_to_import{$policy_name};
        my $import_policy_name = $import_policy->{import_export_properties}->{policy_name};
        my $import_rv;
        eval {
            $import_rv = $self->{client}->import_policy($self->{client}->{session},
                                                        $call_flags,
                                                        $import_policy->{import_export_properties});
             if ($import_rv && $import_rv->{msgs}) {
                  my @msgs = map { $_->{msg} } grep { $_->{level} eq 'warning'} @{$import_rv->{msgs}};
                  if (@msgs) {
                      $self->{warnings}->{$import_policy_name} = \@msgs;
                  }
             }
        };

        if ($import_policy->{delete_temp_file}) {
           unlink $import_policy->{import_export_properties}->{filename};
        }

        if (!$import_rv || !$import_rv->{completed_successfully}) {
            my $err = $@;
            if ($import_rv && $import_rv->{msgs}) {
                $err = join ("\n", map { $_->{msg} } @{$import_rv->{msgs}});
            }
            elsif (ref $err) {
                $err = $err->{display_message} || $err->{error_message};
            }

            my $msg = "Failed to import Policy '$import_policy->{import_export_properties}->{policy_name}' from '$import_policy->{import_export_properties}->{filename}': $err";
            $self->{logger}->error($msg)  if $self->{logger};
            die {"error_message" => $msg,
                 "error_line_number" => $import_policy->{import_export_properties}->{src_config_line}};
        }
    }
    $self->{client}->commit(); ## All updates and imported policies are valid for eachother
    my $change_made = (@policy_names || $prepare_sigs);

    my $deleted_policies = 0;
    if (!$params{dry_run}) {
        my $generate_negsig_collections = 0;
        if ($verify_policies) { 
            my @policies = F5::ASMConfig::Entity::Policy->retrieve_all(dbh => $self->{dbh}, get_criteria => {});
            for my $policy (@policies) {
                if (!$existing_policies{$policy->{policy_name}} && $delete_unused_policies) {
                    if ($policy->{is_active_policy}) {
                        F5::SetActive::set_active(dbh => $self->{dbh},
                                                  is_daemon => 0,
                                                  use_daemon => 0,
                                                  policy_id => $policy->{policy_id},
                                                  update_ramcache => 0,
                                                  do_update_core => 0,
                                                  do_save_policy_history => 0,
                                                  deactivate_policy => 1,
                                                  asm_config_handler => $self->{client});
                    }
                    $self->{client}->delete_policy($self->{client}->{session},
                                                   $call_flags,
                                                   $policy);
                    $deleted_policies++;
                }
                elsif ($existing_policies{$policy->{policy_name}} && ($policy->{is_modified} || !$policy->{is_active_policy})) {
                    $generate_negsig_collections = 1;
                    my $set_active_rv = F5::SetActive::set_active(dbh => $self->{dbh},
                                                                  is_daemon => 0,
                                                                  use_daemon => 0,
                                                                  policy_id => $policy->{policy_id},
                                                                  update_ramcache => 0,
                                                                  do_update_core => 0,
                                                                  expanded_return_value => 1,
                                                                  do_save_policy_history => 0,
                                                                  do_validation => 0,
                                                                  asm_config_handler => $self->{client},
                                                                  generate_negsig_collections => 0,
                                                                 );
                    if (!$set_active_rv->{completed_successfully}) {
                        my $msg = "Failed to compile Policy " . $policy->{policy_name};
                        $self->{logger}->error($msg) if $self->{logger};
                        die {"error_message" => $msg,
                             "error_line_number" => $existing_policies{$policy->{policy_name}}->{src_config_line} };
                    }
                }
            }
            my @policy_paths = values %$policy_path_map;
            $self->create_policy_map(%params,
                                     policies => \@policy_paths);
        }

        if ($prepare_sigs) {
            my $negsig_update = F5::NegativeSignatures::SigFile::Update->new(dbh => $self->{dbh}, logger => $self->{logger});
            $negsig_update->set_sigs_active(asm_config_handler => $self->{client}, skip_set_active => 1);
        }
        elsif ($generate_negsig_collections) {
            my $negative_signatures = F5::NegativeSignatures->new(dbh => $self->{dbh},
                                                                  logger => $self->{logger},
                                                                  use_memory_cache => 1);

            my $collection = F5::NegativeSignatures::Collection->new(dbh => $self->{dbh},
                                                                     logger => $self->{logger},
                                                                     negative_signatures => $negative_signatures,
                                                                    );
            $collection->generate();
        }

        @vs_config = sort {$a->{log_profile_name} cmp $b->{log_profile_name} } @vs_config;
        $self->{config}->{vs_config} = \@vs_config;
        if ($self->{previous_config} && $self->{previous_config}->{vs_config} && are_hashes_equivalent($self->{config}->{vs_config}, $self->{previous_config}->{vs_config})) {
            $self->{logger}->info("Logging configuration matches signature from previous run -- skipping")  if $self->{logger};
        }
        else {
            $self->prepare_all_logging_profiles(vs_config => \@vs_config);
        }
    }
    elsif ($change_made) {
       #cache config changes from dry run
       if ($self->{previous_config}) {
           $self->{config}->{vs_config} = $self->{previous_config}->{vs_config};
       }
       $self->{config}->{prepare_sigs} = $prepare_sigs;
       F5::DbUtils::upsert_singleton(dbh => $self->{dbh},
                                     table_name => "PLC.CONFIG_SET",
                                     rec => {
                                         config_set => YAML::XS::Dump($self->{config}),
                                         last_updated => time(),
                                     }
                                    );
    }
    elsif ($self->{previous_config}) {
       $self->{config}->{vs_config} = $self->{previous_config}->{vs_config};
    }

    if ($deleted_policies && $self->{save_for_reuse} && F5::DbUtils::is_sqlite($self->{dbh})) {
        $self->{dbh}->do('VACUUM PLC');
        $self->{dbh}->do('VACUUM DCC');
    }
}

#----------------------------------------------------------------------
#
sub create_policy_map {
    my ($self, %params) = @_;
    ## Create policy map
    my $dbh = $params{dbh} || $self->{dbh};
    my @policies = $params{policies} ? @{$params{policies}} : ();
    my $policy_map_file = $params{policy_map_location} || $DEFAULT_POLICY_MAP_FILENAME;
    if (!@policies) {
        @policies = F5::ASMConfig::Entity::Policy->retrieve_all(dbh => $dbh, get_criteria => {});
    }
    my @policy_mappings;
    for my $policy (@policies) {
        #bd expects all lines to end in \n, even the last one.
        push (@policy_mappings, "$policy->{description} $policy->{policy_name}\n");
    }
    my $policy_map = join('', @policy_mappings);
    open (my $policy_map_fh, '>', $policy_map_file);
    print $policy_map_fh $policy_map;
}

#----------------------------------------------------------------------

sub are_hashes_equivalent {
    my($hash_a, $hash_b) = @_;

    return YAML::XS::Dump($hash_a) eq YAML::XS::Dump($hash_b);
}

#----------------------------------------------------------------------

sub get_file_md5 {
    my($self, %params) = @_;
    my $filename = $params{filename};
    my $src_config_line = $params{src_config_line};
    my $res;
    eval {
        $res = F5::Utils::File::get_file_md5($filename);
    };
    if (my $err = $@) {
        $self->{logger}->error($err)  if $self->{logger};
        my $error = {"error_message" => $err};
        if ($src_config_line) {
            $error->{error_line_number} = $src_config_line;
        }
        die $error;
    }
    return $res;
}

#----------------------------------------------------------------------

sub get_update_version {
    my ($self, %params) = @_;
    my $update_type = $params{update_type};

    my $rv = {};
    my $update_file_date = $updates{$update_type}->{package}->new(dbh => $self->{dbh})->get_update_file_date();
    if ($self->{"${update_type}_timestamp"} || $update_file_date) {
        $rv->{revision_datetime} = $self->{"${update_type}_timestamp"} || F5::DateTimeUtils::convert_epoch_to_rfc3339(F5::DateTimeUtils::convert_datetime_to_epoch($update_file_date));
    }

    if (-e $updates{$update_type}->{version_file}) {
        $rv->{version} = File::Slurp::read_file($updates{$update_type}->{version_file});
    }

    return $rv;
}

sub get_user_signatures_package_version {
    my ($self, %params) = @_;
    my @rv;
    if ($self->{config}->{user_defined_signatures_hash}) {
        for my $filename (keys %{$self->{config}->{user_defined_signatures_hash}}) {
            my $file_rv;
            $file_rv->{tag} = $self->{config}->{user_defined_signatures_hash}->{$filename}->{tag};
            if ($self->{config}->{user_defined_signatures_hash}->{$filename}->{revisionDatetime}) {
                $file_rv->{revisionDatetime} = $self->{config}->{user_defined_signatures_hash}->{$filename}->{revisionDatetime};
            }
            push (@rv,$file_rv);
        }
    }
    return \@rv;
}

#----------------------------------------------------------------------

sub handle_update {
    my ($self, %params) = @_;
    my $skip_set_sigs_active = $params{skip_set_sigs_active};
    my $call_flags = $params{call_flags} || F5::ASMConfig::CallFlags->new(skip_set_active => 1, skip_set_sigs_active => $skip_set_sigs_active);
    my $filename    = $params{filename};
    my $im_file     = $params{im_file};
    my $file_format = $params{file_format};
    my $update_type = $params{update_type};
    my $validate_requirements = defined $params{validate_requirements} ? $params{validate_requirements} : 1;

    my $prepare_sigs = 0;
    my $policies_to_prepare = {};
    if (!$im_file) {
        my $update_file = $filename || $updates{$update_type}->{update_file};
        my $update;
        eval {
            $update = YAML::XS::LoadFile($update_file);
        };
        if ($update) {
            my $loaded_timestamp = F5::DateTimeUtils::convert_datetime_to_epoch($updates{$update_type}->{package}->new(dbh => $self->{dbh})->get_update_file_date);
            $self->{config}->{"${update_type}_timestamp"} = $update->{revisionDatetime};
            my $timestamp = F5::DateTimeUtils::convert_rfc3339_to_epoch($update->{revisionDatetime});
            if ($loaded_timestamp == $timestamp) {
                if ($self->{previous_config} && (!$self->{previous_config}->{"${update_type}_timestamp"} || ($self->{previous_config}->{"${update_type}_timestamp"} ne $update->{revisionDatetime}))) {
                     $prepare_sigs = 1;
                }
                $self->{"${update_type}_timestamp"} = $update->{revisionDatetime};
                return { prepare_sigs => $prepare_sigs };
            }

            $self->{logger}->info("Loading $update_type revision $update->{revisionDatetime}") if $self->{logger};
            $filename = dirname($update_file) . "/$update->{filename}";
            $im_file = 0;
            $file_format = 'binary';
        }
    };

    if ($filename) {
        if (! -e $filename) {
            my $msg = "Missing update file $filename";
            $self->{logger}->error($msg) if $self->{logger};
            die "$msg\n";
        }
        $self->{logger}->info("Loading $update_type file $filename") if $self->{logger};

        my $update_rpc = $updates{$update_type}->{update_rpc};
        my $object_class = $updates{$update_type}->{object_class};
        my $settings_class = $updates{$update_type}->{settings_class};
        my $update_rv = $self->{client}->$update_rpc($self->{client}->{session},
                                                     $call_flags,
                                                     $object_class->new(filename => $filename,
                                                                        file_format => $file_format),
                                                     $settings_class->new(allow_older_timestamp => 1,
                                                                          update_type => F5::ASMConfig::UpdateType::UPDATE_TYPE_ASM_SIGNATURES(),
                                                                          dry_run => $params{dry_run},
                                                                          im_file => $im_file,
                                                                          new_live_update => !$im_file,
                                                                          validate_requirements => $validate_requirements, ));

        if (!$update_rv || !$update_rv->{completed_successfully}) {
             my $err = $@;
             if ($update_rv && $update_rv->{error_msg}) {
                 $err = $update_rv->{error_msg}
             }
             elsif (ref $err) {
                 $err = $err->{display_message} || $err->{error_message};
             }
             my $msg = "Failed to install $update_type file '$filename': $err";
             $self->{logger}->error($msg)  if $self->{logger};
             die {"error_message" => $msg};
        }

        if ($skip_set_sigs_active) {
            $prepare_sigs = 1;
        }
        if ($update_type eq 'signatures' && !$validate_requirements && $self->{config}->{"${update_type}_timestamp"}) {
            my @violating_sig_reqs = F5::Utils::Policy::get_policy_signature_requirements( dbh => $self->{dbh},
                                                                                           datetime => $self->{config}->{"${update_type}_timestamp"},
                                                                                           sig_tags => [''] );
            for my $sig_req (@violating_sig_reqs) {
                $policies_to_prepare->{$sig_req->{policy}->{policy_name}} = 1;
            }
        }
    }
    else {
        $self->{logger}->info("No $update_type update found.") if $self->{logger};
    }
    return { prepare_sigs => $prepare_sigs,
             prepare_policies => $policies_to_prepare };
}

#----------------------------------------------------------------------

sub handle_user_defined_signatures {
  my ($self, %params) = @_;
  my $call_flags = F5::ASMConfig::CallFlags->new(%{$params{call_flags} || {}}, skip_set_active => 1, skip_set_sigs_active => 1);
  my $validate_requirements = defined $params{validate_requirements} ? $params{validate_requirements} : 1;
  my $prepare_sigs = 0;

  #Note: This should avoid failing the import due to existing policy requirements, but force any such policies that would have caused a failure to be recompiled even if they haven't changed.
  #This will allow for changing signatures and policies together, and correct failure on a mismatch.

  my $existing_tags = F5::DbUtils::get_distinct_values(dbh => $self->{dbh},
                                                       table_name => 'PLC.NEGSIG_SIGNATURES',
                                                       fields => 'sig_tag',
                                                       where_fields => {
                                                                 sig_tag => {
                                                                     values => [''],
                                                                     comparison => 'ne',
                                                                 }
                                                            },);

  my $seen_tags;
  my $policies_to_prepare = {};
  if ($self->{config}->{user_defined_signatures}) {
      for my $file (@{$self->{config}->{user_defined_signatures}}) {
          my $filename = $file->{filename};
          $self->{config}->{user_defined_signatures_hash}->{$filename}->{md5} = $self->get_file_md5(%$file); #$file has: filename, src_config_line
          my $tag;
          my $revisionDatetime;
          my $sigfile;

          if ($self->{previous_config} &&
              ($self->{previous_config}->{user_defined_signatures_hash}->{$filename}->{md5} || '') eq $self->{config}->{user_defined_signatures_hash}->{$filename}->{md5}) {
              $tag = $self->{previous_config}->{user_defined_signatures_hash}->{$filename}->{tag};
              $revisionDatetime = $self->{previous_config}->{user_defined_signatures_hash}->{$filename}->{revisionDatetime};
          }

          if (!$tag) {
              $sigfile = $self->load_file(%$file);
              $tag = $sigfile->{tag};
              $revisionDatetime = $sigfile->{revisionDatetime};
          }

          my $error_message;
          if (!$tag) {
              $error_message = "Missing tag in file '$filename'";
          }
          elsif ($seen_tags->{$tag}++) {
              $error_message = "Duplicate tag '$tag' in file '$filename'";
          }

          if ($error_message) {
              $self->{logger}->error($error_message) if $self->{logger};
              my $error = {"error_message" => $error_message};
              if ($file->{src_config_line}) {
                  $error->{error_line_number} = $file->{src_config_line};
              }
              die $error;
          }

          $self->{config}->{user_defined_signatures_hash}->{$filename}->{revisionDatetime} = $revisionDatetime;
          $self->{config}->{user_defined_signatures_hash}->{$filename}->{tag} = $tag;

          if (!$sigfile) {
              $self->{logger}->debug("Specified signature file ($filename) matches previous run -- skipping") if $self->{logger};
              next;
          }
          $self->{logger}->info("Importing signatures with tag: '$tag' from file '$filename'") if $self->{logger};
          $prepare_sigs = 1;
          my $update_rv = $self->{client}->update_sigfile($self->{client}->{session},
                                                          $call_flags,
                                                          F5::ASMConfig::Sigfile->new(filename => $filename,
                                                                                      flg_is_user_defined => 1,),
                                                          F5::ASMConfig::SigfileUpdateSettings->new(allow_older_timestamp => 1,
                                                                            im_file => 0,
                                                                            update_type => F5::ASMConfig::UpdateType::UPDATE_TYPE_ASM_SIGNATURES(),
                                                                            dry_run => $params{dry_run},
                                                                            validate_requirements => $validate_requirements,
                                                                           ));
          if ($revisionDatetime) {
              my @violating_sig_reqs = F5::Utils::Policy::get_policy_signature_requirements( dbh => $self->{dbh},
                                                                                             datetime => $revisionDatetime,
                                                                                             sig_tags => [$tag] );
              for my $sig_req (@violating_sig_reqs) {
                  $policies_to_prepare->{$sig_req->{policy}->{policy_name}} = 1;
              }
          }
      }
  }

  my @tags_to_delete;
  for my $tag (keys %$existing_tags) {
      if (!$seen_tags->{$tag}) {
          $self->{logger}->info("Removing signatures with tag: $tag") if $self->{logger};
          push (@tags_to_delete, $tag);
      }
  }
  if (@tags_to_delete) {
      $prepare_sigs = 1;
      $call_flags->{search_filters} = [F5::ASMConfig::SearchFilter->new({ field_name => 'sig_tag',
                                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                                          field_value_list => \@tags_to_delete })];
      $self->{client}->delete_many_signatures($self->{client}->{session},
                                              $call_flags,
                                              F5::ASMConfig::Signature->new());
      my @violating_sig_reqs = F5::Utils::Policy::get_policy_signature_requirements( dbh => $self->{dbh},
                                                                                     sig_tags => \@tags_to_delete );
      for my $sig_req (@violating_sig_reqs) {
          $policies_to_prepare->{$sig_req->{policy}->{policy_name}} = 1;
      }
  }

  return { prepare_sigs => $prepare_sigs,
           prepare_policies => $policies_to_prepare,
         };
}

#----------------------------------------------------------------------

sub save {
    my($self, %params) = @_;
    my $output_file = $params{output_file};
    my $applied_config = $params{applied_config};
    my $policy_map_location = $params{policy_map_location} || $DEFAULT_POLICY_MAP_FILENAME;

    my @extra_files;
    my $database_base_dir = F5::Deploy::get_path_by_type("database");
    if ($applied_config) {
        push (@extra_files, $FULL_CONFIG_DIR);
    }
    elsif ($self->{save_for_reuse}) {
        F5::DbUtils::upsert_singleton(dbh => $self->{dbh}, table_name => "PLC.CONFIG_SET",
           rec => {
             config_set => YAML::XS::Dump($self->{config}),
             last_updated => time(),
           });
        push(@extra_files, "$database_base_dir/PLC", "$database_base_dir/DCC");
    }
    else {
        push (@extra_files, "$database_base_dir/DCC");
    }

    my $result = F5::SystemUtils::run_system_tool(command_line_params => ['tar', '-zpc',
                                                             '-f', $output_file,
                                                             $ACCOUNT_DATA_DIR,
                                                             $policy_map_location,
                                                             @extra_files,
                                                            ]);
    if ($result->{completed_successfully}) {
        $result->{filename} = $output_file;
        $result->{file_size} = -s $output_file;
    }
    else {
        $result->{error_message} = $result->{standard_error};
    }
    return $result;
}

#----------------------------------------------------------------------

sub load {
    my($self, %params) = @_;
    my $config_set_file = $params{config_set_file};
    if (! -e $config_set_file) {
        $self->{logger}->warn("Could not find file to load: '$config_set_file'");
        return;
    }
    my $config_set_cache = {  filename => $config_set_file,
                              size => -s $config_set_file,
                              md5  => F5::Utils::File::get_file_md5($config_set_file),
                           };
    my $source_cache_file = F5::Deploy::get_path_by_type('var') . '/config_set_source.yaml';
    if (-e $source_cache_file) {
        my $source_cache = YAML::XS::LoadFile($source_cache_file);
        if (are_hashes_equivalent($config_set_cache, $source_cache)) {
            $self->{logger}->info("Compiled config already loaded from: $config_set_file");
            return;
        }
    }

    my $unzip_dir = "$TMP_DIR/config_set_load.$$";
    make_path($unzip_dir);
    F5::SystemUtils::run_system_tool(command_line_params => ['tar', '-zpx',
                                                             '-f', $config_set_file,
                                                             '-C', $unzip_dir,
                                                            ]);
    find({
          wanted => sub {
              if (-f $File::Find::name) {
                  my $actual_file_location = $File::Find::name;
                  $actual_file_location =~ s/$unzip_dir//;
                  move($File::Find::name, $actual_file_location);
              }
          },
         },
         "$unzip_dir");

    rmtree($unzip_dir);
    YAML::XS::DumpFile($source_cache_file, $config_set_cache);
}

#----------------------------------------------------------------------

sub apply {
    my($self, %params) = @_;

    my $device_config = F5::DeviceConfig->new(dbh => $self->{dbh}, logger => $self->{logger});
    if ($self->{config}) {
        $self->{config}->{device_config} = $device_config->to_hash();
    }
    if ( !$params{force_apply} && $self->{previous_config} && $self->{config} && are_hashes_equivalent($self->{config}, $self->{previous_config}) ) {
        $self->{logger}->info("All configuration matches previous run -- skipping apply")  if $self->{logger};
        return { completed_successfully => 1 };
    }

    $device_config->apply();
    my $rv = F5::SetActive::set_active(dbh => $self->{dbh},
                                       logger => $self->{logger},
                                       do_update_core => 1,
                                       send_sync => 0,
                                       asm_config_handler => $self->{client},
                                       is_daemon => $params{is_daemon},
                                       job => $self->{job},
                                      );
    if ($self->{config}) {
        F5::DbUtils::upsert_singleton(dbh => $self->{dbh}, table_name => "PLC.CONFIG_SET",
           rec => {
             config_set => YAML::XS::Dump($self->{config}),
             last_updated => time(),
           });
    }

    if (!$params{is_daemon}) {
        $self->publish_changes();
    }
    if ($rv) {
        return { completed_successfully => 1 };
    }
}

#----------------------------------------------------------------------

sub publish_changes {
    my ($self, %params) = @_;
    my $logger = $params{logger} || $self->{logger};

    $logger->info("Writing Binary Files");
    my $bd = F5::BdAgent->new(single_run => 1,
                              logger => $logger,
                             );
    $bd->start;
}

#----------------------------------------------------------------------

my %content_format_map = ( "splunk"   => F5::PrepareConf::Base::LOG_REMOTE_STORAGE_SPLUNK(),
                           "arcsight" => F5::PrepareConf::Base::LOG_REMOTE_STORAGE_ARCSIGHT(),
                           "bigiq"    => F5::PrepareConf::Base::LOG_REMOTE_STORAGE_BIGIQ(),
                           "user-defined" => F5::PrepareConf::Base::LOG_REMOTE_STORAGE_REMOTE(),
                           "grpc" => F5::PrepareConf::Base::LOG_REMOTE_STORAGE_GRPC(),
                           "default" => F5::PrepareConf::Base::LOG_REMOTE_STORAGE_REMOTE(),
                         );

my %valid_request_types = ( 'illegal' => 1,
                            'all'     => 1,
                            'all_requests' => 1,
                            'blocked' => 1,
                            'none'    => 1,
                            'illegal-including-staged-signatures' => 1,
                          );

my $MAX_MESSAGE_SIZE = 64 * $K_BYTES;
my $MIN_MESSAGE_SIZE = 1 * $K_BYTES;

sub build_logging_config {
    my ($self, %params) = @_;
    my $policy = $params{policy};
    my $vs_info = $policy->{vs_info} || {};
    my @vs_config;

    my $error_msg;
    my $src_config_line;
    for my $logger_key (keys %$vs_info) {
        my $vs = $vs_info->{$logger_key};
        if ($vs->{logging}) {
            my $vs_name = $vs->{vs_name} || $logger_key;
            my $profile_name = $vs->{profile_name} || "${vs_name}_${logger_key}";
            $src_config_line = delete $vs->{logging}->{src_config_line};
            if (my $log_config_filename = $vs->{logging}->{'$ref'}) {
                my $extra_log_config = $self->load_file(filename => $log_config_filename,
                                                        src_config_line => $src_config_line);
                $vs->{logging}->{filter}  = $extra_log_config->{filter};
                $vs->{logging}->{content} = $extra_log_config->{content};
            }
            my $logging_params = {
                                    logger_type => F5::Mcp::FwLoggerType::FW_LOGGER_TYPE_REMOTE(),
                                    remote_storage => F5::PrepareConf::Base::LOG_REMOTE_STORAGE_REMOTE(),
                                    protocol => F5::Mcp::FwLogRemoteProtocol::FW_LOG_REMOTE_PROTOCOL_TCP(),
                                    facility => F5::Mcp::SyslogFacility::SYSLOG_FACILITY_LOG_LOCAL0(),
                                    report_anomalies => 0,
                                    response_logging => 1, #NONE
                                    logic_operation => 0,
                                    maximum_request_size => 2048,
                                    maximum_header_size => 0,
                                    maximum_query_size => 0,
                                    maximum_entry_length => 2 * $K_BYTES,
                                    address       => $vs->{logging}->{dest_ip} || '127.0.0.1',
                                    port     => $vs->{logging}->{dest_port} || '514',
                                    dest_filename    => $vs->{logging}->{dest_filename},
                                    logging_filter   => $vs->{logging}->{filter},
                                    log_profile_name => $profile_name,
                                    vs_name          => $vs_name,
                                 };

           if ((ref $logging_params->{logging_filter})
               && (my $logging_filter = ($logging_params->{logging_filter}->{request_type} || $logging_params->{logging_filter}->{request_types} || $logging_params->{logging_filter}->{requestType}) )) {
               if (!$valid_request_types{$logging_filter}) {
                   $error_msg = "Invalid request_types: $logging_filter";
                   last;
               }
               if ($logging_filter eq 'none') {
                   next;
               }
               $logging_params->{logging_filter} = $logging_filter;
           }
           if (my $logging_content = $vs->{logging}->{content}) {
               if ($logging_content->{format}) {
                   $logging_params->{remote_storage} = $content_format_map{$logging_content->{format}};
                   if (!$logging_params->{remote_storage}) {
                       $error_msg = "Invalid logging format '$logging_content->{format}'";
                       last;
                   }
                   if ($logging_content->{format} eq 'user-defined') {
                       if (!$logging_content->{format_string}) {
                           $error_msg = "Missing format_string for user-defined format.";
                           last;
                       }
                       $logging_params->{user_string} = $logging_content->{format_string} || $logging_content->{formatString};
                   }
               }
               my $max_request_size = $logging_content->{max_request_size} || $logging_content->{maxRequestSize};
               if (defined($max_request_size) && $max_request_size ne 'any') {
                   if ($max_request_size < 1 || $max_request_size > 2048) {
                       $error_msg = "max_request_size must be between 1 and 2048 inclusive.";
                       last;
                   }
                   $logging_params->{maximum_request_size} = $max_request_size;
               }
               my $max_message_size = $logging_content->{max_message_size} || $logging_content->{maxMessageSize};
               if ($max_message_size) {
                   $logging_params->{maximum_entry_length} = 0;
                   if ($max_message_size =~ /^([0-9]+)(k?)$/i) {
                       my ($bytes, $k) = ($1, $2);
                       $logging_params->{maximum_entry_length} = $bytes;
                       if ($k) {
                           $logging_params->{maximum_entry_length} *= $K_BYTES;
                       }
                   }

                   if ($logging_params->{maximum_entry_length} < $MIN_MESSAGE_SIZE || $logging_params->{maximum_entry_length} > $MAX_MESSAGE_SIZE) {
                       $error_msg = "Invalid max_message_size '$max_message_size'. Must be between 1k-64k.";
                       last;
                   }
               }

               if ($logging_params->{maximum_entry_length} < $logging_params->{maximum_request_size} ) {
                  $error_msg = "Invalid max_message_size '$logging_params->{maximum_entry_length}'. Must be larger than the max_request_size '$logging_params->{maximum_request_size}'";
                  last;
               }

               my $escaping_characters = $logging_content->{escaping_characters} || $logging_content->{escapingCharacters} || [];
               if (! ref($escaping_characters) || reftype($escaping_characters) ne 'ARRAY') {
                   $error_msg = "Expected list for escaping_characters: $escaping_characters.";
                   last;
               }
               if (@$escaping_characters) {
                   $logging_params->{remote_storage_escaping} = [];
                   for my $escaping (@$escaping_characters) {
                       if ( !(defined($escaping->{from}) && defined($escaping->{to})) ) {
                           $error_msg = "Escaping characters must have a 'from' and 'to' defined.";
                           last;
                       }
                       push (@{$logging_params->{remote_storage_escaping}}, { search_for => $escaping->{from}, replace_with => $escaping->{to} } );
                   }
                   last if $error_msg;
               }
               $logging_params->{list_prefix} = $logging_content->{list_prefix} || $logging_content->{listPrefix};
               $logging_params->{list_delimiter} = $logging_content->{list_delimiter} || $logging_content->{listDelimiter};
               $logging_params->{list_suffix} = $logging_content->{list_suffix} || $logging_content->{listSuffix};
           }
           if ($logging_params->{port} !~ /^\d+$/ || $logging_params->{port} < 1 || $logging_params->{port} > 65535) {
               $error_msg = "Invalid logging port '$logging_params->{port}'. Must be numeric and between 1 and 65535.";
               last;
           }
           my $valid_ip_rv = F5::ASMConfig::Entity::Base->normalize_host(field_value => $logging_params->{address}, field_def => {});
           if (!$valid_ip_rv->{completed_successfully}) {
               $error_msg = "Invalid logging destination '$logging_params->{address}'. Must be a valid FQDN or IP address.";
               last;
           }
           if ($valid_ip_rv->{normalized_value}) {
               $logging_params->{address} = $valid_ip_rv->{normalized_value};
           }
  
           $logging_params->{request_type_filter} = $logging_params->{logging_filter} || 'all';
           push (@vs_config, $logging_params);
        }
    }
    if ($error_msg) {
        my $error = {"error_message" => "Logging Configuration Error: $error_msg"};
        if ($src_config_line) {
            $error->{error_line_number} = $src_config_line;
        }
        die $error;
    }
    return \@vs_config;
}

#----------------------------------------------------------------------

sub prepare_all_logging_profiles {
    my ($self, %params) = @_;
    my $dbh    = $params{dbh} || $self->{dbh};
    my $logger = $params{logger} || $self->{logger};
    my $vs_config = dclone($params{vs_config});

    my $batch_insert = F5::BatchInsert->new(dbh => $dbh,
                                            logger => $logger,
                                           );
    my $prepareconf_logging = F5::PrepareConf::LoggingProfile->new(dbh => $dbh,
                                                                   logger => $logger,
                                                                   batch_insert => $batch_insert,
                                                                  );

    my $rv = $prepareconf_logging->prepare(config_set_loggers => $vs_config);
    $batch_insert->flush;
    return $rv;
}

#----------------------------------------------------------------------

sub is_file_json {
    my($self, $filename) = @_;

    open(my $inf, '<', $filename) or die "Failed to read '$filename'\n";
    my $first_part;
    read($inf, $first_part, 50);
    close $inf;

    if ($first_part =~ /^\s*\{/s) {
        return 1;
    }
    return;
}

#----------------------------------------------------------------------

sub load_file {
    my ($self, %params) = @_;
    my $filename = $params{filename};
    my $src_config_line = $params{src_config_line};
    my $rv;
    eval {
        if ($self->is_file_json($filename)) {
            $rv = from_json(File::Slurp::read_file($filename));
        }
        else {
            $rv = YAML::XS::LoadFile($filename);
        }
        if (!ref $rv) {
            die "Bad file format. Expected YAML or JSON.\n";
        }
    };
    if ($@) {
        my $error = {"error_message" => $@};
        if ($src_config_line) {
            $error->{error_line_number} = $src_config_line;
        }
        die $error;
    }
    return $rv;
}

#----------------------------------------------------------------------

sub restore_default_config {
}

#----------------------------------------------------------------------

1;
