package F5::DbUtils;

use strict;
use warnings;
use Carp;

use Memoize;
memoize('get_account_cleanup_tables');
memoize('get_config_sync_tables');
memoize('get_active_policy_tables');
memoize('get_import_policy_tables');
memoize('get_export_policy_flag_tables');
memoize('get_nonpolicy_config_sync_tables');
memoize('get_db_fields');
memoize('get_active_policy_and_config_sync_tables');
memoize('get_export_where_clauses_by_table');
memoize('get_large_record_tables');
memoize('get_task_tables');
memoize('get_ignore_string');
memoize('get_unix_timestamp_string');
memoize('get_now_string');
memoize('get_max_placeholders');

use constant FIND_ONLY_INDEX_FIELDS => 1;
use constant FIND_ONLY_NONINDEX_FIELDS => 2;

my $DEFAULT_CONNECT_ATTEMPT_LIMIT = 3;
my $DEFAULT_CONNECT_ATTEMPT_DELAY = 5;
use F5::Deploy;
use F5::Cfg;
use F5::BigipUtils;
use DBI;
use F5::ProcessTable;
use List::Util;
use F5::BatchInsert;
use F5::Db::Single;
use File::Path;
use File::Copy;
use Scalar::Util qw(reftype);

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(insert_data_to_table foreign_key_scrub get_config_sync_tables);

our $SQL_INIT_FILE_PATH = F5::Deploy::map_path('/ts/install');
our $LOCK_FILE = F5::Deploy::map_path('/ts/lock/ts_db.lock');
my $USERS_CFG_FILE = F5::Deploy::map_path('/etc/ts/common/users.cfg');

sub insert_data_to_table {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $table = $params{table_name};
  my $data = $params{data};
  my $logger = $params{logger};
  my $idmap = $params{idmap};
  my $idmap_strictness = $params{idmap_strictness};
  my $ignore = $params{ignore} ? get_ignore_string($dbh) : "";
  my $log_level_on_missing_idmap = defined $params{log_level_on_missing_idmap} ? $params{log_level_on_missing_idmap} : 'warn';
  my $preserve_id = $params{preserve_id} || 0;
  my $idmap_strictness_callback = $params{idmap_strictness_callback};
  my $field_strictness = $params{field_strictness} || {};
  my $id_field_name;
  my %insert_extra_opts;
  if ($params{ignore}) {
      $insert_extra_opts{do_insert_ignore} = 1;
  }

  my $batch_insert = F5::BatchInsert->new(dbh => $dbh);

  if (exists($params{id_field_name})) {
    $id_field_name = $params{id_field_name};
  }
  else {
    my @db_fields = get_db_fields($dbh, $table, FIND_ONLY_INDEX_FIELDS);
    if (($#db_fields > 0) || !$db_fields[0] || ($db_fields[0] eq 'policy_id')) {
        $id_field_name = '';
    } 
    else {
        $id_field_name = $db_fields[0];
    }
  }

  my %idmap_out;

  # Get field names to insert from DB and data structure

  my @db_fields = get_db_fields($dbh, $table);
  unless (@db_fields) {
    $logger->error("Failed to import policy table '$table'.") if $logger;
    return;
  }

  # get one row from the exported data for fields names
  unless (%$data) {
    $logger->debug("table '$table' is empty") if $logger;
    return {};
  }

  # all used fields anywhere in the dataset
  my %field_list;
  for my $row (values %$data) {
    for my $field (keys %$row) {
      $field_list{$field} = 1;
    }
  }
  my @field_list = keys %field_list;

  @field_list = grep($_ ne $id_field_name, @field_list) if ($id_field_name && !$preserve_id);

  # INSERT DATA

 ROWDATA_LOOP:
  foreach my $rowdata_orig (values %$data) {
    # Copy the original data row to prevent side effect on original data set
    my $rowdata = { %$rowdata_orig };

    # check if any values need substitution of keys
    # NOTE that:
    #    for the substitution of keys, we use the available db fields '@db_fields' only.
    #    while in 'batch_insert->insert_row' (below), we use a list of all used fields anywhere in the dataset '@field_list'.
    foreach my $subst_field (@db_fields) {
      next unless $idmap->{$subst_field};

      if (exists $rowdata->{$subst_field}) {
          if (!defined $rowdata->{$subst_field} &&
              $field_strictness->{$subst_field} &&
              $field_strictness->{$subst_field}->{allow_undef}) {

              next;
          }

          if ( defined $rowdata->{$subst_field} &&
               $field_strictness->{$subst_field}  &&
               $field_strictness->{$subst_field}->{allow_zero} &&
               $rowdata->{$subst_field} == 0 ) {

              next;
          }


          my $val_map = $idmap->{$subst_field};
          if (exists $val_map->{'*'}) {
              $rowdata->{$subst_field} = $val_map->{'*'};
          }
          elsif (defined $rowdata->{$subst_field} && exists $val_map->{$rowdata->{$subst_field}}) {
              $rowdata->{$subst_field} = $val_map->{$rowdata->{$subst_field}};
          }
          elsif ($idmap_strictness) {
              my $row_id = '';
              if ($id_field_name) {
                  $row_id = $rowdata->{$id_field_name};
              }

              if (defined $idmap_strictness_callback) {
                  $idmap_strictness_callback->(row_id => $row_id, 
                                               table => $table, 
                                               logger => $logger,
                                               subst_field => $subst_field, 
                                               rowdata => $rowdata);
              }
              else {
                  if ($log_level_on_missing_idmap) {
                      $logger->$log_level_on_missing_idmap("Row $row_id of table $table is missing $subst_field ($rowdata->{$subst_field}) -- skipping\n" . caller()) if $logger;
                  }
              }
              next ROWDATA_LOOP;
          }
      }
    }

    my $old_id = $rowdata->{$id_field_name} if ($id_field_name);

    my $new_obj;
    $batch_insert->insert_row(rec => $rowdata,
                              table_name => $table,
                              fields => \@field_list,
                              get_deferred_id => \$new_obj,
                              %insert_extra_opts,
                             );

    # map to new id if the table is a parent
    if ($params{idmap_save}) {
        my ($new_id) = $preserve_id ? $old_id : $new_obj;
        $idmap_out{$old_id} = $new_id;
    }
  }

  $batch_insert->flush;
  for my $k (keys %idmap_out) {
      $idmap_out{$k} = ref ($idmap_out{$k}) ? $idmap_out{$k}->get_value : $idmap_out{$k};
  }

  return \%idmap_out;
}

#----------------------------------------------------------------------

sub get_db_fields {
  my($dbh, $table, $get_index, $as_hashref) = @_;

  my $dbname = undef;
  if ($table =~ /^(.*?)\.(.*?)$/) {
    $dbname = $1;
    $table = $2;
  }

  my $sth = $dbh->column_info(undef, $dbname, $table, '%');
  $sth->execute;
  my %primary_fields;
  if (defined($get_index)) {
      my @names = $dbh->primary_key(undef, $dbname, $table);
      %primary_fields = map { ($_ => 1) } @names;
  }
  my $retval = {};
  my $is_sqlite = is_sqlite($dbh);
  while(my $row = $sth->fetchrow_hashref) {
    if (!defined($get_index) || 
        ($get_index == FIND_ONLY_NONINDEX_FIELDS && !$primary_fields{$row->{COLUMN_NAME}}) ||
        ($get_index == FIND_ONLY_INDEX_FIELDS && $primary_fields{$row->{COLUMN_NAME}})) {

        my $default = $row->{COLUMN_DEF};
        if ($is_sqlite && $default) {
            #Remove erroneous quotes around string defaults in sqlite
            $default =~ s/'//g;
            if ($default eq 'NULL') {
                undef $default;
            }

        }

        $retval->{$row->{COLUMN_NAME}} = { column_name => $row->{COLUMN_NAME},
                                           nullable => $row->{NULLABLE},
                                           default  => $default,
                                         };
    }
  }
  if ($as_hashref) {
      return $retval;
  }
  else {
      return keys %$retval;
  }
}

#----------------------------------------------------------------------

sub foreign_key_scrub {
  my($dbh, $dbname, $logger, $foreign_keys_map) = @_;

  my @foreign_keys = @$foreign_keys_map;
  for my $foreign_key (@foreign_keys) {
    my($child_table, $child_field, $parent_table, $parent_field) = ($foreign_key =~ /^(.*?)\.(.*?)=>(.*?)\.(.*?)$/);
    $logger->debug("Checking for illegal references from $child_table.$child_field to $parent_table.$parent_field") if $logger;
    my $find_unmatched_sth = $dbh->prepare("SELECT CHILD.id AS child_id " .
#					   "CHILD.policy_id AS child_policy_id " .
#					   "PARENT.policy_id AS parent_policy_id " .
					   "FROM $dbname.$child_table AS CHILD LEFT OUTER JOIN $dbname.$parent_table AS PARENT " .
					   "ON CHILD.$child_field = PARENT.$parent_field " .
					   "WHERE PARENT.$parent_field IS NULL " . 
					   "AND CHILD.$child_field != 0 ") or
#					   "OR PARENT.policy_id != CHILD.policy_id") or # todo: check that child and parent are in the same policy
					     return _db_error($logger, "Statement prepare failed");

    $find_unmatched_sth->execute or
      return _db_error($logger, "Statement execute failed");
    my $unmatched_ids = $find_unmatched_sth->fetchall_arrayref({});
    for my $row (@$unmatched_ids) {
      if ($row->{parent_policy_id}) {
	# Deleting due to differing policy ids
	$logger->info("Deleting $child_table row '$row->{child_id}' (differing policies: $child_table policy = $row->{child_policy_id}; $parent_table policy = $row->{parent_policy_id}") if $logger;
      } else {
	$logger->info("Deleting $child_table row '$row->{child_id}' (no match found in $parent_table)") if $logger;
      }
      $dbh->do("DELETE FROM $dbname.$child_table WHERE id = ?", undef, $row->{child_id}) or 
	return _db_error($logger, "Delete failed ($child_table, id $row->{child_id}");
    }
  }
  return 1;
}

#----------------------------------------------------------------------

sub _db_error {
  my($logger, $msg) = @_;
  if ($logger && $msg) {
    $logger->error($msg . " - $DBI::errstr");
  }
  return;
}

#----------------------------------------------------------------------

sub get_config_sync_tables {
  my($dbname) = @_;
  return grep_from_sql_file($dbname, 'CONFIG_SYNC:');
}
#----------------------------------------------------------------------

sub get_non_config_sync_tables {
  my($dbname) = @_;
   unless ($dbname) {
    return (get_non_config_sync_tables('DCC'), get_non_config_sync_tables('PLC'));
  }
  
  my @non_config_sync_tables = (); 

  my @all_tables = get_all_tables($dbname);
  my @config_sync_tables = get_config_sync_tables($dbname);
   
  my %config_sync_tables = map { ($_ => 1) } @config_sync_tables;

  @non_config_sync_tables = grep { !$config_sync_tables{$_} } @all_tables;

  return @non_config_sync_tables;
}
#----------------------------------------------------------------------
sub get_all_tables {
  my($dbname) = @_;

  my @tables;
  my $compiled_pattern = qr/CREATE TABLE (\w+)/;
  open(my $inf, get_sql_filename($dbname));
  while(my $line = <$inf>) {
    chomp($line);
    if ($line =~ /$compiled_pattern/) {
      push(@tables, "$dbname.$1");
    }
  }
  close $inf;

  return @tables;
}

#----------------------------------------------------------------------
sub get_asmqkview_exclude_tables {
  my($dbname) = @_;
  return grep_from_sql_file($dbname, 'ASMQKVIEW_EXCLUDE:');
}

#----------------------------------------------------------------------
sub get_asmqkview_include_tables {
  my($dbname) = @_;
  
  my @asmqkview_include_tables = (); 

  my @all_tables = get_all_tables($dbname);
  my @asmqkview_exclude_tables = get_asmqkview_exclude_tables($dbname);
   
  my %asmqkview_exclude_tables = map { ($_ => 1) } @asmqkview_exclude_tables;

  @asmqkview_include_tables = grep { !$asmqkview_exclude_tables{$_} } @all_tables;

  return @asmqkview_include_tables;
}

#----------------------------------------------------------------------
sub get_recover_stat_tables {
  my($dbname) = @_;
  return grep_from_sql_file($dbname, 'RECOVER_STAT:');
}
#----------------------------------------------------------------------
sub get_event_backup_tables {
  my($dbname) = @_;
  return grep_from_sql_file($dbname, 'EVENT_BACKUP:');
}
#----------------------------------------------------------------------
sub get_recover_clear_tables {
  my($dbname) = @_;
  return grep_from_sql_file($dbname, 'RECOVER_CLEAR:');
}
#----------------------------------------------------------------------

sub get_account_cleanup_tables {
  my($dbname) = @_;
  return grep_from_sql_file($dbname, 'ACCOUNT_CLEANUP:');
}

#----------------------------------------------------------------------

sub get_active_policy_tables {
  my($dbname) = @_;
  return grep_from_sql_file($dbname, 'ACTIVE_POLICY:');
}

#----------------------------------------------------------------------

sub get_import_policy_tables {
  my($dbname) = @_;
  return grep_from_sql_file($dbname, 'POLICY_IMPORT:');
}

#----------------------------------------------------------------------

sub get_task_tables {
  my($dbname) = @_;
  return grep_from_sql_file($dbname, 'TSK_TABLE:');
}

#----------------------------------------------------------------------

sub get_large_record_tables {
  my($dbname) = @_;
  return grep_from_sql_file($dbname, 'LARGE_RECORD_TABLE:');
}

#----------------------------------------------------------------------

sub get_export_policy_flag_tables {
  my($dbname) = @_;

  my %export_table_flags;

  my @lines = grep_from_sql_file($dbname, 'POLICY_EXPORT_FLAG:', { include_parens => 1 });
  for my $line (@lines) {
      my($table, $flag) = ($line =~ /^([\w.]+)\((.*?)\)/);
      $export_table_flags{$table} = $flag;
  }

  return %export_table_flags;
}

#----------------------------------------------------------------------

sub get_export_where_clauses_by_table {
  my($dbname) = @_;

  my %export_table_where_clauses;

  my @lines = grep_from_sql_file($dbname, 'EXPORT_WHERE_CLAUSE:', { include_quotes => 1 });
  for my $line (@lines) {
      my($table, $where_clause) = ($line =~ /^([\w.]+)\s*"(.*?)"/);
      $export_table_where_clauses{$table} = $where_clause;
  }

  return %export_table_where_clauses;

}

#----------------------------------------------------------------------

sub get_nonimport_policy_tables {
  my($dbname) = @_;
  return grep_from_sql_file($dbname, 'POLICY_NONIMPORT:');
}

#----------------------------------------------------------------------

sub get_cluster_tables {
  my($dbname) = @_;
  return grep_from_sql_file($dbname, 'CLUSTER_TABLE:');
}


#----------------------------------------------------------------------

sub get_pb_config_tables {
   my($dbname) = @_;
   return grep_from_sql_file($dbname, 'POLICY_BUILDER:'); 
}
#----------------------------------------------------------------------

sub get_nonpolicy_config_sync_tables {
  my($dbname) = @_;
  unless ($dbname) {
    return (get_nonpolicy_config_sync_tables('DCC'), get_nonpolicy_config_sync_tables('PLC'));
  }

  my %import_policy_tables = map { ($_ => 1) } get_import_policy_tables($dbname);
  return grep(!exists $import_policy_tables{$_}, get_config_sync_tables($dbname));
}

#----------------------------------------------------------------------
sub get_active_policy_and_config_sync_tables{
  my($dbname) = @_;

  my $include_traffic_data = defined $ENV{'INCLUDE_TRAFFIC_DATA'}? $ENV{'INCLUDE_TRAFFIC_DATA'} : 0;
  my @rv;
  unless ($dbname) {
    @rv =  (get_config_sync_tables('DCC'), get_config_sync_tables('PLC'),get_active_policy_tables('DCC'));
  }else{
    @rv =  (get_config_sync_tables($dbname),get_active_policy_tables($dbname));
  }

  if (! $include_traffic_data){
    @rv = grep {! ($_ eq 'PLC.NEGSIG_ACCEPT_LOG')} @rv;
  }

  return @rv;

}

#----------------------------------------------------------------------
# TODO - grep_from_sql_file 
sub get_request_log_upgrade_tables {
  return qw(PRX.REQUEST_LOG
            PRX.REQUEST_LOG_CLEARED
            PRX.REQUEST_LOG_PROPERTIES
            PRX.REQUEST_LOG_SYNC_FILES
           );
}

#----------------------------------------------------------------------
# TODO - grep_from_sql_file 
sub get_learning_suggestions_upgrade_tables {
  return qw(PLC.PL_SUGGESTIONS
            PLC.PL_SUGGESTION_SUPPORT_IDS
            PLC.PL_SUGGESTION_VIOLATION_RATING_COUNTS
           );
}

#----------------------------------------------------------------------

sub get_table_schema {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $table_name = $params{table_name};
  my $strip_comments = $params{strip_comments};

  my($create_table, $create_sql) = $dbh->selectrow_array("SHOW CREATE TABLE $table_name");

  if ($strip_comments) {
    $create_sql =~ s/\/\*.*?\*\///s;
  }

  return $create_sql;
}

#----------------------------------------------------------------------

sub select_into_file {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $table_name = $params{table_name};
  my $file_name = $params{file_name};

  my $rv = $dbh->do(<<EndQuery);
SELECT *
  FROM $table_name
  INTO OUTFILE '$file_name'
  FIELDS TERMINATED BY ',' OPTIONALLY ENCLOSED BY '"'
EndQuery

  return $rv;
}

#----------------------------------------------------------------------

sub load_from_file {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $table_name = $params{table_name};
  my $file_name = $params{file_name};

  return unless -f $file_name;

  my $rv = $dbh->do(<<EndQuery);
LOAD DATA INFILE '$file_name'
  INTO TABLE $table_name
  FIELDS TERMINATED BY ',' OPTIONALLY ENCLOSED BY '"'
EndQuery

  return $rv;
}

#----------------------------------------------------------------------
sub grep_from_sql_file {
  my($dbname, $pattern, $params) = @_;

  return unless $dbname;
  my @tables;
  my $compiled_pattern;
  if ($params && $params->{include_parens}) {
      $compiled_pattern = qr/$pattern \s* ([\w\.\(\)]+)/x;
  }
  elsif ($params && $params->{include_quotes}) {
      $compiled_pattern = qr/$pattern \s* ([\w\.]+\s*".*")/x;
  }
  else {
    $compiled_pattern = qr/$pattern \s* ([\w\.]+)/x;
  }
  open(my $inf, get_sql_filename($dbname));
  while(my $line = <$inf>) {
    chomp($line);
    if ($line =~ /$compiled_pattern/) {
      push(@tables, $1);
    }
  }
  close $inf;

  return @tables;
}

#----------------------------------------------------------------------

sub get_sql_filename {
  my($dbname) = @_;

  return $SQL_INIT_FILE_PATH . '/'. lc $dbname . '.sql';
}

#----------------------------------------------------------------------

sub get_dbh {
  my(%arguments) = @_;

  my $dbh;
  my $with_retry = defined $arguments{with_retry} ? $arguments{with_retry} 
                                                  : 1;

  my $connect_attempt_limit = $arguments{connect_attempt_limit} || 0;
  my $connect_attempt_delay = $arguments{connect_attempt_delay} || 1;

  # Shortcut for unified retry limits (this may become the default in the future)
  if ($with_retry) {
      $connect_attempt_limit = defined $arguments{connect_attempt_limit} ? $arguments{connect_attempt_limit} 
                                                                         : $DEFAULT_CONNECT_ATTEMPT_LIMIT;
      $connect_attempt_delay = defined $arguments{connect_attempt_delay} ? $arguments{connect_attempt_delay} 
                                                                         : $DEFAULT_CONNECT_ATTEMPT_DELAY;
  }

  my $connect_attempt_count = 0;
  my $last_exception;

  while (1) {
      my $user_cfg_section = $arguments{user_cfg_section} ||
                             F5::BigipUtils::is_device_bigip() ? 'DATABASE' : 'EM_DATABASE';
      
      my $database_username = F5::BigipUtils::is_docker() ? 'root'
                                                          : $arguments{database_username} || get_mysql_user(section => $user_cfg_section);

      my $database_password = $arguments{database_password} || get_mysql_password(user => $database_username);

      my $database_type = F5::BigipUtils::is_docker() ? "SQLite" 
                                                      : $arguments{database_type} || 
                              F5::Cfg::cfg_get_config_item($USERS_CFG_FILE, $user_cfg_section, 'Driver');
      my $database_name = $arguments{database_name} ||
                          F5::Cfg::cfg_get_config_item($USERS_CFG_FILE, $user_cfg_section, 'Name');

      my $dsn = "dbi:$database_type:$database_name";
      my $raise_error = $arguments{raise_error} || 0;
      my $db_opts = $arguments{db_opts} || {};
    
      # Shortcut for using exceptions (this may become the default in the future)
      if ($arguments{with_exception}) {
          $db_opts->{RaiseError} = 1;
          $db_opts->{PrintError} = 0;
      }
      if ($database_type eq 'SQLite') {
          $db_opts->{sqlite_allow_multiple_statements} = 1;
          $db_opts->{RaiseError} = 1;
          $db_opts->{PrintError} = 0;

          my $database_dir = F5::Deploy::get_path_by_type("database") || '';
          $dsn = "dbi:$database_type:$database_dir/$database_name";
      }

      my $database_hostname = defined($arguments{host}) ? $arguments{host}
                               : F5::Cfg::cfg_get_config_item($USERS_CFG_FILE, $user_cfg_section, 'Hostname');

      if($database_hostname) {
	if($database_hostname eq 'localhost') {
            $database_hostname = '127.0.0.1';
        }
        $dsn .= ":$database_hostname";
        
        if(defined $arguments{port}) {
            $dsn .= ":$arguments{port}"; # default 3306
        }
      }

      if ($arguments{dsn_options}) {
          $dsn .= ";$arguments{dsn_options}";
      }

      $connect_attempt_count++;

      my $check_dbh_rv;
      eval {
          $last_exception = undef;
          $dbh = DBI->connect( $dsn, $database_username, $database_password,
                               {RaiseError => $raise_error,
                                Callbacks => {
                                    begin_work => sub { my ($dbh) = @_; 
                                                        $dbh->{private_f5_callbacks} = [];
                                                        $dbh->{private_f5_callbacks_by_name} = {};
                                                        return;},
                                    rollback   => sub { my ($dbh) = @_;
                                                        $dbh->{private_f5_callbacks} = [];
                                                        $dbh->{private_f5_callbacks_by_name} = {};
                                                        return;},
                                    commit     => sub { my ($dbh) = @_;
                                                        for my $callback (@{$dbh->{private_f5_callbacks}}) {
                                                            $callback->();
                                                        }
                                                        return;},
                                },
                                %$db_opts,
                               } );
          $check_dbh_rv = check_dbh($dbh);
          if ($database_type eq 'SQLite') {
              sqlite_init($dbh);
          }
      };
      if ($@) {
          $last_exception = $@;
      }
      if ($check_dbh_rv || $connect_attempt_count >= $connect_attempt_limit) {
          last;
      }
      sleep $connect_attempt_delay;
  }

  if (defined $last_exception) {
      die $last_exception;
  }

  return $dbh;
}

#---------------------------------------------------------------------
sub get_primary_dbh {
  my(%arguments) = @_;
   if (defined $arguments{master_blade}){

   eval "use PassCrypt;";
   die "Can't load PassCrypt" if ($@);

    my $slot_dbh = get_dbh();

    my $primary_database_password  = $slot_dbh->selectrow_array(<<EndQuery);
SELECT mysql_password
 FROM  CONF_PRIMARY_BLADE_RATE_LIMIT;
EndQuery

    $primary_database_password = PassCrypt::decrypt_password("$primary_database_password");
    my $host = "127.3.0.$arguments{master_blade}";

    my $primary_dbh = get_dbh( host => $host,
                               database_password =>  $primary_database_password, );
  }
}
#----------------------------------------------------------------------

sub refresh_dbh {
  my(%params) = @_;

  my $dbh = $params{dbh};

  return $dbh if (check_dbh($dbh));

  $dbh = F5::DbUtils::get_dbh(%params);
  if(check_dbh($dbh)) {
    $params{logger}->info("successful connect to database") if $params{logger};
    return $dbh;
  }

  return;
}

#---------------------- ----------------------------------------------------------

sub check_dbh {
  my($dbh) = @_;
  return ((defined $dbh) && ($dbh->ping()));
}

#---------------------- ----------------------------------------------------------

sub check_privileges {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $required_privileges = $params{required_privileges};

  return 0 if (!check_dbh($dbh));
  
  # In future, more advanced check can be implemented by select from information_schema.USER_PRIVILEGES and SCHEMA_PRIVILEGES
  my $user_privileges = $dbh->selectcol_arrayref('SHOW GRANTS');

  if ($user_privileges->[0] && $user_privileges->[0] =~ /^GRANT ALL PRIVILEGES ON \*\.\*/) { # like root
      return 1;
  }

  for my $required_privilege (@$required_privileges) {
      if (!grep { $_ =~ /$required_privilege/ } @$user_privileges) {
          return 0;
      }
  }

  return 1;
}

#----------------------------------------------------------------------

sub copy_defaults_table {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $policy_id = $params{policy_id};
  my $dbname = $params{dbname};
  my $table_name = $params{table_name};
  my $defaults_table_name = $params{defaults_table_name} || $params{table_name} . '_DEFAULTS';
  my $ignore = $params{ignore} ? get_ignore_string($dbh) : "";

  if ($table_name =~ s/^(.*)\.//) {
    $dbname ||= $1;
  }

  if(! defined $dbname) {
      $dbname = 'PLC';
  }

  $table_name = "$dbname.$table_name";
  $defaults_table_name =~ s/^.*?\.//;
  $defaults_table_name = "$dbname.$defaults_table_name";

  my %fields = map(($_ => 1), F5::DbUtils::get_db_fields($dbh, $table_name));
  my %default_fields = map(($_ => 1), F5::DbUtils::get_db_fields($dbh, $defaults_table_name));

  for my $k (keys %fields) {
    delete $fields{$k} unless $default_fields{$k};
  }
  delete $fields{'id'};

  my @insert_fields = keys %fields;
  my @select_fields = map("defaults_table.$_", @insert_fields);
  my @tables = ("$defaults_table_name defaults_table");
  my @where_clauses;
  my @bind_values;

  if ($params{insert_values}) {
    for my $field (keys %{$params{insert_values}}) {
      push(@insert_fields, $field);
      push(@select_fields, '?');
      push(@bind_values, $params{insert_values}->{$field});
    }
  }

  if ($params{where_values}) {
    for my $field (keys %{$params{where_values}}) {
      push(@where_clauses, "defaults_table.$field = ?");
      push(@bind_values, $params{where_values}->{$field});
    }
  }

  if ($policy_id) {
    push(@tables, "$dbname.PL_POLICIES pol");
    push(@select_fields, "pol.id");
    push(@insert_fields, "policy_id");
    if ($policy_id ne '*') {
      push(@where_clauses, "pol.id = ?");
      push(@bind_values, $policy_id);
    }
  }

  push(@where_clauses, '1 = 1') unless @where_clauses;
  my $where_clauses = join(' AND ', @where_clauses);

  local $" = ',';
  $dbh->do(<<EndQuery, undef, @bind_values);
INSERT $ignore INTO $table_name (@insert_fields)
     SELECT @select_fields
       FROM @tables
      WHERE $where_clauses
EndQuery

}

#----------------------------------------------------------------------
my $UNREFERENCED_FILE_EXPIRY_TIME = 60 * 5; #5 minutes.

sub delete_unreferenced_files {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $logger = $params{logger};
  my $db_file_refs = $params{db_file_refs};
  my $dir = $params{dir};

  my %db_files;
  for my $db_file_ref (@$db_file_refs) {
    my $db_file_ref_files = $dbh->selectcol_arrayref(<<EndQuery);
SELECT $db_file_ref->{field}
  FROM $db_file_ref->{table_name}
EndQuery

    # grep to prevent 'Use of uninitialized value in list assignment'
    #    at SELECT grace_conf_file FROM DCC.DATA_PROTECTION
    %db_files = (%db_files,
                 map { ($_ => 1) } grep { $_ } @$db_file_ref_files);
  }

  if ($params{db_value_to_filename_conversion_func}) {
    %db_files = map { ($params{db_value_to_filename_conversion_func}->($_) => 1) } keys %db_files;
  }

  my $marked_files = $dbh->selectall_hashref(<<EndQuery, 'filename', undef, $dir);
SELECT *
  FROM DCC.FILE_REFERENCES_FOR_DELETION
  WHERE dir = ?
EndQuery

  my $time = time();
  opendir(my $dir_h, $dir);
  while (my $fs_filename = readdir($dir_h)) {
    next if ($fs_filename =~ /^\.+$/);
    
    $fs_filename = "$dir/$fs_filename";

    if($db_files{$fs_filename}) {
        if ($marked_files->{$fs_filename}) {
            #File is still in use. Remove delete mark.
            delete_row(dbh => $dbh,
                       table_name => 'DCC.FILE_REFERENCES_FOR_DELETION',
                       rec => {filename => $fs_filename});
        }
    }
    else {
        if ($marked_files->{$fs_filename}) {
            if (($time - $marked_files->{$fs_filename}->{delete_time}) > $UNREFERENCED_FILE_EXPIRY_TIME) { 
                if (-d $fs_filename) {
                    $logger->debug("Deleting unused directory $fs_filename") if $logger;
                    rmtree $fs_filename;
                }
                else {
                    $logger->debug("Deleting unused file $fs_filename") if $logger;
                    unlink $fs_filename;
                }
                delete_row(dbh => $dbh,
                           table_name => 'DCC.FILE_REFERENCES_FOR_DELETION',
                           rec => {filename => $fs_filename});
            }
        }
        else {
            insert_row(dbh => $dbh,
                       table_name => 'DCC.FILE_REFERENCES_FOR_DELETION',
                       ignore => 1,
                       rec => {filename => $fs_filename, dir => $dir, delete_time => $time });
        }
    }
    delete $marked_files->{$fs_filename};
  }
  closedir($dir_h);

  #Cleanup orphaned rows
  for my $filename (keys %$marked_files) {
      delete_row(dbh => $dbh,
                 table_name => 'DCC.FILE_REFERENCES_FOR_DELETION',
                 rec => {filename => $filename});
  }

}

#----------------------------------------------------------------------
sub delete_row {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $table_name = $params{table_name};
  my $rec = $params{rec};
  my $fields = $params{fields};

  local $" = ',';
  unless ($fields) {
      $fields = determine_db_fields(row => $rec,
                                    table_name => $table_name,
                                    dbh => $dbh);
  }

  my @conditions = map { "$_ = ?" } @$fields;
  my @where_values = @{$rec}{@$fields};

  my $where_clauses = ' WHERE ' . join(' AND ', @conditions);
  my $rv;
  eval {
        $rv = $dbh->do("DELETE FROM  $table_name $where_clauses",undef,@where_values);
  };

  if ($@) {
      my $eval_error = $@;
      $eval_error =~ s/\s+ at \s+ \S+ \s+ line \s+ \d+ .*$//sx;
      die "Failed on delete from $table_name ($eval_error)\n";
  }

}

#----------------------------------------------------------------------

sub delete_rows {
    my (%params) = @_;

    my $dbh = $params{dbh};
    my $where_fields = $params{where_fields};
    my $table_name = $params{table_name};

    my $constructed_where_clause = construct_where_clause(where_fields => $where_fields);
    my $where_clause = $constructed_where_clause->{where_clause};
    my @where_values = @{$constructed_where_clause->{where_values}};

    my $rv;
    eval {
        $rv = $dbh->do("DELETE FROM $table_name WHERE $where_clause",undef,@where_values);
    };
    if (my $eval_error = $@) {
        $eval_error =~ s/\s+ at \s+ \S+ \s+ line \s+ \d+ .*$//sx;
        die "Failed on delete from $table_name ($eval_error)\n";
    }
}

#----------------------------------------------------------------------

sub insert_row {
  my(%params) = @_;

  my $dbh = $params{dbh} || F5::Db::Single->instance;
  my $table_name = $params{table_name};
  my $fields = $params{fields};
  my $rec = $params{rec};
  my $single_row_insert = 0;
  my $dup_key_fields = $params{on_duplicate_key_update_fields};
  my $do_replace = $params{do_replace};
  my $ignore = ($params{ignore} ? get_ignore_string($dbh) : '');
  my $set_date_fields_to_now = $params{set_date_fields_to_now};
  my $set_date_fields_to_now_unix = $params{set_date_fields_to_now_unix};
  my $last_insert_id_retval = $params{last_insert_id};

  if ($dup_key_fields && ref($rec) eq 'ARRAY') {
    die "Cannot do 'ON DUPLICATE KEY UPDATE' with a multi-row insert\n";
  }

  if (ref($rec) ne 'ARRAY') {
    # single row insert
    $rec = [ $rec ];
    $single_row_insert = 1;
  }

  local $" = ',';

  unless ($fields) {
      $fields = determine_db_fields(row => $rec->[0],
                                    table_name => $table_name,
                                    dbh => $dbh);
  }
  my @single_row_placeholders = ('?') x scalar @$fields;

  my @dup_key_placeholders = ();
  my @dup_key_values = ();
  if (ref($dup_key_fields) eq 'ARRAY') {
    my $dup_key_fields = determine_db_fields(row => { map { $_ => 1 } @$dup_key_fields },
                                             table_name => $table_name,
                                             dbh => $dbh);
    for my $field (@$dup_key_fields) {
        push(@dup_key_values, $rec->[0]->{$field});
        push(@dup_key_placeholders, "$field = ?");
    }
  }

  if ($set_date_fields_to_now) {
    my $now_string = get_now_string(dbh => $dbh);
    push(@$fields, @$set_date_fields_to_now);
    push(@single_row_placeholders, ($now_string) x scalar @$set_date_fields_to_now);
  }

  if ($set_date_fields_to_now_unix) {
    push(@$fields, @$set_date_fields_to_now_unix);
    push(@single_row_placeholders, (get_unix_timestamp_string(dbh => $dbh)) x scalar @$set_date_fields_to_now_unix);
  }

  my @all_row_placeholders = ('(' . join(',', @single_row_placeholders) . ')') x scalar @$rec;

  if(!@all_row_placeholders) {
     if($single_row_insert) { # Called with a hash - default row insertion.
         @all_row_placeholders = ('()');
         $fields = [];
     }
     else { # Called with an empty array - do not insert anything.
         return;
     }
  }

  my $rv;
  eval {
      if ($do_replace) {
          $rv = $dbh->do("REPLACE INTO $table_name (@$fields) VALUES @all_row_placeholders",
                         undef, map { @{$_}{@$fields} } @$rec);
      }
      else {
          if (@dup_key_placeholders) {
              if (is_sqlite($dbh)) {
                  $rv = $dbh->do("REPLACE INTO $table_name (@$fields) VALUES @all_row_placeholders",
                                 undef, map { @{$_}{@$fields} } @$rec);
              }
              else {
                  my @values = map { @{$_}{@$fields} } @$rec;
                  $rv = $dbh->do("INSERT $ignore INTO $table_name (@$fields) VALUES @all_row_placeholders ON DUPLICATE KEY UPDATE " . join(',',@dup_key_placeholders),
                                 undef, (@values , @dup_key_values) );
              }
          }
          else {
              $rv = $dbh->do("INSERT $ignore INTO $table_name (@$fields) VALUES @all_row_placeholders",
                         undef, map { @{$_}{@$fields} } @$rec);
          }
      }

      if ($last_insert_id_retval && defined $rv && $rv ne "0E0") {
          my($last_insert_id) = $dbh->last_insert_id(undef, undef, undef, undef);
          if (is_sqlite($dbh)) {
              $last_insert_id -= $#$rec;
          }
          $$last_insert_id_retval = $last_insert_id;
      }
  };
  if ($@) {
      my $eval_error = $@;
      $eval_error =~ s/\s+ at \s+ \S+ \s+ line \s+ \d+ .*$//sx;
      die "Failed on insert to $table_name ($eval_error)\n";
  }

  return $rv;
}

#----------------------------------------------------------------------

sub determine_db_fields {
    my(%params) = @_;

    my $db_fields;
    if ($params{dbh} && ref($params{dbh}) !~ /::CSV/) {
        $db_fields = get_db_fields($params{dbh}, $params{table_name}, undef, 1);
    }
    return [ grep { ! $db_fields || $db_fields->{$_} } keys %{$params{row}} ];
}

#----------------------------------------------------------------------

sub update_row {
  my(%params) = @_;

  my $dbh = $params{dbh} || F5::Db::Single->instance;
  my $table_name = $params{table_name};
  my $fields = $params{fields};
  my $rec = $params{rec};
  my $set_date_fields_to_now = $params{set_date_fields_to_now};
  my $set_date_fields_to_now_unix = $params{set_date_fields_to_now_unix};
  my $increment_fields = $params{increment_fields};
  my $where_fields = $params{where_fields};

  local $" = ',';
  unless ($fields) {
      $fields = determine_db_fields(row => $rec,
                                    table_name => $table_name,
                                    dbh => $dbh);
  }

  my @update_clauses = map { "$_ = ?" } @$fields;
  my @update_values = @{$rec}{@$fields};

  if ($set_date_fields_to_now) {
    my $now_string = get_now_string(dbh => $dbh);
    push(@update_clauses, map { "$_ = $now_string" } @$set_date_fields_to_now);
  }

  if ($set_date_fields_to_now_unix) {
    my $unix_timestamp_string = get_unix_timestamp_string(dbh => $dbh);
    push(@update_clauses, map { "$_ = $unix_timestamp_string" } @$set_date_fields_to_now_unix);
  }

  if ($increment_fields) {
    push(@update_clauses, map { "$_ = $_ + 1" } @$increment_fields);
  }

  return 0 unless @update_clauses;

  my $constructed_where_clause = construct_where_clause(where_fields => $where_fields);
  my $where_clause = $constructed_where_clause->{where_clause};
  my @where_values = @{$constructed_where_clause->{where_values}};

  my $rv = $dbh->do("UPDATE $table_name SET @update_clauses WHERE $where_clause",
                    undef, @update_values, @where_values);

  return (defined $rv && $rv ne '0E0') ? $rv
                                       : 0;
}

#----------------------------------------------------------------------

sub construct_where_clause {
    my (%params) = @_;
    my $where_fields = $params{where_fields};
    my $literal_where_fields = $params{literal_where_fields};
    my(@where_field_names, @where_values);
    my @where_clauses = (1);

    if ($where_fields) {
        for my $where_field (keys %$where_fields) {
            my $values = $where_fields->{$where_field};
            my $negative_comparison = 0;
            if (ref $values && reftype($values) eq 'HASH') {
                if ($values->{comparison} eq 'ne') {
                    $negative_comparison = 1;
                }
                $values = $values->{values};
            }

            my $field_comparison;
            if (ref $values) {
                my $placeholders = join(',', ('?') x scalar @{$values});
                if ($placeholders) {
                    $field_comparison = $negative_comparison ? 'NOT IN' : 'IN';
                    push (@where_clauses, "$where_field $field_comparison ($placeholders)");
                    push (@where_values, @{$values});
                }
                else {
                     push (@where_clauses, $negative_comparison);
                     last;
                }
            }
            else {
                $field_comparison = $negative_comparison ? '!=' : '=';
                push (@where_clauses, "$where_field $field_comparison ?");
                push (@where_values, $where_fields->{$where_field});
            }
        }
    }

    if ($literal_where_fields) {
        for my $literal_where_field (keys %$literal_where_fields) {
            push (@where_clauses, "$literal_where_field = $literal_where_fields->{$literal_where_field}");
        }
    }

    my $where_clause = join(' AND ', @where_clauses);
    return {where_clause => $where_clause,
            where_values => \@where_values};    
}

#----------------------------------------------------------------------

sub find_lowest_unused_id {
    my (%params) = @_;
    my $dbh = $params{dbh};
    my $table_name = $params{table_name};
    my $id_name = $params{id_name};
    my $where_values = $params{where_values};

    my @join_clauses;
    my @join_bind_values;
    my @where_clauses;
    my @where_bind_values;
    for my $k (keys %$where_values) {
        push(@join_clauses, "A.$k = ?");
        push(@join_bind_values, $where_values->{$k});
        push(@join_clauses, "B.$k = ?");
        push(@join_bind_values, $where_values->{$k});
        push(@where_clauses, "A.$k = ?");
        push(@where_bind_values, $where_values->{$k});
    }
    if (defined $params{min_id_value}) {
        my @min_value_where_clauses = @where_clauses;
        push (@min_value_where_clauses, "A.$id_name = ?");
        my $min_value_where_clause = join(' AND ', @min_value_where_clauses);
      
        my ($min_exists) = $dbh->selectrow_array(<<EndQuery, undef,  @where_bind_values, $params{min_id_value});
SELECT COUNT(*)
    FROM $table_name A
  WHERE $min_value_where_clause
EndQuery

        if (!$min_exists) {
            return $params{min_id_value};
        }

        push(@where_clauses, "A.$id_name >= ?");
        push(@where_bind_values, $params{min_id_value});
    }

    my $join_clauses = '';
    if (@join_clauses) {
        $join_clauses = ' AND ' . join(' AND ', @join_clauses);
    }
    my $where_clauses = '';
    if (@where_clauses) {
        $where_clauses = ' AND ' . join(' AND ', @where_clauses);
    }

    my ($id) = @{$dbh->selectcol_arrayref(<<EndQuery, undef, @join_bind_values, @where_bind_values)};
SELECT MIN(A.$id_name +1)
    FROM $table_name A LEFT JOIN $table_name B
      ON A.$id_name + 1 = B.$id_name
         $join_clauses
      WHERE B.$id_name IS NULL
         $where_clauses
EndQuery

    return $id || 0;
}

#----------------------------------------------------------------------

sub check_table_exists {
    my(%params) = @_;

    my $dbh = $params{dbh};
    my $table_name = $params{table_name};
    my $dbname = $params{dbname};

    my ($table_exists) = grep { $_ eq $table_name } @{$dbh->selectcol_arrayref("SHOW TABLES FROM $dbname")};

    if ($table_exists) {
       return 1;
    }
    else {
       return 0;
    }
}
#----------------------------------------------------------------------#

sub exists_in_table{
  my (%params) = @_;
  my $dbh = $params{dbh};
  my $field_name = $params{field_name};
  my $field_value = $params{field_value};
  my $table_name = $params{table_name};
  my $policy =  $params{policy};
  my $where_fields = $params{where_fields};

  if ($field_name) {
      $where_fields->{$field_name} = $field_value;
  }

  if (defined $policy && $policy->{policy_id} && grep { $_ eq 'policy_id'} F5::DbUtils::get_db_fields($dbh, $table_name) ) {
      $where_fields->{policy_id} = $policy->{policy_id}; 
  }

  my $constructed_where_clause = construct_where_clause(where_fields => $where_fields);
  my $where_clause = $constructed_where_clause->{where_clause};
  my @where_values = @{$constructed_where_clause->{where_values}};

  my $exists_in_table = $dbh->selectrow_array(<<EndQuery, undef, @where_values);
SELECT COUNT(*)
  FROM $table_name
 WHERE $where_clause
EndQuery

  return  $exists_in_table;
}

#----------------------------------------------------------------------

sub get_column_values {
    my (%params) = @_;
    my $dbh = $params{dbh};
    my $field_name = $params{field_name};
    my $table_name = $params{table_name};

    my $constructed_where_clause = construct_where_clause(where_fields => $params{where_fields});
    my $where_clause = $constructed_where_clause->{where_clause};
    my @where_values = @{$constructed_where_clause->{where_values}};


    my $column_values = $dbh->selectcol_arrayref(<<EndQuery, undef, @where_values);
SELECT $field_name
  FROM $table_name
  WHERE $where_clause
EndQuery

   return $column_values;
}

#----------------------------------------------------------------------

sub get_distinct_values {
    my (%params) = @_;
    my $dbh = $params{dbh};
    my $table_name = $params{table_name};
    my $fields = $params{fields};
    if (!ref($fields)) {
        $fields = [$fields];
    }
    my $hash_field = $params{hash_field} || $fields->[0];

    my $constructed_where_clause = construct_where_clause(where_fields => $params{where_fields});
    my $where_clause = $constructed_where_clause->{where_clause};
    my @where_values = @{$constructed_where_clause->{where_values}};

    local $" = ',';
    my $distinct_values = $dbh->selectall_hashref(<<EndQuery, $hash_field, undef, @where_values);
SELECT DISTINCT @$fields
  FROM $table_name
  WHERE $where_clause
EndQuery

    return $distinct_values;
}

#----------------------------------------------------------------------

sub get_mysql_user {
    my (%params) = @_;
    return F5::Cfg::cfg_get_config_item($USERS_CFG_FILE, $params{section} || 'DATABASE', 'User');
}

#----------------------------------------------------------------------

sub get_mysql_password {
    return F5::Cfg::get_mysql_password(@_);
}

#----------------------------------------------------------------------

sub get_active_policy_id {
    my(%params) = @_;
    my $dbh = $params{dbh};
    my $account_id = $params{account_id};
    my $dbname = $params{dbname};

    my $active_policy_id = $dbh->selectrow_array(<<EndQuery, undef, $account_id);
SELECT active_policy_id
  FROM $dbname.ACCOUNTS
 WHERE account_id = ?
EndQuery

  return $active_policy_id;
}

#----------------------------------------------------------------------

sub add_callback {
    my (%params) = @_;
    my $dbh      = $params{dbh};
    my $callback = $params{callback};
    my $name = $params{name};

    my $callbacks = $dbh->{private_f5_callbacks};
    push (@$callbacks, $callback);
    $dbh->{private_f5_callbacks} = $callbacks;
    if ($name) {
        $dbh->{private_f5_callbacks_by_name}->{$name} = {
                                                         'name' => $name,
                                                         'callback' => $callback,
                                                         'callback_params' => $params{callback_params},
                                                        };
    }
}

#----------------------------------------------------------------------

sub get_callback_by_name {
    my (%params) = @_;
    my $dbh      = $params{dbh};
    my $name = $params{name};

    return $dbh->{private_f5_callbacks_by_name}->{$name};
}

#----------------------------------------------------------------------

sub is_mysql_up
{
    return F5::ProcessTable::is_proc_up('mysqld');
}

#----------------------------------------------------------------------

sub select_singleton {
    my(%params) = @_;

    my $dbh = $params{dbh} || F5::Db::Single->instance;
    my $table_name = $params{table_name};

    my $existing_row = $dbh->selectrow_hashref("SELECT * FROM $table_name");

    return $existing_row;
}


#----------------------------------------------------------------------

sub upsert_singleton {
    my(%params) = @_;

    my $dbh = $params{dbh} || F5::Db::Single->instance;
    my $table_name = $params{table_name};
    my $fields = $params{fields};
    my $rec = $params{rec};

    my $existing_row = $dbh->selectrow_hashref("SELECT * FROM $table_name");
    if (! $existing_row) {
        # There is no existing row -- just insert the new values
        return insert_row(%params);
    }

    my $is_update_needed = 0;
    for my $k (keys %$rec) {
        if (defined $existing_row->{$k} != defined $rec->{$k} ||
            (defined $rec->{$k} && $existing_row->{$k} ne $rec->{$k})) {

            $is_update_needed = 1;
            last;
        }
    }

    if ($is_update_needed) {
        return update_row(%params);
    }

    return;
}

#----------------------------------------------------------------------

use DBI::Const::GetInfoType;
sub is_sqlite {
    my $dbh = shift;
    return ($dbh->get_info($GetInfoType{SQL_DBMS_NAME}) || '') eq 'SQLite';
}

#----------------------------------------------------------------------

sub get_db_type {
    my $dbh = shift;
    if (is_sqlite($dbh)) {
        return 'sqlite';
    }
    else {
        return 'mysql';
    }
}

#----------------------------------------------------------------------

sub get_ignore_string {
    my $dbh = shift;
    my $ignore = "IGNORE ";
    if (is_sqlite($dbh)) {
        $ignore = "OR $ignore";
    } 
    return $ignore;
}

#----------------------------------------------------------------------

sub get_unix_timestamp_string {
    my (%params) = @_;
    my $dbh = $params{dbh};
    if (is_sqlite($dbh)) {
        my $field = $params{field} || "'now'";
        return "strftime('%s', $field)"
    }
    else {
        my $field = $params{field} || '';
        return "UNIX_TIMESTAMP($field)";
    }
}

#----------------------------------------------------------------------

sub get_now_string {
    my (%params) = @_;
    my $dbh = $params{dbh};
    my $now_string = 'NOW()';
    if (is_sqlite($dbh)) {
        $now_string = "date('now')";
    }
    return $now_string
}

#----------------------------------------------------------------------

sub get_max_placeholders {
    my (%params) = @_;
    my $dbh = $params{dbh};
    my $max_placeholders = 65400;
    if (is_sqlite($dbh)) {
        $max_placeholders = 240000;
    }
    return $max_placeholders;
}
#----------------------------------------------------------------------

sub sqlite_init {
    my ($dbh) = @_;
    $dbh->do('ATTACH "' . F5::Deploy::get_path_by_type("database") . '/PLC" as PLC');
    $dbh->do('ATTACH "' . F5::Deploy::get_path_by_type("database") . '/DCC" as DCC');
    $dbh->do('PRAGMA foreign_keys = ON');
    $dbh->sqlite_create_function("POW", 2, sub { return (shift) ** (shift) });
    $dbh->sqlite_create_function("IF", 3, sub { return (shift) ? (shift) : (shift) });
    $dbh->sqlite_create_aggregate("BIT_OR", -1, 'sqlite_bit_or');
}

package sqlite_bit_or;
sub new {bless [], shift;}

sub step {
   my ($self, @values) = @_;
   push @$self, @values;
}

sub finalize {
    my $self = shift;
    my $result = 0;
    foreach my $value ( @$self ) {
        $result |= $value;
    }
    return $result;
}

#----------------------------------------------------------------------
1;
