package F5::ImportExportPolicy::Binary;
#******************************************************************************
#copyright 2000-2003 Magnifire Networks Ltd. All Rights Reserved.
#
#FILE: import_export_policy.pl  
#
#USES: Log.pm
#
#CREATED: 19 August 2003 By Ephraim Dan
#
#MODIFIED: 09 October 2006 By Eitan Schuler: Basically refactored to perl module
#
#COMMENTS: provides policy import/export functionality

#*******************************************************************************

# CVS file:
# $Header: //depot/project/golan/TrafficShield/mng/import_export/import_export_policy.pl#8 $

use strict;
use warnings;

use base qw(F5::ImportExportPolicy::Base);

use F5::Logger;
use F5::DbUtils qw/insert_data_to_table/;
use F5::GlobalParametersUpgrade qw/params_upgrade_import/;
use F5::ParamExtractAttribCleanup qw(cleanup_param_extract_attrib);
use F5::NavigationParamsUpgrade qw(upgrade_navigation_params);
use F5::AlternateResponsesFixup;
use F5::AjaxAlternateResponsesFixup;
use F5::License;
use F5::WildcardEntityUpgrade;
use F5::NegativeSignatures;
use F5::SpecialCharactersUnmap;
use F5::ResponseScrubbing;
use F5::XMLSecurityUpgrade;
use F5::UrlContentUpgrade;
use F5::BigipVersionUtils;
use F5::ObjectTypeWildcardUpgrade;
use F5::PolicyXffUpgrade;
use F5::PolicyCsrfUpgrade;
use F5::ResetStaging;
use F5::CommonUpgrade;
use F5::ASMConfig::Entity::AllowedMethod;
use F5::ASMConfig::Entity::PolicyAttributes;
use F5::ASMConfig::Entity::Violation;
use F5::ASMConfig::Entity::ServerTechnology;
use F5::ASMConfig::Entity::PolicySignatureRequirement;
use F5::Cfg;
use F5::CRC;
use F5::ParseUri;
use F5::NegativeSignatures::Set;
use F5::BigipUtils;
use F5::ObjUtils qw(store_fd);
use DBI;
use Getopt::Long;    
use F5::Utils::Rest;
use F5::SystemUtils;
use F5::ThreatCampaigns;
use F5::DbUtils;

use Storable qw/fd_retrieve/;
use vars qw{
    $CURRENT_TS_VER
    $CURRENT_TS_BUILD
    @PLC_EXPORT_TABLES
    $EXPORT_HASH_MAX_BATCH
    $IMPLIED_ENCODING
    $MAX_IMPORT_FILE_VERSION
};

# key_ui_type
use constant  KEY_UI_TYPE_INTEGER           => 0x2;
use constant  KEY_UI_TYPE_INTEGER_CHECK_RANGE => 0x8;
use constant  KEY_UI_TYPE_DECIMAL           => 0x4;
use constant  KEY_UI_TYPE_DECIMAL_CHECK_RANGE => 0x10;
 
use constant OWA_SYSTEM_ID => 1;

# constants
my $RET_OK = 0;
my $RET_NOT_OK = -1;
my $TRUE = 1;
my $FALSE = 0;
my $EXIT_OK = 0;
my $EXIT_NOT_OK = -1;
my $DYNAMIC_CONTENT_VALUE =  2;
my $HEADER_CHARSET_TYPE_CODE = 1;
my $URI_CHARSET_TYPE_CODE = 2;
my $PARAM_NAME_CHARSET_TYPE_CODE = 3;
my $PARAM_VALUE_CHARSET_TYPE_CODE = 4;

my $NORMALIZATION_CODE_ASCII_DECODING = 1;
my $NORMALIZATION_CODE_MULTIPLE_SLASHES = 2;

my %CHARSET_TYPE_UPGRADE_MAP = (0 => $PARAM_VALUE_CHARSET_TYPE_CODE, 
                                1 => undef,
                                2 => undef,
                                100 => $URI_CHARSET_TYPE_CODE,
                                101 => $HEADER_CHARSET_TYPE_CODE,
                                102 => $PARAM_NAME_CHARSET_TYPE_CODE);
$IMPLIED_ENCODING = 'iso-8859-1'; # Encoding to use for policies from pre-multibyte versions (such as Red Mountain)
$MAX_IMPORT_FILE_VERSION = 9;

sub new {
    my($class, %params) = @_;
    my $self = {};
    bless $self, $class;
    $self->init(%params);
    return $self;
}

#--------------------------------------------------------------------
# subroutine definitions:
#--------------------------------------------------------------------


sub get_available_policy_name {
    my($self, %params) = @_;
    return $self->completed_successfully(msg => $self->get_unique_policy_name(
                                                           policy_name => $params{policy_name}, 
                                                           add_info => 0));
}

####################
# EXPORT FUNCTIONS #
####################

sub get_sigfile_timestamp
{
    my ( $self, %params ) = @_;
    # Get the timestamp from the last row
    my $sigfile_timestamp = $self->{dbh}->selectrow_array(<<EndQuery);
SELECT sigfile_timestamp
  FROM PLC.SIGFILE_STATUS
  ORDER BY sigfile_status_id DESC LIMIT 1;
EndQuery
              
    return $sigfile_timestamp;
}

#--------------------------------------------------------------------
sub export_policy
{
    my($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};

    my $policy = $params{policy};
    my $policy_id = $params{policy_id};
    $self->{policy_id} = $policy_id;

    my $filename = $params{filename};

    my %export_data;

    my ($rc, $sql, $sth, $rowdata);

    # Open output file - for output form, the file will be appended to
    # throughout the export process to minimize memory usage.
    my $outf;
    eval {
        my $gzip_cmd = F5::SystemUtils::find_command_path('gzip');
        no warnings;
        open OLD_STDOUT, ">&STDOUT";
        open NEW_STDOUT, ">", $filename;
        open STDOUT, ">&NEW_STDOUT";
        open $outf, "| $gzip_cmd";
        open STDOUT, ">&OLD_STDOUT";
    };

    if ($@)
    {
      return $self->fatal_error(user_msg => "Failed to create exported policy file (open pipe to gzip failed).",
                                log_msg => "failed to open pipe to gzip ($!)");
    }

    my $rv;
    # EXPORT THE POLICY

    # find the policy in PL_POLICIES

    $rc = $self->get_table_data('PL_POLICIES', \%export_data, 'id');

    if ($rc != $RET_OK)
    {
      return $self->fatal_error(user_msg => "Failed to export policy table 'PL_POLICIES'.",
                                log_msg => "failed to get table data from PL_POLICIES");
    }

    if (not exists $export_data{PL_POLICIES}{$policy_id})
    {
      return $self->fatal_error(user_msg => "No such policy_id '$policy_id' was found in the policy database.",
                                log_msg => "no such policy '$policy_id' in PL_POLICIES");
    }

    $self->{policy_name} = $export_data{PL_POLICIES}->{$policy_id}->{name};

    # export the version and build
    $export_data{TS_VER} = $CURRENT_TS_VER;
    $export_data{TS_BUILD} = $CURRENT_TS_BUILD;
    $export_data{_export_file_version} = $MAX_IMPORT_FILE_VERSION;
    $export_data{_negsig_update_file_date} = F5::NegativeSignatures->new($dbh)->get_update_file_date;
    my $license = F5::License->new();
    $export_data{LICENSE} = $license->get_asm_licensed_state_str();

    $export_data{policy_version_time} = $policy->{policy_version_time};
    $export_data{policy_version_device_name} = $policy->{policy_version_device_name};
    $export_data{policy_version_policy_name} = $policy->{policy_version_policy_name};
    $export_data{policy_version_last_change} = $policy->{policy_version_last_change};
    $export_data{sigfile_timestamp} = $self->get_sigfile_timestamp();

    if ($policy->{parent_policy_name} && $policy->{parent_policy_name} ne 'none') {
        $export_data{parent_policy_name} = $policy->{parent_policy_name};
    }

    # Append policy version number to description, and clear out.
    my $current_policy = $export_data{PL_POLICIES}->{$policy_id};

    $self->write_to_export_file($outf, \%export_data) or return $self->fatal_error(msg => "Cannot write to output file.");

    $rc = $self->get_negsig_tables(dbh => $dbh,
                                   outf => $outf,
                                   writer_callback => sub { $self->write_to_export_file($outf, shift); },
                                  );

    $rc = $self->get_violation_tables(dbh => $dbh,
                                      outf => $outf,
                                      writer_callback => sub { $self->write_to_export_file($outf, shift); },
                                     );

    $rc = $self->get_ssl_table(dbh => $dbh,
                               writer_callback => sub {
                                   $self->write_to_export_file($outf, shift);
                               });


    my %export_table_handlers = (
                                 'PL_ALLOWED_METHODS' => {
                                                          row_callback => sub {
                                                              my($row) = @_;
                                                              $row->{method_name} = F5::ASMConfig::Entity::AllowedMethod->get_method_name(dbh => $dbh,
                                                                                                                                          code => $row->{method_code});
                                                          }
                                                         },
                                );

    my %export_table_flags = F5::DbUtils::get_export_policy_flag_tables('PLC');
    my %export_table_where_clauses = F5::DbUtils::get_export_where_clauses_by_table('PLC');
 
    foreach my $table (@PLC_EXPORT_TABLES)
    {

        if ($export_table_flags{"PLC.$table"} && ! $params{$export_table_flags{"PLC.$table"}}) {
            my $export_table_flag = $export_table_flags{"PLC.$table"};
            $logger->debug("Export $export_table_flag not selected -- table $table skipped") if $logger;
            next;
        }

        my $table_hash = {};
        my $writer_callback = sub {
          $self->write_to_export_file($outf, shift);
        };
        if (($export_table_handlers{$table} && $export_table_handlers{$table}->{row_callback}) || $export_table_where_clauses{"PLC.$table"}) {
            $rc = $self->get_table_contents(
                                            dbh => $dbh,
                                            row_callback => $export_table_handlers{$table}->{row_callback},
                                            writer_callback => $writer_callback,
                                            table_name => $table,
                                            store_as_hash => 1,
                                            where_fields => { policy_id => $policy_id },
                                            extra_where_clause => $export_table_where_clauses{"PLC.$table"},
                                           );
        }
        else {
            $rc = $self->get_table_data($table, $table_hash, 'policy_id', $EXPORT_HASH_MAX_BATCH, $writer_callback);
        }

        if ($rc != $RET_OK)
        {
          return $self->fatal_error(user_msg => "Failed to export policy table '$table'.",
                                    log_msg => "failed to export table $table");
        }

    }

    local $SIG{CHLD} = '';
    if (! close($outf))
    {
      return $self->fatal_error(user_msg => "Failed to create exported policy file (close pipe to gzip failed).",
                                log_msg => "failed to close pipe to gzip ($! $?)");
    }

    ##############
    # SUCCESS !! #
    ##############

    return $self->completed_successfully(msg => "Policy '$self->{policy_name}' was successfully exported.");
}

#--------------------------------------------------------------------

sub get_negsig_tables {
    my($self, %params) = @_;

    my $dbh = $params{dbh};
    my $outf = $params{outf};
    my $writer_callback = $params{writer_callback};

    my $set_ids = $dbh->selectcol_arrayref(<<EndQuery, undef, $self->{policy_id});
SELECT set_id
  FROM PLC.PL_POLICY_NEGSIG_SETS
 WHERE policy_id = ?
EndQuery
    
    my $rv = F5::NegativeSignatures::Set->export_set(set_ids => $set_ids, dbh => $dbh);

    $self->write_to_export_file($outf, { NEGSIG_SETS => $rv->{sets_xml} });

    $self->get_table_contents(
                              dbh => $dbh,
                              table_name => 'NEGSIG_SIGNATURES',
                              fields => [ qw(sig_id rev last_update flg_is_user_defined sig_name sig_tag sig_tag_policy) ],
                              row_callback => sub {
                                  my($row) = @_;
                                  delete $row->{sig_name} unless ($row->{flg_is_user_defined});
                              },
                              writer_callback => $writer_callback,
                             );
}

#--------------------------------------------------------------------

sub get_violation_tables {
    my($self, %params) = @_;

    my $dbh = $params{dbh};
    my $outf = $params{outf};
    my $writer_callback = $params{writer_callback};

    my $viol_indexes = $dbh->selectcol_arrayref(<<EndQuery, undef, $self->{policy_id});
SELECT viol_index
  FROM PLC.PL_BLOCKING
 WHERE policy_id = ?
   AND (flg_learn | flg_alarm | flg_reject) = 1
EndQuery

    my @search_filters = (F5::ASMConfig::SearchFilter->new(field_name => 'viol_index',
                                                           field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                           field_value_list => $viol_indexes),
                          F5::ASMConfig::SearchFilter->new(field_name => 'flg_is_user_defined',
                                                           field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                                           field_value => 1),
                         );

    my $xml_content = F5::ASMConfig::Entity::Violation->export_xml(dbh => $dbh,
                                                                   search_filters => [ @search_filters ]);

    $self->write_to_export_file($outf, { VIOLATIONS => $xml_content });
}

#-------------------------------------------------------------------
sub get_ssl_table{
    my($self, %params) = @_;

    my $dbh = $params{dbh};
    my $writer_callback = $params{writer_callback};
    $self->get_table_contents(
                              dbh => $dbh,
                              table_name => 'SSL_CERTIFICATES',
                              fields => [ qw(ssl_cert_id key_name) ],
                              writer_callback => $writer_callback,
                              force_empty => 1,
                             );

}
#--------------------------------------------------------------------
# get_table_contents is designed to replace get_table_data, in that 
# it allows for row-based callbacks and filters.

sub get_table_contents {
  my($self, %params) = @_;

  my $dbh = $params{dbh};
  my $table_name = $params{table_name};
  my $where_fields = $params{where_fields};
  my $extra_where_clause = $params{extra_where_clause};
  my $fields = $params{fields} || [ '*' ];
  my $row_callback = $params{row_callback};
  my $writer_callback = $params{writer_callback};
  my $batch_limit = $params{batch_limit} || $EXPORT_HASH_MAX_BATCH;
  my $force_empty = $params{force_empty};
  my $store_as_hash = $params{store_as_hash};

  my @where_clauses = (1);
  my @bind_values;
  if ($where_fields) {
    my @where_fields = keys %$where_fields;
    push (@where_clauses, map("$_ = ?", @where_fields));
    @bind_values = map($where_fields->{$_}, @where_fields);
  }

  if ($extra_where_clause) {
      push (@where_clauses, $extra_where_clause);
  }

  my $where_clause = join(' AND ', @where_clauses);
  local $" = ',';
  my $sth = $dbh->prepare("SELECT @$fields FROM PLC.$table_name WHERE $where_clause");
  $sth->execute(@bind_values);

  my @table_contents;
  while(my $row = $sth->fetchrow_hashref) {
    if ($row_callback) {
      $row_callback->($row);
    }
    next unless %$row;
    delete $row->{is_inherited};

    push(@table_contents, $row);
    if ($batch_limit && scalar @table_contents % $batch_limit == 0) {
      $writer_callback->({ $table_name => $self->convert_table_contents_to_ref(table_contents => \@table_contents,
                                                                               store_as_hash => $store_as_hash)});
      @table_contents = ();
    }
  }

  if (($batch_limit && scalar @table_contents) || $force_empty) {
      $writer_callback->({ $table_name => $self->convert_table_contents_to_ref(table_contents => \@table_contents,
                                                                               store_as_hash => $store_as_hash)});
  }

  return $RET_OK;
}

#----------------------------------------------------------------------

sub convert_table_contents_to_ref {
    my($self, %params) = @_;

    my $table_contents = $params{table_contents};
    my $retval;

    if ($params{store_as_hash}) {
        my $row_count = 1;
        $retval = { map { (($_->{id} || $row_count++) => $_) } @$table_contents };
    }
    else {
        $retval = $table_contents;
    }

    return $retval;
}

#----------------------------------------------------------------------
# DEPRECATED - superceded by get_table_contents (but still in use!)
#
sub get_table_data
{
    my ($self, $table, $hash, $where_col, $batch_limit, $writer_callback) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};
    my $sql = "SELECT * FROM PLC.$table WHERE $where_col=?";

    my $sth = $dbh->prepare($sql);

    unless ($sth->execute($self->{policy_id})) 
    {
        my $dbi_err = $DBI::err ? $DBI::errstr : 'unknown';
        $logger->error("failed to execute '$sql' ($dbi_err)") if $logger;
        return $RET_NOT_OK;
    }

    my $table_hash;
    $table_hash->{$table} = {};
    my $rowcount = 0;
    while(my $row = $sth->fetchrow_hashref) {
        $rowcount++;
        my $hash_id = $row->{'id'} || $rowcount;
        $table_hash->{$table}->{$hash_id} = { %$row };
        if ($batch_limit && $rowcount % $batch_limit == 0) {
            $writer_callback->($table_hash);
            # If there are no more records after this batch, this will prevent an empty hash
            # or undef from being written in the next call to the writer_callback.
            $table_hash = {};
        }
    }

    if ($batch_limit && scalar values %$table_hash) {
        $writer_callback->($table_hash);
        $table_hash = {}; # This prevents the hash from being copied back to the passed $hash parameter.
    }

    $hash->{$table} = $table_hash->{$table};

    return $RET_OK;
}

#----------------------------------------------------------------------

####################
# IMPORT FUNCTIONS #
####################

#--------------------------------------------------------------------
sub import_policy
{
    my($self, %params) = @_;

    my $filename = $params{filename};
    my $force_import = $params{force};
    my $overwrite_flag = $params{overwrite};
    my $force_policy_id = $params{force_policy_id};
    my $force_policy_name_flg =  $params{force_policy_name_flg};
    my $is_within_ucs_upgrade = $params{is_within_ucs_upgrade};
    my $reset_sigs = $params{reset_sigs};
    my $get_policy_attributes_only = $params{get_policy_attributes_only};
    my $target_account_id = $params{target_account_id};

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};

    $self->{user_name} = $params{creator_name} || $self->{user_name};

    if (! -f $filename)
    {
      return $self->fatal_error(user_msg => "Exported policy file not found!.",
                                log_msg => "no such file '$filename'");
    }
    
    my $inf;
    eval {
        my $gzip_cmd = F5::SystemUtils::find_command_path('gzip');
        no warnings;
        open OLD_STDIN, "<&STDIN"; 
        open NEW_STDIN, "$filename";
        open STDIN, "<&NEW_STDIN";
        open $inf, "$gzip_cmd -q -d -c |";
        open STDIN, "<&OLD_STDIN";
    };

    if ($@)
    {
      return $self->fatal_error(user_msg => "Cannot read file: the file does not appear to be a valid exported policy file.",
                                log_msg => "can't open pipe to read from gzip ($!)");
    }

    my $import_data = $self->read_from_export_file($inf) or return $self->fatal_error(msg => "Cannot read from input file.");

    my $import_file_version;
    if (defined($import_data->{_export_file_version})) {
      $import_file_version = $import_data->{_export_file_version};
    }
    elsif ((! defined $import_data->{TS_VER}) #V30 or older
           || ($import_data->{TS_VER} =~ /\Qts_software_ts.ts.3.1/) #V31
           || ($import_data->{TS_VER} =~ /\Qts_software_ts.ts.5.0.0/
               && $import_data->{TS_BUILD} =~ /^9\.2\.[01]\./)  #pre-Columbia ASM
          ) {
      $import_file_version = 0;
    }
    else {
      $import_file_version = 1;
    }

    my $common_upgrade_version;
    if (!defined $import_data->{TS_VER}) {
        $common_upgrade_version = '3.0.0';
    }
    elsif ($import_data->{TS_VER} =~ /ts\.(\d+\.\d+\.\d+)/) {
        $common_upgrade_version = $1;
        if ($common_upgrade_version =~ /^5\./) {
            # 5.0.0 was used as a fake placeholder in TS_VER for all versions from 9.2.0 through 10.0.0
            $common_upgrade_version = $import_data->{TS_BUILD};
        }
    }
    $self->{common_upgrade}->set_old_version( $common_upgrade_version);

    if ($import_file_version > $MAX_IMPORT_FILE_VERSION && !$force_import) {
      return $self->fatal_error(msg => "Policy file is from a later version of ASM and cannot be imported.");
    }

    if (!defined $import_data->{TS_VER} && !$force_import) {
      return $self->fatal_error(msg => "Policy file is from an unsupported version of ASM and cannot be imported.");
    }

    my $import_bigip_version = $common_upgrade_version;

    if (F5::BigipUtils::is_device_bigip()) {
        my $negsig_update_file_date = F5::NegativeSignatures->new($dbh)->get_update_file_date;
        if ($negsig_update_file_date &&
            $import_data->{_negsig_update_file_date} && 
            $import_data->{_negsig_update_file_date} ne $negsig_update_file_date) {
          $self->add_info(msg => "Attack Signature update file ($negsig_update_file_date) differs from Attack Signature update file used by policy ($import_data->{_negsig_update_file_date})");
        }

    }

    my ($rc, $sql, $sth, $rowdata);

    #########################
    # CHECK THE POLICY NAME #
    #########################

    my ($old_plc_id) = keys %{$import_data->{PL_POLICIES}};
    my $policy_name = $params{policy_name} || $import_data->{PL_POLICIES}{$old_plc_id}{name};
    $policy_name = $self->policy_name_fixup(%params,
                                            policy_name => $policy_name);
    my $orig_policy_name = $policy_name;
    my $policy_type = $params{policy_type} || $import_data->{PL_POLICIES}{$old_plc_id}{policy_type} || F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY();
 
    my ($table, $id_map);

    if (F5::BigipUtils::is_device_bigip()) {
        $policy_name = $self->get_unique_policy_name(
                                                     force_policy_name_flg => $force_policy_name_flg,
                                                     force_policy_id => $force_policy_id,
                                                     overwrite_flag => $overwrite_flag,
                                                     dbh => $self->{dbh},
                                                     policy_name => $policy_name,
                                                    );
    
        $self->{policy_name} = $import_data->{PL_POLICIES}{$old_plc_id}{name} = $policy_name;
        $self->{policy_type} = $import_data->{PL_POLICIES}{$old_plc_id}{policy_type} = $policy_type;
        $import_data->{PL_POLICIES}{$old_plc_id}{rest_uuid} = F5::Utils::Rest::generate_uuid($policy_name);
        $import_data->{PL_POLICIES}{$old_plc_id}{policy_name_crc} = F5::CRC::get_crc($policy_name);
        if (defined($params{policy_description})) {
            $import_data->{PL_POLICIES}{$old_plc_id}{description} = $params{policy_description};
        }
    
        #####################
        # IMPORT THE POLICY #
        #####################
    
    
        # PL_POLICIES
        $table = 'PL_POLICIES';
        # on old versions we need to set custom security level
        $self->pl_policies_fixup_fields($import_data->{PL_POLICIES}, $import_data);
        $self->set_dynamic_session_for_upgrade(table_data => $import_data,
                                               is_within_ucs_upgrade => $is_within_ucs_upgrade,
                                               target_account_id => $target_account_id,
                                              );
    }

    if (! exists $import_data->{PL_POLICIES}->{$old_plc_id}->{flg_encoding_selected}) {
        $import_data->{PL_POLICIES}->{$old_plc_id}->{flg_encoding_selected} = 1;
    }

    $import_data->{PL_POLICIES}->{$old_plc_id}->{encoding} = lc($import_data->{PL_POLICIES}->{$old_plc_id}->{encoding});
    if ($get_policy_attributes_only) {
        # Finish reading file to prevent Broken Pipe error
        while (! eof($inf)) {
            my $dummy = <$inf>;
        }

        $self->{policy} = F5::ASMConfig::Policy->new({
                                                      policy_name => $orig_policy_name,
                                                      encoding => $import_data->{PL_POLICIES}->{$old_plc_id}->{encoding},
                                                      flg_encoding_selected => $import_data->{PL_POLICIES}->{$old_plc_id}->{flg_encoding_selected},
                                                      policy_type => $policy_type,
                                                      parent_policy_name => $import_data->{parent_policy_name},
                                                     });
        return $self->completed_successfully(msg => '');
    }

    my %subst_hash;

    if ($params{external_idmap}) {
        %subst_hash = %{$params{external_idmap}};
    }

    if ( defined($import_data->{PL_POLICIES}->{$old_plc_id}->{policy_type}) &&
         $import_data->{PL_POLICIES}->{$old_plc_id}->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT() &&
         !$import_data->{PL_POLICIES}->{$old_plc_id}->{flg_encoding_selected} ) {
        die "Cannot import a Parent Policy without an encoding\n";
    }

    my $policy_changes_log_params = {};
    my $existing_policies = $dbh->selectall_hashref('SELECT id, account_id, flg_encoding_selected, encoding FROM PLC.PL_POLICIES', 'id');
    if ($overwrite_flag && $existing_policies->{$force_policy_id}) {
      if ( $existing_policies->{$force_policy_id}->{flg_encoding_selected} &&
           $existing_policies->{$force_policy_id}->{encoding} ne lc($import_data->{PL_POLICIES}->{$old_plc_id}->{encoding}) &&
           !$params{allow_encoding_change}) {
          die "Imported and replaced policies have different encodings. You cannot perform this operation.\n";
      }

      $self->overwrite_pl_policies(data => $import_data->{PL_POLICIES},
                                   policy_id => $force_policy_id);

      $policy_changes_log_params->{policy_id} = $force_policy_id;
      $target_account_id = $existing_policies->{$force_policy_id}->{account_id};
      $subst_hash{account_id} = { '*' => $target_account_id };
      $id_map = { $old_plc_id => $force_policy_id };
    }
    else {
        my %insert_options;
        my $original_policy_id;
        $target_account_id = 0;
        $subst_hash{account_id} = { '*' => $target_account_id };
        if ($force_policy_id) {
            my @rows = values %{$import_data->{'PL_POLICIES'}};
            $original_policy_id = $rows[0]->{id};
            $rows[0]->{id} = $force_policy_id;
            $insert_options{preserve_id} = 1;
        }
        $id_map = insert_data_to_table(dbh => $self->{dbh},
                             table_name => "PLC.$table",
                             data => $import_data->{'PL_POLICIES'},
                             idmap => \%subst_hash,
                             idmap_save => 1,
                             idmap_strictness => $self->{enforce_strictness},
                             logger => $self->{logger},
                             %insert_options,
                            );

        if ($force_policy_id) {
            $id_map = { $original_policy_id => $force_policy_id };
        }
    }

    $logger->debug("Importing for account $target_account_id");

    if (not defined $id_map)
    {
      return $self->fatal_error(user_msg => "Failed to import policy table '$table'.",
                                log_msg => "import $table failed");
    }
    
    my $policy_id = $id_map->{$old_plc_id};
    $self->{policy_id} = $policy_id;
    $self->set_alternate_response($policy_id, $import_data);
    $self->set_negsig_quarantine_fields($policy_id, $import_data);

    $self->{policy_version_time} = $import_data->{policy_version_time};
    $self->{policy_version_device_name} = $import_data->{policy_version_device_name};
    $self->{policy_version_policy_name} = $import_data->{policy_version_policy_name};
    $self->{policy_version_last_change} = $import_data->{policy_version_last_change};
    $self->{import_sigfile_timestamp} = $import_data->{sigfile_timestamp};
    

    if (not $policy_id)
    {
      return $self->fatal_error(user_msg => "Failed to determine new policy_id for the imported policy.",
                                log_msg => "can't determine new policy_id!");
    }
    
    # NOTE: must save policy_id in $subst_hash{policy_id}{'*'} since we need access
    # to policy_id for populating policy defaults
    $subst_hash{policy_id} = { '*' => $policy_id };
    $subst_hash{from_id} = { 0 => 0 };

    $subst_hash{viol_index} = { map { ($_ => $_) } @{$dbh->selectcol_arrayref('SELECT viol_index FROM PLC.VIOLATIONS where flg_is_user_defined = 0')} };

    $logger->debug("new policy_id is: $policy_id") if $logger;

    # Call common_upgrade fixups for PL_POLICIES to cache values for later usage
    $self->{common_upgrade}->table_row_fixup(table_name => "PLC.PL_POLICIES",
                                             idmap => \%subst_hash,
                                             table_data => $import_data->{PL_POLICIES},
                                             dbh => $dbh,
                                             policy_id => $self->{policy_id});

    #########################################
    # For export file v2 and later, all tables other than PL_POLICIES are contained in
    # subsequent records in the file.
    my $flows_save_table; # PL_FLOWS and PL_PARAMS must be saved to pass into set_extract_from_id.
    $flows_save_table->{PL_FLOWS} = {};

    my %negsig_set_names;
    my %certificate_names;
    my %imported_tables;
    my $saved_expiration;
    my %import_table_config = ( 
                                'PL_OBJECTS' => { before_import =>  sub { modify_objects_data(shift, $target_account_id) },
                                                  id_map_key => [ 'extract_from_object_id',
                                                                  'from_id',
                                                                  'to_id',
                                                                  'obj_id',
                                                                  'object_id',
                                                                  'revoke_object_id',
                                                                  'prerequisite_object_id'] },
				'PL_HEADERS' => { id_map_key => 'header_id' },
                                'PL_FLOWS' => { id_map_key => 'flow_id',
                                                after_import => sub {
                                                    my $import_table = shift;
                                                    $flows_save_table->{'PL_FLOWS'} = { %{$flows_save_table->{'PL_FLOWS'}},
                                                                                        %$import_table };
                                                    return $RET_OK;
                                                }
                                              },
                                # PL_PARAMS will only be used for v3 or earlier.
                                'PL_PARAMS' => {
                                                 before_import => sub {
                                                   my($import_table, $import_table_name, $import_config, $subst_hash) = @_;
                                                   $flows_save_table->{'PL_PARAMS'} = $import_table;
                                                   $self->set_extract_from_id($flows_save_table, $subst_hash) if ($import_file_version == 0);
                                                   return $RET_OK;
                                                 },
                                                 import => sub { return {} if (params_upgrade_import(@_)) },
                                               },


                                'PL_PARAM_ATTRIBUTES' => { id_map_key => 'param_attrib_id',
                                                           before_import => \&pl_param_attributes_before_fixup,
                                                         },
                                'PL_FILETYPES' => { id_map_key => 'obj_type_id' },

                                # PLC.PL_XML_PROFILE_VALIDATION_FILES.xml_validation_file_id
                                # -----------------------------------------------------------------
                                'PL_XML_PROFILE_VALIDATION_FILES' => {delayed_load => 1,
                                                                     },
                                'PL_XML_VALIDATION_FILES' => {id_map_key => 'xml_validation_file_id',
                                                             },
                                # -----------------------------------------------------------------

                                # PLC.PL_JSON_PROFILE_VALIDATION_FILES.json_validation_file_id
                                # -----------------------------------------------------------------
                                'PL_JSON_PROFILE_VALIDATION_FILES' => {delayed_load => 1,
                                                                     },
                                'PL_JSON_VALIDATION_FILES' => {id_map_key => 'json_validation_file_id',
                                                             },
                                # -----------------------------------------------------------------

                                'PL_GRPC_PROFILE_IDL_FILES' => {delayed_load => 1, },
                                'PL_IDL_FILES' => {id_map_key => 'idl_file_id', },

                                'PL_EXTRACT_ATTRIBUTES' => { id_map_key => 'extract_attrib_id' },
                                'PL_DYNAMIC_FLOWS' => {
                                                          delayed_load => 1, 
                                                          enforce_strictness => 1,
                                                          idmap_strictness_callback => sub { 
                                                              my %args = @_;
                                                              my $msg = "Corruption detected while importing a policy - deleting Dynamic Flow record. Please reconfigure Dynamic Flows.";
                                                              $self->add_warning(msg => $msg, log_level => "warn");

                                                          },
                                                      },
                                'PL_GLOBAL_EXTRACTS' => { id_map_key => 'global_extract_id' },
                                'PL_STATIC_PARAMS' => { renamed_columns => { 'param_id' => 'param_attrib_id'} },
                                'PL_XML_WS_SUBVIOLATIONS' => { force_populate_defaults => 1,
                                                               populate_defaults => generate_populate_defaults('PL_XML_WS_SUBVIOLATIONS_DEFAULTS')
                                                             },
  
                                'PL_ALLOWED_METHODS' => { before_import => sub { $self->pl_allowed_methods_before_fixup(@_) },
                                                        },


                                 'PL_OBJECT_METHOD_OVERRIDES' => {
                                                                   before_import => sub { $self->url_allowed_methods(@_) },
                                                                },

                                'PL_CHARSETS' => { before_import => sub { $self->pl_charsets_before_fixup(shift, $import_file_version) },
                                                   force_populate_defaults => 1,
                                                   populate_defaults => sub { $self->pl_charsets_populate_defaults(@_) },
                                                 },
                                'PL_HIDE_PARAM_VALUES' => { populate_defaults => generate_populate_defaults('PL_HIDE_PARAM_VALUES_DEFAULTS')
                                                          },

                                'PL_ALLOWED_RESPONSE_CODES' => { populate_defaults => generate_populate_defaults('PL_ALLOWED_RESPONSE_CODES_DEFAULTS') 
								},
                                'PL_WEB_SCRAPING_SETTINGS' => {
                                                                before_import => sub { $self->warn_if_web_scraping_enabled(@_) },
                                                                },
                                # The following tables will only be used for v5 or earlier
                               'PL_CONTENT_PROFILES' => { id_map_key => [ 'profile_id', 'websocket_json_profile_id', 'websocket_text_profile_id' ], }, # overridden below for upgrade from Reactivity
                               'PL_OBJECT_CONTENT_PROFILES' => {
                                                                field_strictness => { 
                                                                                      'profile_id' => { allow_undef => 1 }
                                                                                     },
                                                               },
                               # The following tables will only be used for v5 or earlier
                                'PL_PARAMS_META_CHARS' => { renamed_table => 'PL_PARAM_VALUE_METACHARS',
                                                            renamed_columns => { 'param_id' => 'param_attrib_id',
                                                                                 'meta_char_id' => 'metachar' },
                                                            ignore => 1,
                                                          },
                                'PL_ILLEGAL_META_CHARS_POOL' => { import => sub {
                                                                    F5::WildcardEntityUpgrade::get_metachar_ids(@_,
                                                                                                                { dbname => 'PLC' });
                                                                  },
                                                                  id_map_key => 'metachar',
                                                                },
                                'PL_ALLOWED_OBJECTS_REGEXP' => { renamed_table => 'PL_OBJECTS',
                                                                 ignore => 1,
                                                                 before_import => sub {
                                                                   my($import_table, $import_table_name, $import_config, $subst_hash) = @_;
                                                                   F5::WildcardEntityUpgrade::convert_allowed_objects_regexp_to_wildcard($dbh, $logger, $import_table, $import_table_name, $import_config, 'PLC', $target_account_id);
                                                                   return $RET_OK;
                                                                 }
                                                               },
                                'PL_GLOBAL_PARAMS' =>   {
                                                            before_import => sub {
                                                                F5::WildcardEntityUpgrade::convert_param_name_regex_to_wildcard(
                                                                    $dbh, $logger, @_
                                                                );
                                                                return $RET_OK;
                                                            },
                                                        },
                               'NEGSIG_SIGNATURES' =>   {
                                                         import => sub {
                                                           my($dbh, $logger, $import_table, $import_table_name, $import_config, $subst_hash) = @_;
                                                           my $idmap = create_negsig_idmap(dbh => $dbh,
                                                                                                      logger => $logger,
                                                                                                      table_contents => $import_table,
                                                                                                      table_name => $import_table_name,
                                                                                                      entity_name_fields => [ qw(sig_name sig_tag sig_tag_policy)],
                                                                                                      id_field => 'sig_id',
                                                                                                     );
                                                           return $idmap;
                                                         },
                                                         id_map_key => 'sig_id',
                                                        },
                               'NEGSIG_SETS' =>   {
                                                   import => sub {
                                                           my($dbh, $logger, $import_table, $import_table_name, $import_config, $subst_hash) = @_;

                                                           if (ref($import_table)) {
                                                                my $idmap = create_negsig_idmap(dbh => $dbh,
                                                                  logger => $logger,
                                                                  table_contents => $import_table,
                                                                  table_name => $import_table_name,
                                                                  entity_name_fields => [ qw(set_name)],
                                                                  id_field => 'set_id',
                                                                  store_names => \%negsig_set_names,
                                                                );
                                                                return $idmap;
                                                            }
                                                            else {
                                                                my $rvs = F5::NegativeSignatures::Set->import_set(xml_text => $import_table,
                                                                                                                  dbh => $dbh,
                                                                                                                 );

                                                                my $idmap;
                                                                $self->add_negsig_set_warning(idmap => \$idmap, rvs => $rvs);
                                                                # Remove empty sets
                                                                for my $set_rv (@$rvs) {
                                                                    if ($set_rv->{update_status} eq $F5::NegativeSignatures::Set::SET_UPDATE_STATUS_EMPTY) {
                                                                        delete $idmap->{$set_rv->{old_set_id}};
                                                                    }
                                                                }

                                                                return $idmap || {};
                                                           }
                                                   },
                                                   id_map_key => 'set_id',
                                                  },

                               'PL_BLOCKING' => {
                                                    log_level_on_missing_idmap => '',
                                                },

                               'VIOLATIONS' =>   {
                                                   import => sub {
                                                           my($dbh, $logger, $import_table, $import_table_name, $import_config, $subst_hash) = @_;

                                                           my $import_rvs = F5::ASMConfig::Entity::Violation->import_xml(xml_content => $import_table,
                                                                                                                         dbh => $dbh,
                                                                                                                        );

                                                           my $idmap = {};
                                                           $self->complete_user_defined_violation_handling(import_rvs => $import_rvs->{entity_rvs},
                                                                                                           policy_id => $policy_id,
                                                                                                           idmap => $idmap);
                                                           return $idmap;
                                                   },
                                                   id_map_key => 'viol_index',
                                                  },
                               'PL_POLICY_NEGSIG_SIGNATURES' => { 
                                                                 before_import => sub {
                                                                     my($import_table, $import_table_name, $import_config, $subst_hash) = @_;

                                                                     ####################################################################
                                                                     # Import the previous envorced rule only if the ASU version on the origin device is the same
                                                                     # as ASU on the destination device
                                                                     my $system_sigfile_timestamp = $self->get_sigfile_timestamp();
                                                                     if ($self->{import_sigfile_timestamp} &&
                                                                         ($self->{import_sigfile_timestamp} ne $system_sigfile_timestamp) ) {
                                                                         for my $k (keys %$import_table) {
                                                                             my $rec = $import_table->{$k};
                                                                             $rec->{previous_enforced_rule} = '';
                                                                             $rec->{previous_enforced_rule_md5} = '';
                                                                         }
                                                                     }           
                                                                     ####################################################################
                                                                     return $RET_OK;
                                                                 },
                                                                 idmap_strictness_callback => sub {
                                                                     # Do nothing - if the signature is not on this device, simply ignore.
                                                                 },
                                                                 id_field_name => '',
                                                                },
                               'PL_POLICY_NEGSIG_SETS' => {
                                                           before_import => sub { 
                                                             my($import_table, $import_table_name, $import_config, $subst_hash) = @_;
                                                             $self->issue_negsig_set_warnings(table_contents => $import_table,
                                                                                              table_name => $import_table_name,
                                                                                              subst_hash => $subst_hash,
                                                                                              set_names => \%negsig_set_names,
                                                                                             );

                                                             my %seen_negsig_set_ids;
                                                             for my $k (keys %$import_table) {
                                                                 my $rec = $import_table->{$k};
                                                                 if ($seen_negsig_set_ids{$subst_hash->{set_id}->{$rec->{set_id}}}) {
                                                                     delete $import_table->{$k};
                                                                 }
                                                                 $seen_negsig_set_ids{$subst_hash->{set_id}->{$rec->{set_id}}} = 1;
                                                             }
                                                           },
                                                           after_import => sub {
                                                               $self->add_unused_default_negsig_set_warning;
                                                           },
                                                           id_field_name => ''
                                                          },
                               'PL_PARAM_NEGSIG_SIGNATURES' => { id_field_name => '', 
                                                                 log_level_on_missing_idmap => '' },

                               'PL_COOKIE_NEGSIG_SIGNATURES' => { id_field_name => '',
                                                                 log_level_on_missing_idmap => '' },

                               'PL_CONTENT_PROFILE_NEGSIG_SIGNATURES' => { id_field_name => '',
                                                                           log_level_on_missing_idmap => '',
                                                                         },
			       'PL_HEADER_NEGSIG_SIGNATURES' => { id_field_name => '',
                                                                 log_level_on_missing_idmap => '' },
                               'PL_OBJECT_NEGSIG_SIGNATURES' => { id_field_name => '',
                                                                 log_level_on_missing_idmap => '' },

                               'PL_NORMALIZATION' => {
                                                        before_import => sub {
                                                          my($import_table, $import_table_name, $import_config, $subst_hash) = @_;
                                                          remove_unused_normalization_settings($import_table);
                                                        },
                                                        populate_defaults => generate_populate_defaults('PL_NORMALIZATION_DEFAULTS'),
                                                     },
                                'SSL_CERTIFICATES' => {
                                                           import => sub {
                                                                        my($dbh, $logger, $import_table, $import_table_name, $import_config, $subst_hash) = @_;
                                                                        my $idmap = create_certificate_idmap(dbh => $dbh,
                                                                                           logger => $logger,
                                                                                           table_contents => $import_table,
                                                                                           table_name => $import_table_name,
                                                                                           entity_name_field => 'key_name',
                                                                                           id_field => 'ssl_cert_id',
                                                                                           store_names => \%certificate_names,
                                                                                          );
                                                                      return $idmap;
                                                                   },
                                                            id_map_key => 'ssl_cert_id',
                                                        },

                                 'PL_XML_PROFILE_CLIENT_CERTIFICATE' => {

                                                             before_import => sub {
                                                             my($import_table, $import_table_name, $import_config, $subst_hash) = @_;
                                                             $self->issue_certificate_warnings(table_contents => $import_table,
                                                                                              table_name => $import_table_name,
                                                                                              subst_hash => $subst_hash,
                                                                                              cert_names => \%certificate_names,
                                                                                             );
                                                           },
                                                           id_field_name => 'xml_client_id',
                                                         },

                                 'PL_XML_PROFILE_SERVER_CERTIFICATE' => {

                                                             before_import => sub {
                                                             my($import_table, $import_table_name, $import_config, $subst_hash) = @_;
                                                             $self->issue_certificate_warnings(table_contents => $import_table,
                                                                                              table_name => $import_table_name,
                                                                                              subst_hash => $subst_hash,
                                                                                              cert_names => \%certificate_names,
                                                                                             );
                                                           },
                                                           id_field_name => 'xml_server_id',

                                                                         },

                                'PL_POLICY_SESSION_XFF' => {populate_defaults => sub {
                                                                            F5::PolicyXffUpgrade::populate_policy_xff_defaults(
                                                                                dbh => $dbh,
                                                                                logger => $logger,
                                                                                policy_id => $policy_id,);
                                                                            return $RET_OK;},
                                                                    },
                                'PL_POLICY_CSRF' => {populate_defaults => sub {
                                                                            F5::PolicyCsrfUpgrade::populate_policy_csrf_defaults(
                                                                                dbh => $dbh,
                                                                                logger => $logger,
                                                                                policy_id => $policy_id,);
                                                                            return $RET_OK;},
                                                                    },

                                 'PL_POLICY_CSRF_URLS' => { id_map_key => 'csrf_url_id' },

                                 'PL_LOGIN_OBJECTS' => { id_map_key => 'login_object_id',
                                                         after_import => sub {
                                                                  my ($import_table) = @_; 
                                                                  eval{$import_table = (values %$import_table)[0];};
                                                                  if (defined $import_table->{expiration_period}) {
                                                                      $saved_expiration = $import_table->{expiration_period};
                                                                  } 
                                                                  return $RET_OK;},
                                                               },
                                  
                                 'PL_POLICY_PREREQUISITE_ATTRIBUTES' => { populate_defaults => sub {
                                                                               my $ignore = F5::DbUtils::get_ignore_string($dbh);
                                                                               $dbh->do(<<EndQuery, undef, $policy_id, defined $saved_expiration ? $saved_expiration : 0);
INSERT $ignore INTO PLC.PL_POLICY_PREREQUISITE_ATTRIBUTES
(policy_id, expiration_period)
VALUES
(?,?)
EndQuery
                                                                               return $RET_OK;
                                                                          }
                                                                        },
                                'PL_POLICY_BUILDER_CONFIG' => { id_map_key => 'pbconfig_id' },

                                'PL_SCANNER_VULNS' => { field_strictness => {
                                                                                'cookie_id' => { allow_undef => 1 },
                                                                                'object_id' => { allow_undef => 1 },
                                                                                'param_attrib_id' => { allow_undef => 1 },
                                                                            },
                                                      },

                                'PL_ALLOWED_COOKIES' => { id_map_key => 'cookie_id' },

                                'PL_SCANNER_CONFIG' => { id_map_key => 'scanner_config_id' },

                                'PL_SCANNER_SITE_MAP_TRANSMISSION' => { import => sub { 
                                                                            # Ignore this table -- marked as POLICY_IMPORT in order to
                                                                            # ensure that it gets emptied on replace policy,
                                                                            # also may be useful for diagnostics
                                                                            return $RET_OK;
                                                                        }
                                                                      },

                                'PL_POLICY_TEMPLATES_SETTINGS' => {
                                                                      before_import => sub {
                                                                         return $self->validate_policy_template(@_);
                                                                      },
                                                                  },

                                'PL_LOGIN_BRUTE_FORCE_OBJECT_ATTRIBUTES' =>  { field_strictness => {
                                                                                                    'object_id' => { allow_zero => 1 },
                                                                                                   },
                                                                             },

                                'PL_SUGGESTIONS'  =>    {    field_strictness => {
                                                                                   'sig_id' => { allow_undef => 1,
                                                                                                 allow_zero => 1 },
                                                                                 },
                                                        },

                                'PL_CONTENT_PROFILE_VALUE_METACHARS' => {
                                                                         log_level_on_missing_idmap => '',
                                                                        },
                                'PL_WEBSOCKET_URLS' => {
                                                        id_map_key => 'websocket_url_id',
                                                        enforce_strictness => 0,
                                                       },

                                'PL_MICROSERVICES' => {
                                                        id_map_key => 'microservice_id',
                                                        enforce_strictness => 0,
                                                       },

                                'PL_SERVER_TECHNOLOGIES' => {
                                                             before_import => sub {
                                                                 return $self->skip_if_missing_parent_record(data => $_[0],
                                                                                                             table_name => $_[1],
                                                                                                             parent_entity_class => 'F5::ASMConfig::Entity::ServerTechnology',
                                                                                                            );

                                                             },
                                                            },
                                'PL_THREAT_CAMPAIGNS' => {
                                                             before_import => sub {
                                                                 return $self->skip_if_missing_parent_record(data => $_[0],
                                                                                                             table_name => $_[1],
                                                                                                             parent_entity_class => 'F5::ASMConfig::Entity::ThreatCampaign',
                                                                                                             add_warnings => 0,
                                                                                                            );

                                                             },
                                                            },
                                'PL_POLICY_NEGSIG_REQUIREMENTS' => {
                                                                      before_import => sub {
                                                                          if (! $is_within_ucs_upgrade) {
                                                                              return $self->validate_sig_requirements(@_);
                                                                          }
                                                                      },
                                                                   },
                             );


    my @plc_tables_to_read = (@PLC_EXPORT_TABLES,
                              # Import negsig tables for setting up idmap and handling quarantine
                              'NEGSIG_SETS',
                              'NEGSIG_SIGNATURES',
                              'SSL_CERTIFICATES',
                              'VIOLATIONS',
                             );

    for my $renamed_table (keys %{$self->{common_upgrade}->get_renamed_tables}) {
        $renamed_table =~ s/^PLC\.//;
        push(@plc_tables_to_read,
             $renamed_table);
    }

    if ($import_file_version < 4) {
      # Replace PL_PARAM_ATTRIBUTES in the table list with PL_PARAMS.
      # PL_PARAMS was dropped as of Sonoma and broken out into multiple tables.
      # See https://peterpan.f5net.com/cgi-bin/twiki/view/FSPilot/SonomaASMGlobalParametersSDD
      for my $table (@plc_tables_to_read) {
        $table = 'PL_PARAMS' if ($table eq 'PL_PARAM_ATTRIBUTES');
      }
    }

    if ($import_file_version < 6) {
      # Replace PL_PARAM_VALUE_METACHARS in the table list with PL_PARAMS_META_CHARS
      # See https://peterpan.f5net.com/cgi-bin/twiki/view/FSPilot/GolanASMWildcardEntity
      for my $table (@plc_tables_to_read) {
        $table = 'PL_PARAMS_META_CHARS' if ($table eq 'PL_PARAM_VALUE_METACHARS');
      }
      push(@plc_tables_to_read, 
           'PL_ILLEGAL_META_CHARS_POOL',
           'PL_ALLOWED_OBJECTS_REGEXP',
          );

      $import_table_config{'PL_FILETYPES'} = {  id_map_key => 'obj_type_id',
                                                before_import => sub {
                                                  my($data, $table_name, $import_config, $subst_hash) = @_;

                                                  F5::ObjectTypeWildcardUpgrade::add_object_type_wildcards(dbh => $dbh,
                                                                                                           logger => $logger,
                                                                                                           policy_id => $policy_id,
                                                                                                           table_data => $data);
                                                },
                                             };
    }

    if ($import_file_version < 9) {
      # Reactivity needs to be converted to F5 XML config

      my $xml_profile_defense_levels;
      $import_table_config{'PL_CONTENT_PROFILES'} = { 
                                                 id_map_key => 'profile_id',
                                                 after_import => sub { 
                                                   my($import_table, $import_table_name, $import_config, $subst_hash) = @_;
                                                   $xml_profile_defense_levels = F5::XMLSecurityUpgrade::save_xml_defense_levels(
                                                                                                                                 table_data => $import_table,
                                                                                                                                 id_map => $subst_hash,
                                                                                                                                );
                                                   return $RET_OK;
                                                 },
                                                };
      $import_table_config{'PL_XML_PROFILE_DEFENSE_ATTRIBUTES'} = { 
                                                                   before_import => sub {
                                                                     my($import_table, $import_table_name, $import_config, $subst_hash) = @_;
                                                                     F5::XMLSecurityUpgrade::convert_xml_defense_attributes(
                                                                                                                            dbh => $dbh,
                                                                                                                            table_data => $import_table,
                                                                                                                           );
                                                                     return $RET_OK;
                                                                   },
                                                                   after_import => sub {
                                                                     my($import_table, $import_table_name, $import_config, $subst_hash) = @_;
                                                                     F5::XMLSecurityUpgrade::populate_xml_defense_attributes(
                                                                                                                             dbh => $dbh,
                                                                                                                             xml_profile_defense_levels => $xml_profile_defense_levels,
                                                                                                                            );
                                                                     return $RET_OK;
                                                                   }
                                                                  };
      $import_table_config{'PL_XML_PROFILE_VALIDATION_FILES'} = { 
                                                                   after_import => sub {
                                                                     my($import_table, $import_table_name, $import_config, $subst_hash) = @_;
                                                                     F5::XMLSecurityUpgrade::set_primary_enforced_validation_file(
                                                                                                                                  dbh => $dbh,
                                                                                                                                 );
                                                                     return $RET_OK;
                                                                   }
                                                                  };
    }

    my @delayed_tables_data = ();

    while (1) { # Endless loop is broken either when there are no more records in the file 
                # or when all tables have been loaded from the single import_data hash
        my $import_table;
        my $import_table_name;
        if ($import_file_version >= 2) {
            last if (eof($inf));
            $import_table = $self->read_from_export_file($inf) or return $self->fatal_error(msg => "Cannot read from input file.");
            $import_table_name = (keys %$import_table)[0];
            # Skip unless this table is on the list of tables to import. This is for future-proofing.
            next unless (grep($_ eq $import_table_name, @plc_tables_to_read));
            $import_table = (values %$import_table)[0];

            # $import_table can be undef only due to a bug in versions 9.2.2 - 9.4_devel.
            # undef is functionally equivalent to an empty hashref, so set that there.
            $import_table = {} if !defined($import_table);
        } else {
            last unless (@plc_tables_to_read);
            $import_table_name = shift(@plc_tables_to_read);
            $import_table = $import_data->{$import_table_name};
            next unless defined($import_table);
        }

        my $current_table_config = $import_table_config{$import_table_name} || {};
        if (my $new_table_name = $self->{common_upgrade}->get_new_table_name(table_name => "PLC.$import_table_name")) {
            $new_table_name =~ s/^PLC\.//;
            $current_table_config = $import_table_config{$new_table_name} || {};
        }

        if($current_table_config &&  $current_table_config->{delayed_load}) {
            push @delayed_tables_data, {import_table => $import_table, 
                                        import_table_name => $import_table_name, 
                                        current_table_config => $current_table_config,
                                       };
        } else {
            my $rv = $self->single_table_import($import_table, 
                                                $import_table_name, 
                                                $current_table_config,
                                                \%subst_hash);

            $imported_tables{$import_table_name} = 1;

            if ($rv != $RET_OK) {
                return $self->fatal_error(msg => "Import failure for table $import_table_name");
            }

            $self->{common_upgrade}->after_table_complete(table_name => "PLC.$import_table_name",
                                                          policy_id  => $policy_id,
                                                          dbh        => $dbh,
                                                          logger     => $logger);
        }
    }

    for my $delayed_table_data (@delayed_tables_data) {
        my $rv = $self->single_table_import($delayed_table_data->{import_table}, 
                                            $delayed_table_data->{import_table_name}, 
                                            $delayed_table_data->{current_table_config},
                                            \%subst_hash);

        $imported_tables{$delayed_table_data->{import_table_name}} = 1;
        

        if ($rv != $RET_OK) {
            return $self->fatal_error(msg => "Import failure for table $delayed_table_data->{import_table_name}");
        }

        $self->{common_upgrade}->after_table_complete(table_name => "PLC.$delayed_table_data->{import_table_name}",
                                                      policy_id  => $policy_id,
                                                      dbh        => $dbh,
                                                      logger     => $logger);

    }

    for my $table_name (keys %import_table_config) {
      my $config = $import_table_config{$table_name};

      if ( (!$imported_tables{$table_name} && $config->{populate_defaults} ) || 
             ( $config->{populate_defaults} && $config->{force_populate_defaults})) {
        my $rv = $config->{populate_defaults}->($dbh, $logger, undef, $table_name, $config, \%subst_hash);
      }
    }

    $self->{common_upgrade}->add_default_rows(policy_id => $policy_id,
                                              seen_tables => { map { ("PLC.$_" => 1) } keys %imported_tables});

    $self->{common_upgrade}->after_all_tables_complete(policy_id  => $policy_id, dbh => $dbh, idmap => \%subst_hash);

    cleanup_param_extract_attrib(dbh => $dbh,
                                 logger => $logger,
                                 dbname => 'PLC',
                                 policy_id => $policy_id);

    upgrade_navigation_params(dbh => $dbh,
                              logger => $logger,
                              policy_id => $policy_id,
                              dbname => 'PLC');

    F5::ResetStaging::reset_staging_and_tightening(dbh => $dbh,
      logger => $logger,
      policy_id => $policy_id );

    if ($import_file_version < 6) {
      F5::WildcardEntityUpgrade::globalize_charset_definitions(dbh => $dbh,
                                                               logger => $logger,
                                                               policy_id => $policy_id,
                                                               dbname => 'PLC');

      F5::WildcardEntityUpgrade::add_wildcard_entities(dbh => $dbh,
                                                       logger => $logger,
                                                       policy_id => $policy_id,
                                                       dbname => 'PLC');

    }

    if ($import_file_version < 8) {
        F5::SpecialCharactersUnmap::unmap_all(dbh => $dbh,
                                               logger => $logger,
                                               policy_id => $policy_id,
                                               dbname => 'PLC');
    }

    if (F5::BigipVersionUtils::zero_pad_segments($import_bigip_version) lt F5::BigipVersionUtils::zero_pad_segments('11.0.0') ) {
        F5::UrlContentUpgrade::populate_url_content_defaults(dbh => $dbh,
                                                             policy_id => $policy_id,
                                                             idmap => \%subst_hash,
                                                            );
    }

    my $negative_signatures = F5::NegativeSignatures->new($dbh);
    $negative_signatures->add_all_signatures_to_policy(policy_id => $policy_id);

    F5::ThreatCampaigns::add_all_threat_campaigns_to_policy(dbh => $dbh, policy_id => $policy_id, logger => $logger);

    if ($import_file_version < 7 or $reset_sigs) {
      $dbh->do(<<EndQuery, undef, $policy_id);
DELETE FROM PLC.PL_POLICY_NEGSIG_SETS
 WHERE policy_id = ?
EndQuery
      $negative_signatures->add_default_sets_to_policy(policy_id => $policy_id, flg_reject => ($reset_sigs ? 0 : undef));
    }

    F5::ResponseScrubbing::add_defaults(dbh => $dbh,
                                        policy_id => $policy_id,
                                        logger => $logger,
                                        dbname => 'PLC');

    F5::AjaxAlternateResponsesFixup::set_ajax_alternate_response($dbh, $logger, "PLC", $policy_id);

    $self->fix_ws_properties();

    ##############
    # SUCCESS !! #
    ##############
    close($inf);

    if (defined $policy_changes_log_params->{policy_id}) {
        $self->update_pl_policy_changes_log(policy_id => $policy_changes_log_params->{policy_id})
    }

    return $self->completed_successfully(msg => "The operation was completed successfully. The policy name is '$self->{policy_name}'.");
}

#--------------------------------------------------------------------
sub single_table_import {
    my ($self, $import_table, $import_table_name, $import_config, $subst_hash) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};
    # handle renamed table
    if ($import_config->{renamed_table}) {
      $import_table_name = $import_config->{renamed_table};
    }

    # handle renamed columns
    if ($import_config->{renamed_columns}) {
      convert_renamed_columns($import_table, $import_table_name, $import_config->{renamed_columns});
    }

    # before_import
    if ($import_config->{before_import} && $import_table) {
      my $ret = $import_config->{before_import}->($import_table, $import_table_name, $import_config, $subst_hash);
        if ($ret != $RET_OK) {
            return $RET_NOT_OK;
        }
    }

    if (! $import_config->{import}) {
        # F5::CommonUpgrade row_handler
        $self->{common_upgrade}->table_row_fixup(table_name => "PLC.$import_table_name",
                                                 idmap => $subst_hash,
                                                 table_data => $import_table,
                                                 dbh => $dbh,
                                                 policy_id => $self->{policy_id});
    }
    
    my $new_table_name = $self->{common_upgrade}->get_new_table_name(table_name => "PLC.$import_table_name");
    if (defined $new_table_name ) {
        $new_table_name =~ s/^PLC\.//;
        $import_table_name = $new_table_name;
    }

    if ($import_table_name) {
        my $id_map;
        my @id_map_keys;
        if ($import_config->{id_map_key}) {
            if (ref($import_config->{id_map_key}) eq 'ARRAY') {
                @id_map_keys = @{$import_config->{id_map_key}};
            } else {
                @id_map_keys = ($import_config->{id_map_key});
            }
        }
        if ($import_config->{force_populate_defaults} && 
            $import_config->{populate_defaults} && 
            !scalar keys %$import_table) {
          $id_map = $import_config->{populate_defaults}->($dbh, $logger, $import_table, $import_table_name, $import_config, $subst_hash);
        } elsif ($import_config->{import}) {
          $id_map = $import_config->{import}->($dbh, $logger, $import_table, $import_table_name, $import_config, $subst_hash);
        } else {
          my %extra_params;
          for my $copied_extra_param (qw(id_field_name ignore)) {
            if (exists $import_config->{$copied_extra_param}) {
              $extra_params{$copied_extra_param} = $import_config->{$copied_extra_param};
            }
          }
    
          my $idmap_strictness = ($import_config && (exists $import_config->{enforce_strictness})) ? 
                                 $import_config->{enforce_strictness} :
                                 $self->{enforce_strictness};
          my $idmap_strictness_callback = $import_config->{idmap_strictness_callback};
    
          $id_map = insert_data_to_table(dbh => $dbh,
                                         table_name => "PLC.$import_table_name",
                                         data => $import_table,
                                         idmap => $subst_hash,
                                         idmap_save => (@id_map_keys ? $TRUE : $FALSE),
                                         idmap_strictness => $idmap_strictness,
                                         idmap_strictness_callback => $idmap_strictness_callback,
                                         field_strictness => $import_config->{field_strictness},
                                         logger => $logger,
                                         log_level_on_missing_idmap => exists $import_config->{log_level_on_missing_idmap}
                                                                       ? $import_config->{log_level_on_missing_idmap}
                                                                       : 'warn',
                                         %extra_params,
                                        );
        }
    
        if (!defined $id_map) {
            return $RET_NOT_OK;
        }
    
        if (@id_map_keys && $id_map) {
            for my $k (@id_map_keys) {
                if (defined($subst_hash->{$k})) {
                    $subst_hash->{$k} = { %{$subst_hash->{$k}}, %$id_map };
                } else {
                    $subst_hash->{$k} = $id_map;
                }
            }
        }
    }
    # after_import

    if ($import_config->{after_import}) {
        my $ret = $import_config->{after_import}->($import_table, $import_table_name, $import_config, $subst_hash);
        if ($ret != $RET_OK) {
            return $RET_NOT_OK;
        }
    }
    return $RET_OK;
}

#--------------------------------------------------
sub modify_objects_data
{
    my ($data, $acc_id) = @_;
    my ($app_id, $obj);
    foreach my $id (keys %$data)
    {
       if($data->{$id}->{uri} =~ m#^http(s)?://([^:/]+)(?::\d+)?(/[^\t\r\n]*)?#)
       {
          ($app_id, undef, $obj) = F5::ParseUri::parse_unescaped_uri($data->{$id}->{uri});
          $data->{$id}->{app_id} = $app_id;
          $data->{$id}->{uri} = $obj;
       }
       else
       {
          ($app_id, $obj) = ((defined $data->{$id}->{app_id}) ? $data->{$id}->{app_id} : 0 , $data->{$id}->{uri});
       }
       if(! length($obj))
       {
          # unable to parse; ignore this line
          delete $data->{$id};
          next;
       }
    }

    return $RET_OK;
}

#--------------------------------------------------------------------
sub remove_unused_normalization_settings {
    my ($data) = @_;
    foreach my $id (keys %$data) {
       delete $data->{$id} if(($data->{$id}->{normalization_code} == $NORMALIZATION_CODE_ASCII_DECODING) ||
                              ($data->{$id}->{normalization_code} == $NORMALIZATION_CODE_MULTIPLE_SLASHES));
    }

    return $RET_OK;
}
#--------------------------------------------------------------------
sub fix_act_as_methods {

    my($self, $data, $table_name, $import_config, $subst_hash) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};


    for my $k (keys %$data) {
        my $row = $data->{$k};
        if ($row->{method_name}) {
            my $act_as_method_name = F5::ASMConfig::Entity::AllowedMethod->get_method_name(dbh => $self->{dbh},
                                                                                           code => $row->{act_as_method_code}) || 'GET';

            $row->{method_code} = F5::ASMConfig::Entity::AllowedMethod->get_method_code(dbh => $dbh,
                                                                                        name => $row->{method_name},
                                                                                        add_if_missing => 1,
                                                                                        default_act_as => $act_as_method_name,
                                                                                       );
        }
        else {
            # If this is an import from a prior version, disallow PL_ALLOWED_METHODS records
            # with method_code higher than the provided factory METHOD_POOL, to prevent inconsistencies.
            # The highest method_code since v9.4.1 was 37 (the ability to add methods to the pool was only added in v9.4.4.
            if (!($subst_hash->{method_code} && $subst_hash->{method_code}->{$row->{method_code}}) &&
                $row->{method_code} > 37) {

                $self->add_warning(msg => "Allowed methods using user-defined method names from previous versions cannot be imported and must be re-created.");
                delete $data->{$k};
            }
        }
    }
    
    return $RET_OK;
}

#--------------------------------------------------------------------

sub set_alternate_response
{
   my ($self, $policy_id,$data) = @_;

   my $row = (values %{$data->{"PL_POLICIES"}})[0];
   return $RET_OK if not exists $row->{alternate_response_file};
   $row->{policy_id} = $policy_id;
   F5::AlternateResponsesFixup::set_alternate_response($self->{dbh}, $self->{logger}, "PLC", $row);

   return $RET_OK;
}

#--------------------------------------------------------------------

sub set_negsig_quarantine_fields
{
   my ($self, $policy_id,$data) = @_;

   my $row = (values %{$data->{"PL_POLICIES"}})[0];
   return $RET_OK if not exists $row->{flg_quarantine_signatures};
   $row->{policy_id} = $policy_id;

   insert_data_to_table(dbh => $self->{dbh},
                        logger => $self->{logger},
                        table_name => 'PLC.PL_POLICY_NEGSIG_ATTRIBUTES',
                        data => { 1 => $row },
                       );

   return $RET_OK;
}

#--------------------------------------------------------------------

sub set_dynamic_session_for_upgrade {
  my ($self, %params) = @_;

  my $table_data = $params{table_data};
  my $is_within_ucs_upgrade = $params{is_within_ucs_upgrade};
  my $dbh = $self->{dbh};

  my $row = (values %{$table_data->{"PL_POLICIES"}})[0];
  if ($params{target_account_id} && $is_within_ucs_upgrade && ! exists $row->{flg_use_dynamic_session_in_url}) {
    my $dynamic_session_vals = $dbh->selectrow_hashref(<<EndQuery, undef, $params{target_account_id});
SELECT flg_use_dynamic_session_in_url,
       dynamic_session_pattern,
       dynamic_session_description
  FROM DCC.ACCOUNTS
 WHERE account_id = ?
EndQuery

    for my $k (keys %$dynamic_session_vals) {
      $row->{$k} = $dynamic_session_vals->{$k};
    }
  }

  return $RET_OK;
}

#--------------------------------------------------------------------
# check if a given account id exist on DCC

sub check_account_id
{
   my ($self, $input_account_id) = @_;

   my $sql = 'SELECT account_id FROM DCC.ACCOUNTS WHERE account_id = ?';
   if((! defined $self->{dbh}->selectrow_array($sql, undef, $input_account_id)) || $DBI::err)
   {
      my $dbi_err = $DBI::err ? $DBI::errstr : 'unknown';
      return $self->fatal_error(user_msg => "account_id : '$input_account_id' was not found in the database. Provide a valid acount_id.",
                                log_msg => "no account found with account_id='$input_account_id' in ACCOUNTS ($dbi_err)");
   }

   return $TRUE;
}

#--------------------------------------------------------------------
# set the extract from object of dynamic parameter
sub set_extract_from_id
{
   my ($self, $data, $subst_hash) = @_;
   my $logger = $self->{logger};
   my $params_table = 'PL_PARAMS';
   my $flows_table = 'PL_FLOWS';
   my $extract_from_id;
   
   $logger->info("set_extract_from called") if $logger;
   foreach my $id (keys %{$data->{$params_table}})
   {
      my $extract_from_id;
      # only set a value for dynamic parameters that are missing this setting
      if  (defined $data->{$params_table}->{$id}->{'extract_from_object_id'})
      {
         next;
      }
      elsif($data->{$params_table}->{$id}->{'key_input_type'} != $DYNAMIC_CONTENT_VALUE)
      {
         $extract_from_id = 0;
      }
      else
      {
         # find the new object id of the source object in the flow that this parameter is configured on
         my $old_flow_id = $data->{$params_table}->{$id}->{'flow_id'};

         my $old_from_id = $data->{$flows_table}->{$old_flow_id}->{'from_id'};
         if($old_from_id)
         {
            $extract_from_id = $old_from_id;
         }
         else
         {
            $extract_from_id = 0;
         }
      }

      $logger->debug("extract_from_id=$extract_from_id") if $logger;

      $data->{$params_table}->{$id}->{'extract_from_object_id'} = $extract_from_id;

   }
}

#--------------------------------------------------------------------

sub pl_charsets_before_fixup {
  my($self, $data, $import_file_version) = @_;
  if ($self->{incompatible_charsets}) {
    %{$data} = ();
    return $RET_OK;
  }

  fix_old_charset_types($data) if($import_file_version == 0);
  return $RET_OK;
}

#--------------------------------------------------------------------

sub pl_charsets_populate_defaults {
  my($self, $dbh, $logger, $import_table, $import_table_name, $import_config, $subst_hash) = @_;

  $self->insert_charset_defaults(policy_id => $self->{policy_id});
  return $RET_OK;
}

#----------------------------------------------------------------------

sub pl_allowed_methods_before_fixup { 
    my($self, $data, $table_name, $import_config, $subst_hash) = @_;

    my $rv = $self->fix_act_as_methods($data, $table_name, $import_config, $subst_hash);
    return $RET_NOT_OK unless $rv == $RET_OK;

    return remove_duplicates($data, 'method_code');
}
#----------------------------------------------------------------------

sub warn_if_web_scraping_enabled {
    my($self, $data, $table_name, $import_config, $subst_hash) = @_;

    for my $k (keys %$data) {
        my $row = $data->{$k};
        if ($row->{flg_is_enabled} || $row->{flg_web_scraping_block} || $row->{flg_web_scraping_alarm}) {
            $self->add_warning(msg => "Web Scraping configuration (previously used in this security policy) will be ignored. Device-ID mode must be configured in bot profile for this option to work.");
        }
    }

    return $RET_OK;
}

#----------------------------------------------------------------------

sub url_allowed_methods {
    my($self, $data, $table_name, $import_config, $subst_hash) = @_;

    for my $k (keys %$data) {
        my $row = $data->{$k};
        if ($row->{method_name}) {
            F5::ASMConfig::Entity::AllowedMethod->get_method(dbh => $self->{dbh},
                                                             name => $row->{method_name},
                                                             add_if_missing => 1,);
        }
   }

   return $RET_OK;
}

#----------------------------------------------------------------------

sub remove_duplicates {
  my($data, $uniq_col) = @_;
  my %seen_val;

  for my $k (sort { $a <=> $b } keys %$data) {
    if ($seen_val{$data->{$k}->{$uniq_col}}) {
      delete $data->{$k};
    } else {
      $seen_val{$data->{$k}->{$uniq_col}} = 1;
    }
  }

  return $RET_OK;
}

#--------------------------------------------------------------------
# chrset codes changed from old version to V32
sub fix_old_charset_types
{
   my ($data) = @_;
   return unless defined($data);
   foreach my $id (keys %$data)
   {
      if(! defined  $CHARSET_TYPE_UPGRADE_MAP{$data->{$id}->{'charset_type'}})
      {
         delete $data->{$id};
      }
      else
      {
         # old policies may contain incomplete charset (only 255 characters)
         if(length($data->{$id}->{'charset'}) == 255)
         {
             $data->{$id}->{'charset'}.= '0';
         }
         # the type codes had changed from V31 -> New versions
         $data->{$id}->{'charset_type'} = $CHARSET_TYPE_UPGRADE_MAP{$data->{$id}->{'charset_type'}};
      }
      
   }
}
#--------------------------------------------------------------------
# EncryptEntireBodyValue should be reset if no appropriate certificate exists
sub fix_ws_properties {
  my($self) = @_;

  return if F5::BigipUtils::is_docker();
  my $dbh = $self->{dbh};

  $dbh->do(<<EndQuery,undef, $self->{policy_id});
   UPDATE PLC.PL_XML_SOAP_SECURITY as A
LEFT JOIN PLC.PL_XML_PROFILE_CLIENT_CERTIFICATE as B
       ON A.profile_id=B.profile_id 
      SET A.flg_encrypt_whole_body_content = 0
    WHERE B.xml_client_id IS NULL
      AND A.policy_id = ?
EndQuery
}

#--------------------------------------------------------------------
#####################
# GENERAL FUNCTIONS #
#####################
#--------------------------------------------------------------------

sub init
{
    my($self, %params) = @_;

    $self->{msgs} = [];
    $self->{enforce_strictness} = $params{enforce_strictness} || 1;
    $self->{user_name} = $params{user_name} || 'SYSTEM';
    $self->{dbh} = $params{dbh} || F5::DbUtils::get_dbh();
    $self->{logger} = $params{logger};

    # ASM version and build number
    $CURRENT_TS_VER = 'ts_software_ts.ts.' . F5::BigipVersionUtils::bigip_version();
    $CURRENT_TS_BUILD = F5::BigipVersionUtils::bigip_build();
    # tables list
    @PLC_EXPORT_TABLES = F5::DbUtils::get_import_policy_tables('PLC');
    for my $table (@PLC_EXPORT_TABLES) {
      $table =~ s/.*?\.//;
    }

    $self->SUPER::init();

    $EXPORT_HASH_MAX_BATCH = $self->{cfg}->val('GENERAL', 'MaximumExportBatch');
    $self->{common_upgrade} = F5::CommonUpgrade->new(dbh => $self->{dbh},
                                                     logger => $self->{logger},
                                                     notification_handler => $self,
                                                    );
}

#--------------------------------------------------

sub write_to_export_file {
    my($self,$outf, $output_hash) = @_;
    my $rv;
    eval { $rv = store_fd($output_hash, $outf); };
    if (! defined $rv || $@)
    {
        my $err = $@ || 'unknown';
        $self->fatal_error(msg => "Failed to write record to exported policy file (store_fd failed - $err).");
    }
    return 1;
}

#--------------------------------------------------

sub read_from_export_file {
    my ($self, $inf) = @_;
    my $import_data;
    eval { $import_data = fd_retrieve($inf) };

    if (! defined $import_data || $@)
    {
        my $err = $@ || 'unknown';
        $self->fatal_error(user_msg => "Cannot read exported policy file.",
                           log_msg => "failed to fd_retrieve from gzip pipe ($err)");
        return undef;
    }
    return $import_data;
}

#----------------------------------------------------------------------

sub convert_renamed_columns {
  my($import_table, $import_table_name, $renamed_columns) = @_;

  for my $rec (values %$import_table) {
    for my $col (keys %$renamed_columns) {
      if ($rec->{$col}) {
        $rec->{$renamed_columns->{$col}} = $rec->{$col};
        delete $rec->{$col};
      }
    }
  }

  return $RET_OK;
}


#----------------------------------------------------------------------

sub pl_policies_fixup_fields {
  my($self, $data, $basic_configuration) = @_;

  for my $rec (values %$data) {
    $rec->{createdtime} = time;
    delete $rec->{modifytime};
    delete $rec->{modifier_name};
    $rec->{creator_name} = $self->{user_name};

    if (!$rec->{encoding}) {
        $rec->{encoding} = $basic_configuration->{ENCODING} || $IMPLIED_ENCODING;
    }

    $rec->{encoding} = lc($rec->{encoding});
    delete $rec->{policy_version};
    delete $rec->{revision};
    delete $rec->{flg_flowless_model};
  }
}

#----------------------------------------------------------------------

sub pl_param_attributes_before_fixup {
  my($import_table, $import_table_name, $import_config, $subst_hash) = @_;

  for my $rec (values %$import_table) {
    # Remove integer check range or decimal check range flag via mask
    $rec->{key_ui_type} &= ~( KEY_UI_TYPE_INTEGER_CHECK_RANGE | KEY_UI_TYPE_DECIMAL_CHECK_RANGE);
  }

  return $RET_OK;
}

#----------------------------------------------------------------------

sub generate_populate_defaults {
  my($default_table_name) = @_;

  my $populate_defaults_sub = sub {
    my($dbh, $logger, $import_table, $import_table_name, $import_config, $subst_hash) = @_;

    F5::DbUtils::copy_defaults_table(dbh => $dbh,
                                 table_name => $import_table_name,
                                 dbname => 'PLC',
                                 policy_id => $subst_hash ->{policy_id}{'*'},
                                 ignore => 1);
    return $RET_OK;
  };

  return $populate_defaults_sub;
}

#----------------------------------------------------------------------

sub create_negsig_idmap {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $table_contents = $params{table_contents};
  my $logger = $params{logger};
  my $table_name = $params{table_name};
  my $entity_name_fields = $params{entity_name_fields};
  my $id_field = $params{id_field};
  my $store_names = $params{store_names};
  my %idmap;

  $table_contents = [ values %$table_contents ] if (ref($table_contents) eq 'HASH');

  # Get factory values for this table
  my $factory_ids = $dbh->selectcol_arrayref("SELECT $id_field FROM PLC.$table_name WHERE flg_is_user_defined = 0");
  @idmap{@$factory_ids} = @$factory_ids;

  # Try to determine local id for user defined
  for my $row (@$table_contents) {
    next unless ($row->{flg_is_user_defined});
    my $where_fields = {};

    $where_fields->{flg_is_user_defined} = 1;

    for my $entity_name_field ( @$entity_name_fields) {
        $where_fields->{$entity_name_field} = $row->{$entity_name_field} || '';
    }

    my $constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => $where_fields);
    my $where_clause = $constructed_where_clause->{where_clause};
    my @where_values = @{$constructed_where_clause->{where_values}};

    my ($local_id) = $dbh->selectrow_array(<<EndQuery, undef, @where_values);
SELECT $id_field
  FROM PLC.$table_name
 WHERE $where_clause
EndQuery

    $idmap{$row->{$id_field}} = $local_id if $local_id;
  }

  if ($store_names) {
    for my $row (@$table_contents) {
      $store_names->{$row->{$id_field}} = $row->{$entity_name_fields->[0]};
    }
  }

  return \%idmap;
}

#----------------------------------------------------------------------
sub create_certificate_idmap {
  my(%params) = @_;
  my $dbh = $params{dbh};
  my $table_contents = $params{table_contents};
  my $logger = $params{logger};
  my $table_name = $params{table_name};
  my $entity_name_field = $params{entity_name_field};
  my $id_field = $params{id_field};
  my $store_names = $params{store_names};
  my %idmap;
  $table_contents = [ values %$table_contents ] if (ref($table_contents) eq 'HASH');
  for my $row (@$table_contents) {
        my ($local_id) = $dbh->selectrow_array(<<EndQuery, undef, $row->{$entity_name_field});
SELECT $id_field
  FROM PLC.$table_name
WHERE $entity_name_field = ?
EndQuery
  $idmap{$row->{$id_field}} = $local_id if $local_id;
  }
  if ($store_names) {
          for my $row (@$table_contents) {
             $store_names->{$row->{$id_field}} = $row->{$entity_name_field};
           }
   }
 return \%idmap;
}
#----------------------------------------------------------------------
sub issue_certificate_warnings {
  my($self, %params) = @_;
  my $table_contents = $params{table_contents};
  my $table_name = $params{table_name};
  my $subst_hash = $params{subst_hash};
  my $cert_names = $params{cert_names};
  my $dbh = $self->{dbh};
  for my $key (keys %$table_contents) {
    my $row = $table_contents->{$key};
    unless (exists $subst_hash->{ssl_cert_id}->{$row->{ssl_cert_id}}) {
      my $cert_name = $cert_names->{$row->{ssl_cert_id}} || $row->{ssl_cert_id};
      $self->add_warning(msg => qq{Certificate "$cert_name" (previously used in this policy) does not exist on this system.});
      delete $table_contents->{$key};
    }
  }
  return $RET_OK;
}
#----------------------------------------------------------------------
sub issue_negsig_set_warnings {
  my($self, %params) = @_;

  my $table_contents = $params{table_contents};
  my $table_name = $params{table_name};
  my $subst_hash = $params{subst_hash};
  my $set_names = $params{set_names};

  my $dbh = $self->{dbh};

  for my $key (keys %$table_contents) {
    my $row = $table_contents->{$key};
    unless (exists $subst_hash->{set_id}->{$row->{set_id}}) {
      my $set_name = $set_names->{$row->{set_id}} || $row->{set_id};
      $self->add_warning(msg => qq{Signature Set "$set_name" (previously used in this policy) does not exist on this system.});
      delete $table_contents->{$key};
    }
  }

  return $RET_OK;
}

#----------------------------------------------------------------------

my $USER_DEFINED_TEMPLATE_ID = 400000000;

sub validate_policy_template {
    my($self, $data, $import_table_name, $import_config, $subst_hash) = @_;
    my $dbh = $self->{dbh};
    for my $key (keys %$data) {
        my $row = $data->{$key};
        if ($row->{policy_template_id} >= $USER_DEFINED_TEMPLATE_ID ||
            ! F5::DbUtils::exists_in_table(dbh => $dbh,
                                           field_name => 'id',
                                           field_value => $row->{policy_template_id},
                                           table_name => 'PLC.POLICY_TEMPLATES')) {
            delete $data->{$key};
        }
    }
    return $RET_OK;
}

#----------------------------------------------------------------------

sub skip_if_missing_parent_record {
    my($self, %params) = @_;
    my $dbh = $self->{dbh};
    my $data = $params{data};
    my $parent_entity_class = $params{parent_entity_class};
    my $foreign_key_field_name = $params{foreign_key_field_name} || $parent_entity_class->id_field();
    my $add_warnings = defined $params{add_warnings} ? $params{add_warnings} : 1;

    my %missing_record_keys;
    for my $key (keys %$data) {
        my $row = $data->{$key};
        if (! F5::DbUtils::exists_in_table(dbh => $dbh,
                                           field_name => $foreign_key_field_name,
                                           field_value => $row->{$foreign_key_field_name},
                                           table_name => $parent_entity_class->db_table())) {
            $missing_record_keys{$row->{$foreign_key_field_name}} = 1;
            delete $data->{$key};
        }
    }

    if ($add_warnings) {
        my $missing_record_count = scalar keys %missing_record_keys;
        if ($missing_record_count > 1) {
            my $element_type_display_name_plural = $parent_entity_class->element_type_display_name_plural();
            $self->add_warning(msg => "$missing_record_count $element_type_display_name_plural previously used in this security policy were not found on this system. (IDs = " . join(', ', sort keys %missing_record_keys) . ")");
        }
        elsif ( $missing_record_count ) {
            my $element_type_display_name = $parent_entity_class->element_type_display_name();
            $self->add_warning(msg => "$missing_record_count $element_type_display_name previously used in this security policy was not found on this system. (IDs = " . join(', ', sort keys %missing_record_keys) . ")");
        }
    }
    return $RET_OK;
}

#----------------------------------------------------------------------

sub validate_sig_requirements {
    my($self, $data, $import_table_name, $import_config, $subst_hash) = @_;

    my $sig_req_obj = F5::ASMConfig::Entity::PolicySignatureRequirement->new(dbh => $self->{dbh});
    for my $key (keys %{$data}) {
        my $row = $data->{$key};
        my $store_obj = F5::ASMConfig::PolicySignatureRequirement->new(
                                                                        max_revision_datetime => $row->{max_revision_datetime},
                                                                        min_revision_datetime => $row->{min_revision_datetime},
                                                                        sig_tag               => $row->{sig_tag}
                                                                    );
        eval {
            $sig_req_obj->before_store(
                                        store_mode     => 'add',
                                        store_obj      => $store_obj,
                                        existing_obj   => {},
                                        validate_value => 1
                                    );
        };
        if ($@) {
            my $error_msg = $@;
            if (ref($error_msg)) {
                $error_msg = $error_msg->{error_message};
            }
            return $self->fatal_error(msg => $error_msg);
        }
    }
    return $RET_OK;
}

#----------------------------------------------------------------------

1;
