package F5::ImportExportPolicy::XML;

use strict;
use warnings;

use base qw(F5::ImportExportPolicy::Base);

use PassCrypt;

use F5::XML::Twig;
use File::Basename;
use YAML::XS;
use MIME::Base64;
use Time::Local;
use Encode qw(:all);
use Digest::MD5;

# Used to call xmllint externally
use IPC::Open3;
use Symbol;
use File::Spec;

use F5::Deploy;
use F5::CRC;
use F5::BatchInsert;
use F5::DbUtils;
use F5::NegativeSignatures;
use F5::NegativeSignatures::Set;
use F5::UrlContentUpgrade;
use F5::ResetStaging;
use F5::BigipVersionUtils;
use F5::CommonUpgrade;
use F5::ASMConfig::Entity::AllowedMethod;
use F5::ASMConfig::Entity::Violation;
use F5::ASMConfig::Entity::Cookie;
use F5::ASMConfig::Entity::URL;
use F5::ASMConfig::Entity::WebsocketUrl;
use F5::ASMConfig::Entity::Parameter;
use F5::ASMConfig::Entity::BruteForce;
use F5::ASMConfig::Entity::Filetype;
use F5::ASMConfig::Entity::RedirectionDomain;
use F5::ASMConfig::Entity::Policy;
use F5::ASMConfig::Entity::PolicySignatureRequirement;
use F5::NwdUtils;
use F5::XMLUtils;
use F5::XML::Schema::Builder;
use F5::BigipUtils;
use F5::Utils::Rest;
use F5::ASMConfig::Utils;
use F5::ASMConfig::Cache;
use F5::ThreatCampaigns;

my $UNLIKELY_DELIMITER = '###UNLIKELY_DELIMITER###';
my $ASM_IMPORT_EXPORT_XML_CONFIG_FILENAME = F5::Deploy::map_path('/usr/share/ts/config/import_export_xml_policy.yaml');
my $MAX_MEMORY_USAGE = 200_000_000; # 200 MB approximately
my $MEMORY_CHECK_INTERVAL = 200;
my $INTEGRITY_CHECK_REPLACE_STRING = 'UNLIKELY_STRING_FOR_INTEGRITY_CHECK_REPLACE';

use constant RECORD_STATE_DEFER => 'defer';
use constant RECORD_STATE_DEFER_NON_FATAL => 'defer_non_fatal';
use constant FIELD_STATE_IGNORE => 'ignore';

sub new {
  my($class, %params) = @_;
  my $self = {};
  bless $self, $class;
  $self->init(%params);
  return $self;
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;

  if (exists $params{dbh}) {
    $self->{dbh} = $params{dbh};
  }
  else {
    $self->{dbh} = F5::DbUtils::get_dbh();
  }
  $self->{logger} = $params{logger};
  eval {
      require Encode::HanExtra;
  };

  $self->SUPER::init();

  $self->{batch_insert} = F5::BatchInsert->new(dbh => $self->{dbh});
  my $config_filename = $ENV{ASM_IMPORT_EXPORT_XML_CONFIG_FILENAME} || $ASM_IMPORT_EXPORT_XML_CONFIG_FILENAME;
  $self->{config} = YAML::XS::LoadFile($config_filename);
  $self->{pattern_order_last} = {};
  $self->{table_fields} = {};
  $self->{seen_during_import} = {};
  $self->{seen_during_import_tables} = {};
  $self->{foreign_key_save} = {};
  $self->{deferred_records} = [];
  $self->{accumulated_fatal_errors} = [];
  $self->{primary_encoding} = 'utf-8';
  $self->{flg_encoding_selected} = 0;
  $self->{account_id} = 0;
  $self->{common_upgrade} = F5::CommonUpgrade->new(dbh => $self->{dbh},
                                                   logger => $self->{logger},
                                                   upgrade_source => F5::ASMConfig::PolicyFileFormat::POLICY_FILE_FORMAT_XML(),
                                                   notification_handler => $self,
                                                  );
  $self->{idmap} = {};
  $self->{seen_negsig_set_ids} = {};
  $self->{initial_memory_usage} = F5::NwdUtils::get_process_stat()->{vmsize};
  $self->{memory_check_count} = 0;
  $self->{entity_schema} = F5::ASMConfig::Utils::get_entity_schema();

}

#----------------------------------------------------------------------

sub export_policy {
  my($self, %params) = @_;
  
  delete $self->{user_defined_sig_id_to_sig_name}; # Clear cache just in case

  my $dbh = $self->{dbh};
  my $policy_id = $params{policy_id};
  my $minimal_export = $params{minimal_export};
  my $include_scanner_tables = defined $params{include_scanner_tables} ? $params{include_scanner_tables} : 1;

  $self->{policy} = $params{policy};
  $self->{primary_encoding} = $self->get_primary_encoding($policy_id);
  my $base_twig = $self->get_export_base();

  eval {
      for my $node (@{$self->{config}->{nodes}}) {
          $self->export_node(node => $node,
                             policy_id => $policy_id,
                             twig => $base_twig,
                             minimal_export => $minimal_export,
                             include_scanner_tables => $include_scanner_tables,
                            );
      }
  };

  if ($@) {
      my $error_msg = $@;
      if (ref($@)) {
          $error_msg = $@->{error_msg};
      }
      return $self->fatal_error(msg => $error_msg);
  }

  my @deferred_records = @{$self->{deferred_records}};
  for my $deferred_record (@deferred_records) {
    my $defer_func = $deferred_record->{defer_func};
    $self->$defer_func(%$deferred_record);
  }

  my $integrity_check = Digest::MD5->new->add(encode_utf8($base_twig->sprint))->hexdigest;
  $base_twig->root->set_att("integrity_check" => $integrity_check);

  open(my $outf, '>:utf8', $params{filename});
  $base_twig->print($outf);
  close $outf;

  ($self->{policy_name}) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id);
SELECT name
  FROM PLC.PL_POLICIES
 WHERE id = ?
EndQuery
  $self->{policy_id} = $policy_id;

  return $self->completed_successfully(msg => "Policy '$self->{policy_name}' was successfully exported.");
}

#----------------------------------------------------------------------

sub export_node {
  my($self, %params) = @_;

  my $policy_id = $params{policy_id};
  my $node = $params{node};
  my $dbh = $self->{dbh};

  return if ($node->{for_backward_compatibility});

  $self->fixup_node(%params);
  return unless (exists $node->{xpath});
  return if ($node->{import_only});

  if ($node->{policy_export_flag} && ! $params{$node->{policy_export_flag}}) {

      $self->{logger}->debug("Export $node->{policy_export_flag} not selected -- $node->{xpath} skipped") if $self->{logger};
      return;
  }

  my $policy_id_field = ($node->{db_tables}->[0] eq 'PLC.PL_POLICIES' ? 'id' : 'policy_id');
  my @where_clauses = ("$node->{db_tables}->[0].$policy_id_field = ?");
  my @bind_vals = ($policy_id);
  unless ($node->{limit_clause}) {
    $node->{limit_clause} = {};
  }
  if ($params{limit_clause}) {
    $node->{limit_clause} = { %{$node->{limit_clause}},
                              %{$params{limit_clause}},
                            };
  }
  for my $field (keys %{$node->{limit_clause}}) {
    if ($node->{limit_clause}->{$field} =~ s/^!= //) {
      push(@where_clauses, "$field != ?");
    }
    else {
      push(@where_clauses, "$field = ?");
    }
    push(@bind_vals, $node->{limit_clause}->{$field});
  }

  unless ($node->{export_limit_clause}) {
    $node->{export_limit_clause} = {};
  }
  if ($params{export_limit_clause}) {
    $node->{export_limit_clause} = { %{$node->{export_limit_clause}},
                              %{$params{export_limit_clause}},
                            };
  }
  for my $field (keys %{$node->{export_limit_clause}}) {
    if ($node->{export_limit_clause}->{$field} =~ s/^!= //) {
      push(@where_clauses, "$field != ?");
    }
    else {
      push(@where_clauses, "$field = ?");
    }
    push(@bind_vals, $node->{export_limit_clause}->{$field});
  }

  if ($node->{db_table_link}) {
    my $join_where_clause = join(' = ', map { "$_.$node->{db_table_link}->{$_}" } keys %{$node->{db_table_link}});
    push(@where_clauses, $join_where_clause);
  }

  my $where_clause = 'WHERE ' . join(' AND ', @where_clauses);

  my $order_by_clause = '';
  if ($node->{order_by}) {
    $order_by_clause = "ORDER BY $node->{order_by}";
  }
  my $tables = join(',', @{$node->{db_tables}});
  my @select_fields = ('*');
  if ($node->{id_field}) {
    push(@select_fields, "$node->{id_field} as id_field");
  }
  my $select_fields = join(',', @select_fields);

  my $sth = $dbh->prepare(<<EndQuery);
SELECT $select_fields
  FROM $tables
$where_clause
$order_by_clause
EndQuery

  $sth->execute(@bind_vals);

  my $found_nodes = 0;
  while (my $db_rec = $sth->fetchrow_hashref) {
    $found_nodes++;
    # check if memory usage has grown too much
    if ($self->{memory_check_count}++ % $MEMORY_CHECK_INTERVAL == 0) {
        my $current_memory = F5::NwdUtils::get_process_stat()->{vmsize};
        if ($current_memory && $current_memory - $self->{initial_memory_usage} > $MAX_MEMORY_USAGE) {
            die "The security policy is too large and cannot be exported as XML. Please use regular export instead.\n";
        }
    }
   
    if ($node->{defaults} && $node->{defaults}->{export_non_default_only}) {
        my $default_table = $node->{defaults}->{defaults_table_name} || "${tables}_DEFAULTS";
        my $join_field = $node->{defaults}->{export_non_default_only}->{join_field};
        my @where_fields = ($join_field);
        my @where_values = ($db_rec->{$join_field});
        if ($node->{defaults}->{where_values}) {
            for my $field (keys %{$node->{defaults}->{where_values}}) {
                push (@where_fields, $field);
                push (@where_values, $node->{defaults}->{where_values}->{$field});
            }
        }
        my $where_clause = join(" AND ", map { "$_ = ?" } @where_fields);
        my $default_row = $dbh->selectrow_hashref(<<EndQuery, undef, @where_values);
SELECT * 
 FROM $default_table
 WHERE $where_clause
EndQuery
        if ($default_row) {
            my $matches_defaults = 1;
            for my $default_field (keys %$default_row) {
                if (exists $db_rec->{$default_field}               
                    && ((defined($default_row->{$default_field}) != defined($db_rec->{$default_field})) 
                     || (defined($default_row->{$default_field}) && $default_row->{$default_field} ne $db_rec->{$default_field}))
                        ) {
                    $matches_defaults = 0;
                   last;
                }
            }
            next if $matches_defaults;
        }
    }

    if ($node->{db_tables}->[0] eq 'PLC.PL_POLICIES') {
        # Copy other policy attributes (such as policy_version) to %db_rec
        for my $k (keys %{$self->{policy}}) {
            if (! exists $db_rec->{$k}) {
                $db_rec->{$k} = $self->{policy}->{$k};
            }
        }
    }

    my $node_twig;
    if ($node->{add_to_existing_node}) {
      $node_twig = F5::XML::Schema::Builder::add_or_find_node($params{twig}, $node->{xpath});
    }
    else {
      $node_twig = XML::Twig::Elt->new(basename($node->{xpath}));
      $node_twig->paste(last_child => F5::XML::Schema::Builder::add_or_find_node($params{twig}, dirname($node->{xpath})));
    }

    if ($node->{row_func}) {
      my $function_name = $node->{row_func};
        $self->$function_name(node => $node,
                              db_rec => $db_rec,
                              node_twig => $node_twig,
                              mode => 'export',
                             );
    }
    else {
      my $record_state = 0;
      my $save_rec = {};

      for my $field (@{$node->{fields}}) {
        $self->export_field(
                            field => $field,
                            node => $node,
                            element => $node_twig,
                            db_rec => $db_rec,
                            save_rec => $save_rec,
                            record_state => \$record_state,
                           );
        if ($record_state eq 'skip') {
            my $print_field_value = '';
            if ($field->{db_field}) {
                if (defined $db_rec->{$field->{db_field}}) {
                    $print_field_value = "($field->{db_field} = $db_rec->{$field->{db_field}})";
                }
                else {
                    $print_field_value = "($field->{db_field} is undefined)";
                }
            }
            $self->{logger}->debug("Record from $node->{db_tables}->[0] could not be exported $print_field_value -- skipping") if $self->{logger};
            last;
        }
      }

      if ($record_state eq 'skip') {
          $node_twig->delete;
          next;
      }

      if ($node->{limit_clause}) {
          for my $k (keys %{$node->{limit_clause}}) {
              $save_rec->{$k} = $node->{limit_clause}->{$k};
          }
      }

      if ($node->{save_for_foreign_key}) {
        $self->{foreign_key_save}->{$node->{db_tables}->[0]}->{$db_rec->{id_field}} = $save_rec;
      }

      # Add subnodes
      for my $subnode (@{$node->{subnodes}}) {
        if (! $self->check_dependent_field_values(dependent_field_values => $subnode->{included_only_if},
                                                  db_rec => $db_rec)) {
            # Skip this subnode
            next;
        }

        my $parent_id;
        if ($subnode->{subnode_connection}->{parent_db_field}) {
          $parent_id = $db_rec->{$subnode->{subnode_connection}->{parent_db_field}};
        }
        else {
          $parent_id = $db_rec->{id_field};
        }
        $self->export_node(node => $subnode,
                           policy_id => $policy_id,
                           twig => $node_twig,
                           limit_clause => {
                                            $subnode->{subnode_connection}->{child_db_field} => $parent_id
                                           },
                           minimal_export => $params{minimal_export},
                          );
      }
    }
  }
  if (!$found_nodes && $node->{defaults} && !$node->{defaults}->{single_row}) {
      #Export an empty node so the xpath will be found on import, and full defaults will not be created.
      my $node_twig;
      if ($node->{add_to_existing_node}) {
          $node_twig = F5::XML::Schema::Builder::add_or_find_node($params{twig}, $node->{xpath});
      }
      else {
          $node_twig = XML::Twig::Elt->new(basename($node->{xpath}));
          $node_twig->paste(last_child => F5::XML::Schema::Builder::add_or_find_node($params{twig}, dirname($node->{xpath})));
      }
  }
}

#----------------------------------------------------------------------

sub import_policy {
  my($self, %params) = @_;

  # Clear cache just in case
  delete $self->{user_defined_sig_name_to_sig_id};
  delete $self->{sig_ids};

  my $get_policy_attributes_only = $params{get_policy_attributes_only};
  $self->{policy_description} = $params{policy_description};
  $self->{policy_type} = $params{policy_type};
  $self->{creator_name} = $params{creator_name};

  if ($params{encoding}) {
      $self->{primary_encoding} = $params{encoding};
      $self->{flg_encoding_selected} = 1;
  }
  elsif (defined($params{flg_encoding_selected})) {
      $self->{flg_encoding_selected} = $params{flg_encoding_selected};
  }

  if (F5::BigipUtils::is_device_bigip()){
      eval {
        check_schema($params{filename});
      };
      if ($@) {
        return $self->fatal_error(msg => $@);
      }
  }

  my $passes_integrity_check = $self->passes_integrity_check(filename => $params{filename});
  if (! $passes_integrity_check)  {
      $self->add_warning(msg => 'Security policy version information will be ignored, since the file has been modified since it was exported.');
  }

  my $policy_id;
  my $policy_changes_log_params = {};
  my $policy_was_inserted = 0;

  # Build twig handlers
  my %handlers;
  my %start_tag_handlers;
  my %nodes_by_xpath;
  for my $node (@{$self->{config}->{nodes}}) {
    next unless ($node->{xpath});

    unless ($nodes_by_xpath{$node->{xpath}}) {
      $nodes_by_xpath{$node->{xpath}} = [];
    }

    push(@{$nodes_by_xpath{$node->{xpath}}}, $node);
  }

  for my $xpath (keys %nodes_by_xpath) {
    $handlers{$xpath} = sub {
        my($t, $elt) = @_;

        if ($get_policy_attributes_only) {
            # Failsafe in case encoding is missing
            $t->finish_now;
        }
        elsif(!$policy_was_inserted){
            my $rv = $self->insert_new_policy(force_policy_id => $params{force_policy_id}, allow_encoding_change => $params{allow_encoding_change});
            if ($self->{primary_encoding} ne 'utf-8') {
                $self->{encoding_changed} = 1;
            }
            $policy_id = $rv->{policy_id};
            $policy_changes_log_params = $rv->{policy_changes_log_params};
            $policy_was_inserted = 1;
        }

        for my $node (@{$nodes_by_xpath{$xpath}}) {
            if ($node->{prehandle_func}) {
                my $prehandle_func = $node->{prehandle_func};
                $self->$prehandle_func(
                    element => $elt,
                    node => $node,
                    policy_id => $policy_id,
                    import_policy_params => \%params,
                );
            }
            $self->import_node(element => $elt,
                               node => $node,
                               policy_id => $policy_id,
                               import_policy_params => \%params,
                              );
        }
    };

    $start_tag_handlers{$xpath} = sub { 
        $self->{last_start_tag_position} = $_[0]->current_line;
    };
  }

  $handlers{'policy/type'} = sub {
     my ($t, $elm) = @_;
     if (!$self->{policy_type}) {
         my $type = $elm->text;
         my %mapping = reverse %{F5::ASMConfig::Entity::Policy->fields->{policy_type}->{display_value_map}};
         $self->{policy_type} = $mapping{$type};
     }
  };

  $handlers{'policy/parent_policy_name'} = sub {
     my ($t, $elm) = @_;
     $self->{parent_policy_name} = $elm->text;
  };

  $handlers{'policy/policy_version/timestamp'} = sub {
    my ($t, $elm) = @_;
    if ($passes_integrity_check) {
        $self->{policy_version_time} = F5::DateTimeUtils::convert_any_datetime_to_epoch($elm->text);
    }
  };

  $handlers{'policy/policy_version/device_name'} = sub {
    my ($t, $elm) = @_;
    if ($passes_integrity_check) {
        $self->{policy_version_device_name} = $elm->text;
    }
  };

  $handlers{'policy/policy_version/policy_name'} = sub {
    my ($t, $elm) = @_;
    if ($passes_integrity_check) {
        $self->{policy_version_policy_name} = $elm->text;
    }
  };

  $handlers{'policy/policy_version/last_change'} = sub {
    my ($t, $elm) = @_;
    if ($passes_integrity_check) {
        $self->{policy_version_last_change} = $elm->text;
    }
  };

  $handlers{'policy/encoding'} = sub {
    my ($t, $policy_encoding) = @_;
    my $original_encoding = lc($policy_encoding->text);
    if (!$params{encoding}) {
        $self->{primary_encoding} = $original_encoding;
        if (!defined($params{flg_encoding_selected})) {
            $self->{flg_encoding_selected} = 1;
        }
    }
    elsif( $params{encoding} ne $original_encoding ) {
        $self->{encoding_changed} = 1;
    }

    if ($get_policy_attributes_only) {
        $t->finish_now;
    }
    elsif(!$policy_was_inserted){
        my $rv = $self->insert_new_policy(force_policy_id => $params{force_policy_id}, allow_encoding_change => $params{allow_encoding_change});
        $policy_id = $rv->{policy_id};
        $policy_changes_log_params = $rv->{policy_changes_log_params};
        $policy_was_inserted = 1;
    }
  };

  $handlers{'policy/attack_signatures/sigfile_timestamp'} = sub {
     my ($t, $elm) = @_;
     $self->{import_sigfile_timestamp} = $elm->text;
  };

  my $twig = F5::XML::Twig->new(twig_handlers => \%handlers, 
                            start_tag_handlers => {policy => 
                                                     sub {
                                                       my( $t, $policy)= @_;
                                                       $self->get_bigip_version(mode => 'import',
                                                                                field_value => defined $policy->{att}->{bigip_version}
                                                                                               ? $policy->{att}->{bigip_version}
                                                                                               : '10.1.0'
                                                                               ); 
                                                       my $policy_name = $params{policy_name} || $policy->{att}->{name};
                                                       $self->{orig_policy_name} = $self->policy_name_fixup(%params,
                                                                                                            policy_name => $policy_name,);
                                                     },
                                                   %start_tag_handlers,
                                                  }
                           );
  eval {
    $twig->parsefile($params{filename});
  };
  if ($@) {
    my $error_msg = $@;
    if ( !( $error_msg =~ /Failed action/ ) ){
        $error_msg = 'Malformed XML: ' . $error_msg;
    }
    else {
        $self->{msgs} = [];
    }
    $error_msg =~ s/ at \/usr\/local\/share\/.*//g;
    $error_msg =~ s/\n//g;
    $self->{dbh}->rollback;
    return $self->fatal_error(msg => $error_msg, msgs => @{$self->{accumulated_fatal_errors}} ? $self->{accumulated_fatal_errors} : undef);
  }

  $self->{policy_type} ||= F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY(); #To prevent warnings in case it was not set from anywhere else
  if ($get_policy_attributes_only) {
      $self->{policy} = F5::ASMConfig::Policy->new({
                                                    policy_name => $self->{orig_policy_name},
                                                    policy_type => $self->{policy_type},
                                                    encoding => $self->{primary_encoding},
                                                    flg_encoding_selected => $self->{flg_encoding_selected},
                                                    policy_version_time => $self->{policy_version_time},
                                                    policy_version_device_name => $self->{policy_version_device_name},
                                                    policy_version_policy_name => $self->{policy_version_policy_name},
                                                    policy_version_last_change => $self->{policy_version_last_change},
                                                    parent_policy_name => $self->{parent_policy_name},
                                                   });
      $self->{policy_name} = $self->{orig_policy_name};
      return $self->completed_successfully(msg => '');
  }

  if (@{$self->{accumulated_fatal_errors}}) {
      $self->{dbh}->rollback;
      return $self->fatal_error(msgs => $self->{accumulated_fatal_errors});
  }

  # Finish deferred records
  my %check_records = (check_records => 1,
                       num_records  => scalar @{$self->{deferred_records}});
  push (@{$self->{deferred_records}}, \%check_records);

  while (my $deferred_record = shift @{$self->{deferred_records}}) {
    if ($deferred_record->{check_records}) {
        last if !@{$self->{deferred_records}};

        if ($deferred_record->{num_records} > scalar @{$self->{deferred_records}}) {
            $deferred_record->{num_records} = scalar @{$self->{deferred_records}};
            push (@{$self->{deferred_records}}, $deferred_record);
        }
        else {
            if (! grep { $_->{record_state} && $_->{record_state} ne RECORD_STATE_DEFER_NON_FATAL } @{$self->{deferred_records}}) {
                # All remaining deferred records are tolerable - log a warning, but complete successfully
                $self->add_warning(msg => 'Some records were not added due to missing foreign key dependencies.');
                last;
            }

            $self->{dbh}->rollback;
            return $self->fatal_error(msg => 'Malformed XML: Could not resolve foreign key dependence');
        }
    }
    else {
        my $defer_func = $deferred_record->{defer_func};
        $self->$defer_func(%$deferred_record);
    }
  }

  $self->{batch_insert}->flush;

  for my $table_name (keys %{$self->{seen_during_import_tables}}) {
      $self->{common_upgrade}->after_table_complete(table_name => $table_name,
                                                    policy_id  => $policy_id,
                                                    dbh        => $self->{dbh},
                                                    logger     => $self->{logger});

  }

  F5::ASMConfig::Cache::clear_cached_policy(policy => { policy_id => $policy_id },);
  $self->add_policy_defaults(policy_id => $policy_id, twig => $twig);

  $self->{common_upgrade}->after_all_tables_complete(policy_id  => $policy_id, dbh => $self->{dbh}, idmap => $self->{idmap});

  $self->{batch_insert}->flush;

  $self->post_import_fixups(policy_id => $policy_id);

  if (defined $policy_changes_log_params->{policy_id}) {
      $self->update_pl_policy_changes_log(policy_id => $policy_changes_log_params->{policy_id})
  }

  return $self->completed_successfully(msg => "The operation was completed successfully. The security policy name is '$self->{policy_name}'.");
}

#----------------------------------------------------------------------

sub passes_integrity_check {
    my($self, %params) = @_;

    my $md5_accumulator = Digest::MD5->new;
    my $extracted_integrity_check;
    open(my $inf, '<', $params{filename});
    while(defined(my $line = <$inf>)) {
        if (! $extracted_integrity_check) {
            if ($line =~ s/(<policy .*?integrity_check=")(.*?)(")/$1$INTEGRITY_CHECK_REPLACE_STRING$3/) {
                $extracted_integrity_check = $2;
            }
        }
        $md5_accumulator->add($line);
    }
    close $inf;

    return (! $extracted_integrity_check ||
            $extracted_integrity_check eq $md5_accumulator->hexdigest);
}

#----------------------------------------------------------------------


sub update_policy_attributes {
    my ($self, %params) = @_;
    my ($policy_node) = grep { (defined $_->{db_tables}->[0]) && ($_->{db_tables}->[0] eq 'PLC.PL_POLICIES') } @{$self->{config}->{nodes}};
    $self->import_node(
        %params,
        element => $params{element}->parent,
        node => $policy_node,
    );
}

#----------------------------------------------------------------------

sub get_export_base {
  my($self, %params) = @_;
  my $xml_document_template = <<EndDoc;
<?xml version="1.0" encoding="utf-8"?>
<policy />
EndDoc

  my $twig = F5::XML::Twig->new(pretty_print => 'indented',
                            output_text_filter => sub {
                              my $str = shift;
                              return decode($self->{primary_encoding} => $str, Encode::FB_HTMLCREF);
                            }
                           );
  $twig->parse($xml_document_template);

  return $twig;
}

#----------------------------------------------------------------------

sub insert_new_policy {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my %new_policy_rec = (
                        account_id => 0,
                        name => "newly_created_xml_policy_$$",
                        creator_name => $self->{creator_name} || 'SYSTEM',
                        createdtime => time,
                        flg_encoding_selected => $self->{flg_encoding_selected},
                        encoding => $self->{primary_encoding},
                       );

  my $policy_id;
  my $existing_policies;
  if ($params{force_policy_id}) {
      $new_policy_rec{id} = $params{force_policy_id};
      $policy_id = $params{force_policy_id};
      $existing_policies = $dbh->selectall_hashref('SELECT id, account_id, flg_encoding_selected, encoding FROM PLC.PL_POLICIES', 'id');
  }


  my %policy_changes_log_params = ();
  if ($params{force_policy_id} && $existing_policies->{$params{force_policy_id}}) {
      if ($self->{encoding_changed} && !$params{allow_encoding_change}) {
          die "Failed action: Imported and replaced policies have different encodings.\n";
      }

      $self->overwrite_pl_policies(row_data => \%new_policy_rec,
                                   policy_id => $policy_id);

      $policy_changes_log_params{policy_id} = $policy_id;

      $self->{account_id} = $existing_policies->{$params{force_policy_id}}->{account_id};
  }
  else {
      $self->insert_row(table_name => 'PLC.PL_POLICIES',
                        rec => \%new_policy_rec,
                        get_deferred_id => \$policy_id,
                       );

      $policy_id = $policy_id->get_value;
  }
  $self->{policy_id} = $policy_id;

  # Call common_upgrade fixups for PL_POLICIES to cache values for later usage
  my $policy_row_fixup = $self->{common_upgrade}->get_row_fixup_callback(table_name => "PLC.PL_POLICIES");
  $policy_row_fixup->(row => { %new_policy_rec,
                               id => $self->{policy_id},
                             },
                      idmap => $self->{idmap},
                      dbh => $dbh);

  return {policy_id => $policy_id,
          policy_changes_log_params => { %policy_changes_log_params },
         };
}

#----------------------------------------------------------------------

sub import_node {
  my($self, %params) = @_;
  my $element = $params{element};
  my $node = $params{node};

  return if ($node->{for_backward_compatibility});

  $self->fixup_node(node => $node);
  if ($node->{xpath}) {
    $self->{seen_during_import}->{$node->{xpath}} = 1;
    $self->{seen_during_import_tables}->{$node->{db_tables}->[0]} = 1;
  }

  my $record_state = 0;
  my %db_rec;

  for my $field (keys %{$node->{limit_clause}}) {
      if ($node->{limit_clause}->{$field} !~ /^!= /) {
          $db_rec{$field} = $node->{limit_clause}->{$field};
      }
  }

  my $save_rec = {};
  for my $field (@{$node->{fields}}) {
    $self->import_field(field => $field,
                        node => $node,
                        element => $element,
                        save_rec => $save_rec,
                        db_rec => \%db_rec,
                        record_state => \$record_state,
                        policy_id => $params{policy_id},
                       );

  }
  if ($node->{row_func}) {
    my $function_name = $node->{row_func};
    $self->$function_name(mode => 'import',
                          node => $node,
                          db_rec => \%db_rec,
                          element => $element,
                         );
  }

  return unless (%db_rec);
  die if ($record_state eq 'error');
  return if ($record_state eq 'skip');

  if ($params{db_field}) {
    %db_rec = (%db_rec,
               %{$params{db_field}});
  }

  $db_rec{policy_id} = $params{policy_id};
  my ($foreign_key, $foreign_key_save_table);
  if ($node->{save_for_foreign_key}) {

    my @key_fields;
    for my $foreign_key_field (@{$node->{save_for_foreign_key}->{fields}}) {
        my $foreign_key_field_value;
        if (exists $node->{limit_clause}->{$foreign_key_field}) {
            $foreign_key_field_value = $node->{limit_clause}->{$foreign_key_field};
        }
        else {
            $foreign_key_field_value = F5::XMLUtils->get_value_from_xpath(element => $element, xpath => $foreign_key_field);
        }
        push(@key_fields, $foreign_key_field_value);
    }
    $foreign_key = join($UNLIKELY_DELIMITER, map { defined($_) ? $_ : '' } @key_fields);
    my $foreign_key_save;
    if ($params{subnode_foreign_key_save}) {
      $foreign_key_save = $params{subnode_foreign_key_save};
    }
    else {
      $foreign_key_save = $self->{foreign_key_save};
    }

    unless ($foreign_key_save->{$node->{db_tables}->[0]}) {
      $foreign_key_save->{$node->{db_tables}->[0]} = {};
    }
    $foreign_key_save_table = $foreign_key_save->{$node->{db_tables}->[0]};

    if (exists $foreign_key_save_table->{$foreign_key}) {
        $self->add_fatal_error(error_msg => sprintf('Duplicate record \'%s\' for element \'%s\'',
                                                    $save_rec->{'@name'},
                                                    $node->{xpath},
                                                   ),
                               %params,
                              );
        return;
    }
  }

  my @insert_tables = @{$node->{db_tables}};
  for my $i (0..$#{$node->{db_tables}}) {
      if (my $common_upgrade_callback = $self->{common_upgrade}->get_row_fixup_callback(table_name => $node->{db_tables}->[$i])) {
          my $row_state_rv;
          $common_upgrade_callback->(row => \%db_rec,
                                     idmap => $self->{idmap},
                                     dbh => $self->{dbh},
                                     row_state_rv => \$row_state_rv,
                                    );
          if ($row_state_rv && $row_state_rv eq 'skip') {
              return;
          }
      }
      my $new_table_name = $self->{common_upgrade}->get_new_table_name(table_name => $node->{db_tables}->[$i]);
      if (defined $new_table_name) {
          $insert_tables[$i] = $new_table_name ? "PLC.$new_table_name" 
                                               : $new_table_name; #If empty string, keep empty string
      }
  }

  my $deferred_id;
  if ($node->{db_tables}->[0] eq 'PLC.PL_POLICIES') {
    if ($db_rec{name}) {
      # determine next available name
      $db_rec{name} = $self->{orig_policy_name};
      $db_rec{name} = $self->get_unique_policy_name(%{$params{import_policy_params}},
                                                    policy_name => $db_rec{name},);
      $self->{policy_name} = $db_rec{name};
      my $rest_uuid = F5::Utils::Rest::generate_uuid($db_rec{name});
      $db_rec{rest_uuid} = $rest_uuid;
      $db_rec{policy_name_crc} = F5::CRC::get_crc($db_rec{name});
    }

    if ($self->{policy_type}) {
        $db_rec{policy_type} = $self->{policy_type};
    }
    else {
        $self->{policy_type} = $db_rec{policy_type} = F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY();
    }

    if ($self->{policy_description}) {
        $db_rec{description} = $self->{policy_description};
    }
    $db_rec{encoding} = $self->{primary_encoding};
    $db_rec{flg_encoding_selected} = $self->{flg_encoding_selected};

    if( !$db_rec{flg_encoding_selected} && defined($db_rec{policy_type}) &&
        $db_rec{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT() ) {
      die "Failed action: Cannot import a Parent Policy without an encoding\n";
    }

    delete $db_rec{policy_id};
    F5::DbUtils::update_row(dbh => $self->{dbh},
                            table_name => 'PLC.PL_POLICIES',
                            rec => \%db_rec,
                            where_fields => { id => $params{policy_id} });
    $self->{flg_is_protocol_free} = $db_rec{flg_is_protocol_free} || 0;

  }
  elsif ($record_state eq RECORD_STATE_DEFER || $record_state eq RECORD_STATE_DEFER_NON_FATAL) {
    push(@{$self->{deferred_records}},
         { defer_func => 'import_node',
           record_state => $record_state,
           %params });
  }
  elsif (scalar @{$node->{db_tables}} > 1 && $node->{db_table_link}) {
    $self->insert_row(table_name => $insert_tables[0],
                      rec => $self->filter_rec_for_table(rec => \%db_rec,
                                                         table_name => $insert_tables[0],
                                                        ),
                      get_deferred_id => \$deferred_id
                     );
    my %deferred_opts = ();
    if ($deferred_id) {
        $db_rec{$node->{db_table_link}->{$insert_tables[1]}} = $deferred_id;

        if ($node->{id_field} && $node->{id_field} =~ /$insert_tables[1]/) {
            $deferred_opts{get_deferred_id} = \$deferred_id;
        }
    }

    $self->insert_row(table_name => $insert_tables[1],
                      rec => $self->filter_rec_for_table(rec => \%db_rec,
                                                         table_name => $insert_tables[1],
                                                        ),
                      %deferred_opts,
                     );

  }
  elsif ($node->{subnodes} || $node->{save_for_foreign_key}) {
    $self->insert_row(table_name => $insert_tables[0],
                      rec => \%db_rec,
                      get_deferred_id => \$deferred_id
                     );
  }
  else {
    for my $table_name (@insert_tables) {
        $self->insert_row(table_name => $table_name,
                          rec => \%db_rec);
    }
  }

  if ($node->{save_for_foreign_key}) {
    $foreign_key_save_table->{$foreign_key} = $deferred_id;
  }

  if ($node->{subnodes}) {
    for my $subnode (@{$node->{subnodes}}) {
      my $subnode_foreign_key_save = $subnode->{save_for_foreign_key} ? undef :  {};

      my $parent_db_field = $deferred_id;
      if ($subnode->{subnode_connection}->{parent_db_field}) {
        $parent_db_field = $db_rec{$subnode->{subnode_connection}->{parent_db_field}};
      }
      for my $subelement ($element->findnodes($subnode->{xpath})) {
        $self->import_node(element => $subelement,
                           node => $subnode,
                           policy_id => $params{policy_id},
                           db_field => {
                                        $subnode->{subnode_connection}->{child_db_field} => $parent_db_field
                                       },
                           subnode_foreign_key_save => $subnode_foreign_key_save,
                          );
      }
    }
  }

  if ($node->{after_complete}) {
      $self->{batch_insert}->flush;
      my $func_name = $node->{after_complete};
      $self->$func_name(%params);
  }
}

#----------------------------------------------------------------------

sub import_field {
  my($self, %params) = @_;
  my $field = $params{field};
  my $node = $params{node};
  my $element = $params{element};
  my $db_rec = $params{db_rec};
  my $save_rec = $params{save_rec};
  my $record_state = $params{record_state};

  return if ($field->{export_only});

  my $field_value;
  my $import_xpath = defined($field->{import_xpath}) ? $field->{import_xpath}
                                                     : $field->{xpath};
  if (defined($import_xpath)) {    
    $field_value = F5::XMLUtils->get_value_from_xpath(element => $element,
                                               xpath => $import_xpath);

    return unless defined($field_value);
  }

  if ($node->{save_for_foreign_key} && $field->{xpath}) {
    $save_rec->{$field->{xpath}} = $field_value;
  }

  if ($field->{field_func}) {
    my $field_func = $field->{field_func};
    $field_value = $self->$field_func(mode => 'import',
                                      field_value => $field_value,
                                      field => $field,
                                      db_rec => $db_rec,
                                      element => $element,
                                      node => $node,
                                      record_state => $record_state,
                                      policy_id => $params{policy_id},
                                     );
  }
  return if ($record_state eq RECORD_STATE_DEFER || $record_state eq RECORD_STATE_DEFER_NON_FATAL);

  if($field->{encode_output} || ! $field->{field_func}) {
    if ($self->{primary_encoding} ne 'utf-8' && $field_value) {
      $field_value = encode($self->{primary_encoding}, $field_value);
    }
  }
 
  return if ($field->{skip_if_empty} && !length($field_value));
  return unless $field->{db_field};
  $db_rec->{$field->{db_field}} = $field_value;

  if ($field->{implies}) {
    my $implies_func = $field->{implies};
    $self->$implies_func(mode => 'import',
                         field_value => $field_value,
                         field => $field,
                         db_rec => $db_rec,
                         node => $node,
                        );
  }
}

#----------------------------------------------------------------------

sub set_pattern_order {
  my($self, %params) = @_;
  my $field_value = $params{field_value};
  my $db_rec = $params{db_rec};
  my $node = $params{node};

  if ($params{mode} eq 'export') {
    # unneeded
  }
  elsif ($params{mode} eq 'import') {
    if ($field_value) {
      # Wildcard
      unless ($self->{pattern_order_last}->{$node->{db_tables}->[-1]}) {
        $self->{pattern_order_last}->{$node->{db_tables}->[-1]} = 0;
      }
      $db_rec->{pattern_order} = ++$self->{pattern_order_last}->{$node->{db_tables}->[-1]};
    }
    else {
      # Explicit - 'pattern_order' default value in DB is 'NULL'
    }
  }
}

#----------------------------------------------------------------------

sub set_implied_flag {
  my($self, %params) = @_;
  my $field_value = $params{field_value};
  my $db_rec = $params{db_rec};
  my $field = $params{field};

  if ($params{mode} eq 'export') {
  }
  elsif ($params{mode} eq 'import') {
    $db_rec->{$field->{implied_flag}} = ($field_value ? 1 : 0);
  }
}

#----------------------------------------------------------------------

sub set_implied_mask {
  my($self, %params) = @_;
  my $field_value = $params{field_value};
  my $db_rec = $params{db_rec};
  my $field = $params{field};

  if ($params{mode} eq 'export') {
    return ($field_value & $field->{value_mask} ? 'true' : 'false');
  }
  elsif ($params{mode} eq 'import') {
    for my $mask_field (keys %{$field->{implied_mask}}) {
      $db_rec->{$mask_field} |= $field->{implied_mask}->{$mask_field};
    }
  }
}

#----------------------------------------------------------------------

sub convert_to_value_mask {
  my($self, %params) = @_;
  my $field_value = $params{field_value};
  my $db_rec = $params{db_rec};
  my $field = $params{field};

  if ($params{mode} eq 'export') {
    return ($field_value & $field->{value_mask} ? 'true' : 'false');
  }
  elsif ($params{mode} eq 'import') {
    if ($field_value eq 'true' || $field_value eq 'false') {
      unless ($db_rec->{$field->{db_field}}) {
        $db_rec->{$field->{db_field}} = 0;
      }
      if ($field_value eq 'true') {
        return $db_rec->{$field->{db_field}} | $field->{value_mask};
      }
      else {
        return $db_rec->{$field->{db_field}};
      }
    }
    else {
      ${$params{record_state}} = 'error';
      $self->add_fatal_error(error_msg => sprintf('Field \'%s\' may not contain the value \'%s\'.',
                                                  $self->get_display_xpath(%params),
                                                  $field_value
                                                 ),
                             %params,
                            );
    }
  }
}

#----------------------------------------------------------------------

sub warn_if_web_scraping_enabled {
    my ($self, %params) = @_;

    if ($params{mode} eq 'import') {
        my $db_rec = $params{db_rec};
        if ($db_rec->{flg_is_enabled} || $db_rec->{flg_web_scraping_block} || $db_rec->{flg_web_scraping_alarm}) {
            $self->add_warning(msg => "Web Scraping configuration (previously used in this security policy) will be ignored. Device-ID mode must be configured in bot profile for this option to work.");
        }
    }
}

#----------------------------------------------------------------------

sub handle_negsig_signatures {
    my ($self, %params) = @_;

    if ($params{mode} eq 'export') {
        # Nothing to do
    }
    elsif ($params{mode} eq 'import') {
        my $db_rec = $params{db_rec};
        if(exists $db_rec->{name}) { # Calculate user defined signature id based on name
            my $sig_name = $db_rec->{name};
            if(! exists $self->{user_defined_sig_name_to_sig_id}) {
                $self->{user_defined_sig_name_to_sig_id} = $self->{dbh}->selectall_hashref(<<EndQuery, 'sig_name', undef, 1);
SELECT sig_name, sig_id
  FROM PLC.NEGSIG_SIGNATURES
  WHERE flg_is_user_defined = ?;
EndQuery
            }
            if(exists $self->{user_defined_sig_name_to_sig_id}->{$sig_name}) {
                $db_rec->{sig_id} = $self->{user_defined_sig_name_to_sig_id}->{$sig_name}->{sig_id};
                delete $db_rec->{name};
            }
            else { # user defined signature with this name does NOT exists - skip the node
		$self->add_warning(msg => qq{signature with name $sig_name (previously used in this security policy) does not exist on this system.});
                undef %{$db_rec};
            }
        }
        else { # Factory signature - make sure it exists
            if(! exists $self->{sig_ids}) {
                $self->{sig_ids} = $self->{dbh}->selectall_hashref(<<EndQuery, 'sig_id');
SELECT sig_id
  FROM PLC.NEGSIG_SIGNATURES
EndQuery
            }
            if(!exists $self->{sig_ids}->{$db_rec->{sig_id}}) { # Factory signature does not exist
		$self->add_warning(msg => qq{signature with id $db_rec->{sig_id} (previously used in this security policy) does not exist on this system.});
                undef %{$db_rec};
            }
        }
    }
}

#----------------------------------------------------------------------

sub handle_sig_name {
    my ( $self, %params ) = @_;

    if ($params{mode} eq 'export') {
        my $sig_id = $params{field_value};
        if(! exists $self->{user_defined_sig_id_to_sig_name}) {
            $self->{user_defined_sig_id_to_sig_name} = $self->{dbh}->selectall_hashref(<<EndQuery, 'sig_id', undef, 1);
SELECT sig_id, sig_name
  FROM PLC.NEGSIG_SIGNATURES
  WHERE flg_is_user_defined = ?;
EndQuery
        }

        return exists $self->{user_defined_sig_id_to_sig_name}->{$sig_id} ?
                      $self->{user_defined_sig_id_to_sig_name}->{$sig_id}->{sig_name} : undef;
    }
    elsif ($params{mode} eq 'import') {
        my $sig_name = $params{field_value};
        $params{db_rec}->{name} = $sig_name;
        return $sig_name;
    }
}

#----------------------------------------------------------------------

sub get_sigfile_timestamp {
    my($self) = @_;
    # Get the timestamp from the last row
    my $sigfile_timestamp = $self->{dbh}->selectrow_array(<<EndQuery);
SELECT sigfile_timestamp
  FROM PLC.SIGFILE_STATUS
  ORDER BY sigfile_status_id DESC LIMIT 1;
EndQuery

    return $sigfile_timestamp;
}

sub handle_sigfile_timestamp {
    my($self, %params) = @_;
    my $field_value = $params{field_value};
    if ($params{mode} eq 'export') {
        return $self->get_sigfile_timestamp();
    }
    elsif ($params{mode} eq 'import') {
        return $field_value;
    }
}

#----------------------------------------------------------------------

sub obfuscate_encrypt {
    my($field_value, $chunk_size, $arbitrary_delimiter) = @_;

    my $field_value_length = length($field_value);
    my @string_chunks = ();
    for(my $i = 0; $i < $field_value_length; $i += $chunk_size) {
        push @string_chunks, MIME::Base64::encode_base64(PassCrypt::crypt_password(substr($field_value, $i, $chunk_size)));
    }
    return join ($arbitrary_delimiter, @string_chunks);
}

sub obfuscate_decrypt {
    my($field_value, $arbitrary_delimiter) = @_;

    my @string_chunks_to_decrypt = split /$arbitrary_delimiter/, $field_value;
    my @string_chunks = map { PassCrypt::decrypt_password(MIME::Base64::decode_base64($_)) } @string_chunks_to_decrypt;
    return join ('', @string_chunks);
}

sub obfuscate {
    my($self, %params) = @_;
    my $field_value = $params{field_value};
    my $arbitrary_delimiter = ':'; # This delimiter is not part of 'base64'

    if ($params{mode} eq 'export') {
        return obfuscate_encrypt($field_value, 64, $arbitrary_delimiter);
    }
    elsif ($params{mode} eq 'import') {
        return obfuscate_decrypt($field_value, $arbitrary_delimiter);
    }
}

sub obfuscate_previous_enforced_rule {
    my($self, %params) = @_;
    my $field_value = $params{field_value};
    if ($params{mode} eq 'export') {
        return obfuscate_encrypt($field_value, 64, ':');
    }
    elsif ($params{mode} eq 'import') {
        if(!$field_value) {
	    return $field_value;
	}
        if(!exists $self->{system_sigfile_timestamp}) {
	    $self->{system_sigfile_timestamp} = $self->get_sigfile_timestamp();
	}
	# Import the previous envorced rule only if the ASU version on the origin device is the same as ASU on the destination device
        if($self->{import_sigfile_timestamp} &&
           ($self->{import_sigfile_timestamp} eq $self->{system_sigfile_timestamp})) {
            return obfuscate_decrypt($field_value, ':');
        }
        else {
	    return '';
        }
    }   
}

#----------------------------------------------------------------------

sub convert_to_hex {
  my($self, %params) = @_;
  my $field_value = $params{field_value};

  if ($params{mode} eq 'export') {
    return sprintf('0x%x', $field_value);
  }
  elsif ($params{mode} eq 'import') {
    if ($field_value =~ /^0x[0-9a-f]{1,2}$/i) {
      return hex($field_value);
    }
    else {
        ${$params{record_state}} = 'error';
        $self->add_fatal_error(error_msg => sprintf('Field \'%s\' may not contain the value \'%s\'.',
                                                    $self->get_display_xpath(%params),
                                                    $field_value,
                                                   ),
                               %params);
    }
  }
}

#----------------------------------------------------------------------

sub convert_in_chunks {
    my($self, %params) = @_;
    my $field_value = $params{field_value};
    my $field = $params{field};
    my $max_chunk_size = $field->{max_chunk_size};
    my $element = $params{element};

    if ($params{mode} eq 'export') {
        if ($field_value =~ /\000/) {
            $field_value =  MIME::Base64::encode_base64($field_value);
            $element->set_att("base64_encoded" => "1");
        }

        if (length($field_value) > $max_chunk_size) {
            while ($field_value) {
                my $new_node = XML::Twig::Elt->new('chunk');
                $new_node->set_text(substr($field_value, 0, $max_chunk_size, ''));
                $new_node->paste(last_child => $element);
            }
        }
        else {
            $element->set_text($field_value);
        }
    }
    elsif ($params{mode} eq 'import') {
        my @chunks = $element->findnodes('chunk');
        my $retval = '';
        if (@chunks) {
            for my $chunk (@chunks) {
                $retval .= $chunk->text;
            }
        }
        else {
            $retval = $element->text;
        }

        # Decode base64 if enabled
        if ($element->{att}->{base64_encoded}) {
            $retval = MIME::Base64::decode_base64($retval);
        }

        return $retval;
    }
}

#----------------------------------------------------------------------

sub convert_by_value_map {
    my($self, %params) = @_;
    my $field_value = $params{field_value};
    my $field = $params{field};

    if ($params{mode} eq 'export') {
        if (defined($field->{export_value_map}->{$field_value})) {
            return $field->{export_value_map}->{$field_value};
        }
        else {
            if ($field->{if_missing} && $field->{if_missing} eq 'skip') {
                ${$params{record_state}} = 'skip';
            }
            return;
        }
    }
    elsif ($params{mode} eq 'import') {
        if (defined($field->{value_map}->{$field_value})) {
            return $field->{value_map}->{$field_value};
        }
        else {
            # Exact match was not found, handle according to if_missing
            my @if_missing_handlings = split(/,/, $field->{if_missing} || 'error');

            for my $if_missing_handling (@if_missing_handlings) {

                if ($if_missing_handling eq 'loosen') {
                    # Accommodate "loose" formatting - capitalization errors, and leading/trailing spaces
                    my $loose_field_value = loosen_field_value($field_value);
                    for my $lookup_value (keys %{$field->{value_map}}) {
                        if ($loose_field_value eq loosen_field_value($lookup_value)) {
                            return $field->{value_map}->{$lookup_value};
                        }
                    }
                }

                elsif ($if_missing_handling eq 'skip') {
                    ${$params{record_state}} = 'skip';
                    if ($field->{warning_if_missing}) {
                        my $warning_msg = $field->{warning_if_missing};
                        $warning_msg =~ s/\$FIELD_VALUE\$/$field_value/g;
                        $self->add_warning(msg => $warning_msg);
                    }
                    return;
                }
                elsif ($if_missing_handling eq 'error') {

                    ${$params{record_state}} = 'error';
                    $self->add_fatal_error(error_msg => sprintf('Field \'%s\' may not contain the value \'%s\'.',
                                                                $self->get_display_xpath(%params),
                                                                $field_value,
                                                               ),
                                           %params);
                    return;
                }
                elsif ($if_missing_handling eq FIELD_STATE_IGNORE) {
                    if ($field->{warning_if_missing}) {
                        my $warning_msg = $field->{warning_if_missing};
                        $warning_msg =~ s/\$FIELD_VALUE\$/$field_value/g;
                        $self->add_warning(msg => $warning_msg);
                    }
                }
            }
        }
    }
}

#----------------------------------------------------------------------

sub loosen_field_value {
    my($val) = @_;
    $val =~ s/^\s+//;
    $val =~ s/\s+$//;
    $val =~ s/\s+/ /g;
    $val = lc $val;
    return $val;
}

#----------------------------------------------------------------------

sub convert_to_method_pool {
  my($self, %params) = @_;
  my $field_value = $params{field_value};

  if ($params{mode} eq 'export') {
    return F5::ASMConfig::Entity::AllowedMethod->get_method_name(dbh => $self->{dbh},
                                                                 code => $field_value);
  }
  elsif ($params{mode} eq 'import') {
      my $act_as_method_name = 'GET'; # default
      if (my $act_as_element = $params{element}->first_child('act_as')) {
          $act_as_method_name = $act_as_element->text;
      }

    return F5::ASMConfig::Entity::AllowedMethod->get_method_code(dbh => $self->{dbh},
                                                                 name => $field_value,
                                                                 add_if_missing => 1,
                                                                 default_act_as => $act_as_method_name,
                                                                );
  }
}
#----------------------------------------------------------------------

sub add_method_to_pool_if_missing {
  my($self, %params) = @_;
  my $method_name = $params{field_value};

  if ($params{mode} eq 'import') {

      F5::ASMConfig::Entity::AllowedMethod->get_method(dbh => $self->{dbh},
                                                       name => $method_name,
                                                       add_if_missing => 1,);
  }

  return $method_name;
  
}
#----------------------------------------------------------------------

sub get_display_xpath {
  my($self, %params) = @_;
  my $display_xpath = '';
  if ($params{node}->{xpath}) {
    $display_xpath .= $params{node}->{xpath};
  }
  if ($params{field}->{xpath}) {
      if ($display_xpath) {
          $display_xpath .= '/';
      }
      $display_xpath .= $params{field}->{xpath};
  }
  return $display_xpath;
}

#----------------------------------------------------------------------

sub convert_newlines {
  my($self, %params) = @_;
  my $field_value = $params{field_value};

  if ($params{mode} eq 'export') {
    return $field_value;
  }
  elsif ($params{mode} eq 'import') {
    $field_value =~ s/\n/\r\n/gs;
    return $field_value;
  }
}

#----------------------------------------------------------------------
sub convert_by_value_type_boolean {
  my($self, %params) = @_;
  my $field_value = $params{field_value};

  if ($params{mode} eq 'export') {
    return ($field_value ? 'true' : 'false');
  }
  elsif ($params{mode} eq 'import') {
    if ($field_value eq 'true' || $field_value eq '1') {
      return 1;
    }
    elsif ($field_value eq 'false' || $field_value eq '0') {
      return 0;
    }
    else {
        ${$params{record_state}} = 'error';
        $self->add_fatal_error(error_msg => sprintf('Field \'%s\' may not contain the value \'%s\'.',
                                                    $self->get_display_xpath(%params),
                                                    $field_value
                                                   ),
                               %params,
                              );
    }
  }
}

#----------------------------------------------------------------------

sub convert_by_value_type_date {
    my($self, %params) = @_;
    my $field_value = $params{field_value};
  
    if ($params{mode} eq 'export') {
        return F5::DateTimeUtils::convert_epoch_to_rfc3339($field_value);
    }
    elsif ($params{mode} eq 'import') {
        my $retval =  F5::DateTimeUtils::convert_any_datetime_to_epoch($field_value);
        unless ($retval) {
            ${$params{record_state}} = 'error';
            $self->add_fatal_error(error_msg => sprintf('Field \'%s\' may not contain the value \'%s\'.',
                                                        $self->get_display_xpath(%params),
                                                        $field_value,
                                                       ),
                                   %params,
                                  );
        }
        return $retval;
    }
}

#----------------------------------------------------------------------

sub convert_by_value_type_unixdate {
  my($self, %params) = @_;
  my $field_value = $params{field_value};

  if ($params{mode} eq 'export') {
    my @date = localtime($field_value);
    $date[5]+= 1900;
    $date[4]++;
    splice(@date, 6);
    return sprintf("%0.4d-%0.2d-%0.2d %0.2d:%0.2d:%0.2d", reverse @date);
  }
  elsif ($params{mode} eq 'import') {
    return 0 unless $field_value;
    if ($field_value =~ /^(\d{4})-(\d\d)-(\d\d) (\d\d):(\d\d):(\d\d)$/) {
      my @date = reverse split(/[-: ]/, $field_value);
      $date[4]--;
      return timelocal(@date);
    }
    else {
        ${$params{record_state}} = 'error';
        $self->add_fatal_error(error_msg => sprintf('Field \'%s\' may not contain the value \'%s\'.',
                                                    $self->get_display_xpath(%params),
                                                    $field_value,
                                                   ),
                               %params,
                              );
    }
  }
}

#----------------------------------------------------------------------

sub get_integrity_check {
  my($self, %params) = @_;
  my $field_value = $params{field_value};
  my $db_rec = $params{db_rec};

  if ($params{mode} eq 'export') {
      return $INTEGRITY_CHECK_REPLACE_STRING;
  }
}

#----------------------------------------------------------------------

sub get_parent_policy_name {
    my($self, %params) = @_;
    my $db_rec = $params{db_rec};

    if ($params{mode} eq 'export') {
        my ($parent_policy_name) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $db_rec->{id});
SELECT name
  FROM PLC.PL_POLICIES p
   JOIN PLC.PL_POLICY_PARENT pa
  ON p.id = pa.parent_policy_id
 WHERE pa.policy_id = ?
EndQuery

        return $parent_policy_name;
    }

    return $params{field_value};
}

#----------------------------------------------------------------------

sub get_bigip_version {
  my($self, %params) = @_;
  my $field_value = $params{field_value};
  my $db_rec = $params{db_rec};

  if ($params{mode} eq 'export') {
      return F5::BigipVersionUtils::bigip_version();
  }
  elsif ($params{mode} eq 'import') {
      $self->{old_version} = $field_value;
      $self->{common_upgrade}->set_old_version($field_value);
  }
}

#----------------------------------------------------------------------

sub get_from_foreign_key {
    my($self, %params) = @_;
    my $field_value = $params{field_value};
    my $field = $params{field};
    my $db_rec = $params{db_rec};
    my $element = $params{element};

    if ($params{mode} eq 'export') {
        my $foreign_table = $self->{foreign_key_save}->{$field->{lookup_table}};
        my $foreign_rec;
        if ($foreign_table) {
            $foreign_rec = $foreign_table->{$db_rec->{$field->{db_field}}};
        }
        if ($foreign_rec) {
            for my $foreign_link (@{$field->{foreign_links}}) {
                if (! $foreign_link->{lookup_only}) {
                    $self->set_value_for_xpath(value => $foreign_rec->{$foreign_link->{foreign_field}},
                                               xpath => $foreign_link->{xpath},
                                               element => $element);
                }
            }
        }
        else {
            # Record has not been loaded yet - defer this node until later
            push(@{$self->{deferred_records}}, 
                 { defer_func => 'get_from_foreign_key',
                   %params });
        }
    }
    elsif ($params{mode} eq 'import') {
        my $element = $params{element};
        my @lookup_values;

        for my $foreign_link (@{$field->{foreign_links}}) {
            my $foreign_value;
            if ($foreign_link->{xpath}) {
                my $value_from_xpath = ($foreign_link->{value_from_xpath}) ? $foreign_link->{value_from_xpath}
                                                                           : F5::XMLUtils->get_value_from_xpath(element => $element,
                                                                                                                xpath => $foreign_link->{xpath});
                if (! defined $value_from_xpath && defined $foreign_link->{value_if_missing}) {
                    $value_from_xpath = $foreign_link->{value_if_missing};
                }

                if (! defined $value_from_xpath) {
                    # At least 1 xpath couldn't be found - abort
                    ${$params{record_state}} = $field->{if_missing};
                    return;
                }
                else {
                    $foreign_value = $value_from_xpath;
                }
            }
            elsif ($foreign_link->{field_value}) {
                $foreign_value = $foreign_link->{field_value};
            }

            if ($foreign_link->{value_map}) {
                $foreign_value = $foreign_link->{value_map}->{$foreign_value};
            }
            push(@lookup_values, $foreign_value);
        }

        my $foreign_key = join($UNLIKELY_DELIMITER, @lookup_values);
        if ($self->{foreign_key_save}->{$field->{lookup_table}}->{$foreign_key}) {
            ${$params{record_state}} = '';
            return $self->{foreign_key_save}->{$field->{lookup_table}}->{$foreign_key};
        }
        else {
            ${$params{record_state}} = ($field->{if_missing} eq FIELD_STATE_IGNORE ? RECORD_STATE_DEFER_NON_FATAL : RECORD_STATE_DEFER);
            $db_rec->{$field->{db_field}} = { table_name => $field->{lookup_table},
                                              foreign_key => $foreign_key };
        }
    }
}

#----------------------------------------------------------------------

sub insert_row {
  my($self, %params) = @_;
  return if !$params{table_name};
  $self->{batch_insert}->insert_row(%params);
}

#----------------------------------------------------------------------

sub filter_rec_for_table {
  my($self, %params) = @_;
  my $dbh = $self->{dbh};

  my $table_name = $params{table_name};
  my $rec = $params{rec};
  unless ($self->{table_fields}->{$table_name}) {
    my @fields = F5::DbUtils::get_db_fields($dbh, $table_name, F5::DbUtils::FIND_ONLY_NONINDEX_FIELDS());
    $self->{table_fields}->{$table_name} = { map { ($_ => 1) } @fields };
  }

  my %table_rec = map { ($_ => $rec->{$_}) } grep { $self->{table_fields}->{$table_name}->{$_} } keys %$rec;
  return \%table_rec;
}

#----------------------------------------------------------------------

sub fixup_node {
  my($self, %params) = @_;
  my $node = $params{node};
  my $dbh = $self->{dbh};

  return if $node->{fixup_done};

  if ($params{minimal_export}) {
      if ($node->{minimal_export_func}) {
          my $func_name = $node->{minimal_export_func};
          my %minimal_export_node_changes = $self->$func_name(%params);
          for my $key (keys %minimal_export_node_changes) {
              $node->{$key} = $minimal_export_node_changes{$key};
          }
      }

      if ($node->{minimal_export}) {
          for my $key (keys %{$node->{minimal_export}}) {
              $node->{$key} = $node->{minimal_export}->{$key};
          }
      }
  }

  # Copy remainder of node from template node
  if ($node->{copy_from_template}) {
    my($template_node) = grep { $_->{template_name} && $_->{template_name} eq $node->{copy_from_template} } @{$self->{config}->{nodes}};
    my @premerge_fields;
    if ($node->{fields}) {
      @premerge_fields = @{$node->{fields}};
    }
    %$node = (%$template_node,
              %$node);
    # Combine fields
    $node->{fields} = [ @{$template_node->{fields}},
                        @premerge_fields ];
    delete $node->{copy_from_template};
  }

  for my $field (@{$node->{fields}}) {
    # Set field conversion functions
    if ($field->{value_map_lookup} && $dbh) {
      my $value_map = $dbh->selectcol_arrayref(<<EndQuery, { Columns => [1,2] });
SELECT $field->{value_map_lookup}->{value_field},
       $field->{value_map_lookup}->{join_field}
  FROM $field->{value_map_lookup}->{table_name}
EndQuery
      $field->{value_map} = { @$value_map };
      delete $field->{value_map_lookup};
    }

    if ( !defined($field->{export_value_map}) && $field->{value_map}) {
      $field->{export_value_map} = { reverse %{$field->{value_map}} };
    }
    
    unless ($field->{field_func}) {
      if ($field->{value_map}) {
        $field->{field_func} = 'convert_by_value_map';
      }
      elsif ($field->{value_type}) {
        $field->{field_func} = 'convert_by_value_type_' . $field->{value_type};
      }
      elsif ($field->{value_mask}) {
        $field->{field_func} = 'convert_to_value_mask';
      }
    }
  }

  $node->{fixup_done} = 1;
}

#----------------------------------------------------------------------

sub add_negsig_signature_limit_clause {
    my($self, %params) = @_;
    my %minimal_export_opts;

    # Default for scanner (whitehat) policy has signatures disabled
    # All other policies default to signatures enabled
    my($flg_is_whitehat) = $self->{dbh}->selectrow_array('SELECT scanner_type != "" FROM PLC.PL_SCANNER_CONFIG WHERE policy_id = ?',
                                                         undef, $self->{policy}->{policy_id});
    if ($flg_is_whitehat) {
        $minimal_export_opts{limit_clause} = { flg_disabled => 0 };
    }
    else {
        $minimal_export_opts{limit_clause} = { flg_disabled => 1 };
    }
    return %minimal_export_opts;
}

#----------------------------------------------------------------------

sub convert_metachar {
    my ($self, %params) = @_;
    my $mode = $params{mode};
    my $node = $params{node};
    my $db_rec = $params{db_rec};


    my ($default_charset, $section_type) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $self->{primary_encoding}, $db_rec->{charset_type});
SELECT DC.charset, DC.section_type
  FROM DCC.DEFAULT_CHARSETS DC
       JOIN DCC.ENCODINGS E
    ON DC.encoding_id = E.encoding_id
 WHERE E.encoding = ?
   AND DC.charset_type = ?
EndQuery

    if ($mode eq 'export') {
        my $node_twig = $params{node_twig};

        my @chars         = split(//, $db_rec->{charset});
        my @default_chars = split(//, $default_charset);
        for my $i (0..$#chars) {
            if ($chars[$i] != $default_chars[$i]) {
                my $metachar_node = XML::Twig::Elt->new('metachar');
                $metachar_node->paste(last_child => $node_twig);
                $metachar_node->set_att(character => $self->convert_to_hex(mode => 'export', field_value =>  $i));
                $metachar_node->set_text($chars[$i] ? 'allow' : 'disallow');
            }
        }
    }

    if ($mode eq 'import') {
        my $element = $params{element};

        my $charset = $default_charset;
        my @metachar_nodes = $element->findnodes('metachar');
        for my $metachar_node (@metachar_nodes) {
            if ($metachar_node->text eq 'allow') {
                my $character = $self->convert_to_hex(mode => 'import', field_value =>  $metachar_node->{att}->{character});
                substr($charset, $character, 1) = '1';
            }
            elsif ($metachar_node->text eq 'disallow') {
                my $character = $self->convert_to_hex(mode => 'import', field_value =>  $metachar_node->{att}->{character});
                substr($charset, $character, 1) = '0';
            }
        }
        $db_rec->{charset} = $charset;
        $db_rec->{section_type} = $section_type;
    }
}

#----------------------------------------------------------------------

sub check_dependent_field_values {
    my($self, %params) = @_;
    my $dependent_field_values = $params{dependent_field_values} || {};
    my $db_rec = $params{db_rec};

    for my $dependent_field (keys %$dependent_field_values) {
        if (ref($dependent_field_values->{$dependent_field})) {
            if (! grep { $db_rec->{$dependent_field} == $_ } @{$dependent_field_values->{$dependent_field}}) {
                return;
            }
        }
        else {
            if ($db_rec->{$dependent_field} != $dependent_field_values->{$dependent_field}) {
                return;
            }
        }
    }
    return 1;
}

#----------------------------------------------------------------------

sub export_field {
  my($self, %params) = @_;
  my $field = $params{field};
  my $node = $params{node};
  my $element = $params{element};
  my $db_rec = $params{db_rec};
  my $save_rec = $params{save_rec};
  my $record_state = $params{record_state};

  return if ($field->{import_only});

  if (! $self->check_dependent_field_values(dependent_field_values => $field->{included_only_if},
                                            db_rec => $db_rec)) {
      # skip to the next field
      return;
  }

  my $field_value;
  if (defined $field->{db_field}) {
    $field_value = $db_rec->{$field->{db_field}};
    return unless defined($field_value);
  }

  if ($field->{field_func}) {
    my $field_func = $field->{field_func};
    $field_value = $self->$field_func(mode => 'export',
                                      field_value => $field_value,
                                      field => $field,
                                      db_rec => $db_rec,
                                      element => $element,
                                      record_state => $record_state,
                                     );
  }
  return if ($$record_state eq 'skip');
  return if (! defined $field_value && $field->{skip_if_null});
  if ($field->{skip_if_empty} && (! defined $field_value || $field_value eq '')) {
      return;
  }

  if ($field->{implied_mask}) {
    for my $mask_field (keys %{$field->{implied_mask}}) {
      if (! ($db_rec->{$mask_field} & $field->{implied_mask}->{$mask_field})) {
        return;
      }
    }
  }

  if (defined($field->{xpath})) {
    $self->set_value_for_xpath(value => $field_value,
                               element => $element,
                               xpath => $field->{xpath});
  }

  if ($node->{save_for_foreign_key} && $field->{xpath} && grep { $field->{xpath} eq $_ } @{$node->{save_for_foreign_key}->{fields}}) {
    $save_rec->{$field->{xpath}} = $field_value;
  }
}

#----------------------------------------------------------------------

sub set_value_for_xpath {
  my($self, %params) = @_;
  my $xpath = $params{xpath};
  my $element = $params{element};
  my $value = $params{value};

  return unless defined($xpath);
  if ($xpath eq '') {
    $element->set_text($value);
  }
  elsif ($xpath =~ /^(.*?)@(.*)/) {
    my($xpath_only, $attribute) = ($1, $2);
    my $subelement = F5::XML::Schema::Builder::add_or_find_node($element, $xpath_only);
    $subelement->set_att($attribute => $value);
  }
  else {
     F5::XML::Schema::Builder::add_node_with_value($element, $xpath, $value);
  }
}

#----------------------------------------------------------------------
sub add_policy_defaults {
  my($self, %params) = @_;

  my $twig = $params{twig};
  my $policy_id = $params{policy_id};
  my $account_id = $self->{account_id};
  my $dbh = $self->{dbh};

  for my $node (@{$self->{config}->{nodes}}) {
    if ($node->{defaults}) {
      my $defaults_def = $node->{defaults};

      if ($defaults_def->{defaults_xpath} &&
          $twig->findnodes($defaults_def->{defaults_xpath})) {
          next;
      }

      if ($defaults_def->{insert_ignore} ||
          !$node->{xpath} ||
          !$self->{seen_during_import}->{$node->{xpath}}) {

        if ($defaults_def->{single_row}) {
          $self->insert_row(table_name => $node->{db_tables}->[0],
                            rec => { policy_id => $policy_id },
                           );
        }
        elsif ($defaults_def->{defaults_func}) {
          my $defaults_func = $defaults_def->{defaults_func};
          $self->$defaults_func(dbh => $dbh,
                                policy_id => $policy_id,
                                dbname => 'PLC');
        }
        else {
          my %copy_defaults_opts = (%$defaults_def,
                                    table_name => $node->{db_tables}->[0]);

          if ($copy_defaults_opts{insert_ignore}) {
            $copy_defaults_opts{ignore} = 1;
          }

          F5::DbUtils::copy_defaults_table(dbh => $dbh,
                                           policy_id => $policy_id,
                                           %copy_defaults_opts,
                                          );
        }
      }
    }
  }

  $self->{common_upgrade}->add_default_rows(policy_id => $policy_id,
                                            seen_tables => $self->{seen_during_import_tables});

  if ($self->{incompatible_charsets}) {
    $dbh->do(<<EndQuery, undef, $policy_id);
DELETE FROM PLC.PL_CHARSETS
      WHERE policy_id = ?
EndQuery
  }
  $self->insert_charset_defaults(policy_id => $policy_id,
                                 account_id => $account_id);

  my $negative_signatures = F5::NegativeSignatures->new(dbh => $dbh, dbname => 'PLC');
  $negative_signatures->add_all_signatures_to_policy(policy_id => $policy_id);

  F5::ThreatCampaigns::add_all_threat_campaigns_to_policy(dbh => $dbh, policy_id => $policy_id, logger => $self->{logger});
}

#----------------------------------------------------------------------

sub add_default_negsig_sets {
  my($self, %params) = @_;

  my $negative_signatures = F5::NegativeSignatures->new(dbh => $params{dbh}, dbname => 'PLC');
  $negative_signatures->add_default_sets_to_policy(policy_id => $params{policy_id});
}

#----------------------------------------------------------------------

sub add_default_cookie {
    my ($self, %params) = @_;
    return if $self->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT();
    F5::ASMConfig::Entity::Cookie->new(dbh => $params{dbh},
                                       policy => { policy_id => $params{policy_id}},
                                       object => { cookie_name => '*',
                                                   is_wildcard => 1,
                                                   has_tightening => F5::ASMConfig::TighteningMode::TIGHTENING_MODE_EXCEPTION(),
                                                   cookie_enforcement_type => F5::ASMConfig::CookieEnforcementType::COOKIE_ENFORCEMENT_TYPE_ALLOWED(),
                                                   in_staging => 1,
                                                 },
                                      )->store_add(suppress_events => 1);
}

#----------------------------------------------------------------------

sub add_default_filetype {
    my ($self, %params) = @_;
    return if $self->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT();
    F5::ASMConfig::Entity::Filetype->new(dbh => $params{dbh},
                                         policy => { policy_id => $params{policy_id} },
                                         object => { filetype_name => '*',
                                                     is_wildcard => 1,
                                                     flg_check_req_len => 0,
                                                     flg_check_qs_len  => 0,
                                                     flg_check_pd_len  => 0,
                                                     flg_check_url_len => 0,
                                                   },
                                        )->store_add(suppress_events => 1);
}

#----------------------------------------------------------------------

sub add_default_objects {
    my ($self, %params) = @_;
    return if $self->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT();

    my @protocols = (0);
    if (!$self->{flg_is_protocol_free}) {
        push (@protocols, 1);
    }

    for my $is_https (@protocols) {
        my $url = F5::ASMConfig::Entity::URL->new(dbh => $params{dbh},
                                                  policy => { policy_id => $params{policy_id} },
                                                  preload_default => 1,
                                                  object => {
                                                      app_id => $is_https,
                                                      url_name => '*',
                                                      is_wildcard => 1,
                                                      has_tightening => F5::ASMConfig::TighteningMode::TIGHTENING_MODE_NEVER,
                                                      wildcard_sequence => 1,
                                                      in_staging => 1,
                                                      last_tightening_time => time(),
                                                      staged_since => time(),
                                                      last_updated => time(),
                                                      description => '',
                                                  },
                                                 )->store_add(suppress_events => 1);
    }

}

#----------------------------------------------------------------------

my $WEBSOCKET_VERSION = '12.1.0';

sub add_default_websocket_urls {
    my ($self, %params) = @_;
    return if $self->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT();

    if ($self->{old_version} && F5::BigipVersionUtils::compare_versions(
                                                                        version1 => $self->{old_version},
                                                                        version2 => $WEBSOCKET_VERSION,
                                                                        ) < 0 ) {
        #Do not add websockets if we are importing an older version.
        return;
    }

    my @ws_protocols = (F5::ASMConfig::WebsocketProtocol::WS_PROTOCOL_WS());
    if (!$self->{flg_is_protocol_free}) {
        push (@ws_protocols, F5::ASMConfig::WebsocketProtocol::WS_PROTOCOL_WSS());
    }


    my $default_plain_text_profile = F5::ASMConfig::Entity::WebsocketUrl->get_missing_default_text_profile(
                                                      dbh => $params{dbh},
                                                      policy => { policy_id => $params{policy_id} }, );

    if ($default_plain_text_profile->{__new_entity}) {
        $default_plain_text_profile->store_add(suppress_events => 1);
    }


    for my $ws_protocol (@ws_protocols) {
        F5::ASMConfig::Entity::WebsocketUrl->new(dbh => $params{dbh},
                                                  policy => { policy_id => $params{policy_id} },
                                                  preload_default => 1,
                                                  object => {
                                                      protocol => $ws_protocol,
                                                      websocket_url_name => '*',
                                                      is_wildcard => 1,
                                                      has_tightening => F5::ASMConfig::TighteningMode::TIGHTENING_MODE_NEVER,
                                                      wildcard_sequence => 1,
                                                      in_staging => 1,
                                                      last_tightening_time => time(),
                                                      staged_since => time(),
                                                      last_updated => time(),
                                                      description => '',
                                                      flg_wildcard_includes_slash => 1,
                                                      flg_websocket_message_format_text => 1,
                                                      websocket_plain_text_profile => $default_plain_text_profile,
                                                      flg_check_payload => 1,
                                                  },
                                                 )->store_add(suppress_events => 1);
    }

}

#----------------------------------------------------------------------

sub add_default_parameter {
    my ($self, %params) = @_;
    return if $self->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT();
    F5::ASMConfig::Entity::Parameter->new(dbh => $params{dbh},
                                          policy => { policy_id => $params{policy_id} },
                                          object => { param_name => '*',
                                                      is_wildcard => 1,
                                                      in_staging => 1,
                                                      param_level => F5::ASMConfig::ParamLevel::PARAM_LEVEL_GLOBAL(),
                                                      param_data_type => F5::ASMConfig::ParamDataType::PARAM_DATA_ALPHA_NUMERIC(),
                                                      param_value_type => F5::ASMConfig::ParamValueType::PARAM_VALUE_USER_INPUT(),
                                                    },
                                         )->store_add(suppress_events => 1);

}

#----------------------------------------------------------------------

sub add_default_redirection_domain {
    my ($self, %params) = @_;
    return if $self->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT();
    F5::ASMConfig::Entity::RedirectionDomain->new(dbh => $params{dbh},
                                                  policy => { policy_id => $params{policy_id} },
                                                  object => { domain_name          => '*',
                                                              is_wildcard          => 1,
                                                              wildcard_sequence    => 1,
                                                              last_tightening_time => time(),
                                                              last_updated         => time(),
                                                              has_tightening       => F5::ASMConfig::TighteningMode::TIGHTENING_MODE_ALWAYS,
                                                            },
                                                 )->store_add(suppress_events => 1);

}

#----------------------------------------------------------------------

sub add_default_brute_force {
    my ($self, %params) = @_;
    return if $self->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT();
    F5::ASMConfig::Entity::BruteForce->new(dbh => $params{dbh},
                                           policy => { policy_id => $params{policy_id} },
                                           object => {
                                                      flg_is_default => 1,
                                                      object_id => 0,
                                                      flg_enabled => 1,
                                                     },                                         
                                      )->store_add(suppress_events => 1);
}

#----------------------------------------------------------------------

sub build_schema {
  my($self, %params) = @_;

  my $output_filename = $params{output_filename};
  my $xml_document_template = <<EndDoc;
<?xml version="1.0" encoding="utf-8"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
  <xs:include schemaLocation="negsig/sets.xsd"/>
</xs:schema>
EndDoc

  my $twig = F5::XML::Twig->new(pretty_print => 'indented');
  $twig->parse($xml_document_template);

  for my $node (@{$self->{config}->{nodes}}) {
        F5::XML::Schema::Builder::create_node_schema( node => $node,
                                                      base_twig => $twig,
                                                      node_handler => $self,
                                                    );
  }

  # Add minOccurs = 0 to all other elements
  for my $element ($twig->findnodes('//xs:element')) {
       if (!$element->{att}->{minOccurs} && (! $element->{att}->{name} || $element->{att}->{name} ne 'policy')) {
          $element->set_att(minOccurs => 0);
       }
       if ($element->{att}->{name} && $element->{att}->{name} eq 'set') {
            delete $element->{att}->{type};
            $element->{att}->{ref} = $element->{att}->{name};
            delete $element->{att}->{name};
       }
  }

  open(my $outf, '>', $output_filename);
  $twig->print($outf);
  close $outf;

}

#----------------------------------------------------------------------

sub check_schema {
  my($filename) = @_;

  my $error_msg = "The security policy file does not conform to the schema and cannot be imported.";

  F5::XML::Schema::Builder::check_schema( filename => $filename, 
                                          schema => F5::Deploy::map_path('/ts/var/schema/import_export_xml.xsd'),
                                          flg_no_network => 1,
                                          error_msg => $error_msg);
}

#----------------------------------------------------------------------

sub post_import_fixups {
  my($self, %params) = @_;
  my $dbh = $self->{dbh};
  my $policy_id = $params{policy_id};

  # Fix flg_parse_dynamic_params
  $dbh->do(<<EndQuery, undef, $policy_id);
UPDATE PLC.PL_OBJECTS
   SET flg_parse_dynamic_params = 1
 WHERE id IN ( SELECT * FROM (
    SELECT o.id FROM PLC.PL_OBJECTS o,
       PLC.PL_OBJECT_EXTRACTS oe,
       PLC.PL_EXTRACT_ATTRIBUTES ea
   WHERE o.id = oe.obj_id
     AND oe.extract_attrib_id = ea.id
     AND o.policy_id = ?
     AND ea.dynamic_param_type <> 0) subquery)
EndQuery

  F5::ResetStaging::reset_staging_and_tightening(
    dbh => $dbh,
    policy_id => $policy_id,
  );

  F5::UrlContentUpgrade::populate_url_content_defaults(dbh => $dbh,
                                                       policy_id => $policy_id,
                                                       idmap => $self->{idmap},
                                                      );
}

#----------------------------------------------------------------------

sub negsig_set_definition {
  my($self, %params) = @_;
  my $field = $params{field};
  my $field_value = $params{field_value};
  my $element = $params{element};

  if ($params{mode} eq 'export') {
      
      my $set = F5::NegativeSignatures::Set->export_set(set_ids => [$field_value], dbh => $self->{dbh})->{sets_twig}->first_child;
      if (F5::NegativeSignatures::Set->new(set_id => $field_value, dbh => $self->{dbh})->set_flg_is_user_defined) {
          return $set;
      }

      for my $child_elm ($set->children) {
          $child_elm->delete;
      }

      return $set;
  }
  elsif ($params{mode} eq 'import') {

      if ((defined $element->first_child('set')) && ($element->first_child('set')->has_children)) {
          my $rv = F5::NegativeSignatures::Set->import_set(twig => $element, dbh => $self->{dbh});
          my $idmap;
    
          $self->add_negsig_set_warning(idmap => \$idmap, rvs => $rv);

          my $new_set_id = $rv->[0]->{new_set_id};

          if ($rv->[0]->{update_status} eq $F5::NegativeSignatures::Set::SET_UPDATE_STATUS_EMPTY || $self->{seen_negsig_set_ids}->{$new_set_id}) {
	      # Prevent empty sets
              # Prevent duplicate set_ids from conflicting
              ${$params{record_state}} = 'skip';
              return;
          }
          $self->{seen_negsig_set_ids}->{$new_set_id} = 1;
          return $new_set_id;
      }
      else {
          if (! $field->{all_sets}) {
              $field->{all_sets} = { @{$self->{dbh}->selectcol_arrayref(<<EndQuery, { Columns => [1,2] })} };
SELECT set_id, set_name
  FROM PLC.NEGSIG_SETS
EndQuery
              $field->{sets_by_name} = { reverse %{$field->{all_sets}} };
          }

          if (my $set_element = $element->first_child('set')) {
              if ($set_element->{att}->{name}) {
                  $field_value = $field->{sets_by_name}->{$set_element->{att}->{name}};
              }
              elsif ($set_element->{att}->{id}) {
                  $field_value = $set_element->{att}->{id};
              }
          }
          else {
              $field_value = $element->{att}->{set_id};
          }
          return unless $field_value;

          if ($field->{all_sets}->{$field_value}) {
              return $field_value;
          }
          else {
              ${$params{record_state}} = 'skip';
              my $set_name = $params{element}->{att}->{name};
              if ($set_name) {
                  $self->add_warning(msg => qq{Signature Set "$set_name" (previously used in this security policy) does not exist on this system.});
              }
              else {
                  $self->add_warning(msg => qq{Signature Set with set_id $field_value (previously used in this security policy) does not exist on this system.});
              }
          }
      }
  }
  elsif ($params{mode} eq 'build_schema') {

    return XML::Twig::Elt->new('xs:element' => { ref => 'set',
                                                 minOccurs => 0,
                                                 maxOccurs => 1,
                                               });
  }
}

#----------------------------------------------------------------------

sub user_defined_violation_definition {
    my($self, %params) = @_;
    my $field = $params{field};
    my $field_value = $params{field_value};
    my $element = $params{element};

    if ($params{mode} eq 'export') {

        my $violation = F5::ASMConfig::Entity::Violation->new(dbh => $self->{dbh},
                                                              get_criteria => { viol_index => $field_value },
                                                             );

        if ($violation->{flg_is_user_defined}) {
            my $xml_twig = $violation->to_xml;
            $xml_twig->set_tag('properties');

            # <properties> node in XML policy differs from <violation> in XML violation export
            # in the following ways:
            # - viol_index is stored in properties/internal_id, instead of violation/@id attribute
            # - viol_name is taken from properties/icontrol_name (forced already to be identical),
            #     instead of violation/@name
            $self->set_value_for_xpath(element => $xml_twig,
                                       xpath => 'internal_id',
                                       value => $xml_twig->{att}->{id});
            $xml_twig->del_att('name');
            $xml_twig->del_att('id');

            return $xml_twig;
        }
    }
    elsif ($params{mode} eq 'import') {

        if ((defined $element->first_child('properties')) && ($element->first_child('properties')->has_children)) {
            # Translate properties element in XML policy to XML violation
            my $properties = $element->first_child('properties');
            $properties->set_tag('violation');

            # Copy viol_name from icontrol_name (see note on export above)
            if (my $icontrol_name_elm = $properties->first_child('icontrol_name')) {
                $properties->set_att('name' => $icontrol_name_elm->text);
            }
            # Move internal_id to @id attribute (see note on export above)
            if (my $internal_id_elm = $properties->first_child('internal_id')) {
                $properties->set_att('id' => $internal_id_elm->text);
            }
            # Remove id and name from <violation> element, since viol_index will be taken from this field
            $element->del_att('id');
            $element->del_att('name');

            my $rv;
            eval {
                $rv = F5::ASMConfig::Entity::Violation->import_xml_entity(element => $properties,
                                                                          dbh => $self->{dbh},
                                                                          return_object_fields => [ 'viol_index' ],
                                                                          store_values => { flg_is_deleted => 0 },
                                                                         );

                $self->complete_user_defined_violation_handling(import_rvs => [ $rv ],
                                                                policy_id => $params{policy_id});

            };
            if ($@) {
                my $error_msg = $@;
                if (ref($error_msg) && $error_msg->{error_message}) {
                    $error_msg = $error_msg->{error_message};
                }
                $error_msg = "User-defined violation could not be added: $error_msg";
                return $self->add_fatal_error(error_msg => $error_msg);
            }

            if ($rv && $rv->{entity} && $rv->{entity}->{viol_index}) {
                return $rv->{entity}->{viol_index};
            }
        }
    }
    return;
}

#----------------------------------------------------------------------

sub add_fatal_error {
    my($self, %params) = @_;

    my $error_msg = $params{error_msg};
    my $twig;
    if ($params{field_element}) {
        $twig = $params{field_element}->twig;
    }
    elsif ($params{element}) {
        $twig = $params{element}->twig;
    }
    if ($twig) {
        $error_msg .= " (line $self->{last_start_tag_position}-" . $twig->current_line . ')';
    }

    push(@{$self->{accumulated_fatal_errors}}, $error_msg);
}

#----------------------------------------------------------------------

sub handle_signature_requirements {
    my ($self, %params) = @_;
    my $element        = $params{element};
    my $mode           = $params{mode};
    my $db_rec         = $params{db_rec};
    my @signature_sets = $element->findnodes('/policy/attack_signatures/signature_set');

    if ($mode eq "import") {
        my $sig_sets_tag_map = {};
        if (!$self->{sig_tags}) {
            $self->{sig_tags} = $self->{dbh}->selectall_hashref(<<EndQuery, 'sig_tag');
SELECT DISTINCT(sig_tag)
FROM PLC.NEGSIG_SIGNATURES
EndQuery
        }

        for my $signature_set (@signature_sets) {
            my $set_value = $signature_set->first_child_trimmed_text('set');
            if (length $set_value) {
                my $set =  $signature_set->first_child('set');
                my $filter = $set->first_child_trimmed_text('filter');
                if (length $filter) {
                    my $tag_name = $set->first_child('filter')->first_child_trimmed_text('sig_tag_val');
                    if (length $tag_name) {
                        $sig_sets_tag_map->{'tag'}->{$tag_name} = 1;
                    }
                    # Check 'sig_tag_op' is present or not. If not present we will get 'undef' below
                    my $tag_op = $set->first_child('filter')->first_child_matches('sig_tag_op');
                    if (! defined $tag_op) {
                        $sig_sets_tag_map->{'tag_op'}->{'all'} = 1;
                    }
                }
                else {
                    my $type = $set->first_child_trimmed_text('flg_is_manual');
                    if (length $type && $type) {
                        $sig_sets_tag_map->{'type'}->{'manual'} = 1;
                    }
                    else {
                        my $name = $set->{'att'}->{'name'};
                        if (length $name) {
                            my $set_id = $self->{dbh}->selectrow_array(<<EndQuery, undef, $name);
SELECT set_id
FROM PLC.NEGSIG_SETS WHERE set_name = ?
EndQuery
                            if ($set_id) {
                                my ($sig_tag_val, $sig_tag_op) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $set_id);
SELECT sig_tag_val, sig_tag_op
FROM PLC.NEGSIG_SET_FILTERS WHERE set_id = ?
EndQuery
                                if ($sig_tag_val) {
                                    $sig_sets_tag_map->{'tag'}->{$sig_tag_val} = 1;
                                }
                                if (!$sig_tag_op) {
                                    $sig_sets_tag_map->{'tag_op'}->{'all'} = 1;
                                }
                            }
                        }
                    }
                }
            }
        }
        if (@signature_sets) {
            if (!$sig_sets_tag_map->{'tag'}->{$db_rec->{sig_tag}}
                && !$sig_sets_tag_map->{'type'}->{'manual'}
                && !$sig_sets_tag_map->{'tag_op'}->{'all'})
            {
                $self->add_warning(msg => "Policy requires signature tag '$db_rec->{sig_tag}', but does not include any signature of that tag.");
            }
        }
        for my $tag (keys %{$sig_sets_tag_map->{'tag'}}) {
            if (length($tag) && !$self->{sig_tags}->{$tag}) {
                $self->add_warning(msg => "Tag '$tag' is not required in policy but exists in 'signature_set'.");
            }
        }
        my $store_obj = F5::ASMConfig::PolicySignatureRequirement->new(
                                                                        max_revision_datetime => $db_rec->{max_revision_datetime},
                                                                        min_revision_datetime => $db_rec->{min_revision_datetime},
                                                                        sig_tag               => $db_rec->{sig_tag}
                                                                    );
        my $obj = F5::ASMConfig::Entity::PolicySignatureRequirement->new(dbh => $self->{dbh});
        eval {
            $obj->before_store(
                                %params,
                                store_mode     => 'add',
                                store_obj      => $store_obj,
                                existing_obj   => {},
                                validate_value => 1
                            );
        };
        if ($@) {
            my $error_msg = $@;
            if (ref($error_msg)) {
                $error_msg = $error_msg->{error_message};
            }
            return $self->add_fatal_error(error_msg => $error_msg);
        }
    }
}

#----------------------------------------------------------------------

1;
