package F5::Logger;

use Data::Dumper;
use FindBin;

use F5::Deploy;
use F5::Log;
use F5::Cfg;
use F5::ASM::COAPI;
use F5::Utils::File;
use F5::DateTimeUtils;
use F5::Syslog;
use F5::BigipUtils;
use File::stat;

F5::ASM::COAPI::Login('root');

# propagate F5::Log constants to using packages
#    for seamless shift from F5::Log to F5::Logger
use Exporter ();
@ISA = qw(Exporter);
@EXPORT_OK = qw($INF $DBG $ERR $F_ERR $TRACE);
%EXPORT_TAGS = (LEGACY_ERROR_CONSTANTS => [qw($INF $DBG $ERR $F_ERR $TRACE)]);

$Data::Dumper::Sortkeys = 1;
$Data::Dumper::Useqq = 1;

use constant PRINT_LOG => 1;
use constant PRINT_TERM => 2;
use constant PRINT_ALL => 3;

our $ALL_MODULES = 'all';

our %LOG_LEVEL_CONSTANT_MAP = (
                              $F_ERR => 1,
                              $ERR => 2,
                              # WARN would be 3, but is currently unused due to the limitations of Log.pm
                              $INF => 4,
                              $DBG => 5,
                              $TRACE => 6,
                             );

our %LOG_LEVEL_NAME_MAP = (
                          'fatal' => 1,
                          'error' => 2,
                          'warn' => 2, # Elevated to error due to the limiations of Log.pm
                          'info' => 4,
                          'debug' => 5,
                          'trace' => 6,
                          );

my %LOG_LEVEL_SYSLOG_MAP = (
                            $F_ERR => 'alert',
                            $ERR => 'err',
                            $INF => 'info',
                            $DBG => 'debug',
                            $TRACE => 'info',
                           );

my $LOGGING_CONF_FILE = F5::Deploy::map_path('/etc/ts/tools/asm_logging.conf');
my $ENV_LOG_LEVEL_FILE = F5::Deploy::map_path('/ts/var/env_log_level.yaml');
my $CONFIG_RECHECK_INTERVAL = 5;

sub new {
  my($class, $filename, $print_target, $params) = @_;

  my %params = (filename => $filename,
                dbvar => 'none',
               );
  if (defined $print_target) {
      $params{print_target} = $print_target;
  }

  if ($params) {
      %params = (%params, %$params);
  }

  return $class->get_logger(%params);
}


#----------------------------------------------------------------------

sub get_logger {
    my ($class, %params) = @_;

    my $self = bless({}, $class);
    $self->init(%params);

    return undef unless $self->{filename};
    return $self;
}

#----------------------------------------------------------------------
# temporary. for seamless shift from F5::Log to F5::Logger

sub log_print {
  my ($self, $level, $action, $message) = @_;
  $self->call_logprint($level, $message, $action);
}

#----------------------------------------------------------------------

sub log {
    my ($self, %params) = @_;
    $self->call_logprint(uc $params{log_level}, $params{msg}, $params{action}, %params);
}

#----------------------------------------------------------------------

sub get_config {
    my ($self, %params) = @_;
    my $logger_name = $params{logger_name} || $self->{logger_name};

    # Look for default values in config file
    if (-e $LOGGING_CONF_FILE && F5::Cfg::cfg_get_config_sections($LOGGING_CONF_FILE, $logger_name)) {

        for my $cfg_item (qw(filename dbvar log_level)) {
            my $item_value = F5::Cfg::cfg_get_config_item($LOGGING_CONF_FILE, $logger_name, $cfg_item);

            if (defined $item_value && ! defined $params{$cfg_item}) {
                $params{$cfg_item} = $item_value;
            }
        }
    }
    return %params;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    my $logger_name = $params{logger_name} || $self->{logger_name} || $FindBin::Script;
    $self->{logger_name} = $logger_name;

    if (-e $LOGGING_CONF_FILE) {
        $self->{config_mtime} = stat($LOGGING_CONF_FILE)->mtime;
    }
    
    if (exists $ENV{F5WAF_LOG_TO_STDOUT}) {
        $print_target = 3;
        STDOUT->autoflush(1);
        STDERR->autoflush(1);
    }
    else {
       $print_target = 1;
    }

    %params = (
               log_level => F5::BigipUtils::get_default_log_level() || 'debug',
               umask => 0077,
               dbvar => $logger_name,
               print_target => $print_target,
               $self->get_config(%params),
              );

    $self->{last_config_check} = time();
    $self->{dbvar} = $params{dbvar};
    $self->{filename} = F5::Deploy::map_path($params{filename});
    $self->{print_target} = $params{print_target};
    $self->log_level($self->get_log_level(%params));
    $self->{umask} = $params{umask};
    $self->{throttling_times} ||= {};
}

#----------------------------------------------------------------------

sub get_log_level {
    my ($self, %params) = @_;
    my $dbvar = $params{dbvar} || $self->{dbvar};
    my $logger_name = $params{logger_name} || $self->{logger_name} || '';
    my $log_level = $params{log_level} || $self->{log_level};

    my %env_log_levels = $self->get_env_log_levels();
    if ($env_log_levels{$dbvar}) {
        return $env_log_levels{$dbvar};
    }
    if ($env_log_levels{$logger_name}) {
        return $env_log_levels{$logger_name};
    }
    if ($env_log_levels{$ALL_MODULES}) {
        return $env_log_levels{$ALL_MODULES};
    }

    # Check for existence of config file, to avoid impossible MCP lookup during build-time
    if (-e $LOGGING_CONF_FILE && $dbvar ne 'none') {
        my $query = F5::ASM::COAPI::Query->new("db_variable", { name => "log.asm.$dbvar.level"});
        if ($query && @$query) {
            my ($obj) = @$query;
            if ($obj->{value}) {
                return $obj->{value};
            }
            elsif ($obj->{default}) {
                return $obj->{default};
            }
        }
        else {
            $self->{dbvar} = 'none';  #Must have been a bad dbvar
        }
    }

    return $log_level;
}

#----------------------------------------------------------------------

sub get_env_log_levels {
    my($self, %params) = @_;

    my @full_env_log_level;
    if ($ENV{F5WAF_LOG_LEVEL}) {
        @full_env_log_level = split(/\s*,\s*/, $ENV{F5WAF_LOG_LEVEL});
    }
    else {
        for my $env_var (keys %ENV) {
            if ($env_var =~ /^F5WAF_LOG_LEVEL_(.*)/) {
                my $module_name = lc($1);
                push(@full_env_log_level, "$module_name=$ENV{$env_var}");
            }
        }
    }

    if (! @full_env_log_level && -e $ENV_LOG_LEVEL_FILE) {
        my $env_log_level_file = YAML::XS::LoadFile($ENV_LOG_LEVEL_FILE);
        @full_env_log_level = ($env_log_level_file->{env_log_level});
    }

    my %env_log_levels;
    for my $log_part (@full_env_log_level) {
        next if (! $log_part);
        my($module_name, $level) = split(/\s*=\s*/, $log_part);
        if (! $level) {
            if (exists $LOG_LEVEL_NAME_MAP{$module_name}) {
                # module_name is a lone value, which is a level for "all"
                $level = $module_name;
                $module_name = $ALL_MODULES;
            }
            else {
                # Otherwise, ignore this log_part -- it seems to be a module_name
                # without a log level.
                next;
            }
        }
        $env_log_levels{$module_name} = $level;
    }

    return %env_log_levels;
}

#----------------------------------------------------------------------

sub update_env_log_level_file {
    my($self, %params) = @_;

    YAML::XS::DumpFile($ENV_LOG_LEVEL_FILE,
                       {
                        env_log_level => $params{env_log_level},
                       });
    update_logging_file();
}

#----------------------------------------------------------------------

sub update_logging_file {
    F5::Utils::File::touch($LOGGING_CONF_FILE);
}

#----------------------------------------------------------------------

sub call_logprint {
  my($self, $level, $msg, $log_opts, %params) = @_;

  if( F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $CONFIG_RECHECK_INTERVAL,
                                                         start_time => $self->{last_config_check}) == 0) {
      if($self->{config_mtime} && -e $LOGGING_CONF_FILE && $self->{config_mtime} != stat($LOGGING_CONF_FILE)->mtime) {
          $self->init();
      }
      else {
          $self->{last_config_check} = time();
      }
  }
      
  my $log_time;
  my $action;

  if ($log_opts) {
      if (ref($log_opts)) {
          $action = $log_opts->{action};
          $log_time = $log_opts->{log_time};
      }
      else {
          # For backward compatibility, when the last argument was used to override action
          $action = $log_opts;
      }
  }

  unless ($action) {
      # Determine action if unspecified
      my $action = (caller(2))[3];
      $action =~ s/^.*:://;
  }

  if ($params{throttling_interval}) {
      my @caller = caller(1);
      my $origin = "$caller[1]:$caller[2]"; # Calling source filename and line number
      if (! $self->{throttling_times}->{$origin} ||
          F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $params{throttling_interval},
                                                             start_time => $self->{throttling_times}->{$origin}) == 0) {

          $self->{throttling_times}->{$origin} = time();
      }
      elsif ($params{throttled_log_level}) {
          $level = uc $params{throttled_log_level};
      }
      else {
          # No throttled_log_level provided -- throttled logs are skipped
          return;
      }
  }

  return if (!$self->is_level($level));

  if (ref($msg)) {
      eval {
          $msg = Dumper($msg);
      };
      if ($@) {
          $Data::Dumper::Useqq = 0;
          $msg = Dumper($msg);
          $Data::Dumper::Useqq = 1;
      }
  }

  if ($params{syslog}) {
      F5::Syslog::ts_syslog($LOG_LEVEL_SYSLOG_MAP{$level}, $msg);
  }

  logprint($level, $action, $msg, $self->{filename}, $self->{print_target}, $log_time, { umask => $self->{umask} });
}

#----------------------------------------------------------------------

sub is_level {
  my($self, $level) = @_;

  return (($self->{print_target}) &&
      ($self->log_level >= $LOG_LEVEL_CONSTANT_MAP{$level}));
}

#----------------------------------------------------------------------

sub log_level {
  my($self, $new_log_level) = @_;

  if (defined($new_log_level)) {
    if ($new_log_level =~ /\D/) {
      $self->{log_level} = $LOG_LEVEL_NAME_MAP{$new_log_level};
    }
    else {
      $self->{log_level} = $new_log_level;
    }
  }

  return $self->{log_level};
}

#----------------------------------------------------------------------

sub info {
  shift->call_logprint($INF, @_);
}

sub debug {
  shift->call_logprint($DBG, @_);
}

sub trace {
  shift->call_logprint($TRACE, @_);
}

sub warn {
  # Log.pm has no warn level, so we elevate it to an error.
  shift->call_logprint($ERR, @_);
}

sub error {
  shift->call_logprint($ERR, @_);
}

sub fatal {
  shift->call_logprint($F_ERR, @_);
}

#----------------------------------------------------------------------

sub is_info {
  shift->is_level($INF);
}

sub is_debug {
  shift->is_level($DBG);
}

sub is_trace {
  shift->is_level($TRACE);
}

sub is_warn {
  shift->is_level($ERR);
}

sub is_error {
  shift->is_level($ERR);
}

sub is_fatal {
  shift->is_level($F_ERR);
}

#----------------------------------------------------------------------

1;

