package F5::MBIP::CtmmSidecar;

use strict;
use warnings;

use Net::NATS::Client;
use YAML::XS;
use File::Find;
use File::Basename;
use File::Slurp;
use File::Copy;
use File::Path;
use MIME::Base64;
use Google::ProtocolBuffers::Dynamic;
use JSON;
use Digest::MD5;

use F5::Deploy;
use F5::ASMConfig::GuaranteedClient;
use F5::ConfigSet;
use F5::SetActive;
use F5::Db::Single;
use F5::Log::Job;

our $config = YAML::XS::LoadFile(F5::Deploy::get_path_by_type("etc") . "/tools/mbip_ctmm_sidecar_config.yaml");
my $PULSE_MSG_INTERVAL = 30;
my $NATS_CONNECT_TIMEOUT = 15;
our $are_protobufs_loaded = 0;

our %HANDLE_ELEMENTS = map { ($_ => 1) }
  qw(waf_policy
     waf_remote_log_profile
     application_security);

our %LOG_PROFILE_FILTER = (0 => 'illegal',
                           1 => 'illegal-including-staged-signatures',
                           2 => 'all',
                          );

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = bless {}, $class;

    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    for my $k (keys %$config) {
        $self->{$k} = $config->{$k};
    }

    for my $k (keys %params) {
        $self->{$k} = $params{$k};
    }

    for my $k (keys %ENV) {
        if ($k =~ /^F5WAF_MBIP_SIDECAR_(.*)$/) {
            $self->{lc $1} = $ENV{$k};
        }
    }
    
    $self->{asmconfig_client} = F5::ASMConfig::GuaranteedClient::get_client(username => 'mbip_sidecar');
    F5::Db::Single->init;
    $self->{dbh} = F5::Db::Single->instance();
    $self->{config_set} = F5::ConfigSet->new(dbh => $self->{dbh},
                                             logger => $self->{logger},
                                             client => $self->{asmconfig_client});

    $self->{messages_handled_count} = 0;
    $self->{log_profiles_by_name} = {};
    $self->{log_profiles_by_id} = {};
    $self->{application_security_objs} = {};
    if ($self->{initial_full_dir}) {
        $self->{initial_full_dir} = F5::Deploy::map_path($self->{initial_full_dir});
        $self->{source_full_dir} = F5::Deploy::map_path($self->{source_full_dir});
        $self->create_initial_full_dir;
    }
}

#----------------------------------------------------------------------

sub reinit {
    my($self, %params) = @_;

    $self->load_protobuf_definitions();
    if (! $self->{nats_client}) {
        $self->connect_to_nats();
    }
}

#----------------------------------------------------------------------

sub connect_to_nats {
    my($self, %params) = @_;

    $self->{logger}->debug("Connecting to NATS") if $self->{logger};
    my $username_prefix = '';
    if ($self->{nats_username}) {
        $username_prefix = $self->{nats_username};
        if ($self->{nats_password}) {
            $username_prefix .= ":$self->{nats_password}";
        }
        $username_prefix .= '@';
    }

    my $connect_rv;
    eval {
        # CPAN module Net::NATS::Client uses module URI under the hood in order to parse the URI. The
        # username and password cannot be extracted if an unknown protocol is used. Thus, although the
        # actual protocol is NATS internal protocol, this is listed as "ftp" in order for URI parsing to
        # work correctly.
        $self->{nats_client} = Net::NATS::Client->new(uri => 'ftp://' . "$username_prefix$self->{nats_host}:$self->{nats_port}",
                                                      socket_args => { Timeout => $NATS_CONNECT_TIMEOUT },
                                                     );

        $connect_rv = $self->{nats_client}->connect();
    };
    if ($@ || ! $connect_rv) {
        my $error_msg = $@ || '';
        $error_msg =~ s| at /\S+ line \d+$||;
        if ($error_msg) {
            $error_msg = "($error_msg)";
        }
        $self->{logger}->error("Couldn't connect to NATS $error_msg") if $self->{logger};
        delete $self->{nats_client};
        return;
    }
    $self->{logger}->info("Connected to NATS") if $self->{logger};

    my $subscription = $self->{nats_client}->subscribe($self->{nats_subject}, sub {
                                                           my ($message) = @_;
                                                           eval {
                                                               $self->handle_message($message);
                                                           };
                                                           if ($@) {
                                                               $self->{logger}->error("Failed to handle NATS message: $@") if $self->{logger}; 
                                                           }
                                                  });
}

#----------------------------------------------------------------------
# Protobuf defintions should only be loaded once for a process. Once .proto files from a particular
# path have been loaded, this will be intentionally skipped.
sub load_protobuf_definitions {
    my($self) = @_;

    return if ($are_protobufs_loaded);

    if (! -d $self->{protobuf_file_path}) {
        $self->{logger}->info("Couldn't find protobuf defintions") if $self->{logger};
        return;
    }
    $self->{logger}->debug("Loading protobuf defintions") if $self->{logger};
    my $protobuf = Google::ProtocolBuffers::Dynamic->new($self->{protobuf_file_path});
    my $was_protobuf_file_found = 0;
    find({
          no_chdir => 1,
          follow_fast => 1,
          wanted => sub {
              my $filename = basename($File::Find::name);
              if (-f $File::Find::name && $filename =~ /\.proto$/) {
                  $protobuf->load_file($filename);
                  $was_protobuf_file_found = 1;
              }
          },
         },
         $self->{protobuf_file_path});

    if ($was_protobuf_file_found) {
        $protobuf->map({package => $self->{protobuf_base_package}, prefix => $self->{protobuf_base_package}});
        $are_protobufs_loaded = 1;
        $self->{logger}->info("Loaded protobuf defintions") if $self->{logger};
    }
}

#----------------------------------------------------------------------

sub run {
    my($self) = @_;

    my $last_pulse_time = time;
    my @potential_poll_times;
    while (1) {
        $self->reinit();
        if ($PULSE_MSG_INTERVAL) {
            push(@potential_poll_times,
                 F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $PULSE_MSG_INTERVAL,
                                                                    start_time => $last_pulse_time));
        }

        my $poll_time = List::Util::min(@potential_poll_times);
        if ($self->{nats_client}) {
            eval {
                $self->{nats_client}->wait_for_op($poll_time);
            };
            if ($@) {
                my $error_msg = $@ || '';
                $error_msg =~ s| at /\S+ line \d+$||;
                if ($error_msg) {
                    $error_msg = "($error_msg)";
                }
                $self->{logger}->error("Couldn't poll NATS server $error_msg") if $self->{logger};
                delete $self->{nats_client};
            }
        }

        if ($PULSE_MSG_INTERVAL && 
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $PULSE_MSG_INTERVAL,
                                                               start_time => $last_pulse_time) == 0) {

            $self->log_pulse_message;
            $last_pulse_time = time;
        }
    }
}

#----------------------------------------------------------------------

sub log_pulse_message {
    my($self) = @_;

    $self->{logger}->info("$self->{messages_handled_count} messages handled since startup") if $self->{logger};
}

#----------------------------------------------------------------------

sub handle_message {
    my($self, $message) = @_;
    my $proto_struct = $self->{protobuf_base_struct};
    my $message_rec = $proto_struct->decode($message->data);

    if ($message_rec && $message_rec->{delta} && $message_rec->{delta}->{changes}) {
        $self->{messages_handled_count}++;
        for my $element_rec (@{$message_rec->{delta}->{changes}}) {
            my($element_name, $element) = %$element_rec;
            $self->{logger}->debug("Received $element_name message (operation = $element->{op})") if $self->{logger};
            $self->{logger}->trace($element) if $self->{logger};

            if ($HANDLE_ELEMENTS{$element_name}) {
                for my $change (@{$element->{changes}}) {
                    for my $k (keys %$change) {
                        $element->{$k} = $change->{$k};
                    }
                }

                my $function_name;
                if ($element->{op} == ctmm::operation::OBJ_CREATE_IF()) {
                    $function_name = 'add_or_update_';
                }
                elsif ($element->{op} == ctmm::operation::OBJ_DELETE()) {
                    $function_name = 'delete_';
                }
                else {
                    $self->{logger}->error("Unrecognized element op code ($element->{op})") if $self->{logger};
                    return;
                }
                $function_name .= $element_name;
                my %job_attributes = (logger => $self->{logger},
                                      operationId => $function_name,
                                      component => 'mbip_sidecar',
                                     );
                if ($element->{name}) {
                    my $rest_element_name = $element_name;
                    $rest_element_name =~ s/_(.)/\U$1/g;
                    $job_attributes{$rest_element_name . "Name"} = $element->{name};
                }
                if ($element->{importContent}) {
                    $job_attributes{policySize} = length($element->{importContent});
                    $job_attributes{policyChecksum} = Digest::MD5->new->add($element->{importContent})->hexdigest;
                }
                my $job = F5::Log::Job->new(%job_attributes);
                $job->set_started();
                $self->$function_name(%$element, job => $job);
                $job->set_completed();
            }
        }
    }
}

#----------------------------------------------------------------------

sub delete_waf_policy {
    my($self, %params) = @_;
    my $policy_name = $params{name};
    $self->{logger}->info("Deleting policy $policy_name") if $self->{logger};

    my $call_flags = F5::ASMConfig::CallFlags->new();

    my $policy_criteria = F5::ASMConfig::Policy->new(policy_name => $policy_name);

    $self->{asmconfig_client}->delete_policy($self->{asmconfig_client}->{session},
                                             $call_flags,
                                             $policy_criteria);
}

#----------------------------------------------------------------------


sub add_or_update_waf_policy {
    my($self, %params) = @_;

    my $import_policy_name = $params{name};

    $self->{logger}->info("Applying updated policy $import_policy_name") if $self->{logger};
    my $call_flags = F5::ASMConfig::CallFlags->new();

    my $filename;
    my $fh;
    ($fh, $filename) = File::Temp::tempfile();
    print $fh $params{importContent};
    close $fh;

    $self->{logger}->info("Preparing importing for policy $import_policy_name")  if $self->{logger};
    my $import_export_properties = F5::ASMConfig::ImportExportProperties->new(filename => $filename,
                                                                              policy_name => $import_policy_name,
                                                                              force_policy_name_flg => 1,
                                                                              overwrite => 1,
                                                                             );

    my $import_rv = $self->{asmconfig_client}->import_policy($self->{asmconfig_client}->{session},
                                                $call_flags,
                                                $import_export_properties);

    unlink $filename;

    my $set_active_rv = F5::SetActive::set_active(dbh => $self->{dbh},
                                                  logger => $self->{logger},
                                                  job => $params{job},
                                                  is_daemon => 0,
                                                  use_daemon => 0,
                                                  policy_id => $import_rv->{policy}->{policy_id},
                                                  update_ramcache => 0,
                                                  do_update_core => 1,
                                                  expanded_return_value => 1,
                                                  do_save_policy_history => 0,
                                                  asm_config_handler => $self->{asmconfig_client});
    if (!$set_active_rv->{completed_successfully}) {
        my $msg = "Failed to compile Policy $import_policy_name";
        $self->{logger}->error($msg)  if $self->{logger};
        die {"error_message" => $msg};
    }
}

#----------------------------------------------------------------------

sub add_or_update_waf_remote_log_profile {
    my($self, %params) = @_;

    my $log_profile = \%params;
    $self->{logger}->info("Adding logging profile $log_profile->{name} ($log_profile->{id})") if $self->{logger};

    my @ip_address_segments = unpack('C*', pack('Q', $log_profile->{addr}->{a2}));
    $log_profile->{ip_address} = join('.', @ip_address_segments[4..7]);
    $self->{log_profiles_by_id}->{$log_profile->{id}} = $log_profile;
    $self->{log_profiles_by_name}->{$log_profile->{name}} = $log_profile;

    $self->prepare_logging_profiles(%params);
}

#----------------------------------------------------------------------

sub delete_waf_remote_log_profile {
    my($self, %params) = @_;

    $self->{logger}->info("Deleting logging profile $params{name}") if $self->{logger};
    my $log_profile = $self->{log_profiles_by_name}->{$params{name}};
    delete $self->{log_profiles_by_name}->{$log_profile->{name}};
    delete $self->{log_profiles_by_id}->{$log_profile->{id}};

    $self->prepare_logging_profiles();
}

#----------------------------------------------------------------------

sub add_or_update_application_security {
    my($self, %params) = @_;

    my $waf_log_profile_id = $params{wafLogProfileId} || '';
    $self->{logger}->info("Adding application security $params{name} (virtualServerId = $params{virtualServerId}, wafLogProfileId = $waf_log_profile_id)") if $self->{logger};
    $self->{application_security_objs}->{$params{name}} = { %params };

    $self->prepare_logging_profiles(%params);
}

#----------------------------------------------------------------------

sub delete_application_security {
    my($self, %params) = @_;

    $self->{logger}->info("Deleting application security $params{name}") if $self->{logger};
    delete $self->{application_security_objs}->{$params{name}};

    $self->prepare_logging_profiles(%params);
}

#----------------------------------------------------------------------

sub prepare_logging_profiles {
    my($self, %params) = @_;

    $self->{logger}->debug("Updating logging profiles") if $self->{logger};

    my @vs_configs;
    for my $security_obj (values %{$self->{application_security_objs}}) {
        if (! $security_obj->{wafLogProfileId}) {
            $self->{logger}->info("WAF Security $security_obj->{name} has no logging profile -- skipping") if $self->{logger};
            next;
        }

        my $log_profile = $self->{log_profiles_by_id}->{$security_obj->{wafLogProfileId}};
        if (! $log_profile) {
            $self->{logger}->info("WAF Log Profile $security_obj->{wafLogProfileId} doesn't exist -- skipping") if $self->{logger};
            next;
        }
        my $logging_config = { logging => {
                                           dest_ip => $log_profile->{ip_address},
                                           dest_port => $log_profile->{port},
                                           filter => $LOG_PROFILE_FILTER{$log_profile->{filterRequestType}},
                                           content => {
                                                       max_message_size => '64k',
                                                      },
                                          },
                             };

        my $vs_config = $self->{config_set}->build_logging_config(policy => { vs_info => { $security_obj->{virtualServerId} => $logging_config } });
        push(@vs_configs, @$vs_config);
    }

    $self->{config_set}->prepare_all_logging_profiles(vs_config => \@vs_configs);
    my $set_active_rv = F5::SetActive::set_active(dbh => $self->{dbh},
                                                  logger => $self->{logger},
                                                  job => $params{job},
                                                  do_update_core => 1,
                                                  asm_config_handler => $self->{asmconfig_client});
}

#----------------------------------------------------------------------

sub create_initial_full_dir {
    my($self) = @_;

    my $set_active_rv = F5::SetActive::set_active(dbh => $self->{dbh},
                                                  logger => $self->{logger},
                                                  do_update_core => 1,
                                                  asm_config_handler => $self->{asmconfig_client});

    mkpath($self->{initial_full_dir});
    find({
          no_chdir => 1,
          follow_fast => 1,
          wanted => sub {
              my $filename = basename($File::Find::name);
              copy($File::Find::name, "$self->{initial_full_dir}/$filename");
          },
         },
         $self->{source_full_dir});
}

#----------------------------------------------------------------------

1;
