package F5::MBIP::Sidecar;

use strict;
use warnings;

use Net::NATS::Client;
use YAML::XS;
use File::Spec;
use File::Find;
use File::Basename;
use File::Slurp;
use File::Copy;
use File::Path;
use MIME::Base64;
use F5::ProtocolBuffers;
use JSON;
use POSIX;
use Digest::MD5;
use Scalar::Util qw(reftype);

use F5::DateTimeUtils;
use F5::Deploy;
use F5::ASMConfig::GuaranteedClient;
use F5::ConfigSet;
use F5::SetActive;
use F5::Db::Single;
use F5::Log::Job;
use F5::DeviceConfig;
use F5::SystemUtils;
use F5::Captcha;
use F5::Utils::File;

our $are_protobufs_loaded = 0;

our $icb_heartbeat_message_type = 'declTmm::dpcMessageV2';

our %HANDLE_ELEMENTS = map { ($_ => 1) }
  qw(waf_policy
     waf_remote_log_profile
     application_security);

our %LOG_PROFILE_FILTER = (0 => 'illegal',
                           1 => 'illegal-including-staged-signatures',
                           2 => 'all',
                          );

our %EXCEPTION_MESSAGE_REASON = (
    0 => 'unspecified',
    1 => 'oversized',
    2 => 'delete',
);

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = bless {}, $class;

    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    my $config = YAML::XS::LoadFile($params{config_file});
    for my $k (keys %$config) {
        $self->{$k} = $config->{$k};
    }

    for my $k (keys %params) {
        $self->{$k} = $params{$k};
    }

    for my $k (keys %ENV) {
        if ($k =~ /^F5WAF_MBIP_SIDECAR_(.*)$/) {
            $self->{lc $1} = $ENV{$k};
        }
    }

    $self->{asmconfig_client} = F5::ASMConfig::GuaranteedClient::get_client(username => 'mbip_sidecar');
    F5::Db::Single->init;
    $self->{dbh} = F5::Db::Single->instance();
    $self->{config_set} = F5::ConfigSet->new(dbh => $self->{dbh},
                                             logger => $self->{logger},
                                             client => $self->{asmconfig_client});

    $self->{messages_handled_count} = 0;
    $self->{log_profiles_by_name} = {};
    $self->{log_profiles_by_id} = {};
    $self->{application_security_objs} = {};

    my $device_config = F5::DeviceConfig->new(dbh => $self->{dbh}, logger => $self->{logger});
    $device_config->apply();

    $self->{captcha} = F5::Captcha->new(dbh => $self->{dbh},
                                        logger => $self->{logger},
                                        asmconfig_client => $self->{asmconfig_client});
    $self->{captcha}->generate_captchas(generate_count => 1, out_dir_suffix => "_init", check_if_captcha_in_use => 1); # generate a single captcha on startup - blocking
    $self->{is_captcha_in_use} = 0;

    if ($self->{initial_full_dir}) {
        $self->{initial_full_dir} = F5::Deploy::map_path($self->{initial_full_dir});
        $self->{source_full_dir} = F5::Deploy::map_path($self->{source_full_dir});
        $self->create_initial_full_dir;
    }

    if ($self->{init_marker_filename}) {
        F5::Utils::File::touch($self->{init_marker_filename});
    }

    $self->{generation} = 0;
    $self->{sequence} = 0;
    $self->{last_sub_id} = 0;
    $self->{entity_map} = {};
}

#----------------------------------------------------------------------

sub reinit {
    my($self, %params) = @_;

    my $rv_proto = $self->load_protobuf_definitions();

    my $rv_nats = 1;
    if ($rv_proto && !$self->{nats_client}) {
        $rv_nats = $self->connect_to_nats();
    }

    return $rv_proto && $rv_nats;
}

#----------------------------------------------------------------------

sub connect_to_nats {
    my($self, %params) = @_;

    $self->{logger}->debug("Connecting to NATS") if $self->{logger};
    my $username_prefix = '';
    if ($self->{nats_username}) {
        $username_prefix = $self->{nats_username};
        if ($self->{nats_password}) {
            $username_prefix .= ":$self->{nats_password}";
        }
        $username_prefix .= '@';
    }

    my $connect_rv;
    eval {
        # CPAN module Net::NATS::Client uses module URI under the hood in order to parse the URI. The
        # username and password cannot be extracted if an unknown protocol is used. Thus, although the
        # actual protocol is NATS internal protocol, this is listed as "ftp" in order for URI parsing to
        # work correctly.
        my %socket_args;
        if ($self->{nats_connect_timeout}) {
            $socket_args{Timeout} = $self->{nats_connect_timeout};
        }
        if ($self->{ssl_ca_file}) {
            $socket_args{SSL_ca_file} = $self->{ssl_ca_file};
        }
        if ($self->{ssl_cert_file}) {
            $socket_args{SSL_cert_file} = $self->{ssl_cert_file};
        }
        if ($self->{ssl_key_file}) {
            $socket_args{SSL_key_file} = $self->{ssl_key_file};
        }
        if ($self->{ssl_version}) {
            $socket_args{SSL_version} = $self->{ssl_version};
        }
        if ($self->{ssl_cipher_list}) {
            $socket_args{SSL_cipher_list} = $self->{ssl_cipher_list};
        }
        $self->{nats_client} = Net::NATS::Client->new(uri => 'ftp://' . "$username_prefix$self->{nats_host}:$self->{nats_port}",
                                                      socket_args => \%socket_args,
                                                     );

        $connect_rv = $self->{nats_client}->connect();
    };
    if ($@ || ! $connect_rv) {
        my $error_msg = $@ || '';
        $error_msg =~ s| at /\S+ line \d+$||;
        if ($error_msg) {
            $error_msg = "($error_msg)";
        }
        $self->{logger}->error("Couldn't connect to NATS $error_msg") if $self->{logger};
        delete $self->{nats_client};
        return;
    }
    $self->{logger}->info("Connected to NATS") if $self->{logger};

    my $icb_subscription = $self->{nats_client}->subscribe($self->{nats_subject_icb}, sub {
                                                           my ($message) = @_;
                                                           eval {
                                                               $self->handle_icb_message(message => $message->data);
                                                           };
                                                           # this error is likely to be an "ASMConfigException" object
                                                           if (my $err = $@) {
                                                               $self->{logger}->error("Failed to handle NATS message:") if $self->{logger};
                                                               $self->{logger}->error($err) if $self->{logger};
                                                           }
                                                  });

    my $jobproxy_subscription = $self->{nats_client}->subscribe($self->{nats_subject_jobproxy_subscribe}, sub {
                                                           my ($message) = @_;
                                                           eval {
                                                               $self->handle_jobproxy_message(message => $message->data);
                                                           };
                                                           # this error is likely to be an "ASMConfigException" object
                                                           if (my $err = $@) {
                                                               $self->{logger}->error("Failed to handle NATS message:") if $self->{logger};
                                                               $self->{logger}->error($err) if $self->{logger};
                                                           }
                                                  });

    return 1;
}

#----------------------------------------------------------------------
# Protobuf defintions should only be loaded once for a process. Once .proto files from a particular
# path have been loaded, this will be intentionally skipped.

sub load_protobuf_definitions {
    my($self) = @_;

    return 1 if ($are_protobufs_loaded);

    if (! -d $self->{protobuf_file_path}) {
        $self->{logger}->info("Couldn't find protobuf defintions") if $self->{logger};
        return $are_protobufs_loaded;
    }

    $self->{logger}->debug("Loading protobuf defintions") if $self->{logger};

    $are_protobufs_loaded = F5::ProtocolBuffers::init(base_dir => $self->{protobuf_file_path},
                                                      package => $self->{protobuf_base_package},
                                                      imports_dir => $self->{protobuf_imports_dir},
                                                     );

    if ($are_protobufs_loaded) {
        $self->{logger}->info("Loaded protobuf defintions") if $self->{logger};
    }

    return $are_protobufs_loaded;
}

#----------------------------------------------------------------------

sub send_icb_heartbeat {
    my($self) = @_;

    if (! $are_protobufs_loaded) {
        $self->{logger}->debug("Skipping ICB heartbeat (protobuf definitions are not loaded)") if $self->{logger};
        return;
    }

    if (! $self->{nats_client}) {
        $self->{logger}->debug("Skipping ICB heartbeat (NATS Client not connected)") if $self->{logger};
        return;
    }

    my $message = $icb_heartbeat_message_type->new();
    # default accessors names are: 'set_' + protobuf message field name
    $message->set_generation($self->{generation});
    $message->set_seq_start($self->{sequence});
    $self->{logger}->trace("Created $icb_heartbeat_message_type message:") if $self->{logger};
    $self->{logger}->trace($message) if $self->{logger};

    my $bytes = $icb_heartbeat_message_type->encode($message);

    if (!$self->{generation} && !$self->{sequence}) {
        $self->{logger}->info("Sending initial heartbeat message") if $self->{logger};
    }

    $self->{nats_client}->request($self->{heartbeat_subject}, $bytes , sub {
                                      my ($reply) = @_;
                                      if (defined $reply && $self->{logger}) {
                                          $self->{logger}->trace("ICB reply:");
                                          $self->{logger}->trace($reply);
                                      }
    });
    $self->{logger}->debug("Sent heartbeat message") if $self->{logger};
}

#----------------------------------------------------------------------

sub run {
    my($self) = @_;

    my $last_pulse_time = time;
    my $last_captcha_time = 0;
    my $last_icb_heartbeat_time = 0;

    while (1) {
        $self->reinit();

        my @potential_poll_times;

        if ($self->{icb_heartbeat_interval}) {
            push(@potential_poll_times,
                 F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{icb_heartbeat_interval},
                                                                    start_time => $last_icb_heartbeat_time));
        }
        if ($self->{pulse_msg_interval}) {
            push(@potential_poll_times,
                 F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{pulse_msg_interval},
                                                                    start_time => $last_pulse_time));
        }

        my $poll_time = List::Util::min(grep { $_ } @potential_poll_times) || 5;
        if ($self->{nats_client}) {
            eval {
                $self->{logger}->trace("Going to poll for $poll_time seconds") if $self->{logger};
                my $rv = $self->{nats_client}->wait_for_op($poll_time);
                my $display_rv = $rv || 'undef';
                $self->{logger}->trace("After poll, rv = $display_rv") if $self->{logger};

                if (! $rv) {
                    my $connection = $self->{nats_client}->connection();
                    if ($connection) {
                        if ($connection->eof()) {
                            $self->{logger}->info("NATS connection was terminated") if $self->{logger};
                            delete $self->{nats_client};
                        }
                        if (my $error_msg = $connection->error()) {
                            $self->{logger}->error("NATS connection ended with error ($error_msg)") if $self->{logger};
                            delete $self->{nats_client};
                        }
                    }
                    else {
                        $self->{logger}->info("NATS connection was lost") if $self->{logger};
                        delete $self->{nats_client};
                    }
                }
            };
            if ($@) {
                my $error_msg = $@ || '';
                $error_msg =~ s| at /\S+ line \d+$||;
                if ($error_msg) {
                    $error_msg = "($error_msg)";
                }
                $self->{logger}->error("Couldn't poll NATS server $error_msg") if $self->{logger};
                delete $self->{nats_client};
            }
        }
        else {
            $self->{logger}->trace("Sleeping for $poll_time seconds") if $self->{logger};
            sleep $poll_time;
        }

        if ($self->{pulse_msg_interval} &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{pulse_msg_interval},
                                                               start_time => $last_pulse_time) == 0) {
            $self->log_pulse_message;
            $last_pulse_time = time;
        }
        if ($self->{captcha_generate_interval} && $self->{is_captcha_in_use} &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{captcha_generate_interval},
                                                               start_time => $last_captcha_time) == 0) {
            $self->spawn_generate_captchas;
            $last_captcha_time = time;
        }
        if ($self->{icb_heartbeat_interval} &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{icb_heartbeat_interval},
                                                               start_time => $last_icb_heartbeat_time) == 0) {

            $self->send_icb_heartbeat();
            $last_icb_heartbeat_time = time;
        }

        if ($self->{current_captcha_pid}) {
            $self->reap_spawned_processes();
        }
    }
}

#----------------------------------------------------------------------

sub log_pulse_message {
    my($self) = @_;

    $self->{logger}->info("$self->{messages_handled_count} messages handled since startup") if $self->{logger};
}

#----------------------------------------------------------------------

sub expand_proto_any_type_messages {
    my($message) = @_;

    if (ref($message) && ref($message) eq 'google.protobuf::Any') {
        $message = decode_single_any_type_message($message);
    }

    if (reftype($message) eq 'HASH') {
        for my $k (keys %$message) {
            if (ref($message->{$k})) {
                $message->{$k} = expand_proto_any_type_messages($message->{$k});
            }
        }
    }
    elsif (reftype($message) eq 'ARRAY') {
        for my $i (0..$#$message) {
            if (ref($message->[$i])) {
                $message->[$i] = expand_proto_any_type_messages($message->[$i]);
            }
        }
    }

    return $message;
}

#----------------------------------------------------------------------

sub decode_single_any_type_message {
    my($any_type_message) = @_;

    my $proto_type = $any_type_message->{type_url};
    $proto_type =~ s|.*/||;
    $proto_type =~ s|\.|::|;

    return $proto_type->decode($any_type_message->{value});
}

#----------------------------------------------------------------------

sub get_ref_suffix {
    my($struct) = @_;

    my $ref_suffix = '';
    if (ref($struct)) {
        $ref_suffix = ref($struct);
        $ref_suffix =~ s/^.*:://;
    }
    return $ref_suffix;
}

#----------------------------------------------------------------------

sub handle_icb_message {
    my($self, %params) = @_;

    my $message = $params{message};

    my $base_proto_type = $self->{protobuf_base_package} . '::' . $self->{protobuf_base_struct};
    my $message_rec;
    eval {
        $message_rec = $base_proto_type->decode($message);
        expand_proto_any_type_messages($message_rec);
    };
    if (my $err = $@) {
        $self->{logger}->error("Couldn't decode $base_proto_type message ($err)") if $self->{logger};
        my $message_as_base64 = encode_base64($message);
        $self->{logger}->debug("Received message = \"$message_as_base64\"") if $self->{logger};
        return;
    }

    $self->{logger}->trace($message_rec) if $self->{logger};
    if (! $message_rec || ! %$message_rec || ! $message_rec->{embedded}) {
        $self->{logger}->debug("Empty message -- skipping") if $self->{logger};
        return;
    }
    # Top-level dataPlaneWrapperV2 only contains "embedded", which is the real content
    $message_rec = $message_rec->{embedded};

    my $message_type = get_ref_suffix($message_rec);

    if ($message_type eq 'unsatisfiableMsgV2' &&
        $message_rec->{generation} == $self->{generation} &&
        $message_rec->{seq_start} == $self->{sequence}) {

        $self->{logger}->info("Received unsatisfiable message -- resetting sequence") if $self->{logger};
        $self->{generation} = 0;
        $self->{sequence} = 0;
        $self->{last_sub_id} = 0;
        return;
    }

    if ($message_type eq 'createMsgV2') {
        my %header_defaults = (generation => 0,
                               seq_end => 0,
                               seq_start => 0,
                               end_flag => 0,
                               sub_id => 0,
                              );
        my $header = $message_rec->{header} || {};
        $header = { %header_defaults, %$header };

        if ($header->{generation} != $self->{generation}) {
            $self->{logger}->info("Skipping message from different generation (generation = $header->{generation})") if $self->{logger};
            return;
        }

        if ($header->{seq_start} != $self->{sequence}) {
            $self->{logger}->debug("Skipping message from different sequence (seq_start = $header->{seq_start}, seq_end = $header->{seq_end})") if $self->{logger};
            return;
        }

        if ($header->{sub_id} && $header->{sub_id} != $self->{last_sub_id} + 1) {
            $self->{logger}->info("Missing sub-message (current = $header->{sub_id}, last received = $self->{last_sub_id})  -- resetting sequence") if $self->{logger};
            $self->{generation} = 0;
            $self->{sequence} = 0;
            $self->{last_sub_id} = 0;
            return;
        }

        for my $entity (@{$message_rec->{embedded}}) {
            my $action = 'add_or_update';
            my $display_action = 'add/update';
            my $entity_type = get_ref_suffix($entity);

            if ($entity_type eq 'exceptionObject') {
                my $reason = $EXCEPTION_MESSAGE_REASON{$entity->{reason}};

                if ($reason eq 'unspecified') {
                    $self->{logger}->info("Unspecified exception in message (seq_start = $header->{seq_start}, seq_end = $header->{seq_end}") if $self->{logger};
                    next;
                }
                if ($reason eq 'oversized') {
                    $self->{logger}->info("Oversized exception in message (seq_start = $header->{seq_start}, seq_end = $header->{seq_end}") if $self->{logger};
                    next;
                }

                if ($reason eq 'delete') {
                    # For some reason, ICB sends deleted entities as an exceptionObject with the "delete" reason and the entity (with only id populated) as
                    # "embedded" under the exceptionObject.
                    # For our purposes, this means that the current entity is the embedded message.
                    $action = 'delete';
                    $display_action = 'delete';
                    $entity = $entity->{embedded};
                    $entity_type = get_ref_suffix($entity);
                }
            }

            if ($HANDLE_ELEMENTS{$entity_type}) {
                my $entity_id = $entity->{id} || '';
                my $entity_name = '';
                if ($entity->{name}) {
                    $entity_name = $entity->{name};
                }
                elsif ($self->{entity_map}->{$entity_id}) {
                    $entity_name = $self->{entity_map}->{$entity_id}->{name};
                }

                my $function_name = "${action}_$entity_type";
                $self->{logger}->info("Applying $display_action for $entity_type \"$entity_name\" (id = $entity_id)") if $self->{logger};

                my $job = $self->create_job(operation_id => $function_name,
                                            entity => $entity,
                                            entity_type => $entity_type,
                                            entity_name => $entity_name,
                                           );
                eval {
                    $job->set_started();
                    $self->$function_name(%$entity, job => $job);
                    $job->set_completed();
                };
                if ($@) {
                    $job->set_failed();
                }
            }
            else {
                $self->{logger}->debug("Skipping unhandled entity (type = $entity_type)") if $self->{logger};
            }
        }

        if ($header->{end_flag} == 1) {
            $self->{generation} = $header->{generation} || 0;
            $self->{sequence} = $header->{seq_end} || 0;
            $self->{last_sub_id} = 0;
            $self->{logger}->debug("Updated current sequence to $self->{sequence}") if $self->{logger};
        }
        else {
            $self->{last_sub_id} = $header->{sub_id};
            $self->{logger}->debug("Updated current sub_id to $self->{last_sub_id}") if $self->{logger};
        }
    }
}

#----------------------------------------------------------------------

sub create_job {
    my($self, %params) = @_;
    my $operation_id = $params{operation_id};
    my $entity = $params{entity};
    my $entity_type = $params{entity_type};
    my $entity_name = $params{entity_name};

    my %job_attributes;
    my $get_attributes_func = "get_job_attributes_for_$entity_type";
    if ($self->can($get_attributes_func)) {
        %job_attributes = $self->$get_attributes_func(%params);

    }

    if ($entity_name && $entity_type) {
        my $rest_element_name = $entity_type;
        $rest_element_name =~ s/_(.)/\U$1/g;
        $job_attributes{$rest_element_name . "Name"} = $entity_name;
    }

    return F5::Log::Job->new(logger => $self->{logger},
                             operationId => $operation_id,
                             component => 'mbip_sidecar',
                             %job_attributes,
                            );
}

#----------------------------------------------------------------------

sub get_job_attributes_for_waf_policy {
    my($self, %params) = @_;
    my $entity = $params{entity};

    my %job_attributes;
    my $import_content;
    if ($entity->{filename}) {
        $import_content = File::Slurp::read_file($entity->{filename});
    }
    elsif ($entity->{importContent}) {
        $import_content = $entity->{importContent};
    }

    if ($import_content) {
        $job_attributes{policySize} = length($import_content);
        $job_attributes{policyChecksum} = Digest::MD5->new->add($import_content)->hexdigest;
    }

    return %job_attributes;
}

#----------------------------------------------------------------------

sub handle_jobproxy_message {
    my($self, %params) = @_;
    my $message = $params{message};
    eval {
        $message = decode_json($message);
    };
    if ($@) {
        $self->{logger}->error("Cannot parse JobProxy message '$message': $@");
        return;
    }

    $self->{logger}->trace("Handling JobProxy message:") if $self->{logger};
    $self->{logger}->trace($message) if $self->{logger};

    my $jobproxy_role = $message->{Role};
    $self->{logger}->debug("Received JobProxy message with role $jobproxy_role") if $self->{logger};
    if ($jobproxy_role ne $self->{job_proxy_role}) {
        $self->{logger}->debug("Job role is not $self->{job_proxy_role}, skipping job.") if $self->{logger};
        return;
    }

    my $message_type = $message->{MessageType};
    if ($message_type eq "Broadcast") {
        $self->accept_jobproxy_job;
    }
    elsif ($message_type eq "Content") {
        # TODO: validate policy here, for now just forwarding the job to icb.
        $self->forward_jobproxy_job_to_icb(id => $message->{Id});
    }
}

#----------------------------------------------------------------------

sub accept_jobproxy_job {
    my($self) = @_;
    my %response = (messageType => "Request",
                    role => $self->{job_proxy_role},
                   );
    $self->send_jobproxy_response(response => \%response);
}

#----------------------------------------------------------------------

sub forward_jobproxy_job_to_icb {
    my($self, %params) = @_;
    my $jobproxy_job_id = $params{id};
    my %response = (messageType => "Response",
                    id => $jobproxy_job_id,
                    status => "PENDING",
                    role => "AGENTCONFIG",
                   );
    $self->send_jobproxy_response(response => \%response);
}

#----------------------------------------------------------------------

sub send_jobproxy_response {
    my($self, %params) = @_;
    my $response = $params{response};
    $self->{logger}->trace("Sending JobProxy response :") if $self->{logger};
    $self->{logger}->trace($response) if $self->{logger};
    my $json_response = encode_json $response;
    $self->{nats_client}->publish($self->{nats_subject_jobproxy_publish}, $json_response);
}

#----------------------------------------------------------------------

sub delete_waf_policy {
    my($self, %params) = @_;
    my $api_uuid = $params{id};

    my $policy_name;
    if ($self->{entity_map}->{$api_uuid} && $self->{entity_map}->{$api_uuid}->{type} eq 'waf_policy') {
        $policy_name = $self->{entity_map}->{$api_uuid}->{name};
    }
    if (! $policy_name) {
        $self->{logger}->info("Policy not found -- skipping delete (api_uuid: $api_uuid)") if $self->{logger};
        return;
    }

    $self->{logger}->info("Deleting policy $policy_name (api_uuid: $api_uuid)") if $self->{logger};

    my $call_flags = F5::ASMConfig::CallFlags->new();

    my $policy_criteria = F5::ASMConfig::Policy->new(policy_name => $policy_name);

    $self->{asmconfig_client}->delete_policy($self->{asmconfig_client}->{session},
                                             $call_flags,
                                             $policy_criteria);

    delete $self->{entity_map}->{$api_uuid};
}

#----------------------------------------------------------------------

sub add_or_update_waf_policy {
    my($self, %params) = @_;
    my $api_uuid = $params{id};

    my $import_policy_name = $params{name};
    my $filename = $params{filename};

    $self->{entity_map}->{$api_uuid} = { type => 'waf_policy',
                                         name => $import_policy_name,
                                       };

    $self->{logger}->info("Applying updated policy $import_policy_name (api_uuid: $api_uuid)") if $self->{logger};

    my $call_flags = F5::ASMConfig::CallFlags->new();

    if (! $filename) {
        my $fh;
        ($fh, $filename) = File::Temp::tempfile();
        print $fh $params{importContent};
        close $fh;
    }

    if (! -f $filename || -z $filename) {
        $self->{logger}->error("No such policy file or empty file at: $filename") if $self->{logger};
        return;
    }

    $self->{logger}->info("Preparing importing for policy $import_policy_name")  if $self->{logger};
    my $import_export_properties = F5::ASMConfig::ImportExportProperties->new(filename => $filename,
                                                                              policy_name => $import_policy_name,
                                                                              force_policy_name_flg => 1,
                                                                              overwrite => 1,
                                                                             );

    my $import_rv = $self->{asmconfig_client}->import_policy($self->{asmconfig_client}->{session},
                                                $call_flags,
                                                $import_export_properties);

    unlink $filename;

    my $set_active_rv = F5::SetActive::set_active(dbh => $self->{dbh},
                                                  logger => $self->{logger},
                                                  job => $params{job},
                                                  is_daemon => 0,
                                                  use_daemon => 0,
                                                  policy_id => $import_rv->{policy}->{policy_id},
                                                  update_ramcache => 0,
                                                  do_update_core => 1,
                                                  expanded_return_value => 1,
                                                  do_save_policy_history => 0,
                                                  asm_config_handler => $self->{asmconfig_client});
    if (!$set_active_rv->{completed_successfully}) {
        my $msg = "Failed to compile Policy $import_policy_name";
        $self->{logger}->error($msg)  if $self->{logger};
        die {"error_message" => $msg};
    }

    if (! $self->{is_captcha_in_use}) {
        $self->{is_captcha_in_use} = $self->{captcha}->is_captcha_in_use;
    }
}

#----------------------------------------------------------------------

sub add_or_update_waf_remote_log_profile {
    my($self, %params) = @_;

    my $log_profile = \%params;
    $self->{logger}->info("Adding logging profile $log_profile->{name} ($log_profile->{id})") if $self->{logger};

    my @ip_address_segments = unpack('C*', pack('Q', $log_profile->{addr}->{a2}));
    $log_profile->{ip_address} = join('.', @ip_address_segments[4..7]);
    $self->{log_profiles_by_id}->{$log_profile->{id}} = $log_profile;
    $self->{log_profiles_by_name}->{$log_profile->{name}} = $log_profile;

    $self->prepare_logging_profiles(%params);
}

#----------------------------------------------------------------------

sub delete_waf_remote_log_profile {
    my($self, %params) = @_;

    $self->{logger}->info("Deleting logging profile $params{name}") if $self->{logger};
    my $log_profile = $self->{log_profiles_by_name}->{$params{name}};
    delete $self->{log_profiles_by_name}->{$log_profile->{name}};
    delete $self->{log_profiles_by_id}->{$log_profile->{id}};

    $self->prepare_logging_profiles();
}

#----------------------------------------------------------------------

sub add_or_update_application_security {
    my($self, %params) = @_;

    my $waf_log_profile_id = $params{wafLogProfileId} || '';
    $self->{logger}->info("Adding application security $params{name} (virtualServerId = $params{virtualServerId}, wafLogProfileId = $waf_log_profile_id)") if $self->{logger};
    $self->{application_security_objs}->{$params{name}} = { %params };

    $self->prepare_logging_profiles(%params);
}

#----------------------------------------------------------------------

sub delete_application_security {
    my($self, %params) = @_;

    $self->{logger}->info("Deleting application security $params{name}") if $self->{logger};
    delete $self->{application_security_objs}->{$params{name}};

    $self->prepare_logging_profiles(%params);
}

#----------------------------------------------------------------------

sub prepare_logging_profiles {
    my($self, %params) = @_;

    $self->{logger}->debug("Updating logging profiles") if $self->{logger};

    my @vs_configs;
    for my $security_obj (values %{$self->{application_security_objs}}) {
        if (! $security_obj->{wafLogProfileId}) {
            $self->{logger}->info("WAF Security $security_obj->{name} has no logging profile -- skipping") if $self->{logger};
            next;
        }

        my $log_profile = $self->{log_profiles_by_id}->{$security_obj->{wafLogProfileId}};
        if (! $log_profile) {
            $self->{logger}->info("WAF Log Profile $security_obj->{wafLogProfileId} doesn't exist -- skipping") if $self->{logger};
            next;
        }
        my $logging_config = { logging => {
                                           dest_ip => $log_profile->{ip_address},
                                           dest_port => $log_profile->{port},
                                           filter => $LOG_PROFILE_FILTER{$log_profile->{filterRequestType}},
                                           content => {
                                                       max_message_size => '64k',
                                                      },
                                          },
                             };

        my $vs_config = $self->{config_set}->build_logging_config(policy => { vs_info => { $security_obj->{virtualServerId} => $logging_config } });
        push(@vs_configs, @$vs_config);
    }

    $self->{config_set}->prepare_all_logging_profiles(vs_config => \@vs_configs);
    my $set_active_rv = F5::SetActive::set_active(dbh => $self->{dbh},
                                                  logger => $self->{logger},
                                                  job => $params{job},
                                                  do_update_core => 1,
                                                  asm_config_handler => $self->{asmconfig_client});
}

#----------------------------------------------------------------------

sub create_initial_full_dir {
    my($self) = @_;

    my $set_active_rv = F5::SetActive::set_active(dbh => $self->{dbh},
                                                  logger => $self->{logger},
                                                  do_update_core => 1,
                                                  asm_config_handler => $self->{asmconfig_client});

    mkpath($self->{initial_full_dir});
    find({
          no_chdir => 1,
          follow_fast => 1,
          wanted => sub {
              my $filename = basename($File::Find::name);
              copy($File::Find::name, "$self->{initial_full_dir}/$filename");
          },
         },
         $self->{source_full_dir});
}

#----------------------------------------------------------------------

sub spawn_generate_captchas {
    my($self, %params) = @_;

    if ($self->{current_captcha_pid} = fork) {
        # in parent
        $self->{logger}->debug("Forked to the parent with child pid $self->{current_captcha_pid}") if $self->{logger};
    }
    else {
        # in child
        exit($self->{captcha}->generate_captchas(%params) || 0);
    }
}

#----------------------------------------------------------------------

sub reap_spawned_processes {
    my($self, %params) = @_;

    if ($self->{current_captcha_pid}) {
        my $rv = waitpid($self->{current_captcha_pid}, WNOHANG);
        if ($rv) {
            my $exit_code = $? >> 8;
            $self->{logger}->debug("Generate captchas: child exited with code: $exit_code") if $self->{logger};
            delete $self->{current_captcha_pid};
        }
    }
}

#----------------------------------------------------------------------

1;
