package F5::NegativeSignatures::Rule;

use strict;
use warnings;

use F5::RE2;
use MIME::Base64;
use Data::MessagePack;

use constant BUFFER_TYPE_ALL_CONTENT => 0x1;
use constant BUFFER_TYPE_URI => 0x2;
use constant BUFFER_TYPE_URI_ONLY => 0x4;
use constant BUFFER_TYPE_PARAM => 0x8;
use constant BUFFER_TYPE_HEADER => 0x10;
use constant BUFFER_TYPE_URI_WILDCARD_OBJ_ONLY => 0x20;
use constant BUFFER_TYPE_BODY => 0x40;
use constant BUFFER_TYPE_PARAM_NORMALIZED => 0x80; 
use constant BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY => 0x400;
use constant BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY => 0x1000;
use constant BUFFER_TYPE_VALUE_CONTENT_XML_ONLY => 0x100; 
use constant BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY => 0x8; 
use constant BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY_NORMALIZED => 0x800;
use constant BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY_NORMALIZED => 0x2000;
use constant BUFFER_TYPE_VALUE_CONTENT_XML_ONLY_NORMALIZED => 0x200; 
use constant BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY_NORMALIZED => 0x80; 
use constant BUFFER_TYPE_VALUE_CONTENT_PLAIN_TEXT_ONLY => 0x8000;
use constant BUFFER_TYPE_VALUE_CONTENT_PLAIN_TEXT_NORMALIZED => 0x10000;
use constant BUFFER_TYPE_VALUE_CONTENT => (BUFFER_TYPE_VALUE_CONTENT_XML_ONLY | BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY | BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY | BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY | BUFFER_TYPE_VALUE_CONTENT_PLAIN_TEXT_ONLY);
use constant BUFFER_TYPE_VALUE_CONTENT_NORMALIZED => (BUFFER_TYPE_VALUE_CONTENT_XML_ONLY_NORMALIZED | BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY_NORMALIZED | BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY_NORMALIZED | BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY_NORMALIZED | BUFFER_TYPE_VALUE_CONTENT_PLAIN_TEXT_NORMALIZED);
use constant BUFFER_TYPE_USER_AGENT => 0x4000;

use constant FORMAT_NUMERIC => 1;
use constant FORMAT_NONNEGATIVE_NUMERIC => 2;

my %format_validate = (
                       FORMAT_NUMERIC() => {
                                            format_name => 'numeric',
                                            validate_func => sub { return ($_[0] =~ /^-?\d+$/); },
                                           },
                       FORMAT_NONNEGATIVE_NUMERIC() => {
                                            format_name => 'non-negative numeric',
                                            validate_func => sub { return ($_[0] =~ /^\d+$/); },
                                           },
                      );

use Carp;
use JSON;

use Exporter ();
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(
                     BUFFER_TYPE_ALL_CONTENT
                     BUFFER_TYPE_URI
                     BUFFER_TYPE_URI_ONLY
                     BUFFER_TYPE_PARAM
                     BUFFER_TYPE_HEADER
                     BUFFER_TYPE_BODY
                     BUFFER_TYPE_URI_WILDCARD_OBJ_ONLY
                     BUFFER_TYPE_PARAM_NORMALIZED
                     BUFFER_TYPE_VALUE_CONTENT
                     BUFFER_TYPE_VALUE_CONTENT_NORMALIZED
                     BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY
                     BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY
                     BUFFER_TYPE_VALUE_CONTENT_XML_ONLY
                     BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY
                     BUFFER_TYPE_VALUE_CONTENT_PLAIN_TEXT_ONLY
                     BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY_NORMALIZED
                     BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY_NORMALIZED
                     BUFFER_TYPE_VALUE_CONTENT_XML_ONLY_NORMALIZED
                     BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY_NORMALIZED
                     BUFFER_TYPE_VALUE_CONTENT_PLAIN_TEXT_NORMALIZED
                     BUFFER_TYPE_USER_AGENT
                  );

our %EXPORT_TAGS = (constants => [ @EXPORT_OK ]);


my %buffer_type_text = (
                        BUFFER_TYPE_ALL_CONTENT() => 'content',
                        BUFFER_TYPE_URI() => 'URI content',
                        BUFFER_TYPE_URI_ONLY() => 'URI content',
                        BUFFER_TYPE_PARAM() => 'HTTP-based value content',
                        BUFFER_TYPE_HEADER() => 'header content',
                        BUFFER_TYPE_BODY() => 'body content',
                        BUFFER_TYPE_URI_WILDCARD_OBJ_ONLY() => 'URI content',
                        BUFFER_TYPE_PARAM_NORMALIZED() => 'HTTP-based value content',
                        BUFFER_TYPE_VALUE_CONTENT() => 'value content',
                        BUFFER_TYPE_VALUE_CONTENT_NORMALIZED() => 'value content',
                        BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY() => 'JSON-based value content',
                        BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY() => 'GWT-based value content',
                        BUFFER_TYPE_VALUE_CONTENT_XML_ONLY() => 'XML-based value content',
                        BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY() => 'HTTP-based value content',
                        BUFFER_TYPE_VALUE_CONTENT_PLAIN_TEXT_ONLY() => 'Plain text value content',
                        BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY_NORMALIZED() => 'JSON-based value content',
                        BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY_NORMALIZED() => 'GWT-based value content',
                        BUFFER_TYPE_VALUE_CONTENT_XML_ONLY_NORMALIZED() => 'XML-based value content',
                        BUFFER_TYPE_VALUE_CONTENT_PLAIN_TEXT_NORMALIZED() => 'Plain text value content',
                        BUFFER_TYPE_USER_AGENT() => 'User-Agent header content',
                       );

my %buffer_type_groups = (
                          general => [ 
                                      BUFFER_TYPE_ALL_CONTENT,
                                      BUFFER_TYPE_URI,
                                      BUFFER_TYPE_URI_ONLY,
                                      BUFFER_TYPE_HEADER,
                                      BUFFER_TYPE_BODY,
                                      BUFFER_TYPE_URI_WILDCARD_OBJ_ONLY,
                                      BUFFER_TYPE_USER_AGENT,
                                     ],
                          value => [
                                    BUFFER_TYPE_VALUE_CONTENT,
                                    BUFFER_TYPE_VALUE_CONTENT_NORMALIZED,
                                   ],
                          'XML-based value' => [
                                                BUFFER_TYPE_VALUE_CONTENT_XML_ONLY,
                                                BUFFER_TYPE_VALUE_CONTENT_XML_ONLY_NORMALIZED,
                                               ],
                          'JSON-based value' => [
                                                BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY,
                                                BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY_NORMALIZED,
                                               ],
                          'GWT-based value' => [
                                                BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY,
                                                BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY_NORMALIZED,
                                               ],
                          'HTTP-based value' => [
                                                 BUFFER_TYPE_PARAM,
                                                 BUFFER_TYPE_PARAM_NORMALIZED,
                                                 BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY,
                                                 BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY_NORMALIZED,
                                               ],
                          'Plain text value' => [
                                                  BUFFER_TYPE_VALUE_CONTENT_PLAIN_TEXT_ONLY,
                                                  BUFFER_TYPE_VALUE_CONTENT_PLAIN_TEXT_NORMALIZED,
                                                ]
                         );

my %group_for_buffer_type;
for my $group (keys %buffer_type_groups) {
  for my $buffer_type (@{$buffer_type_groups{$group}}) {
    $group_for_buffer_type{$buffer_type} = $group;
  }
}

my %regexp_types = ( 'pcre' => 1,
                     're2' => 2,
                     'ipp' => 3,
                   );

my $regexp_types_by_sig_type = { asm => {
                                         'pcre' => 1,
                                         're2' => 2,
                                         'ipp' => 2,
                                        },
                                 dos => {
                                         'ipp' => 3,
                                        }
                               };

my @buffer_types = (
                    {
                     context => 'request',
                     snort_keyword => 'content',
                     buffer_type => BUFFER_TYPE_ALL_CONTENT(),
                     response_context => 'response',
                    },
                    {
                     context => 'uri-with-query-string',
                     snort_keyword => 'uricontent',
                     buffer_type => BUFFER_TYPE_URI(),
                     regexp_flag => 'U',
                    },
                    {
                     context => 'uri-without-query-string',
                     snort_keyword => 'uricontent',
                     buffer_type => BUFFER_TYPE_URI_ONLY(),
                     additional_snort => 'objonly;',
                     regexp_flag => 'U',
                    },
                    {
                     context => 'parameter-cookie-contentprofile',
                     snort_keyword => 'valuecontent',
                     buffer_type => BUFFER_TYPE_VALUE_CONTENT_NORMALIZED(),
                     additional_snort => 'norm;',
                     regexp_flag => 'V',
                    },
                    {
                     context => 'parameter-cookie',
                     snort_keyword => 'valuecontent',
                     buffer_type => BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY_NORMALIZED(),
                     additional_snort => 'httponly; norm;',
                     regexp_flag => 'V',
                    },
                    {
                     context => 'xml',
                     snort_keyword => 'valuecontent',
                     buffer_type => BUFFER_TYPE_VALUE_CONTENT_XML_ONLY_NORMALIZED(),
                     additional_snort => 'xmlonly; norm;',
                     regexp_flag => 'V',
                    },
                    {
                     context => 'json',
                     snort_keyword => 'valuecontent',
                     buffer_type => BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY_NORMALIZED(),
                     additional_snort => 'jsononly; norm;',
                     regexp_flag => 'V',
                    },
                    {
                     context => 'gwt',
                     snort_keyword => 'valuecontent',
                     buffer_type => BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY_NORMALIZED(),
                     additional_snort => 'gwtonly; norm;',
                     regexp_flag => 'V',
                    },
                    {
                     context => 'header',
                     snort_keyword => 'headercontent',
                     buffer_type => BUFFER_TYPE_HEADER(),
                     regexp_flag => 'H',
                    },
                   );

my @comparisons = (
                        {
                         negation => 1,
                         regexp_type => $regexp_types{re2},
                         comparison => 'regexp-does-not-match',
                        },
                        {
                         negation => 0,
                         regexp_type => $regexp_types{re2},
                         comparison => 'regexp-match',
                        },
                        {
                         negation => 1,
                         regexp_type => 0,
                         comparison => 'string-does-not-match',
                        },
                        {
                         negation => 0,
                         regexp_type => 0,
                         comparison => 'string-match',
                        },
                  );

my %context_map = map { ($_->{context} => $_) } @buffer_types;
my %response_context_map = map { ($_->{response_context} => $_) } grep { $_->{response_context} } @buffer_types;
my %comparison_map = map { ($_->{comparison} => $_) } @comparisons;
my %buffer_type_map = map { ($_->{buffer_type} => $_) } @buffer_types;
# When a rule is composed of only URI_ONLY keywords, they are automatically converted to URI_WILDCARD_OBJ_ONLY (handled
# in parse_rule).  For simple keyword conversion purposes, these buffer types are equivalent.
$buffer_type_map{BUFFER_TYPE_URI_WILDCARD_OBJ_ONLY()} = $buffer_type_map{BUFFER_TYPE_URI_ONLY()};

sub new {
  my($class, %params) = @_;

  if ($params{packed_rule_obj}) {
      %params = (%{Data::MessagePack->unpack(F5::Utils::Encode::decode_base64($params{packed_rule_obj}))}, %params);
  }

  my $self = { 
               keywords => $params{keywords} || [],
               attribs  => $params{attribs},
               parse_errors => $params{parse_errors} || [],
               parse_warnings => $params{parse_warnings} || [],
               apply_to => $params{apply_to},
               threshold => $params{threshold},
               sig_type => $params{sig_type} || 'asm',
               logger => $params{logger},
               sig_id => $params{sig_id},
             };
  bless $self, $class;

  if (!@{$self->{keywords}}) {
      $self->parse_rule(rule_string => $params{rule_string}, );
  }

  return $self;
}

#----------------------------------------------------------------------

sub packed_rule_obj {
    my ($self, %params) = @_;
    my %rule_as_hash = %$self;
    return MIME::Base64::encode_base64(Data::MessagePack->pack(\%rule_as_hash), '');
}

#----------------------------------------------------------------------

sub parse_rule {
  my($self, %params) = @_;

  my $rule_string = $params{rule_string};
  my $apply_to = $self->{apply_to};
  my $sig_type = $self->{sig_type};

  my %matching_delimiter = ('(' => ')',
                            '{' => '}',
                            '[' => ']',
                            '<' => '>');
  my %pcre_regexp_flags = map { ($_,1) } qw(s i m x);
  my %pcre_snort_flags = ( 
                          U => { buffer_type => BUFFER_TYPE_URI,
                                 apply_to => 'Request',
                               },
                          O => { buffer_type => BUFFER_TYPE_URI_ONLY,
                                 apply_to => 'Request',
                               },
                          P => { buffer_type => BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY,
                                 apply_to => 'Request',
                               },
                          H => { buffer_type => BUFFER_TYPE_HEADER,
                                 apply_to => 'Request',
                               },
                          N => { buffer_type => BUFFER_TYPE_PARAM_NORMALIZED,
                                 apply_to => 'Request',
                               },
                          V => { buffer_type => BUFFER_TYPE_VALUE_CONTENT,
                                 apply_to => 'Request',
                               },
#                          B => { buffer_type => BUFFER_TYPE_BODY,
#                               },
                         );

  my $element_definitions = {
                             msg => { type => 'top level',
                                      quoted => 1 },
                             reference => { type => 'top level',
                                            allow_repeated => 1,
                                          },
                             dsize => { type => 'top level',
                                        warn_on_usage => 1,
                                      },
                             threshold => { type => 'top level'},
                             flowbits => { type => 'top level',
                                           allow_repeated => 1,
                                           warn_on_usage => 1,
                                         },
                             classtype => { type => 'top level'},
                             sid => { type => 'top level'},
                             rev => { type => 'top level'},
                             flow => { type => 'top level',
                                       warn_on_usage => 1,
                                     },
                             content => { type => 'keyword',
                                          quoted => 1,
                                          buffer_type => BUFFER_TYPE_ALL_CONTENT,
                                          sig_types => [qw(asm)],
                                        },
                             uricontent => { type => 'keyword',
                                             quoted => 1,
                                             buffer_type => BUFFER_TYPE_URI,
                                             apply_to => 'Request',
                                           },
                             paramcontent => { type => 'keyword',
                                               quoted => 1,
                                               buffer_type => BUFFER_TYPE_PARAM,
                                               apply_to => 'Request',
                                               sig_types => [qw(asm)],
                                             },
                             headercontent => { type => 'keyword',
                                                quoted => 1,
                                                buffer_type => BUFFER_TYPE_HEADER,
                                                apply_to => 'Request',
                                              },
#                             bodycontent => { type => 'keyword',
#                                              quoted => 1,
#                                              buffer_type => BUFFER_TYPE_BODY },
                             valuecontent => { type => 'keyword',
                                               quoted => 1,
                                               buffer_type => BUFFER_TYPE_VALUE_CONTENT,
                                               apply_to => 'Request',
                                               sig_types => [qw(asm)],
                                             },
                             pcre => { type => 'keyword',
                                       quoted => 1,
                                       sig_types => [qw(asm)],
                                     },
                             re2 => { type => 'keyword',
                                      quoted => 1,
                                      sig_types => [qw(asm)],
                                    },
                             ipp => { type => 'keyword',
                                      quoted => 1,
                                    },
                             objonly => { type => 'keyword modifier',
                                          default_value => 1,
                                          change_to_buffer_type => {
                                                                    BUFFER_TYPE_URI() => BUFFER_TYPE_URI_ONLY,
                                                                   },
                                          sig_types => [qw(asm)],
                                        },
                             norm => { type => 'keyword modifier',
                                       default_value => 1,
                                       change_to_buffer_type => {
                                                                 BUFFER_TYPE_PARAM() => BUFFER_TYPE_PARAM_NORMALIZED,
                                                                 BUFFER_TYPE_VALUE_CONTENT() => BUFFER_TYPE_VALUE_CONTENT_NORMALIZED,
                                                                 BUFFER_TYPE_VALUE_CONTENT_XML_ONLY() => BUFFER_TYPE_VALUE_CONTENT_XML_ONLY_NORMALIZED,
                                                                 BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY() => BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY_NORMALIZED,
                                                                 BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY() => BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY_NORMALIZED,
                                                                 BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY() => BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY_NORMALIZED,
                                                                 BUFFER_TYPE_VALUE_CONTENT_PLAIN_TEXT_ONLY() => BUFFER_TYPE_VALUE_CONTENT_PLAIN_TEXT_NORMALIZED,
                                                                },

                                       sig_types => [qw(asm)],
                                     },
                             jsononly => { type => 'keyword modifier',
                                          default_value => 1,
                                          change_to_buffer_type => {
                                                                    BUFFER_TYPE_VALUE_CONTENT() => BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY,
                                                                    BUFFER_TYPE_VALUE_CONTENT_NORMALIZED() => BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY_NORMALIZED,
                                                                   },
                                           sig_types => [qw(asm)],
                                         },
                             gwtonly => { type => 'keyword modifier',
                                          default_value => 1,
                                          change_to_buffer_type => {
                                                                    BUFFER_TYPE_VALUE_CONTENT() => BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY,
                                                                    BUFFER_TYPE_VALUE_CONTENT_NORMALIZED() => BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY_NORMALIZED,
                                                                   },
                                          sig_types => [qw(asm)],
                                  },
                             xmlonly => { type => 'keyword modifier',
                                          default_value => 1,
                                          change_to_buffer_type => {
                                                                    BUFFER_TYPE_VALUE_CONTENT() => BUFFER_TYPE_VALUE_CONTENT_XML_ONLY,
                                                                    BUFFER_TYPE_VALUE_CONTENT_NORMALIZED() => BUFFER_TYPE_VALUE_CONTENT_XML_ONLY_NORMALIZED,
                                                                   },
                                          sig_types => [qw(asm)],
                                        },
                             httponly => { type => 'keyword modifier',
                                           default_value => 1,
                                           change_to_buffer_type => {
                                                                     BUFFER_TYPE_VALUE_CONTENT() => BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY,
                                                                     BUFFER_TYPE_VALUE_CONTENT_NORMALIZED() => BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY_NORMALIZED,
                                                                   },
                                           sig_types => [qw(asm)],
                                         },
                             plaintextonly => { type => 'keyword modifier',
                                          default_value => 1,
                                          change_to_buffer_type => {
                                                                    BUFFER_TYPE_VALUE_CONTENT() => BUFFER_TYPE_VALUE_CONTENT_PLAIN_TEXT_ONLY,
                                                                    BUFFER_TYPE_VALUE_CONTENT_NORMALIZED() => BUFFER_TYPE_VALUE_CONTENT_PLAIN_TEXT_NORMALIZED,
                                                                   },
                                          sig_types => [qw(asm)],
                                        },
                             rawbytes => { type => 'keyword modifier',
                                           default_value => 1,
                                         },
                             nocase => { type => 'keyword modifier',
                                         default_value => 1 },
                             isdataat => { type => 'top level',
                                           warn_on_usage => 1,
                                           format => FORMAT_NONNEGATIVE_NUMERIC,
                                         },
                             isdataat_relative => { type => 'keyword modifier',
                                                    format => FORMAT_NONNEGATIVE_NUMERIC,
                                                  },
                             depth => { type => 'keyword modifier',
                                        format => FORMAT_NONNEGATIVE_NUMERIC,
                                      },
                             offset => { type => 'keyword modifier',
                                        format => FORMAT_NONNEGATIVE_NUMERIC,
                                       },
                             within => { type => 'keyword modifier',
                                         relative => 1,
                                         format => FORMAT_NONNEGATIVE_NUMERIC,
                                         sig_types => [qw(asm)],
                                       },
                             distance => { type => 'keyword modifier',
                                           relative => 1,
                                           format => FORMAT_NONNEGATIVE_NUMERIC,
                                           sig_types => [qw(asm)],
                                         },
                             'useragentonly' => { type => 'keyword modifier',
                                                  default_value => 1,
                                                  change_to_buffer_type => {
                                                                            BUFFER_TYPE_HEADER() => BUFFER_TYPE_USER_AGENT,
                                                                           },
                                        },
                            };

  for my $k (keys %$element_definitions) {
      if ($element_definitions->{$k}->{sig_types} && ! grep($_ eq $sig_type, @{$element_definitions->{$k}->{sig_types}})) {
          delete $element_definitions->{$k};
      }
  }

  unless ($rule_string) {
    $self->add_parse_error("Rule is empty");
  }
  my @elements;

  if ($rule_string) {
      # Remove routing info at the beginning of an externally obtained Snort rule
      if ($rule_string =~ /^ \s* (.*?) \s+ \( \s* (.*?) \s* \) \s* $/x) {
          my ($routing_data, $rule_only) = ($1,$2);
          my @routing_data = split(/\s+/, $routing_data);
          if (scalar @routing_data == 7) {
              $rule_string = $rule_only;
          }
      }

      $rule_string =~ s/^\s*//;
      unless ($rule_string =~ s/;\s*$//) {
          $self->add_parse_error("Missing terminating semicolon on final tag");
      }
      @elements = split(/\s*;\s+/, $rule_string);
      unless (@elements) {
          $self->add_parse_error("Rule is empty");
      }
  }

  for my $elem (@elements) {
    my($key, $value) = split(/\s*:\s*/, $elem, 2);
    my $orig_value = $value;

    # isdataat is a top-level element, isdataat:x,relative is a keyword modifier
    if ($key eq 'isdataat' && $value =~ s/\s*,\s*relative$//) {
      $key = 'isdataat_relative';
    }

    my $element_definition = $element_definitions->{$key};

    my $found_tag_error = 0;
    unless ($element_definition) {
      $self->add_parse_error("Invalid tag: $key");
      $found_tag_error++;
    }

    if ($element_definition->{apply_to} && $apply_to && $element_definition->{apply_to} ne $apply_to) {
      $self->add_parse_error("Invalid tag for $apply_to: $key");
      $found_tag_error++;
    }

    unless ($elem =~ /:/ ||
            (exists $element_definitions->{$elem} &&
             $element_definitions->{$elem}->{default_value})) {
      $self->add_parse_error("Tag found with no value: $elem");
      $found_tag_error++;
    }

    if ($element_definition->{warn_on_usage}) {
      $self->add_parse_warning("Snort tag unsupported by ASM: $key");
    }

    if (exists $element_definition->{format}) {
      if (!$format_validate{$element_definition->{format}}->{validate_func}->($value)) {
        my $format_name = $format_validate{ $element_definition->{format} }->{format_name};
        $self->add_parse_error("Invalid format for $key: $value (must be $format_name)");
        $found_tag_error++;
      }
    }

    next if ($found_tag_error);

    $value = $element_definition->{default_value} unless defined($value);

    my $negation = 0;
    if ($element_definition->{type} eq 'keyword' && $value =~ s/^!//) {
        if ($sig_type eq 'dos') {
            $self->add_parse_error("Negation not allowed: $key");
        }
        else {
            $negation = 1;
        }
    }

    if ($element_definition->{quoted}) {
      unless ($value =~ s/^\s*"(.*)"\s*$/$1/) {
        $self->add_parse_error("Unquoted content: $key");
      }

      if ($value =~ /"/) {
        $self->add_parse_error("Unescaped quote in $key: $value");
        next;
      }
    }

    if ($element_definition->{type} eq 'top level') {
      if (exists $self->{attribs}->{$key}) {

        if ($element_definition->{allow_repeated}) {
          if (!ref($self->{attribs}->{$key})) {
            $self->{attribs}->{$key} = [ $self->{attribs}->{$key} ];
          }

          push(@{$self->{attribs}->{$key}}, $value);
        }
        else {
          $self->add_parse_error("Repeated tag: $key");
          next;
        }
      }
      else {
        if ($key eq 'threshold') {
          my $threshold_attribs = $self->parse_threshold($value);
          next unless $threshold_attribs;
          $self->{threshold} = $threshold_attribs;
        }
        else {
          $self->{attribs}->{$key} = $value;
        }
      }
    }
    elsif ($element_definition->{type} eq 'keyword') {
      my $keyword = {};
      $keyword->{negation} = $negation;

      $keyword->{regexp_type} = $regexp_types_by_sig_type->{$sig_type}->{$key} || 0;
      if ($regexp_types{$key}) {
        $keyword->{buffer_type} = BUFFER_TYPE_ALL_CONTENT;
        my ($pcre_begin_delimiter) = ($value =~ /^m?(.)/);
        if ($pcre_begin_delimiter =~ /\w/) {
          $self->add_parse_error("Invalid delimiter: $pcre_begin_delimiter");
          next;
        }

        my $pcre_end_delimiter = $matching_delimiter{$pcre_begin_delimiter} || $pcre_begin_delimiter;

        my ($pattern, $match_flags);
        if ($value =~ /^m?\Q$pcre_begin_delimiter\E(.*)\Q$pcre_end_delimiter\E(.*)$/s) {
          ($pattern, $match_flags) = ($1, $2);

          my $found_regexp_error = 0;

          if ($key eq 'pcre' || $key eq 'ipp') {
              # Validate pcre in perl
              eval "my \$temp = qr/\$pattern/;";
              if ($@) {
                  my $eval_error = $@;
                  $eval_error =~ s/ at \(eval .*$//s;
                  $self->add_parse_error("Invalid PCRE: $orig_value ($eval_error)");
                  $found_regexp_error++;
              }
          }

          if ($key eq 're2') {
              my $validate_rv = F5::RE2::is_re2_valid($pattern);
	      if ($validate_rv) {
		  $self->add_parse_error("Invalid RE2: $orig_value ($validate_rv)");
		  $found_regexp_error++;
	      }
          }

          if ($match_flags) {
            my @match_flags = split(//, $match_flags);

            if (my @unknown_flags = grep(!$pcre_regexp_flags{$_} && !$pcre_snort_flags{$_} && $_ ne 'R', @match_flags)) {
              for my $unknown_flag (@unknown_flags) {
                $self->add_parse_error("Unknown regular expression flag: $unknown_flag");
              }
              $found_regexp_error++;
            }


            my @regexp_flags = grep($pcre_regexp_flags{$_}, @match_flags);
            if (@regexp_flags) {
              $pattern = '(?' . join('', @regexp_flags) . ")$pattern";
            }
            my @snort_flags = grep($pcre_snort_flags{$_}, @match_flags);
            if (@snort_flags) {
              if (scalar @snort_flags > 1) {
                $self->add_parse_error("Mutually exclusive regular expression flags used: " . join(',', @snort_flags));
                $found_regexp_error++;
              }
              else {
                if ($pcre_snort_flags{$snort_flags[0]}->{apply_to} && $apply_to &&
                    $pcre_snort_flags{$snort_flags[0]}->{apply_to} ne $apply_to) {
                  $self->add_parse_error("Invalid regular expression flag for $apply_to: $snort_flags[0]");
                  $found_regexp_error++;
                }
                else {
                  $keyword->{buffer_type} = $pcre_snort_flags{$snort_flags[0]}->{buffer_type};
                }
              }
            }

            if (grep($_ eq 'R', @match_flags)) {
              $keyword->{is_relative} = 1;
            }
          }
          next if $found_regexp_error;

          # Remove unnecessary escaping backslash from delimiter
          $pattern =~ s/\\(\Q$pcre_end_delimiter\E)/$1/g;

          $keyword->{text} = $pattern;
        }
        else {
          $self->add_parse_error("Regular expression not delimited properly: $orig_value");
          next;
        }
      }
      else {
        # Unescape pipe-escapes in non-pcre match text
        $value =~ s/\|(.*?)\|/$self->convert_pipe_escape($1)/ge;

        $keyword->{text} = $value;

        $keyword->{buffer_type} = (exists $element_definition->{buffer_type} ?
                                   $element_definition->{buffer_type} : BUFFER_TYPE_ALL_CONTENT);
      }

      push(@{$self->keywords}, $keyword);
    }
    elsif ($element_definition->{type} eq 'keyword modifier') {
      unless (@{$self->keywords}) {
        $self->add_parse_error("Content modifier on non-content tag: $key");
        next;
      }

      if (scalar @{$self->keywords} < 2 && $element_definition->{relative}) {
        $self->add_parse_error("Relative modifier cannot be applied to the first content tag: $key");
        next;
      }

      if (exists $element_definition->{change_to_buffer_type}) {
        my $last_keyword = $self->keywords->[-1];
        if (!exists $element_definition->{change_to_buffer_type}->{$last_keyword->{buffer_type}}) {
          $self->add_parse_error("Modifier cannot be applied to $buffer_type_text{$last_keyword->{buffer_type}} buffer type: $key");
          next;
        }
        $last_keyword->{buffer_type} = $element_definition->{change_to_buffer_type}->{$last_keyword->{buffer_type}};
      }

      if (exists $self->keywords->[-1]->{$key}) {
        $self->add_parse_error("Repeated modifier on content tag: $key");
        next;
      }

      $self->keywords->[-1]->{$key} = $value;
      if ($element_definition->{relative}) {
        $self->keywords->[-1]->{is_relative} = 1;
      }
    }
  }

  my @relative_keyword_modifiers = grep($element_definitions->{$_}->{type} eq 'keyword modifier' &&
                                        $element_definitions->{$_}->{relative}, 
                                        keys %$element_definitions);
  for my $i (1..$#{$self->keywords}) {
    my $last_buffer_type_group = $group_for_buffer_type{$self->keywords->[$i - 1]->{buffer_type}};
    my $this_buffer_type_group = $group_for_buffer_type{$self->keywords->[$i]->{buffer_type}};
    if ($last_buffer_type_group ne $this_buffer_type_group) {
      $self->add_parse_error(ucfirst $last_buffer_type_group . " content and $this_buffer_type_group content cannot be combined in a single rule.");
    }

    my @violating_modifiers = grep(exists $self->keywords->[$i]->{$_}, @relative_keyword_modifiers);
    if (@violating_modifiers &&
        $self->keywords->[$i - 1]->{buffer_type} != $self->keywords->[$i]->{buffer_type}) {
      $self->add_parse_error("Relative modifer not applicable to differing buffer types: " . 
                             join(',', @violating_modifiers));
    }
  }

  if ($sig_type eq 'dos' && grep($_->{buffer_type} == BUFFER_TYPE_HEADER, @{$self->keywords})) {
      $self->add_parse_error("headercontent cannot be used without useragentonly");
  }

  # If the rule only contains BUFFER_TYPE_URI_ONLY keywords, change them all to BUFFER_TYPE_URI_WILDCARD_OBJ_ONLY
  if (!grep($_->{buffer_type} != BUFFER_TYPE_URI_ONLY, @{$self->keywords})) {
    for my $keyword (@{$self->keywords}) {
      $keyword->{buffer_type} = BUFFER_TYPE_URI_WILDCARD_OBJ_ONLY;
    }
  }

  if ($self->parse_errors) {
    die join("\n", $self->parse_errors) . "\n";
  }
}

#----------------------------------------------------------------------

sub parse_threshold {
  my($self, $value) = @_;
  my $found_threshold_error = 0;
  my %threshold_attribs;
  if ($value =~ /^\s* type \s+ (limit|threshold|both) \s* , 
                 \s+ track \s+ (by_src|by_dst) \s*, 
                 \s+ count \s+ (\d+) \s* , 
                 \s+ seconds \s+ (\d+) \s* $/x) {
    ($threshold_attribs{type}, $threshold_attribs{track}, $threshold_attribs{count}, 
     $threshold_attribs{seconds}) = ($1, $2, $3, $4);

    if ($threshold_attribs{type} ne 'limit') {
      $self->add_parse_error("Unsupported threshold type: $threshold_attribs{type}");
      $found_threshold_error++;
    }

    if ($threshold_attribs{track} ne 'by_src') {
      $self->add_parse_error("Unsupported threshold track: $threshold_attribs{track}");
      $found_threshold_error++;
    }

    if ($threshold_attribs{count} != 1) {
      $self->add_parse_error("Unsupported threshold count: $threshold_attribs{count}");
      $found_threshold_error++;
    }

    if ($found_threshold_error) {
      return;
    }
    else {
      return \%threshold_attribs;
    }
  }
}

#----------------------------------------------------------------------

sub convert_pipe_escape {
  my($self, $hex) = @_;
  $hex =~ s/ //g;

  my @hex_pairs = unpack("(A2)*", $hex);
  if (my @invalid_pairs = grep { $_ !~ /[0-9a-f]{2}/i } @hex_pairs) {
    $self->add_parse_error("Invalid escaped hex string: " . join(',', @invalid_pairs));
    return '';
  }
  return join('', map { chr(hex($_)) } @hex_pairs);
}

#----------------------------------------------------------------------

sub keyword_count {
  my($self) = @_;
  return scalar @{$self->keywords};
}

#----------------------------------------------------------------------

sub non_regexp_keyword_count {
  my($self) = @_;
  return scalar grep(!$_->{regexp_type} && !$_->{negation}, @{$self->keywords});
}

#----------------------------------------------------------------------

sub regexp_keywords {
  my($self) = @_;
  return [ grep($_->{regexp_type}, @{$self->keywords}) ];
}

#----------------------------------------------------------------------

sub flg_has_valuecontent_http {
  my($self) = @_;
  if (scalar grep (($_->{buffer_type} & BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY) ||
                   ($_->{buffer_type} & BUFFER_TYPE_VALUE_CONTENT_HTTP_ONLY_NORMALIZED),
                   @{$self->keywords})) {
    return 1;
  }
  return 0;
}

#----------------------------------------------------------------------

sub flg_has_valuecontent_xml {
  my($self) = @_;
  if (scalar grep (($_->{buffer_type} & BUFFER_TYPE_VALUE_CONTENT_XML_ONLY) ||
                   ($_->{buffer_type} & BUFFER_TYPE_VALUE_CONTENT_XML_ONLY_NORMALIZED),
                   @{$self->keywords})) {
    return 1;
  }
  return 0;
}

#----------------------------------------------------------------------

sub flg_has_valuecontent_json {
  my($self) = @_;
  if (scalar grep (($_->{buffer_type} & BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY) ||
                   ($_->{buffer_type} & BUFFER_TYPE_VALUE_CONTENT_JSON_ONLY_NORMALIZED),
                   @{$self->keywords})) {
    return 1;
  }
  return 0;
}
#----------------------------------------------------------------------

sub flg_has_valuecontent_gwt {
  my($self) = @_;
  if (scalar grep (($_->{buffer_type} & BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY) ||
                   ($_->{buffer_type} & BUFFER_TYPE_VALUE_CONTENT_GWT_ONLY_NORMALIZED),
                   @{$self->keywords})) {
    return 1;
  }
  return 0;
}
#----------------------------------------------------------------------

sub flg_has_valuecontent_plain_text {
  my($self) = @_;
  if (scalar grep (($_->{buffer_type} & BUFFER_TYPE_VALUE_CONTENT_PLAIN_TEXT_ONLY) ||
                   ($_->{buffer_type} & BUFFER_TYPE_VALUE_CONTENT_PLAIN_TEXT_NORMALIZED),
                   @{$self->keywords})) {
    return 1;
  }
  return 0;
}
#----------------------------------------------------------------------

sub flg_has_headercontent {
  my($self) = @_;
  if (scalar grep (($_->{buffer_type} & BUFFER_TYPE_HEADER),
                   @{$self->keywords})) {
    return 1;
  }
  return 0;
}

#----------------------------------------------------------------------

sub flg_has_uricontent {
  my($self) = @_;
  if ((scalar grep { ($_->{buffer_type} & BUFFER_TYPE_URI) ||
                     ($_->{buffer_type} & BUFFER_TYPE_URI_ONLY) ||
                     ($_->{buffer_type} & BUFFER_TYPE_URI_WILDCARD_OBJ_ONLY)
                   }
                   @{$self->keywords}) ||
      $self->flg_has_requestcontent) {

    return 1;
  }
  return 0;
}

#----------------------------------------------------------------------

sub flg_has_requestcontent {
  my($self) = @_;
  if ($self->{apply_to} eq 'Request' &&
      scalar grep (($_->{buffer_type} & BUFFER_TYPE_ALL_CONTENT),
                   @{$self->keywords})) {
    return 1;
  }
  return 0;
}

#----------------------------------------------------------------------

sub flg_has_responsecontent {
    my($self) = @_;
    if ($self->{apply_to} eq 'Response' &&
        scalar grep (($_->{buffer_type} & BUFFER_TYPE_ALL_CONTENT),
                     @{$self->keywords})) {
        return 1;
    }
    return 0;
}

#----------------------------------------------------------------------

sub add_parse_error {
  my($self, $message) = @_;
  push(@{$self->{parse_errors}}, $message);
}

#----------------------------------------------------------------------

sub parse_errors {
  return @{$_[0]->{parse_errors}};
}

#----------------------------------------------------------------------

sub add_parse_warning {
  my($self, $message) = @_;
  push(@{$self->{parse_warnings}}, $message);
}

#----------------------------------------------------------------------

sub parse_warnings {
  return @{$_[0]->{parse_warnings}};
}
#----------------------------------------------------------------------

sub sig_references {
  my($self) = @_;
  my $sig_references = $self->{attribs}->{reference};
  if (!$sig_references) {
    $sig_references = [];
  }
  elsif (!ref($sig_references)) {
    $sig_references = [ $sig_references ];
  }

  for my $sig_reference (@$sig_references) {
    $sig_reference =~ s|^url,https?://|url,|;
  }

  my @sig_references = grep($_ ne 'url,', @$sig_references);
  return \@sig_references;
}

#----------------------------------------------------------------------

sub buffer_type_mask {
  my($self) = @_;

  my $buffer_type_mask = 0;
  for my $keyword (@{$self->{keywords}}) {
    $buffer_type_mask |= $keyword->{buffer_type};
  }

  if ($self->isdataat) {
    $buffer_type_mask |= BUFFER_TYPE_ALL_CONTENT;
  }

  return $buffer_type_mask;
}

#----------------------------------------------------------------------
# to_simple_keywords() returns a list of keyword hashes for simplified signature creation.
# This is not to be confused with the internal keywords returned by keywords().
sub to_simple_keywords {
    my($self) = @_;
    my @simple_keywords;
    my %allowed_simple_keyword_attributes = map { ($_ => 1) } qw(buffer_type
                                                                 regexp_type
                                                                 negation
                                                                 text
                                                                 norm
                                                                 nocase
                                                                 httponly
                                                                 xmlonly
                                                                 jsononly
                                                                 gwtonly
                                                                 objonly
                                                               );

    for my $internal_keyword (@{$self->{keywords}}) {

        if (my @disallowed_simple_keyword_attributes = grep { ! $allowed_simple_keyword_attributes{$_ } } keys %$internal_keyword) {
            $self->{logger}->trace("Could not convert to keywords with specified attributes (" . join(', ', @disallowed_simple_keyword_attributes) . ")") if $self->{logger};
            return [];
        }

        if ($internal_keyword->{text} =~ /[\cA-\cZ]/) {
            $self->{logger}->trace("Could not convert to keywords with unprintable characters") if $self->{logger};
            return [];
        }

        my $simple_keyword = { 'value' => $internal_keyword->{text} };
        if (! $buffer_type_map{$internal_keyword->{buffer_type}}) {
            $self->{logger}->trace("Could not convert to keywords with buffer type $internal_keyword->{buffer_type}") if $self->{logger};
            return [];

        }

        if ($self->{apply_to} eq 'Response') {
            $simple_keyword->{context} = $buffer_type_map{$internal_keyword->{buffer_type}}->{response_context};
            if (! $simple_keyword->{context}) {
                $self->{logger}->trace("Could not convert to keywords with buffer type $internal_keyword->{buffer_type} for response signature") if $self->{logger};
                return [];

            }

        }
        else {
            $simple_keyword->{context} = $buffer_type_map{$internal_keyword->{buffer_type}}->{context};
        }

        my($comparison) = grep { ($_->{negation} == $internal_keyword->{negation} || 0) &&
                                   ($_->{regexp_type} == $internal_keyword->{regexp_type} || 0) } @comparisons;

        if (! $comparison) {
            $self->{logger}->trace("Comparison could not be determined for keyword '$simple_keyword->{value}'") if $self->{logger};
            return [];
        }

        $simple_keyword->{comparison} = $comparison->{comparison};

        $simple_keyword->{match_case} = $internal_keyword->{nocase} ? 0 : 1;
        push(@simple_keywords, $simple_keyword);
    }

    return \@simple_keywords;
}

#----------------------------------------------------------------------

sub from_simple_keywords {
    my($self, %params) = @_;

    my $apply_to = $params{apply_to};
    if (! $apply_to && ref($self)) {
        $apply_to = $self->{apply_to};
    }
    $apply_to ||= 'Request';

    my $rule = '';
    for my $simple_keyword (@{$params{keywords}}) {
        if (! $simple_keyword->{context}) {
            die "Context was not specified\n";
        }
        my $buffer_type;
        if ($apply_to eq 'Response') {
            if (! $response_context_map{$simple_keyword->{context}}) {
                die "Unknown context $simple_keyword->{context}\n";
            }
            $buffer_type = $response_context_map{$simple_keyword->{context}};
        }
        else {
            if (! $context_map{$simple_keyword->{context}}) {
                die "Unknown context $simple_keyword->{context}\n";
            }
            $buffer_type = $context_map{$simple_keyword->{context}};
        }

        if (! $simple_keyword->{comparison}) {
            die "Comparison was not specified\n";
        }
        if (! $comparison_map{$simple_keyword->{comparison}}) {
            die "Unknown comparison $simple_keyword->{comparison}\n";
        }
        my $comparison = $comparison_map{$simple_keyword->{comparison}};

        if (! defined $simple_keyword->{'value'} || $simple_keyword->{'value'} eq '') {
            die "Keyword may not be empty\n";
        }

        if ($comparison->{regexp_type}) {
            $rule .= 're2:';
            if ($comparison->{negation}) {
                $rule .= '!';
            }
            my $value = $simple_keyword->{'value'};
            $value =~ s|(?!\\)/|\\/|g;
            $value =~ s|"|\\x22|g;
            my $regexp_flag = $buffer_type->{regexp_flag} || '';
            $rule .= "\"/$value/$regexp_flag\"; "
        }
        else {
            $rule .= $buffer_type->{snort_keyword} . ':';
            if ($comparison->{negation}) {
                $rule .= '!';
            }
            my $value = $simple_keyword->{'value'};
            $value =~ s/(["'|;]+)/'|' .  pipe_escape($1) . '|'/ge;

            $rule .= '"' . $value . '"; '
        }
        if (defined($simple_keyword->{match_case}) && $simple_keyword->{match_case} == 0) {
            $rule .= 'nocase; ';
        }
        if ($buffer_type->{additional_snort}) {
            $rule .= $buffer_type->{additional_snort} . ' ';
        }
    }
    $rule =~ s/\s*$//;

    # Parse the constructed rule only in order to determine if the resulting rule is illegal
    # (for example, if conflicting contexts are used together).
    my $rule_obj = $self->new(rule_string => $rule, apply_to => $apply_to);

    # Return constructed rule string (not parsed rule object)
    return $rule;
}

#----------------------------------------------------------------------

sub pipe_escape {
    return join(" ", map { unpack("H*", $_) } split(//, $_[0]));
}

#----------------------------------------------------------------------
# Accessors

sub keywords {
  return $_[0]->{keywords};
}

#----------------------------------------------------------------------

sub sig_id {
  return $_[0]->{sig_id};
}

#----------------------------------------------------------------------

sub threshold {
  my($self, $threshold_key) = @_;
  return unless $self->{threshold};
  return $self->{threshold}->{$threshold_key};
}

#----------------------------------------------------------------------

sub isdataat {
  my($self) = @_;
  return $self->{attribs}->{isdataat};
}

1;
