package F5::NegativeSignatures::Utils;

use strict;
use warnings;
use F5::ASMConfig::Entity::SignatureSystem;
use F5::Utils::Rest;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(find_or_add_system find_or_add_attack_type);

use Memoize;
use F5::Deploy;
use F5::DbUtils;

memoize('find_or_add_system');
memoize('find_or_add_attack_type');
memoize('find_or_add_system_group');
memoize('get_attack_type_name');
memoize('get_attack_type_id');
memoize('get_set_id_name');
memoize('get_schema_attributes', NORMALIZER => 'memoize_only_without_arguments');
my $get_schema_attributes_unique_count = 2;

my $XML_SCHEMA_URI = "http://www.w3.org/2001/XMLSchema";
our $NEGSIG_COMMON_SCHEMA_FILENAME = $ENV{ASM_NEGSIG_COMMON_SCHEMA_FILENAME} || F5::Deploy::map_path('/ts/var/schema/negsig/common.xsd');

#----------------------------------------------------------------------

sub get_attack_type_name {
    my($class, %params) = @_;

    my $dbh = $params{dbh};
    my $attack_type_id = $params{attack_type_id};
    my $plc_dbname = 'PLC';

    my ($attack_type_name) = $dbh->selectrow_array(<<EndQuery, undef, $attack_type_id);
SELECT attack_type_name
  FROM $plc_dbname.NEGSIG_ATTACK_TYPES
 WHERE attack_type_id = ?
EndQuery
  
    return $attack_type_name;
}

#------------------------------------------------------------------------

sub get_attack_type_id {
    my($class, %params) = @_;

    my $dbh = $params{dbh};
    my $attack_type_name = $params{attack_type_name};
    my $plc_dbname = 'PLC';

    my ($attack_type_id) = $dbh->selectrow_array(<<EndQuery, undef, $attack_type_name);
SELECT attack_type_id
  FROM $plc_dbname.NEGSIG_ATTACK_TYPES
 WHERE attack_type_name = ?
EndQuery
  
    return $attack_type_id;
}

#----------------------------------------------------------------------

sub get_set_id_name {
    my(%params) = @_;

    my $dbh = $params{dbh};
    my $dbname = $params{dbname};
    my $attack_type_id = $params{attack_type_id};

    my ($set_id, $set_name) = $dbh->selectrow_array(<<EndQuery, undef, $attack_type_id);
SELECT ns.set_id, ns.set_name
  FROM PLC.NEGSIG_SET_FILTERS nsf,
       PLC.NEGSIG_SETS ns
 WHERE nsf.set_id = ns.set_id
   AND nsf.attack_type_id = ?
EndQuery

    return {set_id => $set_id, set_name => $set_name}; 

}

#----------------------------------------------------------------------

sub find_or_add_system {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $dbname = $params{dbname};
  my $system_name = $params{system_name};
  my $add_if_missing = defined $params{add_if_missing} ? $params{add_if_missing} : 1;

  my ($system_id) = $dbh->selectrow_array("SELECT system_id FROM $dbname.NEGSIG_SYSTEMS WHERE system_name = ? OR system_display_name = ?",
                                          undef, $system_name, $system_name);
  unless ($system_id) {
      if (! $add_if_missing) {
          return;
      }
      my $system_display_name = $system_name;
      my $system_group = '';

      my $schema_attributes = get_schema_attributes();
      my %display_names = %{$schema_attributes->{display_names}};
      if (exists $display_names{system_name} && exists $display_names{system_name}->{$system_name}) {
          $system_display_name = $display_names{system_name}->{$system_name};
      }
      if (exists $schema_attributes->{internal_ids}->{system_name} && 
          exists $schema_attributes->{internal_ids}->{system_name}->{$system_name}) {
          $system_id = $schema_attributes->{internal_ids}->{system_name}->{$system_name};
      }

      if (exists $schema_attributes->{system_groups}->{system_name} && 
          exists $schema_attributes->{system_groups}->{system_name}->{$system_name}) {
          $system_group = find_or_add_system_group(dbh => $dbh,
                                                   dbname => $dbname,
                                                   system_group => $schema_attributes->{system_groups}->{system_name}->{$system_name}
                                                   );
      }

      my %rec = (
                 system_name => $system_name,
                 system_display_name => $system_display_name,
                 flg_is_hidden => 0,
                 system_group => $system_group,
                 rest_uuid => F5::Utils::Rest::generate_uuid($system_name),
                );
      my %last_insert_param;
      if ($system_id) {
          $rec{system_id} = $system_id;
      }
      else {
          $last_insert_param{last_insert_id} = \$system_id;
      }

      F5::DbUtils::insert_row(
                              dbh => $dbh,
                              table_name => "$dbname.NEGSIG_SYSTEMS",
                              rec => \%rec,
                              %last_insert_param,
                             );
  }

  return $system_id;
}

#----------------------------------------------------------------------

sub find_or_add_system_group {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $dbname = $params{dbname};
  my $system_group = $params{system_group};

  my ($system_group_id) = $dbh->selectrow_array("SELECT system_group_id FROM $dbname.NEGSIG_SYSTEM_GROUPS WHERE system_group = ?",
                                          undef, $system_group);
  unless ($system_group_id) {
      my $system_group_display_name = $system_group;
      my $system_group_order = 0;

      my $schema_attributes = get_schema_attributes();
      my %display_names = %{$schema_attributes->{display_names}};
      if (exists $display_names{system_group} && exists $display_names{system_group}->{$system_group}) {
          $system_group_display_name = $display_names{system_group}->{$system_group};
      }
      if (exists $schema_attributes->{internal_ids}->{system_group} &&
          exists $schema_attributes->{internal_ids}->{system_group}->{$system_group}) {
          $system_group_id = $schema_attributes->{internal_ids}->{system_group}->{$system_group};
      }
      if (exists $schema_attributes->{order_ids}->{system_group} &&
          exists $schema_attributes->{order_ids}->{system_group}->{$system_group}) {
          $system_group_order = $schema_attributes->{order_ids}->{system_group}->{$system_group};
      }

      my %rec = (
                 system_group => $system_group,
                 system_group_display_name => $system_group_display_name,
                 system_group_order => $system_group_order,
                 rest_uuid => F5::Utils::Rest::generate_uuid($system_group),
                );
      my %last_insert_param;
      if ($system_group_id) {
          $rec{system_group_id} = $system_group_id;
      }
      else {
          $last_insert_param{last_insert_id} = \$system_group_id;
      }

      F5::DbUtils::insert_row(
                              dbh => $dbh,
                              table_name => "$dbname.NEGSIG_SYSTEM_GROUPS",
                              rec => \%rec,
                              %last_insert_param,
                             );
  }

  return $system_group_id;
}


#----------------------------------------------------------------------

sub find_or_add_attack_type {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $dbname = $params{dbname};
  my $attack_type_name = $params{attack_type_name};
  my $find_only = $params{find_only};

  my ($attack_type_id) = $dbh->selectrow_array("SELECT attack_type_id FROM $dbname.NEGSIG_ATTACK_TYPES WHERE attack_type_name = ?",
                                               undef, $attack_type_name);

  return $attack_type_id if $find_only;

  unless ($attack_type_id) {
    if (! $ENV{ASM_GENERATE_NEGSIG_CSV}) {
        die "Attack type $attack_type_name is not defined and cannot be used\n";
    }
    my $attack_type_display_name = $attack_type_name;

    my $schema_attributes = get_schema_attributes();
    my %display_names = %{$schema_attributes->{display_names}};
    if (exists $display_names{attack_type} && exists $display_names{attack_type}->{$attack_type_name}) {
      $attack_type_display_name = $display_names{attack_type}->{$attack_type_name};
    }
    if (exists $schema_attributes->{internal_ids}->{attack_type} && 
        exists $schema_attributes->{internal_ids}->{attack_type}->{$attack_type_name}) {
      $attack_type_id = $schema_attributes->{internal_ids}->{attack_type}->{$attack_type_name};
    }

    if ($attack_type_id) {
      $dbh->do("INSERT INTO $dbname.NEGSIG_ATTACK_TYPES (attack_type_id, attack_type_name, attack_type_display_name) VALUES (?,?,?)",
               undef, $attack_type_id, $attack_type_name, $attack_type_display_name);
    } else {
        $dbh->do("INSERT INTO $dbname.NEGSIG_ATTACK_TYPES (attack_type_name, attack_type_display_name) VALUES (?,?)",
                 undef, $attack_type_name, $attack_type_display_name);
        ($attack_type_id) = $dbh->last_insert_id(undef, undef, undef, undef);
    }
  }

  return $attack_type_id;
}

#----------------------------------------------------------------------

sub get_schema_attributes {

  my($schema_filename) = @_;

  $schema_filename ||= $NEGSIG_COMMON_SCHEMA_FILENAME;
  eval "use F5::XML::Twig;";
  die "Can't load F5::XML::Twig" if ($@);

  my $schema_attributes = { display_names => {},
                            internal_ids => {},
                            order_ids => {},
                          };

  my $twig = F5::XML::Twig->new(twig_handlers => { 'xsd:enumeration[@asm:displayName]' => sub {
                                                 my($twig, $elm) = @_;

                                                 my $element_name_elm = $elm->parent('xsd:element');
                                                 $schema_attributes->{display_names}->{$element_name_elm->att('name')}->{$elm->att('value')} =
                                                   $elm->att('asm:displayName');
                                               },
                                               'xsd:enumeration[@asm:internalID]' => sub {
                                                 my($twig, $elm) = @_;

                                                 my $element_name_elm = $elm->parent('xsd:element');
                                                 $schema_attributes->{internal_ids}->{$element_name_elm->att('name')}->{$elm->att('value')} =
                                                   $elm->att('asm:internalID');
                                               },
                                               'xsd:enumeration[@asm:system_group]' => sub {
                                                 my($twig, $elm) = @_;

                                                 my $element_name_elm = $elm->parent('xsd:element');
                                                 $schema_attributes->{system_groups}->{$element_name_elm->att('name')}->{$elm->att('value')} =
                                                   $elm->att('asm:system_group');
                                               },
                                               'xsd:enumeration[@asm:orderID]' => sub {
                                                 my($twig, $elm) = @_;

                                                 my $element_name_elm = $elm->parent('xsd:element');
                                                 $schema_attributes->{order_ids}->{$element_name_elm->att('name')}->{$elm->att('value')} =
                                                   $elm->att('asm:orderID');
                                               },

                                             },
                            map_xmlns => {
                                          'http://www.f5.com/asm' => 'asm',
                                          $XML_SCHEMA_URI => 'xsd',
                                         },
                           );

  $twig->safe_parsefile($schema_filename);

  return $schema_attributes;
}

#----------------------------------------------------------------------

sub memoize_only_without_arguments {
    if (@_) {
        return $get_schema_attributes_unique_count++;
    }
    return 1;
}

#----------------------------------------------------------------------

sub create_and_assign_system_signatures_set {
    my(%params) = @_;

    my $asm_config_handler = $params{asm_config_handler};
    my $session = $params{session};
    my $call_flags = $params{call_flags};
    my $policy = $params{policy};
    my $get_messages_only = $params{get_messages_only};
    my $system_name = $params{system_name};

    my $system_id = find_or_add_system(dbh => $asm_config_handler->{dbh},
                                       system_name => $system_name,
                                       dbname => 'PLC');
    my $sig_set_system = F5::ASMConfig::SignatureSetSystem->new( system_id => $system_id);
    my @systems = ($sig_set_system);
    my %sigset_params = (systems => \@systems);
    my $base_set_name = "$system_name Signatures";

    my ($sig_set, $set_name) = get_signature_set(%params, base_set_name => $base_set_name, sigset_params => \%sigset_params);

    my ($set_id, @rv);
    if (!$get_messages_only) {
      if($sig_set) {
          if ( $sig_set->{return_value} && $sig_set->{return_value}->{set_id} ) {
              $set_id = $sig_set->{return_value}->{set_id};
          }
          else {
              $set_id = $sig_set->{set_id};
          }
      }
    }
    else {
        if (!$sig_set) {
            my $msg = "Signature Set '$set_name' for system '$system_name' will be created and assigned to the policy.";
            push( @rv, $msg );
        }
        else {
            $set_id = $sig_set->{set_id};
        }
    }

   if ($set_id) {
        my $policy_sig_set = F5::ASMConfig::Entity::PolicySignatureSet->new(get_criteria => {set_id => $set_id},
                                                                            dbh => $asm_config_handler->{dbh},
                                                                            policy => $policy );
        if (!$policy_sig_set) {
           if (!$get_messages_only) {
               $asm_config_handler->add_policy_signature_set(
                                           $asm_config_handler->{session},
                                           $call_flags,
                                           $policy,
                                           F5::ASMConfig::PolicySignatureSet->new({set_id => $set_id, flg_alarm => 1, flg_reject => 1, flg_learn => 1}));
          }
          else {
              my $msg = "Signature Set '$sig_set->{set_name}' will be assigned to the policy.";
              push (@rv, $msg )
          }
       }
   }
   return { return_value => \@rv } ;
}

#----------------------------------------------------------------------

sub get_signature_set {
    my(%params) = @_;

    my $asm_config_handler = $params{asm_config_handler};
    my $call_flags = $params{call_flags};
    my $policy = $params{policy};
    my $get_messages_only = $params{get_messages_only};
    my $base_set_name = $params{base_set_name};
    my $sigset_params = $params{sigset_params};

    my $policy_negsig_attributes = $asm_config_handler->get_policy_negsig_attributes(
                                                        $asm_config_handler->{session},
                                                        $call_flags,
                                                        $policy,
                                                        F5::ASMConfig::PolicyNegsigAttributes->new({}));

    my $set_name_suffix = '';
    my $negsig_accuracy_level = $policy_negsig_attributes->{min_accuracy_for_auto_added_signatures};
    if ($negsig_accuracy_level == F5::ASMConfig::SignatureAccuracy::SIGNATURE_ACCURACY_HIGH()) {
        $set_name_suffix = ' (High Accuracy)';
    }
    elsif ($negsig_accuracy_level == F5::ASMConfig::SignatureAccuracy::SIGNATURE_ACCURACY_MEDIUM()) {
        $set_name_suffix = ' (High/Medium Accuracy)';
    }

    my $set_name = $base_set_name. $set_name_suffix;
    $sigset_params->{set_name} = $set_name;
    my $sig_set = F5::ASMConfig::Entity::SignatureSet->new(get_criteria => {set_name => $set_name},
                                                           dbh => $asm_config_handler->{dbh});

    if (!$get_messages_only) {
      if (! $sig_set) {
          if (! $sigset_params) {
              my $base_sigset = F5::ASMConfig::Entity::SignatureSet->new(get_criteria => {set_name => $base_set_name},
                                                                         dbh => $asm_config_handler->{dbh});
              $base_sigset->{set_name} = $set_name;
              map {delete $base_sigset->{$_}} qw(flg_is_user_defined set_id rest_uuid signature_set_filter);
              $sigset_params = $base_sigset;
          }

          if ($negsig_accuracy_level > F5::ASMConfig::SignatureAccuracy::SIGNATURE_ACCURACY_LOW()) {
              $sigset_params->{accuracy_op} = "ge";
              $sigset_params->{accuracy_val} = $negsig_accuracy_level;
          }
          $sig_set = $asm_config_handler->import_signature_set(
                         $asm_config_handler->{session},
                         $call_flags,
                         F5::ASMConfig::SignatureSet->new( %$sigset_params ));
      }
    }

    return ($sig_set, $set_name);
}

#----------------------------------------------------------------------

1;
