package F5::Policy::Overlay;
use strict;
use warnings;

use F5::BigipUtils;
use F5::Deploy;
use F5::Logger;
use F5::ObjUtils;

use YAML::XS;
use JSON::PP;
use Scalar::Util qw(reftype);

my $overlays = {};
my %common_overlays = ();

my $POLICY_OVERLAY = F5::Deploy::map_path('/usr/share/ts/config/declarative_policy_overlay.yaml');

#----------------------------------------------------------------------

sub new {
    my ($class, %params) = @_;
    my $self = {};
    bless $self, $class;
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my ($self, %params) = @_;
    $self->{logger} = $params{logger} || F5::Logger->get_logger(logger_name => 'policy_overlay');
    $self->{msgs} = [];
    $self->{seen_warnings} = {};
    my $overlay_file = $params{overlay_file} || $POLICY_OVERLAY;
    if ($overlays->{$overlay_file}) {
        $self->{overlay} = $overlays->{$overlay_file};
    }
    elsif (-f $overlay_file) {
        local $YAML::XS::Boolean = "JSON::PP";
        $self->{overlay} = $overlays->{$overlay_file} = YAML::XS::LoadFile($overlay_file)->{policy};
        if (!F5::BigipUtils::is_feature_enabled('allow_staging')) {
            $common_overlays{performStaging} = JSON::PP::false;
            $common_overlays{learn} = JSON::PP::false;
        }
    }
    else {
        $self->{logger}->debug("No overlay defined") if $self->{logger};
    }
}

#----------------------------------------------------------------------

sub get_overlay {
    my ($self, %params) = @_;
    my $overlay = $params{overlay} || $self->{overlay} || {};
    if ($params{path}) {
        return exists ($overlay->{$params{path}}) ? $overlay->{$params{path}} : {} ;
    }
    return $overlay;
}

#----------------------------------------------------------------------

sub is_supported {
    my ($self, %params) = @_;
    my $overlay = $self->get_overlay(%params);
    if ($overlay && $overlay->{__schema_ignore}) {
        return 0;
    }
    else {
        return 1;
    }
}

#----------------------------------------------------------------------

sub apply_overlay {
    my ($self, %params) = @_;
    my $declarative_policy = $params{declarative_policy};
    my $overlay = $params{overlay} || $self->{overlay};
    my $base_namespace = $params{namespace} || "";

    if (!$overlay) {
         return { policy_element => $declarative_policy };
    }

    if ($overlay->{__schema_overlay} && $overlay->{__schema_overlay}->{required_fields}) {
        my $schema_overlay = $overlay->{__schema_overlay};
        for my $required_field (@{$schema_overlay->{required_fields}}) {
            if ( ! exists($declarative_policy->{$required_field}) ) {
                ## Invalid value
                my $warning_message_field = "'$base_namespace/$required_field'";
                if ($schema_overlay->{warning_message_field} && exists($declarative_policy->{$schema_overlay->{warning_message_field}}) ) {
                    $warning_message_field .= " for $schema_overlay->{warning_message_field} '$declarative_policy->{$schema_overlay->{warning_message_field}}'";
                }
                if ($schema_overlay->{skip_when_invalid}) {
                    $self->add_warning(msg => "$warning_message_field is unsupported.");
                    return {skip => 1};
                }
                elsif ($schema_overlay->{value_when_invalid}) {
                    my $overlay_value = $schema_overlay->{value_when_invalid};
                    $self->add_warning(msg => "$warning_message_field was empty, but must be '$overlay_value'.");
                    $declarative_policy->{$required_field} = $overlay_value;
                }
            }
        }
    }

    for my $entry (keys %$declarative_policy) {
        my $namespace = "$base_namespace/$entry";
        my $policy_reftype = reftype($declarative_policy->{$entry});
        my $overlay_exists = exists($overlay->{$entry});
        my $overlay_pointer = $overlay_exists ? $overlay->{$entry} : {};
        my $overlay_reftype = reftype($overlay_pointer);

        #Skip entirely
        if ($overlay_exists && $overlay_reftype && $overlay_reftype eq 'HASH' && exists($overlay_pointer->{__schema_ignore}) ) {
            my $msg = "Element '$namespace' is unsupported.";
            my $export_overlay = ($params{is_in_declarative_export} && exists $overlay_pointer->{__export_overlay}) ? $overlay_pointer->{__export_overlay} : undef;
            if ($export_overlay) {
                my $action = $export_overlay->{action};
                if ($action eq 'ignore' ||
                    ($action eq 'ignore_if_empty' && F5::ObjUtils::is_empty($declarative_policy->{$entry}) ) ) {
                    undef $msg;
                }
                elsif ($action eq 'ignore_on_condition') {
                    #Currently only valid for hash objects
                    my %condition = %{$export_overlay->{condition}};
                    my $all_match = 1;
                    for my $field (keys %condition) {
                        if (defined($declarative_policy->{$entry}->{$field}) && $condition{$field} ne $declarative_policy->{$entry}->{$field}) {
                            $all_match = 0;
                            last;
                        }
                    }
                    if ($all_match) {
                        undef $msg;
                    }
                }
            }
            if ($msg) {
                if ($export_overlay && $export_overlay->{action} eq 'warn') {
                    $msg = $export_overlay->{warning_text};
                }
                $self->add_warning(msg => $msg);
            }
            delete $declarative_policy->{$entry};
            next;
        }

        #Handle policy collection, recurse deeper if there are more overlays to apply or if there are common overlays.
        if ($policy_reftype && $policy_reftype eq 'ARRAY' &&
            ($overlay_exists || %common_overlays) ) {
            #This case must exclude an array of scalars like allowedResponseCodes
            my $collection_reftype = reftype($declarative_policy->{$entry}->[0]);
            if ($collection_reftype && $collection_reftype eq 'HASH') {
                my @collection;
                for my $policy_element (@{$declarative_policy->{$entry}}) {
                    my $rv = $self->apply_overlay(%params,
                                                  overlay => $overlay_pointer,
                                                  namespace => $namespace,
                                                  declarative_policy => $policy_element);
                    if (!$rv->{skip}) {
                        push (@collection, $rv->{policy_element});
                    }
                }
                $declarative_policy->{$entry} = \@collection;
            }
        }
        elsif ($overlay_exists) {
            #Scalar value
            if (!$overlay_reftype || $overlay_reftype eq 'SCALAR') {
                if (!defined ($declarative_policy->{$entry}) || !defined($overlay_pointer) || "$declarative_policy->{$entry}" ne "$overlay_pointer") {
                    my $former_value = defined($declarative_policy->{$entry}) ? $declarative_policy->{$entry} : '';
                    my $overlay_value = defined($overlay_pointer) ? $overlay_pointer : '';
                    $self->add_warning(msg => "$namespace must be '$overlay_value' (was '$former_value').");
                }
                delete $declarative_policy->{$entry}; #static values are the default, so they can be safely removed.
            }
            elsif ($overlay_reftype eq 'HASH') {
                if (my $schema_overlay = $overlay_pointer->{__schema_overlay} ) {
                    if ($schema_overlay->{allowed_values} || $schema_overlay->{excluded_values}) {
                        if ($schema_overlay->{allowed_values} && !$schema_overlay->{allowed_values_hash}) {
                            $schema_overlay->{allowed_values_hash} = { map { ($_ => 1) } @{$schema_overlay->{allowed_values}} };
                        }
                        if ($schema_overlay->{excluded_values} && !$schema_overlay->{excluded_values_hash}) {
                            $schema_overlay->{excluded_values_hash} = { map { ($_ => 1) } @{$schema_overlay->{excluded_values}} };
                        }
                        if (($schema_overlay->{allowed_values_hash} && !$schema_overlay->{allowed_values_hash}->{$declarative_policy->{$entry}}) ||
                            ($schema_overlay->{excluded_values_hash} && $schema_overlay->{excluded_values_hash}->{$declarative_policy->{$entry}})) {
                            ## Invalid value
                            if ($schema_overlay->{ignore_when_invalid}) {
                                $self->add_warning(msg => "$namespace value '$declarative_policy->{$entry}' is invalid.");
                                delete $declarative_policy->{$entry};
                                next;
                            }
                            elsif ($schema_overlay->{skip_when_invalid}) {
                                $self->add_warning(msg => "$namespace value '$declarative_policy->{$entry}' is unsupported.");
                                return {skip => 1};
                            }
                            elsif ($schema_overlay->{value_when_invalid}) {
                                my $former_value = defined($declarative_policy->{$entry}) ? $declarative_policy->{$entry} : '';
                                my $overlay_value = $schema_overlay->{value_when_invalid};
                                $self->add_warning(msg => "$namespace must be '$overlay_value' (was '$former_value').");
                                $declarative_policy->{$entry} = $overlay_value;
                            }
                        }
                    }
                }

                if ($policy_reftype && $policy_reftype eq 'HASH') {
                    my $rv = $self->apply_overlay(%params,
                                                  overlay => $overlay_pointer,
                                                  namespace => $namespace,
                                                  declarative_policy => $declarative_policy->{$entry});
                    if ($rv->{skip}) {
                        delete $declarative_policy->{$entry};
                        return {skip => 1};
                    }
                    $declarative_policy->{$entry} = $rv->{policy_element};
                }
            }
            elsif ($overlay_reftype eq 'ARRAY') {
                die "Found invalid reftype in policy overlay at '$namespace'\n";
            }
        }

        if (exists $common_overlays{$entry} && exists $declarative_policy->{$entry}) {
            if ("$common_overlays{$entry}" ne "$declarative_policy->{$entry}") {
                my $msg = $params{is_in_declarative_export} ? "$namespace value $declarative_policy->{$entry} is unsupported"
                                                            : "$namespace must be set to '$common_overlays{$entry}'";
                $self->add_warning(msg => $msg);
            }
            delete $declarative_policy->{$entry};
        }
    }
    return {policy_element => $declarative_policy};
}

#----------------------------------------------------------------------

sub add_warning {
  my ($self, %params) = @_;

  my $user_msg = $params{user_msg} || $params{msg};
  my $log_msg = $params{log_msg} || $params{msg};

  my $log_level = (defined $params{log_level}) ? $params{log_level} : "info";

  #Only add unique warnings.
  return if $self->{seen_warnings}->{$user_msg}++;

  push(@{$self->{msgs}}, {
                          msg_level => 'warning',
                          msg_text => $user_msg,
                          }
      );

  $self->{logger}->$log_level($log_msg) if $self->{logger};
}

#----------------------------------------------------------------------

sub clear_msgs {
    my ($self) = @);
    $self->{msgs} = [];
}

#----------------------------------------------------------------------

sub get_msgs {
  my ($self) = @_;
  return $self->{msgs};
}

#----------------------------------------------------------------------

1;
