package F5::PrepareConf::Hsl;

use strict;
use warnings;

use base 'F5::PrepareConf::Base';

use DBI;

# tsdev includes

use F5::DbUtils;
use F5::IpUtils;
use F5::Hsl;
use F5::PrepareConf::Base qw(ERRMSG_DBH_LOST RET_OK RET_NOT_OK FALSE TRUE);

#defaults

use constant HSL_DELETE_PROFILE => 'delete';

sub prepare {
    my ($self, %params) = @_;

    my $logger = $params{logger} || $self->{logger};
    my $hsl_profile_id = $params{hsl_profile_id};
    my $hsl_profile_options = $params{hsl_profile_options};
    my $dbh = $self->{dbh};
     

    my $sql = "SELECT hsl_profile_name, hsl_profile_type_id FROM PLC.HSL_PROFILES WHERE hsl_profile_id=?";
    my ($hsl_profile_name, $hsl_profile_type_id) = $dbh->selectrow_array($sql, undef, $hsl_profile_id);

    if(!$hsl_profile_type_id || ( (defined $hsl_profile_options) && ($hsl_profile_options eq HSL_DELETE_PROFILE) ))
    {
        $self->empty_table("DCC.HSL_PROFILES", undef, {"DCC.HSL_PROFILES.hsl_profile_id" => $hsl_profile_id});
        return RET_OK;
    }

    # Calculate a new alarm mask
    $sql  = "SELECT BIT_OR(A.violation_code)  FROM PLC.HSL_PROFILE_VIOLATION_CONFIG A, PLC.HSL_VIOLATION_CODES B WHERE A.flg_alarm=? AND A.hsl_profile_id=? AND A.violation_code=B.violation_code";
    my ($alarm_mask) = $dbh->selectrow_array($sql, undef, 1, $hsl_profile_id);

    # Calculate a new reject mask
    $sql  = "SELECT BIT_OR(A.violation_code)  FROM PLC.HSL_PROFILE_VIOLATION_CONFIG A, PLC.HSL_VIOLATION_CODES B WHERE A.flg_reject=? AND A.hsl_profile_id=? AND A.violation_code=B.violation_code";
    my ($reject_mask) = $dbh->selectrow_array($sql, undef, 1, $hsl_profile_id);

    # Calculate a new allowed command mask
    $sql = "SELECT BIT_OR(A.command_code) FROM PLC.HSL_PROFILE_COMMAND_CONFIG A, PLC.HSL_COMMAND_CODES B WHERE  A.flg_is_enabled=? AND A.hsl_profile_id=? AND A.command_code=B.command_code";
    my ($allowed_command_mask) = $dbh->selectrow_array($sql, undef, 1, $hsl_profile_id);

    # Copy data from PLC to DCC, insert new profile or update the old one
    local $" = ','; # so that lists inside double-quotes will be comma-separated
    my @fields_tbl_hsl_profiles = qw(hsl_profile_id hsl_profile_name hsl_profile_type_id reject_mask alarm_mask allowed_command_mask);

    my $ignore = F5::DbUtils::get_ignore_string($dbh);

    my @placeholders = ('?') x @fields_tbl_hsl_profiles;
    my $sql_insert = "INSERT $ignore INTO DCC.HSL_PROFILES (@fields_tbl_hsl_profiles) VALUES (@placeholders)";
    my $sql_update = "UPDATE DCC.HSL_PROFILES SET hsl_profile_name=?,reject_mask=?,alarm_mask=?,allowed_command_mask=? WHERE hsl_profile_id=$hsl_profile_id";
    my @bind_values = ($hsl_profile_id, $hsl_profile_name, $hsl_profile_type_id, $reject_mask, $alarm_mask, $allowed_command_mask);
    my @bind_values_update = ( $hsl_profile_name, $reject_mask, $alarm_mask, $allowed_command_mask);

    my $sth = $dbh->prepare_cached($sql_insert) or $self->{logger}->error(qq{$DBI::errstr on "$sql_insert"});
    $sth->execute(@bind_values)
        or $self->{logger}->error(qq{$DBI::errstr on "$sql_insert" values=(@bind_values)});
    $sth = $dbh->prepare_cached($sql_update) or $self->{logger}->error(qq{$DBI::errstr on "$sql_update"});
    $sth->execute(@bind_values_update)
        or $self->{logger}->error(qq{$DBI::errstr on "$sql_update" values=(@bind_values_update)});
   
    # handle specific profile configuration 

    if ( $hsl_profile_type_id == F5::Hsl::SMTP)
    {
      # Copy SMTP configuration from from PLC to DCC
      # 1. Empty DCC.HSL_PROFILES_SMTP_CONFIG table
      # 2. Copy  PLC.HSL_PROFILES_SMTP_CONFIG content to DCC.HSL_PROFILES_SMTP_CONFIG table

      $self->empty_table("DCC.HSL_PROFILES_SMTP_CONFIG", undef, {"DCC.HSL_PROFILES_SMTP_CONFIG.hsl_profile_id" => $hsl_profile_id});
      $sql = "INSERT $ignore INTO DCC.HSL_PROFILES_SMTP_CONFIG SELECT * FROM PLC.HSL_PROFILES_SMTP_CONFIG WHERE hsl_profile_id=$hsl_profile_id";
      my $sth = $dbh->prepare_cached($sql) or $self->{logger}->error(qq{$DBI::errstr on "$sql"});
      $sth->execute() or $self->{logger}->error(qq{$DBI::errstr on "$sql"});     
    
      # Copy SMTP address configuration
      # 1. Empty DCC.HSL_PROFILES_SMTP_ADDRESS_CONFIG 
      # 2. Copy  PLC.HSL_PROFILES_SMTP_ADDRESS_CONFIG content to DCC.HSL_PROFILES_SMTP_ADDRESS_CONFIG table

      $self->empty_table("DCC.HSL_PROFILES_SMTP_ADDRESS_CONFIG", undef, {"DCC.HSL_PROFILES_SMTP_ADDRESS_CONFIG.hsl_profile_id" => $hsl_profile_id});

       my @plc_fields = qw(hsl_profile_id address_type address);
       my $plc_sth = $dbh->prepare("SELECT @plc_fields from  PLC.HSL_PROFILES_SMTP_ADDRESS_CONFIG WHERE hsl_profile_id= ?");
       $plc_sth->execute($hsl_profile_id);
         while(my $plc_row = $plc_sth->fetchrow_hashref) {
            my $ip_address_or_domain =  $plc_row->{address};
            my $route_domain = 0;
            if ( $ip_address_or_domain =~ /^(.*)\%([0-9]+)$/g ){ # We want to validate and split adress field only in case it is with route domain
                    my $split_ip_route = F5::IpUtils::split_ip_route( ip_address => $plc_row->{address} , die_on_error => 0 );
                    if (! $split_ip_route->{completed_successfully}) {
                        $logger->info("Could not split [$plc_row->{address}] into ip and route") if $logger;
                        next;
                    }
                    $ip_address_or_domain = $split_ip_route->{ip_address};
                    $route_domain = $split_ip_route->{route_domain};
            }

           $self->insert_row(
                      table_name => 'DCC.HSL_PROFILES_SMTP_ADDRESS_CONFIG',
                      ignore => 1,
                      rec => { hsl_profile_id => $plc_row->{hsl_profile_id} ,
                               address_type =>  $plc_row->{address_type},
                               address => $ip_address_or_domain,
                               route_domain => $route_domain,
                             });
         } 


    } elsif (  $hsl_profile_type_id == F5::Hsl::FTP){
       # Copy FTP configuration from from PLC to DCC
       # 1. Empty DCC.HSL_PROFILES_FTP_CONFIG table
       # 2. Copy  PLC.HSL_PROFILES_FTP_CONFIG content to DCC.HSL_PROFILES_FTP_CONFIG table

       $self->empty_table("DCC.HSL_PROFILES_FTP_CONFIG", undef, {"DCC.HSL_PROFILES_FTP_CONFIG.hsl_profile_id" => $hsl_profile_id});
       $sql = "INSERT $ignore INTO DCC.HSL_PROFILES_FTP_CONFIG SELECT * FROM PLC.HSL_PROFILES_FTP_CONFIG WHERE hsl_profile_id=$hsl_profile_id";
       my $sth = $dbh->prepare_cached($sql) or error(qq{$DBI::errstr on "$sql"});
       $sth->execute() or error(qq{$DBI::errstr on "$sql"});
     } 
    
    return RET_OK;
}

1;
