package F5::PrepareConf::LoggingProfile;

use strict;
use warnings;

use base 'F5::PrepareConf::Base';
use F5::Mcp::FwLoggerType;

use DBI;
use Memoize;
use JSON;
use Bit::Vector;

# tsdev includes
use F5::CRC;
use F5::IpUtils;
use F5::ASM::COAPI;
use F5::ServerLookup;
use F5::Mcp::FwLoggerType;
use F5::Mcp::FwLogRemoteProtocol;
use F5::Mcp::FwLogFilterKey;
use F5::Mcp::SyslogFacility;
use F5::ASMConfig::Entity::RequestLogger;
use F5::BigipUtils; 

# NOTE: these constants should be equivalent to MCP enum fw_log_filter_key_t
use constant {
    LOG_FILTER_KEY_REQUEST_TYPE => 10,
    LOG_FILTER_KEY_PROTOCOL => 11,
    LOG_FILTER_KEY_RESPONSE_CODE => 12,
    LOG_FILTER_KEY_HTTP_METHOD => 13,
    LOG_FILTER_KEY_LOGIN_RESULT => 14,
    LOG_FILTER_KEY_SEARCH_ALL => 0,
    LOG_FILTER_KEY_SEARCH_IN_HEADERS => 6,
    LOG_FILTER_KEY_SEARCH_IN_POST_DATA => 5,
    LOG_FILTER_KEY_SEARCH_IN_QUERY_STRING => 4,
    LOG_FILTER_KEY_SEARCH_IN_REQUEST => 1,
    LOG_FILTER_KEY_SEARCH_IN_URI => 3,
    LOG_FILTER_KEY_LOG_CHALLENGE_FAILURE_REQUESTS => 15,
};

# Log destination pool uses different enum log_pool_protocol_t
use constant {
    LOG_POOL_PROTOCOL_TCP => 0,
    LOG_POOL_PROTOCOL_UDP => 1,
    LOG_REMOTE_PROTOCOL_UDP => 0,
    LOG_REMOTE_PROTOCOL_TCP => 1,
};

use constant LOGGER_FILTER_ALL                                       => 0;
use constant LOGGER_REQUEST_TYPE_ILLEGAL                             => 1;
use constant LOGGER_REQUEST_TYPE_ILLEGAL_INCLUDING_STAGED_SIGNATURES => 2;
use constant LOGGER_REQUEST_TYPE_BLOCKED                             => 3;
use constant LOGGER_PROTOCOL_HTTP                                    => 1;
use constant LOGGER_PROTOCOL_HTTPS                                   => 2;
use constant LOGGER_PROTOCOL_WS                                      => 4;
use constant LOGGER_PROTOCOL_WSS                                     => 8;
use constant LOGGER_ENTITY_ASM                                       => 1;
use constant LOGGER_ENTITY_HSL                                       => 4;

# NOTE: these constants should be equivalent to remote_log_dest_type_t
use constant {
    DESTINATION_TYPE_IP_ADDRESS => 1,
    DESTINATION_TYPE_HOSTNAME => 2,
    DESTINATION_TYPE_FILE => 3,
};

my $DEFAULT_STORAGE_FORMAT = "";
if (defined(F5::BigipUtils::get_default_log_format_fields())) {
    $DEFAULT_STORAGE_FORMAT = join(",", map { /%/ ? $_ : qq{$_="%$_%"} } @{F5::BigipUtils::get_default_log_format_fields()});
}

my $GRPC_STORAGE_FORMAT = "";
if (defined(F5::BigipUtils::get_grpc_log_format_fields())) {
    $GRPC_STORAGE_FORMAT = join(",", map { /%/ ? $_ : qq{$_="%$_%"} } @{F5::BigipUtils::get_grpc_log_format_fields()});
}

my $ARCSIGHT_STORAGE_FORMAT = join(' ', qw(CEF:0|F5|ASM|%s|%s|%s|%d|dvchost=%s
                                           dvc=%s
                                           cs1=%s cs1Label=policy_name
                                           cs2=%s cs2Label=http_class_name
                                           deviceCustomDate1=%s deviceCustomDate1Label=policy_apply_date
                                           externalId=%s
                                           act=%s
                                           cn1=%d cn1Label=response_code
                                           src=%s
                                           spt=%d
                                           dst=%s
                                           dpt=%d
                                           requestMethod=%s
                                           app=%s
                                           cs5=%s cs5Label=x_forwarded_for_header_value
                                           rt=%s
                                           deviceExternalId=%d
                                           cs4=%s cs4Label=attack_type
                                           cs6=%s cs6Label=geo_location
                                           c6a1=%s c6a1Label=device_address
                                           c6a2=%s c6a2Label=source_address
                                           c6a3=%s c6a3Label=destination_address
                                           c6a4=%s c6a4Label=ip_address_intelligence
                                           msg=%s
                                           suid=%s
                                           suser=%s
                                           cn2=%d cn2Label=violation_rating
                                           cn3=%llu cn3Label=device_id
                                           microservice=%s
                                           request=%s%.*s
                                           cs3Label=full_request cs3=%s
                                          ));

use constant ARCSIGHT_SEARCH_FOR            => qq(=);
use constant ARCSIGHT_REPLACE_WITH          => qq(\\=);
use constant SPLUNK_SEARCH_FOR            => qq(");
use constant SPLUNK_REPLACE_WITH          => qq(%22);
use constant BIGIQ_SEARCH_FOR            => qq(");
use constant BIGIQ_REPLACE_WITH          => qq(%22);

my $SPLUNK_FORMAT_STRING = join(',', qw(unit_hostname="%unit_hostname%"
                                        management_ip_address="%management_ip_address%"
                                        management_ip_address_2="%management_ip_address_2%"
                                        http_class_name="%http_class_name%"
                                        web_application_name="%http_class_name%"
                                        policy_name="%policy_name%"
                                        policy_apply_date="%policy_apply_date%"
                                        violations="%violations%"
                                        support_id="%support_id%"
                                        request_status="%request_status%"
                                        response_code="%response_code%"
                                        ip_client="%ip_client%"
                                        route_domain="%route_domain%"
                                        method="%method%"
                                        protocol="%protocol%"
                                        query_string="%query_string%"
                                        x_forwarded_for_header_value="%x_forwarded_for_header_value%"
                                        sig_ids="%sig_ids%"
                                        sig_names="%sig_names%"
                                        date_time="%date_time%"
                                        severity="%severity%"
                                        attack_type="%attack_type%"
                                        geo_location="%geo_location%"
                                        ip_address_intelligence="%ip_address_intelligence%"
                                        username="%username%"
                                        session_id="%session_id%"
                                        src_port="%src_port%"
                                        dest_port="%dest_port%"
                                        dest_ip="%dest_ip%"
                                        sub_violations="%sub_violations%"
                                        virus_name="%virus_name%"
                                        violation_rating="%violation_rating%"
                                        websocket_direction="%websocket_direction%"
                                        websocket_message_type="%websocket_message_type%"
                                        device_id="%device_id%"
                                        staged_sig_ids="%staged_sig_ids%"
                                        staged_sig_names="%staged_sig_names%"
                                        threat_campaign_names="%threat_campaign_names%"
                                        staged_threat_campaign_names="%staged_threat_campaign_names%"
                                        blocking_exception_reason="%blocking_exception_reason%"
                                        captcha_result="%captcha_result%"
                                        microservice="%microservice%"
                                        tap_event_id="%tap_event_id%"
                                        tap_vid="%tap_vid%"
                                        vs_name="%vs_name%"
                                        sig_cves="%sig_cves%"
                                        staged_sig_cves="%staged_sig_cves%"
                                        uri="%uri%"
                                        fragment="%fragment%"
                                        request="%request%"
                                        response="%response%"
                                       ));

my $BIGIQ_FORMAT_STRING = join(',', qw(unit_hostname="%unit_hostname%"
                                        management_ip_address="%management_ip_address%"
                                        management_ip_address_2="%management_ip_address_2%"
                                        http_class_name="%http_class_name%"
                                        web_application_name="%http_class_name%"
                                        policy_name="%policy_name%"
                                        policy_apply_date="%policy_apply_date%"
                                        violations="%violations%"
                                        support_id="%support_id%"
                                        request_status="%request_status%"
                                        response_code="%response_code%"
                                        ip_client="%ip_client%"
                                        route_domain="%route_domain%"
                                        method="%method%"
                                        protocol="%protocol%"
                                        query_string="%query_string%"
                                        x_forwarded_for_header_value="%x_forwarded_for_header_value%"
                                        sig_ids="%sig_ids%"
                                        sig_names="%sig_names%"
                                        date_time="%date_time%"
                                        severity="%severity%"
                                        attack_type="%attack_type%"
                                        geo_location="%geo_location%"
                                        ip_address_intelligence="%ip_address_intelligence%"
                                        username="%username%"
                                        session_id="%session_id%"
                                        src_port="%src_port%"
                                        dest_port="%dest_port%"
                                        dest_ip="%dest_ip%"
                                        sub_violations="%sub_violations%"
                                        virus_name="%virus_name%"
                                        microservice="%microservice%"
                                        tap_event_id="%tap_event_id%"
                                        tap_vid="%tap_vid%"
                                        uri="%uri%"
                                        violation_details="%violation_details%"
                                        violation_rating="%violation_rating%"
                                        websocket_direction="%websocket_direction%"
                                        websocket_message_type="%websocket_message_type%"
                                        compression_method="%compression_method%"
                                        device_id="%device_id%"
                                        staged_sig_ids="%staged_sig_ids%"
                                        staged_sig_names="%staged_sig_names%"
                                        threat_campaign_names="%threat_campaign_names%"
                                        staged_threat_campaign_names="%staged_threat_campaign_names%"
                                        blocking_exception_reason="%blocking_exception_reason%"
                                        mobile_application_name="%mobile_application_name%"
                                        mobile_application_version="%mobile_application_version%"
                                        client_type="%client_type%"
                                        captcha_result="%captcha_result%"
                                        header="%headers%"
                                        fragment="%fragment%"
                                        sig_cves="%sig_cves%"
                                        staged_sig_cves="%staged_sig_cves%"
                                        request="%request%"
                                        response="%response%"
                                       ));

our $LOGGING_TABLES = [ qw(DCC.LOG_LOCAL_STORAGE_CONFIGURATION
                          DCC.REMOTE_STORAGE_CONFIGURATION
                          DCC.REMOTE_STORAGE_FORMAT_ESCAPING
                          DCC.LOG_APPIQ_CONFIG
                          DCC.REMOTE_FILTER
                          DCC.LOCAL_FILTER
                          DCC.LOG_APPIQ_FILTER)
                     ];

our $MAPPING_TABLES = [ qw(DCC.VS_LOG_PROFILE) ];

my $BIGINT_WIDTH = 65; # Don't use the zeroth bit
my $DEFAULT_PORT = 514;
my $K_BYTES = 1024;

memoize('get_storage_fields',
        NORMALIZER => sub { return '' });
memoize('change_storage_format');
memoize('get_bigint_mask');


sub prepare {
    my ($self, %params) = @_;

    my $flg_load_complete = $params{flg_load_complete};
    my $change_made = 0;
    my $seen_vs_name_crcs = {};
    my $seen_log_profile_name_crcs = {};
    my $do_cleanup = defined $params{do_cleanup} ? $params{do_cleanup} : 1;
    my $remove_unseen_vs_names = defined $params{remove_unseen_vs_names} ? $params{remove_unseen_vs_names} : 0;
    my $remove_unseen_log_profile_names = defined $params{remove_unseen_log_profile_names} ? $params{remove_unseen_log_profile_names} : 0;

    $self->{logger}->debug("Preparing logging profiles") if $self->{logger};

    if ($flg_load_complete) {
        $self->empty_table($LOGGING_TABLES);
        $self->empty_table($MAPPING_TABLES);
        $change_made = 1;
    }

    if ($flg_load_complete || exists $params{vs_names}) {
        my $rv = $self->prepare_from_mcp_update(%params,
                                                seen_vs_name_crcs => $seen_vs_name_crcs,
                                               );
        $change_made ||= $rv;
    }
    if (exists $params{config_set_loggers}) {
        my $rv = $self->prepare_config_set_loggers(%params,
                                                   seen_vs_name_crcs => $seen_vs_name_crcs,
                                                   seen_log_profile_name_crcs => $seen_log_profile_name_crcs,
                                                  );
        $change_made ||= $rv;
        # Loading from Config Set implies remove all other VS names
        $remove_unseen_vs_names = 1;
        $remove_unseen_log_profile_names = 1;
    }

    if ($flg_load_complete || $params{policy}) {
        my $rv = $self->prepare_request_loggers(%params,
                                                seen_vs_name_crcs => $seen_vs_name_crcs,
                                               );
        $change_made ||= $rv;
    }

    if ($do_cleanup) {
        $self->cleanup(seen_vs_name_crcs => $seen_vs_name_crcs,
                       seen_log_profile_name_crcs => $seen_log_profile_name_crcs,
                       remove_unseen_vs_names => $remove_unseen_vs_names,
                       remove_unseen_log_profile_names => $remove_unseen_log_profile_names,
                      );
    }

    if ($change_made) {
        $self->prepare_negsig_enforcer_names();
    }

    $self->{logger}->debug("Finished preparing logging profiles") if $self->{logger};
    return $change_made;
}

#----------------------------------------------------------------------

sub prepare_from_mcp_update {
    my ($self, %params) = @_;

    my $flg_load_complete = $params{flg_load_complete};
    my $vs_names = $params{vs_names};
    my $log_profile_names = $params{profile_names};
    my $publisher_names = $params{publisher_names};
    my $destination_names = $params{destination_names};
    my $dest_pool_names = $params{dest_pool_names};
    my $pool_names = $params{pool_names};

    my $seen_vs_name_crcs = $params{seen_vs_name_crcs};

    my $dbh = $self->{dbh};
    my $change_made = 0;

    $self->{logger}->debug("Preparing MCP logging profiles") if $self->{logger};
    # Now handle:
    #   - VSs that had logging profiles assigned/changed 
    #   - Logging profiles that had configuration changes

    # are there any logging profiles assigned to VSs ? we got nothing to do if not...
    my $vs_log_profiles = F5::ASM::COAPI::Query->new('fw_vs_log_profile');
    if ($vs_log_profiles && @$vs_log_profiles) {
        for my $vs_log_profile (@$vs_log_profiles) {

            # was a relevant VS updated?
            my $vs_name = $vs_log_profile->{virtual_server_name};
            my $vs_name_crc = F5::CRC::get_crc32($vs_name);
            $seen_vs_name_crcs->{$vs_name_crc} = 1;
            my $vs_updated = grep { $_ eq $vs_name } @$vs_names;

            # was a relevant log prof updated?
            my $log_profile_name = $vs_log_profile->{profile_name};
            my $profile_updated = grep { $_ eq $log_profile_name } @$log_profile_names;

            # if not full update and no changes - next
            if (!$flg_load_complete) {
                next unless $vs_updated || $profile_updated;
            }

            # look for our logging profile
            my ($log_application, $log_protocol);
            # is it application security type ?
            $log_application = F5::ASM::COAPI::Query->new(
                'fw_log_application', {profile_name => $log_profile_name}
            );
            # or, is it protocol security type ? (mutually exclusive on a VS)
            if (!@$log_application) {
                $log_protocol = F5::ASM::COAPI::Query->new(
                    'fw_log_protocol_transfer', {profile_name => $log_profile_name}
                );
                next unless @$log_protocol;
            }

            # if a vs was updated - we want to refresh the related profile as well
            $profile_updated = 1 if $vs_updated;

            # if a profile was updated - we want to refresh the related VSs as well
            $vs_updated = 1 if $profile_updated;

            # calculate the required CRCs
            my $log_profile_name_crc = F5::CRC::get_crc32($log_profile_name);

            # clear relevant DCC tables
            if ($vs_updated && !$flg_load_complete) {
                my $rv = $self->empty_table($MAPPING_TABLES, undef, {"vs_name_crc" => $vs_name_crc, "log_profile_name_crc" => $log_profile_name_crc});
                $change_made ||= $rv;
            }
            if ($profile_updated && !$flg_load_complete) {
                my $rv = $self->empty_table($LOGGING_TABLES, undef, {"log_profile_name_crc" => $log_profile_name_crc});
                $change_made ||= $rv;
            }

            # hash of params for use in method calls below
            my %prepare_params = (log_application => $log_application->[0],
                                  log_profile_name => $log_profile_name,
                                  log_profile_name_crc => $log_profile_name_crc);

            # update application security logging profiles only
            if ( ($profile_updated || $flg_load_complete) && $log_application && @$log_application) {
                $prepare_params{server_addresses} = F5::ASM::COAPI::Query->new('fw_log_server_address', {parent_profile => $log_application->[0]->{profile_name}});
                $prepare_params{request_filters} = F5::ASM::COAPI::Query->new('fw_log_request_filter', {parent_profile => $log_application->[0]->{profile_name}});

                $self->prepare_log_application_local_storage_conf(%prepare_params);
                my $server_found = $self->prepare_log_application_remote_storage_conf(%prepare_params);
                $self->prepare_log_application_filter(server_found => $server_found, %prepare_params);
                $change_made = 1;
            }
            # or, update protocol security logging profiles only (mutually exclusive)
            elsif ( $profile_updated || $flg_load_complete ||
                    ($publisher_names && @$publisher_names) || ($dest_pool_names && @$dest_pool_names) ||
                    ($destination_names && @$destination_names) || ($pool_names && @$pool_names) ) {

                if ($log_protocol && @$log_protocol && $log_protocol->[0]->{publisher}) {
                    my $rv = $self->prepare_log_protocol_remote_storage_conf(
                        publisher_name => $log_protocol->[0]->{publisher},
                        log_profile_name => $log_profile_name,
                        log_profile_name_crc => $log_profile_name_crc,
                        publisher_names => $publisher_names,
                        destination_names => $destination_names,
                        dest_pool_names => $dest_pool_names,
                        pool_names => $pool_names,
                        profile_updated => $profile_updated || $flg_load_complete,
                    );
                    $change_made ||= $rv;
                }
            }
 
            # update log prof to VSs mapping
            if (@$log_application && ($vs_updated || $flg_load_complete)) {
               $self->prepare_log_prof_mapping_conf(
                   vs_name_crc => $vs_name_crc,
                   log_profile_name_crc => $log_profile_name_crc,
                   storage_type => $log_application->[0]->{logger_type},
               );
               $change_made = 1;
            }

        } # END OF: for my $vs_log_profile 
    } 
    else {
        # No assigned profiles found - Clearing MAPPING tables...
        my $rv = $self->empty_table($MAPPING_TABLES);
        $change_made ||= $rv;
    }

    return $change_made;
}

#----------------------------------------------------------------------

sub cleanup {
    my($self, %params) = @_;

    $self->{logger}->debug("Cleaning old logging profiles") if $self->{logger};
    my $change_made = 0;
    my $remove_unseen_vs_names = $params{remove_unseen_vs_names};
    my $remove_unseen_log_profile_names = $params{remove_unseen_log_profile_names};
    my $seen_vs_name_crcs = $params{seen_vs_name_crcs} || {};
    my $seen_log_profile_name_crcs = $params{seen_log_profile_name_crcs} || {};

    if ($remove_unseen_vs_names) {
        my $all_vs_name_crcs =  $self->{dbh}->selectall_hashref(<<"EndQuery", 'vs_name_crc');
SELECT DISTINCT vs_name_crc
  FROM DCC.VS_LOG_PROFILE
EndQuery

        my @vs_name_crcs_to_remove;
        for my $vs_name_crc (keys %$all_vs_name_crcs) {
            if (! $seen_vs_name_crcs->{$vs_name_crc}) {
                $self->{logger}->trace("Removing unseen VS (vs_name_crc = $vs_name_crc)") if $self->{logger};
                push @vs_name_crcs_to_remove, $vs_name_crc;
            }
        }
        if (@vs_name_crcs_to_remove) {
            my $rv = $self->empty_table($MAPPING_TABLES, undef, { "vs_name_crc" => \@vs_name_crcs_to_remove });
            $change_made ||= $rv;
        }
    }
    if ($remove_unseen_log_profile_names) {
       my $all_log_profile_name_crcs =  $self->{dbh}->selectall_hashref(<<"EndQuery", 'log_profile_name_crc');
SELECT DISTINCT log_profile_name_crc
  FROM DCC.VS_LOG_PROFILE
EndQuery

        my @log_profile_name_crcs_to_remove;
        for my $log_profile_name_crc (keys %$all_log_profile_name_crcs) {
            if (! $seen_log_profile_name_crcs->{$log_profile_name_crc}) {
                $self->{logger}->trace("Removing unseen log profile (log_profile_name_crc = $log_profile_name_crc)") if $self->{logger};
                push @log_profile_name_crcs_to_remove, $log_profile_name_crc;
            }
        }
        if (@log_profile_name_crcs_to_remove) {
            my $rv = $self->empty_table($MAPPING_TABLES, undef, { "log_profile_name_crc" => \@log_profile_name_crcs_to_remove });
            $change_made ||= $rv;
        }
    }

    my $all_log_profile_name_crcs = $self->{dbh}->selectall_hashref(<<"EndQuery", 'log_profile_name_crc');
SELECT log_profile_name_crc
  FROM DCC.LOG_LOCAL_STORAGE_CONFIGURATION
 UNION
SELECT log_profile_name_crc
  FROM DCC.REMOTE_STORAGE_CONFIGURATION
EndQuery

    my $in_use_log_profile_name_crcs =  $self->{dbh}->selectall_hashref(<<"EndQuery", 'log_profile_name_crc');
SELECT DISTINCT log_profile_name_crc
  FROM DCC.VS_LOG_PROFILE
 UNION
SELECT log_profile_name_crc
  FROM DCC.ACCOUNT_LOG_PROFILE
EndQuery

    my @log_profile_name_crcs_to_remove;
    for my $log_profile_name_crc (keys %$all_log_profile_name_crcs) {
        if (! $in_use_log_profile_name_crcs->{$log_profile_name_crc}) {
            $self->{logger}->trace("Removing unused log profile (log_profile_name_crc = $log_profile_name_crc)") if $self->{logger};
            push @log_profile_name_crcs_to_remove, $log_profile_name_crc;
        }
    }
    if (@log_profile_name_crcs_to_remove) {
        my $rv = $self->empty_table($LOGGING_TABLES, undef, { "log_profile_name_crc" => \@log_profile_name_crcs_to_remove });
        $change_made ||= $rv;
    }

    return $change_made;
}

#----------------------------------------------------------------------

sub prepare_log_prof_mapping_conf {
    my ($self, %params) = @_;
    my $storage_type = defined($params{storage_type}) ? $params{storage_type} : F5::Mcp::FwLoggerType::FW_LOGGER_TYPE_REMOTE();

    $self->{logger}->trace("Adding VS/logging profile association (vs_name_crc = $params{vs_name_crc}, log_profile_name_crc = $params{log_profile_name_crc})") if $self->{logger};
    $self->insert_row(
                      table_name => 'DCC.VS_LOG_PROFILE',
                      do_replace => 1,
                      rec => { 
                               vs_name_crc => $params{vs_name_crc},
                               log_profile_name_crc => $params{log_profile_name_crc},
                               storage_type => $storage_type,
                             }
                     );
}

#----------------------------------------------------------------------

sub prepare_log_application_local_storage_conf {
    my ($self, %params) = @_;

    my $log_application = $params{log_application};
    my $log_profile_name = $params{log_profile_name};
    my $log_profile_name_crc = $params{log_profile_name_crc};

    return if ($log_application->{logger_type} != F5::Mcp::FwLoggerType::FW_LOGGER_TYPE_LOCAL());

    $self->insert_row(
                      table_name => 'DCC.LOG_LOCAL_STORAGE_CONFIGURATION',
                      do_replace => 1,
                      rec => { 
                               log_profile_name => $log_profile_name,
                      	       log_profile_name_crc => $log_profile_name_crc,
                               guaranteed_logging => $log_application->{guarantee_logging},
                               response_logging => $log_application->{response_logging}, # equivalent enums
                               response_guaranteed_logging => $log_application->{guarantee_response_logging}
                             }
                     );
}

#----------------------------------------------------------------------

sub change_storage_format {
    my (%params) = @_;

    my $storage_format_template = $params{storage_format_template};

    my $storage_fields = get_storage_fields();
    my $max_id = List::Util::max(map { $_->{id} } @$storage_fields);
    my $bitmap_width = $max_id > $BIGINT_WIDTH ? $max_id : $BIGINT_WIDTH;

    my @bitmap = ('0') x $bitmap_width;
    my $storage_format_fmt_template = $storage_format_template;
    $storage_format_fmt_template =~ s/([{}])/$1$1/g;
    for my $field (@$storage_fields) {
        if ($storage_format_template =~ s/\%\Q$field->{name}\E\%/\%$field->{id}\$\Q$field->{format}\E/g) {
             $storage_format_fmt_template =~ s/\%\Q$field->{name}\E\%/\{$field->{id}\}/g;
            # This field occurred at least once -- add to the bitmap
            $bitmap[$field->{id}] = 1;
        }
    }

    return { storage_format => $storage_format_template,
             storage_format_fmt => $storage_format_fmt_template,
             bitmap => \@bitmap };
}

#----------------------------------------------------------------------

sub prepare_log_application_remote_storage_conf {
    my ($self, %params) = @_;

    my $logger = $params{logger} || $self->{logger};
    my $log_application = $params{log_application};
    my $log_profile_name = $params{log_profile_name};
    my $log_profile_name_crc = $params{log_profile_name_crc};
    my $server_addresses = $params{server_addresses} || [];

    return if (($log_application->{remote_storage} == F5::PrepareConf::Base::LOG_REMOTE_STORAGE_NONE()) ||
               ($log_application->{logger_type} != F5::Mcp::FwLoggerType::FW_LOGGER_TYPE_REMOTE()));

    for my $server_address (@$server_addresses) {
        my ($remote_storage_conf, $remote_storage_escaping);
        $remote_storage_conf->{log_profile_name} = $remote_storage_escaping->{log_profile_name} = $log_profile_name;
        $remote_storage_conf->{log_profile_name_crc} = $remote_storage_escaping->{log_profile_name_crc} = $log_profile_name_crc;
        $remote_storage_conf->{entity} = LOGGER_ENTITY_ASM; # by default (bitmap)
        $remote_storage_conf->{protocol} = $log_application->{protocol}; # equivalent enums
        $remote_storage_conf->{list_prefix} = $log_application->{list_prefix} || '';
        $remote_storage_conf->{list_delimiter} = $log_application->{list_delimiter} || ',';
        $remote_storage_conf->{list_suffix} = $log_application->{list_suffix} || '';

        my $search_for = $params{search_for};
        my $replace_with = $params{replace_with};
        $remote_storage_conf->{dest_type} = DESTINATION_TYPE_IP_ADDRESS;

        if (!$log_application->{dest_filename} ) {
            if (! F5::IpUtils::is_ip_address($server_address->{address}) ) {
                $remote_storage_conf->{dest_type} = DESTINATION_TYPE_HOSTNAME;
                $remote_storage_conf->{server_hostname} = $server_address->{address};
                $remote_storage_conf->{server_ip} = '';
                $remote_storage_conf->{route_domain} = '';
                $remote_storage_conf->{ip_with_route_domain} = '';
            }
            else {
                my $split_ip_route = F5::IpUtils::split_ip_route( ip_address =>  $server_address->{address} , die_on_error => 0 );
                if (! $split_ip_route->{completed_successfully}) {
                    $logger->info("Could not split [$server_address->{address}] into ip and route") if $logger;
                    next;
                }

                $remote_storage_conf->{server_ip} =  $split_ip_route->{ip_address};
                $remote_storage_conf->{route_domain} = $split_ip_route->{route_domain};
                $remote_storage_conf->{ip_with_route_domain} =  $server_address->{address}; 
            }
            # NOTE: COAPI attempts to find a services entry for the given IpPort (in /etc/services)
            $remote_storage_conf->{server_port} = getservbyname($server_address->{port}, '');
            $remote_storage_conf->{server_port} = $server_address->{port} unless defined $remote_storage_conf->{server_port};
        }
        else {
            $remote_storage_conf->{server_ip} = '';
            $remote_storage_conf->{route_domain} = '';
            $remote_storage_conf->{ip_with_route_domain} = '';
            $remote_storage_conf->{server_port} = '';
            $remote_storage_conf->{dest_type} = DESTINATION_TYPE_FILE;
        }

        $remote_storage_conf->{facility} = $log_application->{facility} + 16; # equivalent enums with offset 16
        $remote_storage_conf->{severity_id} = 6; # info
        $remote_storage_conf->{flg_log_anomaly_attacks} = $log_application->{report_anomalies};
        $remote_storage_conf->{flg_log_cloud_security} = $log_application->{report_cloud_security};
        $remote_storage_conf->{response_logging} = $log_application->{response_logging}; # equivalent enums
        $remote_storage_conf->{max_req_len} = $log_application->{maximum_request_size};
        $remote_storage_conf->{max_hdrs_len} = $log_application->{maximum_header_size};
        $remote_storage_conf->{max_qs_len} = $log_application->{maximum_query_size};
        $remote_storage_conf->{max_log_length} = $log_application->{maximum_entry_length};
        $remote_storage_conf->{dest_filename} = $log_application->{dest_filename} || "";

        if ($log_application->{remote_storage} == F5::PrepareConf::Base::LOG_REMOTE_STORAGE_ARCSIGHT()) { # in arcsight use the hardcoded string
            $remote_storage_conf->{entity} |= F5::PrepareConf::Base::LOGGER_ENTITY_ARCSIGHT();
            $remote_storage_conf->{storage_format} = $ARCSIGHT_STORAGE_FORMAT;
            $remote_storage_conf->{storage_format_fmt} = $ARCSIGHT_STORAGE_FORMAT;
            $remote_storage_conf->{storage_format_entities_bitmap} = get_bigint_mask(('1') x $BIGINT_WIDTH);

            $search_for = ARCSIGHT_SEARCH_FOR;
            $replace_with = ARCSIGHT_REPLACE_WITH;
        }
        else {
            if ($log_application->{remote_storage} == F5::PrepareConf::Base::LOG_REMOTE_STORAGE_SPLUNK()) {
                $remote_storage_conf->{entity} |= F5::PrepareConf::Base::LOGGER_ENTITY_SPLUNK();
                $log_application->{user_string} = $SPLUNK_FORMAT_STRING;

                $search_for = SPLUNK_SEARCH_FOR;
                $replace_with = SPLUNK_REPLACE_WITH;
            }
            elsif ($log_application->{remote_storage} == F5::PrepareConf::Base::LOG_REMOTE_STORAGE_GRPC()) {
                $log_application->{user_string} = $GRPC_STORAGE_FORMAT;

                $search_for = SPLUNK_SEARCH_FOR;
                $replace_with = SPLUNK_SEARCH_FOR x 2;
            }
            elsif ($log_application->{remote_storage} == F5::PrepareConf::Base::LOG_REMOTE_STORAGE_BIGIQ()) {
                $remote_storage_conf->{entity} |= F5::PrepareConf::Base::LOGGER_ENTITY_BIGIQ();
                $log_application->{user_string} = $BIGIQ_FORMAT_STRING;

                $search_for = BIGIQ_SEARCH_FOR;
                $replace_with = BIGIQ_REPLACE_WITH;
            }
            else {
                $search_for ||= SPLUNK_SEARCH_FOR;
                $replace_with ||= SPLUNK_SEARCH_FOR x 2;
            }

            if ($log_application->{user_string} ne "") {
                $remote_storage_conf->{storage_format} = $log_application->{user_string};
            }
            else {
                my @field_list;

                for my $field_name (@{$log_application->{fields}}) {
                    my $field_format = $log_application->{field_format} || '"%v"'; # default for CSV
                    $field_format =~ s/\%k/$field_name/g;
                    $field_format =~ s/\%v/\%$field_name\%/g;
                    push(@field_list, $field_format);
                }

                $remote_storage_conf->{storage_format} = join($log_application->{field_delimiter}, @field_list);
            }

            my $rv = change_storage_format(storage_format_template => $remote_storage_conf->{storage_format});
            $remote_storage_conf->{storage_format} = $rv->{storage_format};
            $remote_storage_conf->{storage_format_fmt} = $rv->{storage_format_fmt};
            my @bitmap = @{$rv->{bitmap}};

            $remote_storage_conf->{storage_format_entities_bitmap} = get_bigint_mask(@bitmap);
        }

        $self->insert_row(table_name => 'DCC.REMOTE_STORAGE_CONFIGURATION',
                          do_replace => 1,
                          rec => $remote_storage_conf);
        if (!$log_application->{remote_storage_escaping} || !@{$log_application->{remote_storage_escaping}}) {
            $log_application->{remote_storage_escaping} = [{search_for => $search_for, replace_with => $replace_with}];
        }
        for my $escaping (@{$log_application->{remote_storage_escaping}}) {
            $self->insert_row(table_name => 'DCC.REMOTE_STORAGE_FORMAT_ESCAPING',
                              rec => {%$remote_storage_escaping, search_for => $escaping->{search_for}, replace_with => $escaping->{replace_with} },
                              do_replace => 1);
        }
    }
    
    return scalar @$server_addresses;
}

#----------------------------------------------------------------------

sub prepare_log_application_filter {
    my ($self, %params) = @_;

    my $log_application = $params{log_application};
    my $log_profile_name = $params{log_profile_name};
    my $log_profile_name_crc = $params{log_profile_name_crc};
    my $server_found = $params{server_found};
    my $request_filters = $params{request_filters} || [];

    my ($local_filter, $remote_filter);
    $local_filter->{log_profile_name} = $remote_filter->{log_profile_name} = $log_profile_name;
    $local_filter->{log_profile_name_crc} = $remote_filter->{log_profile_name_crc} = $log_profile_name_crc;
    $local_filter->{flg_log_challenge_failure_requests} = $remote_filter->{flg_log_challenge_failure_requests} = 0;
    $local_filter->{entity} = LOGGER_ENTITY_ASM;
    $local_filter->{lrn_mask} = 0;
    $remote_filter->{start_time} = 0;
    $remote_filter->{end_time} = 0;
    $local_filter->{logic_relation} = $remote_filter->{logic_relation} = $log_application->{logic_operation}; # equivalent enums

    for my $request_filter (@$request_filters) {
        if ($request_filter->{key} == LOG_FILTER_KEY_REQUEST_TYPE) {
            $local_filter->{request_type} = $remote_filter->{request_type}
                = ($request_filter->{value_list}[0] eq 'illegal') ? LOGGER_REQUEST_TYPE_ILLEGAL
                : ($request_filter->{value_list}[0] eq 'illegal-including-staged-signatures') ? LOGGER_REQUEST_TYPE_ILLEGAL_INCLUDING_STAGED_SIGNATURES
                : ($request_filter->{value_list}[0] eq 'blocked') ? LOGGER_REQUEST_TYPE_BLOCKED
                : LOGGER_FILTER_ALL;
        }
        elsif ($request_filter->{key} == LOG_FILTER_KEY_PROTOCOL) {
            $local_filter->{protocols} = LOGGER_FILTER_ALL;
            if (grep { $_ eq 'http' } @{$request_filter->{value_list}}) { $local_filter->{protocols} |= LOGGER_PROTOCOL_HTTP;  }
            if (grep { $_ eq 'https'} @{$request_filter->{value_list}}) { $local_filter->{protocols} |= LOGGER_PROTOCOL_HTTPS; }
            if (grep { $_ eq 'ws'   } @{$request_filter->{value_list}}) { $local_filter->{protocols} |= LOGGER_PROTOCOL_WS;    }
            if (grep { $_ eq 'wss'  } @{$request_filter->{value_list}}) { $local_filter->{protocols} |= LOGGER_PROTOCOL_WSS;   }
            $remote_filter->{protocols} = $local_filter->{protocols};
        }
        elsif ($request_filter->{key} == LOG_FILTER_KEY_RESPONSE_CODE) {
            $local_filter->{respones_code} = $remote_filter->{respones_code}
                = ($request_filter->{value_list}[0] eq 'all') ? ''
                : join(',', @{$request_filter->{value_list}});
        }
        elsif ($request_filter->{key} == LOG_FILTER_KEY_HTTP_METHOD) {
            if ($request_filter->{value_list}[0] eq 'all') {
            	$local_filter->{http_methods} = $remote_filter->{http_methods} = '';
            }
            else {
                my $http_methods = F5::ASM::COAPI::Query->new('asm_http_method');
                my @method_list;

                for my $method_name (@{$request_filter->{value_list}}) {
                    my ($method_code) = map { $_->{code} }
                                        grep { $_->{leaf_name} eq $method_name }
                                        @$http_methods;
                    push(@method_list, $method_code) if defined $method_code;
                }

                $local_filter->{http_methods} = $remote_filter->{http_methods} = join(',', @method_list);
            }
        }
        elsif ($request_filter->{key} == LOG_FILTER_KEY_LOGIN_RESULT) {
           	$local_filter->{login_results} = LOGGER_FILTER_ALL;
            
	    my %login_result_mask_values = ( "login-result-successful" => 1,
                              	      	     "login-result-failed" => 2,
                                             "login-result-unknown" => 4,
					    );

	    $local_filter->{login_results} = List::Util::sum(map { $login_result_mask_values{$_} } @{$request_filter->{value_list}});
       	    $remote_filter->{login_results} = $local_filter->{login_results};
        }
        elsif ($request_filter->{key} == LOG_FILTER_KEY_LOG_CHALLENGE_FAILURE_REQUESTS) {
            $local_filter->{flg_log_challenge_failure_requests} = $remote_filter->{flg_log_challenge_failure_requests} = ($request_filter->{value_list}[0] eq 'enabled') ? 1 : 0 ;
        }
        elsif ($request_filter->{key} != LOG_FILTER_KEY_SEARCH_ALL) { # all behaves like nothing
            $local_filter->{apply_pattern_pair1_to} = $remote_filter->{apply_pattern_to} = $request_filter->{key};
            $local_filter->{pair1_pattern1} = $remote_filter->{pattern1} = $request_filter->{value_list}[0];
        }
    }

    $local_filter->{apply_pattern_pair2_to} = 0;
    $local_filter->{pair1_pattern2} = $local_filter->{pair2_pattern1} = $local_filter->{pair2_pattern2} = '';
    
    # Complete default values for unspecified filters
    $local_filter->{request_type} = $remote_filter->{request_type} = LOGGER_REQUEST_TYPE_ILLEGAL unless exists $local_filter->{request_type};
    $local_filter->{protocols} = $remote_filter->{protocols} = LOGGER_FILTER_ALL unless exists $local_filter->{protocols};
    $local_filter->{respones_code} = $remote_filter->{respones_code} = '' unless exists $local_filter->{respones_code};
    $local_filter->{http_methods} = $remote_filter->{http_methods} = '' unless exists $local_filter->{http_methods};
    $local_filter->{login_results} = $remote_filter->{login_results} = LOGGER_FILTER_ALL unless exists $local_filter->{login_results};

    if ($log_application->{logger_type} == F5::Mcp::FwLoggerType::FW_LOGGER_TYPE_LOCAL()) {
        $self->insert_row(table_name => 'DCC.LOCAL_FILTER',
                          do_replace => 1,
                          rec => $local_filter);
    }
    elsif ($log_application->{remote_storage} != F5::PrepareConf::Base::LOG_REMOTE_STORAGE_NONE() &&
           $server_found && ($log_application->{logger_type} == F5::Mcp::FwLoggerType::FW_LOGGER_TYPE_REMOTE())) {
        $self->insert_row(table_name => 'DCC.REMOTE_FILTER',
                          do_replace => 1,
                          rec => $remote_filter);
    } 
}

#----------------------------------------------------------------------

sub prepare_log_protocol_remote_storage_conf {
    my ($self, %params) = @_;

    my $logger = $params{logger} || $self->{logger};
    my $publisher_name = $params{publisher_name};
    my $log_profile_name = $params{log_profile_name};
    my $log_profile_name_crc = $params{log_profile_name_crc};
    my $publisher_names = $params{publisher_names};
    my $destination_names = $params{destination_names};
    my $dest_pool_names = $params{dest_pool_names};
    my $pool_names = $params{pool_names};
    my $profile_updated = $params{profile_updated};
    
    my $log_publisher_destination = F5::ASM::COAPI::Query->new('log_publisher_destination', {log_publisher_name => $publisher_name});
    return unless ($log_publisher_destination && @$log_publisher_destination && $log_publisher_destination->[0]->{log_destination_name});
    my $destination_name = $log_publisher_destination->[0]->{log_destination_name}; # multiple destinations are disallowed for protocol publisher

    # Initial values
    my $entity = LOGGER_ENTITY_HSL; # bitmap
    my $facility = 0; # local0
    my $severity_id = 6; # info
    my ($storage_format, $storage_format_entities_bitmap, $search_for, $replace_with, $dest_filename);

    my $dest_pool_name;
    my $log_destination_remote = F5::ASM::COAPI::Query->new('log_destination_remote_syslog', {name => $destination_name});
    
    if ($log_destination_remote && @$log_destination_remote) {
    	$facility = $log_destination_remote->[0]->{default_facility};
        $severity_id = $log_destination_remote->[0]->{default_severity};
        $storage_format = '';
        $storage_format_entities_bitmap = get_bigint_mask(('0') x $BIGINT_WIDTH);
        $search_for = SPLUNK_SEARCH_FOR;
        $replace_with = SPLUNK_SEARCH_FOR x 2;
        $dest_pool_name = $log_destination_remote->[0]->{remote_hsl_name};
    }
    else { # name of log destination is unique across all its types
        my $log_destination_bigiq = F5::ASM::COAPI::Query->new('log_destination_bigiq', {name => $destination_name});

        if ($log_destination_bigiq && @$log_destination_bigiq) {
            $entity |= F5::PrepareConf::Base::LOGGER_ENTITY_BIGIQ();

            my $rv = change_storage_format(storage_format_template => $BIGIQ_FORMAT_STRING);
            $storage_format = $rv->{storage_format};
            my @bitmap = @{$rv->{bitmap}};

            $storage_format_entities_bitmap = get_bigint_mask(@bitmap);
            $search_for = BIGIQ_SEARCH_FOR;
            $replace_with = BIGIQ_REPLACE_WITH;
            $dest_pool_name = $log_destination_bigiq->[0]->{forward_to};
        }
        else {
            my $log_destination_splunk = F5::ASM::COAPI::Query->new('log_destination_splunk', {name => $destination_name});

            if ($log_destination_splunk && @$log_destination_splunk) {
                $entity |= F5::PrepareConf::Base::LOGGER_ENTITY_SPLUNK();

                my $rv = change_storage_format(storage_format_template => $SPLUNK_FORMAT_STRING);
                $storage_format = $rv->{storage_format};
                my @bitmap = @{$rv->{bitmap}};

                $storage_format_entities_bitmap = get_bigint_mask(@bitmap);
                $search_for = SPLUNK_SEARCH_FOR;
                $replace_with = SPLUNK_REPLACE_WITH;
                $dest_pool_name = $log_destination_splunk->[0]->{forward_to};
            }
            else {
                my $log_destination_arcsight = F5::ASM::COAPI::Query->new('log_destination_arcsight', {name => $destination_name});

                if ($log_destination_arcsight && @$log_destination_arcsight) {
                    $entity |= F5::PrepareConf::Base::LOGGER_ENTITY_ARCSIGHT();
                    $storage_format = $ARCSIGHT_STORAGE_FORMAT;
                    $storage_format_entities_bitmap = get_bigint_mask(('1') x $BIGINT_WIDTH);
                    $search_for = ARCSIGHT_SEARCH_FOR;
                    $replace_with = ARCSIGHT_REPLACE_WITH;
                    $dest_pool_name = $log_destination_arcsight->[0]->{forward_to};
                }
           }
        }
    }

    return unless $dest_pool_name; # found and not none
    my $dest_pool = F5::ASM::COAPI::Query->new('log_destination_remote_hsl', {name => $dest_pool_name});
    return unless ($dest_pool && @$dest_pool && $dest_pool->[0]->{pool_name});
    my $pool_name = $dest_pool->[0]->{pool_name};

    # was anything updated? ...
    my $publisher_updated = grep { $_ eq $publisher_name } @$publisher_names;
    my $destination_updated = grep { $_ eq $destination_name } @$destination_names;
    my $dest_pool_updated = grep { $_ eq $dest_pool_name } @$dest_pool_names;
    my $pool_updated = grep { $_ eq $pool_name } @$pool_names;

    # ... if so - delete and go on to insert
    if ($profile_updated || $publisher_updated || $destination_updated || $dest_pool_updated || $pool_updated) {
        $self->empty_table($LOGGING_TABLES, undef, {"log_profile_name" => $log_profile_name});
    } else {
        return;
    }

    my $protocol = ($dest_pool->[0]->{protocol}  == LOG_POOL_PROTOCOL_UDP) ? LOG_REMOTE_PROTOCOL_UDP : LOG_REMOTE_PROTOCOL_TCP;
    my $pool_members = F5::ASM::COAPI::Query->new('pool_member', {pool_name => $pool_name});

    for my $pool_member (@$pool_members) {
        my ($remote_storage_conf, $remote_storage_escaping);
        $remote_storage_conf->{log_profile_name} = $remote_storage_escaping->{log_profile_name} = $log_profile_name;
        $remote_storage_conf->{log_profile_name_crc} = $remote_storage_escaping->{log_profile_name_crc} = $log_profile_name_crc;
        $remote_storage_conf->{entity} = $entity;
        $remote_storage_conf->{protocol} = $protocol;

        $pool_member->{addr} = F5::ServerLookup::get_ip_address($pool_member->{addr});
        my $split_ip_route = F5::IpUtils::split_ip_route( ip_address =>  $pool_member->{addr} , die_on_error => 0 );
        if (! $split_ip_route->{completed_successfully}) {
            $logger->info("Could not split [$pool_member->{addr}] into ip and route") if $logger;
            next;
        }

        $remote_storage_conf->{server_ip} =  $split_ip_route->{ip_address};
        $remote_storage_conf->{route_domain} = $split_ip_route->{route_domain};
        $remote_storage_conf->{ip_with_route_domain} =  $pool_member->{addr};

        # NOTE: COAPI attempts to find a services entry for the given IpPort (in /etc/services)
        $remote_storage_conf->{server_port} = getservbyname($pool_member->{port}, '');
        $remote_storage_conf->{server_port} = $pool_member->{port} unless defined $remote_storage_conf->{server_port};

        $remote_storage_conf->{facility} = $facility + 16; # equivalent enums with offset 16
        $remote_storage_conf->{severity_id} = $severity_id;
        $remote_storage_conf->{storage_format} = $storage_format;
        $remote_storage_conf->{storage_format_entities_bitmap} = $storage_format_entities_bitmap;
        $remote_storage_escaping->{search_for} = $search_for;
        $remote_storage_escaping->{replace_with} = $replace_with;

        # Default values
        $remote_storage_conf->{flg_log_anomaly_attacks} = 0;
        $remote_storage_conf->{flg_log_cloud_security} = 0;
        $remote_storage_conf->{response_logging} = 1;
        $remote_storage_conf->{max_req_len} = 0;
        $remote_storage_conf->{max_hdrs_len} = 0;
        $remote_storage_conf->{max_qs_len} = 0;
        $remote_storage_conf->{max_log_length} = 1024;

        $self->insert_row(table_name => 'DCC.REMOTE_STORAGE_CONFIGURATION',
                          do_replace => 1,
                          rec => $remote_storage_conf);

        $self->insert_row(table_name => 'DCC.REMOTE_STORAGE_FORMAT_ESCAPING',
                          rec => $remote_storage_escaping,
                          do_replace => 1);
    }
    return 1;
}

#----------------------------------------------------------------------

sub get_bigint_mask {
    my(@bits) = @_;
    my $vec = Bit::Vector->new_Bin(256, join('', reverse @bits[1..$#bits]));
    return pack("H*", $vec->to_Hex);
}
#----------------------------------------------------------------------

sub get_storage_fields {
    return F5::ASM::COAPI::Query->new('fw_log_storage_field');
}

#----------------------------------------------------------------------

sub prepare_config_set_loggers {
    my($self, %params) = @_;
    my $config_set_loggers = $params{config_set_loggers};
    my $seen_vs_name_crcs = $params{seen_vs_name_crcs};
    my $seen_log_profile_name_crcs = $params{seen_log_profile_name_crcs};
    my $change_made = 0;

    $self->{logger}->debug("Preparing config set loggers") if $self->{logger};

    for my $config_set_logger (@$config_set_loggers) {
        $config_set_logger->{vs_name_crc} ||= F5::CRC::get_crc32($config_set_logger->{vs_name});
        $seen_vs_name_crcs->{$config_set_logger->{vs_name_crc}} = 1;

        $self->prepare_logging_profile(log_profile => $config_set_logger);
        $seen_log_profile_name_crcs->{$config_set_logger->{log_profile_name_crc}} = 1;
        $self->prepare_log_prof_mapping_conf(%$config_set_logger);
        $change_made = 1;
    }
    return $change_made;
}

#----------------------------------------------------------------------

my %FORMAT_TYPE_MAP = ( F5::ASMConfig::RequestLoggerFormatType::REQUEST_LOGGER_FORMAT_TYPE_SPLUNK() => F5::PrepareConf::Base::LOG_REMOTE_STORAGE_SPLUNK(),
                        F5::ASMConfig::RequestLoggerFormatType::REQUEST_LOGGER_FORMAT_TYPE_ARCSIGHT() => F5::PrepareConf::Base::LOG_REMOTE_STORAGE_ARCSIGHT(),
                        F5::ASMConfig::RequestLoggerFormatType::REQUEST_LOGGER_FORMAT_TYPE_BIGIQ() => F5::PrepareConf::Base::LOG_REMOTE_STORAGE_BIGIQ(),
                        F5::ASMConfig::RequestLoggerFormatType::REQUEST_LOGGER_FORMAT_TYPE_DEFAULT() => F5::PrepareConf::Base::LOG_REMOTE_STORAGE_REMOTE(),
                        0 => F5::PrepareConf::Base::LOG_REMOTE_STORAGE_REMOTE(),
                      );
my $UNLIKELY_DELIMITER = '###UNLIKELY_DELIMITER###';

sub prepare_request_loggers {
    my($self, %params) = @_;
    my $policy = $params{policy};
    my $change_made = 0;

    my @policy_name_crcs;
    if ($policy && $policy->{policy_name}) {
        @policy_name_crcs =  (F5::CRC::get_crc($policy->{policy_name}));
        $self->{logger}->debug("Preparing request logger for $policy->{policy_name}") if $self->{logger};
    }
    else {
        $self->{logger}->debug("Preparing all request loggers") if $self->{logger};
    }
    my $rv = $self->empty_table('DCC.ACCOUNT_LOG_PROFILE', @policy_name_crcs);
    $change_made ||= $rv;

    my $where_fields = {};
    if ($policy && $policy->{policy_id}) {
        $where_fields->{policy_id} = $policy->{policy_id};
    }
    else {
        # Only apply active policies
        $where_fields->{"policy.account_id"} = { comparison => 'ne',
                                                 values => 0 };
    }

    my $constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => $where_fields);
    my $where_clause = $constructed_where_clause->{where_clause};
    my @where_values = @{$constructed_where_clause->{where_values}};

    my $request_loggers = $self->{dbh}->selectall_arrayref(<<EndQuery, { Slice => {} }, @where_values);
SELECT request_logger.*, policy.name as policy_name
  FROM PLC.PL_REQUEST_LOGGERS request_logger,
       PLC.PL_POLICIES policy
 WHERE policy.id = request_logger.policy_id
   AND $where_clause
EndQuery

    for my $request_logger (@$request_loggers) {
        $request_logger->{log_profile_name} = join('', $request_logger->{policy_name}, $UNLIKELY_DELIMITER, $request_logger->{request_logger_name});
        ($request_logger->{address}, $request_logger->{port}) = split(/:/, $request_logger->{destination});
        if ($request_logger->{format_type}) {
            $request_logger->{remote_storage} = $FORMAT_TYPE_MAP{$request_logger->{format_type}};
        }
        else {
            $request_logger->{user_string} = $request_logger->{format_string};
        }

        $request_logger->{request_type_filter} = 'all';
        my $logger_filters = $self->{dbh}->selectall_arrayref(<<EndQuery, { Slice => {} }, $request_logger->{request_logger_id});
SELECT *
  FROM PLC.PL_REQUEST_LOGGERS_FILTERS
 WHERE request_logger_id = ?
EndQuery

        for my $logger_filter (@$logger_filters) {
            my $field_values = [];
            if ($logger_filter->{field_values}) {
                $field_values = JSON::decode_json($logger_filter->{field_values});
            }
            if ($logger_filter->{field_name} eq 'request_status') {
                my %field_values = map { ($_ => 1) } @$field_values;

                if ($field_values{illegal}) {
                    $request_logger->{request_type_filter} = 'illegal';
                }
                elsif ($field_values{alerted} && $field_values{blocked}) {
                    $request_logger->{request_type_filter} = 'illegal';
                }
                elsif ($field_values{blocked}) {
                    $request_logger->{request_type_filter} = 'blocked';
                }
            }
        }

        $self->prepare_logging_profile(log_profile => $request_logger);
        $self->prepare_policy_request_logger_mapping(log_profile => $request_logger,
                                                     policy_name => $request_logger->{policy_name});
        $change_made = 1;
    }

    return $change_made;
}

#----------------------------------------------------------------------

sub prepare_logging_profile {
    my($self, %params) = @_;
    my $log_profile = $params{log_profile};

    $self->{logger}->debug("Preparing log profile $log_profile->{log_profile_name}") if $self->{logger};

    $log_profile->{log_profile_name_crc} = F5::CRC::get_crc32($log_profile->{log_profile_name});
    # No need to empty_table for all LOGGING_TABLES, since all inserts to these tables will be with do_replace flag.
    # Additionally, these logging profiles can only be remote, and thus cannot switch to/from local.
    # However the DCC.REMOTE_STORAGE_CONFIGURATION table still needs to be emptied, as the address and port are both part of the unique keys, so insert/replace will leave old records behind.
    # Also the DCC.REMOTE_STORAGE_FORMAT_ESCAPING table needs to be emptied as changing the escaping characters won't remove the old ones, as they are not unique keys
    $self->empty_table('DCC.REMOTE_STORAGE_CONFIGURATION', undef, {"log_profile_name_crc" => $log_profile->{log_profile_name_crc}} );
    $self->empty_table('DCC.REMOTE_STORAGE_FORMAT_ESCAPING', undef, {"log_profile_name_crc" => $log_profile->{log_profile_name_crc}} );

    my $log_application = {
                           logger_type => F5::Mcp::FwLoggerType::FW_LOGGER_TYPE_REMOTE(),
                           remote_storage => F5::PrepareConf::Base::LOG_REMOTE_STORAGE_REMOTE(),
                           user_string => $DEFAULT_STORAGE_FORMAT, # user_string is ignored if remote_storage is not LOG_REMOTE_STORAGE_REMOTE (i.e splunk,bigiq..)
                           protocol => F5::Mcp::FwLogRemoteProtocol::FW_LOG_REMOTE_PROTOCOL_TCP(),
                           facility => F5::Mcp::SyslogFacility::SYSLOG_FACILITY_LOG_LOCAL0(),
                           report_anomalies => 0,
                           response_logging => 1, #NONE
                           logic_operation => 0,
                           maximum_request_size => 2048,
                           maximum_header_size => 0,
                           maximum_query_size => 0,
                           maximum_entry_length => $log_profile->{max_message_size} || 64 * $K_BYTES,
                           %$log_profile,
                          };

    my $server_found = $self->prepare_log_application_remote_storage_conf(log_application => $log_application,
                                                                          log_profile_name => $log_profile->{log_profile_name},
                                                                          log_profile_name_crc => $log_profile->{log_profile_name_crc},
                                                                          server_addresses => [
                                                                                               {
                                                                                                address => $log_profile->{address},
                                                                                                port => $log_profile->{port} || $DEFAULT_PORT,
                                                                                               },
                                                                                              ],
                                                                         );

    $self->prepare_log_application_filter(log_application => $log_application,
                                          log_profile_name => $log_application->{log_profile_name},
                                          log_profile_name_crc => $log_profile->{log_profile_name_crc},
                                          server_found => $server_found,
                                          request_filters => [
                                                              { key => F5::Mcp::FwLogFilterKey::FW_LOG_FILTER_KEY_REQUEST_TYPE(),
                                                                value_list => [ $log_profile->{request_type_filter} ],
                                                              },
                                                             ]);
}

#----------------------------------------------------------------------

sub prepare_policy_request_logger_mapping {
    my($self, %params) = @_;
    my $policy_name = $params{policy_name};
    my $log_profile = $params{log_profile};

    my $policy_name_crc = F5::CRC::get_crc($policy_name);

    $self->{logger}->trace("Adding policy/logging profile mapping for $log_profile->{log_profile_name}") if $self->{logger};
    $self->insert_row(table_name => 'DCC.ACCOUNT_LOG_PROFILE',
                      do_replace => 1,
                      rec => {
                              policy_name_crc => $policy_name_crc,
                              log_profile_name_crc => $log_profile->{log_profile_name_crc},
                             });

}

#----------------------------------------------------------------------

1;
