package F5::PrepareFiles;

use strict;
use Cwd;

use Sys::Hostname;

use F5::Deploy;
use F5::Logger qw(:LEGACY_ERROR_CONSTANTS);

use F5::Cfg;
use F5::SetActive;
use F5::DbUtils;
use F5::Provisioning;
use F5::PrepareConf::Base;
use F5::ASM::COAPI;
use F5::PrepareConf::BotDefense;
use F5::Utils::File;
use F5::BdCfgTables;

use Digest::MD5;
use File::Copy;
use File::Basename;
use File::Path;
use File::Find;
use Data::MessagePack;
use F5::MessagePack;
use YAML::XS;

use MIME::Base64;
use Time::HiRes;
use POSIX qw/strftime/;

use F5::BigipVersionUtils;
use F5::ProtobufDataFile;

use vars qw{$RET_OK
	    $RET_NOT_OK
	    $LOG_FILE
	    $PRINT_TARGET
	    $DEBUG
	    };

#default log file for this module, may be changed from upper level
$LOG_FILE = 'prepare_files.log';
$PRINT_TARGET = 3;
my $CONF_FILE = "/etc/ts/dcc/prepare_files.cfg";
my $CHECKSUM_FILE = F5::Deploy::map_path('/ts/recovery/internal_config_checksum');
my $DATA_FILE_SUFFIX = 'msgp';

my $DCC_TABLE_CFG;
my $LOGGER;

my $ROW_ACTION_KEEP = 1;
my $ROW_ACTION_GET_NEXT = 2;
my $ROW_ACTION_FINISHED = 3;
$DEBUG = 0;

$RET_OK = 0;
$RET_NOT_OK = -1;

#--------------------------------------------------------------------------------
sub PrepareSiteTree
{
    my (%params) = @_;

    my $site_id = $params{site_id};
    my $test_site_dir = $params{test_site_dir};

    my $site_dir = (defined $test_site_dir) ? $test_site_dir. '/files/site_1' : '/ts/files/site_' . $site_id;
    pf_logprint($INF,undef,"started, site_pid=$site_id\n");

    my $conf_dir = F5::Deploy::map_path($site_dir . '/config');

    F5::Utils::File::delete_files_recursively($conf_dir) if(-d $conf_dir);

    my @SITE_CONFIG_DIRS = qw(full);
    for my $dir (@SITE_CONFIG_DIRS) {
        File::Path::make_path("$conf_dir/$dir");
    }

    return $RET_OK;
}

#--------------------------------------------------------------------------------
sub PrepareInternalConfigFiles
{
    my (%params) = @_;

    my $site_id = $params{site_id};
    my $test_site_dir = $params{test_site_dir};
    my $dbh = $params{dbh};

    my $dest_dir = (defined $test_site_dir) ? $test_site_dir : F5::Deploy::map_path('/ts/files/site_' . $site_id.'/config');
    pf_logprint($INF,undef,"started, site_id=$site_id\n");

    if ( !$dbh)
    {
	pf_logprint($ERR, 'PrepareInternalConfigFiles', "Missing dbh");
	return $RET_NOT_OK;
    }

    if (!F5::BigipUtils::is_docker()) {
        if (open (my $fh, '>', $CHECKSUM_FILE)) {
            my ($junk, $current_checksum) = $dbh->selectrow_array("CHECKSUM TABLE DCC.INTERNAL_CONFIG_ITEMS");
            print $fh $current_checksum;
            close ($fh);
        }
        else {
            pf_logprint($ERR, 'PrepareInternalConfigFiles', "Could not open $CHECKSUM_FILE");
            return $RET_NOT_OK;
        }
    }

    my $sql = <<EndQuery;
SELECT cfg_files.file_name,
       cfg_sections.section_name,
       cfg_items.item_id,
       cfg_items.field_name,
       cfg_items.current_value,
       cfg_items.internal_remark
 FROM DCC.INTERNAL_CONFIG_FILES cfg_files
   JOIN DCC.INTERNAL_CONFIG_SECTIONS cfg_sections USING (file_id)
   JOIN DCC.INTERNAL_CONFIG_ITEMS cfg_items USING (section_id)
 WHERE cfg_files.file_name IS NOT NULL
 ORDER BY cfg_sections.section_order, cfg_items.item_order, cfg_items.item_id
EndQuery

    # read all config items

    my $sth = $dbh->prepare($sql);

    if (not $sth)
    {
	pf_logprint($ERR, 'DBI->prepare',"'$sql': '$DBI::errstr'");
	return $RET_NOT_OK;
    }

    if (not $sth->execute())
    {
	pf_logprint($ERR, 'DBI->execute',"'$sql': '$DBI::errstr'");
	return $RET_NOT_OK;
    }

    # get all rows

    my %config_hash = ();

    # keep track of order of sections
    my $order = 1;

    while(my $row = $sth->fetchrow_hashref)
    {
        my ($file, $section, $item_id, $key, $value, $comment) =
            @$row{qw(file_name section_name item_id field_name current_value internal_remark)};

        # avoid using undef
        $section = '' if not defined $section;

        if (! defined $file || $file eq '')
        {
            pf_logprint($INF, undef, "No file_name for item_id='$item_id', skipping");
            next;
        }

        $config_hash{$file}{$section} ||= { order => $order, items => [] };

        # make sure value has no CRLF
        $value =~ s/\r?\n/ /g;

        # check for comment before item itself
        if (defined $comment && $comment ne '')
        {
            # make sure comment has no CRLF
            $comment =~ s/\r?\n/ /g;
            push @{ $config_hash{$file}{$section}{items} }, "// $comment\n";
        }

        push @{ $config_hash{$file}{$section}{items} }, "$key=$value\n";

        $order++;
    }

    $sth->finish;

    # write out cfg files

    my $ret = $RET_OK;

    foreach my $filename (keys %config_hash)
    {
        my $output_file = "$dest_dir/$filename.cfg";

        pf_logprint($DBG, undef, "writing config file $output_file");

        my $filehash = $config_hash{$filename};

        if (not open OUTPUT, '>', $output_file)
        {
            pf_logprint($ERR, undef, "Cannot open cfg file $output_file ($!)");
            $ret = $RET_NOT_OK;
            next;
        }

        my $cfg_file_contents = '';
        # loop through sections in order
        foreach my $section (sort { $filehash->{$a}{order} <=> $filehash->{$b}{order} } keys %$filehash)
        {
            # pf_logprint($DBG, undef, "writing section '$section'");

            # only write section if there is one defined
            if ($section ne '')
            {
                $cfg_file_contents .= "\n[$section]\n\n";
            }

            # loop through each item
            foreach my $value (@{ $filehash->{$section}{items} })
            {
                $cfg_file_contents .= $value;
            }
        }

        print OUTPUT $cfg_file_contents;
        close OUTPUT;

        if ($filename eq 'bd') {
            F5::DbUtils::insert_row(dbh => $dbh,
                                    table_name => 'DCC.INTERNAL_PARAMETERS',
                                    rec => { raw_parameters_ini => $cfg_file_contents },
                                    do_replace => 1,
                                   );
        }
    }

    pf_logprint($INF,undef,"ended, site_id=$site_id\n");

    return $ret;
}

#--------------------------------------------------------------------------------
sub PrepareCoreFiles
{
    my (%params) = @_;

    my $site_id = $params{site_id};
    my $dbh = $params{dbh};
    my $site_dir = (defined $params{test_site_dir}) ? $params{test_site_dir}
                                                    : "/ts/files/site_$site_id";
    my $saved_umask = umask(0000);

    my $conf_dir = F5::Deploy::map_path("$site_dir/config");
    pf_logprint($INF,undef,"started, site_pid=$site_id\n");

    set_bigip_info($dbh);

    # Signature names may be affected by changes in INTERNAL_CONFIG_ITEMS, and need to be
    # prepared in case remote_logger_include_sig_risk_accuracy has changed in the interim.
    my $prepare_conf = F5::PrepareConf::Base->new(dbh => $dbh);
    $prepare_conf->prepare_negsig_enforcer_names;
    $prepare_conf->prepare_server_technologies_fingerprints;
    $prepare_conf->prepare_icap_server();
    $prepare_conf->prepare_database_protection_server();
    $prepare_conf->prepare_viol_rating;
    $prepare_conf->prepare_enforcer_accounts();
    $prepare_conf->persist_to_mcp();
    $prepare_conf->prepare_data_protection();
    $prepare_conf->prepare_deception_response_pages();
    $prepare_conf->prepare_deception_traps();

    my $prepare_core_files_result = prepare_core_files(dbh => $dbh,
                                                       dest_dir => $conf_dir,
                                                       prepare_account_tables => 1,
                                                      );

    my $bot_defense = F5::PrepareConf::BotDefense->new(dbh => $dbh,
                                                       logger => $LOGGER,
                                                      );
    $bot_defense->prepare_bot_defense_profiles();

    pf_logprint($INF,undef,"ended, site_pid=$site_id\n");
    umask($saved_umask);
    return ($prepare_core_files_result->{completed_successfully} ? $RET_OK : $RET_NOT_OK);
}
#--------------------------------------------------------------------------------
sub set_bigip_info {
    my ($dbh) = @_;

    F5::DbUtils::update_row(dbh => $dbh,
                            table_name => 'DCC.BIG_IP_UNIT_INFO',
                            rec => {
                                    software_version => F5::BigipVersionUtils::bigip_version(),
                                    flg_is_asm_provisioned => F5::Provisioning::is_module_enabled('asm'),
                                   },
                           );
}

#--------------------------------------------------------------------------------
sub PrepareCoreFilesUpdate {
    my (%params) = @_;

    my $site_id = $params{site_id};
    my $dbh = $params{dbh};
    my $site_dir = (defined $params{test_site_dir}) ? $params{test_site_dir}
                                                    : "/ts/files/site_$site_id";
    my $conf_dir = F5::Deploy::map_path("$site_dir/config");

    my $logger = $LOGGER;
    my $policy = $params{policy};
    my $original_policy_name_crc = $params{original_policy_name_crc};

    $logger->info("Started BD data files prepare") if $logger;

    my $prepare_core_files_result = prepare_core_files(dbh => $dbh,
                                                       dest_dir => $conf_dir,
                                                       policy => $policy,
                                                       generate_update => 1,
                                                       original_policy_name_crc => $original_policy_name_crc,
                                                       prepare_account_tables => $params{prepare_account_tables},
                                                       modified_dcc_tables => $params{modified_dcc_tables},
                                                       job => $params{job},
                                                      );

# Uncomment for debugging only
###    save_conf_update_files_for_debug($site_dir);
    $logger->info("Completed BD data files prepare") if $logger;

    return { completed_successfully => $prepare_core_files_result->{completed_successfully},
             update_dir => $prepare_core_files_result->{target_dirs}->{deltas},
           };
}

#--------------------------------------------------------------------------------

sub cleanup_update_files {
    my (%params) = @_;

    my $site_id = $params{site_id} || 1;
    my $site_dir = (defined $params{test_site_dir}) ? $params{test_site_dir} : "/ts/files/site_$site_id";
    my $update_base_dir = F5::Deploy::map_path("$site_dir/config");

    # Preserve the most recent dir, since it will be needed for future delta calculation
    opendir(my $dirh, $update_base_dir);
    my @previous_dirs = sort { $a <=> $b } grep { /^\d+$/} readdir($dirh);
    closedir($dirh);

    F5::Utils::File::cleanup_files(logger => $LOGGER,
                                   dir => $update_base_dir,
                                   max_size => 25 * 1024 * 1024, # 25 MB
                                   retain_dirs => [ $previous_dirs[-1] ],
                                   min_age => 60,
                                  );
}

#----------------------------------------------------------------------
# prepare_core_files builds the full DCC config files, either for all accounts or
# for a particular account (if specified). Wherever possible, config is separated
# into distinct files for each account.
#
# Additionally, prepare_core_files builds update DCC config difference files,
# relative to the last configuration. The update files are composed by iterating
# simultaneously through the DCC database cursor, aka the "fresh" row, and the
# last full config file, aka the "full" row. The current fresh row and current
# full row are compared lexically to determine:
#
#  - If the fresh row keys precede the full row keys, this is a new fresh row and
#    it is added to the update as "+1" operation.
#
#  - If the full row keys precede the fresh row keys, this is a deleted full row
#    and it is added to the update as "-1" operation.
#
#  - If the fresh and full rows have matching keys, but differ in value, this is
#    an updated row and it is added to the update as "0" operation.
#
sub prepare_core_files {
    my(%params) = @_;

    my $dbh = $params{dbh};
    my $dest_dir = $params{dest_dir};
    my $logger = $LOGGER;
    my $generate_update = $params{generate_update} || 0;
    my $policy = $params{policy};
    my $original_policy_name_crc = $params{original_policy_name_crc};
    my $prepare_account_tables = $params{prepare_account_tables};
    my $job = $params{job};
    my %modified_dcc_tables;
    if ($params{modified_dcc_tables}) {
        %modified_dcc_tables = map { ($_ => 1) } @{$params{modified_dcc_tables}};
    }

    my $target_dirs = get_target_dirs(dest_dir => $dest_dir);

    prepare_violations_for_mcp(%params);
    prepare_attack_types_for_mcp(%params);

    my $policy_name_crc = 0;
    if ($original_policy_name_crc) {
        $policy_name_crc = $original_policy_name_crc;
    }
    elsif ($policy && $policy->{policy_name_crc}) {
        $policy_name_crc = $policy->{policy_name_crc};
    }

    my @generated_dcc_tables;

    for my $dcc_table (@$DCC_TABLE_CFG) {
        if (exists $dcc_table->{read_from_dcc_database} && ! $dcc_table->{read_from_dcc_database}) {
            next;
        }
        if ($dcc_table->{split_config_by_account} && ! $prepare_account_tables) {
            # Skip account-specific table for non-account update (such as for logging profile)
            next;
        }

        if (%modified_dcc_tables && ! $modified_dcc_tables{$dcc_table->{table}}) {
            $logger->trace("Skipping $dcc_table->{table} (not specified in table list)") if $logger;
            next;
        }

        push(@generated_dcc_tables, $dcc_table);
        $logger->info("Preparing core files for $dcc_table->{table}") if $logger;

        my ($cfg_query, @bind_params);
        my $full_file = {};
        my ($full_row, $fresh_row);

        my $update_add_file = { dest_dir => $target_dirs->{deltas},
                                prefix =>  sprintf('%04d__', $#generated_dcc_tables + 1),
                                dcc_table => $dcc_table,
                               };
        my $update_del_file = $update_add_file;
        if ($dcc_table->{split_adds_and_deletes}) {
            # Adds and deletes for a particular DCC table may need to be split in order to allow foreign keys to
            # switch over to the new record before the old one is deleted (such as XML profiles, which are referenced
            # from URLs). For such table, deletes will be handled at the end, after all adds and updates have completed.
            #
            # To ensure that foreign key order is respected, deletes must be handled in reverse order. This is accomplished
            # by subtracting the current update index from 10k (there should be fewer than 5k DCC tables for a long long time).
            $update_del_file = { %$update_add_file,
                                 prefix =>  sprintf('%04d__', 10_000 - $#generated_dcc_tables),
                               };
        }

        my $fresh_file = { dest_dir => "$target_dirs->{new_full}",
                           dcc_table => $dcc_table,
                         };

        my @full_files;
        if ($generate_update) {
            if ($dcc_table->{split_config_by_account}) {
                if ($policy_name_crc) {
                    @full_files = ("$target_dirs->{previous_full}/$dcc_table->{output_file}__account_$policy_name_crc.$DATA_FILE_SUFFIX");
                }
                else {
                    @full_files = (sort glob("$target_dirs->{previous_full}/$dcc_table->{output_file}__account*.$DATA_FILE_SUFFIX"));
                }
            }
            else {
                @full_files = ("$target_dirs->{previous_full}/$dcc_table->{output_file}.$DATA_FILE_SUFFIX");
            }
        }

        if ($dcc_table->{legacy_query}) {
            $cfg_query = $dcc_table->{legacy_query};
        }
        else {
            my @field_names = ( map { $_->{name} } @{$dcc_table->{message_fields}} );

            if ($dcc_table->{split_config_by_account} && ! $dcc_table->{fields_by_name}->{policy_name_crc}) {
                unshift(@field_names, 'policy_name_crc');
            }
            my $field_names = join(', ', @field_names);

            my @where_clauses;
            if ($policy_name_crc && $dcc_table->{split_config_by_account}) {
                push(@where_clauses, 'policy_name_crc = ?');
                push(@bind_params, $policy_name_crc);
            }
            my $where_clauses;
            if (@where_clauses) {
                $where_clauses = 'WHERE ' . join(' AND ', @where_clauses);
            }

            my $order_clause = '';
            if ($dcc_table->{order_by_fields} && @{$dcc_table->{order_by_fields}}) {
                if (F5::DbUtils::is_sqlite($dbh)) {
                    $order_clause .= " ORDER BY " . join(', ', map { "CAST($_ AS TEXT)" } @{$dcc_table->{order_by_fields}});
                }
                else {
                    $order_clause .= " ORDER BY " . join(', ', map { "BINARY $_ " } @{$dcc_table->{order_by_fields}});
                }
            }

            $cfg_query = <<EndQuery;
SELECT $field_names
  FROM $dcc_table->{table}
$where_clauses
$order_clause
EndQuery
        }

        $logger->trace("About to execute $cfg_query") if $logger;
        my $values_sth = $dbh->prepare($cfg_query);
        $values_sth->execute(@bind_params);

        my $fresh_row_action = $ROW_ACTION_GET_NEXT;
        my $full_row_action = $ROW_ACTION_GET_NEXT;
        if (! $generate_update || $dcc_table->{send_all_data_in_update} || ! @full_files) {
            $full_row_action = $ROW_ACTION_FINISHED;
        }

        my $md5 = Digest::MD5->new;

        while ($fresh_row_action != $ROW_ACTION_FINISHED || $full_row_action != $ROW_ACTION_FINISHED) {

            if ($fresh_row_action == $ROW_ACTION_GET_NEXT) {
                # Read next row from db cursor
                my @row = $values_sth->fetchrow_array();
                if (! @row) {
                    $fresh_row = {};
                    $fresh_row_action = $ROW_ACTION_FINISHED;
                }
                else {
                    my @row_compare_policy_name_crc;
                    if ($dcc_table->{split_config_by_account} && ! $dcc_table->{fields_by_name}->{policy_name_crc}) {
                        @row_compare_policy_name_crc = (shift(@row));
                    }

                    if ($dcc_table->{has_transformable_fields}) {
                        for my $nullable_field_index (@{$dcc_table->{nullable_field_indexes}}) {
                            if (! defined($row[$nullable_field_index])) {
                                $row[$nullable_field_index] = '';
                            }
                        }
                    }

                    # Compose cfg line from db
                    $fresh_row = { compare_fields => [ @row_compare_policy_name_crc, @row[@{$dcc_table->{order_by_field_indexes}}] ] };

                    my @row_key_values;
                    my $row_index = 0;
                    for my $dcc_field (@{$dcc_table->{message_fields}}) {
                        $fresh_row->{rec}->{element}->{$dcc_field->{name}} = $row[$row_index];
                        if (defined $dcc_field->{is_primary_key} && $dcc_field->{is_primary_key}) {
                            push @row_key_values, $row[$row_index];
                        }
                        $row_index++;
                    }

                    my $delim = ',';
                    # stringified for md5 calculation
                    #     Otherwise ['hello', 'world'] and ['hell', 'oworld'] will have the same MD5.
                    for my $val (@row_key_values) {
                        $md5->add($val, $delim);
                    }
                    $fresh_row->{rec}->{key_md5} = $md5->hexdigest;
                    $md5->reset;

                    # stringified for md5 calculation
                    for my $val (@row) {
                        $md5->add($val, $delim);
                    }
                    $fresh_row->{rec}->{full_md5} = $md5->hexdigest;
                    $md5->reset;

                    if (! $dcc_table->{send_all_data_in_update}) {
                        $fresh_row->{rec}->{action} = 'add';
                    }

                    if ($dcc_table->{split_config_by_account}) {
                        if ($dcc_table->{fields_by_name}->{policy_name_crc}) {
                            $fresh_row->{policy_name_crc} = $row[$dcc_table->{fields_by_name}->{policy_name_crc}->{field_index}];
                        }
                        else {
                            $fresh_row->{policy_name_crc} = $row_compare_policy_name_crc[0];
                        }
                    }

                    write_cfg_line(output_file => $fresh_file,
                                   row => $fresh_row);

                    $fresh_row_action = $ROW_ACTION_KEEP;
                }
            }

            if ($full_row_action == $ROW_ACTION_GET_NEXT) {
                while ($full_row_action == $ROW_ACTION_GET_NEXT) {
                    if ($full_file->{iter}) {
                        $full_row->{rec} = $full_file->{iter}->get_next();

                        if ($full_row->{rec}) {
                            last;
                        }
                        else {
                            $full_file->{iter} = undef;
                            $full_row = {};
                        }
                    }

                    if (! @full_files) {
                        $full_row_action = $ROW_ACTION_FINISHED;
                        last;
                    }
                    else {
                        $full_file->{filename} = shift(@full_files);

                        if (! -e $full_file->{filename}) {
                            next;
                        }
                        if ($full_file->{filename} =~ /account_(\d+)/) {
                            $full_file->{policy_name_crc} = $1;
                        }
                        $full_file->{iter} = F5::MessagePack::Iterator->new(filename => $full_file->{filename});
                    }
                }

                if ($full_row->{rec}) {
                    delete $full_row->{compare_fields};
                    if ($dcc_table->{split_config_by_account}) {
                        $full_row->{policy_name_crc} = $full_file->{policy_name_crc};
                    }

                    $full_row_action = $ROW_ACTION_KEEP;
                }
            }

            if ($full_row_action == $ROW_ACTION_FINISHED && $fresh_row_action == $ROW_ACTION_FINISHED) {
                last;
            }

            if ($generate_update) {
                if ($dcc_table->{send_all_data_in_update}) {
                    if ($fresh_row_action != $ROW_ACTION_FINISHED) {
                        write_cfg_line(output_file => $update_add_file,
                                       row => $fresh_row);
                        $fresh_row_action = $ROW_ACTION_GET_NEXT;
                    }
                }
                elsif ($fresh_row->{rec} && $full_row->{rec} &&
                       $fresh_row->{rec}->{full_md5} eq $full_row->{rec}->{full_md5}) {

                    $full_row_action = $ROW_ACTION_GET_NEXT;
                    $fresh_row_action = $ROW_ACTION_GET_NEXT;
                }
                else {
                    if ($full_row->{rec} && ! $full_row->{compare_fields}) {
                        my @full_fields = ( map { $full_row->{rec}->{element}->{$_->{name}} } @{$dcc_table->{message_fields}} );

                        my @row_compare_policy_name_crc;
                        if ($dcc_table->{split_config_by_account} && ! $dcc_table->{fields_by_name}->{policy_name_crc}) {
                            @row_compare_policy_name_crc = ($full_file->{policy_name_crc});
                        }

                        $full_row->{compare_fields} = [ @row_compare_policy_name_crc, @full_fields[@{$dcc_table->{order_by_field_indexes}}] ];
                    }

                    my $field_comparison = compare_fields($fresh_row, $full_row);

                    if (! $fresh_row->{rec} || ($full_row->{compare_fields} && $field_comparison > 0)) {
                        # Add delete record to update
                        write_cfg_line(output_file => $update_del_file,
                                       row => $full_row,
                                       cfg_line_change => '-1');
                        $full_row_action = $ROW_ACTION_GET_NEXT;
                    }
                    elsif (! $full_row->{rec} || ($fresh_row->{compare_fields} && $field_comparison < 0)) {
                        # Newly added record
                        write_cfg_line(output_file => $update_add_file,
                                       row => $fresh_row);
                        $fresh_row_action = $ROW_ACTION_GET_NEXT;
                    }
                    else {
                        if ($full_row->{rec}->{full_md5} ne $fresh_row->{rec}->{full_md5}) {
                            # Update values for existing record
                            # Updates are handled in the adds file, even for split add/delete tables
                            write_cfg_line(output_file => $update_add_file,
                                           row => $fresh_row,
                                           cfg_line_change => '0');
                        }
                        $full_row_action = $ROW_ACTION_GET_NEXT;
                        $fresh_row_action = $ROW_ACTION_GET_NEXT;
                    }
                }
            }
            else {
                $fresh_row_action = $ROW_ACTION_GET_NEXT;
            }
        }

        if ($update_add_file->{fh}) {
            close($update_add_file->{fh});
        }
        if ($update_del_file->{fh}) {
            close($update_del_file->{fh});
        }
        if ($fresh_file->{fh}) {
            close($fresh_file->{fh});
        }

        $values_sth->finish;
    }

    my @input_dirs = ("$target_dirs->{new_full}");
    if ($target_dirs->{deltas}) {
        push(@input_dirs, $target_dirs->{deltas});
    }
    my @previous_dirs;
    if ($target_dirs->{previous_full}) {
        @previous_dirs = ($target_dirs->{previous_full});
    }
    F5::ProtobufDataFile::convert_protobuf_data_files(
                                                      input_dirs => \@input_dirs,
                                                      previous_dirs => \@previous_dirs,
                                                      logger => $logger,
                                                     );

    if ($target_dirs->{previous_full}) {
        copy_from_previous_full(%params,
                                generated_dcc_tables => { map { ($_->{output_file} => 1) } @generated_dcc_tables },
                                target_dirs => $target_dirs,
                               );
    }

    my $manifest_file = prepare_bd_cfg_manifest(target_dirs => $target_dirs,
                                                job => $job,
                                               );
    my $protobuf_manifest_file = F5::ProtobufDataFile::convert_bd_cfg_manifest(manifest_file => $manifest_file);
    File::Copy::copy($manifest_file, "$dest_dir/" . basename($manifest_file));
    File::Copy::copy($protobuf_manifest_file, "$dest_dir/" . basename($protobuf_manifest_file));

    return { completed_successfully => 1,
             target_dirs => $target_dirs,
           };
}

#--------------------------------------------------------------------------------

sub get_target_dirs {
    my(%params) = @_;

    my $dest_dir = $params{dest_dir};
    opendir(my $dirh, $dest_dir);
    my @previous_dirs = sort { $a <=> $b } grep { /^\d+$/} readdir($dirh);
    closedir($dirh);

    my $target_dirs;
    if (@previous_dirs) {
        my $previous_generation = $previous_dirs[-1];
        my $new_generation = sprintf('%03d', $previous_generation + 1);
        $target_dirs = {
                        previous_generation => $previous_generation + 0,
                        previous_full => "$dest_dir/$previous_generation/full",
                        previous_top_level => "$dest_dir/$previous_generation",
                        new_generation => $new_generation + 0,
                        new_full => "$dest_dir/$new_generation/full",
                        new_top_level => "$dest_dir/$new_generation",
                        deltas => "$dest_dir/$new_generation/deltas",
                       };
    }
    else {
        $target_dirs = {
                        new_full => $dest_dir . '/'. sprintf('%03d', 1) . "/full",
                        new_generation => 1,
                       };
    }
    mkpath($target_dirs->{new_full});
    if ($target_dirs->{deltas}) {
        mkpath($target_dirs->{deltas});
    }

    return $target_dirs;
}

#--------------------------------------------------------------------------------

sub compare_fields {
    my($left_row, $right_row) = @_;

    if (! $left_row->{compare_fields} || ! @{$left_row->{compare_fields}}) {
        if (! $right_row->{compare_fields} || ! @{$right_row->{compare_fields}}) {
            return 0;
        }
        else {
            return 1;
        }
    }
    elsif (! $right_row->{compare_fields} || ! @{$right_row->{compare_fields}}) {
        return -1;
    }

    for my $i (0..$#{$left_row->{compare_fields}}) {
        if (my $cmp_result = $left_row->{compare_fields}->[$i] cmp $right_row->{compare_fields}->[$i]) {
            return $cmp_result;
        }
    }
    return 0;
}

#--------------------------------------------------------------------------------

sub write_cfg_line {
    my(%params) = @_;

    my $output_file = $params{output_file};
    my $row = $params{row};
    my $cfg_line_change = $params{cfg_line_change};

    my $new_filename;
    if ($row->{policy_name_crc} && $output_file->{dcc_table}->{split_config_by_account}) {
        if (! $output_file->{policy_name_crc} || $row->{policy_name_crc} ne $output_file->{policy_name_crc}) {
            $new_filename = "$output_file->{dest_dir}/$output_file->{prefix}$output_file->{dcc_table}->{output_file}__account_$row->{policy_name_crc}";
            $output_file->{policy_name_crc} = $row->{policy_name_crc};
        }
    }
    elsif (! $output_file->{fh}) {
        $new_filename = "$output_file->{dest_dir}/$output_file->{prefix}$output_file->{dcc_table}->{output_file}";
    }

    if (defined $cfg_line_change) {
        $row->{rec}->{action} = ($cfg_line_change eq '-1') ? 'delete' : 'update' ;
    }

    if ($new_filename) {
        if ($output_file->{fh}) {
            close($output_file->{fh});
        }
        open($output_file->{fh}, '>', $new_filename . ".$DATA_FILE_SUFFIX");
        $output_file->{encoder} = Data::MessagePack->new()->canonical();
    }
    my $encoded_rec = $output_file->{encoder}->pack($row->{rec});
    print { $output_file->{fh} } $encoded_rec;
}

#--------------------------------------------------------------------------------

sub copy_from_previous_full {
    my(%params) = @_;

    my $dest_dir = $params{dest_dir};
    my $target_dirs = $params{target_dirs};
    my $policy = $params{policy};
    my $generated_dcc_tables = $params{generated_dcc_tables};
    my $logger = $params{logger};
    my $original_policy_name_crc = $params{original_policy_name_crc};

    my $policy_name_crc;
    if ($original_policy_name_crc) {
        $policy_name_crc = $original_policy_name_crc;
    }
    elsif ($policy && $policy->{policy_name}) {
        $policy_name_crc = F5::CRC::get_crc($policy->{policy_name});
    }

    find({wanted => sub {
              my $filename = $File::Find::name;
              return unless (-f $filename);

              my $copy_prev_file_to_new_full = 0;
              # Ignore full files that are account-specific, but not from the current account
              if ($policy_name_crc) {
                  if ($filename =~ /__account_(\d+)/) {
                      my $file_policy_name_crc = $1;
                      if ($file_policy_name_crc != $policy_name_crc) {
                          $copy_prev_file_to_new_full = 1;
                      }
                  }
              }

              my $base_filename = basename($filename);

              if (! $copy_prev_file_to_new_full) {
                  my $base_filename_without_account = $base_filename;
                  $base_filename_without_account =~ s/(__account_\d+)?\.[^.]+$//;

                  if (! $generated_dcc_tables->{$base_filename_without_account}) {
                      $copy_prev_file_to_new_full = 1;
                  }
              }

              if ($copy_prev_file_to_new_full) {
                  my $new_filename = "$target_dirs->{new_full}/$base_filename";
                  File::Copy::copy($filename, $new_filename);
              }
          },
          no_chdir => 1,
         },
         $target_dirs->{previous_full});
}

#--------------------------------------------------------------------------------

sub prepare_bd_cfg_manifest {
    my(%params) = @_;
    my $target_dirs = $params{target_dirs};
    my $job = $params{job};
    my ($time, $micro) = Time::HiRes::gettimeofday();
    my $datetime_micros = sprintf('%d%06d',$time, $micro);

    my $bd_cfg_manifest;
    if ($target_dirs->{previous_full} && -e "$target_dirs->{previous_top_level}/bd_cfg_manifest.yaml") {
        $bd_cfg_manifest = YAML::XS::LoadFile("$target_dirs->{previous_top_level}/bd_cfg_manifest.yaml");

        # Remove generations with missing (i.e. cleaned) files
        for my $i (reverse 0..$#{$bd_cfg_manifest->{generations}}) {
            my $generation = $bd_cfg_manifest->{generations}->[$i];
            my $remove_this_generation = 0;

            for my $file_list_field (qw(fullFiles deltaFiles dependentFiles)) {
                for my $file (@{$generation->{$file_list_field}}) {
                    if (! -e $file) {
                        $remove_this_generation = 1;
                        last;
                    }
                }
            }

            if ($remove_this_generation) {
                splice(@{$bd_cfg_manifest->{generations}}, $i, 1);
            }
        }
    }
    else {
        # No previous manifest -- create a new one
        $bd_cfg_manifest = {
                             generations => [],
                           };
    }
    if (! $bd_cfg_manifest->{baseConfigId} && ! $ENV{F5WAF_IS_BUILD_TIME}) {
        $bd_cfg_manifest->{baseConfigId} = hostname . "__$datetime_micros";
    }

    my $new_generation = { generationIndex => $target_dirs->{new_generation},
                           datetimeMicros => $datetime_micros,
                           fullFiles => [ map { { filename => $_ } } get_full_files_by_dir($target_dirs->{new_full}) ],
                         };

    if ($target_dirs->{deltas}) {
        $new_generation->{deltaFiles} = [ map { { filename => $_ } } get_delta_files_by_dir($target_dirs->{deltas}) ];
    }

    if ($job) {
        $new_generation->{jobId} = $job->{jobId};
    }

    unshift(@{$bd_cfg_manifest->{generations}}, $new_generation);

    my $manifest_filename = "$target_dirs->{new_top_level}/bd_cfg_manifest.yaml";
    YAML::XS::DumpFile($manifest_filename, $bd_cfg_manifest);
    return $manifest_filename;
}

#----------------------------------------------------------------------

sub get_delta_files_by_dir {
    my($base_dir) = @_;

    my @rv;
    for my $file (sort glob("$base_dir/*")) {
        if (! corresponding_protobuf_file_exists($file)) {
            push(@rv, $file);
        }
    }
    return @rv;
}

#----------------------------------------------------------------------

sub get_full_files_by_dir {
    my($base_dir) = @_;

    my @rv;
    my %table_file_map;
    for my $file (glob("$base_dir/*")) {
        if (! corresponding_protobuf_file_exists($file)) {
            my $table_output_file = basename($file);
            $table_output_file =~ s/(__account_\d+)?\.[^.]+$//;
            $table_output_file =~ s/^.*__//;
            if (! $table_file_map{$table_output_file}) {
                $table_file_map{$table_output_file} = [];
            }

            push(@{$table_file_map{$table_output_file}}, $file);
        }
    }

    my $dcc_table_config = F5::BdCfgTables::get_dcc_table_config();
    for my $dcc_table (@$dcc_table_config) {
        if ($table_file_map{$dcc_table->{output_file}}) {
            push(@rv, @{$table_file_map{$dcc_table->{output_file}}});
        }
    }
    return @rv;
}

#----------------------------------------------------------------------

sub corresponding_protobuf_file_exists {
    my($filename) = @_;

    if ($filename =~ /\.msgp$/) {
        my $corresponding_protobuf_file = $filename;
        $corresponding_protobuf_file =~ s/\.msgp$/.dat/;
        if (-e $corresponding_protobuf_file) {
            return 1;
        }
    }
    return;
}

#--------------------------------------------------------------------------------
sub pf_logprint
{
    $LOGGER->log_print(shift, shift, shift);
}

#----------------------------------------------------------------------

sub init {
    my(%params) = @_;
    my $dbh = $params{dbh};

    $PRINT_TARGET = F5::Cfg::cfg_get_config_item($CONF_FILE, "DEBUG",'PrintTarget');
    $LOG_FILE = '/ts/log/' . F5::Cfg::cfg_get_config_item($CONF_FILE, "DEBUG",'LogFile');
    $LOGGER = $params{logger} || F5::Logger->new($LOG_FILE, $PRINT_TARGET);

    $DEBUG = F5::Cfg::cfg_get_config_item($CONF_FILE, "DEBUG",'DEBUG');

    $DCC_TABLE_CFG = F5::BdCfgTables::get_dcc_table_config(dbh => $dbh);
}

#----------------------------------------------------------------------

# Violations need to be populated in MCP for AVR usage

sub prepare_violations_for_mcp {
    my(%params) = @_;

    my $dbh = $params{dbh};

    my($db_violations) = $dbh->selectall_hashref(<<EndQuery, 'viol_index');
SELECT viol_index, description
  FROM DCC.PARSER_STATUS_CODES
EndQuery

    my $mcp_violations_query = F5::ASM::COAPI::Query->new('asm_violation');
    my %mcp_violations = map { ($_->{code} => $_) } @$mcp_violations_query;

    my %all_viol_indexes = map { ($_ => 1) } (keys %mcp_violations, keys %$db_violations);

    for my $viol_index (keys %all_viol_indexes) {
        my $mcp_violation = $mcp_violations{$viol_index};
        my $db_violation = $db_violations->{$viol_index};

        if ($db_violation) {
            if (! $mcp_violation || $mcp_violation->{name} ne $db_violation->{description}) {
                # Add or update
                if ( $mcp_violation) {
                    $mcp_violation->delete;
                }

                $mcp_violation = F5::ASM::COAPI::Object->new('asm_violation');
                $mcp_violation->{code} = $viol_index;
                $mcp_violation->{name} = $db_violation->{description};
                $mcp_violation->save;
            }
        }
        elsif ($mcp_violation) {
            # No longer exists in MySQL -- needs to be deleted from MCP
            $mcp_violation->delete;
        }
    }
}

#----------------------------------------------------------------------

sub prepare_attack_types_for_mcp {
    my(%params) = @_;

    my $dbh = $params{dbh};

    my($db_attack_types) = $dbh->selectall_hashref(<<EndQuery, 'attack_type_bit_index');
SELECT attack_type_bit_index, attack_type_name
  FROM DCC.ATTACK_TYPE_CODES
EndQuery

    my $mcp_attack_types_query = F5::ASM::COAPI::Query->new('asm_attack_type');
    my %mcp_attack_type_bit_indexes = map { ($_->{attack_type_bit_index} => $_) } @$mcp_attack_types_query;

    my %all_attack_type_bit_indexes = map { ($_ => 1) } (keys %mcp_attack_type_bit_indexes, keys %$db_attack_types);

    for my $attack_type_bit_index (keys %all_attack_type_bit_indexes) {
        my $mcp_attack_type = $mcp_attack_type_bit_indexes{attack_type_bit_index};
        my $db_attack_type = $db_attack_types->{$attack_type_bit_index};

        if ($db_attack_type) {
            if (! $mcp_attack_type || $mcp_attack_type->{name} ne $db_attack_type->{attack_type_name}) {
                # Add or update
                if ( $mcp_attack_type) {
                    $mcp_attack_type->delete;
                }

                $mcp_attack_type = F5::ASM::COAPI::Object->new('asm_attack_type');
                $mcp_attack_type->{attack_type_bit_index} = $attack_type_bit_index;
                $mcp_attack_type->{name} = $db_attack_type->{attack_type_name};
                $mcp_attack_type->save;
            }
        }
        elsif ($mcp_attack_type) {
            # No longer exists in MySQL -- needs to be deleted from MCP
            $mcp_attack_type->delete;
        }
    }
}

#----------------------------------------------------------------------

# Debugging utility to track full/fresh/update files generated by each run sub
sub save_conf_update_files_for_debug {
    my($site_dir) = @_;

    find({wanted => sub {
              my $filename = $File::Find::name;
              return unless -f $filename;
              my $new_file = $filename;
              $new_file =~ s|$site_dir|/var/tmp/conf_copy/$$|;
              mkpath(dirname($new_file));
              copy($filename, $new_file);
          },
          no_chdir => 1,
         },
         $site_dir);
}

#----------------------------------------------------------------------

#package return:
1;
