package F5::ProcessHandler;

use strict;
use warnings;

eval {
  local ($^W) = 0;
  require 'syscall.ph';
  require 'sys/resource.ph';
};
use Bit::Vector;
use IO::Handle;
use File::Find;
use YAML::XS;
use File::Copy;
use File::Basename;
use Time::HiRes qw(sleep time);

use F5::Deploy;
use F5::Cfg;
use F5::SystemUtils;
use F5::NwdUtils;
use F5::NwdUtils::UpdateGTMScore;
use F5::Logger;
use F5::Cluster::Utils;
use F5::ProcessTable;
use F5::BigIpLog;
use F5::ASMReady;
use F5::Provisioning;

use constant {
    CHECK_PROC_INTERVAL => 0.1,
    KILL_PROC_INTERVAL => 0.5,
    SOFT_KILL_TIME_OUT => 10,
    HARD_KILL_TIME_OUT => 15,
    RLIM_INFINITY => '7fffffffffffffff',
    ALL_CPUS_AFFINITY => '0xFFFFFFFF',
    TASKSET_EXEC => '/bin/taskset',
    PID_FILE_FOLDER => '/var/run',
    DCC_CFG_FILE => F5::Deploy::map_path('/etc/ts/dcc/dcc.cfg'),
};

my $process_config;

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = bless {}, $class;
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    $self->{logger} = $params{logger} || F5::Logger->new('/ts/log/process_handler.log');

    if (! $process_config) {
        $process_config = YAML::XS::LoadFile(F5::Deploy::map_path('/usr/share/ts/config/asm_processes.yaml'));

        $self->{bigstart_values} = F5::Cluster::Utils::get_bigstart_values(qw(get_number_physical_cpu split_planes get_total_memory get_tmm_count));
        $self->{bigstart_values}->{last_cpu_id} = $self->{bigstart_values}->{get_number_physical_cpu} ? ($self->{bigstart_values}->{get_number_physical_cpu} - 1)
                                                                                                      : 1;
        $self->{bigstart_values}->{split_planes} ||= 0;

        my $asm_split_planes = 
            (F5::Cluster::Utils::get_mcp_val('scheduler.splitplanes.asmopt')) ne 'false' ? 1 : 0 ;

        $self->{asm_split_planes} = $self->{bigstart_values}->{split_planes} & $asm_split_planes;
        $self->{logger}->debug("setting split_planes action to:[".$self->{asm_split_planes}."]");
    }

}

#----------------------------------------------------------------------

sub get_process_def {
    my ($self, %params) = @_;
    my $process = $params{process};
    my $process_def = $process_config->{processes}->{$process};

    # !!! before anything else !!! - split planes values override regular values
    if ($self->{asm_split_planes} && $process_def->{split_planes}) {
        %$process_def = (%$process_def, %{$process_def->{split_planes}});
    }
    return $process_def;
}

#----------------------------------------------------------------------

sub start_process_group_by_module {
    my($self, %modules) = @_;

    my $logger = $self->{logger};

    my %started_by_module;
    for my $module (keys %modules) {
        my $group = 'startup_' . $module;
        for my $process (@{$process_config->{process_groups}->{$group}->{processes}}) {
            if ($started_by_module{$process}) {
                $logger->info("Process '$process' is already started for module '" . $started_by_module{$process} . "'");
            }
            else {
                $logger->info("Starting process '$process' for module '$module'");
                $started_by_module{$process} = $module;
                $self->start_process($process);
            }
        }
    }
}

#----------------------------------------------------------------------

sub start_process_group {
    my($self, $group) = @_;

    my $logger = $self->{logger};

    $logger->info("Starting process group $group");
    for my $process (@{$process_config->{process_groups}->{$group}->{processes}}) {
        $self->start_process($process);
    }
}

#----------------------------------------------------------------------

sub start_process {
    my($self, $process) = @_;

    my $logger = $self->{logger};

    my $p_pid = $$; # remember parent pid
    F5::SystemUtils::daemonize(); # create grandchild

    if ($p_pid != $$) { # grandchild only does this
        my $grandchild_pid = $$;

        my $process_def = $self->get_process_def(process => $process);

        # set maximum core file size limit
        if (my $size = $process_def->{max_core_size}) {
            if ($size eq 'unlimited') {
                my $vector = Bit::Vector->new_Hex(64,RLIM_INFINITY);
                $size = $vector->to_Dec();
            }
            $logger->debug("setting max core file size:[$size] for pid:[$grandchild_pid] proc:[$process]");
            my $rlim_t = pack "L!L!",$size,$size;
            syscall(&SYS_setrlimit,&RLIMIT_CORE,$rlim_t);
        }

        # set maximum opened file handles limit
        if (my $files = $process_def->{max_files_opened}) {
            $logger->debug("setting max opened files:[$files] for pid:[$grandchild_pid] proc:[$process]");
            my $rlim_t = pack "L!L!",$files,$files;
            syscall(&SYS_setrlimit,&RLIMIT_NOFILE,$rlim_t);
        }

        # set maximum stack size limit
        if (my $stack = $process_def->{max_stack_size}) {
            $logger->debug("setting max stack size:[$stack] for pid:[$grandchild_pid] proc:[$process]");
            my $rlim_t = pack "L!L!",$stack,$stack;
            syscall(&SYS_setrlimit,&RLIMIT_STACK,$rlim_t);
        }

        # redirect STDOUT to file
        if ($process_def->{stdout_to_file}) {
            $logger->debug("redirecting STDOUT to file:[".$process_def->{stdout_to_file}."] for pid:[$grandchild_pid] proc:[$process]");
            open OUTPUT, '>>', $process_def->{stdout_to_file} or die $!;
            STDOUT->fdopen( \*OUTPUT, 'w' ) or die $!;
        }

        # set umask
        if ($process_def->{mask}) {
            $logger->debug("setting umask:[".$process_def->{mask}."] for pid:[$grandchild_pid] proc:[$process]");
            umask($process_def->{mask});
        }

        # set environment variables
        if ($process_def->{env}) {
            for my $k (keys %{$process_def->{env}}) {
                if ($process_def->{env}->{$k} eq '') {
                    $logger->debug("deleting env variable:[$k] for pid:[$grandchild_pid] proc:[$process]");
                    delete $ENV{$k};
                }
                else {
                    $logger->debug("setting env variable:[$k] value:[".$process_def->{env}->{$k}."] for pid:[$grandchild_pid] proc:[$process]");
                    $ENV{$k} = $process_def->{env}->{$k};
                }
            }
        }

        $process_def->{save_pid_file} = 1 unless defined $process_def->{save_pid_file};
        # write .pid file
        if ($process_def->{save_pid_file}) {
            my $pid_file = $process_def->{pid_file} || PID_FILE_FOLDER . '/' . $process . '.pid';
            $logger->debug("setting pid file:[$pid_file] for pid:[$grandchild_pid] proc:[$process]");
            open(my $outf, '>', $pid_file);
            print $outf $grandchild_pid;
            close $outf;
        }

        # update cfg files (.ini style)
        if (my $cfg = $process_def->{cfg}) {
            my ($section,$key,$val) = split(',',$cfg->{update});
            $logger->debug("setting section:[$section] key:[$key] val:[$val] in file:[".$cfg->{file}."] for pid:[$grandchild_pid] proc:[$process]");
            F5::Cfg::cfg_set_config_item($cfg->{file}, $section, $key, $val);
        }

        # set nice value
        my $nice = $process_def->{nice};
        if (defined $nice) {
            $logger->debug("setting nice:[$nice] for pid:[$grandchild_pid] proc:[$process]");
            setpriority(0, $grandchild_pid, $nice);
        }

        # pin process to cpu (set affinity)
        if (my $affinity = $process_def->{affinity}) {
            my $vector = Bit::Vector->new_Hex($self->{bigstart_values}->{get_number_physical_cpu},0);
            $vector->bit_flip($self->{bigstart_values}->{last_cpu_id});

            my $taskset_mask = ( $affinity eq 'last' )
                ? '0x'. $vector->to_Hex() : ALL_CPUS_AFFINITY ;

            $logger->debug("setting affinity:[$taskset_mask] for pid:[$grandchild_pid] proc:[$process]");
            system(TASKSET_EXEC, '-p', $taskset_mask, $grandchild_pid); #doesn't spawn a shell
        }

        # run the actual job
        if (my $func = $process_def->{start_func}) {
            no strict 'refs';
            $self->$func();
            exit(1); # something went wrong - should never get here
        }
        elsif (my $path = $process_def->{path}) {
            $path = F5::Deploy::map_path($path);
            # 'exec' is the default execution method, other can be supported as well
            my $method = ($process_def->{exec_method}) ? $process_def->{exec_method} : 'exec';

            if ($method eq 'exec') {
                exec($path);
                exit(1); # something went wrong - should never get here
            }
            elsif ($method eq 'system') {
                # note that the 'setpriority', affinity and '$ENV' manipulation above DO take affect
                #    on the process that we are about to system to the background (I was surprised)...
                system("$path &");
                exit(1); # you should get here but YOU SHALL NOT PASS!!!

                # Using 'system' in a double fork (grandchild) created a few un-expected issues:
                # (1): after system MUST exit immediatelly. otherwise get this error - 'An instance (pid dddd) is already running'
                #          where 'dddd' is the pid of the current grandchild.
                # (2): the 'system &' results in a new (other than $$) pid but could not write it to '.pid' file
                #          because MUST exit immediatelly after system. (see previous note)
            }
        }
    } # grandchild scope

    # the original parent is the only process that leaves this subroutine in one piece
}

#----------------------------------------------------------------------

sub kill_process_group {
    my($self, $group) = @_;

    my $logger = $self->{logger};

    if ($group eq 'all') {
        $self->kill_ts_extra();
    }

    $logger->info("Killing process group $group");
    my @target_proc_names;
    my $is_bd_agent_killed = 0;
    for my $process (@{$process_config->{process_groups}->{$group}->{processes}}) {
        $is_bd_agent_killed = 1 if $process eq 'bd_agent';
        push(@target_proc_names, $self->get_process_names($process));
    }

    $self->do_kill(@target_proc_names);

    if ($is_bd_agent_killed) {
        F5::ASMReady::remove_bd_agent_config_end_lock_file();
    }
}

#----------------------------------------------------------------------

sub kill_process {
    my($self, $process) = @_;

    my $logger = $self->{logger};

    $logger->info("Killing process $process");
    $self->do_kill($self->get_process_names($process));

    if ($process eq 'bd_agent') {
        F5::ASMReady::remove_bd_agent_config_end_lock_file();
    }
}

#----------------------------------------------------------------------

sub get_process_names {
    my($self, $process) = @_;
    my $process_def = $process_config->{processes}->{$process};
    if ($process_def->{process_names}) {
        return @{$process_def->{process_names}};
    }

    return F5::Deploy::map_path($process_def->{path});
}

#----------------------------------------------------------------------

sub do_kill {
    my($self, @target_proc_names) = @_;

    my $logger = $self->{logger};
    my %target_proc_names = map { (basename($_) => 1) } @target_proc_names;

    my $kill_signal = 'TERM';
    my $start_time = time;
    my $last_kill_time = 0;

    while (1) {
        my $procs = F5::ProcessTable::get_process_table();
        my @kill_pids;

        for my $pid (keys %$procs) {
            if ($procs->{$pid}->{basename} && $target_proc_names{$procs->{$pid}->{basename}}) {
                $logger->debug("Adding $pid to kill list ($procs->{$pid}->{cmdline})");
                push(@kill_pids, $pid);
            }
        }

        if (! @kill_pids) {
            $logger->debug("All processes have been killed");
            last;
        }

        if (time - $start_time > SOFT_KILL_TIME_OUT) {
            $logger->debug("Switching to KILL signal");
            $kill_signal = 'KILL';
        }

        elsif (time - $start_time > HARD_KILL_TIME_OUT) {
            $logger->info("Couldn't kill - giving up");
            last;
        }

        if (time - $last_kill_time > KILL_PROC_INTERVAL) {
            $logger->debug("Sending kill signal $kill_signal");
            kill($kill_signal, @kill_pids);
            $last_kill_time = time;
        }
        sleep(CHECK_PROC_INTERVAL);
    }

    $logger->debug("Kill processes complete");
    return;
}

#----------------------------------------------------------------------

sub start_bd {
    my($self) = @_;

    my $logger = $self->{logger};

    my $MIN_ALLOCATED_MEMORY = 300 * 1024 * 1000; # this is bd default
    my $MED_ALLOCATED_MEMORY = 450 * 1024 * 1024;
    my $MAX_ALLOCATED_MEMORY = 1024 * 1024 * 1024;
    my $LARGE_MEMORY_THRESHOLD_IN_KB = 16438928;
    my $ULTRA_LARGE_MEMORY_THRESHOLD_IN_KB = 33554432;
    my $MAX_ACCOUNTS_MEM_THRESHOLD = 1677722; # 1.6 * 1024 * 1024 (rounded up); # 1.6 GB
    my $HIGH_END_ACCOUNTS_THRESHOLD = 1024;
    my $LOW_END_ACCOUNTS_THRESHOLD = 250;
    my $MIN_MALLOC_ARENA_POOLS = 4;
    my $MAX_MALLOC_ARENA_POOLS = 8;
    my $MALLOC_ARENA_POOLS_THRDS_THRESHOLD = 12;

    my $threads = $self->{bigstart_values}->{get_number_physical_cpu};
    my $tmm_count = $self->{bigstart_values}->{get_tmm_count};

    # this happens when the bigstart 'split_planes' value is true, regardless of the 'asm_split_planes' value
    if ($self->{bigstart_values}->{split_planes}) {
        my $nice = F5::Cluster::Utils::get_mcp_val('tmplugin.splitplanes.nice');
        $ENV{TMPLUGIN_NICE} = $nice;
        $logger->debug("setting [ENV{TMPLUGIN_NICE} = $nice]");
    }

    my $avail_memory = $self->{bigstart_values}->{get_total_memory};
    my $xml_memory = 0;
    if ( $avail_memory >= $ULTRA_LARGE_MEMORY_THRESHOLD_IN_KB ){
	$xml_memory = $MAX_ALLOCATED_MEMORY;
    }
    elsif ( $avail_memory >= $LARGE_MEMORY_THRESHOLD_IN_KB ) {
	$xml_memory = $MED_ALLOCATED_MEMORY;
    }
    else {
	$xml_memory = $MIN_ALLOCATED_MEMORY;
    }
	
    my $mhz = F5::NwdUtils::get_proc_cpuinfo_cpu_mhz();
    my $bd_memory = F5::NwdUtils::get_relevant_bd_memory(logger => $logger);

    # IMPORTANT :
    #    the same validation is done in 'tm_daemon/mcp/asm/validation/AsmPolicy.cpp' at 'AsmPolicyValidator::run'
    #    the validation logic has to be the same in both places
    my $max_accounts = ($bd_memory > $MAX_ACCOUNTS_MEM_THRESHOLD) ? $HIGH_END_ACCOUNTS_THRESHOLD : $LOW_END_ACCOUNTS_THRESHOLD ;

    # set this for later use by F5::Utils::Account::validate_create_account
    F5::Cfg::cfg_set_config_item(DCC_CFG_FILE, 'GENERAL', 'MaxAccounts', $max_accounts);

    my $scheduler = F5::Cluster::Utils::get_mcp_val('tmplugin.scheduler');
    $ENV{TMPLUGIN_SCHEDULER} = $scheduler;
    $logger->debug("setting [ENV{TMPLUGIN_SCHEDULER} = $scheduler]");

    $ENV{MALLOC_ARENA_MAX} = ($threads >= $MALLOC_ARENA_POOLS_THRDS_THRESHOLD) ? $MAX_MALLOC_ARENA_POOLS : $MIN_MALLOC_ARENA_POOLS;

    my $cmd = "./bd tmm_count $tmm_count proc_cpuinfo_cpu_mhz $mhz total_xml_memory  $xml_memory total_umu_max_size $bd_memory";

    if (F5::Provisioning::is_module_enabled("dos")) {
	$cmd .= ' dos_provisioned 1';
    }
    $logger->info("bd exec line is: [$cmd]");

    chdir(F5::Deploy::map_path('/usr/share/ts/bin'));
    exec(split(/\s+/,$cmd));
}

#----------------------------------------------------------------------

# used by 'kill_process_group' when group is 'all'
sub kill_ts_extra {
    # Remove gtm_vs_score objects
    if ( !(F5::Cluster::Utils::is_clustering_enabled() ) || (F5::Cluster::Utils::is_master_blade() ) ) {
        my %ASM_vs = ();
        F5::NwdUtils::UpdateGTMScore::clean_up(ASM_vs=>\%ASM_vs,log_in=>1);
    }

    # Remove all GUI session files
    find({
          wanted => sub {
              if (-f $File::Find::name &&
                  $File::Find::name =~ m|^/tmp/sess_|) {

                  unlink($File::Find::name);
              }
          },
          no_chdir => 1,
         },
         '/tmp');
}

#----------------------------------------------------------------------

sub restart_process {
    my($self, $process) = @_;

    $self->{logger}->debug("Restarting $process");
    $self->kill_process($process);
    $self->start_process($process);
}

#----------------------------------------------------------------------

sub wait_for_process {
    my($self, $process) = @_;

    if ($process eq 'bd_agent') {
        F5::ASMReady::wait_for_bd_agent_config_end(logger => $self->{logger});
    }
}

#----------------------------------------------------------------------

sub handle_actions {
    my($self, @process_actions) = @_;

    for my $process_action (@process_actions) {
        my($action, $process) = split(/ /, $process_action, 2);

        $action .= '_process';

        $self->$action($process);
    }
}

#----------------------------------------------------------------------

1;
