package F5::RequestLog;

use strict;
use warnings;

use File::Find;
use File::Basename;
use File::Path;
use List::Util;
use Time::HiRes;
use POSIX qw(strftime);
use Math::BigInt;
use Data::MessagePack;
use Data::MessagePack::Stream;

use F5::Deploy;
use F5::ASMConfig::Entity::BruteForceLoginPageStatus;
use F5::ASMConfig::Entity::Policy;
use F5::Provisioning;
use F5::BigIpLog;
eval {
  require bd::Main;
  require bot_defense_event_log::BotDefenseEventLog;
};
use F5::Utils::Rest;
use F5::Db::Partition;
use F5::BatchInsert;
use F5::LrnPipe;
use F5::IpAddr;
use F5::Violation::Utils;
use F5::DbClean;
use F5::ASMConfig::Client;
use F5::Cluster::Utils;
use F5::IO::LrnPipe;
use F5::IO::ASMConfig::EventListener;
use F5::IO::Selector;
use F5::Utils::File;
use F5::Cfg;
use F5::Db::Single;
use F5::DateTimeUtils;
use Encode qw(:all);
eval {
  require Encode::HanExtra;
};
use Thrift::BinaryProtocol;
use Thrift::MemoryBuffer;
use File::stat;
use YAML::XS;
use F5::CRC;
use F5::ASMConfig::Entity::Webhook;
use F5::ASMConfig::Entity::RequestLog;

use constant SIZEOF_INT => 4;

use constant CLEAN_DB_CFG_FILE  =>  F5::Deploy::map_path('/etc/ts/tools/asmlogd_clean_db.yaml');
use constant ASMLOGD_CFG_FILE  =>  F5::Deploy::map_path('/etc/ts/tools/asmlogd.cfg');
use constant WEBHOOKS_CFG_FILE  =>  F5::Deploy::map_path('/etc/ts/tools/webhooks.cfg');

# as defined in '../tm_app/bigstart/DiskProvisioning.pm'
# there's only one at this point, but in future there might be more
our $ASM_DATA_PARTITION = '/var/asmdata1';

my @ASM_PIPE_TYPES = qw(
                       RQL
                       PBRQL
                       BOTDEFENSERQL
                  );

my @DHD_PIPE_TYPES = qw(
                       BOTDEFENSERQL
                  );

my @BRUTE_FORCE_CLEANUP_TABLES = qw(
                                   PRX.BRUTE_FORCE_LOGIN_PAGE_STATUS
                                   PRX.BRUTE_FORCE_ATTACK_LOG
                                  );

my %BF_DB_TABLE_TO_REST_UNIQUE_FIELD = (
                                'PRX.BRUTE_FORCE_MITIGATED_USERNAMES' => 'username',
                                'PRX.BRUTE_FORCE_MITIGATED_IPS' => 'ip_address',
                                'PRX.BRUTE_FORCE_MITIGATED_DEVICE_IDS' => 'device_id',
                                'PRX.BRUTE_FORCE_LEAKED_CREDENTIALS' => 'username',
                               );

my %REPEATED_SUB_MESSAGE_TO_DB_TABLE = (
                                mitigated_usernames => 'PRX.BRUTE_FORCE_MITIGATED_USERNAMES',
                                mitigated_ips => 'PRX.BRUTE_FORCE_MITIGATED_IPS',
                                mitigated_device_ids => 'PRX.BRUTE_FORCE_MITIGATED_DEVICE_IDS',
                                leaked_credentials => 'PRX.BRUTE_FORCE_LEAKED_CREDENTIALS',
                               );

my %BOT_DEFENSE_REPEATED_FIELD_TO_DB_TABLE = (
                                anomaly_categories => { db_table => 'PRX.BOT_DEFENSE_ANOMALY_CATEGORIES',
                                                        db_field => 'anomaly_category',
                                                      },
                                anomalies => { db_table => 'PRX.BOT_DEFENSE_ANOMALIES',
                                               db_field => 'anomaly',
                                             },
                               );

my @MAIN_SUB_MESSAGES = qw(
                         attack_start_info
                         mitigation_stats
                        );

my @IGNORED_SUB_MESSAGES = qw(
                            attack_end_info
                           );

our $PULSE_MSG_INTERVAL = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'pulse_msg_interval') || 30;
our $PROTOBUF_FILE_PATH = $ASM_DATA_PARTITION . '/request_log';
our $PROTOBUF_FILE_INDIVIDUAL_MAX_SIZE = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'protobuf_file_individual_max_size');
our $PROTOBUF_FILE_TOTAL_MAX_SIZE = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'protobuf_file_total_max_size');
our $BATCH_INSERT_MAXIMUM_FLUSH_TIME = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'batch_insert_maximum_flush_time');
our $BATCH_INSERT_MAXIMUM_BATCH_SIZE = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'batch_insert_maximum_batch_size');
our $SESSION_TRACKING_MAXIMUM_FLUSH_TIME = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'session_tracking_maximum_flush_time');
our $SESSION_TRACKING_MAXIMUM_BATCH_SIZE = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'session_tracking_maximum_batch_size');
our $FLUSH_ON_NEXT_EMPTY_POLL = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'flush_on_next_empty_poll') || 0;
our $CONNECT_ATTEMPT_MINIMUM_INTERVAL = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'connect_attempt_minimum_interval') || 5;
our $LOCAL_INSERT_RATE_LIMIT_COUNT = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'local_insert_rate_limit_count');
our $LOCAL_INSERT_RATE_LIMIT_INTERVAL = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'local_insert_rate_limit_interval');
our $LOCAL_INSERT_BOT_DEFENSE_RATE_LIMIT_COUNT = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'local_insert_bot_defense_rate_limit_count');
our $LOCAL_INSERT_BOT_DEFENSE_RATE_LIMIT_INTERVAL = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'local_insert_bot_defense_rate_limit_interval');

our $PRIMARY_LOAD_FILE_INSERT_RATE = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'primary_load_file_insert_rate');
our $PRIMARY_LOADABLE_FILE_QUEUE_CACHE_INTERVAL = 60 * 60; # 1 hours

our $SECONDARY_CREATE_FILE_PATH = $ASM_DATA_PARTITION . '/cluster/request_log/create';
our $SECONDARY_TRANSFER_FILE_PATH = $ASM_DATA_PARTITION . '/cluster/request_log/transfer';
our $PRIMARY_ACK_TRANSFER_FILE_PATH = $ASM_DATA_PARTITION . '/cluster/request_log/ack';

our $SECONDARY_MAXIMUM_ROTATE_TIME = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'secondary_maximum_rotate_time');
our $SECONDARY_MAXIMUM_FILE_SIZE = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'secondary_maximum_file_size');
our $SECONDARY_RATE_LIMIT_COUNT = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'secondary_rate_limit_count');
our $SECONDARY_RATE_LIMIT_INTERVAL = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'secondary_rate_limit_interval');
our $SECONDARY_BOT_DEFENSE_RATE_LIMIT_COUNT = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'secondary_bot_defense_rate_limit_count');
our $SECONDARY_BOT_DEFENSE_RATE_LIMIT_INTERVAL = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'secondary_bot_defense_rate_limit_interval');

our $DB_PARTITION_ROW_LIMIT = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'db_partition_row_limit') || 100_000;
our $DB_PARTITION_MAX_NUMBER = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'asmlogd', 'db_partition_max_number') || 100;

our $GET_PROTOBUF_LAST_FILE_ID = 0;
our $GET_PROTOBUF_FH;

our $WEBHOOK_BATCH_SEND_FLUSH_TIME = F5::Cfg::cfg_get_config_item(WEBHOOKS_CFG_FILE, 'webhooks', 'webhook_batch_send_flush_time') || 5;
our $WEBHOOK_NUMBER_OF_MESSAGES_IN_QUEUE = F5::Cfg::cfg_get_config_item(WEBHOOKS_CFG_FILE, 'webhooks', 'webhook_number_of_messages_in_queue') || 100;

my @CLUSTER_REQUEST_LOG_FILE_PATHS = ($SECONDARY_CREATE_FILE_PATH,
                                      $SECONDARY_TRANSFER_FILE_PATH,
                                      $PRIMARY_ACK_TRANSFER_FILE_PATH
                                     );

my $policies_webhook_event_hash = {};
my @POLICY_WEBHOOK_FIELDS = keys %{F5::ASMConfig::Entity::Policy->webhook_fields};

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;

    my $self = bless({}, $class);
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;
    $self->{dbh} = $params{dbh};
    $self->{logger} = $params{logger};

    $self->check_pipes_provisioning;
    $self->update_cluster_state;
}

#----------------------------------------------------------------------

sub run_insert {
    my($self) = @_;

    my $last_flush = time;
    my $last_pulse_time = time;
    my $last_session_tracking_flush = time;
    my $last_webhooks_flush = time;

    $self->{webhooks_logs} = ();
    $self->{number_of_throttled_messages} = 0;

    while (1) {

        my $poll_time = undef;

        my @potential_poll_times;
        if ($self->{current_primary_load_file}) {
            push(@potential_poll_times, 0);
        }
        elsif ($self->{is_cluster_secondary} && $self->{current_secondary_file} &&
               $self->{current_secondary_file}->{file_size}) {

            push(@potential_poll_times,
                 F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $SECONDARY_MAXIMUM_ROTATE_TIME,
                                                                    start_time => $self->{current_secondary_file}->{start_time}));
        }
        elsif (my $oldest_record_time = $self->{batch_insert}->oldest_record_time) {
            if ($FLUSH_ON_NEXT_EMPTY_POLL) {
                push(@potential_poll_times, 0);
            }
            else {
                push(@potential_poll_times,
                     F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $BATCH_INSERT_MAXIMUM_FLUSH_TIME,
                                                                        start_time => $oldest_record_time));
            }
        }

        if ($SESSION_TRACKING_MAXIMUM_FLUSH_TIME) {
              push(@potential_poll_times,
                     F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $SESSION_TRACKING_MAXIMUM_FLUSH_TIME,
                                                                        start_time => $last_session_tracking_flush));
        }

        if ($PULSE_MSG_INTERVAL) {
            push(@potential_poll_times,
                 F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $PULSE_MSG_INTERVAL,
                                                                    start_time => $last_pulse_time));
        }
        
        for my $pipe_type (@{$self->{pipe_types}}) {
            if (! $self->{pipes}->{$pipe_type}) {
                my $time_remaining = F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $CONNECT_ATTEMPT_MINIMUM_INTERVAL,
                                                                                        start_time => $self->{last_pipe_connect_attempt_time}->{$pipe_type});

                if (! $time_remaining) {
                   $self->open_pipe(pipe_type => $pipe_type);
                    $self->{last_pipe_connect_attempt_time}->{$pipe_type} = time;
                    push(@potential_poll_times, $CONNECT_ATTEMPT_MINIMUM_INTERVAL);
                }
                else {
                    push(@potential_poll_times, $time_remaining);
                }
            }
        }

        if (! $self->{asmconfig_event_listener}) {
            my $time_remaining = F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $CONNECT_ATTEMPT_MINIMUM_INTERVAL,
                                                                                    start_time => $self->{last_asmconfig_connect_attempt_time});

            if (! $time_remaining) {
                $self->open_asmconfig_event_listener;
                $self->{last_asmconfig_connect_attempt_time} = time;
                push(@potential_poll_times, $CONNECT_ATTEMPT_MINIMUM_INTERVAL);
            }
            else {
                push(@potential_poll_times, $time_remaining);
            }
        }

        if (@potential_poll_times) {
            $poll_time = List::Util::min(@potential_poll_times);
        }

        $self->{logger}->trace("About to poll with max time-out of: $poll_time") if $self->{logger};
        my @readers = $self->{selector}->can_read($poll_time);

        if ($self->{is_cluster_secondary}) {
            $self->rotate_secondary_file;
        }

        for my $reader (@readers) {
            $reader->handle_read;
        }

        if ($self->{current_primary_load_file}) {
            $self->load_primary_from_current_secondary;
        }

        if ($self->{batch_insert}->oldest_record_time &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $BATCH_INSERT_MAXIMUM_FLUSH_TIME,
                                                               start_time => $self->{batch_insert}->oldest_record_time) == 0 ||
            (! @readers && $FLUSH_ON_NEXT_EMPTY_POLL)) {

            $self->{logger}->debug("Flushing to database") if $self->{logger};
            $self->{batch_insert}->flush;
        }

        if ($SESSION_TRACKING_MAXIMUM_FLUSH_TIME &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $SESSION_TRACKING_MAXIMUM_FLUSH_TIME,
                                                               start_time => $last_session_tracking_flush) == 0) {
            if ($self->{sa_dp_count}) {
                $self->{logger}->debug("Flushing session tracking data") if $self->{logger};
                $self->update_session_tracking;
            }
            $last_session_tracking_flush = time;
        }

        if ($PULSE_MSG_INTERVAL && 
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $PULSE_MSG_INTERVAL,
                                                               start_time => $last_pulse_time) == 0) {

            $self->log_pulse_message;
            $self->check_pipes_provisioning;
            $last_pulse_time = time;
        }

        if ($WEBHOOK_BATCH_SEND_FLUSH_TIME &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $WEBHOOK_BATCH_SEND_FLUSH_TIME,
                                                               start_time => $last_webhooks_flush) == 0) {

            $self->{logger}->debug("Flushing logs messages of webhooks") if $self->{logger};
            $self->send_webhook_events;
            $last_webhooks_flush = time;
        }
    }

    # Should never reach here
    $self->{logger}->warn("Unexpectedly ended") if $self->{logger};
}

#----------------------------------------------------------------------

sub check_pipes_provisioning {
    my($self) = @_;

    if (F5::Provisioning::is_module_enabled("asm")) {
        $self->{logger}->debug("ASM is enabled, will start asm pipes") if $self->{logger};
        $self->{pipe_types} = \@ASM_PIPE_TYPES;
    }
    elsif (F5::Provisioning::is_module_enabled("dos")) {
        $self->{logger}->debug("DOS is enabled, will start dos pipes") if $self->{logger};
        $self->{pipe_types} = \@DHD_PIPE_TYPES;
    }
    else {
        $self->{logger}->debug("ASM and DOS are disabled, not starting pipes") if $self->{logger};
        $self->{pipe_types} = [];
    }
}

#----------------------------------------------------------------------

sub update_session_tracking {
    my ($self, %params) = @_;
    # the sa data point that we are about to add to persistence should already
    #    exist in session_db. bd was supposed to place it there first.
    # we should not add if this is not the case.
    # sa object to be used in add and search below

    my $asm_config_client;
    if ($self->{asmconfig_event_listener}) {
        $asm_config_client = $self->{asmconfig_event_listener}->get_client;
    }
    return unless $asm_config_client;

    for my $policy_id (keys %{$self->{sa_dp}}) {
        for my $scope (keys %{$self->{sa_dp}->{$policy_id}}) {
            my @values = keys %{$self->{sa_dp}->{$policy_id}->{$scope}};
            next unless scalar @values;
            my $call_flags = F5::ASMConfig::CallFlags->new({
                                                   is_asynchronous => 1,
                                                   ignored_errors => [
                                                                       F5::ASMConfig::IgnoredError->new({
                                                                                                          error_code => F5::ASMConfig::ErrorCode::KNOWN_DB_ERRORS,
                                                                                                        })
                                                                     ],
                                                   search_filters => [
                                                                       F5::ASMConfig::SearchFilter->new( field_name => 'data_point_value',
                                                                                                         field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN,
                                                                                                         field_value_list => \@values,
                                                                                                       ),
                                                                     ],
                                                 });
             my $sa_dp_obj = F5::ASMConfig::SessionAwarenessDataPoint->new({
                   data_point_action => F5::ASMConfig::SessionAwarenessAction::SESSION_ACTION_BLOCK_ALL,
                   data_point_scope => $scope,
                  });
             my $store_sa_dp_obj = F5::ASMConfig::SessionAwarenessDataPoint->new({
                   is_persistent => 1,
                  });

             eval {
                $asm_config_client->update_many_session_awareness_data_points($asm_config_client->{session},
                                                                              $call_flags,
                                                                              F5::ASMConfig::Policy->new({ policy_id => $policy_id }),
                                                                              $sa_dp_obj,
                                                                              $store_sa_dp_obj);
            };
            if ($@) {
                $self->{logger}->error("Session Awareness records could not be added.");
            }
        }
    }
    $self->{sa_dp} = {};
    $self->{sa_dp_count} = 0;
}

#----------------------------------------------------------------------

sub log_pulse_message {
    my($self) = @_;

    my @additional_info;
    my @additional_bot_defense_info;
    my @additional_attack_id_info;
    my @additional_login_object_info;
    my @additional_pipe_info;

    # last handled records of each type
    if ($self->{last_support_id_handled}) {
        push(@additional_info, "last support id handled = $self->{last_support_id_handled}");
    }
    if ($self->{last_bot_defense_support_id_handled}) {
        push(@additional_bot_defense_info, "last Bot Defense support id handled = $self->{last_bot_defense_support_id_handled}");
    }
    if ($self->{last_attack_id_handled}) {
        push(@additional_attack_id_info, "last attack id handled = $self->{last_attack_id_handled}");
    }
    if ($self->{last_login_object_handled}) {
        push(@additional_login_object_info, "last login object handled = $self->{last_login_object_handled}");
    }

    # per pipe type, records handled counters, pre-rate-limiting
    for my $pipe_type (@{$self->{pipe_types}}) {
        if ($self->{"pipe_" . $pipe_type . "_requests_handled_count"}) {
            push(@additional_pipe_info, $self->{"pipe_" . $pipe_type . "_requests_handled_count"} . " records handled from $pipe_type pipe type");
        }
    }

    # on primary, from all pipes, post-rate-limiting
    if ($self->{local_stored_count}) {
        push(@additional_pipe_info, "locally stored $self->{local_stored_count} requests from all pipe types");
        if ($self->{pb_handled_count}) {
            push(@additional_pipe_info, "of which $self->{pb_handled_count} requests were from Policy Builder Traffic Learning");
        }
    }

    # on secondary, from all pipes, post-rate-limiting
    if ($self->{secondary_handled_count}) {
        push(@additional_pipe_info, "wrote $self->{secondary_handled_count} requests to secondary sync files from all pipe types");
    }

    # on primary, from all secondaries
    #     loading these is not rate-limited on primary
    #     probably since they were already rate-limited on the secondaries
    if ($self->{load_from_secondary_count}) {
        push(@additional_info, "stored $self->{load_from_secondary_count} requests from secondary sync files");
    }

    # rate limited records count of each type
    if ($self->{rate_limit_skipped_count}) {
        push(@additional_info, "skipped $self->{rate_limit_skipped_count} requests due to rate limit");
    }
    if ($self->{rate_limit_bot_defense_skipped_count}) {
        push(@additional_bot_defense_info, "skipped $self->{rate_limit_bot_defense_skipped_count} Bot Defense requests due to rate limit");
    }

    my $additional_info = '';
    if (@additional_info) {
        $additional_info = '(' . join(', ', @additional_info) . ')';
    }

    my $additional_bot_defense_info = '';
    if (@additional_bot_defense_info) {
        $additional_bot_defense_info = '(' . join(', ', @additional_bot_defense_info) . ')';
    }

    my $additional_attack_id_info = '';
    if (@additional_attack_id_info) {
        $additional_attack_id_info = '(' . join(', ', @additional_attack_id_info) . ')';
    }

    my $additional_login_object_info = '';
    if (@additional_login_object_info) {
        $additional_login_object_info = '(' . join(', ', @additional_login_object_info) . ')';
    }

    my $additional_pipe_info = '';
    if (@additional_pipe_info) {
        $additional_pipe_info = '(' . join(', ', @additional_pipe_info) . ')';
    }

    $self->{logger}->info("$self->{requests_handled_count} request records handled since startup $additional_info") if $self->{logger};
    $self->{logger}->info("$self->{bot_defense_events_handled_count} Bot Defense records handled since startup $additional_bot_defense_info") if $self->{logger};
    $self->{logger}->info("$self->{attacks_handled_count} attack records handled since startup $additional_attack_id_info") if $self->{logger};
    $self->{logger}->debug("$self->{login_objects_handled_count} login object records handled since startup $additional_login_object_info") if $self->{logger};
    $self->{logger}->debug("$self->{all_pipe_messages_count} pipe messages handled since startup $additional_pipe_info") if $self->{logger};

}

#----------------------------------------------------------------------

sub insert_record {
    my($self, %params) = @_;

    my $raw_protobuf = $params{raw_protobuf};

    my $rec;
    if ($self->check_is_bot_defense_request(%params)) {
        my $bot_defense_event_log_protobuf = $self->{bot_defense_event_log_protobuf};
        $bot_defense_event_log_protobuf->unpack($raw_protobuf);

        my $rec = $bot_defense_event_log_protobuf->to_hashref();
        $self->insert_bot_defense_record(%params,
                                         rec => $rec,
                                        );
    }
    else {
        my $bd_protobuf = $self->{bd_protobuf};
        $bd_protobuf->unpack($raw_protobuf);

        my $rec = $bd_protobuf->to_hashref();
        $self->insert_bd_record(%params,
                                rec => $rec,
                               );
    }
}

#----------------------------------------------------------------------

sub insert_bot_defense_record {
    my($self, %params) = @_;

    my $bot_defense_record = $params{rec};
    $self->{logger}->trace($bot_defense_record) if $self->{logger};

    $self->{bot_defense_events_handled_count}++;
    $self->{last_bot_defense_support_id_handled} = $bot_defense_record->{support_id};

    if ($bot_defense_record->{is_update_request}) {
        $self->{logger}->debug("Updating Bot Defense event with support_id = $bot_defense_record->{support_id}") if $self->{logger};
        $self->{batch_insert}->update_row(table_name => 'PRX.BOT_DEFENSE_EVENT_LOG',
                                          rec => $bot_defense_record,
                                          fields => [ qw(challenge_failure_reason) ],
                                          where_fields => { support_id => $bot_defense_record->{support_id} },
                                         );
        return;
    }

    $self->{last_bot_defense_event_log_id}++;

    # mobile_debugger_enabled_device, as received from protobuf message, is a string one of: 'Yes', 'No', 'N/A'
    #   it is represented, in DB and APIs, as a string of equal value
    $bot_defense_record->{mobile_device_debugger_mode} = $bot_defense_record->{mobile_debugger_enabled_device};

    # mobile_debugger_enabled_device, as represented in DB and APIs, is a boolean,
    #   based on it's string value, as received from protobuf message
    $bot_defense_record->{mobile_debugger_enabled_device} = $bot_defense_record->{mobile_debugger_enabled_device} eq 'Yes' ? 1 : 0 ;

    # mobile_is_human_behaviour, as received from protobuf message, is a string one of: 'Yes', 'No', 'N/A'
    #   it is represented, in DB and APIs, as a string of equal value
    $bot_defense_record->{mobile_human_behavior_mode} = $bot_defense_record->{mobile_is_human_behaviour};

    # mobile_is_human_behaviour, as represented in DB and APIs, is a boolean,
    #   based on it's string value, as received from protobuf message
    $bot_defense_record->{mobile_is_human_behaviour} = $bot_defense_record->{mobile_is_human_behaviour} eq 'Yes' ? 1 : 0 ;

    $bot_defense_record->{bot_defense_event_log_internal_id} = $self->{last_bot_defense_event_log_id};
    $bot_defense_record->{client_ip} = $self->ip_address_as_string(ip_field => $bot_defense_record->{client_ip});
    $bot_defense_record->{dest_ip} = $self->ip_address_as_string(ip_field => $bot_defense_record->{dest_ip});
    $bot_defense_record->{device_id} = $self->session_id_as_hex(session_id => $bot_defense_record->{device_id}) || '';
    $bot_defense_record->{session_id} = $self->session_id_as_hex(session_id => $bot_defense_record->{session_id}) || '';

    my $partition = "Common";
    my $virtual_server_name = $bot_defense_record->{virtual_server_name};
    if ($virtual_server_name && $virtual_server_name =~ m|/(\w+)/|) {
        $partition = $1;
    }
    $bot_defense_record->{partition_id} = $partition;

    for my $repeated_field (keys %BOT_DEFENSE_REPEATED_FIELD_TO_DB_TABLE) {
        my $db_field = $BOT_DEFENSE_REPEATED_FIELD_TO_DB_TABLE{$repeated_field}->{db_field};
        delete $bot_defense_record->{$db_field};
    }

    $self->{logger}->debug("Logging Bot Defense event with support_id = $bot_defense_record->{support_id}") if $self->{logger};

    $self->{batch_insert}->insert_row(table_name => 'PRX.BOT_DEFENSE_EVENT_LOG',
                                      do_insert_ignore => 1,
                                      rec => $bot_defense_record);

    push(@{$self->{recent_bot_defense_insert_times}}, { insert_time => Time::HiRes::time(),
                                                      });

    for my $repeated_field (keys %BOT_DEFENSE_REPEATED_FIELD_TO_DB_TABLE) {
        my $db_table = $BOT_DEFENSE_REPEATED_FIELD_TO_DB_TABLE{$repeated_field}->{db_table};
        my $db_field = $BOT_DEFENSE_REPEATED_FIELD_TO_DB_TABLE{$repeated_field}->{db_field};
        if ($bot_defense_record->{$repeated_field} && ref($bot_defense_record->{$repeated_field}) eq 'ARRAY') {
            for my $repeated_field_value (@{$bot_defense_record->{$repeated_field}}) {
                my $repeated_record = {};
                $repeated_record->{bot_defense_event_log_internal_id} = $self->{last_bot_defense_event_log_id};
                $repeated_record->{$db_field} = $repeated_field_value;
                $self->{batch_insert}->insert_row(table_name => $db_table,
                                                  rec => $repeated_record,
                                                  do_insert_ignore => 1,
                                                 );
            }
        }
    }
}

#----------------------------------------------------------------------

sub insert_bd_record {
    my($self, %params) = @_;

    my $bd_rec = $params{rec};
    $self->{logger}->trace($bd_rec) if $self->{logger};

    my $policy_id;
    if ( $bd_rec->{policy_name_crc} ) {
        $policy_id = $self->{value_map}->{policy_name_crc_to_policy_id}->{$bd_rec->{policy_name_crc}};
    }

    if (! $policy_id ) {
        $self->load_policy_mapping(policy_name_crc => $bd_rec->{policy_name_crc});
        $policy_id = $self->{value_map}->{policy_name_crc_to_policy_id}->{$bd_rec->{policy_name_crc}} if $bd_rec->{policy_name_crc};
    }
    if (! $policy_id ) {
        my $account_identifier = 'N/A';;
        if ( $bd_rec->{policy_name} ) {
            $account_identifier = "for policy '$bd_rec->{policy_name}'";
        }
        elsif ( $bd_rec->{policy_name_crc} ) {
            $account_identifier = "for policy name crc $bd_rec->{policy_name_crc}";
        }
        elsif ( $bd_rec->{data} && $bd_rec->{data}->{request} && $bd_rec->{data}->{request}->{vs_name}) {
            $account_identifier = "with virtual server '$bd_rec->{data}->{request}->{vs_name}'";
        }

        $self->{logger}->info("Policy mapping for account $account_identifier could not be loaded -- skipping record") if $self->{logger};
        return;
    }

    # according to Data types, as defined in 'message Data' at -
    #     TrafficShield/common/asm_proto_msg/bd.proto
    if ($bd_rec->{data}->{request} || $bd_rec->{data}->{response}) {
        $self->insert_request_log(%params,
                                  bd_rec => $bd_rec,
                                  policy_id => $policy_id,
                                 );
    }
    elsif ($bd_rec->{data}->{brute_force_attack}) {
        $self->insert_attack_log(%params,
                                 bd_rec => $bd_rec,
                                 policy_id => $policy_id,
                                );
    }
    elsif ($bd_rec->{data}->{brute_force_login_page_status}) {
        $self->insert_login_page_status(%params,
                                 bd_rec => $bd_rec,
                                 policy_id => $policy_id,
                                );
    }
    else {
        $self->{logger}->error("Unrecognized bd record type -- skipping record") if $self->{logger};
    }
}

#----------------------------------------------------------------------

sub insert_login_page_status {
    my($self, %params) = @_;
    my $bd_rec = $params{bd_rec};
    my $policy_id = $params{policy_id};

    $self->{login_objects_handled_count}++;
    my $login_page_status = $bd_rec->{data}->{brute_force_login_page_status};

    $self->{logger}->debug("Logging login page status of '$login_page_status->{login_object}'") if $self->{logger};
    $self->{last_login_object_handled} = $login_page_status->{login_object};

    $login_page_status->{policy_id} = $policy_id;

    my $security_policy_entity = F5::ASMConfig::Entity::Policy->new( object => {
                                                                                policy_name => $self->{value_map}->{policy_id_to_policy_name}->{$policy_id}
                                                                               }
                                                                   );

    my $login_page_status_entity = F5::ASMConfig::Entity::BruteForceLoginPageStatus->new( object => { 
                                                                                                      security_policy => $security_policy_entity,
                                                                                                      vs_name => $login_page_status->{vs_name},
                                                                                                      login_object_rest_uuid => $login_page_status->{login_object_rest_uuid},
                                                                                                    }
                                                                                        );

    $login_page_status->{rest_uuid} = $login_page_status_entity->get_rest_uuid();

    $self->{batch_insert}->insert_row(table_name => 'PRX.BRUTE_FORCE_LOGIN_PAGE_STATUS',
                                      rec => $login_page_status,
                                      do_replace => 1,
                                     );

    push(@{$self->{recent_insert_times}}, { insert_time => Time::HiRes::time(),
                                          });
}

#----------------------------------------------------------------------

my %BRUTE_FORCE_ATTACK_STATUS = (
    F5::ASMConfig::BruteForceAttackLogAttackStatus::BRUTE_FORCE_ATTACK_LOG_ATTACK_STATUS_STARTED() => "Started",
    F5::ASMConfig::BruteForceAttackLogAttackStatus::BRUTE_FORCE_ATTACK_LOG_ATTACK_STATUS_ENDED()  => "Finished",
);

my %BRUTE_FORCE_DETECTION_MODE = (
    F5::ASMConfig::BruteForceAttackLogDetectionMode::BRUTE_FORCE_ATTACK_LOG_DETECTION_MODE_NUMBER_OF_FAILED_LOGINS_INCREASED() => "Distributed Attack",
    F5::ASMConfig::BruteForceAttackLogDetectionMode::BRUTE_FORCE_ATTACK_LOG_DETECTION_MODE_EXTREME_NUMBER_OF_CREDENTIALS_STUFFING_DICTIONARY_MATCHES() => "Credentials Stuffing Attack",
    F5::ASMConfig::BruteForceAttackLogDetectionMode::BRUTE_FORCE_ATTACK_LOG_DETECTION_MODE_NUMBER_OF_FAILED_LOGINS_INCREASED_AND_EXTREME_NUMBER_OF_CREDENTIALS_STUFFING_DICTIONARY_MATCHES() => "Distributed and Credentials Stuffing Attack",
);

my %BRUTE_FORCE_MITIGATION_ACTION = (
    F5::ASMConfig::BruteForceAttackLogCurrentMitigation::BRUTE_FORCE_ATTACK_LOG_CURRENT_MITIGATION_ALARM() => "Alarm",
    F5::ASMConfig::BruteForceAttackLogCurrentMitigation::BRUTE_FORCE_ATTACK_LOG_CURRENT_MITIGATION_ALARM_AND_CAPTCHA() => "Alarm and CAPTCHA",
    F5::ASMConfig::BruteForceAttackLogCurrentMitigation::BRUTE_FORCE_ATTACK_LOG_CURRENT_MITIGATION_ALARM_AND_CLIENT_SIDE_INTEGRITY() => "Alarm and Client Side Integrity",
    F5::ASMConfig::BruteForceAttackLogCurrentMitigation::BRUTE_FORCE_ATTACK_LOG_CURRENT_MITIGATION_ALARM_AND_CLIENT_SIDE_INTEGRITY_CAPTCHA() => "Alarm and Client Side Integrity then CAPTCHA",
);

sub insert_attack_log {
    my($self, %params) = @_;
    my $bd_rec = $params{bd_rec};
    my $policy_id = $params{policy_id};

    my $brute_force_attack = $bd_rec->{data}->{brute_force_attack};

    $self->{logger}->debug("Logging attack_id = $brute_force_attack->{attack_id}") if $self->{logger};
    $self->{last_attack_id_handled} = $brute_force_attack->{attack_id};
    $self->{attacks_handled_count}++;

    my $repeated_info_by_db_table;
    for my $sub_message_kind (keys %REPEATED_SUB_MESSAGE_TO_DB_TABLE) {
        $repeated_info_by_db_table->{$REPEATED_SUB_MESSAGE_TO_DB_TABLE{$sub_message_kind}} = $brute_force_attack->{$sub_message_kind};
    }

    for my $sub_message_kind (@MAIN_SUB_MESSAGES) {
        for my $field (keys %{$brute_force_attack->{$sub_message_kind}}) {
            $brute_force_attack->{$field} = $brute_force_attack->{$sub_message_kind}->{$field};
        }
    }

    $brute_force_attack->{policy_id} = $policy_id;

    $self->{batch_insert}->insert_row(table_name => 'PRX.BRUTE_FORCE_ATTACK_LOG',
                                      rec => $brute_force_attack,
                                      do_replace => 1,
                                     );

    push(@{$self->{recent_insert_times}}, { insert_time => Time::HiRes::time(),
                                          });

    for my $db_table (values %REPEATED_SUB_MESSAGE_TO_DB_TABLE) {
        if ($repeated_info_by_db_table->{$db_table} && ref($repeated_info_by_db_table->{$db_table}) eq 'ARRAY') {
            for my $info_record (@{$repeated_info_by_db_table->{$db_table}}) {
                $info_record->{attack_id} = $brute_force_attack->{attack_id};
                $info_record->{rest_uuid} = F5::Utils::Rest::generate_uuid($info_record->{$BF_DB_TABLE_TO_REST_UNIQUE_FIELD{$db_table}});
                $self->{batch_insert}->insert_row(table_name => $db_table,
                                                  rec => $info_record,
                                                  do_replace => 1,
                                                 );
            }
        }
    }

    my $event_name = $BRUTE_FORCE_ATTACK_STATUS{$brute_force_attack->{attack_status}};
    if ($event_name) {
        my $policy_name = $self->{value_map}->{policy_id_to_policy_name}->{$policy_id};
        my $login_url = $brute_force_attack->{login_object};
        my $vs_name = $brute_force_attack->{vs_name};
        my $attack_type = $BRUTE_FORCE_DETECTION_MODE{$brute_force_attack->{anomaly_attack_type}} || 'N/A';
        my $mitigation_action = $BRUTE_FORCE_MITIGATION_ACTION{$brute_force_attack->{current_mitigation}} || 'N/A';

        my $msg = "Brute force attack: $event_name, Policy name: $policy_name, VS name: $vs_name, Object: $login_url, Attack type: $attack_type, Mitigation action: $mitigation_action";
        F5::BigIpLog::bigip_ts_ts_info_msg_log("[SECEV] $msg"); # /var/log/asm
    }
}

#----------------------------------------------------------------------

sub insert_request_log {
    my($self, %params) = @_;
    my $raw_protobuf = $params{raw_protobuf};
    my $bd_rec = $params{bd_rec};
    my $policy_id = $params{policy_id};

    my $protobuf_file_rec = $self->write_to_protobuf_file(protobuf_msg => $raw_protobuf);
    return if !$protobuf_file_rec;

    my $next_request_log_id = $self->{last_request_log_id} + 1;
    $self->{requests_handled_count}++;

    my $request = $bd_rec->{data}->{request} || {};
    my $response = $bd_rec->{data}->{response} || {};
    my $url = '';
    my $username = $request->{username} || '';
    my $request_schema = 0;
    if ($request->{url}) {
        $url = $request->{url}->{name} || '';
        $request_schema = $request->{url}->{traffic_app_id} || 0;
        if ($bd_rec->{is_websockets}) {
            if ($request_schema) {
                $request_schema = 4;
            }
            else {
                $request_schema = 3;
            }
        }
    }

    my $encoding = $self->{value_map}->{policy_id_to_encoding}->{$policy_id};
    
    if ($encoding && $encoding ne 'utf-8') {
         Encode::from_to($url, $encoding, 'utf-8');
         Encode::from_to($username, $encoding, 'utf-8');
    }

    my $matched_microservice = '';
    if ( $request->{microservice}->{host} && $request->{microservice}->{url} ){
        $matched_microservice =  $request->{microservice}->{url} =~ m/^\// ? "$request->{microservice}->{host}"."$request->{microservice}->{url}" :
                                                                             "$request->{microservice}->{host}"."/"."$request->{microservice}->{url}";
    }

    my $db_rec = {
                  'id' => $next_request_log_id,
                  'support_id' => $request->{support_id} || 0,
                  'previous_support_id' => $request->{previous_support_id} || 0,
                  'support_id_suffix' => substr($request->{support_id} || '', -4),
                  'policy_id' => $policy_id,
                  'src_ip' => $self->ip_address_as_string(ip_field => $request->{source_ip}),
                  'src_port' => $request->{src_port},
                  'dest_ip' => $self->ip_address_as_string(ip_field => $request->{dest_ip}),
                  'dest_port' => $request->{dest_port},
                  'request_schema' => $request_schema,
                  'log_time' => $request->{log_time},
                  'method_code' => $request->{method_code}, # todo: convert to method name when switching to request_log
                  'url' => $url,
                  'response_code' => $bd_rec->{data}->{response}->{status} || 0,
                  'geo_location_country_code' => $request->{geo_location},
                  'attack_types_mask' => $request->{attack_types_status},
                  'flg_viewed' => 0,
                  'username' => $username,
                  'session_id' => $self->session_id_as_hex(session_id => $request->{session_id}) || '',
                  'device_id' => $self->session_id_as_hex(session_id => $request->{device_id}) || '',
                  'flg_req_blocked' => $request->{flg_req_blocked} || 0,
                  'flg_req_truncated' => $request->{raw_req_data}->{flag_req_truncated} || 0,
                  'is_unblock_request' => $request->{is_unblock_request} || 0,
                  'slot_number' => $request->{slot_number} || 0,
                  'iprep_threat' => $request->{iprep_threat} || 0,
                  'severity_id' => $request->{severity_id} || 0,
                  'has_violations' => has_violations($request->{enforced_violations}->{viol_set}),
                  'viol_set_enforce' => $request->{enforced_violations}->{viol_set},
                  'viol_set_staging' => $request->{tightening_status_code}->{viol_set},
                  'violation_rating' => $request->{violation_rating} || 0,
                  'response_exclusion_reason' => $response->{exclusion_reason} || 0,
                  'flg_resp_compressed' => $response->{is_compressed} || 0,
                  'flg_resp_truncated' => $response->{is_truncated} || 0,
                  'protobuf_file_seek_pos' => $protobuf_file_rec->{seek_pos},
                  'protobuf_file_id' => $protobuf_file_rec->{file_id},
                  'protobuf_record_size' => $protobuf_file_rec->{record_size},
                  'flg_display' => $bd_rec->{should_log_locally},
                  'suggestion_keys' => $params{suggestion_keys},
                  'host_header' => $request->{host_header} || '',
                  'virtual_server_name' => $request->{vs_name} || '',
                  'microservice_url' => $request->{microservice}->{url} || '',
                  'microservice_host_name' => $request->{microservice}->{host} || '',
                  'matched_microservice' => $matched_microservice,
                  'challenge_failure_reason' => $request->{challenge_failure_reason} || '1',
                  'enforcement_action' => $request->{enforcement_action} || '1',
                  'enforced_by' => $request->{enforced_by} || '1',
                  'request_status' => $request->{request_status} || '1',
                  'sig_cves' => (defined $request->{sig_cves}) ? join(',', @{$request->{sig_cves}}) : '',
                  'operation_id' => $request->{url}->{operation_id} || '',
                  'flg_has_likely_false_positive_violations' => $request->{has_likely_false_positive_violations},
                 };

    $self->{last_support_id_handled} = $db_rec->{support_id};

    if ($bd_rec->{data}->{request}->{is_update_request}) {
        $self->{logger}->debug("Updating event with support_id = $db_rec->{support_id}") if $self->{logger};
        $self->{batch_insert}->update_row(table_name => 'PRX.REQUEST_LOG',
                                          rec => $db_rec,
                                          fields => [ qw(challenge_failure_reason) ],
                                          where_fields => { support_id => $db_rec->{support_id} },
                                         );
        return;
    }

    $self->{last_request_log_id} = $next_request_log_id;
    $self->adjust_partitions(next_request_log_id => $next_request_log_id);

    if ($bd_rec->{data}->{request}->{sa_lock_scopes_bitmask}) {
        $self->add_session_awareness_lock( bd_rec => $bd_rec,);
    }

    $self->{logger}->debug("Logging support_id = $db_rec->{support_id}") if $self->{logger};
    $self->{batch_insert}->insert_row(table_name => 'PRX.REQUEST_LOG',
                                      do_insert_ignore => 1,
                                      rec => $db_rec);

    $self->handle_webhook_request_log(db_rec => $db_rec);

    push(@{$self->{recent_insert_times}}, { insert_time => Time::HiRes::time(),
                                          });
}

#----------------------------------------------------------------------

sub send_webhook_events {
    my($self, %params) = @_;

    my $asm_config_client;
    if ($self->{asmconfig_event_listener}) {
        $asm_config_client = $self->{asmconfig_event_listener}->get_client;
    }

    if (!$asm_config_client) {
        $self->{logger}->info("Could not connect to asmconfig") if $self->{logger};
        return;
    }

    my @element_list = ();
    my $request_webhooks_fields = F5::ASMConfig::Entity::RequestLog->webhook_fields();
    for my $db_rec (@{$self->{webhooks_logs}}) {
        my %request_log_obj = map { $_ => $db_rec->{$request_webhooks_fields->{$_}->{db_field_name}} } keys %$request_webhooks_fields;
        my $request_log = F5::ASMConfig::RequestLog->new(%request_log_obj,
                                                         request_log_id => $db_rec->{id}->bstr(),
                                                         log_time => $db_rec->{log_time},
                                                         enforcement_state_rating => $db_rec->{violation_rating},
                                                         request_policy => $db_rec->{policy});

        my $element = F5::ASMConfig::Element->new({ element_type => F5::ASMConfig::ElementType::REQUEST_LOG,
                                                    request_log => $request_log});
        push(@element_list, $element);
    }

    my $total_record_count = @element_list;
    if (!$total_record_count) {
        return;
    }

    $self->{webhooks_logs} = ();

    my $search_result = F5::ASMConfig::SearchResult->new(total_record_count => $total_record_count,
                                                          element_list => \@element_list);
    my $call_flags = F5::ASMConfig::CallFlags->new(is_asynchronous => 1);
    my $event = F5::ASMConfig::Event->new(element => $element_list[0],
                                          event_time => $element_list[0]->{request_log}->{log_time},
                                          search_result => $search_result,
                                          event_type => F5::ASMConfig::EventType::ADD);

    $self->{logger}->debug("Flushing all webhooks logs") if $self->{logger};
    eval {
        $asm_config_client->send_event($asm_config_client->{session},
                                       $call_flags,
                                       $event);
    };
    if (my $err = $@) {
        $self->{logger}->error("Failed to send webhooks logs events") if $self->{logger};
        $self->{logger}->error($err) if $self->{logger};
    }
}

#----------------------------------------------------------------------

sub handle_webhook_request_log {
    my($self, %params) = @_;
    my $db_rec = $params{db_rec};

    my $webhook_event_viol_rating = {F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_ILLEGAL_HTTP_REQUEST              => 1,
                                     F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_HTTP_REQUEST_NEEDING_EXAMINATION  => 3,
                                     F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_LIKELY_MALICIOUS_HTTP_REQUEST     => 4 };

    if (!$policies_webhook_event_hash->{$db_rec->{policy_id}}) {
        return;
    }

    if (!$db_rec->{violation_rating}) {
        return;
    }

    for my $webhook_event (keys %{$webhook_event_viol_rating}) {
        if (($db_rec->{has_violations}) && ($db_rec->{violation_rating} >= $webhook_event_viol_rating->{$webhook_event}) && $policies_webhook_event_hash->{$db_rec->{policy_id}}->{$webhook_event}) {
            my $need_sort = 0;
            my $webhook_log = { %$db_rec,
                                policy => $self->{value_map}->{policy_id_to_webhook_fields}->{$db_rec->{policy_id}} };

            if (($self->{webhooks_logs}) && (scalar @{$self->{webhooks_logs}}) >= $WEBHOOK_NUMBER_OF_MESSAGES_IN_QUEUE) {
                $self->{number_of_throttled_messages}++;
                my $last_element_in_array = $self->{webhooks_logs}[-1];
                if ($last_element_in_array->{violation_rating} < $db_rec->{violation_rating}) {
                    pop (@{$self->{webhooks_logs}});
                    push (@{$self->{webhooks_logs}}, $webhook_log);
                    $need_sort = 1;
                }
            }
            else {
                push (@{$self->{webhooks_logs}}, $webhook_log);
                $need_sort = 1;
            }

            if ($need_sort) {
                my @sorted_logs = sort { $b->{violation_rating} <=> $a->{violation_rating} || $a->{log_time} cmp $b->{log_time}} @{$self->{webhooks_logs}};
                $self->{webhooks_logs} = \@sorted_logs;
            }
            last;
        }
    }

    $self->{logger}->debug("Throttled webhook number of log messages: $self->{number_of_throttled_messages}") if $self->{logger};
}

#----------------------------------------------------------------------

my %SESSION_AWARENESS_SCOPES = (
                                F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_USER() => 'username',
                                F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_SESSION() => 'session_id',
                                F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_IP() => 'source_ip',
                                F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_DEVICE() => 'device_id',
                               );

sub add_session_awareness_lock {
    my ($self, %params) = @_;
    my $bd_rec = $params{bd_rec};
    return unless $self->{value_map}->{policy_name_crc_to_session_awareness}->{$bd_rec->{policy_name_crc}};
    return unless $self->{sa_dp_count} < $SESSION_TRACKING_MAXIMUM_BATCH_SIZE;
    my $policy_id = $self->{value_map}->{policy_name_crc_to_policy_id}->{$bd_rec->{policy_name_crc}};

    for my $scope (keys %SESSION_AWARENESS_SCOPES) {
        if ($bd_rec->{data}->{request}->{sa_lock_scopes_bitmask} & (2 ** ($scope - 1))) {

            my $data_point_value = $bd_rec->{data}->{request}->{$SESSION_AWARENESS_SCOPES{$scope}};
            if ($scope == F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_IP) {
                $data_point_value = $self->ip_address_as_string(ip_field => $data_point_value);
            }
            elsif (($scope == F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_SESSION) ||
                   ($scope == F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_DEVICE)) {
                $data_point_value = Math::BigInt->new($data_point_value)->as_hex;
                $data_point_value =~ s/^0x//;
            }
            if ($self->{sa_dp_count} < $SESSION_TRACKING_MAXIMUM_BATCH_SIZE && !$self->{sa_dp}->{$policy_id}->{$scope}->{$data_point_value}) {
                $self->{sa_dp}->{$policy_id}->{$scope}->{$data_point_value} = 1;
                $self->{sa_dp_count}++;
            }
        }
    }
 
}

#----------------------------------------------------------------------

my %tighten_violations = (
    F5::Violation::VIOL_PARAMETER() => 1,
    F5::Violation::VIOL_URL()       => 1,
    F5::Violation::VIOL_FILETYPE()  => 1,
    F5::Violation::VIOL_REDIRECT()  => 1,
);

sub has_violations {
    my( $violations_set_idx) = @_;
    
    my @violation_ids = F5::Violation::Utils::get_indexes_from_viol_set($violations_set_idx);
    if (scalar @violation_ids) {
        return 1;
    }
    
    return 0;
}

#----------------------------------------------------------------------

sub init_run_insert {
    my($self, %params) = @_;

    # Create all paths on initial startup (partition is created and mounted at ASM startup)
    for my $create_path (($PROTOBUF_FILE_PATH,
                          $SECONDARY_CREATE_FILE_PATH,
                          $SECONDARY_TRANSFER_FILE_PATH,
                          $PRIMARY_ACK_TRANSFER_FILE_PATH,
                         )) {
        if (! -e $create_path) {
            mkpath($create_path);
        }
    }

    $self->reinitialize_policy_mapping();
    $self->{bd_protobuf} = bd::Main->new();
    $self->{bot_defense_event_log_protobuf} = bot_defense_event_log::BotDefenseEventLog->new();
    $self->{rql_serializer} = Thrift::BinaryProtocol->new(Thrift::MemoryBuffer->new());

    $self->{db_clean} = F5::DbClean->new(dbh => $self->{dbh},
                                         logger => $self->{logger},
                                         config_file => CLEAN_DB_CFG_FILE,
                                         after_clean => sub {
                                             my (%after_clean_params) = @_;
                                             $self->after_clean_handler(%after_clean_params);
                                         },
                                        );

    $self->{batch_insert} = F5::BatchInsert->new(dbh => $self->{dbh},
                                                 db_handler => $self->{db_clean},
                                                 batch_size => $BATCH_INSERT_MAXIMUM_BATCH_SIZE,
                                                 logger => $self->{logger});

    $self->{requests_handled_count} = 0;
    $self->{bot_defense_events_handled_count} = 0;
    $self->{attacks_handled_count} = 0;
    $self->{login_objects_handled_count} = 0;
    $self->{local_stored_count} = 0;
    $self->{pb_handled_count} = 0;
    $self->{secondary_handled_count} = 0;
    $self->{all_pipe_messages_count} = 0;
    $self->{load_from_secondary_count} = 0;
    $self->{rate_limit_skipped_count} = 0;
    $self->{rate_limit_bot_defense_skipped_count} = 0;
    $self->{recent_insert_times} = [];
    $self->{recent_bot_defense_insert_times} = [];
    $self->{last_support_id_handled} = undef;
    $self->{last_bot_defense_support_id_handled} = undef;
    $self->{last_attack_id_handled} = undef;
    $self->{last_login_object_handled} = undef;
    $self->{sa_dp} = {};
    $self->{sa_dp_count} = 0;

    $self->{messagepack} = Data::MessagePack->new();

    $self->{selector} = F5::IO::Selector->new(sleep_on_poll_when_no_connections => 1);
    for my $pipe_type (@{$self->{pipe_types}}) {
        $self->open_pipe(pipe_type => $pipe_type);
        $self->{last_pipe_connect_attempt_time}->{$pipe_type} = time;
        $self->{"pipe_" . $pipe_type . "_requests_handled_count"} = 0;
    }
    $self->open_asmconfig_event_listener;
    $self->{last_asmconfig_connect_attempt_time} = time;

    my $dbh = $self->{dbh} || F5::Db::Single->instance;
    $self->load_existing_protobuf_files(dbh => $dbh);

    ($self->{last_bot_defense_event_log_id}) = $dbh->selectrow_array('SELECT MAX(bot_defense_event_log_internal_id) FROM PRX.BOT_DEFENSE_EVENT_LOG');
    $self->{last_bot_defense_event_log_id} = Math::BigInt->new($self->{last_bot_defense_event_log_id} || 0);

    $self->{primary_loadable_files} = $dbh->selectall_hashref('SELECT * FROM PRX.REQUEST_LOG_SYNC_FILES',
                                                              'sync_filename');
    # todo: complete sync file in "loading" state
    $self->update_primary_loadable_files;

    # end all ongoing attacks
    my $attack_end_status = F5::ASMConfig::BruteForceAttackLogAttackStatus::BRUTE_FORCE_ATTACK_LOG_ATTACK_STATUS_ENDED();
    F5::DbUtils::update_row(dbh => $dbh,
                            table_name => 'PRX.BRUTE_FORCE_ATTACK_LOG',
                            rec => { attack_status => $attack_end_status },
                            set_date_fields_to_now_unix => [ 'mitigation_end_time' ],
                            where_fields => { mitigation_end_time => 0 }
                           );

    # blow away all login page status reports, bd will resend the relevant ones again
    $dbh->do("DELETE FROM PRX.BRUTE_FORCE_LOGIN_PAGE_STATUS");

    # cleanup possible leftovers due to unexpected asmlogd restart
    $self->cleanup_brute_force();
}

#----------------------------------------------------------------------

sub open_pipe {
    my($self, %params) = @_;
    my $pipe_type = $params{pipe_type};
    $self->{logger}->debug("Connecting to pipe $pipe_type") if $self->{logger};
    $self->{pipes}->{$pipe_type} = F5::IO::LrnPipe->new(pipe_type => $pipe_type,
                                             logger => $self->{logger},
                                             selector => $self->{selector},
                                             read_message_handler => sub {
                                                 $self->pipe_message_received(msg => $_[1], 
                                                                              pipe_type => $pipe_type);
                                             },
                                             socket_close_handler => sub {
                                                 my ($handle, $errno) = @_;
                                                 $self->{logger}->info("Pipe close received") if $self->{logger};
                                                 $handle->{selector}->remove(shift) if $handle->{selector};
                                                 $handle->close if $handle;

                                                 delete $self->{pipes}->{$pipe_type};
                                             },
                                            );

}

#----------------------------------------------------------------------

sub open_asmconfig_event_listener {
    my($self) = @_;

    $self->{logger}->debug("Connecting event listener") if $self->{logger};
    my $element_types = [
                         F5::ASMConfig::ElementEventType->new({ event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                                element_type => F5::ASMConfig::ElementType::REQUEST_LOG_SYNC}),
                         F5::ASMConfig::ElementEventType->new({ event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                                element_type => F5::ASMConfig::ElementType::CLUSTER_STATE}),
                         F5::ASMConfig::ElementEventType->new({ event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                                element_type => F5::ASMConfig::ElementType::POLICY}),
                         F5::ASMConfig::ElementEventType->new({ event_type => F5::ASMConfig::EventType::ALL_TYPES,,
                                                                element_type => F5::ASMConfig::ElementType::SET_ACTIVE}),
                         F5::ASMConfig::ElementEventType->new({ event_type => F5::ASMConfig::EventType::ALL_TYPES,,
                                                                element_type => F5::ASMConfig::ElementType::UCS_CONFIG_LOAD}),
                         F5::ASMConfig::ElementEventType->new({ event_type => F5::ASMConfig::EventType::ALL_TYPES,,
                                                                element_type => F5::ASMConfig::ElementType::WEBHOOK}),
                         # todo: add events for Method to cache
                         # todo: load methods and policy directly on startup (since catchup is disabled, as it should be)
                        ];

    my $event_subscription = F5::ASMConfig::EventSubscription->new({element_event_types => $element_types,
                                                                    send_topology_suppressed_events => 1,
                                                                    send_catchup_events => 0,
                                                                   });

    $self->{asmconfig_event_listener} = F5::IO::ASMConfig::EventListener->new(event_subscription => $event_subscription,
                                                                              logger => $self->{logger},
                                                                              read_message_handler => sub {
                                                                                  $self->handle_asmconfig_event(event => $_[1]);
                                                                              },
                                                                              socket_close_handler => sub {
                                                                                  my ($handle, $errno) = @_;
                                                                                  $self->{logger}->info("Event listener socket close received") if $self->{logger};
                                                                                  $handle->{selector}->remove(shift) if $handle->{selector};
                                                                                  $handle->close if $handle;

                                                                                  delete $self->{asmconfig_event_listener};
                                                                              },
                                                                              selector => $self->{selector},
                                                                             );

    # Since catchup events are disabled, initial values need to be loaded after
    # asmconfig event subscription is started up (for cluster state, policy/account mappings, etc.)
    $self->update_cluster_state;
    $self->load_policy_mapping;
}

#----------------------------------------------------------------------

sub init_webhooks_cache {
    my($self, %params) = @_;
    my $policy = $params{policy};

    my $asm_config_client;
    if ($self->{asmconfig_event_listener}) {
        $asm_config_client = $self->{asmconfig_event_listener}->get_client;
    }

    if (!$asm_config_client) {
        $self->{logger}->info("Could not connect to asmconfig -- policy mapping will be loaded as needed") if $self->{logger};
        return;
    }

    eval {
            my $search_result = $asm_config_client->search_webhook($asm_config_client->{session},
                                                                   undef,
                                                                   F5::ASMConfig::Policy->new({policy_id => $policy->{policy_id}}));

            for my $element (@{$search_result->{element_list}}) {
                $policies_webhook_event_hash->{$policy->{policy_id}}->{$element->{webhook}->{trigger_event}}->{$element->{webhook}->{webhook_id}} = $element->{webhook};
                $self->{value_map}->{policy_id_to_webhook_fields}->{$policy->{policy_id}} = $policy;
            }
    };
    if (my $err = $@) {
        $self->{logger}->error("Failed to search webhooks") if $self->{logger};
        $self->{logger}->error($err) if $self->{logger};
    }
}

#----------------------------------------------------------------------

sub load_policy_mapping {
    my($self, %params) = @_;

    my $asmconfig_client;
    if ($self->{asmconfig_event_listener}) {
        $asmconfig_client = $self->{asmconfig_event_listener}->get_client;
    }

    if (! $asmconfig_client) {
        $self->{logger}->info("Could not connect to asmconfig -- policy mapping will be loaded as needed") if $self->{logger};
        return;
    }

    $asmconfig_client->{session}->{partition} = F5::ASMConfig::Constants::PARTITION_ALL; # otherwise we default to search in /Common
    my $call_flags = F5::ASMConfig::CallFlags->new(selected_fields => [ qw(policy_id policy_name_crc),
                                                                        @POLICY_WEBHOOK_FIELDS ]);
    if ($params{policy_name_crc}) {
        $call_flags->{search_filters} = [ F5::ASMConfig::SearchFilter->new(field_name => 'policy_name_crc',
                                                                           field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ,
                                                                           field_value => $params{policy_name_crc}) ];
    }
    my $policies_result = $asmconfig_client->search_policy($asmconfig_client->{session},
                                                           $call_flags);

    for my $policy (map { $_->{policy} } @{$policies_result->{element_list}}) {
        $self->update_policy_mapping(policy => $policy,
                                     update_session_awareness => 1);
        $self->init_webhooks_cache(policy => $policy);
    }
}

#----------------------------------------------------------------------

sub update_policy_mapping {
    my($self, %params) = @_;
    my $policy = $params{policy};

    my $policy_name_crc = F5::CRC::get_crc($policy->{policy_name});
    my $old_policy_name_crc = $self->{value_map}->{policy_id_to_policy_name_crc}->{$policy->{policy_id}};
    $self->{logger}->debug("Mapping account $policy_name_crc to policy $policy->{policy_id} '$policy->{policy_name}'") if $self->{logger};

    $self->{value_map}->{policy_name_crc_to_policy_id}->{$policy_name_crc} = $policy->{policy_id};
    if ($params{update_session_awareness}) {
        my $dbh = $params{dbh} || $self->{dbh} || F5::Db::Single->instance;
        my ($flg_session_awareness) = $dbh->selectrow_array(<<EndQuery, undef, $policy_name_crc);
SELECT flg_session_awareness_on
 FROM DCC.ACCOUNT_SESSION_AWARENESS_SETTINGS
 WHERE policy_name_crc = ?
EndQuery
        if ( defined $flg_session_awareness ) { 
            $self->{value_map}->{policy_name_crc_to_session_awareness}->{$policy_name_crc} = $flg_session_awareness;
        }
        else { 
            delete $self->{value_map}->{policy_name_crc_to_session_awareness}->{$policy_name_crc};
        }
    }
    
    if ($old_policy_name_crc && $old_policy_name_crc != $policy_name_crc &&
           $self->{value_map}->{policy_name_crc_to_policy_id}->{$old_policy_name_crc} == $policy->{policy_id}) {
        #remove invalid references
        delete $self->{value_map}->{policy_name_crc_to_session_awareness}->{$old_policy_name_crc};
        delete $self->{value_map}->{policy_name_crc_to_policy_id}->{$old_policy_name_crc};
    }

    $self->{value_map}->{policy_id_to_policy_name_crc}->{$policy->{policy_id}} = $policy_name_crc;
    $self->{value_map}->{policy_id_to_policy_name}->{$policy->{policy_id}} = $policy->{policy_name};
    $self->{value_map}->{policy_id_to_encoding}->{$policy->{policy_id}} = $policy->{encoding};
}

#----------------------------------------------------------------------

sub pipe_message_received {
    my ($self, %params) = @_;
    my $msg = $params{msg};
    my $pipe_type = $params{pipe_type};

    $self->{all_pipe_messages_count}++;

    if ($pipe_type eq 'RQL') {
        $self->{"pipe_" . $pipe_type . "_requests_handled_count"}++;
        return $self->protobuf_message_received(raw_protobuf => $msg,
                                                request_type => 'request',
                                               );
    }
    elsif ($pipe_type eq 'PBRQL') {
        $self->{"pipe_" . $pipe_type . "_requests_handled_count"}++;
        $self->{rql_serializer}->{trans}->resetBuffer($msg);
        my $request_log_protobuf = F5::ASMConfig::RequestLogProtobuf->new();
        eval {
            $request_log_protobuf->read($self->{rql_serializer});
        };
        if ($@) {
            $self->{logger}->info("Could not parse record from $pipe_type: $@");
        }
        else { 
            return $self->protobuf_message_received(raw_protobuf => $request_log_protobuf->{protobuf_bin},
                                                    suggestion_keys => $request_log_protobuf->{suggestion_keys},
                                                    request_type => 'pb_request',
                                                   );
        }
    }
    elsif ($pipe_type eq 'BOTDEFENSERQL') {
        $self->{"pipe_" . $pipe_type . "_requests_handled_count"}++;
        $self->{rql_serializer}->{trans}->resetBuffer($msg);
        my $bot_defense_event_log_protobuf = F5::ASMConfig::BotDefenseEventLogProtobuf->new();
        eval {
            $bot_defense_event_log_protobuf->read($self->{rql_serializer});
        };
        if ($@) {
            $self->{logger}->info("Could not parse record from $pipe_type: $@");
        }
        else { 
            return $self->protobuf_message_received(raw_protobuf => $bot_defense_event_log_protobuf->{protobuf_bin},
                                                    request_type => 'bot_defense_request',
                                                   );
        }
    }
}

#----------------------------------------------------------------------

sub rate_limit {
    my($self, %params) = @_;

    my $rate_limit_count;
    my $rate_limit_interval;

    my $is_bot_defense_request = $self->check_is_bot_defense_request(%params);
    if ($is_bot_defense_request) {
        $rate_limit_count = ($self->{is_cluster_secondary} ? $SECONDARY_BOT_DEFENSE_RATE_LIMIT_COUNT : $LOCAL_INSERT_BOT_DEFENSE_RATE_LIMIT_COUNT);
        $rate_limit_interval = ($self->{is_cluster_secondary} ? $SECONDARY_BOT_DEFENSE_RATE_LIMIT_INTERVAL : $LOCAL_INSERT_BOT_DEFENSE_RATE_LIMIT_INTERVAL);
    }
    else {
        $rate_limit_count = ($self->{is_cluster_secondary} ? $SECONDARY_RATE_LIMIT_COUNT : $LOCAL_INSERT_RATE_LIMIT_COUNT);
        $rate_limit_interval = ($self->{is_cluster_secondary} ? $SECONDARY_RATE_LIMIT_INTERVAL : $LOCAL_INSERT_RATE_LIMIT_INTERVAL);
    }

    my $insert_times_queue = ($is_bot_defense_request) ? 'recent_bot_defense_insert_times' : 'recent_insert_times';
    my $rate_limit_skipped_count_field = ($is_bot_defense_request) ? 'rate_limit_bot_defense_skipped_count' : 'rate_limit_skipped_count';
    my $record_type = ($is_bot_defense_request) ? 'Bot Defense request' : 'request';

    if (scalar @{$self->{$insert_times_queue}} >= $rate_limit_count) {
        my $now = Time::HiRes::time;
        # Clear out all elements in insert_times_queue that are older than rate_limit_interval.
        # get_time_until_maximum_interval will return 0 for any element for which insert_time is logically
        # rate_limit_interval seconds ago (accommodating for clock changes).
        while (@{$self->{$insert_times_queue}} &&
               ! F5::DateTimeUtils::get_time_until_maximum_interval(current_time => $now,
                                                                    start_time => $self->{$insert_times_queue}->[0]->{insert_time},
                                                                    maximum_interval => $rate_limit_interval)) {

            shift(@{$self->{$insert_times_queue}});
        }

        if (scalar @{$self->{$insert_times_queue}} >= $rate_limit_count) {
            $self->{logger}->debug("Skipping $record_type record due to rate limit") if $self->{logger};
            $self->{$rate_limit_skipped_count_field}++;
            return 1;
        }
    }

    return 0;
}

#----------------------------------------------------------------------

sub check_is_bot_defense_request {
    my($self, %params) = @_;
    return ($params{request_type} && $params{request_type} eq 'bot_defense_request');
}

#----------------------------------------------------------------------

sub check_is_pb_request {
    my($self, %params) = @_;
    return ($params{request_type} && $params{request_type} eq 'pb_request');
}

#----------------------------------------------------------------------

sub protobuf_message_received {
    my($self, %params) = @_;

    return if $self->rate_limit(%params);

    if ($self->{is_cluster_secondary}) {
        $self->write_to_secondary_file(%params);
        $self->{secondary_handled_count}++;
    }
    else {
        $self->{local_stored_count}++;

        # different insert types have their own internal handled counters
        $self->insert_record(%params);

        if ($self->check_is_pb_request(%params)) {
            $self->{pb_handled_count}++;
        }
    }
}

#----------------------------------------------------------------------

sub update_webhooks_cache {
    my($self, %params) = @_;
    my $event = $params{event};

    if ($event->{element}->{element_type} == F5::ASMConfig::ElementType::WEBHOOK) {
        my $webhook = $event->{element}->{webhook};
        if (($event->{event_type} == F5::ASMConfig::EventType::ADD) || ($event->{event_type} == F5::ASMConfig::EventType::UPDATE)) {
            $policies_webhook_event_hash->{$event->{policy}->{policy_id}}->{$webhook->{trigger_event}}->{$webhook->{webhook_id}} = $webhook;
        } elsif ($event->{event_type} == F5::ASMConfig::EventType::DELETE) {
            delete $policies_webhook_event_hash->{$event->{policy}->{policy_id}}->{$webhook->{trigger_event}}->{$webhook->{webhook_id}};
        }
    } elsif ($event->{element}->{element_type} == F5::ASMConfig::ElementType::POLICY) {
        if ($event->{event_type} == F5::ASMConfig::EventType::ADD) {
            $self->init_webhooks_cache(policy => $event->{policy});
        } elsif ($event->{event_type} == F5::ASMConfig::EventType::DELETE) {
            delete $policies_webhook_event_hash->{$event->{policy}->{policy_id}};
        }
    }
}

#----------------------------------------------------------------------

sub handle_asmconfig_event {
    my($self, %params) = @_;

    my $event = $params{event};

    $self->{logger}->trace("asmconfig event received") if $self->{logger};
    $self->{logger}->trace($event) if $self->{logger};

    if ($event->{element}->{element_type} == F5::ASMConfig::ElementType::CLUSTER_STATE) {
        $self->{logger}->debug("Cluster state update received") if $self->{logger};
        $self->update_cluster_state;
    }
    elsif ($event->{element}->{element_type} == F5::ASMConfig::ElementType::REQUEST_LOG_SYNC) {
        $self->{logger}->debug("Sync file detected") if $self->{logger};
        $self->update_primary_loadable_files;
    }
    elsif ($event->{element}->{element_type} == F5::ASMConfig::ElementType::POLICY) {
        $self->update_policy_mapping(policy => $event->{element}->{policy});
        if ($event->{event_type} == F5::ASMConfig::EventType::DELETE) {
            $self->cleanup_brute_force(policy_id => $event->{element}->{policy}->{policy_id});
        }
        $self->update_webhooks_cache(event => $event);
    }
    elsif ($event->{element}->{element_type} == F5::ASMConfig::ElementType::SET_ACTIVE) {
        $self->update_policy_mapping(policy => $event->{policy},
                                     update_session_awareness => 1);
        $self->cleanup_brute_force(policy_id => $event->{policy}->{policy_id});
        if ($event->{event_type} == F5::ASMConfig::EventType::UPDATE) {
            $self->{value_map}->{policy_id_to_webhook_fields}->{$event->{policy}->{policy_id}} = $event->{policy};
        }
    }
    elsif ($event->{element}->{element_type} == F5::ASMConfig::ElementType::UCS_CONFIG_LOAD) {
        $self->reinitialize_policy_mapping();
        $policies_webhook_event_hash = {};
        $self->load_policy_mapping();
    }
    elsif ($event->{element}->{element_type} == F5::ASMConfig::ElementType::WEBHOOK) {
        $self->update_webhooks_cache(event => $event);
    }
}

#----------------------------------------------------------------------

sub reinitialize_policy_mapping {
    my($self) = @_;

    $self->load_existing_partition_data;

    $self->{value_map} = { policy_name_crc_to_policy_id => {},
                           policy_id_to_policy_name_crc => {},
                           policy_id_to_policy_name => {},
                           policy_name_crc_to_session_awareness => {},
                           policy_id_to_encoding => {},
                           policy_id_to_webhook_fields => {},
                         };
}

#----------------------------------------------------------------------

sub load_existing_protobuf_files {
    my($self, %params) = @_;

    my $logger = $self->{logger};

    $self->{total_protobuf_file_size} = 0;
    $self->{protobuf_files} = {};

    find({ wanted => sub {
               return unless (-f $File::Find::name);
               return unless (basename($File::Find::name) =~ /^\d+$/);

               my $file_id = basename($File::Find::name) + 0;
               $self->{protobuf_files}->{$file_id} = {
                                                      file_size => -s $File::Find::name,
                                                      file_id => $file_id,
                                                      file_name => $File::Find::name,
                                                     };
           },
           no_chdir => 1,
         },
         $PROTOBUF_FILE_PATH);

    my $dbh = $params{dbh} || $self->{dbh} || F5::Db::Single->instance;
    if (! %{$self->{protobuf_files}}) {
        my($max_file_id) = $dbh->selectrow_array('SELECT MIN(protobuf_file_id) FROM PRX.REQUEST_LOG');
        my $new_file_id = ($max_file_id || 0) + 1;
        $self->{protobuf_files}->{$new_file_id} = { file_id => $new_file_id,
                                                    file_name => $self->get_protobuf_filename(file_id => $new_file_id),
                                                    file_size => 0 };
    }

    $self->{total_protobuf_file_size} = List::Util::sum(map { $_->{file_size} } values %{$self->{protobuf_files}});
    my $max_protobuf_file_id = List::Util::max(keys %{$self->{protobuf_files}});
    $self->{current_protobuf_file} = $self->{protobuf_files}->{$max_protobuf_file_id};

    # todo: Check last protobuf file that last record was not cut off

    $self->load_existing_partition_data;
}

#----------------------------------------------------------------------

sub load_existing_partition_data {
    my($self, %params) = @_;

    my $logger = $self->{logger};
    my $dbh = $params{dbh} || $self->{dbh} || F5::Db::Single->instance;

    # Load initial id and existing partition data
    ($self->{last_request_log_id}) = $dbh->selectrow_array('SELECT MAX(id) FROM PRX.REQUEST_LOG');
    $self->{last_request_log_id} = Math::BigInt->new($self->{last_request_log_id} || 0);

    $self->{partitions} = F5::Db::Partition::retrieve_partitions_info(dbh => $dbh,
                                                                      table_name => 'PRX.REQUEST_LOG',
                                                                      logger => $logger,
                                                                     );

    for my $partition (@{$self->{partitions}}) {
        $partition->{max_id} = Math::BigInt->new($partition->{PARTITION_DESCRIPTION});
    }
    $self->{partitions} = [ sort { $a->{max_id} <=> $b->{max_id} } @{$self->{partitions}} ];
    $self->adjust_partitions;
}

#----------------------------------------------------------------------

sub adjust_partitions {
    my($self, %params) = @_;
    my $next_request_log_id = $params{next_request_log_id};
    my $table_name = $params{table_name} || 'PRX.REQUEST_LOG';
    my $logger = $self->{logger};

    my $max_id = 1;
    if ($self->{partitions} && @{$self->{partitions}}) {
        $max_id = $self->{partitions}->[-1]->{max_id};
    }

    if (! $next_request_log_id || $next_request_log_id >= $max_id) {
        my $dbh = $params{dbh} || $self->{dbh} || F5::Db::Single->instance;

        my $next_partition_first_id = $max_id + $DB_PARTITION_ROW_LIMIT;

        my $number_of_partitions = 0;
        if (ref($self->{partitions}) eq 'ARRAY') {
            $number_of_partitions = scalar @{$self->{partitions}};
        }

        if ($number_of_partitions >= $DB_PARTITION_MAX_NUMBER) {
            my $partitions_num_drop = $number_of_partitions - $DB_PARTITION_MAX_NUMBER + 1;

            for (my $i = 0; $i < $partitions_num_drop; $i++) {
                my $drop_partition = shift @{$self->{partitions}};
                F5::Db::Partition::delete_partition(dbh => $dbh,
                                                    table_name => 'PRX.REQUEST_LOG',
                                                    partition_name => $drop_partition->{PARTITION_NAME},
                                                    logger => $logger,
                                                   );
            }
        }

        my $partition_name = F5::Db::Partition::add_partition(dbh => $dbh,
                                                              table_name => $table_name,
                                                              next_partition_first_id => $next_partition_first_id,
                                                              logger => $logger,
                                                             );

        # max_id is a bit misleading, cause in fact a partition will hold entries w/ IDs less then the value of 'max_id'
        push(@{$self->{partitions}},
             { PARTITION_NAME => $partition_name,
               max_id => $next_partition_first_id });
    }
}

#----------------------------------------------------------------------

sub ip_address_as_binary {
    my($self, %params) = @_;
    my $ip_field = $params{ip_field};

    return pack("L4",
                $ip_field->{ip0},
                $ip_field->{ip1},
                $ip_field->{ip2},
                $ip_field->{ip3});
}

#----------------------------------------------------------------------

sub ip_address_as_string {
    my($self, %params) = @_;
    my $ip_field = $params{ip_field};

    my $ip_address_str = '';
    if ($ip_field && ref($ip_field) && exists $ip_field->{ip0}) {

        $ip_address_str = F5::IpAddr::ip_ntop(pack("L4",
                                                   $ip_field->{ip0},
                                                   $ip_field->{ip1},
                                                   $ip_field->{ip2},
                                                   $ip_field->{ip3}));

        if ($ip_field->{rtdom_id}) {
            # route domain 0 can be safely ignored
            $ip_address_str .=  "%" . $ip_field->{rtdom_id};
        }
    }
    return $ip_address_str;
}

#----------------------------------------------------------------------

sub session_id_as_hex {
    my($self, %params) = @_;

    my $hex_session_id = '';
    if ($params{session_id}) {
        eval {
            my $vec = Bit::Vector->new_Dec(128, $params{session_id});
            $hex_session_id = lc $vec->to_Hex();
            $hex_session_id =~ s/^0+//;
        };
    }
    return $hex_session_id;
}

#----------------------------------------------------------------------

sub get_protobuf_filename {
    my($self, %params) = @_;
    my $file_id = $params{file_id};

    return sprintf("$PROTOBUF_FILE_PATH/%.10d", $file_id);
}

#----------------------------------------------------------------------

sub get_protobuf {
    my($self, %params) = @_;
    my $rec = $params{rec};

    if (! $GET_PROTOBUF_LAST_FILE_ID || $rec->{protobuf_file_id} != $GET_PROTOBUF_LAST_FILE_ID) {
        my $file_name = $self->get_protobuf_filename(file_id => $rec->{protobuf_file_id});
        return undef unless -f $file_name;
        open($GET_PROTOBUF_FH, '<', $file_name);
        $GET_PROTOBUF_LAST_FILE_ID = $rec->{protobuf_file_id};
    }
    seek($GET_PROTOBUF_FH, $rec->{protobuf_file_seek_pos}, 0);
    my $protobuf_contents;
    read($GET_PROTOBUF_FH, $protobuf_contents, $rec->{protobuf_record_size});
    return $protobuf_contents;
}

#----------------------------------------------------------------------

sub after_clean_handler {
    my($self, %params) = @_;

    if ($params{table_name} eq 'PRX.BRUTE_FORCE_ATTACK_LOG') {
        $self->delete_unreferenced_brute_force_attack_data();
    }
    else {
        $self->delete_unused_protobuf_files;
    }
}

#----------------------------------------------------------------------

sub cleanup_brute_force {
    my($self, %params) = @_;
    my $policy_id = $params{policy_id};
    my $dbh = $self->{dbh} || F5::Db::Single->instance;

    for my $orphan_db_table (@BRUTE_FORCE_CLEANUP_TABLES) {
        $self->delete_orphaned_records(main_db_table => 'PLC.PL_POLICIES',
                                       main_db_field => 'id',
                                       orphan_db_table => $orphan_db_table,
                                       orphan_db_field => 'policy_id',
                                       policy_id => $policy_id,
                                      );

        $self->delete_orphaned_records(main_db_table => 'PLC.PL_LOGIN_OBJECTS',
                                       main_db_field => 'rest_uuid',
                                       orphan_db_table => $orphan_db_table,
                                       orphan_db_field => 'login_object_rest_uuid',
                                       additional_join_constraint => 'AND orphan_table.policy_id = main_table.policy_id',
                                       policy_id => $policy_id,
                                      );
    }

    $self->delete_unreferenced_brute_force_attack_data();
}

#----------------------------------------------------------------------

sub delete_unreferenced_brute_force_attack_data {
    my($self, %params) = @_;
    my $dbh = $self->{dbh} || F5::Db::Single->instance;

    for my $orphan_db_table (values %REPEATED_SUB_MESSAGE_TO_DB_TABLE) {
        $self->delete_orphaned_records(main_db_table => 'PRX.BRUTE_FORCE_ATTACK_LOG',
                                       main_db_field => 'attack_id',
                                       orphan_db_table => $orphan_db_table,
                                       orphan_db_field => 'attack_id',
                                      );
    }
}

#----------------------------------------------------------------------

sub delete_orphaned_records {
    my($self, %params) = @_;
    my $main_db_table = $params{main_db_table};
    my $main_db_field = $params{main_db_field};
    my $orphan_db_table = $params{orphan_db_table};
    my $orphan_db_field = $params{orphan_db_field};
    my $additional_join_constraint = $params{additional_join_constraint} || '';
    my $policy_id = $params{policy_id};

    $self->{logger}->debug("Flushing to database") if $self->{logger};
    $self->{batch_insert}->flush if $self->{batch_insert};

    my $dbh = $self->{dbh} || F5::Db::Single->instance;

    my $where_fields = {};
    if ($policy_id) {
      $where_fields->{'orphan_table.policy_id'} = $policy_id;
    }

    my $constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => $where_fields);
    my $where_clause = $constructed_where_clause->{where_clause};
    my @where_values = @{$constructed_where_clause->{where_values}};

    my ($db_name,$table_name) = split(/\./,$orphan_db_table);

    $self->{logger}->debug("Cleaning orphaned records from $db_name.$table_name") if $self->{logger};

    my $orphan_delete_count_result = $dbh->do(<<EndQuery, undef, @where_values);
DELETE orphan_table
  FROM $db_name.$table_name AS orphan_table
  LEFT JOIN $main_db_table AS main_table
         ON orphan_table.$orphan_db_field = main_table.$main_db_field
         $additional_join_constraint
 WHERE main_table.$main_db_field IS NULL
       AND $where_clause
EndQuery

    $self->{logger}->debug("Cleaned $orphan_delete_count_result rows from $db_name.$table_name") if $self->{logger};
}

#----------------------------------------------------------------------

sub delete_unused_protobuf_files {
    my($self) = @_;
    my $dbh = $self->{dbh} || F5::Db::Single->instance;
    my $logger = $self->{logger};

    my($min_file_id) = $dbh->selectrow_array('SELECT MIN(protobuf_file_id) FROM PRX.REQUEST_LOG') || 0;

    for my $file_id (keys %{$self->{protobuf_files}}) {
        if ($file_id < $min_file_id) {
            $self->{logger}->info("Dropping protobuf file $self->{protobuf_files}->{$file_id}->{file_name}") if $self->{logger};

            unlink($self->{protobuf_files}->{$file_id}->{file_name});
            $self->{total_protobuf_file_size} -= $self->{protobuf_files}->{$file_id}->{file_size};
            delete $self->{protobuf_files}->{$file_id};
        }
    }

    my($min_request_log_id) = $dbh->selectrow_array('SELECT MIN(id) FROM PRX.REQUEST_LOG');
    $min_request_log_id = Math::BigInt->new($min_request_log_id);

    while (@{$self->{partitions}}) {
        my $partition = $self->{partitions}->[0];
        if ($partition->{max_id} < $min_request_log_id) {
            $self->{logger}->info("Dropping partition $partition->{PARTITION_NAME}");

            F5::Db::Partition::delete_partition(dbh => $dbh,
                                                table_name => 'PRX.REQUEST_LOG',
                                                partition_name => $partition->{PARTITION_NAME},
                                                logger => $logger,
                                               );

            shift(@{$self->{partitions}});
        }
        else {
            last;
        }
    }

    $dbh->do('DELETE FROM PRX.REQUEST_LOG_PROPERTIES WHERE request_log_id < ?', undef,
             $min_request_log_id);
    
    $dbh->do('DELETE FROM PRX.REQUEST_LOG_TAGS WHERE request_log_id < ?', undef,
             $min_request_log_id);

    $dbh->do(<<EndQuery, undef, 0, $min_request_log_id);
UPDATE PRX.REQUEST_LOG_CLEARED 
  SET request_log_id = ? 
 WHERE request_log_id < ?
EndQuery

}

#----------------------------------------------------------------------

sub write_to_protobuf_file {
    my($self, %params) = @_;

    my $protobuf_msg = $params{protobuf_msg};
    my $protobuf_msg_length = length($protobuf_msg);

    my $retval = {
                  file_id => $self->{current_protobuf_file}->{file_id},
                  seek_pos => $self->{current_protobuf_file}->{file_size} + SIZEOF_INT,
                  record_size => $protobuf_msg_length,
                 };

    my $error;
    my $actual_written_data_size = 0;
    open(my $outf, '>>', $self->{current_protobuf_file}->{file_name}) or $error = $!;
    if ( !$error ) {
       print $outf pack('N', $protobuf_msg_length) or $error = $!;
       $actual_written_data_size += SIZEOF_INT if !$error;
       print $outf $protobuf_msg or $error = $!;
       $actual_written_data_size += $protobuf_msg_length if !$error;
       close $outf or $error = $!;
    }

    $self->{current_protobuf_file}->{file_size} += $actual_written_data_size;
    $self->{total_protobuf_file_size} += $actual_written_data_size;

    if ( $error ) {
         $self->{logger}->info("Write to file $self->{current_protobuf_file}->{file_name} failed ($error)") if $self->{logger};
         $self->cleanup_unrelated_protobuf_files;
         my $file_status = stat($self->{current_protobuf_file}->{file_name});
         my $size = $file_status->size;
         my $actual_written_data_size = $size - $self->{current_protobuf_file}->{file_size};
         $self->{current_protobuf_file}->{file_size} += $actual_written_data_size;
         $self->{total_protobuf_file_size} += $actual_written_data_size;
         return;
    }

    if ($self->{current_protobuf_file}->{file_size} > $PROTOBUF_FILE_INDIVIDUAL_MAX_SIZE) {
        my $previous_file_id = $self->{current_protobuf_file}->{file_id};

        $self->{protobuf_files}->{$previous_file_id + 1} = { file_id => $previous_file_id + 1,
                                                             file_name => $self->get_protobuf_filename(file_id => $previous_file_id + 1),
                                                             file_size => 0,
                                                           };
        $self->{current_protobuf_file} = $self->{protobuf_files}->{$previous_file_id + 1};
        $self->{logger}->info("Now writing to file $self->{current_protobuf_file}->{file_name}") if $self->{logger};
    }

    if ($self->{total_protobuf_file_size} > $PROTOBUF_FILE_TOTAL_MAX_SIZE) {
        my $file_id_to_drop = List::Util::min(keys %{$self->{protobuf_files}});
        my $dbh = $params{dbh} || $self->{dbh} || F5::Db::Single->instance;
        $dbh->do('DELETE FROM PRX.REQUEST_LOG WHERE protobuf_file_id <= ?', undef, $file_id_to_drop);

        $self->delete_unused_protobuf_files;
    }

    return $retval;
}

#----------------------------------------------------------------------

sub cleanup_unrelated_protobuf_files {
    my($self, %params) = @_;

    my $dbh = $params{dbh} || $self->{dbh} || F5::Db::Single->instance;

    my $request_log_file_ids =  $dbh->selectall_hashref(<<EndQuery,'protobuf_file_id');
SELECT DISTINCT protobuf_file_id
  FROM PRX.REQUEST_LOG
EndQuery

    my $request_log_file_names;
    for my $request_log_file_id (keys %$request_log_file_ids){
        my $request_log_file_name = F5::RequestLog->new()->get_protobuf_filename(file_id =>  $request_log_file_id);
         $request_log_file_names->{$request_log_file_name} = 1;
    }

    find(
        {
          wanted => sub {
                          my $name = $File::Find::name;
                          if (-f $name && !$request_log_file_names->{$name}){
                              $self->{logger}->info("Deleting unused file $name") if $self->{logger};
                              unlink $name;
                          }
                       },
          no_chdir => 1,
       }, $PROTOBUF_FILE_PATH);

}
#----------------------------------------------------------------------

sub write_to_secondary_file {
    my($self, %params) = @_;

    $self->{logger}->debug("Saving record, of $params{request_type} type, to $self->{current_secondary_file}->{file_name}") if $self->{logger};

    my $message_to_pack = {};
    $message_to_pack->{request_type} = $params{request_type};
    $message_to_pack->{raw_protobuf} = $params{raw_protobuf};

    my $packed_message = $self->{messagepack}->pack($message_to_pack);

    open(my $outf, '>>', $self->{current_secondary_file}->{file_name});
    print $outf $packed_message;
    close $outf;

    $self->{current_secondary_file}->{file_size} += length($packed_message);
}

#----------------------------------------------------------------------

sub rotate_secondary_file {
    my($self, %params) = @_;

    if ($self->{current_secondary_file} && $self->{current_secondary_file}->{file_size}) {
        # Check if existing secondary file should be moved for csyncd transfer
        if ($self->{current_secondary_file}->{file_size} > $SECONDARY_MAXIMUM_FILE_SIZE ||
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $SECONDARY_MAXIMUM_ROTATE_TIME,
                                                               start_time => $self->{current_secondary_file}->{start_time}) == 0) {

            my $secondary_integrity_filename = $self->get_secondary_integrity_filename(filename => $self->{current_secondary_file}->{file_name});
            $self->{logger}->debug("Writing integrity check file $secondary_integrity_filename") if $self->{logger};
            YAML::XS::DumpFile($secondary_integrity_filename,
                               { filename => $self->{current_secondary_file}->{file_name},
                                 file_size => -s $self->{current_secondary_file}->{file_name},
                               });
            rename($self->{current_secondary_file}->{file_name},
                   $SECONDARY_TRANSFER_FILE_PATH . '/'. basename($self->{current_secondary_file}->{file_name}));
            rename($secondary_integrity_filename,
                   $SECONDARY_TRANSFER_FILE_PATH . '/'. basename($secondary_integrity_filename));

            $self->{logger}->info("Secondary file " . basename($self->{current_secondary_file}->{file_name}) . " moved for transfer to primary") if $self->{logger};
            delete $self->{current_secondary_file};

            F5::Utils::File::cleanup_files(glob_pattern => "$SECONDARY_TRANSFER_FILE_PATH/*__slot_" . $self->{slot_number},
                                           age => 15 * 60, # 15 minutes
                                           max_retained_files => 30,
                                           max_size => 100_000_000, # 100 MB (approx)
                                          );

        }
    }

    if (! $self->{current_secondary_file}) {
        # This is either the first write to secondary file, or the previous secondary
        # file was just moved for csyncd transfer

        my $filename = "$SECONDARY_CREATE_FILE_PATH/request_log__" . filename_formatted_datetime() . "__slot_" . $self->{slot_number};
        $self->{current_secondary_file} = {
                                           start_time => time,
                                           file_name => $filename,
                                           file_size => 0,
                                          };
        $self->{logger}->info("Logging secondary requests to $filename") if $self->{logger};
    }
}

#----------------------------------------------------------------------

sub get_secondary_integrity_filename {
    my($self, %params) = @_;
    my $filename = $params{filename};

    return dirname($filename) . '/integrity_check___' . basename($filename);
}

#----------------------------------------------------------------------

sub update_cluster_state {
    my($self) = @_;
    $self->{is_cluster_secondary} = F5::Cluster::Utils::is_clustering_enabled() && ! F5::Cluster::Utils::is_master_blade();
    $self->{is_cluster_primary} = F5::Cluster::Utils::is_clustering_enabled() && F5::Cluster::Utils::is_master_blade();
    $self->{slot_number} = F5::Cluster::Utils::get_slot_number();
    if ($self->{is_cluster_primary}) {
        $self->{logger}->info("Current device is a cluster primary") if $self->{logger};
    }
    elsif ($self->{is_cluster_secondary}) {
        delete $self->{current_primary_load_file};
        $self->{logger}->info("Current device is a cluster secondary") if $self->{logger};
    }
    if ($self->{slot_number}) {
        $self->{logger}->info("Current device is slot $self->{slot_number}") if $self->{logger};
    }
}

#----------------------------------------------------------------------

sub load_primary_from_current_secondary {
    my($self) = @_;

    my $load_file_fh;
    my $insert_record_count = 0;

    while (1) {
        if (! $self->{current_primary_load_file}) {
            last;
        }

        my $sync_filename = $self->{current_primary_load_file}->{sync_filename};
        if (! $load_file_fh) {
            if (! open($load_file_fh, '<', $sync_filename)) {
                my $cause = $! || 'unknown';
                $self->{logger}->info("Could not open $sync_filename, reason: $cause") if $self->{logger};
                # Sync file could not be read -- skip, and go to next
                $self->update_primary_loadable_files(final_status_for_current => 'partial');
                $load_file_fh = undef;
                delete $self->{messagepack_unpacker};
                next;
            }

            if (my $seek_pos = $self->{current_primary_load_file}->{seek_pos}) {
                if (! seek($load_file_fh, $seek_pos, 0)) {
                    $self->{logger}->info("Could not seek $sync_filename to $seek_pos") if $self->{logger};
                    close $load_file_fh;
                    # Sync file could not be seeked -- skip, and go to next
                    $self->update_primary_loadable_files(final_status_for_current => 'partial');
                    $load_file_fh = undef;
                    delete $self->{messagepack_unpacker};
                    next;
                }
            }
        }

        if (eof($load_file_fh)) {
            close $load_file_fh;
            $self->update_primary_loadable_files(final_status_for_current => 'complete');
            $load_file_fh = undef;
            delete $self->{messagepack_unpacker};
            next;
        }

        if (! $self->{messagepack_unpacker}) {
            $self->{messagepack_unpacker} = Data::MessagePack::Stream->new();
        }

        $self->{logger}->debug("Loading sync record from $self->{current_primary_load_file}->{sync_filename}") if $self->{logger};

        my $buf;
        read($load_file_fh, $buf, 1024);
        $self->{messagepack_unpacker}->feed($buf);

        while ($self->{messagepack_unpacker}->next) {
            my $unpacked_message = $self->{messagepack_unpacker}->data;

            my $request_type = $unpacked_message->{request_type};
            my $raw_protobuf = $unpacked_message->{raw_protobuf};

            $self->insert_record(raw_protobuf => $raw_protobuf,
                                 request_type => $request_type,
                                 slot_number => $self->{current_primary_load_file}->{slot_number}, # Unused for now, but potentially useful
                                );

            $self->{load_from_secondary_count}++;

            $insert_record_count++;
        }

        if ($insert_record_count >= $PRIMARY_LOAD_FILE_INSERT_RATE) {
            last;
        }
    }

    if ($self->{current_primary_load_file} && $load_file_fh) {
        $self->{current_primary_load_file}->{seek_pos} = tell($load_file_fh);
        close $load_file_fh;
    }
}

#----------------------------------------------------------------------

sub update_primary_loadable_files {
    my($self, %params) = @_;

    return unless ($self->{is_cluster_primary});

    F5::Utils::File::cleanup_files(dir => $PRIMARY_ACK_TRANSFER_FILE_PATH,
                                   age => 15 * 60, # 15 minutes
                                   max_retained_files => 500,
                                  );

    if ($params{final_status_for_current} && $self->{current_primary_load_file}) {
        $self->update_secondary_final_status(%params);
    }

    find({ wanted => sub {
               return unless (-f $File::Find::name);
               return unless (basename($File::Find::name) =~ /^request_log/);

               my ($timestamp, $slot_number) = ($File::Find::name =~ /request_log__(\d+_\d+)__slot_(\d+)/);
               my $secondary_integrity_filename = $self->get_secondary_integrity_filename(filename => $File::Find::name);
               my $secondary_integrity;
               if (! -f $secondary_integrity_filename) {
                   $self->{logger}->info("Secondary file $File::Find::name is missing integrity check file -- skipping") if $self->{logger};
                   return;
               }
               else {
                   eval {
                       $secondary_integrity = YAML::XS::LoadFile($secondary_integrity_filename);
                   };
                   if ($secondary_integrity && $secondary_integrity->{file_size} &&
                       $secondary_integrity->{file_size} != -s $File::Find::name) {

                       $self->{logger}->info("Secondary file $File::Find::name does not match integrity check -- skipping") if $self->{logger};
                       return;
                   }
               }

               if ($slot_number && $slot_number == $self->{slot_number}) {
                   $self->{logger}->info("Skipping $File::Find::name (self slot id: $self->{slot_number})") if $self->{logger};
                   return;
               }

               if (! $self->{primary_loadable_files}->{$File::Find::name}) {
                   $self->{primary_loadable_files}->{$File::Find::name} = {
                                                                           sync_filename => $File::Find::name,
                                                                           status => 'pending',
                                                                           slot_number => $slot_number,
                                                                           timestamp => $timestamp,
                                                                          };
               }
           },
           no_chdir => 1,
         },
         $SECONDARY_TRANSFER_FILE_PATH);

    # Clean out older non-existing files, from loadable file queue
    for my $loadable_file (values %{$self->{primary_loadable_files}}) {
        if (! -f $loadable_file->{sync_filename} &&
            defined $loadable_file->{final_status_time} && 
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $PRIMARY_LOADABLE_FILE_QUEUE_CACHE_INTERVAL,
                                                               start_time => $loadable_file->{final_status_time}) == 0) {
            $self->{logger}->trace("Secondary file $loadable_file->{sync_filename} removed from loadable file queue") if $self->{logger};
            delete $self->{primary_loadable_files}->{$loadable_file->{sync_filename}};
        }
    }

    if ($self->{current_primary_load_file} && $self->{current_primary_load_file}->{status} eq 'loading') {
        return;
    }

    my @next_files_to_handle = sort { $a->{sync_filename} cmp $b->{sync_filename} }  grep { $_->{status} eq 'pending' } values %{$self->{primary_loadable_files}};
    while (my $next_file_to_handle = shift(@next_files_to_handle)) {
        $self->{current_primary_load_file} = $next_file_to_handle;
        if (! -f $self->{current_primary_load_file}->{sync_filename}) {
            $self->{logger}->info("Secondary file $self->{current_primary_load_file}->{sync_filename} does not exist -- skipping") if $self->{logger};
            $self->update_secondary_final_status(final_status_for_current => 'missing');
            next;
        }

        $self->{current_primary_load_file}->{status} = 'loading';

        my $dbh = $params{dbh} || $self->{dbh} || F5::Db::Single->instance;
        F5::DbUtils::insert_row(dbh => $dbh,
                                table_name => 'PRX.REQUEST_LOG_SYNC_FILES',
                                rec => $self->{current_primary_load_file},
                               );
        $self->{db_clean}->inserted_rows(row_count => 1,
                                         table_name => 'PRX.REQUEST_LOG_SYNC_FILES',
                                        );
        $self->{logger}->info("Begin load from secondary sync file $self->{current_primary_load_file}->{sync_filename}") if $self->{logger};
        last;
    }
}

#----------------------------------------------------------------------

sub update_secondary_final_status {
    my($self, %params) = @_;

    if (! $self->{current_primary_load_file}->{sync_filename}) {
        $self->{logger}->info("Finished load from unknown secondary sync file (final status = $params{final_status_for_current})") if $self->{logger};
        delete $self->{current_primary_load_file};
        return;
    }

    $self->{current_primary_load_file}->{status} = $params{final_status_for_current};
    $self->{current_primary_load_file}->{final_status_time} = time;

    my $dbh = $params{dbh} || $self->{dbh} || F5::Db::Single->instance;
    F5::DbUtils::update_row(dbh => $dbh,
                            table_name => 'PRX.REQUEST_LOG_SYNC_FILES',
                            rec => $self->{current_primary_load_file},
                            fields => [ qw(status final_status_time) ],
                            where_fields => { sync_filename => $self->{current_primary_load_file}->{sync_filename} },
                           );
    $self->{logger}->info("Finished load from secondary sync file $self->{current_primary_load_file}->{sync_filename} (final status = $params{final_status_for_current})") if $self->{logger};

    my $rv = open(my $outf, '>', $PRIMARY_ACK_TRANSFER_FILE_PATH . '/' . basename($self->{current_primary_load_file}->{sync_filename}));
    if ($rv) {
        print $outf "final_status_time=$self->{current_primary_load_file}->{final_status_time}\n";
        print $outf "status=$self->{current_primary_load_file}->{status}\n";
        close $outf;
    }

    delete $self->{current_primary_load_file};
}

#----------------------------------------------------------------------

sub filename_formatted_datetime {
    my($time) = @_;
    $time ||= time;
    return strftime("%Y%m%d_%H%M%S", localtime($time));
}

#----------------------------------------------------------------------

sub cleanup_secondary_sync_files {
    my($self) = @_;

    for my $dir (@CLUSTER_REQUEST_LOG_FILE_PATHS) {
        F5::Utils::File::cleanup_files(dir => $dir,
                                       age => 60 * 60, # 60 minutes
                                       logger => $self->{logger},
                                      );
    }

    if ($self->{is_cluster_secondary}) {
        find({ wanted => sub {
                   return unless (-f $File::Find::name);

                   my $secondary_sync_file = $SECONDARY_TRANSFER_FILE_PATH . '/' . basename($File::Find::name);
                   if (-f $secondary_sync_file) {
                       unlink($secondary_sync_file);
                       $self->{logger}->info("Deleting request log sync file $secondary_sync_file") if $self->{logger};

                       my $secondary_integrity_filename = $self->get_secondary_integrity_filename(filename => $secondary_sync_file);
                       if (-f $secondary_integrity_filename) {
                           unlink($secondary_integrity_filename);
                           $self->{logger}->debug("Deleting request log sync integrity file $secondary_integrity_filename") if $self->{logger};
                       }
                   }
               },
               no_chdir => 1,
             },
             $PRIMARY_ACK_TRANSFER_FILE_PATH);
    }
}

#----------------------------------------------------------------------

1;
