package F5::ThreatCampaignFile;

use strict;
use warnings;

use File::Find;
use File::Path;
use File::Copy;
use File::Basename;
use File::Slurp;
use Digest::MD5;
use YAML::XS;

# Used to call gpg externally
use IPC::Open3;
use Symbol;
use File::Spec;

use F5::BigipUtils;
use F5::DbUtils;
use F5::Deploy;
use F5::Syslog;
use F5::BigipVersionUtils;
use F5::DateTimeUtils;
use F5::ASMConfig::ASMConfig;
use F5::ThreatCampaigns;
use F5::Logger;
use F5::ASMConfig::Entity::PolicyThreatCampaign;
use F5::LockFactory;
use F5::ASMConfig::Entity::ThreatCampaign;
use F5::ASMConfig::Entity::PolicyThreatCampaignAttributes;


my $DEFAULT_TEMP_BASE_DIR = F5::Deploy::map_path("/ts/var/tmp");
my $GPG_CMD = '/usr/bin/gpg';

#----------------------------------------------------------------------

sub new {
  my($class, %params) = @_;

  my $self = {};
  bless $self, $class;
  $self->init(%params);

  return $self;
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;

  $self->{filename} = $params{filename};
  $self->{base_dir} = $params{base_dir};
  $self->{dbh} = $params{dbh};
  $self->{dbname} = $params{dbname} || 'PLC';
  $self->{output_filename} = $params{output_filename};
  $self->{gpg_sign} = (exists $params{gpg_sign} ? $params{gpg_sign} : !$params{threat_campaign_file_format});
  $self->{threat_campaign_file_format} = $params{threat_campaign_file_format} || 'XML';
  $self->{temp_base_dir} = $params{temp_base_dir} || $DEFAULT_TEMP_BASE_DIR;

  if ($self->{filename}) {
    if (! -e $self->{filename}) {
      F5::Syslog::ts_syslog( 'warning',
                             "Threat Campaign file not found (%s)",
                             $self->{filename}
                           );

      $self->{filename} = '';
    }
    else {
        $self->unzip;
    }
  }
}

#----------------------------------------------------------------------

sub unzip {
  my($self) = @_;

  my $decrypted_filename;
  my $decrypted_is_temporary = 0;

  if ($self->{gpg_sign}) {
    $decrypted_is_temporary = 1;
    my $keyring_path = F5::Deploy::map_path("/ts/share/negsig/gpg_asm_sigfile_installer");
    $decrypted_filename = "$self->{temp_base_dir}/sigfile.tmp.$$.decrypted_sigfile.tgz";

    open(my $null1, ">", File::Spec->devnull);
    open(my $null2, File::Spec->devnull);
    my $gpg_err_fh = gensym;
    my $gpg_pid = open3($null1, $null2, $gpg_err_fh, 
                        $GPG_CMD, qw(--always-trust --yes --homedir), $keyring_path,
                        '--output', $decrypted_filename,
                        '--decrypt', $self->{filename});
    my $gpg_err = join('', <$gpg_err_fh>);
    waitpid($gpg_pid, 0);
    if ($?) {
      die "Threat Campaign file is not valid\n";
    }
  }
  else {
    $decrypted_filename = $self->{filename};
  }

  $self->{base_dir} = "$self->{temp_base_dir}/sigfile.tmp.$$";
  $self->{base_dir_is_temporary} = 1;
  mkdir $self->{base_dir};
  system('tar','-zmxf', $decrypted_filename, '-C', $self->{base_dir});
  unlink $decrypted_filename if $decrypted_is_temporary;
  if ($self->{threat_campaign_file_format} eq 'BINARY') {
      $self->{filename} = glob("$self->{base_dir}/*");
  }
}

#----------------------------------------------------------------------

sub finish {
  my($self) = @_;

  my $dbh = $self->{dbh};

  my %threat_campaign_file_status = (
                         bigip_version => F5::BigipVersionUtils::bigip_version(),
                         threat_campaigns_updated_count => 0,
                         threat_campaigns_added_count => 0,
                         threat_campaigns_deleted_count => 0,
                         load_datetime => F5::DateTimeUtils::convert_epoch_to_datetime(time()),
                        );

  $threat_campaign_file_status{threat_campaign_file_timestamp} = $self->timestamp || $threat_campaign_file_status{load_datetime};

  F5::DbUtils::insert_row(dbh => $dbh,
                          table_name => "$self->{dbname}.THREAT_CAMPAIGN_FILE_STATUS",
                          rec => \%threat_campaign_file_status
                          );

  if ($self->{base_dir_is_temporary}) {
    rmtree($self->{base_dir});
  }
}

#----------------------------------------------------------------------

sub validate_version {
    my($self) = @_;

    if (! -e $self->get_timestamp_file) {
        my $error_msg = "The uploaded Threat Campaign update file does not match the current version of BIG-IP.";
        die {
              error_code   => F5::ASMConfig::ErrorCode::INVALID_SIGFILE_VERSION(),
              display_message => $error_msg,
              error_message => $error_msg,
              error_level  => 'info',
            };
    }
}

#----------------------------------------------------------------------

sub get_update_file_date {
    my($self) = @_;
    return F5::ThreatCampaigns::get_update_file_date(dbh => $self->{dbh});
}

#----------------------------------------------------------------------

sub validate_timestamp {
    my($self, %params) = @_;

    my $current_timestamp = F5::ThreatCampaigns::get_update_file_date(dbh => $self->{dbh});
    if ($self->timestamp lt $current_timestamp && ! $params{allow_older_timestamp}) {
        my $error_msg = "Threat Campaign update file is older than currently installed version";
        die {
              error_code   => F5::ASMConfig::ErrorCode::OLDER_SIGFILE_TIMESTAMP(),
              display_message => $error_msg,
              error_message => $error_msg,
              error_level  => 'info',
            };
    }
}

#----------------------------------------------------------------------

sub get_timestamp_file {
    my($self) = @_;
    return "$self->{base_dir}/timestamps/" . F5::BigipVersionUtils::bigip_version() . ".yaml";
}

#----------------------------------------------------------------------

sub timestamp {
  my($self) = @_;

  if ($self->{timestamp}) {
      return $self->{timestamp};
  }
  my $timestamp;
  my $timestamp_filename = $self->get_timestamp_file;
  if (-e $timestamp_filename) {
      my $timestamp_file = YAML::XS::LoadFile($timestamp_filename);
      $timestamp = $timestamp_file->{last_update};
  }
  else {
      ($timestamp) = $self->{dbh}->selectrow_array("SELECT MAX(last_updated) FROM $self->{dbname}.THREAT_CAMPAIGNS");
  }
  $self->{timestamp} = $timestamp;
  return $timestamp;
}

#----------------------------------------------------------------------

sub load_threat_campaigns {
    my($self, %params) = @_;
    
    my $dir = "$self->{base_dir}/campaigns";
    my $dbh = $self->{dbh};
    my $logger = $self->{logger} || F5::Logger->new('/ts/log/threat_campaigns.log');
    my $threat_campaign_file_format = $self->{threat_campaign_file_format} || 'XML';
    my $filename = $self->{filename};

    my $locker = F5::LockFactory::get_locker(lock_filename => $F5::DbUtils::LOCK_FILE,
                                             logger => $logger,
                                            );
    $locker->lock();

    if ($threat_campaign_file_format eq 'XML' && ! -d $dir ) {
        my $error_msg = "Cannot load Threat Campaigns from unknown directory ($dir)";
        $logger->error($error_msg) if $logger;
        $locker->unlock();

        # message for die should be less informative because it will shown to end user in GUI
        die "Cannot load Threat Campaigns\n";
    }

    my @threat_campaigns = F5::ASMConfig::Entity::ThreatCampaign->retrieve_all(dbh => $dbh,
                                                                               logger => $logger,);
    my %existing_threat_campaigns = map { ( $_->{name} => $_ ) } @threat_campaigns;

    my $load_rv;
    if ($threat_campaign_file_format eq 'XML') {
        $load_rv = $self->load_from_xml(%params,
                                        dir => $dir,
                                        dbh => $dbh,
                                        logger => $logger,
                                        existing_threat_campaigns => \%existing_threat_campaigns);

    }
    elsif ($threat_campaign_file_format eq 'BINARY') {
        $load_rv = $self->load_from_binary(%params,
                                           dbh => $dbh,
                                           filename => $filename,
                                           logger => $logger,
                                           existing_threat_campaigns => \%existing_threat_campaigns);
    }
    return unless $load_rv;

    my $update_count = 0;
    my $add_count = 0;
    my $current_total_count = 0;
    my @new_threat_campaigns;
    my @match_rule_modified_campaign_ids;
    my @threat_campaigns_modified;

    for my $modified_threat_campaign (@$load_rv) {
        $current_total_count++;
        if ($modified_threat_campaign->{change_level}) {
            if ($modified_threat_campaign->{change_level} == F5::ASMConfig::ThreatCampaignChangeLevel::THREAT_CAMPAIGN_CHANGE_LEVEL_UPDATED) {
                $update_count++;
                if ($modified_threat_campaign->{rules_changed}) {
                    push(@match_rule_modified_campaign_ids, $modified_threat_campaign->{threat_campaign_id});
                }
            }
            if ($modified_threat_campaign->{change_level} == F5::ASMConfig::ThreatCampaignChangeLevel::THREAT_CAMPAIGN_CHANGE_LEVEL_ADDED) {
                $add_count++;
                push (@new_threat_campaigns, $modified_threat_campaign->{threat_campaign});
            }
            push (@threat_campaigns_modified, F5::ASMConfig::ThreatCampaignFileModifiedThreatCampaign->new(
                                                  change_level => $modified_threat_campaign->{change_level},
                                                  threat_campaign_id => $modified_threat_campaign->{threat_campaign_id} ,
                                                  threat_campaign_name => $modified_threat_campaign->{threat_campaign_name},
                                                  threat_campaign_display_name => $modified_threat_campaign->{threat_campaign_display_name}));
        }
    }

    F5::ThreatCampaigns::create_policy_threat_campaigns(dbh => $dbh, new_threat_campaigns => \@new_threat_campaigns, logger => $logger);

    F5::ThreatCampaigns::campaign_rules_updated(%params, dbh => $dbh, match_rule_modified_campaign_ids => \@match_rule_modified_campaign_ids, );

    my $delete_count = 0;
    for my $threat_campaign_name (keys %existing_threat_campaigns) {
        if (! $existing_threat_campaigns{$threat_campaign_name}->{_preserved}) {
            my $threat_campaign = $existing_threat_campaigns{$threat_campaign_name};
            #We do not add the deleted campaigns to the threat_campaigns_modified return to avoid foreign key failures.
            $logger->debug("Remove threat campaign $existing_threat_campaigns{$threat_campaign_name}->{threat_campaign_id}") if $logger;
            $existing_threat_campaigns{$threat_campaign_name}->delete();
            $delete_count++;
        }
    }

    $locker->unlock();

    $logger->info("Finish Loading Threat Campaigns") if $logger;

    return {
            completed_successfully => 1,
            threat_campaigns_count => $current_total_count,
            threat_campaigns_updated_count => $update_count,
            threat_campaigns_added_count => $add_count,
            threat_campaigns_deleted_count => $delete_count,
            threat_campaigns_modified => \@threat_campaigns_modified,
           };
}

#----------------------------------------------------------------------------------------

sub load_from_xml {
    my ($self, %params) = @_;

    my $dir = $params{dir};
    my $dbh = $params{dbh};
    my $logger = $params{logger};
    my $existing_threat_campaigns = $params{existing_threat_campaigns};

    $logger->info("Loading Threat Campaigns: $dir") if $logger;

    my @loaded_threat_campaigns;
    find({ wanted => sub {
            my $filename = $File::Find::name;
            return unless (-f $filename);
            return unless ($filename =~ /\.xml$/);
            $logger->debug("Parsing Threat Campaign file $filename") if $logger;

            my $twig = F5::XML::Twig->new(
                twig_handlers => {
                    'threat_campaign' => sub {
                        my($twig, $elm) = @_;
                        my $threat_campaign;
                        eval {
                            $threat_campaign = F5::ASMConfig::Entity::ThreatCampaign->new(dbh => $dbh, xml => $elm, logger => $logger);
                        };
                        if ($@) {
                            $logger->error("Could not load Threat Campaign $filename ($@)") if $logger;
                            return;
                        }

                        if ($threat_campaign->{status} == F5::ASMConfig::ThreatCampaignStatus::THREAT_CAMPAIGN_STATUS_ACTIVE() &&
                              ! (ref($threat_campaign->{match_rules}) && scalar(@{$threat_campaign->{match_rules}})) ) {

                            $logger->info("Threat Campaign file $filename has no match rules") if $logger;
                            return;
                        }

                        my $existing_threat_campaign = $existing_threat_campaigns->{ $threat_campaign->{name} };
                        my $load_rv = $self->load_threat_campaign(
                                               dbh => $dbh,
                                               logger => $logger,
                                               existing_threat_campaign => $existing_threat_campaign,
                                               threat_campaign => $threat_campaign,
                                              );
                        return unless $load_rv;
                        push (@loaded_threat_campaigns, $load_rv);
                    }
                }
            );

            my $rc = $twig->safe_parsefile($filename);
            if (!$rc) {
                my $error_msg = "Cannot load Threat Campaign ($filename): $@";
                $logger->error($error_msg) if $logger;
                return;
            }
        }
    }, $dir);

    return \@loaded_threat_campaigns;
}

#----------------------------------------------------------------------------------------

sub load_from_binary {
    my ($self, %params) = @_;

    my $filename = $params{filename};
    my $dbh = $params{dbh};
    my $logger = $params{logger};
    my $existing_threat_campaigns = $params{existing_threat_campaigns};

    $logger->info("Loading Threat Campaigns: $filename") if $logger;

    my $mp = Data::MessagePack->new();
    my $packed_data = F5::Utils::File::file_slurp($filename);
    my $threat_campaign_file = $mp->unpack($packed_data);
    my @loaded_threat_campaigns;

    if ( defined $threat_campaign_file->{revisionDatetime} ) {
        $self->{timestamp} = F5::DateTimeUtils::convert_epoch_to_datetime( F5::DateTimeUtils::convert_rfc3339_to_epoch ($threat_campaign_file->{revisionDatetime}) );
    }


    if (F5::BigipUtils::is_feature_enabled('validate_software_version')) {
        my $current_bigip_ver = F5::BigipVersionUtils::bigip_version();

        if ( defined $threat_campaign_file->{softwareVersion} ) {
            if (($threat_campaign_file->{softwareVersion} !~ /^\d{1,4}\.\d{1,4}\.\d{1,4}(\.|$)/) ){
                die {error_message => "'softwareVersion' format is incorrect."};
            }
            elsif ( F5::BigipVersionUtils::compare_versions(version1 => $threat_campaign_file->{softwareVersion},
                                                            version2 => $current_bigip_ver) > 0 ){
                die {error_message => "'softwareVersion' should be not higher than $current_bigip_ver"};
            }
        }
    }

    if (defined($threat_campaign_file->{readme})) {
        $self->{readme} = $threat_campaign_file->{readme};
    }

    for my $object (@{$threat_campaign_file->{threat_campaigns}}) {
        #loaded binary objects are packaged from other systems and may need their internal campaign id removed
        delete $object->{threat_campaign_id};
        my $threat_campaign = F5::ASMConfig::Entity::ThreatCampaign->new(dbh => $dbh, logger => $logger, object => $object);

        my $existing_threat_campaign = $existing_threat_campaigns->{ $threat_campaign->{name} };
        my $load_rv = $self->load_threat_campaign(
                                dbh => $dbh,
                                logger => $logger,
                                existing_threat_campaign => $existing_threat_campaign,
                                threat_campaign => $threat_campaign,
                              );
        return unless $load_rv;
        push (@loaded_threat_campaigns, $load_rv);
    }

    return \@loaded_threat_campaigns;
}

#----------------------------------------------------------------------------------------

sub load_threat_campaign {
    my ($self, %params) = @_;
    my $dbh = $params{dbh};
    my $existing_threat_campaign = $params{existing_threat_campaign};
    my $threat_campaign = $params{threat_campaign};
    my $logger = $params{logger};

    my $change_level;
    my $rules_changed;
    if ($existing_threat_campaign) {
        $rules_changed = F5::ASMConfig::Entity::ThreatCampaign->has_value_changed(
                                                                                  existing_obj => $existing_threat_campaign,
                                                                                  store_obj => $threat_campaign,
                                                                                  check_fields => 'match_rules',);
        $existing_threat_campaign->{_preserved} = 1;
        my $rv = $existing_threat_campaign->store_update(store_obj => $threat_campaign, suppress_events => 1, logger => $logger);
        my $db_change_made = $rv->{db_change_made};
        if ($db_change_made ) {
            $threat_campaign = $existing_threat_campaign;
            $change_level = F5::ASMConfig::ThreatCampaignChangeLevel::THREAT_CAMPAIGN_CHANGE_LEVEL_UPDATED;
        }
    }
    else {
        eval {
            $threat_campaign->store_add(store_obj => $threat_campaign);
        };
        my $err = $@;
        if ($err) {
            if (ref($err)) {
                $err = $err->{error_message};
            }
            $logger->error("Fail to create Threat Campaign \"$threat_campaign->{display_name}\": $err") if $logger;
            return;
        }
        $change_level = F5::ASMConfig::ThreatCampaignChangeLevel::THREAT_CAMPAIGN_CHANGE_LEVEL_ADDED;
    }

    return { change_level => $change_level,
             threat_campaign_id => $threat_campaign->{threat_campaign_id},
             threat_campaign_name => $threat_campaign->{name},
             threat_campaign_display_name => $threat_campaign->{display_name},
             rules_changed => $rules_changed,
             threat_campaign => $threat_campaign,
            };
}

#----------------------------------------------------------------------

sub create_threat_campaign_file {
  my ($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $threat_campaign_iterator = $params{threat_campaign_iterator};

  require Data::Structure::Util;

  my @threat_campaigns = $threat_campaign_iterator->get_all();

  F5::ObjUtils::remove_unfreezable_references(\@threat_campaigns);
  Data::Structure::Util::unbless(\@threat_campaigns);

  my $threat_campaign_file = { 
                               threat_campaigns => \@threat_campaigns,
                             };

  my $last_threat_campaign_file = $self->{dbh}->selectrow_hashref(<<EndQuery);
SELECT threat_campaign_file_timestamp
    FROM $self->{dbname}.THREAT_CAMPAIGN_FILE_STATUS
ORDER BY threat_campaign_file_status_id DESC
   LIMIT 1
EndQuery

  $threat_campaign_file->{revisionDatetime} = F5::DateTimeUtils::convert_epoch_to_rfc3339(F5::DateTimeUtils::convert_datetime_to_epoch($last_threat_campaign_file->{threat_campaign_file_timestamp}));

  my $dir = "$self->{temp_base_dir}/threat_campaign_file.tmp.$$";
  my $tmp_filename = "threat_campaigns.bin";
  mkpath $dir;

  open(my $threat_campaign_file_fh, ">", "$dir/$tmp_filename");
  my $mp = Data::MessagePack->new();
  my $packed_data = $mp->pack($threat_campaign_file);
  print $threat_campaign_file_fh $packed_data;
  close $threat_campaign_file_fh;

  delete $threat_campaign_file->{threat_campaigns};

  my $zipped_filename = $tmp_filename . '.tgz';
  my $metadata_filename = "threat_campaign_update.yaml";

  $threat_campaign_file->{checksum} = Digest::MD5::md5_base64($packed_data);
  $threat_campaign_file->{filename} = $zipped_filename;
  YAML::XS::DumpFile("$dir/$metadata_filename", $threat_campaign_file);

  my @cmd = ('tar', '-zcf', "$dir/$zipped_filename", '-C', $dir,  $tmp_filename);
  my $res =  F5::SystemUtils::run_system_tool( command_line_params => \@cmd );

  @cmd = ('tar', '-zcf', $self->{output_filename}, '-C', $dir, $metadata_filename, $zipped_filename);
  $res =  F5::SystemUtils::run_system_tool( command_line_params => \@cmd );

  rmtree($dir);
}

#----------------------------------------------------------------------

1;

