package F5::Utils::File;

use strict;
use warnings;
use File::stat;
use File::Find;
use File::Basename;
use File::Copy;
use File::Path;
use Fcntl ':mode';
use F5::ASMConfig::Handler::TaskBase;
use F5::Utils::Encode;
use F5::Utils::Rest;
use F5::BigipUtils;
use F5::SystemUtils;
use Digest::MD5;


our $DEFAULT_TIMEOUT = 20;

#----------------------------------------------------------------------
sub cleanup_files {
    my(%params) = @_;

    my $logger = $params{logger};
    my $start_time = time();
    my $consider_future_files_as_older = defined $params{consider_future_files_as_older} ? $params{consider_future_files_as_older} : 0;

    my @files;
    if ($params{dir}) {
        find({
              wanted => sub {
                  # Do not remove root directory being cleaned
                  return if ($File::Find::name eq $params{dir});

                  push(@files, { filename => $File::Find::name });
              },
              bydepth => 1,
              no_chdir => 1,
              follow => 0,
             },
             $params{dir});
    }
    else {
        @files = map { { filename => $_ } } glob($params{glob_pattern});
    }

    for my $file (@files) {
        my $stat = stat($file->{filename});
        if (! $stat) {
            # File has vanished in the meantime -- silently mark for deletion to remove it from list
            $file->{skip} = 1;
            $logger->debug("Marking file $file->{filename} for cleanup (no longer present on filesystem)") if $logger;
            next;
        }
        $file->{mtime} = $stat->mtime;
        if ($file->{mtime} > $start_time && $consider_future_files_as_older) {
            # If consider_future_files_as_older is set, files with future timestamps will be treated as
            # much older, but will still be sorted according to their (fake) time order
            $file->{mtime} = $file->{mtime} - $start_time;
        }
        $file->{size} = $stat->size;
        $file->{is_dir} = S_ISDIR($stat->mode);
        $file->{clean_directory} = $params{clean_dirs};
    }

    my $skipped_num = 0;
    my $skipped_size = 0;

    if($params{min_retained_files_in_dir}) {
        my @files_by_mtime = sort { $b->{mtime} <=> $a->{mtime} } @files;
        my $from_dir;
        for my $i (0..$#files_by_mtime) {
            my $file = $files_by_mtime[$i];
            my $file_name = $file->{filename};
            my $dir = dirname($file_name);
            $from_dir->{$dir} ||= 0;
            if($from_dir->{$dir} < $params{min_retained_files_in_dir}) {
                $from_dir->{$dir}++;
                $file->{skip} = 1;
                $skipped_num++;
                $skipped_size += $file->{size};
                $logger->debug("Skipping file $file->{filename} (minimum retained files in directory)") if $logger;
            }
        }
    }

    if ($params{min_age}) {
        for my $file (@files) {
            if ((time - $file->{mtime}) < $params{min_age}) {
                $file->{skip} = 1;
                $skipped_num++;
                $skipped_size += $file->{size};
                $logger->debug("Skipping file $file->{filename} (minimum age)") if $logger;
            }
        }
    }

    if ($params{retain_dirs}) {
        for my $retain_dir (@{$params{retain_dirs}}) {
            my $retain_dir_regex = qr/^$retain_dir/;
            for my $file (@files) {
                if ($file->{filename} =~ $retain_dir_regex) {
                    $logger->debug("Skipping file $file->{filename} (found in retain dir $retain_dir)") if $logger;
                    $file->{skip} = 1;
                    $skipped_num++;
                    $skipped_size += $file->{size};
                }
            }
        }
    }

    if ($skipped_num) {
        if ($params{min_retained_files}) {
            $params{min_retained_files} -= $skipped_num;
        }
        if ($params{max_size}) {
            $params{max_size} -= $skipped_size;
            $logger->debug("Skipped files size was $skipped_size, remaining max size is $params{max_size}") if $logger;
        }
        if ($params{max_retained_files}) {
            $params{max_retained_files} = $params{max_retained_files} > $skipped_num ? $params{max_retained_files} - $skipped_num : 0;
        }
    }

    @files = grep { ! $_->{skip} } @files;

    if (defined $params{age}) {
        for my $file (@files) {
            if ((time - $file->{mtime}) > $params{age}) {
                $file->{to_delete} = 1;
                $logger->debug("Marking file $file->{filename} for cleanup (exceeds age limit)") if $logger;
            }
        }
        delete_marked_files(\@files);
    }

    if (defined $params{max_size}) {
        my @files_by_mtime = sort { $b->{mtime} <=> $a->{mtime} } @files;
        my $total_size = 0;

        for my $i (0..$#files_by_mtime) {
            my $file = $files_by_mtime[$i];
            $total_size += $file->{size};
            $logger->trace("Total thus far = $total_size -- checking $file->{filename} with size $file->{size}") if $logger;

            if (defined $params{min_retained_files} && ($i + 1) <= $params{min_retained_files}) {
                next;
            }

            if ($total_size > $params{max_size}) {
                $file->{to_delete} = 1;
                $logger->debug("Marking file $file->{filename} for cleanup (max disk usage exceeded)") if $logger;
            }
        }
        delete_marked_files(\@files);
    }

    if (defined($params{max_retained_files}) && scalar @files > $params{max_retained_files}) {
        my @files_by_mtime = sort { $b->{mtime} <=> $a->{mtime} } @files;

        for my $i ($params{max_retained_files} .. $#files_by_mtime) {
            $files_by_mtime[$i]->{to_delete} = 1;
            $logger->debug("Marking file $files_by_mtime[$i]->{filename} for cleanup (max retained file count exceeded)") if $logger;
        }
        delete_marked_files(\@files);
    }
}

#----------------------------------------------------------------------

sub delete_marked_files {
    my($files) = @_;

    my @files_to_delete = sort { $b->{filename} cmp $a->{filename} }  grep { $_->{to_delete} } @$files;
    for my $file (@files_to_delete) {
        if ($file->{is_dir}) {
            if ($file->{clean_directory}) {
                rmtree($file->{filename});
            }
            # Only remove empty directories
            if (! glob("$file->{filename}/*")) {
                rmdir($file->{filename});
            }
        }
        else {
            unlink($file->{filename});
        }
    }

    @$files = grep { ! $_->{to_delete} } @$files;
}

#----------------------------------------------------------------------

sub touch {
    my ($file) = @_;
    my $now = time;
    my $fh;
    utime ($now, $now, $file)
       || (open ($fh, ">>", $file) && close($fh))
       || die ("Couldn't touch file: $!\n");

}

#----------------------------------------------------------------------
sub file_slurp {
    my($filename) = @_;
    open(my $inf, '<', $filename);
    my($retval) = join('', <$inf>);
    close $inf;

    return $retval;
}

#----------------------------------------------------------------------

sub file_dump {
    my(%params) = @_;
    my $filename = $params{filename};
    my $write_options = $params{write_options} || '';

    open(my $outf, ">$write_options" , $filename) || die "Failed to write to $filename ($!)";
    print $outf $params{contents} || die "Failed to write to $filename ($!)";
    close $outf || die "Failed to write to $filename ($!)";

    return 1;
}

#----------------------------------------------------------------------

sub get_file_md5 {
    my $filename = $_[0];
    if (open(my $inf, '<', $filename)) {
        return Digest::MD5->new->addfile($inf)->hexdigest;
    }
    die "Failed to read $filename ($!)";
}

#----------------------------------------------------------------------

sub get_data_from_link {
    my(%params) = @_;

    my $link = $params{link};
    my $user = $params{user} || '';
    my $parse_json = $params{parse_json};
    my $basic_auth = $params{basic_auth};

    my $file = F5::Utils::File::get_file_from_link(link => $link,
                                                   user => $user,
                                                   basic_auth => $basic_auth);
    my $filepath = $file->{filepath};
    my $rv = {};

    local $SIG{ALRM} = sub { die {error_message => "Could not access $filepath"} };
    alarm 2;
    if ($parse_json) {
        $rv = F5::YAML::Schema::Utils::parse_json(filename => $filepath);
    }
    else {
        $rv->{data} = file_slurp($filepath);
        $rv->{size} = length($rv->{data});
        $rv->{filename} = $file->{filename};
        $rv->{completed_successfully} = 1;
    }
    alarm 0;

    if (!$rv->{completed_successfully}) {
       my $error_message = $file->{error_message} || $rv->{error_message};
       die {error_message => "Failed parsing file : $error_message."};
    }

    return $rv;
}

#----------------------------------------------------------------------

sub get_file_from_link {
    my(%params) = @_;

    my $link = $params{link};
    my $user = $params{user} || '';
    my $basic_auth = $params{basic_auth};
    my $timeout = defined ($params{timeout}) ? $params{timeout} : $DEFAULT_TIMEOUT;

    my $is_external_link = 0;
    my $resource_path = "";

    if ($link =~ m/^https?:\/\/([^\?]*)/i) {
        $resource_path = $1;
        $is_external_link = 1;
    }
    elsif ($link =~ m/^file:\/\/(.*)/i) {
        $resource_path = $1;
    }
    else {
        die {error_message => "Wrong link format, should be 'file:///FILE_NAME' or 'http(s)://DOMAIN_NAME/FILE_NAME'"};
    }

    my $filename = basename($resource_path);
    my $dirname;
    if (!$is_external_link) {
        $dirname = dirname($resource_path);
        if (length($dirname) == 1) {
            #Ignore "." or "/" directory.
            undef $dirname;
        }
        elsif ($dirname =~ /^[^\/]/) {
            # Path must be specified with three slashes, so dirname must start with /
            die {error_message => "Wrong link format, should be 'file:///FILE_NAME'"};
        }
    }

    my $file;
    if (F5::BigipUtils::is_feature_enabled('include_username_in_uploaded_filenames')) {
        $file = F5::Utils::Rest::get_full_file_path(username => $user, filename => $filename, dir => $dirname );
    }
    else {
        $file = F5::Utils::Rest::get_full_file_path(filename => $filename, dir => $dirname );
    }

    # Download a file from an external link, standalone or CMI active will download the file from the link,
    # the CMI standby unit will read the file from the local sync directory.
    if ($is_external_link) {
        my $ua = F5::LWP::UserAgent->new(timeout => $timeout);
        my $http_request = HTTP::Request->new('GET', $link);

        if (defined $basic_auth && ref($basic_auth) eq 'HASH' &&
            exists $basic_auth->{user} && exists $basic_auth->{passwordBase64}) {
            my $username = $basic_auth->{user};
            my $password = F5::Utils::Encode::decode_base64($basic_auth->{passwordBase64});
            $http_request->authorization_basic($username, $password);
        }
        my $response = $ua->request($http_request);

        if (! $response->is_success) {
           my $err_msg = $response->status_line;
           die {error_message => "Failed http GET request to '$link', status: $err_msg"};
        }
        F5::Utils::File::file_dump(filename => $file,
                                   contents => $response->content);
    }

    if (! -f $file ) {
        die {error_message => "Link file '$file' is not a proper file"};
    }

    my $rv = {};
    $rv->{filepath} = $file;
    $rv->{filename} = $filename;

    if ($dirname) {
       #In case of accessing a file on the local system, always display the same error with no content for security.
       $rv->{error_message} = "Link file '$file' is not a proper file";
    }

    return $rv;
}

#----------------------------------------------------------------------

sub remove_files_only {
  my($dir) = @_;

  find({ no_chdir => 1,
         follow => 1,
         follow_skip => 2,
         wanted => sub {
           my $filename = $File::Find::name;

           if (-f $filename) {
             unlink $filename;
           }
         },
       },
       $dir);
}

#----------------------------------------------------------------------

sub copy_files_recursively {
    my(%params) = @_;

    return move_files_recursively(%params,
                                  file_action => 'copy',
                                 );
}

#----------------------------------------------------------------------

sub move_files_recursively {
  my(%params) = @_;

  my $from_dir = $params{from_dir};
  my $to_dir = $params{to_dir};
  my $file_action = $params{file_action} || 'move';

  find({ no_chdir => 1,
         follow => 1,
         follow_skip => 2,
         wanted => sub {
           my $from_filename = $File::Find::name;

           return unless (-f $from_filename);

           my $to_filename = $from_filename;
           $to_filename =~ s|^$from_dir|$to_dir|;

           unless (-d dirname($to_filename)) {
             mkpath(dirname($to_filename));
           }

           if ($file_action eq 'move') {
               move($from_filename, $to_filename);
           }
           elsif ($file_action eq 'copy') {
               copy($from_filename, $to_filename);
           }
         },
       },
       $from_dir);
}

#----------------------------------------------------------------------
# Due to a minor bug in rmtree in File::Path v1.08, a spurious warning is
# printed, due to failure to chmod a directory owned by another user
# (even if write permission is granted). Thus, delete_files_recursively
# replicates rmtree simplistically.
sub delete_files_recursively {
  my($root) = @_;

  find({
        no_chdir => 1,
        bydepth => 1,
        wanted => sub {
          if (-d $File::Find::name) {
            rmdir($File::Find::name) or
              warn "Could not remove directory $File::Find::name\n";
          }
          else {
            unlink($File::Find::name) or
              warn "Could not remove file $File::Find::name\n";
          }
        }},
       $root);
}

#----------------------------------------------------------------------

sub get_filetype {
    my (%params) = @_;
    my $filename = $params{filename};
    my $valid_types = $params{valid_types};

    my $rv = {completed_successfully => 0};

    if ($filename && -f $filename) {
        my @cmd = ('/usr/bin/file', $filename);
        my $res =  F5::SystemUtils::run_system_tool( command_line_params => \@cmd,
                                                     timeout => 60,
                                                    );
        if ($res && $res->{completed_successfully}) {
            my $filetype_check_output = $res->{standard_output};
            $filetype_check_output =~ s/$filename//;
            chomp($filetype_check_output);
            $rv->{filetype} = $filetype_check_output;
            $rv->{completed_successfully} = 1;
        }
        elsif ($res) {
            $rv->{error_message} = $res->{standard_error};
        }
    }

    return $rv;
}


#----------------------------------------------------------------------

sub check_filetype {
    my (%params) = @_;
    my $filename = $params{filename};
    my $valid_types = $params{valid_types};
    my $rv = get_filetype(%params);

    if ($rv->{completed_successfully} && $rv->{filetype} && $valid_types) {
        if (! grep { index($rv->{filetype}, $_) >= 0 } @$valid_types) {
            $rv->{completed_successfully} = 0;
            $rv->{error_message} =  "Invalid file '$filename', expected file format(s): '" . join (',', @$valid_types) . "'";
        }
    }
    return $rv;
}

#----------------------------------------------------------------------

sub extract_archive {
    my (%params) = @_;
    my $filename = $params{filename};
    my $source_dir = $params{source_dir};
    my $target_dir = $params{target_dir};

    my $full_path = "$source_dir/$filename";
    #File must be tarfile for now
    my $filetype_rv = check_filetype(filename => $full_path, valid_types => ['gzip', 'tar archive']);
    if (!$filetype_rv->{completed_successfully}) {
        return $filetype_rv;
    }
    my @cmd = ('tar', 'xf', $full_path, '-C', $target_dir);
    my $res = F5::SystemUtils::run_system_tool( command_line_params => \@cmd );
    return $res;
}

#----------------------------------------------------------------------
1;
