package F5::Utils::Rest;

use Digest::MD5;
use F5::ModuleLoader;
use F5::ASMConfig::ASMConfig;
use F5::ASMConfig::Cache;
use F5::ASMConfig::Handler::TaskBase;
use F5::Utils::URI;
use F5::BigipUtils;
use File::Basename;
use File::Path;
use Encode;
use Parse::RecDescent;
use LWP::UserAgent;
use HTTP::Request;
use JSON -convert_blessed_universally;
use utf8;

use strict;
use warnings;

sub to_camel_case {
    my ($name) = @_;
    $name = lcfirst($name);
    $name =~ s/[_\-\s]+(\w)/\U$1/g;
    return $name;
}

#----------------------------------------------------------------------

my $ERROR_KIND = ':resterrorresponse';

sub local_request {
    my (%params) = @_;
    my $uri        = $params{uri};
    my $method     = $params{method} || 'GET';
    my $body       = $params{body};
    my $host       = $params{host} || '127.0.0.1';
    my $user       = defined($params{user}) ? $params{user} : 'admin';
    my $port       = defined($params{port}) ? $params{port} : 8100;
    my $proto      = $params{proto} || 'http';
    my $parameters = $params{parameters};
    my $encode_body = defined($params{encode_body}) ? $params{encode_body} : ref ($body);
    my $parse_response = defined($params{parse_reponse}) ? $params{parse_reponse} : 1;
    my $parse_error    = defined($params{parse_error}) ? $params{parse_error} : 1;

    my $logger = $params{logger};

    $uri = "$proto://$user\@$host:$port$uri";
    if ($parameters) {
        $uri .= '?' . F5::Utils::URI::build_query_string($parameters);
    }
    my $ua = LWP::UserAgent->new;
    my $request = HTTP::Request->new($method => $uri);
    if ($body) {
        if ($encode_body) {
            my $json = JSON->new->convert_blessed->allow_blessed;
            $body = $json->encode($body);
        }
        $request->content($body);
    }
    $logger->debug("Making local REST call to $uri") if $logger;
    $logger->trace($request) if $logger;
    my $response = $ua->request($request);
    $logger->trace($response) if $logger;
    my $rv = $response->content;
    if ($response->is_success || ($parse_error && $rv =~ /$ERROR_KIND/)) {
        my $json_obj;
        eval {
            if (!$response->is_success || $parse_response) {
                $json_obj = decode_json($rv);
            }
        };
        if ($json_obj && $json_obj->{kind} && $json_obj->{kind} eq $ERROR_KIND) {
            my $message = $json_obj->{message} || $json_obj->{errorMessage};
	    $logger->info("Failed local request with uri: $uri and body: $body. Error: $message") if $logger;
            die "$json_obj->{code}: $message\n";
        }
        if ($response->is_success) {
            return $parse_response ? $json_obj : $rv;
        }
    }

    my $error_message = "$response->status_line: $rv";
    $logger->info("Failed local request with uri: $uri and body: $body. Error: $error_message") if $logger;
    die "$error_message\n";
}
    

#----------------------------------------------------------------------

sub get_full_file_path {
    my (%params) = @_;
    my $filename = $params{filename};
    my $username = $params{username};
    my $dir;

    if (F5::BigipUtils::is_feature_enabled('standalone') && F5::BigipUtils::is_feature_enabled('allow_file_reference_by_absolute_path') && $params{dir}) {
        $dir = $params{dir};
    }
    else {
        $dir = $F5::ASMConfig::Handler::TaskBase::REST_DIR;
        mkpath($dir);
    }

    my $path = "$dir/";
    if ($username) {
        $path .= $username . "~";
    }
    $path .= $filename;
    return $path;
}

#----------------------------------------------------------------------

sub generate_uuid {
    my ($value) = @_;
    if (utf8::is_utf8($value)) {
        $value = encode('utf8',$value);
    }
    my $uuid = Digest::MD5::md5_base64($value);
    #URL Safe character replacement.
    $uuid =~ tr|+/=|\-_|;
    return $uuid;
}

#----------------------------------------------------------------------

sub extract_policy_from_link {
    my (%params) = @_;
    my $link = $params{'link'};
    my $dbh = $params{dbh};
    my $policy;
    if (my ($rest_uuid) = ($link =~ m|/policies/([^/]*)| )) {
        $rest_uuid =~ s/\?.*$//;
        $policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $dbh, policy => {rest_uuid => $rest_uuid});
    }
    return $policy;
}

#----------------------------------------------------------------------

sub extract_rest_uuid_from_link {
    my (%params) = @_;
    my $link = $params{'link'};
    my $rest_uuid = basename($link);
    $rest_uuid =~ s/\?.*$//;
    return $rest_uuid;
}

#----------------------------------------------------------------------

sub is_expanded {
    my (%params) = @_;
    my $expand_fields_ptr = $params{expand_fields};
    return unless $expand_fields_ptr;
    my @path_parts = split('/', $params{path});
   
    for my $path_part (@path_parts) {
        $expand_fields_ptr = $expand_fields_ptr->{$path_part} || $expand_fields_ptr->{$path_part . 'Reference'} || $expand_fields_ptr->{'*'};
        return unless $expand_fields_ptr;
    }
    return 1;
}

#----------------------------------------------------------------------


sub populate_uuids {
    my (%params) = @_;
    my $dbh = $params{dbh};
    my $handler = $params{asm_config_handler};
    my $policy_id = $params{policy_id};
    my $limit_db_names_list = $params{limit_db_names} || [];
    my %limit_db_names = map { ($_ => 1) } @$limit_db_names_list;

    my @rest_entities;

    if ($params{rest_entities}) {
        @rest_entities = @{$params{rest_entities}};
    }
    elsif ($handler) {
        @rest_entities = map {$_->{class}} grep { $_->{class} && defined ($_->{class}->fields()->{rest_uuid}) } values %{$handler->{element_types}};
    }
    else {
        my @entity_classes = F5::ModuleLoader::load_from_base_class('F5::ASMConfig::Entity::Base');
        @rest_entities = grep { $_->can('fields') && defined ($_->fields()->{rest_uuid}) } @entity_classes;
    }

    my $policy;

    if ($policy_id) {
        $policy = { policy_id => $policy_id };
        @rest_entities = grep { $_->policy_id_field } @rest_entities;
    }

    for my $class (@rest_entities) {
        if (%limit_db_names) {
            my $db_table = $class->db_table || '';
            $db_table =~ /^(.*?)\./;
            my $db_name = $1 || '';
            if (! $limit_db_names{$db_name}) {
                next;
            }
        }

        my $iterator = $class->retrieve_all(dbh => $dbh,
                                            policy => $policy,
                                            get_criteria => { rest_uuid => '' },
                                            for_populate_uuids => 1);
        while (my $entity = $iterator->get_next()) {
            $entity->store_update(%params,
                                  suppress_events => 1,
                                  store_obj => {},
                                  for_populate_uuids => 1,
                                  validate_value => 0,  #Empty update to populate the rest_uuid
                                 );
        }
    }
}

#----------------------------------------------------------------------

sub combine_expressions {
    my ($call, $left, $conj, $right) = @_;
    my @expressions;
    my @iter = ($left,$right);   
    my %collapsed_expressions;
    $conj = lc($conj);
    while (my $expr = shift @iter) {
        if ($expr->{expressions}) {
            if (!$conj || $conj eq $expr->{conj}) {
                # If compound expression conj is same as the other compound expression's, collapse the expressions together
                push (@iter, @{$expr->{expressions}});
            }
            else {
                push (@expressions, $expr);
            }
        }
        elsif (my $cmp = $expr->{'cmp'}) {
            if ($cmp eq "ne" || $cmp eq "eq") {
                my $key = "$expr->{field}__$cmp";
                my $existing = $collapsed_expressions{$key} || {};
                my @value_list = @{$expr->{value_list}};
                push (@value_list, @{$existing->{value_list}}) if defined($existing->{value_list});

                $collapsed_expressions{$key} = { field => $expr->{field},
                                       'cmp' => $expr->{'cmp'},
                                       value_list => \@value_list,
                                       is_grouped => $expr->{is_grouped},
                                     };
            }
            else {
                push (@expressions, $expr);
            }
        }
    }
    push (@expressions, values %collapsed_expressions);
    if (@expressions > 1) {
        if ($conj eq 'or') {
            die "Compound expressions using OR on different search fields are unsupported.\n";
        }
        return { expressions => \@expressions, conj => $conj };
    }
    else {
        return shift @expressions;
    }
}

my $grammar = q{
    CONJ : /and|or/i
    CMP : /eq|ne|ge|gt|le|lt|like|not like/i
    IN : /in/i
    NUM : /[+-]?[0-9]*[.]?[0-9]+/
    BOOL : /true|false/
      { my $bool = $item[1];
        $bool = ($bool eq 'true') ? JSON::true : JSON::false;
        $return = $bool;
      }
    STRING: /[\w\/-]+/
    NOT: /not /i
    EXISTS : /has /i
    QUOTE: /'([^']*('')*)*'/
      { my $text = $item[1];
        $text = substr $item[1], 1, -1;
        $text =~ s/''/'/g;
        $return = $text;
      }

    STRING_FUNCTION: /contains|endswith|startswith|substringof/i

    VALUE: NUM|BOOL|STRING|QUOTE
    VALUE_LIST : VALUE /,/ VALUE_LIST
        { $return = [ $item[1], ref($item[3]) ? @{$item[3]} : $item[3] ] }
    VALUE_LIST : VALUE
        { $return = [ $item[1] ] }

    expression : NOT(?) STRING_FUNCTION /\(/ STRING /,/ QUOTE /\)/ (CMP BOOL)(?)
        { my $boolean_matches_toggle = 1;
          if (@{$item[1]}) {
              $boolean_matches_toggle *= -1;
          }
          if (my @cmp_items = @{$item[8]}) {
              if (lc($cmp_items[0]) eq "ne") {
                  $boolean_matches_toggle *= -1;
              }
              if (lc($cmp_items[1]) eq "false") {
                  $boolean_matches_toggle *= -1;
              }
          }
          my $cmp = ($boolean_matches_toggle < 0) ? 'not like'
                                                  : 'like';
          my $value = $item[6];
          $value =~ s/([_%\\\])/\\\$1/g;   #Three slashes are evidently needed to correctly insert the escape character
          my $function = lc($item[2]);
          if ($function eq 'contains' || $function eq 'substringof') {
              $value = "%$value%";
          }
          elsif ($function eq 'endswith') {
              $value = "%$value";
          }
          elsif ($function eq 'startswith') {
              $value = "$value%";
          }
          $return = { 'field' => $item[4], 'cmp' => $cmp, value_list => [$value]};
        }
    expression : NOT(?) EXISTS STRING
        { $return = {'field' => $item[3], 'cmp' => (@{$item[1]} ? 'not exists' : 'exists'), 'value_list' => [] }; }
    expression : STRING CMP VALUE
        { $return = {'field' => $item[1], 'cmp' => lc($item[2]), 'value_list' => [ $item[3] ] } }
    expression : STRING NOT(?) IN '(' VALUE_LIST ')'
        { $return = {'field' => $item[1], 'cmp' => (@{$item[2]} ? 'ne' : 'eq'), 'value_list' => $item[5] } }
    expression : '(' expressions ')'
        {
          my $rv = $item[2];
          if ( $rv->{expressions} ) {
              $_->{is_grouped} = 1 foreach @{$rv->{expressions}};
          }
          $return = $rv;
        }


    expressions: expression CONJ expressions
        {
          $return = F5::Utils::Rest::combine_expressions(@item);
        }
    expressions: expression

    startrule : expressions
};

#----------------------------------------------------------------------

my $parser;
sub parse_odata_filter {
    my (%params) = @_;
    if (!$parser) {
        $parser = Parse::RecDescent->new($grammar);
    }
    
    my $result = $parser->startrule($params{filter}) or die "Can not parse \$filter: '$params{filter}'\n";
    my $filters = $result->{expressions} ? $result->{expressions} : [$result];
    for my $filter (@$filters) {
        my @values = @{$filter->{value_list}};
        if (@{$filter->{value_list}} > 1) {
            $filter->{'cmp'} = ($filter->{'cmp'} eq 'eq' ? 'in' : 'not in');
        }
        else {
            ($filter->{value}) = @{$filter->{value_list}};
            delete $filter->{value_list};
        }
    }
    return $filters;
}

#----------------------------------------------------------------------

my $aggregate_fields_grammar = q{
    AVG : /AVG/i
    SUM : /SUM/i
    MAX : /MAX/i
    MIN : /MIN/i
    FUNC : AVG|SUM|MAX|MIN
    AS  : /as/i
    STRING: /[\w\/-]+/
    ALIAS: AS STRING
        {
            $return = $item[2];
        }

    expression : FUNC '(' STRING ')' ALIAS(?)
        {
          my $func = ucfirst(lc($item[1]));
          my $alias = shift (@{$item[5]});
          my $name = $alias ? $alias : "$item[3]$func";
          $return = { field => $item[3], 'func' => $func, 'name' => $name };
        }

    startrule : expression(s /,/)
};

#----------------------------------------------------------------------

my $aggregate_fields_parser;
sub parse_aggregate_fields {
    my (%params) = @_;
    if (!$aggregate_fields_parser) {
        $aggregate_fields_parser = Parse::RecDescent->new($aggregate_fields_grammar);
    }
    my $aggregate_fields;
    my $parsed_aggregate_fields = $aggregate_fields_parser->startrule($params{aggregatefields}) or die "Can not parse aggregatefields: '$params{aggregatefields}'\n";
    for my $parsed_aggregate_field (@$parsed_aggregate_fields) {
        $aggregate_fields->{$parsed_aggregate_field->{field}}->{$parsed_aggregate_field->{name}} = $parsed_aggregate_field->{func};
    }
    return $aggregate_fields;
}

#


1;
