package F5::Utils::Templates;

use strict;
use warnings;

use F5::Deploy;
use F5::ASMConfig::Handler;
use F5::ASMConfig::Entity::PolicyTemplate;
use F5::DbUtils;
use F5::Logger;
use YAML::XS;

#----------------------------------------------------------

sub compile_templates {
    my (%params) = @_;
    my $logger = $params{logger} || F5::Logger->new('/ts/log/compile_templates.log');
    my $dbh = $params{dbh} || F5::DbUtils::get_dbh();
    my $client = $params{asm_config_client} ||  F5::ASMConfig::Handler->new(dbh => $dbh, username => "compile_templates", logger => $logger);

    my $allowed_templates = get_allowed_templates();

    my $restore_max_memory_size = $client->{max_memory_size};
    $client->{max_memory_size} = 1024 * 1024 * 1024;

    $client->begin_work();
    my $templates_iter = F5::ASMConfig::Entity::PolicyTemplate->retrieve_all(dbh => $dbh, get_criteria => {});

    while (my $template = $templates_iter->get_next) {
        if ($allowed_templates && ! $allowed_templates->{$template->{icontrol_name}}) {
            $logger->debug("Skipping template $template->{icontrol_name}") if $logger;
            next;
        }

        $logger->info("Compile $template->{icontrol_name}");
        my $policy = $client->add_policy($client->{session}, F5::ASMConfig::CallFlags->new(skip_mcp => 1, force_populate_rest_uuid => 1), 
                                         F5::ASMConfig::Policy->new(policy_template => $template,
                                                                    policy_name => $template->{icontrol_name},
                                                                    encoding => "utf-8",
                                                                    flg_encoding_selected => 1));
    }

    $client->rollback();

    $client->{max_memory_size} = $restore_max_memory_size;
}

#----------------------------------------------------------

sub delete_uneeded_templates {
    my (%params) = @_;
    my $logger = $params{logger} || F5::Logger->new('/ts/log/delete_templates.log');
    my $dbh = $params{dbh} || F5::DbUtils::get_dbh();
    my $client = $params{asm_config_client} ||  F5::ASMConfig::Handler->new(dbh => $dbh, username => "delete_uneeded_templates", logger => $logger);

    my $allowed_templates = get_allowed_templates();
    if (! $allowed_templates) {
        $logger->info("All templates allowed.");
        return;
    }

    my $templates_iter = F5::ASMConfig::Entity::PolicyTemplate->retrieve_all(dbh => $dbh, get_criteria => {});

    while (my $template = $templates_iter->get_next) {
        if ($allowed_templates->{$template->{icontrol_name}}) {
            $logger->info("Keep $template->{icontrol_name}");
            next;
        }

        my $template_file = $template->{template_file};
        unlink($template_file);
        $template_file =~ s/\.xml/.bin/;
        if (-e $template_file) {
            unlink($template_file);
        }
        $template->{flg_is_user_defined} = 1;
        $template->delete();
        $logger->info("Delete $template->{icontrol_name}");
    }
}

#----------------------------------------------------------

sub get_allowed_templates {
    # Get list of templates
    my $templates_file = F5::Deploy::map_path('/usr/share/ts/config/factory_templates.yaml');
    my $allowed_templates;
    if (-f $templates_file) {
        my @template_list = @{YAML::XS::LoadFile($templates_file)->{templates}};
        $allowed_templates = { map {$_ => 1} @template_list };
    }
    return $allowed_templates;
}

#----------------------------------------------------------

1;

