package F5::XMLProfile;

use strict;
use warnings;

use LWP::Simple;
use F5::Deploy;
use F5::XML::Twig;

use F5::XML::Schema::Processor;
use F5::XML::WSDL::Convert;
use File::Path;
use File::Copy;
use File::stat;
use File::Basename;
use Text::Iconv;
use F5::XML::Xpath::xpathprocessor;
use Digest::MD5;
use F5::Certificates;
use F5::ParseUri;
use F5::DbUtils;
use F5::ContentProfile;
use F5::ASMConfig::EasyClient;
use F5::ASMConfig::Entity::XmlProfile;
use List::Util;
use F5::Utils::Net;

use constant APP_ID_HTTP             => 0;
use constant APP_ID_HTTPS            => 1;

my $ARBITRARY_DIVIDER = '__//ARBITRARY_DIVIDER//__';
my $TEMP_SCHEMA_BASE_DIR = F5::Deploy::map_path("/ts/var/tmp");
my $COMPILED_SCHEMA_BASE_DIR = F5::Deploy::map_path("/ts/var/xml/compiled_schemas");
my $DEFENSE_CONFIG_BASE_DIR = F5::Deploy::map_path("/ts/var/xml/defense_config");
my $REQUEST_XPATH_BASE_DIR = F5::Deploy::map_path("/ts/var/xml/request_xpath");
my $RESPONSE_XPATH_BASE_DIR = F5::Deploy::map_path("/ts/var/xml/response_xpath");
my $REQUEST_BINARY_FULL_PATH_BASE = F5::Deploy::map_path("/ts/var/account/xml/request_xpath");
my $RESPONSE_BINARY_FULL_PATH_BASE = F5::Deploy::map_path("/ts/var/account/xml/response_xpath");
my $REQUEST_BINARY_FULL_PATH;
my $RESPONSE_BINARY_FULL_PATH;

my $XML_DEFENSE_ATTRIBUTE_ALLOW_DTDS = 9;
my $XML_DEFENSE_ATTRIBUTE_ALLOW_PROCESSING_INSTRUCTIONS = 20;

my $MAX_TOTAL_FILESIZE_COMPILED_SCHEMAS = 100_000_000;
my $MAX_TOTAL_FILESIZE_DEFENSE_PROFILES = 5_000_000;
my $MAX_TOTAL_FILESIZE_REQUEST_XPATH    = 5_000_000;
my $MAX_TOTAL_FILESIZE_RESPONSE_XPATH    = 5_000_000;
my $MAX_EXTENSION_LEN = 255;

my $SENSITIVE_XML_PARAMS_XPATH = '/policy/profile/sensitive';
my $WSS_XPATH = '/policy/profile/wss';
my $ICAP_XPATH = '/policy/profile/icap';

use constant  XML_DEFENSE_MAX_VALUE_LENGTH => 4;

my @response_elements_list = ();
my @request_elements_list = ();
my @predefined_decrypt_elements = ();
my @predefined_encrypt_elements = ();

my %prefix_namespace_mapping;

my @flags_xpath_mapping = (
    {
       flag_name    => 'flg_verify_body',
       apply_to     => 'Request',
       xpath_list   =>  [qw(/S11:Envelope/S11:Body /S12:Envelope/S12:Body)],
    },

   {
       flag_name    => 'flg_encrypt_whole_body_content',
       apply_to     => 'Response',
       encrypted_part => 'content',
       xpath_list   =>  [qw(/S11:Envelope/S11:Body /S12:Envelope/S12:Body)],
    },

   {
       flag_name    => 'flg_enforce_timestamp',
       apply_to     => 'Request',
       xpath_list   =>  [qw(//wsse:Security/wsu:Timestamp)],
    },


);


sub handle_validation_files {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $profile_id = $params{profile_id} || 0;
  my $asm_config_handler = $params{asm_config_handler}; 
  my $xml_profile = $params{xml_profile};
  my $policy = $params{policy};
  my $store_obj = F5::ASMConfig::XmlProfile->new();
                                                            
  
  if (!$xml_profile) {
      $xml_profile = F5::ASMConfig::Entity::XmlProfile->new(dbh => $dbh,
                                                            get_criteria => {profile_id => $profile_id});
      if (!$xml_profile) {
           return {
            completed_successfully => 0,
            messages => [ 'No profile found' ],
           };
      }
      $policy = $xml_profile->policy;
  }
                  
  local $" = ',';

  my $flg_follow_schema_links;
  if (exists $params{flg_follow_schema_links}) {
    $flg_follow_schema_links = $params{flg_follow_schema_links};
  }
  else {
    $flg_follow_schema_links = $xml_profile->{flg_follow_schema_links};
  }

  my $schema_document;
  my %schema_composition_mappings;

  my $compiled_schema_filename = "$COMPILED_SCHEMA_BASE_DIR/xml_profile_${profile_id}";

  my $profile_validation_files = $xml_profile->{validation_files} || [];

  my $validation_file_count = scalar @$profile_validation_files;

  if ($validation_file_count == 0) {
    if ($profile_id) {
        $dbh->do(<<EndQuery, undef, $profile_id);
DELETE 
  FROM PLC.PL_XML_PROFILE_SOAP_METHODS
 WHERE profile_id = ?
EndQuery
    }

    return {
            completed_successfully => 1,
            messages => [ 'No validation files for this profile' ],
           };
  }

  for my $profile_validation_file (@$profile_validation_files) {
      if (!$profile_validation_file->{xml_validation_file}->{file_contents}) {
          $profile_validation_file->{xml_validation_file} = F5::ASMConfig::Entity::XmlValidationFile->new(dbh => $dbh,
                                                                                                          get_criteria => $profile_validation_file->{xml_validation_file});
      }
  }

  my $latest_verification_file_time = List::Util::max( map { $_->{xml_validation_file}->{last_updated} } @$profile_validation_files );

  if (!$params{force_regenerate} && -e $compiled_schema_filename && stat($compiled_schema_filename)->ctime > $latest_verification_file_time) {
    if ($params{save_update}) {
      add_objects_from_wsdl(dbh => $dbh,
                            profile_id => $profile_id,
                            xml_profile => $xml_profile,
                            asm_config_handler => $asm_config_handler);
    }

    return {
            completed_successfully => 1,
            messages => [ 'All schemas updated successfully.' ],
            compiled_schema_filename => $compiled_schema_filename,
           };
  }

  F5::ContentProfile::clean_old_files(dir => $COMPILED_SCHEMA_BASE_DIR,
                  max_size => $MAX_TOTAL_FILESIZE_COMPILED_SCHEMAS);

  my $temp_schema_dir = "$TEMP_SCHEMA_BASE_DIR/xml_schema_$$";
  mkdir $temp_schema_dir;

  my $file_count = 0;
  my $primary_validation_file;
  for my $profile_validation_file (@$profile_validation_files) {
    my $validation_file = $profile_validation_file->{xml_validation_file};
    $file_count++;
    $validation_file->{local_filename} = "$temp_schema_dir/$file_count.xsd";
    open(my $outf, '>', $validation_file->{local_filename});
    print $outf $validation_file->{file_contents};
    close $outf;

    $validation_file->{is_wsdl} = F5::XML::WSDL::Convert::is_wsdl(xml_contents => $validation_file->{file_contents},
                                                                  flg_follow_schema_links => 0);

    # Save memory by releasing the file contents before storing the row
    delete $validation_file->{file_contents};
    if ($profile_validation_file->{flg_is_primary_enforced}) {
      $primary_validation_file = $validation_file;
    }
    else {
      $schema_composition_mappings{$profile_validation_file->{import_url}} = $validation_file->{local_filename};
    }
  }

  my $wsdl_attribs;
  if ($primary_validation_file->{is_wsdl}) {
    eval {
      $wsdl_attribs = F5::XML::WSDL::Convert::convert_wsdl(wsdl_filename => $primary_validation_file->{local_filename},
                                                           schema_composition_mappings => \%schema_composition_mappings,
                                                           schema_dir => $temp_schema_dir,
                                                           flg_follow_schema_links => $flg_follow_schema_links,
                                                          );
    };
    if ($@) {
      my $error_msg = $@;
      $error_msg =~ s|\s+ at \s+ /\S+ \s+ line \s+ \d .*||sx;
      $error_msg =~ s/\n//g;
      return {
              completed_successfully => 0,
              messages => [ "Malformed WSDL: $error_msg" ],
              delete_file => 1,
             };
    }

    unless ($wsdl_attribs->{completed_successfully}) {
      return {
              completed_successfully => 0,
              missing_files => $wsdl_attribs->{missing_files},
              messages => $wsdl_attribs->{error_msg},
             };
    }

    $primary_validation_file->{local_filename} = $wsdl_attribs->{schema_document};
    %schema_composition_mappings = (%schema_composition_mappings,
                                    %{$wsdl_attribs->{schema_composition_mappings}});

    if (!$primary_validation_file->{file_type} || ($primary_validation_file->{file_type} != F5::ASMConfig::XmlValidationFileType::XML_VALIDATION_FILE_TYPE_WSDL())) {
      $primary_validation_file->{file_type} = F5::ASMConfig::XmlValidationFileType::XML_VALIDATION_FILE_TYPE_WSDL();
      $dbh->do("UPDATE PLC.PL_XML_VALIDATION_FILES SET file_type = ? WHERE xml_validation_file_id = ?",
               undef, F5::ASMConfig::XmlValidationFileType::XML_VALIDATION_FILE_TYPE_WSDL(), $primary_validation_file->{xml_validation_file_id});
    }

    update_soap_methods(methods => $wsdl_attribs->{methods},
                        dbh => $dbh,
                        xml_profile => $xml_profile,
                        store_obj => $store_obj,
                        profile_id => $profile_id);

    # Disallow DTD and Processing Instructions
    $store_obj->{defense_attributes} = [ F5::ASMConfig::XmlDefense->new(def_attrib_value => 0,
                                                                        def_attrib_id => $XML_DEFENSE_ATTRIBUTE_ALLOW_DTDS,),
                                         F5::ASMConfig::XmlDefense->new(def_attrib_value => 0,
                                                                        def_attrib_id => $XML_DEFENSE_ATTRIBUTE_ALLOW_PROCESSING_INSTRUCTIONS,)];
  }

  save_wsdl_objects(wsdl_objects => $wsdl_attribs->{soap_address},
                    dbh => $dbh,
                    xml_profile => $xml_profile,
                    profile_id => $profile_id);

  my $proxy_params = F5::Utils::Net::get_proxy_settings();
  local $ENV{HTTP_PROXY} = $proxy_params->{proxy_uri}; 
  local $ENV{HTTP_PROXY_USERNAME} = $proxy_params->{proxy_username};
  local $ENV{HTTP_PROXY_PASSWORD} = $proxy_params->{proxy_password};
  local $ENV{HTTPS_PROXY} = $proxy_params->{proxy_uri}; 
  local $ENV{HTTPS_PROXY_USERNAME} = $proxy_params->{proxy_username};
  local $ENV{HTTPS_PROXY_PASSWORD} = $proxy_params->{proxy_password};
  $LWP::Simple::ua->env_proxy();

  my $temp_blob_filename = "$temp_schema_dir/processed_schema";
  open(my $blob_fh, '>', $temp_blob_filename);
  my $xsp = F5::XML::Schema::Processor->new(
                                            allow_external_references => $flg_follow_schema_links,
                                            schema_document => $primary_validation_file->{local_filename},
                                            schema_composition_mappings => \%schema_composition_mappings,
                                            blob_fh => $blob_fh,
                                           );

  eval {
    $xsp->xparse;
    $xsp->print_binary;

    close $blob_fh;

    # Move blob to permanent location
    move($temp_blob_filename, $compiled_schema_filename);
    chmod(0664, $compiled_schema_filename);
  };

  rmtree($temp_schema_dir);

  if ($@) {
    my $error_msg = $@;
    $error_msg =~ s/\s+ at \s+ \/\S+ \s+ line \s+ \d .*//sx;
    $error_msg =~ s/Missing base argument/Missing base configuration file/g;

    return {
            completed_successfully => 0,
            messages => [ "An error has occurred while processing the schema or WSDL document. ($error_msg)" ],
            delete_file => 1,
           };
  }

  if (not $xsp->OK) {
    my $errors = [ $xsp->errors ];

    if ( $errors->[0] =~ /===>Exception:/) {
        my $error_msg = $errors->[0];
        $error_msg =~ s/===>Exception://sx;
        return {
                 completed_successfully => 0,
                 messages => [ "An error has occurred while processing the schema or WSDL document. $error_msg" ],
                 delete_file => 1,
               };
    }
    else{
        return {
                 completed_successfully => 0,
                 missing_files => [ $xsp->missing_files ],
                 messages => $errors,
               };
    }
  }

  if ($params{save_update} && $profile_id) {
    $xml_profile->store_update(store_obj => $store_obj,
                               suppress_events => 1,);
    add_objects_from_wsdl(dbh => $dbh,
                          profile_id => $profile_id,
                          asm_config_handler => $asm_config_handler );
  }
  else {
      $xml_profile->{soap_methods} = $store_obj->{soap_methods};
      $xml_profile->{defense_attributes} = $store_obj->{defense_attributes};
  }

  return {
          completed_successfully => 1,
          messages => [ 'All schemas updated successfully.' ],
          compiled_schema_filename => $compiled_schema_filename,
          xml_profile => $xml_profile,
         };

}

#----------------------------------------------------------------------

sub update_soap_methods {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $profile_id = $params{profile_id};
  my $xml_profile = $params{xml_profile};
  my $policy_id = $xml_profile->{policy}->{policy_id};
  my $store_obj = $params{store_obj};
  local $" = ',';

  my @methods;
  for my $method (values %{$params{methods}}) {
      push (@methods, F5::ASMConfig::SoapMethod->new({soap_method_name => $method->{method_name},
                                                      allowed_soap_action => $method->{soap_action} || '',
                                                      flg_allowed => 1,
                                                      soap_method_namespace => $method->{namespace}}));
  }

  my $existing_methods = $xml_profile->hashify_nested_struct_list(nested_def => $xml_profile->fields->{soap_methods},
                                                                  nested_recs => $xml_profile->{soap_methods});
  my $new_methods      = $xml_profile->hashify_nested_struct_list(nested_def => $xml_profile->fields->{soap_methods},
                                                                  nested_recs => \@methods);

  my @updated_methods;
  for my $key (sort keys %$new_methods) {
      my $method = $new_methods->{$key};
      if ($existing_methods->{$key}) {
          $method->{flg_allowed} = $existing_methods->{$key}->{flg_allowed};
      }
      push (@updated_methods, $method);
  }

  $store_obj->{soap_methods} = \@updated_methods;
}

#----------------------------------------------------------------------
sub create_xpath_files{
  my(%params) = @_;
  my $dbh = $params{dbh};
  my $profile_id = $params{profile_id};

  F5::ContentProfile::clean_old_files(dir => $REQUEST_XPATH_BASE_DIR,
                  max_size => $MAX_TOTAL_FILESIZE_REQUEST_XPATH);

  F5::ContentProfile::clean_old_files(dir => $RESPONSE_XPATH_BASE_DIR,
                  max_size => $MAX_TOTAL_FILESIZE_RESPONSE_XPATH);


  my $request_xpath_binary = "$REQUEST_XPATH_BASE_DIR/xml_profile_${profile_id}";
  my $response_xpath_binary = "$RESPONSE_XPATH_BASE_DIR/xml_profile_${profile_id}";


   @request_elements_list = ();
   @response_elements_list = ();

   %prefix_namespace_mapping = ();

  my $prefix_namespace_list = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} } , $profile_id);
SELECT user_prefix,user_namespace
  FROM PLC.PL_XML_SOAP_PREFIX_NAMESPACE_MAPPING
 WHERE profile_id = ?
 UNION
   SELECT user_prefix,user_namespace FROM PLC.PL_XML_SOAP_PREDEFINED_PREFIX_NAMESPACE_MAPPING
EndQuery

  for my $prefix_namespace(@$prefix_namespace_list){
     $prefix_namespace_mapping{$prefix_namespace->{user_prefix}} = $prefix_namespace->{user_namespace};
  }


 my $soap_element_list =  $dbh->selectall_arrayref(<<EndQuery);
SELECT xpath_name,ws_action
  FROM PLC.PL_XML_SOAP_PREDEFINED_SPECIFIC_ELEMENTS
EndQuery

 my %element_lists = ();

 for my $element(@$soap_element_list){
  push @{$element_lists{$element->[1]}}, $element->[0];
 }

  @predefined_decrypt_elements = @{$element_lists{decrypt}};
  @predefined_encrypt_elements = @{$element_lists{encrypt}};


  my $xpath_elements = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} } , $profile_id);
SELECT xpath_name,apply_to,encrypted_part
  FROM PLC.PL_XML_SOAP_SPECIFIC_ELEMENTS
 WHERE profile_id = ?
EndQuery

  for my $xpath_element(@$xpath_elements){
     if ($xpath_element->{apply_to} eq 'Request'){
        push(@request_elements_list, $xpath_element->{xpath_name});
     }else{
        push(@response_elements_list,{ xpath => $xpath_element->{xpath_name},encrypted_part => $xpath_element->{encrypted_part} });
     }

  }

 add_flags_xpath( dbh => $dbh,profile_id => $profile_id);

 my @request_queries = (@predefined_decrypt_elements,@request_elements_list);

 my $request_parser = F5::XML::Xpath::xpathprocessor->new(
    xpath_namespaces_hash     => \%prefix_namespace_mapping,
    xpath_queries_list        => \@request_queries,
  );

  my @response_elements = map $_->{xpath}, @response_elements_list;

  my @response_queries = (@predefined_encrypt_elements,@response_elements);

  my $response_parser = F5::XML::Xpath::xpathprocessor->new(
    xpath_namespaces_hash     => \%prefix_namespace_mapping,
    xpath_queries_list        => \@response_queries,
  );

  my $proxy_params = F5::Utils::Net::get_proxy_settings();
  local $ENV{HTTP_PROXY} = $proxy_params->{proxy_uri}; 
  local $ENV{HTTP_PROXY_USERNAME} = $proxy_params->{proxy_username};
  local $ENV{HTTP_PROXY_PASSWORD} = $proxy_params->{proxy_password};
  local $ENV{HTTPS_PROXY} = $proxy_params->{proxy_uri}; 
  local $ENV{HTTPS_PROXY_USERNAME} = $proxy_params->{proxy_username};
  local $ENV{HTTPS_PROXY_PASSWORD} = $proxy_params->{proxy_password};

  $request_parser->xparse;
  $response_parser->xparse;
  

  open(my $request_bin_fh, '>', $request_xpath_binary);
  open(my $response_bin_fh, '>', $response_xpath_binary);

  $request_parser->print_blob($request_xpath_binary);
  $response_parser->print_blob($response_xpath_binary);

  close $request_bin_fh;
  close $response_bin_fh;

  chmod(0664, $request_xpath_binary);
  chmod(0664, $response_xpath_binary);

  open(my $request_md5_fh, '<', $request_xpath_binary);
  my $request_md5_sum = Digest::MD5->new->addfile( $request_md5_fh )->hexdigest;
  close($request_md5_fh);

  $REQUEST_BINARY_FULL_PATH = $REQUEST_BINARY_FULL_PATH_BASE."__".$request_md5_sum;

 open(my $response_md5_fh, '<', $response_xpath_binary);
  my $response_md5_sum = Digest::MD5->new->addfile( $response_md5_fh )->hexdigest;
  close($response_md5_fh);

  $RESPONSE_BINARY_FULL_PATH = $RESPONSE_BINARY_FULL_PATH_BASE."__".$response_md5_sum;


  return {
          completed_successfully => 1,
          messages => [ 'xpath files created successfully.' ],
          request_xpath_filename => $request_xpath_binary,
          response_xpath_filename => $response_xpath_binary,
         };
}
#----------------------------------------------------------------------
sub add_flags_xpath{

  my(%params) = @_;
  my $dbh = $params{dbh};
  my $profile_id = $params{profile_id};

  my @select_fields = map { $_->{flag_name} } @flags_xpath_mapping;

  my $db_flag_values = $dbh->selectrow_hashref(<<EndQuery, { Slice => {} } , $profile_id);
SELECT @select_fields
  FROM PLC.PL_XML_SOAP_SECURITY
 WHERE profile_id = ?
EndQuery


  for my $flag (@flags_xpath_mapping) {
       if ($db_flag_values->{$flag->{flag_name}}){
          for my $xpath_name (@{$flag->{xpath_list}}){
            if ($flag->{apply_to} eq 'Request'){
                 push(@request_elements_list,$xpath_name);
            }
            else{
                 push(@response_elements_list,{ xpath => $xpath_name, encrypted_part => $flag->{encrypted_part}} );
            }
         }
      }
   }


}
#----------------------------------------------------------------------
sub create_defense_config {
  my(%params) = @_;
  my $dbh = $params{dbh};
  my $profile_id = $params{profile_id};

  F5::ContentProfile::clean_old_files(dir => $DEFENSE_CONFIG_BASE_DIR,
                  max_size => $MAX_TOTAL_FILESIZE_DEFENSE_PROFILES);

  my $defense_config_template = <<EndDoc;
<?xml version="1.0" encoding="utf-8"?>
<policy xmlns="http://www.f5.com/ASM/xmlparser.conf"/>
EndDoc

  my $defense_twig = F5::XML::Twig->new(pretty_print => 'indented');
  $defense_twig->parse($defense_config_template);

  my $defense_attribs_sth = $dbh->prepare(<<EndQuery);
  SELECT def_attrib_xpath, policy_def_attrib.def_attrib_value
    FROM PLC.XML_DEFENSE_ATTRIBUTES def_attrib,
         PLC.PL_XML_PROFILE_DEFENSE_ATTRIBUTES policy_def_attrib
   WHERE def_attrib.def_attrib_id = policy_def_attrib.def_attrib_id
     AND policy_def_attrib.profile_id = ? AND policy_def_attrib.flg_check_attrib = ?
ORDER BY def_attrib.def_attrib_id
EndQuery

  $defense_attribs_sth->execute($profile_id,1);
  while(my $def_attrib = $defense_attribs_sth->fetchrow_hashref) {
    F5::ContentProfile::add_node_with_value( $defense_twig,
                         $def_attrib->{def_attrib_xpath},
                         $def_attrib->{def_attrib_value});
  }

  add_sensitive_params_nodes( defense_twig => $defense_twig, 
                              profile_id => $profile_id,
                              dbh => $dbh);

 my $ws_security_flag = $dbh->selectrow_array(<<EndQuery,undef,$profile_id);
SELECT flg_is_wss_security
  FROM PLC.PL_CONTENT_PROFILES
 WHERE id  = ?
EndQuery

  if($ws_security_flag){
      add_ws_security_nodes(defense_twig => $defense_twig,
                            profile_id => $profile_id,
                            dbh => $dbh);
  }
  add_base64_enc_data_nodes( defense_twig => $defense_twig,
                             profile_id => $profile_id,
                             dbh => $dbh);
  my $defense_config_filename = "$DEFENSE_CONFIG_BASE_DIR/xml_profile_${profile_id}";

  open(my $defense_config_fh, '>', $defense_config_filename);
  $defense_twig->print($defense_config_fh);
  close $defense_config_fh;

  chmod(0664, $defense_config_filename);

  return {
          completed_successfully => 1,
          messages => [ 'Defense configuration created successfully.' ],
          defense_config_filename => $defense_config_filename,
         };


}
#---------------------------------------------------------------------
sub add_ws_security_nodes{
  my (%params) = @_ ;

  my $defense_twig = $params{defense_twig};
  my $profile_id = $params{profile_id};
  my $dbh = $params{dbh};


  my @response_elements_nodes;
  my @request_elements_nodes;
  my @verify_elements_nodes;  

  for my $element(@request_elements_list){
     push(@request_elements_nodes, XML::Twig::Elt->new('query' => $element) );
  }

  for my $element(@response_elements_list){
     push(@response_elements_nodes, XML::Twig::Elt->new('query' =>  {type => $element->{encrypted_part} }, $element->{xpath}) );
  }


  my $db_field_values = $dbh->selectrow_hashref(<<EndQuery, { Slice => {} } , $profile_id);
SELECT encryption_algorithm,
       key_transport_algorithm,
       signature_algorithm,
       canonicalization_algorithm,
       flg_add_timestamp,
       flg_enforce_timestamp,
       flg_enable_response_elements,
       flg_enable_request_elements,
       flg_enable_response_role_actor,
       flg_enable_request_role_actor,
       flg_verify_all,
       flg_decrypt_all,
       create_timestamp_interval,
       timestamp_maximal_interval,
       response_ws_action
  FROM PLC.PL_XML_SOAP_SECURITY
 WHERE profile_id = ?
EndQuery

 my $client_certificates = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} } , $profile_id);
SELECT key_name,key_file_path,flg_skip_validation,A.ssl_cert_id
  FROM PLC.SSL_CERTIFICATE_FILES_PATH as A,
      PLC.SSL_CERTIFICATES as B
 WHERE A.ssl_cert_id = B.ssl_cert_id
  AND B.ssl_cert_id IN (
                      SELECT ssl_cert_id
                        FROM PLC.PL_XML_PROFILE_CLIENT_CERTIFICATE
                       WHERE profile_id = ? )
EndQuery

 my $server_certificates = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} } , $profile_id);
SELECT key_name,key_file_path,flg_skip_validation,A.ssl_cert_id
  FROM PLC.SSL_CERTIFICATE_FILES_PATH as A,
      PLC.SSL_CERTIFICATES as B
 WHERE A.ssl_cert_id = B.ssl_cert_id
  AND B.ssl_cert_id IN (
                      SELECT ssl_cert_id
                        FROM PLC.PL_XML_PROFILE_SERVER_CERTIFICATE
                       WHERE profile_id = ? )
EndQuery


  my $request_queries_dat_file = $REQUEST_BINARY_FULL_PATH;
  my $response_queries_dat_file = $RESPONSE_BINARY_FULL_PATH;
  my @client_credentials_nodes;
  my @server_credentials_nodes;
  my @role_actor_request_nodes;
  my @role_actor_response_nodes;


  my $role_actor_values = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} } , $profile_id);
SELECT role_actor, apply_to
  FROM PLC.PL_XML_ROLE_ACTOR
 WHERE profile_id = ?
EndQuery


  for $role_actor_values(@$role_actor_values){
     if($role_actor_values->{apply_to} eq 'Request' &&  $db_field_values->{flg_enable_request_role_actor} ){
       push(@role_actor_request_nodes, XML::Twig::Elt->new('uri' => $role_actor_values->{role_actor}) );
     }
     elsif($role_actor_values->{apply_to} eq 'Response' && $db_field_values->{flg_enable_response_role_actor} ){
       push(@role_actor_response_nodes, XML::Twig::Elt->new('uri' => $role_actor_values->{role_actor}) );
     }

  }

  for my $client_credential(@$client_certificates){

    if (! -e $client_credential->{key_file_path}){
       F5::Certificates::write_pem_file(ssl_cert_id => $client_credential->{ssl_cert_id});
    }
 
    push(@client_credentials_nodes, XML::Twig::Elt->new(
                                     'client_data',
                                     XML::Twig::Elt->new('key_name' => $client_credential->{key_name}),
                                     XML::Twig::Elt->new('key_file' => $client_credential->{key_file_path}),
                                     XML::Twig::Elt->new('skip_cert_check' => $client_credential->{flg_skip_validation}),
                                    )
         );

  }

  for my $server_credential(@$server_certificates){

    if (! -e $server_credential->{key_file_path}){
       F5::Certificates::write_pem_file(ssl_cert_id => $server_credential->{ssl_cert_id});
    }


    push(@server_credentials_nodes, XML::Twig::Elt->new(
                                     'server_data',
                                     XML::Twig::Elt->new('key_name' => $server_credential->{key_name}),
                                     XML::Twig::Elt->new('key_file' => $server_credential->{key_file_path}),
                                     XML::Twig::Elt->new('skip_cert_check' => $server_credential->{flg_skip_validation}),
                                    )
        );

  }

  my $num_clients = scalar @client_credentials_nodes;
  my $num_servers=  scalar @server_credentials_nodes;


  if (!$db_field_values->{flg_enable_response_elements} || !( $num_servers || $num_clients) ){
   @response_elements_nodes = ();  
  }

  if (!$db_field_values->{flg_enable_request_elements} || !$num_clients){
   @request_elements_nodes = ();
  }

  
  my @wss_nodes;
  if($db_field_values->{response_ws_action} eq 'encrypt_before_sign'){
    $db_field_values->{response_ws_action} = 'encrypt_then_sign';
  }

   if($db_field_values->{response_ws_action} eq 'sign_before_encrypt'){
    $db_field_values->{response_ws_action} = 'sign_then_encrypt';
  }
  
  push(@wss_nodes, XML::Twig::Elt->new('encrypt_and_sign_operation' => $db_field_values->{response_ws_action}) );


  if ($num_clients == 0) {
    $db_field_values->{flg_verify_all} = 0;
  }

  push(@wss_nodes, XML::Twig::Elt->new('request_functionality',
                                                           { decrypt => $db_field_values->{flg_decrypt_all}, 
                                                             verify => $db_field_values->{flg_verify_all} }));
  if($db_field_values->{flg_add_timestamp} && $num_servers){
     push(@wss_nodes,XML::Twig::Elt->new('timestamp',
                                                           { enforce => $db_field_values->{flg_enforce_timestamp}},
                                                           XML::Twig::Elt->new('sign_timestamp' => $db_field_values->{flg_add_timestamp}),
                                                           XML::Twig::Elt->new('place_in_security_header' => $db_field_values->{flg_add_timestamp}),
                                                           XML::Twig::Elt->new('create_timestamp_interval' => $db_field_values->{create_timestamp_interval}),
                                                           XML::Twig::Elt->new('timestamp_maximal_interval' => $db_field_values->{timestamp_maximal_interval})));
  }
  else{
     push(@wss_nodes,XML::Twig::Elt->new('timestamp',
                                                           { enforce => $db_field_values->{flg_enforce_timestamp}},
                                                           XML::Twig::Elt->new('timestamp_maximal_interval' => $db_field_values->{timestamp_maximal_interval})));
  }
 
  my $wss_node = XML::Twig::Elt->new(
                                     'wss',
                                      @wss_nodes,
  
                                     XML::Twig::Elt->new('request_queries_dat_file' => $request_queries_dat_file),
                                     XML::Twig::Elt->new('response_queries_dat_file' => $response_queries_dat_file),

                                     XML::Twig::Elt->new('specific_elements',
                                                           XML::Twig::Elt->new('encrypt_and_sign_specific_elements',  
                                                                   { items =>  scalar @response_elements_nodes },
                                                                   @response_elements_nodes ),

                                                           XML::Twig::Elt->new('verify_specific_elements',
                                                                   { items => scalar @request_elements_nodes },
                                                                   @request_elements_nodes )),
  

                                     XML::Twig::Elt->new('encryption_algorithm' => $db_field_values->{encryption_algorithm}),
                                     XML::Twig::Elt->new('key_transport_algorithm' => $db_field_values->{key_transport_algorithm}),
                                     XML::Twig::Elt->new('signature_algorithm' => $db_field_values->{signature_algorithm}),
                                     XML::Twig::Elt->new('canonicalization_algorithm' => $db_field_values->{canonicalization_algorithm}),


                                     XML::Twig::Elt->new('client_credentials' => 
                                                           { items => scalar @client_credentials_nodes },
                                                           @client_credentials_nodes),

                                     XML::Twig::Elt->new('server_credentials' => 
                                                           { items => scalar @server_credentials_nodes },
                                                           @server_credentials_nodes),

                                     XML::Twig::Elt->new('role_actor',
                                                    XML::Twig::Elt->new('req_role_actor_names' =>
                                                                        { items => scalar @role_actor_request_nodes },
                                                                        @role_actor_request_nodes),

                                                    XML::Twig::Elt->new('res_role_actor_names' =>
                                                                        { items => scalar @role_actor_response_nodes },
                                                                        @role_actor_response_nodes) ),

                                    );

   @response_elements_list = ();
   @request_elements_list = ();

  $wss_node->paste(last_child => F5::ContentProfile::add_or_find_node($defense_twig, dirname($WSS_XPATH)));

}
#---------------------------------------------------------------------
sub add_base64_enc_data_nodes{
  my (%params) = @_ ;

  my $defense_twig = $params{defense_twig};
  my $profile_id = $params{profile_id};
  my $dbh = $params{dbh};

  my @base64_enc_data_elements;


  my $base64_enc_data = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} } , $profile_id);
SELECT namespace, base64_param_name, base64_param_type
  FROM PLC.PL_CONTENT_BASE64_ENC_DATA
 WHERE profile_id = ?
EndQuery

 for my $entry (@$base64_enc_data){
      push(@base64_enc_data_elements, create_base64_enc_data(base64_enc_data => $entry,
                                                          profile_id => $profile_id,
                                                          dbh => $dbh))
 }
 return unless @base64_enc_data_elements;

 my $icap = XML::Twig::Elt->new('icap',
                                  { items => scalar @base64_enc_data_elements },
                                  @base64_enc_data_elements);

 $icap->paste(last_child => F5::ContentProfile::add_or_find_node($defense_twig, dirname($ICAP_XPATH)));
}
#---------------------------------------------------------------------
sub add_sensitive_params_nodes{
  my (%params) = @_ ;

  my $defense_twig = $params{defense_twig};
  my $profile_id = $params{profile_id};
  my $dbh = $params{dbh};

  my @sensitve_data_elements;


  my $sensitive_params = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} } , $profile_id);
SELECT namespace, sensitive_param_name, sensitive_param_type
  FROM PLC.PL_CONTENT_SENSITIVE_PARAMS
 WHERE profile_id = ?
EndQuery

 for my $sense_param (@$sensitive_params){
      push(@sensitve_data_elements, create_sensitive_data(sense_param => $sense_param,
                                                          profile_id => $profile_id, 
                                                          dbh => $dbh))
  }

  return unless @sensitve_data_elements;

  my $sensitive = XML::Twig::Elt->new('sensitive',
                                   { items => scalar @sensitve_data_elements },
                                   @sensitve_data_elements);

  $sensitive->paste(last_child => F5::ContentProfile::add_or_find_node($defense_twig, dirname($SENSITIVE_XML_PARAMS_XPATH)));

}
#---------------------------------------------------------------------
sub create_sensitive_data {

  my (%params) = @_ ;

  my $sense_param = $params{sense_param};
  my $profile_id = $params{profile_id};
  my $dbh = $params{dbh};
  
  my $type =  ($sense_param->{sensitive_param_type}  eq 'Element') ? 1 : 2;
  my $namespace = F5::ContentProfile::convert_to_utf8(string => $sense_param->{namespace}, 
                                  profile_id => $profile_id, 
                                  dbh => $dbh);

  my $name = F5::ContentProfile::convert_to_utf8(string => $sense_param->{sensitive_param_name}, 
                             profile_id => $profile_id,  
                             dbh => $dbh);

  my $new_node = XML::Twig::Elt->new(
                                     'data',
                                     XML::Twig::Elt->new('namespace' => $namespace),
                                     XML::Twig::Elt->new('name' => $name),
                                     XML::Twig::Elt->new('type' => $type),
                                    );
}
#----------------------------------------------------------------------
sub create_base64_enc_data{
  my (%params) = @_ ;

  my $base64_enc_data = $params{base64_enc_data};
  my $profile_id = $params{profile_id};
  my $dbh = $params{dbh};

  my $type =  ($base64_enc_data->{base64_param_type}  eq 'Element') ? 1 : 2;
  my $namespace = F5::ContentProfile::convert_to_utf8(string => $base64_enc_data->{namespace},
                                  profile_id => $profile_id,
                                  dbh => $dbh);

  my $name = F5::ContentProfile::convert_to_utf8(string => $base64_enc_data->{base64_param_name},
                             profile_id => $profile_id,
                             dbh => $dbh);

  my $new_node = XML::Twig::Elt->new(
                                     'data',
                                     XML::Twig::Elt->new('namespace' => $namespace),
                                     XML::Twig::Elt->new('name' => $name),
                                     XML::Twig::Elt->new('type' => $type),
                                    );
}
#----------------------------------------------------------------------
sub set_xml_defense_attribute {
  my(%params) = @_;

  my $dbh = $params{dbh};
  $dbh->do(<<EndQuery, undef, $params{def_attrib_value}, $params{def_attrib_id}, $params{profile_id});
UPDATE PLC.PL_XML_PROFILE_DEFENSE_ATTRIBUTES
   SET def_attrib_value = ?
 WHERE def_attrib_id = ?
   AND profile_id = ?
EndQuery
}

#----------------------------------------------------------------------

sub save_wsdl_objects {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $profile_id = $params{profile_id};
  return unless $profile_id;
  my($policy_id) = $dbh->selectrow_array('SELECT policy_id FROM PLC.PL_CONTENT_PROFILES WHERE id = ?', undef, $profile_id);

  local $" = ',';

  $dbh->do(<<EndQuery, undef, $policy_id,$profile_id);
DELETE FROM PLC.PL_XML_PROFILE_OBJECTS
 WHERE policy_id = ?
  AND profile_id = ?
EndQuery

  for my $wsdl_object (@{$params{wsdl_objects}}) {
    my $rec = {};
    $rec->{object_uri} = $wsdl_object;
    $rec->{policy_id} = $policy_id;
    $rec->{profile_id} = $profile_id;
    my @fields = keys %$rec;
    my @placeholders = ('?') x scalar @fields;

    $dbh->do(<<EndQuery, undef, @{$rec}{@fields});
INSERT INTO PLC.PL_XML_PROFILE_OBJECTS
  (@fields) VALUES (@placeholders)
EndQuery
  }
}

#----------------------------------------------------------------------

sub add_objects_from_wsdl {
  my(%params) = @_;

  my $profile_id = $params{profile_id};
  my $dbh = $params{dbh};
  my $asm_config_handler =  $params{asm_config_handler};
  if (! defined $asm_config_handler) {
      eval {
          $asm_config_handler = F5::ASMConfig::EasyClient->new(username => 'xml_profile');
      };

      if (! defined $asm_config_handler) {
          die "Could not open connection\n";
      }
  }

  my $xml_prof = F5::ASMConfig::XmlProfile->new();
  $xml_prof->{profile_id} = $profile_id;
  my $call_flags = F5::ASMConfig::CallFlags->new();
  my $session = $asm_config_handler->{session};

  my $policy_id = $dbh->selectrow_array(<<EndQuery, undef, $profile_id);
SELECT policy_id
  FROM PLC.PL_CONTENT_PROFILES
 WHERE id = ?
EndQuery

  my $policy = F5::ASMConfig::Policy->new({policy_id => $policy_id});
  my $xml_profile;

  eval{
     $xml_profile = $asm_config_handler->get_xml_profile($session,$call_flags,$policy,$xml_prof);
   };

  if ($@ || (! $xml_profile) ) {
    my $error = defined $@->{error_message} ? $@->{error_message} : '';
    die "failed to load xml profile. ".$error;
  }

  my $objects = $dbh->selectcol_arrayref(<<EndQuery, undef, $profile_id);
SELECT object_uri
  FROM PLC.PL_XML_PROFILE_OBJECTS
 WHERE profile_id = ?
EndQuery


  for my $object (@$objects) {
    my($app_id, $domain, $uri, $query) = F5::ParseUri::parse_uri($object);
    if (! $domain && (! defined($uri) || $uri eq '')) {
        # Invalid URI - skip
        next;
    }

    if (defined($domain) && !defined($uri)) {
        # Put in a default uri if the specified url has just a domain
        $uri = "/";
    }
    my $is_encrypted_flag = ($app_id eq APP_ID_HTTPS) ? 1 : 0;
    # check if the object is inside the policy
    my $obj_prop = F5::ASMConfig::URL->new({ app_id => $is_encrypted_flag, 
                                             url_name => $uri,
                                             is_wildcard => 0});
    my $url;

    my $search_url = $asm_config_handler->search_url($session,$call_flags,$policy,$obj_prop);


    if ( $search_url->{total_record_count} == 0) {
       # add a new url
          $obj_prop->{user_config_level} = 'advanced';
          $url = $asm_config_handler->add_url($session,$call_flags,$policy,$obj_prop);

    }else{
         $url = $search_url->{element_list}->[0]->{url};
    }

   # attach the object to the xml profile
    my $content_profile = F5::ASMConfig::ContentProfile->new({profile_id => $profile_id});
    my $url_content_profile = F5::ASMConfig::UrlContentProfile->new({ url => $url,
                                                                     content_profile => $content_profile,
                                                                     header_order => 0,
                                                                     header_name => '*',
                                                                     header_value => '*',
                                                                     enforcement_type => F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_XML} );

    $call_flags->{add_on_updating_nonexistent_record} = 1;
    my $object_xml;
    eval{
       $object_xml = $asm_config_handler->update_url_content_profile($session,$call_flags,$policy,$url_content_profile);
    };

    if ( $@  || ! $object_xml) {
        my $error = defined $@->{error_message} ? $@->{error_message} : '';
        die "failed to create new URL/xml connection. ".$error ;
    }

    # add filetype
    my $filetype_name = get_extension($url->{url_name});

    my $filetype_prop = F5::ASMConfig::Filetype->new({ filetype_name => $filetype_name,
                                                       is_wildcard => 0,
                                                       flg_check_url_len => 1,
                                                       url_len => length($url->{url_name}),
                                                       flg_check_req_len => 0,
                                                       req_len => 0,
                                                       flg_check_qs_len => 0,
                                                       qs_len => 0,
                                                       flg_check_pd_len => 0,
                                                       pd_len => 0,
                                                      });

    $call_flags->{ignored_errors} = [
                                     F5::ASMConfig::IgnoredError->new(error_code => F5::ASMConfig::ErrorCode::DUPLICATE_RECORD),
                                     F5::ASMConfig::IgnoredError->new(error_code => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT),
                                    ];
    $asm_config_handler->add_filetype($session,$call_flags,$policy,$filetype_prop);
  }
}
#----------------------------------------------------------------------
sub clean_unreferenced_profile_files {
  my(%params) = @_;

  for my $dir (($COMPILED_SCHEMA_BASE_DIR, $DEFENSE_CONFIG_BASE_DIR)) {
    F5::DbUtils::delete_unreferenced_files(dbh => $params{dbh},
                                       logger => $params{logger},
                                       db_file_refs => [
                                                        { table_name => 'PLC.PL_CONTENT_PROFILES',
                                                          field => 'id' },
                                                       ],
                                       db_value_to_filename_conversion_func => sub {
                                         return "$dir/xml_profile_" . shift;
                                       },
                                       dir => $dir);
  }
}
#----------------------------------------------------------------------
sub get_extension {
 my ($name) = @_;
 my $ext =  ($name =~ /\.([^.\/]{1,$MAX_EXTENSION_LEN})$/) ? $1 : 'no_ext';
 return $ext;
    
}
#-----------------------------------------------------------------------
1;
