package F5::ASMConfig::Handler::ImportVulnerabilitiesTask;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::TaskBase/;
use F5::ASMConfig::Entity::ImportVulnerabilitiesTask;
use F5::Utils::Rest;
use MIME::Base64;

our @handled_functions = (
                            {
                             function_name => 'add_import_vulnerabilities_task',
                             criteria_index => 3,
                             store_index    => 4,
                            },
                            {
                              function_name => 'update_import_vulnerabilities_task',
                              criteria_index => 3,
                              store_index => 4
                            },
                            { 
                             function_name => 'execute_import_vulnerabilities_task',
                            },
                            {
                             function_name => 'delete_import_vulnerabilities_task',
                            },
                            {
                             function_name => 'delete_many_import_vulnerabilities_tasks',
                            },
                         );

our $entity_class = 'F5::ASMConfig::Entity::ImportVulnerabilitiesTask';

#----------------------------------------------------------------------

sub add_import_vulnerabilities_task {
  my($self, @rpc_params) = @_;
  return F5::ASMConfig::Handler::TaskBase::add_task(self => $self,
                                                    rpc_params => \@rpc_params,
                                                    entity_class => $entity_class,
                                                    task_type => 'import_vulnerabilities',);
}

#----------------------------------------------------------------------

sub update_import_vulnerabilities_task {
    my ($self, @rpc_params) = @_;   
    return F5::ASMConfig::Handler::TaskBase::update_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                         task_type => 'import_vulnerabilities',);
}

#----------------------------------------------------------------------

sub import_vulnerabilities {
    my (%params) = @_;
    my $self       = $params{self};
    my $session    = $params{session};
    my $call_flags = $params{call_flags};
    my $task       = $params{task};


    my $policy = $task->{target_policy};
    if (!$policy) {
        return $self->handle_error(action => 'execute',
                                   error_thrown => 'Policy not found',
                                   error_code => F5::ASMConfig::ErrorCode::MISSING_POLICY(),
                                   call_flags => $call_flags,
                                   entity_class => $entity_class,
                                   thrift_obj => $task,
                                  );
    }


    my $filename = $task->{filename};
    my $contents = $task->{file_contents};
    if ($filename) {
        $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{filename} );
    }
    if ($task->{file_contents}) {
        if (!$filename) {
            $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{rest_uuid} );
        }
        open (my $FH, '>', $filename) || $self->handle_error(action => 'execute',
                                                             entity_class => $entity_class,
                                                             thrift_obj  => $task,
                                                             error_code  => F5::ASMConfig::ErrorCode::PERMISSION_DENIED(),
                                                             msg         => "Could not open $filename for writing.",
                                                             error_level  => "info", );
       if ($task->{is_base64}) {
           $task->{file_contents} = MIME::Base64::decode_base64($task->{file_contents});
           $task->{is_base64} = 0;
       }
       print $FH $task->{file_contents};
       close ($FH);
    }

    my $rv;
    $policy = F5::ASMConfig::Cache::get_cached_policy(policy => $policy, dbh => $self->{dbh});

    my $scanner_action_fields = F5::ASMConfig::ScannerActionsFields->new(scanner_vuln_data_xml => $task->{file_contents},
                                                                         xml_file_path => $filename,
                                                                         host_names => $task->{domain_names},
                                                                         import_all_host_names => $task->{import_all_domain_names},
                                                                         get_host_names_only => $task->{only_get_domain_names},
                                                                         copy_file => ($task->{filename} ? 1 : 0),
                                                                         scan_id => $task->{scan_id},
                                                                         subscription_id => $task->{subsciption_id});
                                                                              
                                                                         
#??    2: optional list<i32> scanner_vuln_ids,

    eval {
        $self->begin_work($self->{session});
        $rv = $self->update_scanner_vuln_data($session, $call_flags, $policy, $scanner_action_fields);
        $self->commit($self->{session});
    };

    if ($@) {
        $task->{message} = $@->{error_message};
        $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_FAILURE();
    }
    else {
        $task->{vulnerable_hosts} = $rv->{vuln_hosts};
        $task->{vulnerability_count} = $rv->{total_vuln_count};
        $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_COMPLETED();
    }

    if ($task->{file_contents} && !$task->{filename}) {
        unlink($filename);
    }
}


#----------------------------------------------------------------------
sub execute_import_vulnerabilities_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::execute_task(self => $self,
                                                          rpc_params => \@rpc_params,
                                                          entity_class => $entity_class,
                                                          task_type => 'import_vulnerabilities',
                                                          execute_func => sub { import_vulnerabilities(@_); },
                                                          );
}

#----------------------------------------------------------------------

sub delete_import_vulnerabilities_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                        );
}

#----------------------------------------------------------------------

sub delete_many_import_vulnerabilities_tasks {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_many_tasks(self => $self,
                                                               rpc_params => \@rpc_params,
                                                               entity_class => $entity_class,
                                                              );
}

#----------------------------------------------------------------------

1;
